(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: print.ml,v 1.3 2003/06/27 20:34:24 simeon Exp $ *)

(* Module Print
   Description:
     This module implements pretty-printing for the XQuery ASTs.
*)

open Namespace
open Error
open Occurrence
open Datatypes
open Xquery_ast
open Gmisc

open Format

(********************)
(* Global variables *)
(********************)

let result_flag = ref true
    (* true if printing results with namespaces prefixes
       false otherwise *)

(********************************)
(* Print common data structures *)
(********************************)

(* Base types *)

let string_of_basety bt =
  match bt with
  | ATString -> "xsd:string"
  | ATBoolean -> "xsd:boolean"
  | ATDecimal -> "xsd:decimal"
  | ATFloat -> "xsd:float"
  | ATDouble -> "xsd:double"
  | ATDuration -> "xsd:duration"
  | ATDateTime -> "xsd:dateTime"
  | ATTime -> "xsd:time"
  | ATDate -> "xsd:date"
  | ATGYearMonth -> "xsd:gYearMonth"
  | ATGYear -> "xsd:gYear"
  | ATGMonthDay -> "xsd:gMonthDay"
  | ATGDay -> "xsd:gDay"
  | ATGMonth -> "xsd:gMonth"
  | ATHexBinary -> "hexBinary"
  | ATBase64Binary -> "xsd:base64Binary"
  | ATAnyURI -> "xsd:anyURI"
  | ATQName -> "xsd:QName"
  | ATNOTATION -> "xsd:NOTATION"
  | ATInteger -> "xsd:integer"
  | ATInt -> "xsd:int"
  | ATYearMonthDuration -> "xdt:yearMonthDuration"
  | ATDayTimeDuration -> "xdt:dayTimeDuration"
  | ATUntyped -> "xdt:untypedAtomic"

let print_basety ff bt =
  fprintf ff "%s" (string_of_basety bt)

(* Base values *)

let string_of_baseval bv =
  match bv with
  | AtomicString s -> "\"" ^ s ^ "\""
  | AtomicBoolean b -> string_of_bool b
  | AtomicDecimal i -> Decimal._string_of_decimal i
  | AtomicFloat f -> string_of_float f
  | AtomicDouble d -> string_of_float d
  | AtomicDuration t ->  "\"" ^ t ^ "\""
  | AtomicDateTime t ->  "\"" ^ (DateTime.string_of_dateTime t) ^ "\""
  | AtomicTime t ->  "\"" ^ t ^ "\""
  | AtomicDate t ->  "\"" ^ t ^ "\""
  | AtomicGYearMonth t ->  "\"" ^ t ^ "\""
  | AtomicGYear t ->  "\"" ^ t ^ "\""
  | AtomicGMonthDay t ->  "\"" ^ t ^ "\""
  | AtomicGDay t ->  "\"" ^ t ^ "\""
  | AtomicGMonth t ->  "\"" ^ t ^ "\""
  | AtomicHexBinary t ->  "\"" ^ t ^ "\""
  | AtomicBase64Binary t ->  "\"" ^ t ^ "\""
  | AtomicAnyURI u -> AnyURI._string_of_uri u
  | AtomicQName q -> "\"" ^ (Sym.anon_string q) ^ "\""
  | AtomicNOTATION n -> "\"" ^ n ^ "\""
  | AtomicInteger i -> Decimal._string_of_integer i
  | AtomicInt i -> Decimal._string_of_int i
  | AtomicYearMonthDuration s -> "\"" ^ s ^ "\""
  | AtomicDayTimeDuration s -> "\"" ^ s ^ "\""
  | AtomicUntyped s -> "\"" ^ s ^ "\""

let print_baseval ff bv =
  fprintf ff "%s" (string_of_baseval bv)

(* QNames *)

let print_uqname ff qn =
  let simplified_s =
    let s = (Namespace.string_of_uqname qn) in
    match s with
    | "*:*" -> "*"
    | _ -> s
  in
  fprintf ff "%s" simplified_s

let print_uqname_escape ff qn =
  let simplified_s =
    let s = (Namespace.string_of_uqname qn) in
    match s with
    | "*:*" -> "*"
    | "item" -> ":item"
    | "atomic" -> ":atomic"
    | "numeric" -> ":numeric"
    | "document" -> ":document"
    | "text" -> ":text"
    | "some" -> ":some"
    | "every" -> ":every"
    | "for" -> ":for"
    | "let" -> ":let"
    | "in" -> ":in"
    | "satisfies" -> ":satisfies"
    | "return" -> ":return"
    | "returns" -> ":returns"
    | "then" -> ":then"
    | "end" -> ":end"
    | "else" -> ":else"
    | "to" -> ":to"
    | "where" -> ":where"
    | "intersect" -> ":intersect"
    | "case" -> ":case"
    | "stable" -> ":stable"
    | "ascending" -> ":ascending"
    | "descending" -> ":descending"
    | "default" -> ":default"
    | "as" -> ":as"
    | "element" -> ":element"
    | "attribute" -> ":attribute"
    | "type" -> ":type"
    | "node" -> ":node"
    | _ -> s
  in
  fprintf ff "%s" simplified_s

let print_rqname ff qn =
  let rqname_string_fun qn =
    if !result_flag then
      let (uri,ncname) = qn in
      Namespace.string_of_uqname (Namespace.get_uri_prefix uri, ncname)
    else
      Namespace.string_of_rqname qn
  in
  fprintf ff "%s" (rqname_string_fun qn)

(* Symbols *)

let print_rattr_symbol ff rsya =
  let rattr_string_fun =
    if !result_flag then
      Sym.rattr_prefix_string
    else
      Sym.rattr_string 
  in
  fprintf ff "%s" (rattr_string_fun rsya)

let print_relem_symbol ff rsye =
  let relem_string_fun =
    if !result_flag then
      Sym.relem_prefix_string
    else
      Sym.relem_string 
  in
  fprintf ff "%s" (relem_string_fun rsye)

let print_rgroup_symbol ff rsyg =
  let rgroup_string_fun =
    if !result_flag then
      Sym.rgroup_prefix_string
    else
      Sym.rgroup_string 
  in
  fprintf ff "%s" (rgroup_string_fun rsyg)

let print_attr_symbol ff sya =
  let attr_string_fun =
    if !result_flag then
      Sym.string_prefix_of_attr_symbol
    else
      Sym.string_of_attr_symbol
  in
  fprintf ff "%s" (attr_string_fun sya)

let print_elem_symbol ff sye =
  let elem_string_fun =
    if !result_flag then
      Sym.string_prefix_of_elem_symbol
    else
      Sym.string_of_elem_symbol
  in
  fprintf ff "%s" (elem_string_fun sye)

let print_group_symbol ff syg =
  let group_string_fun =
    if !result_flag then
      Sym.string_prefix_of_group_symbol
    else
      Sym.string_of_group_symbol
  in
  fprintf ff "%s" (group_string_fun syg)

(* operators *)

let string_of_bop bop =
  match bop with
  | BEIntersect ->
      "intersect"
  | BEUnion ->
      "union"
  | BEExcept ->
      "except"
  | BEBar ->
      "|"
  | BEAnd ->
      "and"
  | BEOr ->
      "or"
  | BEPrecedes ->
      "<<"
  | BEFollows ->
      ">>"
  | BEEq ->
      "eq"
  | BEEqual ->
      "="
  | BENEq -> 
      "ne"
  | BENEqual -> 
      "!="
  | BEIs ->
      "is"
  | BEIsNot ->
      "isnot"
  | BELtOp ->
      "lt"
  | BELt ->
      "<"
  | BEGtOp ->
      "gt"
  | BEGt ->
      ">"
  | BELte ->
      "le"
  | BELteq ->
      "<="
  | BEGteq ->
      ">="
  | BEGte ->
      "ge"
  | BEMult ->
      "*"
  | BEPlus ->
      "+"
  | BEMinus ->
      "-"
  | BEDiv ->
      "div"
  | BEIDiv ->
      "idiv"
  | BEMod ->
      "mod"

let print_bop ff bop =
  fprintf ff "%s" (string_of_bop bop)

let string_of_uop uop =
  match uop with
  | UEPlus ->
      "+"
  | UEMinus ->
      "-"

let print_uop ff uop =
  fprintf ff "%s" (string_of_uop uop)

(* axis *)

let string_of_axis a =
  match a with
  | Ancestor ->
      "ancestor"
  | Ancestor_or_self ->
      "ancestor-or-self"
  | Attribute ->
      "attribute"
  | Child ->
      "child"
  | Descendant ->
      "descendant"
  | Descendant_or_self ->
      "descendant-or-self"
  | Namespace ->
      "namespace"
  | Parent ->
      "parent"
  | Self ->
      "self"

let print_axis ff a =
  fprintf ff "%s" (string_of_axis a)

(* node test *)

let print_node_kind ff nk =
  match nk with
  | PIKind None ->
      fprintf ff "pi()"
  | PIKind (Some s) ->
      fprintf ff "pi(\"%s\")" s
  | CommentKind ->
      fprintf ff "comment()"
  | TextKind ->
      fprintf ff "text()"
  | AnyKind ->
      fprintf ff "node()"

let print_node_test ff nt =
  match nt with
  | PNameTest qn ->
      fprintf ff "%a" print_uqname_escape qn
  | PNodeKindTest nk ->
      fprintf ff "%a" print_node_kind nk

let print_cnode_test ff nt =
  match nt with
  | CPNameTest qn ->
      fprintf ff "%a" print_rqname qn
  | CPNodeKindTest nk ->
      fprintf ff "%a" print_node_kind nk

(*************************)
(* Print sort parameters *)
(*************************)


let print_sortkind ff sk =
  match sk with
  | Ascending -> fprintf ff "ascending"
  | Descending -> fprintf ff "descending"

let print_emptysortkind ff esk =
  match esk with
  | EmptyGreatest -> fprintf ff "empty greatest"
  | EmptyLeast -> fprintf ff "empty least"

let print_stablekind ff sk =
  match sk with
  | Stable -> fprintf ff "stable "
  | NonStable -> ()

(********************)
(* Print namespaces *)
(********************)

let print_namespaces ff () =
  let all_ns_bindings =
    let keep_binding x y z =
      match x with
      | NSWildcardPrefix ->
	  z
      | _ ->
	  if (Namespace.is_built_in_namespace (x,y))
	  then
	    z
	  else
	    (x,y) :: z
    in
    Hashtbl.fold keep_binding Namespace.prefix_uri_table []
  in
  let rec print_ns ff pfl=
    match pfl with
    | [] -> ()
    | (prefix,uri) :: [] ->
	begin
	  match prefix with
	  | NSDefaultElementPrefix ->
	      fprintf ff "default element namespace = %s@\n@\n" (string_of_uri uri)
	  | NSDefaultFunctionPrefix ->
	      fprintf ff "default function namespace = %s@\n@\n" (string_of_uri uri)
	  | NSWildcardPrefix ->
	      ()
	  | (NSPrefix nc) ->
	      fprintf ff "declare namespace %s = %s@\n@\n" nc (string_of_uri uri)
	end
    | (prefix,uri) :: pfl' ->
	begin
	  match prefix with
	  | NSDefaultElementPrefix ->
	      fprintf ff "default element namespace = %s@\n%a" (string_of_uri uri) print_ns pfl'
	  | NSDefaultFunctionPrefix ->
	      fprintf ff "default function namespace = %s@\n%a" (string_of_uri uri) print_ns pfl'
	  | NSWildcardPrefix ->
	      ()
	  | (NSPrefix nc) ->
	      fprintf ff "declare namespace %s = %s@\n%a" nc (string_of_uri uri) print_ns pfl'
	end
  in
  print_ns ff all_ns_bindings

(*******************)
(* Print new types *)
(*******************)

(* 
let rec print_nameclass ff nc =
  match nc with
  | NCNameTest l ->
      fprintf ff "%a" print_symbol l
  | NCNameClassRef l ->
      fprintf ff "@[nameclass %a@]" print_qname l
  | NCNeg nc' ->
      fprintf ff "@[-(%a)@]" print_nameclass nc'
  | NCPlus (nc1,nc2) ->
      fprintf ff "@[(%a)@ +@ (%a)@]" print_nameclass nc1 print_nameclass nc2
  | NCMinus  (nc1,nc2) ->
      fprintf ff "@[(%a)@ -@ (%a)@]" print_nameclass nc1 print_nameclass nc2
*)

let print_deriv ff d =
  match d with
  | None ->
      fprintf ff ""
  | Some (TRestriction tn) ->
      fprintf ff "@ restriction %a" print_uqname_escape tn
  | Some (TExtension tn) ->
      fprintf ff "@ extension %a" print_uqname tn

(* Note: 
     Here is, for each type constructor, the code used for precedence
     in the pretty-printer.

   4 - TAttrRef, TAttrLocal, TElemRef, TElemLocal, TGroupRef, TEmpty,
       TNone, TBase, TDocument, TText
   3 - TBound
   2 - TAmp
   1 - TComma
   0 - TBar

     Higher number indicates higher precedence.

   - Jerome and Phil

 *)

let rec print_model ff m =
  print_model_prec 0 ff m

and print_model_prec p ff m =
  match m.pmodel_desc with
  | TAttrRef l ->
      fprintf ff "attribute %a" print_attr_symbol l
  | TAttrLocal (l, tm) ->
      fprintf ff "@[<hv 2>attribute %a {@,%a@;<0 -2>}@]" print_attr_symbol l print_typed_model tm
  | TElemRef l ->
      fprintf ff "element %a" print_elem_symbol l
  | TElemLocal (l, d, tm) ->
      fprintf ff "@[<hv 2>element %a%a {@,%a@;<0 -2>}@]" print_elem_symbol l print_deriv d print_typed_model tm
  | TGroupRef l ->
      fprintf ff "type %a" print_group_symbol l
  | TBound (m',l,h) ->
      begin
	match (l,h) with
	| (UP_INT 0, UNBOUNDED) ->
	    fprintf ff "%a*" (print_model_prec 3) m'
	| (UP_INT 1, UNBOUNDED) ->
	    fprintf ff "%a+" (print_model_prec 3) m'
	| (UP_INT 0, UP_INT 1) ->
	    fprintf ff "%a?" (print_model_prec 3) m'
	| (UNBOUNDED, _) ->
	    fprintf ff "none"
	| _ ->
	    fprintf ff "%a (minoccurs %s) (maxoccurs %s)" (print_model_prec 3) m' (string_of_occurs l) (string_of_occurs h)
      end
  | TComma (m1,m2) ->
      if p > 1 then
        fprintf ff "@[<hv 1>(%a,@ %a)@]" (print_model_prec 1) m1 (print_model_prec 1) m2
      else
        fprintf ff "%a,@ %a" (print_model_prec 1) m1 (print_model_prec 1) m2
  | TEmpty ->
      fprintf ff "()"
  | TBar (m1,m2) -> 
      if p > 0 then
	fprintf ff "@[<hv 1>(%a |@ %a)@]" (print_model_prec 0) m1 (print_model_prec 0) m2
      else
	fprintf ff "%a |@ %a" (print_model_prec 0) m1 (print_model_prec 0) m2
  | TNone ->
      fprintf ff "none" 
  | TAmp (m1,m2) -> 
      if p > 2 then
	fprintf ff "@[<hv 1>(%a &@ %a)@]" (print_model_prec 2) m1 (print_model_prec 2) m2
      else
	fprintf ff "%a &@ %a" (print_model_prec 2) m1 (print_model_prec 2) m2
  | TBase bt ->
      fprintf ff "%a" print_basety bt
  | TDocument m ->
      fprintf ff "@[<hv 2>document {@,%a@;<0 -2>}@]" (print_model_prec 0) m
  | TText ->
      fprintf ff "text" 
  | TTuple rnm_list ->
      fprintf ff "@[<hov 2>[@,%a@;<0 -2>]@]" print_tuple_model rnm_list

and print_tuple_model ff rnm_list =
  match rnm_list with
  | [] -> ()
  | (rname,m1) :: [] ->
      fprintf ff "%a : %a" print_uqname rname print_model m1
  | (rname,m1) :: rnm_list' ->
      fprintf ff "%a:$%a,@ %a" print_uqname rname print_model m1 print_tuple_model rnm_list'

and print_typed_model ff tm =
  match tm.ptyped_model_desc with
(*| TTypeRef tn ->
      fprintf ff "@[type %a@]" print_qname tn
  | TNamedType (nc,d,m) ->
      fprintf ff "@[type %a@[%a@] { @[%a@] }" print_nameclass nc print_deriv d print_model m *)
  | TAnonymousType m ->
      fprintf ff "%a" print_model m

(* Print a model handle *)

let print_model_handle ff model_handle =
  match model_handle with
  | Some model ->
      print_model ff model
  | None ->
      fprintf ff "#None#"

(* Print optional model *)

let print_optmodel ff om =
  match om with
  | None ->
      ()
  | Some m ->
      print_model ff m

(* Print datatypes *)

let print_datatype_kind ff dtk =
  match dtk with
  | DTAttrRef None ->
      fprintf ff "attribute"
  | DTAttrRef (Some attr_symbol) ->
      fprintf ff "attribute %a" print_attr_symbol attr_symbol
  | DTElemRef None ->
      fprintf ff "element"
  | DTElemRef (Some elem_symbol) ->
      fprintf ff "element %a" print_elem_symbol elem_symbol
  | DTTypeRef group_symbol ->
      fprintf ff "type %a" print_group_symbol group_symbol
  | DTNode ->
      fprintf ff "node"
  | DTItem ->
      fprintf ff "item"
  | DTAtomic ->
      fprintf ff "atomic"
  | DTNumeric ->
      fprintf ff "numeric"
  | DTText ->
      fprintf ff "text"
  | DTEmpty ->
      fprintf ff "empty"
  | DTBase bt ->
      fprintf ff "%a" print_basety bt
  | DTElemOfType group_symbol ->
      fprintf ff "element of type %a" print_group_symbol group_symbol
  | DTElemOfBase bt ->
      fprintf ff "element of type %a" print_basety bt
  | DTDocument None ->
      fprintf ff "document" 
  | DTDocument (Some elem_symbol) ->
      fprintf ff "document %a" print_elem_symbol elem_symbol
  | DTDocumentOfType group_symbol ->
      fprintf ff "document of type %a" print_group_symbol group_symbol

let print_occurence ff occ =
  match occ with
  | None ->
      ()
  | Some (l,h) ->
      begin
	match (l,h) with
	| (UP_INT 0, UNBOUNDED) ->
	    fprintf ff "*"
	| (UP_INT 1, UNBOUNDED) ->
	    fprintf ff "+"
	| (UP_INT 0, UP_INT 1) ->
	    fprintf ff "?"
	| _ ->
	    raise (Query (Internal_Error ("Cannot have complex minOccurs and maxOccurs in a Datatype")))
      end

let print_datatype ff dt =
  match dt.pdatatype_desc with
  | (dtk,occ) ->
      fprintf ff "%a%a" print_datatype_kind dtk print_occurence occ

let print_optdatatype ff odt =
  match odt with
  | None -> ()
  | Some dt ->
      fprintf ff " as %a " print_datatype dt

(****************************)
(* Print XQuery expressions *)
(****************************)

(* Note:

     Here is, for each type expression, the code used for precedence
     in the pretty-printer.

  22 - ESubtype, EIntersects
  21 - EVar, EScalar, EApp
  20 - EPath        | with path_expr = PAxis,
       ERoot, ESelf, EParent
  19 - EPath        | with path_expr = PStepQualifiers
  18 - EPath        | with path_expr = PSlash, PSlashSlash
  17 - EValidate, ECast, ETreat, EElemFixed, EElemComputed,
       EAttrFixed, EAttrComputed, EDocument, EText, ETextComputed,
       EPI, EComment
  16 - EUnaryOp
  15 - EBinaryOp    | with binop = BEIntersect, BEExcept
  14 - EBinaryOp    | with binop = BEUnion, BEBar
  13 - EBinaryOp    | with binop = BEMult, BEDiv, BEIDiv, BEMod
  12 - EBinaryOp    | with binop = BEPlus, BEMinus
  11 - ERange
  10 - EBinaryOp    | with binop = BEPrecedes, BEFollows, BEEq, BENEq,
                                   BELtOp, BELte, BEGtOp, BEGte,
                                   BEEqual, BENEqual, BEIs, BEIsNot,
                                   BELt, BEGt, BELteq, BEGteq
  9  - EInstanceOf
  8  - EIf
  7  - ETypeswitch
  6  - ESome, EEvery
  5  - EFLWR
  4  - EUnordered
  3  - EBinaryOp    | with binop = BEAnd
  2  - EBinaryOp    | with binop = BEOr
  1  - ESort
  0  - EList

*)

(* Expression *)

let rec print_expr_body ff e =
  print_expr_prec 1 ff e; fprintf ff "\n@?"

and print_expr ff e =
  print_expr_prec 0 ff e

and print_expr_prec p ff e =
  match e.pexpr_desc with
  | ESort (e1,sc) ->
      if p > 1
      then
	fprintf ff "@[<hv 1>(@[<hv 0>%a@;<1 0>%a@])@]" (print_expr_prec 1) e1 print_sort_by_clause sc
      else
	fprintf ff "@[<hv 0>%a@;<1 0>%a@]" (print_expr_prec 1) e1 print_sort_by_clause sc
  | EBinaryOp (e1,op,e2) ->
      begin
	match op with
	| BEOr ->
	    if p > 2
	    then
	      fprintf ff "@[<hv 1>(@[<hov 1>%a@;<1 0>@[<hov 1>%a@;<1 0>%a@]@])@]" (print_expr_prec 2) e1 print_bop op (print_expr_prec 2) e2
	    else
	      fprintf ff "@[<hov 1>%a@;<1 0>@[<hov 1>%a@;<1 0>%a@]@]" (print_expr_prec 2) e1 print_bop op (print_expr_prec 2) e2

	| BEAnd ->
	    if p > 3
	    then
	      fprintf ff "@[<hv 1>(@[<hov 1>%a@;<1 0>@[<hov 1>%a@;<1 0>%a@]@])@]" (print_expr_prec 3) e1 print_bop op (print_expr_prec 3) e2
	    else
	      fprintf ff "@[<hov 1>%a@;<1 0>@[<hov 1>%a@;<1 0>%a@]@]" (print_expr_prec 3) e1 print_bop op (print_expr_prec 3) e2

	| BEPrecedes
	| BEFollows

	| BEEq
	| BENEq
	| BELtOp
	| BELte
	| BEGtOp
	| BEGte
	    
	| BEEqual
	| BENEqual
	| BEIs
	| BEIsNot
	| BELt
	| BEGt
	| BELteq
	| BEGteq ->
	    if p > 10
	    then
	      fprintf ff "@[<hv 1>(@[<hov 1>%a@;<1 0>@[<hov 1>%a@;<1 0>%a@]@])@]" (print_expr_prec 10) e1 print_bop op (print_expr_prec 10) e2
	    else
	      fprintf ff "@[<hov 1>%a@;<1 0>@[<hov 1>%a@;<1 0>%a@]@]" (print_expr_prec 10) e1 print_bop op (print_expr_prec 10) e2
	    
	| BEPlus
	| BEMinus ->
	    if p > 12
	    then
	      fprintf ff "@[<hv 1>(@[<hov 1>%a@;<1 0>@[<hov 1>%a@;<1 0>%a@]@])@]" (print_expr_prec 12) e1 print_bop op (print_expr_prec 12) e2
	    else
	      fprintf ff "@[<hov 1>%a@;<1 0>@[<hov 1>%a@;<1 0>%a@]@]" (print_expr_prec 12) e1 print_bop op (print_expr_prec 12) e2
	    
	| BEMult
	| BEDiv
	| BEIDiv
	| BEMod ->
	    if p > 13
	    then
	      fprintf ff "@[<hv 1>(@[<hov 1>%a@;<1 0>@[<hov 1>%a@;<1 0>%a@]@])@]" (print_expr_prec 13) e1 print_bop op (print_expr_prec 13) e2
	    else
	      fprintf ff "@[<hov 1>%a@;<1 0>@[<hov 1>%a@;<1 0>%a@]@]" (print_expr_prec 13) e1 print_bop op (print_expr_prec 13) e2

	| BEUnion
	| BEBar ->
	    if p > 14
	    then
	      fprintf ff "@[<hv 1>(@[<hov 1>%a@;<1 0>@[<hov 1>%a@;<1 0>%a@]@])@]" (print_expr_prec 14) e1 print_bop op (print_expr_prec 14) e2
	    else
	      fprintf ff "@[<hov 1>%a@;<1 0>@[<hov 1>%a@;<1 0>%a@]@]" (print_expr_prec 14) e1 print_bop op (print_expr_prec 14) e2

	| BEIntersect
	| BEExcept ->
	    if p > 15
	    then
	      fprintf ff "@[<hv 1>(@[<hov 1>%a@;<1 0>@[<hov 1>%a@;<1 0>%a@]@])@]" (print_expr_prec 15) e1 print_bop op (print_expr_prec 15) e2
	    else
	      fprintf ff "@[<hov 1>%a@;<1 0>@[<hov 1>%a@;<1 0>%a@]@]" (print_expr_prec 15) e1 print_bop op (print_expr_prec 15) e2

      end
  | EFLWR (fls, w, o, r) ->
      if p > 5
      then
	fprintf ff "@[<hv 1>(%a%a%a%a)@]" print_flclauses fls print_where w print_order o print_return r
      else
	fprintf ff "@[<hv 0>%a%a%a%a@]" print_flclauses fls print_where w print_order o print_return r
  | EIf (e1,e2,e3) ->
      if p > 8
      then
	fprintf ff "@[<hv 1>(if (%a)@;<1 0>@[<hv 2>then@;<1 0>%a@]@;<1 0>@[<hv 2>else@;<1 0>%a@])@]" (print_expr_prec 0) e1 (print_expr_prec 8) e2 (print_expr_prec 8) e3
      else
	fprintf ff "@[<hv 0>if (%a)@;<1 0>@[<hv 2>then@;<1 0>%a@]@;<1 0>@[<hv 2>else@;<1 0>%a@]@]" (print_expr_prec 0) e1 (print_expr_prec 8) e2 (print_expr_prec 8) e3
  | ESome (bdl,e2) ->
      if p > 6
      then
	fprintf ff "@[<hv 1>(@[<hv 0>@[<hv 5>some %a@]@ @[<hv 2>satisfies@ %a@]@])@]" print_bindings_list bdl (print_expr_prec 6) e2
      else
	fprintf ff "@[<hv 0>@[<hv 5>some %a@]@ @[<hv 2>satisfies@ %a@]@]" print_bindings_list bdl (print_expr_prec 6) e2
  | EEvery (bdl,e2) ->
      if p > 6
      then
	fprintf ff "@[<hv 1>(@[<hv 0>@[<hv 5>every %a@]@ @[<hv 2>satisfies@ %a@]@])@]" print_bindings_list bdl (print_expr_prec 6) e2
      else
	fprintf ff "@[<hv 0>@[<hv 5>every %a@]@ @[<hv 2>satisfies@ %a@]@]" print_bindings_list bdl (print_expr_prec 6) e2
  | ETypeswitch (e', cs) ->
      if p > 7
      then
	fprintf ff "@[<hv 1>(@[<hv 2>@[<hv 2>typeswitch (@,%a@;<0 -2>)@]@;<1 0>%a@])@]" (print_expr_prec 0) e' print_cases cs

      else
	fprintf ff "@[<hv 2>@[<hv 2>typeswitch (@,%a@;<0 -2>)@]@;<1 0>%a@]" (print_expr_prec 0) e' print_cases cs
  | EInstanceOf (e',m) ->
      if p > 9
      then
	fprintf ff "@[<hv 1>(%a@ @[<hv 2>instance of@ %a@])@]" (print_expr_prec 9) e' print_datatype m
      else
	fprintf ff "@[%a@ @[<hv 2>instance of@ %a@]@]" (print_expr_prec 9) e' print_datatype m
  | ERange (e1,e2) ->
      if p > 11
      then
	fprintf ff "@[<hv 1>(%a to %a)@]" (print_expr_prec 11) e1 (print_expr_prec 11) e2
      else
	fprintf ff "%a to %a" (print_expr_prec 11) e1 (print_expr_prec 11) e2
  | EUnaryOp (uop,e') ->
      if p > 16
      then
	fprintf ff "@[<hv 1>(%a%a)@]" print_uop uop (print_expr_prec 16) e'
      else
	fprintf ff "%a%a" print_uop uop (print_expr_prec 16) e'
  | ERoot ->
      fprintf ff "/"
  | EPath pe ->
      fprintf ff "%a" (print_path_expr p) pe
  | ESelf ->
      fprintf ff "."
  | EParent ->
      fprintf ff ".."
  | EVar vn ->
      fprintf ff "$%a" print_uqname vn
  | EScalar bv ->
      fprintf ff "%a" print_baseval bv
  | EApp (fn,es) ->
      fprintf ff "@[<hv 2>%a(@,%a@;<0 -2>)@]" print_uqname fn print_fun_exprs es
  | EList es ->
      fprintf ff "%a" (print_expr_sequence p) es
  | ECast (e',m) ->
      if p > 17
      then
	fprintf ff "@[<hv 1>(@[<hv 2>cast as %a (@,%a@;<0 -2>)@])@]" print_datatype m (print_expr_prec 0) e'
      else
	fprintf ff "@[<hv 2>cast as %a (@,%a@;<0 -2>)@]" print_datatype m (print_expr_prec 0) e'
  | ECastable (e',m) ->
      if p > 17
      then
	fprintf ff "@[<hv 1>(%a@ castable as %a)@]" (print_expr_prec 0) e' print_datatype m
      else
	fprintf ff "%a@ castable as %a" (print_expr_prec 0) e' print_datatype m
  | ETreat (e',m) ->
      if p > 17
      then
	fprintf ff "@[<hv 1>(@[<hv 2>treat as %a (@,%a@;<0 -2>)@])@]" print_datatype m (print_expr_prec 0) e'
      else
	fprintf ff "@[<hv 2>treat as %a (@,%a@;<0 -2>)@]" print_datatype m (print_expr_prec 0) e'
  | EValidate (e',m) ->
      if p > 17
      then
	fprintf ff "(@[<hv 1>validate as %a (%a)@])" print_datatype m (print_expr_prec 17) e'
      else
	fprintf ff "@[<hv 1>validate as %a (%a)@]" print_datatype m (print_expr_prec 17) e'
  | EElemFixed (se, es1, es2) ->
      begin
	match es2 with
	| [] ->
	    fprintf ff "<%a%a/>" print_elem_symbol se print_attrs es1
	| _ ->
	    fprintf ff "@[<hv 2><%a%a>@,%a@;<0 -2></%a>@]" print_elem_symbol se print_attrs es1 print_elem_contents es2 print_elem_symbol se
      end
  | EElemComputed (es0, es1) ->
      begin
	match es1 with
	| [] ->
	    fprintf ff "@[<hv 2>element {@,%a@;<0 -2>} {}@]" (print_expr_prec p) es0
	| _ ->
	    fprintf ff "@[<hv 2>element {@,%a@;<0 -2>} {@,%a@;<0 -2>}@]" (print_expr_prec p) es0 print_elem_contents es1
      end
  | EAttrFixed (sa,es) ->
      fprintf ff "@[<hv 2>attribute %a {@,%a@;<0 -2>}@]" print_attr_symbol sa print_exprs es
  | EAttrComputed (es0,es1) ->
      fprintf ff "@[<hv 2>attribute {@,%a@;<0 -2>} {@,%a@;<0 -2>}@]" (print_expr_prec p) es0 print_exprs es1
  | EText t ->
      fprintf ff "@[<hv 2>text {@,%s@;<0 -2>}@]" t
  | ETextComputed es ->
      fprintf ff "@[<hv 2>text {@,%a@;<0 -2>}@]" print_exprs es
  | EDocument es ->
      fprintf ff "@[<hv 2>document {@,%a@;<0 -2>}@]" print_exprs es
  | EPI (target,pi_content) ->
      fprintf ff "<?%s %s?>" target pi_content
  | EComment c ->
      fprintf ff "<!--%s-->" c
  | ESubtype (m1,m2) ->
      fprintf ff "@[<hv 2>assert as (@,(%a) <: (%a)@;<0 -2>)@]" print_datatype m1 print_datatype m2
  | EIntersects (m1,m2) ->
      fprintf ff "@[<hv 2>assert as (@,(%a) /\\ (%a)@;<0 -2>)@]" print_datatype m1 print_datatype m2
  | EUnordered e ->
      if p > 4
      then
	fprintf ff "@[<hv 1>(@[<hv 2>unordered@ %a@])@]" (print_expr_prec 4) e
      else
	fprintf ff "@[<hv 2>unordered@ %a@]" (print_expr_prec 4) e

and print_sort_by_clause ff (stablekind,ssl) =
  fprintf ff "@[<hv 2>%asort by (@,%a@;<0 -2>)@]" print_stablekind stablekind print_sortspeclist ssl

and print_order_by_clause ff (stablekind,ssl) =
  fprintf ff "@[<hv 2>%aorder by@ %a@]" print_stablekind stablekind print_sortspeclist ssl

and print_bindings_list ff bdl =
  match bdl with
  | [] -> ()
  | (odt,vn,ce1) :: obdl ->
      fprintf ff "@[<hv -3>$%a%a in@ %a@]%a" print_uqname vn print_optdatatype odt print_expr ce1 print_other_bindings_list obdl

and print_other_bindings_list ff bdl =
  match bdl with
  | [] -> ()
  | (odt,vn,ce1) :: obdl ->
      fprintf ff ",@ @[<hv -3>$%a%a in@ %a@]%a" print_uqname vn print_optdatatype odt print_expr ce1 print_other_bindings_list obdl

and print_sortspeclist ff ssl =
  match ssl with
  | [x] ->
      fprintf ff "@[%a@]" print_sortspec x
  | x :: ssl' ->
      fprintf ff "@[%a@],@ @[%a@]" print_sortspec x print_sortspeclist ssl'
  | [] ->
      raise (Query (Internal_Error "Empty sort kind"))

and print_sortspec ff sk =
  match sk with
  | (e,sk,esk) ->
      fprintf ff "@[%a@] @[%a@] @[%a@]" (print_expr_prec 0) e print_sortkind sk print_emptysortkind esk

and print_flclauses ff fls =
  match fls with
  | [x] ->
      begin
	match x.pfl_desc with
	| ELet (odt,vn,e') ->
	    fprintf ff "@[<hv 2>let $%a%a :=@ %a@]" print_uqname vn print_optdatatype odt (print_expr_prec 5) e'
	| EFor (odt,vn,ovn,e') ->
	    begin
	      match ovn with
	      | None ->
		  fprintf ff "@[<hv 2>for $%a%a in@ %a@]"  print_uqname vn print_optdatatype odt (print_expr_prec 5) e'
	      | Some vn' ->
		  fprintf ff "@[<hv 2>for $%a%a at $%a in@ %a@]" print_uqname vn print_optdatatype odt print_uqname vn' (print_expr_prec 5) e'
	    end
      end
  | x :: fls' ->
      begin
	match x.pfl_desc with
	| ELet (odt,vn,e') ->
	    fprintf ff "@[<hv 2>let $%a%a :=@ %a@]%a" print_uqname vn print_optdatatype odt (print_expr_prec 5) e' print_let_clauses fls'
	| EFor (odt,vn,ovn,e') ->
	    begin
	      match ovn with
	      | None ->
		  fprintf ff "@[<hv 2>for $%a%a in@ %a@]%a" print_uqname vn print_optdatatype odt (print_expr_prec 5) e' print_for_clauses fls'
	      | Some vn' ->
		  fprintf ff "@[<hv 2>for $%a%a at $%a in@ %a@]%a" print_uqname vn print_optdatatype odt print_uqname vn' (print_expr_prec 5) e' print_for_clauses fls'
	    end
      end
  | [] ->
      raise (Query (Internal_Error "Empty flower"))

and print_for_clauses ff fls =
  match fls with
  | [x] ->
      begin
	match x.pfl_desc with
	| ELet (odt,vn,e') ->
	    fprintf ff "@;<1 0>@[<hv 2>let $%a%a :=@ %a@]" print_uqname vn print_optdatatype odt (print_expr_prec 5) e'
	| EFor (odt,vn,ovn,e') ->
	    begin
	      match ovn with
	      | None ->
		  fprintf ff ",@;<1 4>@[<hv -2>$%a%a in@ %a@]" print_uqname vn print_optdatatype odt (print_expr_prec 5) e'
	      | Some vn' ->
		  fprintf ff ",@;<1 4>@[<hv -2>$%a%a at $%a in@ %a@]" print_uqname vn print_optdatatype odt print_uqname vn' (print_expr_prec 5) e'
	    end
      end
  | x :: fls' ->
      begin
	match x.pfl_desc with
	| ELet (odt,vn,e') ->
	    fprintf ff "@;<1 0>@[<hv 2>let $%a%a :=@ %a@]%a" print_uqname vn print_optdatatype odt (print_expr_prec 5) e' print_let_clauses fls'
	| EFor (odt,vn,ovn,e') ->
	    begin
	      match ovn with
	      | None ->
		  fprintf ff ",@;<1 4>@[<hv -2>$%a%a in@ %a@]%a" print_uqname vn print_optdatatype odt (print_expr_prec 5) e' print_for_clauses fls'
	      | Some vn' ->
		  fprintf ff ",@;<1 4>@[<hv -2>$%a%a at $%a in@ %a@]%a" print_uqname vn print_optdatatype odt print_uqname vn' (print_expr_prec 5) e' print_for_clauses fls'
	    end
      end
  | [] ->
      raise (Query (Internal_Error "Empty nested for expression"))

and print_let_clauses ff fls =
  match fls with
  | [x] ->
      begin
	match x.pfl_desc with
	| ELet (odt,vn,e') ->
	    fprintf ff ",@;<1 4>@[<hv -2>$%a%a :=@ %a@]" print_uqname vn print_optdatatype odt (print_expr_prec 5) e'
	| EFor (odt,vn,ovn,e') ->
	    begin
	      match ovn with
	      | None ->
		  fprintf ff "@;<1 0>@[<hv 2>for $%a%a in@ %a@]" print_uqname vn print_optdatatype odt (print_expr_prec 5) e'
	      | Some vn' ->
		  fprintf ff "@;<1 0>@[<hv 2>for $%a%a at $%a in@ %a@]" print_uqname vn print_optdatatype odt print_uqname vn' (print_expr_prec 5) e'
	    end
      end
  | x :: fls' ->
      begin
	match x.pfl_desc with
	| ELet (odt,vn,e') ->
	    fprintf ff ",@;<1 4>@[<hv -2>$%a%a :=@ %a@]%a"  print_uqname vn print_optdatatype odt (print_expr_prec 5) e' print_let_clauses fls'
	| EFor (odt,vn,ovn,e') ->
	    begin
	      match ovn with
	      | None ->
		  fprintf ff "@;<1 0>@[<hv 2>for $%a%a in@ %a@]%a" print_uqname vn print_optdatatype odt (print_expr_prec 5) e' print_for_clauses fls'
	      | Some vn' ->
		  fprintf ff "@;<1 0>@[<hv 2>for $%a%a at $%a in@ %a@]%a" print_uqname vn print_optdatatype odt print_uqname vn' (print_expr_prec 5) e' print_for_clauses fls'
	    end
      end
  | [] ->
      raise (Query (Internal_Error "Empty nested for expression"))

and print_where ff w =
  match w with
  | None ->
      ()
  | Some e ->
      fprintf ff "@;<1 0>@[<hv 2>where@;<1 0>%a@]" (print_expr_prec 5) e

and print_order ff o =
  match o with
  | None ->
      ()
  | Some oc ->
      fprintf ff "@;<1 0>%a" print_order_by_clause oc

and print_return ff r =
  fprintf ff "@;<1 0>@[<hv 2>return@;<1 0>%a@]" (print_expr_prec 5) r

and print_cases ff cs =
  match cs with
  | [] ->
      ()
  | (p,ovn,e) :: cs' ->
      begin
	match p.ppattern_desc with
	| Case m ->
	    fprintf ff "case %a as %a return@;<1 2>%a@;%a" print_optvariable ovn print_datatype m (print_expr_prec 7) e print_cases cs'
	| Default ->
	    fprintf ff "default %areturn@;<1 2>%a@;%a" print_optvariable ovn (print_expr_prec 7) e print_cases cs'
      end

and print_optvariable ff ovn =
  match ovn with
  | None -> ()
  | Some vn -> fprintf ff "$%a " print_uqname vn

and print_optvariable_as ff ovn =
  match ovn with
  | None -> ()
  | Some vn -> fprintf ff "$%a as " print_uqname vn

and print_path_expr p ff pe =
  match pe with
  | PSlash (e1,e2) ->
      if p > 18 then
	fprintf ff "@[<hv 1>(@[<hv 1>%a/@,%a@])@]" (print_expr_prec 18) e1 (print_expr_prec 18) e2
      else
	fprintf ff "@[<hv 1>%a/@,%a@]" (print_expr_prec 18) e1 (print_expr_prec 18) e2
  | PSlashSlash (e1,e2) ->
      if p > 18 then
	fprintf ff "@[<hv 1>(@[<hv 1>%a//@,%a@])@]" (print_expr_prec 18) e1 (print_expr_prec 18) e2
      else
	fprintf ff "@[<hv 1>%a//@,%a@]" (print_expr_prec 18) e1 (print_expr_prec 18) e2
  | PAxis (Child, nt) ->
      fprintf ff "%a" print_node_test nt
  | PAxis (Attribute, nt) ->
      fprintf ff "@@%a" print_node_test nt
  | PAxis (a,nt) ->
      fprintf ff "%a::%a" print_axis a print_node_test nt
  | PStepQualifiers (e',sqs) ->
      if p > 19 then
	fprintf ff "@[<hv 1>(@[<hv 1>%a@,%a@])@]" (print_expr_prec 19) e' print_step_qualifiers sqs
      else
	fprintf ff "@[<hv 1>%a@,%a@]" (print_expr_prec 19) e' print_step_qualifiers sqs
	
and print_step_qualifiers ff sqs =
  match sqs with
  | [] ->
      ()
  | sq :: sqs' ->
      fprintf ff "%a@;<0 1>%a" print_step_qualifier sq print_step_qualifiers sqs'
	
and print_step_qualifier ff sq =
  match sq.pstep_qualifier_desc with
  | PredicateQualifier e ->
      fprintf ff "[%a]" (print_expr_prec 0) e

and print_exprs ff es =
  match es with
  | [] ->
      ()
  | [e] ->
      fprintf ff "%a" (print_expr_prec 0) e
  | e :: es' ->
      fprintf ff "%a,@ %a" (print_expr_prec 0) e print_exprs es'

and print_fun_exprs ff es =
  match es with
  | [] ->
      ()
  | [e] ->
      fprintf ff "%a" (print_expr_prec 1) e
  | e :: es' ->
      fprintf ff "%a,@ %a" (print_expr_prec 1) e print_fun_exprs es'

and print_expr_sequence p ff es =
  match es with
  | [] ->
      fprintf ff "()"
  | [e] ->
      fprintf ff "%a" (print_expr_prec p) e
  | e :: es' ->
      if p > 0
      then
	fprintf ff "@[<hov 1>(%a,@ %a)@]" (print_expr_prec 0) e (print_expr_sequence 0) es'
      else
	fprintf ff "@[<hov 1>%a,@ %a@]" (print_expr_prec 0) e (print_expr_sequence 0) es'

and print_attrs ff es =
  match es with
  | [] ->
      ()
  | e :: es' ->
      fprintf ff "@ %a%a" print_attr e print_attrs es'

and print_attr ff e =
  match e.pexpr_desc with
  | EAttrFixed (sa,[e]) ->
      fprintf ff "%a=\"%a\"" print_attr_symbol sa print_attr_content e
  | EAttrFixed (sa,es) ->
      fprintf ff "%a=\"@[<hv 2>{@,%a@;<0 -2>}@]\"" print_attr_symbol sa print_exprs es
  | _ ->
      fprintf ff "@[<hv 2>{@,%a@;<0 -2>}@]" (print_expr_prec 0) e

and print_attr_content ff e =
  match e.pexpr_desc with
  | EText s ->
      fprintf ff "%s" s
  | EScalar (AtomicUntyped s) ->
      fprintf ff "%s" s
  | _ ->
      fprintf ff "@[<hv 2>{@,%a@;<0 -2>}@]" (print_expr_prec 0) e

and print_elem_contents ff ecs =
  let filtered_ecs =
    let is_not_whitespace_only_text e =
      match e.pexpr_desc with
      | EText t -> not(Whitespace.whitespace_only t)
      | _ -> true
    in
    List.filter is_not_whitespace_only_text ecs
  in
  print_elem_contents_aux ff filtered_ecs

and print_elem_contents_aux ff ecs =
  match ecs with
  | [] ->
      ()
  | e :: [] ->
      fprintf ff "%a" print_elem_content e
  | e :: ecs' ->
      fprintf ff "%a@,%a" print_elem_content e print_elem_contents_aux ecs'

and print_elem_content ff e =
  match e.pexpr_desc with
  | EElemFixed (se, es1, es2) ->
      begin
	match es2 with
	| [] ->
	    fprintf ff "<%a@[%a@]/>" print_elem_symbol se print_attrs es1
	| _ ->
	    fprintf ff "@[<hv 2><%a%a>@,%a@;<0 -2></%a>@]" print_elem_symbol se print_attrs es1 print_elem_contents es2 print_elem_symbol se
      end
  | EText t ->
      fprintf ff "%s" t
  | EPI (target,pi) ->
      fprintf ff "<?%s %s?>" target pi
  | EComment c ->
      fprintf ff "<!--%s-->" c
  | _ ->
      fprintf ff "@[<hv 2>{@,%a@;<0 -2>}@]" (print_expr_prec 0) e

(************************)
(* Print XQuery updates *)
(************************)

let print_insert_location ff il =
  match il with
  | UAsLastInto e1 ->
      fprintf ff "@[<hv 2>as last into %a@]" (print_expr_prec 0) e1
  | UAsFirstInto e1 ->
      fprintf ff "@[<hv 2>as first into %a@]" (print_expr_prec 0) e1
  | UInto e1 ->
      fprintf ff "@[<hv 2>into %a@]" (print_expr_prec 0) e1
  | UAfter e1 ->
      fprintf ff "@[<hv 2>after %a@]" (print_expr_prec 0) e1
  | UBefore e1 ->
      fprintf ff "@[<hv 2>before %a@]" (print_expr_prec 0) e1

let print_simple_update ff su =
  match su.psimple_update_desc with
  | UInsert (e1, il) ->
      fprintf ff "@[@[<hv 2>insert@ %a@]@ %a@]" (print_expr_prec 0) e1 print_insert_location il
  | UReplace (Normal_Replace, e1, e2) ->
      fprintf ff "@[@[<hv 2>replace@ %a@]@ with %a@]" (print_expr_prec 0) e1 (print_expr_prec 0) e2
  | UReplace (Value_Of_Replace, e1, e2) ->
      fprintf ff "@[@[<hv 2>replace value of@ %a@]@ with %a@]" (print_expr_prec 0) e1 (print_expr_prec 0) e2
  | UDelete e1 ->
      fprintf ff "@[<hv 2>delete@ %a@]" (print_expr_prec 0) e1
  | UEmpty ->
      fprintf ff "()"

let rec print_block_update ff bu =
  match bu with
  | [] ->
      ()
  | su :: bu' ->
      fprintf ff "@[%a]%a" print_simple_update su print_block_update bu'

let print_inside_complex_update ff cu =
  match cu.pcomplex_update_desc with
  | UCond (e1,u2,u3) ->
	fprintf ff "@[<hv 0>if (%a)@;<1 0>@[<hv 2>then@;<1 0>%a@]@;<1 0>@[<hv 2>else@;<1 0>%a@]@]" (print_expr_prec 0) e1 print_simple_update u2 print_simple_update u3
  | UFLWR (fls, w, o, bu) ->
      fprintf ff "@[<hv 0>%a%a%a%a@]" print_flclauses fls print_where w print_order o print_block_update bu

let print_complex_update ff cu =
  fprintf ff "@[<hv 2>update@ %a@]" print_inside_complex_update cu

let print_update ff u =
  match u.pupdate_desc with
  | USimple su ->
      fprintf ff "%a" print_simple_update su
  | UComplex cu ->
      fprintf ff "%a" print_complex_update cu

let print_update_body ff u =
  print_update ff u; fprintf ff "\n@?"

(* Top-level statements *)

let print_statement ff s =
  match s.pstatement_desc with
  | SExpression expr ->
      fprintf ff "%a" print_expr expr
  | SUpdate update ->
      fprintf ff "%a" print_update update

let print_statement_body ff u =
  print_statement ff u; fprintf ff "\n@?"

let rec print_statements ff ss =
  match ss with
  | [] ->
      ()
  | s :: [] ->
      fprintf ff "%a" print_statement s
  | s :: ss' ->
      fprintf ff "%a;@\n%a" print_statement s print_statements ss'

(* Sort declarations *)

let print_sort_decl ff sd =
  match sd.psort_decl_desc with
  | TypeDecl (tn,m) ->
      fprintf ff "@[<hv 2>define type %a {@,%a@;<0 -2>}@]" print_group_symbol tn print_model m
  | ElemDecl (l,tm) ->
      fprintf ff "@[<hv 2>define element %a {@,%a@;<0 -2>}@]" print_elem_symbol l print_typed_model tm
  | AttrDecl (l,tm) ->
      fprintf ff "@[<hv 2>define attribute %a {@,%a@;<0 -2>}@]" print_attr_symbol l print_typed_model tm

let rec print_sort_decls ff sds =
  match sds with
  | [] ->
      ()
  | sd :: sds' ->
      fprintf ff "%a@\n%a" print_sort_decl sd print_sort_decls sds'


(* Context declarations *)

let print_context_decl ff cd =
  match cd.pcontext_decl_desc with
  | ENamespaceDecl (ncname,uri) ->
      fprintf ff "declare namespace %s = %s" ncname (string_of_uri uri)
  | EDefaultElementNamespaceDecl uri ->
      fprintf ff "default element namespace = %s" (string_of_uri uri)
  | EDefaultFunctionNamespaceDecl uri ->
      fprintf ff "default function namespace = %s" (string_of_uri uri)
  | ESchemaDecl (ncname_opt, schemaname, schema_location_opt) ->
      begin
	match (ncname_opt,schema_location_opt) with
	| (None,None) ->
	    fprintf ff "import schema %s" schemaname
	| (Some ncname,None)->
	    fprintf ff "import schema namespace %s = %s" ncname schemaname
	| (None,Some location)->
	    fprintf ff "import schema %s at %s" schemaname location
	| (Some ncname,Some location)->
	    fprintf ff "import schema namespace %s = %s at %s" ncname schemaname location
      end
  | EDefaultSchemaDecl schemaname ->
      fprintf ff "import schema default element namespace = %s" schemaname

let rec print_context_decls ff cds =
  match cds with
  | [] ->
      ()
  | cd :: cds' ->
      fprintf ff "%a@\n%a" print_context_decl cd print_context_decls cds'

(* Function definitions *)

let rec print_datatypes ff dts =
  match dts with
  | m :: []  ->
      fprintf ff "(%a)" print_datatype m
  | m :: dts'  ->
      fprintf ff "(%a),@ %a" print_datatype m print_datatypes dts'
  | [] -> 
      ()

let rec print_varmodelparams ff vms =
  match vms with
  | (v,dt) :: []  ->
      fprintf ff "$%a as %a" print_uqname v print_datatype dt
  | (v,dt) :: ms'  ->
      fprintf ff "$%a as %a,@ %a" print_uqname v print_datatype dt print_varmodelparams ms'
  | [] ->
      ()

let print_function_def ff fd =
  match fd.pfunction_def_desc with
  | (fn, (dts,odt), fbody) ->
      begin
	match fbody with
	| EFunctionBltIn _ ->
	    fprintf ff "@[<hv 2>@[<hv 2>define function signature@ %a(@,%a@;<0 2>)@]@ as %a@]" print_uqname fn print_datatypes dts print_datatype odt
	| EFunctionUser (vars, e) ->
	    let varmodels = 
	      try
		List.combine vars dts
	      with
	      | Invalid_argument _ ->
		  raise (Query (Expr_Error(fd.pfunction_loc, ("Function signature mismatch"))))
	    in
	    fprintf ff "@[<hv 2>@[<hv 2>define function@ %a(@,%a@;<0 -2>)@] as %a {@,%a@;<0 -2>}@]" print_uqname fn print_varmodelparams varmodels print_datatype odt print_expr e
      end
  
let rec print_function_defs ff fds =
  match fds with
  | [] ->
      ()
  | fd :: fds' ->
      fprintf ff "%a@\n%a" print_function_def fd print_function_defs fds'

(* Global variables *)

let rec print_var_decl ff vd =
  match vd.pvar_decl_desc with
  | (vn, e) ->
      fprintf ff "@[<hv 2>define global $%a {@,%a@;<0 -2>}@]" print_uqname vn print_expr e

let rec print_var_decls ff vds =
  match vds with
  | [] ->
      ()
  | vd :: vds' ->
      fprintf ff "%a@\n%a" print_var_decl vd print_var_decls vds'

(* Keys *)

let print_key_decl ff kd =
  match kd.pkey_def_desc with
  | (kn, e1, e2) ->
      fprintf ff "@[<hv 2>define key %s {@,%a@;<0 -2>} {@,%a@;<0 -2>}@]" kn print_expr e1 print_expr e2

let rec print_key_decls ff kds =
  match kds with
  | [] ->
      ()
  | kd :: kds' ->
      fprintf ff "%a@\n%a" print_key_decl kd print_key_decls kds'

(* Modules *)

let print_module ff qm =
  print_context_decls ff qm.pmodule_contexts;
  print_sort_decls    ff qm.pmodule_sorts;
  print_var_decls     ff qm.pmodule_vars;
  print_key_decls     ff qm.pmodule_keys;
  print_function_defs ff qm.pmodule_functions;
  print_statements    ff qm.pmodule_exprs

let print_module_body ff m =
  print_module ff m; fprintf ff "\n@?"


(*********************)
(* Print Core XQuery *)
(*********************)

(* Expression *)

let rec print_cexpr ff e =
  if (!Conf.print_type_annotations)
  then
    let model_handle = e.pcexpr_type in
    fprintf ff "@[<hv 2>(@[%a@])@ @[%a@]@;<0 -2>@]" print_model_handle model_handle print_cexpr_desc e
  else
    print_cexpr_desc ff e

and print_cexpr_desc ff e =
  match e.pcexpr_desc with
  | CESort (v, e1,stablekind,e2,sk,esk) ->
      fprintf ff "{-- %s in --} (@[%a@]) @[%a@] sort by (@[%a@] @[%a@] @[%a@])" (string_of_uqname v) print_cexpr e1 print_stablekind stablekind print_cexpr e2 print_sortkind sk print_emptysortkind esk
  | CEUnordered e ->
      fprintf ff "unordered (@[%a@])" print_cexpr e
  | CEFor (odt,v,ovn,e1,e2) ->
      begin
	match ovn with
	| None ->
	    fprintf ff "@[<hv 2>for $%a%a in (@,%a@;<0 -2>)@]@;<1 0>@[<hv 2>return@;<1 0>%a@]" print_uqname v print_optdatatype odt print_cexpr e1 print_cexpr e2
	| Some vn' ->
	    fprintf ff "@[<hv 2>for $%a%a at $%a in (@,%a@;<0 -2>)@]@;<1 0>@[<hv 2>return@;<1 0>%a@]" print_uqname v print_optdatatype odt print_uqname vn' print_cexpr e1 print_cexpr e2
      end
  | CELet (odt,v,e1,e2) ->
      fprintf ff "@[<hv 2>let $%a%a := (@,%a@;<0 -2>)@]@;<1 0>@[<hv 2>return@;<1 0>%a@]" print_uqname v print_optdatatype odt print_cexpr e1 print_cexpr e2
  | CEIf (e1,e2,e3) ->
      fprintf ff "@[<hv 2>if (@,%a@;<0 -2>)@]@;<1 0>@[<hv 2>then (@,%a@;<0 -2>)@]@;<1 0>@[<hv 2>else (@,%a@;<0 -2>)@]" print_cexpr e1 print_cexpr e2 print_cexpr e3
  | CETypeswitch (e', cs) ->
      fprintf ff "@[<hv 2>typeswitch@,(%a)@;%a@]" print_cexpr e' print_ccases cs
  | CEVar v ->
      fprintf ff "$%a" print_uqname v
  | CEScalar bv ->
      fprintf ff "%a" print_baseval bv
  | CEText t ->
      fprintf ff "text { \"%s\" }" t
  | CETextComputed e ->
      fprintf ff "text { @[%a@] }" print_cexpr e
  | CEPI (target,pi_content) ->
      fprintf ff "<?%s %s?>" target pi_content
  | CEComment c ->
      fprintf ff "<!--%s-->" c
  | CEDocument e ->
      fprintf ff "document { @[%a@] }" print_cexpr e
  | CEApp (s,params) ->
      fprintf ff "%a(@[%a@])"print_rqname s print_params params 
  | CESeq (e1,e2) ->
      fprintf ff "%a,@ %a" print_cexpr e1 print_cexpr e2
  | CEEmpty ->
      fprintf ff "()"
  | CEElem (l, e') ->
      begin
	match e'.pcexpr_desc with
	| CEEmpty ->
	    fprintf ff "element %a {}" print_relem_symbol l 
	| _ ->
	    fprintf ff "@[<hv 2>element %a {@,%a@;<0 -2>@]}" print_relem_symbol l print_cexpr e'
      end
  | CEAnyElem (e1, e2) ->
      begin
	match e2.pcexpr_desc with
	| CEEmpty ->
	    fprintf ff "@[<hv 2>element {@,%a@;<0 -2>@]} {}" print_cexpr e1
	| _ ->
	    fprintf ff "@[<hv 2>element {@,%a@;<0 -2>@]} @[<hv 2>{@,%a@;<0 -2>@]}" print_cexpr e1 print_cexpr e2
      end
  | CEAttr(l,e') ->
      begin
	match e'.pcexpr_desc with
	| CEEmpty ->
	    fprintf ff "attribute %a {}" print_rattr_symbol l 
	| _ ->
	    fprintf ff "@[<hv 2>attribute %a {@,%a@;<0 -2>@]}" print_rattr_symbol l print_cexpr e'
      end
  | CEAnyAttr(e1,e2) ->
      begin
	match e2.pcexpr_desc with
	| CEEmpty ->
	    fprintf ff "@[<hv 2>attribute {@,%a@;<0 -2>@]} {}" print_cexpr e1
	| _ ->
	    fprintf ff "@[<hv 2>attribute {@,%a@;<0 -2>@]} @[<hv 2>{@,%a@;<0 -2>@]}" print_cexpr e1 print_cexpr e2
      end
  | CEError cexpr ->
      fprintf ff "@[<hv 2>xf:error(@,%a@;<0 -2>)@]" print_cexpr cexpr
  | CETreat (e',m) ->
      fprintf ff "@[<hv 2>treat as %a (@,%a@;<0 -2>)@]" print_model m print_cexpr e' 
  | CEValidate (e',m) ->
      fprintf ff "@[<hv 2>validate as %a (@,%a@;<0 -2>)@]" print_model m print_cexpr e' 
  | CECast (e',m) ->
      fprintf ff "@[<hv 2>cast as %a (@,%a@;<0 -2>)@]" print_model m print_cexpr e' 
  | CECastable (e',m) ->
      fprintf ff "@[<hv 2>%a@ castable as %a@]" print_model m print_cexpr e' 
  | CESubtype (m1,m2) ->
      fprintf ff "@[<hv 2>assert as (@,%a@ <:@ %a@;<0 -2>)@]" print_model m1 print_model m2
  | CEIntersects (m1,m2) ->
      fprintf ff "@[<hv 2>assert as (@,%a@ /\\@ %a@;<0 -2>)@]" print_model m1 print_model m2
  | CEForwardAxis (a,nt) ->
      fprintf ff "%a::%a" print_axis a print_cnode_test nt
  | CEReverseAxis (a,nt) ->
      fprintf ff "%a::%a" print_axis a print_cnode_test nt
  | CESome  (odt, vn, ce1, ce2) ->
      fprintf ff "some $%a%a@ in %a@ satisfies %a" print_uqname vn print_optdatatype odt print_cexpr ce1 print_cexpr ce2
  | CEEvery (odt, vn, ce1, ce2) ->
      fprintf ff "every $%a%a in %a@ satisfies %a" print_uqname vn print_optdatatype odt print_cexpr ce1 print_cexpr ce2
  | CEKeyRef (keyname,keyexpr) ->
      fprintf ff "glx:key-ref(%s,%a)" keyname print_cexpr keyexpr
  | CECreateTuple rname_list ->
      fprintf ff "@[<hov 2>[@,%a@;<0 -2>]@]" print_tuple_create rname_list
  | CEAccessTuple (ce1,rname) ->
      fprintf ff "@[<hv 1>(@,%a@;<0 -1>#%a)@]" print_cexpr ce1 print_uqname rname
  | CEForTuple (v,e1,e2) ->
      fprintf ff "@[<hv 2>for-tuple $%a in (@,%a@;<0 -2>)@]@;<1 0>@[<hv 2>return@;<1 0>%a@]" print_uqname v print_cexpr e1 print_cexpr e2

and print_tuple_create ff rname_list =
  match rname_list with
  | [] -> ()
  | (rname,ce1) :: [] ->
      fprintf ff "%a : %a" print_uqname rname print_cexpr ce1
  | (rname,ce1) :: rname_list' ->
      fprintf ff "%a:$%a,@ %a" print_uqname rname print_cexpr ce1 print_tuple_create rname_list'

and print_params ff es =
  match es with
  | e :: []  ->
      fprintf ff "(@[%a@])" print_cexpr e
  | e :: es  ->
      fprintf ff "(@[%a@]),@ @[%a@]" print_cexpr e print_params es
  | [] -> 
      ()

and print_ccases ff cs =
  match cs with
  | [] ->
      ()
  | (p,ovn,e) :: cs' ->
      fprintf ff "@[%a@] return@;<1 2>@[%a@]@;@[%a@]" print_cpattern (p, ovn) print_cexpr e print_ccases cs'

and print_cpattern ff (p, ovn) =
  match p.ppattern_desc with
  | Case m ->
      fprintf ff "case %a%a" print_optvariable_as ovn print_datatype m
  | Default ->
      fprintf ff "default %a" print_optvariable ovn 

and print_opt_cexpr ff opt_cexpr =
  match opt_cexpr with
  | None -> ()
  | Some cexpr -> fprintf ff "%a" print_cexpr cexpr

let print_core_expr ff e =
  result_flag := true;
  fprintf ff "%a%a\n@?" print_namespaces () print_cexpr e;
  result_flag := false

(* Function definition *)

let print_cfunction_def ff fd =
  match fd.pcfunction_def_desc with
  | (fn, (input_dt_ms,(odt, om)), fbody) ->
      let dts = List.map fst input_dt_ms in
      begin
	match fbody with
	| CEFunctionBltIn _ ->
	    fprintf ff "define function signature@ %a (@[%a@])@;<0 2> as @[%a@] @\n" print_rqname fn print_datatypes dts print_datatype odt
	| CEFunctionUser (vars, e) ->
	    let varmodels = 
	      try
		List.combine vars dts
	      with
	      | Invalid_argument _ ->
		  raise (Query (Expr_Error(fd.pcfunction_loc, ("Function signature mismatch"))))
	    in
	    fprintf ff "define function@ %a (@[%a@])@;<0 2> as @[%a@] @[{ @[%a@] }@] @\n" print_rqname fn print_varmodelparams varmodels print_datatype odt print_cexpr e
      end

(* MF: pretty printing is very finicky. The @, below is essential 
   to getting good line breaks between function definitions *)
let rec print_cfunction_defs ff fds =
  match fds with
  | [] ->
      ()
  | fd :: fds' ->
      fprintf ff "@[%a@,@]%a" print_cfunction_def fd print_cfunction_defs fds'

(* Variable declaration *)

let print_cvar_decl ff vd =
  match vd.pcvar_decl_desc with
  | (vn, e) ->
      fprintf ff "define global@ $%a { @[%a@]@ }" print_uqname vn print_cexpr e

let rec print_cvar_decls ff vds =
  match vds with
  | [] ->
      ()
  | vd :: vds' ->
      fprintf ff "@[%a@]@\n%a" print_cvar_decl vd print_cvar_decls vds'

(* Toplevel query *)

let print_query ff q =
  fprintf ff "@[%a@]" print_cexpr q

let rec print_queries ff qs =
  match qs with
  | q :: [] ->
      fprintf ff "@[%a@]@\n" print_query q
  | q :: qs' ->
      fprintf ff "@[%a@],@\n%a" print_query q print_queries qs'
  | [] ->
      ()

(* Statements *)

let print_statement ff s =
  match s.pcstatement_desc with
  | CSExpression ce ->
      fprintf ff "@[%a@]" print_cexpr ce
  | CSUpdate cu ->
      raise (Query (Prototype "Printing: Updates not supported yet"))

let rec print_statements ff ss =
  match ss with
  | s :: [] ->
      fprintf ff "@[%a@]@\n" print_statement s
  | s :: ss' ->
      fprintf ff "@[%a@],@\n%a" print_statement s print_statements ss'
  | [] ->
      ()

(* Complete query *)

let print_cquery_module ff qm =
  print_cfunction_defs ff qm.pcmodule_functions;
  print_cvar_decls     ff qm.pcmodule_vars;
  print_statements     ff qm.pcmodule_exprs

let rec print_cquery_module_list ff qms =
  match qms with
  | [] ->
      ()
  | qm :: qms' ->
      fprintf ff "{-- Query module --}@\n@[%a@]@\n%a" print_cquery_module qm print_cquery_module_list qms'

(* Toplevel print functions *)

let print_result_type ff m =
  result_flag := true;
  fprintf ff "%a%a\n@?" print_namespaces () print_model m;
  result_flag := false

let print_schema ff sds =
  result_flag := true;
  fprintf ff "%a%a\n@?" print_namespaces () print_sort_decls sds;
  result_flag := false

(**********************)
(* Print core updates *)
(**********************)

let print_cinsert_location ff cil =
  match cil with
  | CUAsLastInto ce1 ->
      fprintf ff "@[<hv 2>as last into %a@]" print_cexpr ce1
  | CUAsFirstInto ce1 ->
      fprintf ff "@[<hv 2>as first into %a@]" print_cexpr ce1
  | CUInto ce1 ->
      fprintf ff "@[<hv 2>into %a@]" print_cexpr ce1
  | CUAfter ce1 ->
      fprintf ff "@[<hv 2>after %a@]" print_cexpr ce1
  | CUBefore ce1 ->
      fprintf ff "@[<hv 2>before %a@]" print_cexpr ce1

let print_csimple_update ff csu =
  match csu.pcsimple_update_desc with
  | CUInsert (ce1, cil) ->
      fprintf ff "@[@[<hv 2>insert@ %a@]@ %a@]" print_cexpr ce1 print_cinsert_location cil
  | CUReplace (Normal_Replace, e1, e2) ->
      fprintf ff "@[@[<hv 2>replace@ %a@]@ with %a@]" print_cexpr e1 print_cexpr e2
  | CUReplace (Value_Of_Replace, e1, e2) ->
      fprintf ff "@[@[<hv 2>replace value of@ %a@]@ with %a@]" print_cexpr e1 print_cexpr e2
  | CUDelete ce1 ->
      fprintf ff "@[<hv 2>delete@ %a@]" print_cexpr ce1
  | CUEmpty ->
      fprintf ff "()"

let rec print_csimple_update_list ff csul =
  match csul with
  | [] -> ()
  | csu :: csul' ->
      fprintf ff "%a@ %a" print_csimple_update csu print_csimple_update_list csul'

let print_inside_ccomplex_update ff ccu =
  match ccu.pccomplex_update_desc with
  | CUCond (ce1,cu2,cu3) ->
	fprintf ff "@[<hv 0>if (%a)@;<1 0>@[<hv 2>then@;<1 0>%a@]@;<1 0>@[<hv 2>else@;<1 0>%a@]@]" print_cexpr ce1 print_csimple_update cu2 print_csimple_update cu3
  | CUForTuple (v,ce1,cul2) ->
      fprintf ff "@[<hv 2>for-tuple $%a in (@,%a@;<0 -2>)@]@;<1 0>@[<hv 2>return@;<1 0>@[%a@]@]" print_uqname v print_cexpr ce1 print_csimple_update_list cul2

let print_ccomplex_update ff ccu =
  fprintf ff "@[<hv 2>update@ %a@]" print_inside_ccomplex_update ccu

let print_cupdate ff cu =
  match cu.pcupdate_desc with
  | CUSimple csu ->
      fprintf ff "%a" print_csimple_update csu
  | CUComplex ccu ->
      fprintf ff "%a" print_ccomplex_update ccu

(***************)
(* Print stubs *)
(***************)

(* Datatype *)

let printf_datatype s dt =
  printf_stub s print_datatype dt

let bprintf_datatype s dt =
  bprintf_stub s print_datatype dt

let printf_datatype_kind s dt =
  printf_stub s print_datatype_kind dt

let bprintf_datatype_kind s dt =
  bprintf_stub s print_datatype_kind dt

(* Type *)

let printf_model s m =
  printf_stub s print_model m

let bprintf_model s m =
  bprintf_stub s print_model m

(* Full XQuery expression *)

let printf_expr s e =
  printf_stub s print_expr_body e

let bprintf_expr s e =
  bprintf_stub s print_expr_body e

let fprintf_expr f s e =
  fprintf_stub f s print_expr_body e

(* Full XQuery update *)

let fprintf_update f s u =
  fprintf_stub f s print_update_body u

(* XQuery statement *)

let fprintf_statement f s u =
  fprintf_stub f s print_statement_body u

(* XQuery module *)

let fprintf_module f s m =
  fprintf_stub f s print_module_body m

(* Core expression *)

let fprintf_cexpr f s e =
  fprintf_stub f s print_core_expr e

let printf_cexpr s e =
  printf_stub s print_core_expr e

let bprintf_cexpr s e =
  bprintf_stub s print_core_expr e

(* Core update *)

let fprintf_cupdate f s u =
  fprintf_stub f s print_cupdate u

(* Function definition *)

let fprintf_cfunction_def f s fd =
  fprintf_stub f s print_cfunction_def fd

let printf_cfunction_def s fd =
  printf_stub s print_cfunction_def fd

let bprintf_cfunction_def s fd =
  bprintf_stub s print_cfunction_def fd

(* Variable declaration *)

let fprintf_cvar_decl f s vd =
  fprintf_stub f s print_cvar_decl vd

let printf_cvar_decl s vd =
  printf_stub s print_cvar_decl vd

let bprintf_cvar_decl s vd =
  bprintf_stub s print_cvar_decl vd

(* Sort declaration *)

let printf_sort_decl s sd =
  printf_stub s print_sort_decl sd

let bprintf_sort_decl s sd =
  bprintf_stub s print_sort_decl sd

(* Core query module list *)

let printf_cquery_module_list s qms =
  printf_stub s print_cquery_module_list qms

let bprintf_cquery_module_list s qms =
  bprintf_stub s print_cquery_module_list qms

(*****************************)
(* Top-level print functions *)
(*****************************)

let fprintf_result_type c s m =
  fprintf_stub c s print_result_type m 

let printf_result_type s m =
  printf_stub s print_result_type m

let bprintf_result_type s m =
  bprintf_stub s print_result_type m

let fprintf_schema c s sds =
  fprintf_stub c s print_schema sds

let printf_schema s sds =
  printf_stub s print_schema sds

let bprintf_schema s sds =
  bprintf_stub s print_schema sds

