(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: xml_astutil.ml,v 1.2 2003/05/28 14:47:38 simeon Exp $ *)

open Namespace
open Datatypes
open Xml_ast
open Error

(* Module Xml_astutil
   Description:
     This module implements some useful operations the XML abstract
     syntax tree.
*)

(***********************)
(* Accessor operations *)
(***********************)

(* Note:
     These functions are used by the validation algorithm.  For that
     reason, they are performing namespaces resolution while walking
     down the tree.
   - Jerome
 *)

let step_within_xml_attribute nse a =
  match a with
  | (l, at) ->
      let rl = Resolve.resolve_attr_symbol nse l in
      (rl, at)

let step_within_xml_text nse e =
  match e with
  | XText t ->
      t
  | _ ->
      raise (Query (Internal_Error "Wrong step inside text in XML AST"))

let step_within_xml_pi nse e =
  match e with
  | XPI (ncn,s) ->
      (ncn,s)
  | _ ->
      raise (Query (Internal_Error "Wrong step inside PI in XML AST"))

let step_within_xml_comment nse e =
  match e with
  | XComment s ->
      s
  | _ ->
      raise (Query (Internal_Error "Wrong step inside Comment in XML AST"))

(* Steps within a forest *)

let rec get_next_xml_node wsmode forest =
  if ((wsmode = Whitespace.Preserve) || !Conf.xml_whitespace)
  then
    begin
      match forest with
      | node :: forest' ->
	  (Some node, forest')
      | [] ->
	  (None, [])
    end
  else
    begin
      match forest with
	(* End of the forest *)
      | [] ->
	  (None, [])
        (* Merge consecutive text nodes *)
      | (XText t1) :: (XText t2) :: forest' ->
	  get_next_xml_node wsmode (XText (t1^t2) :: forest')
        (* Process a single text node according to whitespace rules *)
      | (XText t) :: forest' ->
	  if (Whitespace.whitespace_only t)
	  then
	    get_next_xml_node wsmode forest'
	  else
	    (Some (XText t), forest')
	(* Otherwise process the next node *)
      | node :: forest' ->
	  (Some node, forest')
    end

let rec get_next_xml_node_no_whitespace forest =
  match forest with
  | [] ->
      (None, [])
  | (XText t1) :: (XText t2) :: forest' ->
      get_next_xml_node_no_whitespace (XText (t1^t2) :: forest')
  | (XText t) :: forest' ->
      if (Whitespace.whitespace_only t)
      then
	get_next_xml_node_no_whitespace forest'
      else
	(Some (XText t), forest')
  | node :: forest' ->
      (Some node, forest')

(* Note:
     Takes an attribute symbol and expression computing that attribute
     and returns the corresponding new prefix/namespace binding or fails
   - Jerome *)
  
let extract_ns_att s t =
  match s with
  | Sym.UAttrSym s ->
      let (prefix, localname) = Sym.uattr_name s in
      let ns_prefix =
	match prefix with
	| (NSPrefix "xmlns") ->
	    NSPrefix localname
	| NSDefaultElementPrefix ->
	    if localname = "xmlns"
	    then
	      NSDefaultElementPrefix
	    else
	      raise Not_found
	| _ ->
	    raise Not_found
      in
      let uri =
	try
	  (* Hack for now! - Jerome *)
	  ignore(anyURI_of_untyped t);
	  NSUri t
	with
	| _ ->
	    raise (Query (Mapping_Failure ("Namespace attribute xmlns does not contain URI")))
      in
      (ns_prefix, uri)
  | Sym.RAttrSym _ ->
      raise Not_found

let extract_att attname s t =
  match s with
  | Sym.UAttrSym s ->
      let (prefix, localname) = Sym.uattr_name s in
      if (prefix, localname) = attname
      then
	t
      else
	raise Not_found
  | _ ->
      raise Not_found

let rec extract_att_from_attlist attname attlist =
  match attlist with
  | [] -> (None,[])
  | att :: attlist' ->
      match att with
      | (s, t) ->
	  let (found_att, other_att) = extract_att_from_attlist attname attlist' in
	  begin
	    try
	      let att_content = extract_att attname s t in
	      (Some att_content, att :: other_att)
	    with
	    | Not_found ->
		(found_att, (att :: other_att))
	  end

let extract_ws_attribute wsmode attlist =
  let (omode, other_atts) = extract_att_from_attlist (NSPrefix "xml", "space") attlist in
  let mode =
    match omode with
    | None -> wsmode
    | Some "preserve" -> Whitespace.Preserve
    | Some "default" -> Whitespace.Default
    | Some _ ->
	raise (Query (Mapping_Failure "xml:space attribute must contain eithe \"preserve\" or \"default\""))
  in
  (mode, other_atts)

(* Extract all the attribute namespaces *)

let rec get_xml_ns_attributes att_el =
  match att_el with
  | [] -> ([],[])
  | att_e :: att_el' ->
      let (new_nss', other_att') = get_xml_ns_attributes att_el' in
      match att_e with
      | (s, elist) ->
	  begin
	    try
	      ((extract_ns_att s elist) :: new_nss', (*att_e ::*) other_att')
	    with
	    | Not_found ->
		(new_nss', (att_e :: other_att'))
	  end

(* Steps within a node *)

let step_within_xml_element nse wsmode e =
  match e with
  | XElem (l, ae, ee) ->
      let rl = Resolve.resolve_elem_symbol nse l in
      let (ns_atts, ae') = get_xml_ns_attributes ae in
      let (wsmode', ae'') = extract_ws_attribute wsmode ae' in
      (wsmode', ns_atts, rl, ae'', ee)
  | _ ->
      raise (Query (Internal_Error "Wrong step inside element in XML AST"))

