(***********************************************************************)
(*                                                                     *)
(*                                GALAX                                *)
(*                             XQuery Engine                           *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: xquery_astutil.ml,v 1.2 2003/05/21 15:03:48 simeon Exp $ *)

(* Module Xquery_astutil
   Description:
     This module implements some useful operations the XQuery abstract
     syntax tree.
*)

open Finfo
open Namespace
open Datatypes
open Xquery_ast

open Error

(* Built-in variable '.' *)

let dot      = (NSPrefix "glx", "dot")
let sequence = (NSPrefix "glx", "sequence")
let position = (NSPrefix "glx", "position")
let last     = (NSPrefix "glx", "last")

(**************************)
(* AST creation functions *)
(**************************)

(* New types AST creation *)

let mkdatatype d =
  { pdatatype_desc = d;
    pdatatype_loc = parsing_locinfo () }

let fmkdatatype d l =
  { pdatatype_desc = d;
    pdatatype_loc = l }

let mkmodel d =
  { pmodel_desc = d;
    pmodel_loc = parsing_locinfo () }

let mktyped_model d =
  { ptyped_model_desc = d;
    ptyped_model_loc = parsing_locinfo () }

let mksort_decl d =
  { psort_decl_desc = d;
    psort_decl_loc = parsing_locinfo () }

let fmkmodel d l =
  { pmodel_desc = d;
    pmodel_loc = l }

let fmktyped_model d l =
  { ptyped_model_desc = d;
    ptyped_model_loc = l }

let fmksort_decl d l =
  { psort_decl_desc = d;
    psort_decl_loc = l }

let mkanonymous_type m =
  fmktyped_model (TAnonymousType m) m.pmodel_loc

(* Common data structures AST creation *)

let mkpattern d =
  { ppattern_desc = d;
    ppattern_loc = parsing_locinfo () }

let fmkpattern d l =
  { ppattern_desc = d;
    ppattern_loc = l }

(* XQuery AST creation *)

let mkexpr d =
  { pexpr_desc = d;
    pexpr_loc = parsing_locinfo () }

let mkfl_expr d =
  { pfl_desc = d;
    pfl_loc = parsing_locinfo () }

let mkfunction_def d =
  { pfunction_def_desc = d;
    pfunction_loc = parsing_locinfo () }

let mkvar_decl d =
  { pvar_decl_desc = d;
    pvar_decl_loc = parsing_locinfo () }

let mkkey_def d =
  { pkey_def_desc = d;
    pkey_def_loc = parsing_locinfo () }

let mkcontext_decl d =
  { pcontext_decl_desc = d;
    pcontext_decl_loc = parsing_locinfo () }

let fmkexpr d l =
  { pexpr_desc = d;
    pexpr_loc = l }

let fmkfl_expr d l =
  { pfl_desc = d;
    pfl_loc = l }

let fmkfunction_def d l =
  { pfunction_def_desc = d;
    pfunction_loc = l }

let fmkvar_decl d l =
  { pvar_decl_desc = d;
    pvar_decl_loc = l }

let fmkkey_def d l =
  { pkey_def_desc = d;
    pkey_def_loc = l }

let fmkcontext_decl d l =
  { pcontext_decl_desc = d;
    pcontext_decl_loc = l }

(* Update AST creation *)

let mkupdate u =
    { pupdate_desc = u;
      pupdate_loc = parsing_locinfo () }

let fmkupdate u l =
    { pupdate_desc = u;
      pupdate_loc = l }

let mksimple_update su =
    { psimple_update_desc = su;
      psimple_update_loc = parsing_locinfo () }

let fmksimple_update su l =
    { psimple_update_desc = su;
      psimple_update_loc = l }

let mkcomplex_update su =
    { pcomplex_update_desc = su;
      pcomplex_update_loc = parsing_locinfo () }

let fmkcomplex_update su l =
    { pcomplex_update_desc = su;
      pcomplex_update_loc = l }

(* Statement AST creation *)

let mkstatement u =
    { pstatement_desc = u;
      pstatement_loc = parsing_locinfo () }

let fmkstatement u l =
    { pstatement_desc = u;
      pstatement_loc = l }


(* The empty query *)

let empty_query_module =
  { pmodule_sorts = [];
    pmodule_functions = [];
    pmodule_contexts = [];
    pmodule_vars = [];
    pmodule_keys = [];
    pmodule_exprs = [] }


(* Core AST creation *)

let mkcexpr d =
  { pcexpr_desc = d;
    pcexpr_type = None;
    pcexpr_origin = None;
    pcexpr_loc = parsing_locinfo () }

let mkstep_qualifier d =
  { pstep_qualifier_desc = d;
    pstep_qualifier_loc = parsing_locinfo () }

let mkcfunction_def d =
  { pcfunction_def_desc = d;
    pcfunction_loc = parsing_locinfo () }

let mkcvar_decl d =
    { pcvar_decl_desc = d;
      pcvar_decl_loc = parsing_locinfo () }

let mkckey_def d =
    { pckey_def_desc = d;
      pckey_def_loc = parsing_locinfo () }

let fmkcexpr d m e l =
  { pcexpr_desc = d;
    pcexpr_type = m;
    pcexpr_origin = e;
    pcexpr_loc = l }

let fmkstep_qualifier d l =
  { pstep_qualifier_desc = d;
    pstep_qualifier_loc = l }

let fmkcfunction_def d l=
  { pcfunction_def_desc = d;
    pcfunction_loc = l }

let fmkcvar_decl d l =
    { pcvar_decl_desc = d;
      pcvar_decl_loc = l }

let fmkckey_def d l =
    { pckey_def_desc = d;
      pckey_def_loc = l }

(* Core update AST creation *)

let mkcupdate cu =
    { pcupdate_desc = cu;
      pcupdate_origin = None;
      pcupdate_loc = parsing_locinfo () }

let fmkcupdate cu u l =
    { pcupdate_desc = cu;
      pcupdate_origin = u;
      pcupdate_loc = l }

let mkcsimple_update csu =
    { pcsimple_update_desc = csu;
      pcsimple_update_loc = parsing_locinfo () }

let fmkcsimple_update csu l =
    { pcsimple_update_desc = csu;
      pcsimple_update_loc = l }

let mkccomplex_update csu =
    { pccomplex_update_desc = csu;
      pccomplex_update_loc = parsing_locinfo () }

let fmkccomplex_update csu l =
    { pccomplex_update_desc = csu;
      pccomplex_update_loc = l }

(* Core statement AST creation *)

let mkcstatement cu =
    { pcstatement_desc = cu;
      pcstatement_loc = parsing_locinfo () }

let fmkcstatement cu l =
    { pcstatement_desc = cu;
      pcstatement_loc = l }

(* Core module AST creation *)

let fmkcquery_main_module_from_library_module cqm cstatements =
    { pcmodule_functions = cqm.pcmodule_functions;
      pcmodule_vars      = cqm.pcmodule_vars;
      pcmodule_keys      = cqm.pcmodule_keys;
      pcmodule_exprs     = cstatements }


(**************************)
(* AST accessor functions *)
(**************************)

(* Note:
     Takes an attribute symbol and expression computing that attribute
     and returns the corresponding new prefix/namespace binding or fails
   - Jerome
 *)

let extract_text_from_expression_list el =
  match el with
  | [] -> ""
  | [e1] ->
      begin
	match e1.pexpr_desc with
	| EText t -> t
	| EScalar sv -> erase_atomic_value sv
	| _ ->
	    raise (Query (Internal_Error "complex expression in attribute"))
      end
  | _ ->
      raise (Query (Internal_Error "many expressions in attribute"))

let extract_ns_att s el =
  match s with
  | Sym.UAttrSym s ->
      let (prefix, localname) = Sym.uattr_name s in
      let ns_prefix =
	match prefix with
	| (NSPrefix "xmlns") ->
	    NSPrefix localname
	| NSDefaultElementPrefix ->
	    if localname = "xmlns"
	    then
	      NSDefaultElementPrefix
	    else
	      raise Not_found
	| _ ->
	    raise Not_found
      in
      let t = extract_text_from_expression_list el in
      let uri =
	try
	  (* Hack for now! - Jerome *)
	  ignore(anyURI_of_untyped t);
	  NSUri t
	with
	| _ ->
	    raise (Query (Mapping_Failure ("Namespace attribute xmlns does not contain URI")))
      in
      (ns_prefix, uri)
  | Sym.RAttrSym _ ->
      raise Not_found

(* Extract all the attribute namespaces *)

let rec get_ns_attributes att_el =
  match att_el with
  | [] -> ([],[])
  | att_e :: att_el' ->
      let (new_nss', other_att') = get_ns_attributes att_el' in
      match att_e.pexpr_desc with
      | EAttrFixed (s, elist) ->
	  begin
	    try
	      ((extract_ns_att s elist) :: new_nss', att_e :: other_att')
	    with
	    | Not_found ->
		(new_nss', (att_e :: other_att'))
	  end
      |	_ ->
	  raise (Query (Internal_Error ("non-valid attribute inside element")))

let get_expression_from_statement s = 
  match s.pstatement_desc with 
    SExpression e -> e
  | _ -> raise (Query (Internal_Error ("Expected expression, found update statement")))

let get_cexpression_from_cstatement cs = 
  match cs.pcstatement_desc with 
    CSExpression ce -> ce
  | _ -> raise (Query (Internal_Error ("Expected core expression, found core update statement")))

(*******************************)
(* Returns the expression kind *)
(*******************************)

(* Note: This is used for monitoring
   - Jerome *)


type cexpr_kind =
  | CEKSort
  | CEKUnordered
  | CEKFor
  | CEKLet
  | CEKIf
  | CEKTypeswitch
  | CEKVar
  | CEKScalar
  | CEKDocument
  | CEKPI
  | CEKComment
  | CEKText
  | CEKTextComputed
  | CEKApp
  | CEKSeq
  | CEKEmpty
  | CEKElem
  | CEKAnyElem
  | CEKAttr
  | CEKAnyAttr
  | CEKError
  | CEKTreat
  | CEKValidate
  | CEKCast
  | CEKCastable
  | CEKAxis
  | CEKSubtype
  | CEKIntersects
  | CEKSome
  | CEKEvery
  | CEKKeyRef
  | CEKCreateTuple
  | CEKAccessTuple
  | CEKForTuple


let get_cexpr_kind cexpr =
  match cexpr.pcexpr_desc with
  | CESort _ ->
      CEKSort
  | CEUnordered _ ->
      CEKUnordered
  | CEFor _ ->
      CEKFor
  | CELet _ ->
      CEKLet
  | CEIf _ ->
      CEKIf
  | CETypeswitch _ ->
      CEKTypeswitch
  | CEVar _ ->
      CEKVar
  | CEScalar _ ->
      CEKScalar
  | CEDocument _ ->
      CEKDocument
  | CEPI _ ->
      CEKPI
  | CEComment _ ->
      CEKComment
  | CEText _ ->
      CEKText
  | CETextComputed _ ->
      CEKTextComputed
  | CEApp _ ->
      CEKApp
  | CESeq _ ->
      CEKSeq
  | CEEmpty ->
      CEKEmpty
  | CEElem _ ->
      CEKElem
  | CEAnyElem _ ->
      CEKAnyElem
  | CEAttr _ ->
      CEKAttr
  | CEAnyAttr _ ->
      CEKAnyAttr
  | CEError _ ->
      CEKError
  | CETreat _ ->
      CEKTreat
  | CEValidate _ ->
      CEKValidate
  | CECast _ ->
      CEKCast
  | CECastable _ ->
      CEKCastable
  | CEForwardAxis _ ->
      CEKAxis
  | CEReverseAxis _ ->
      CEKAxis
  | CESubtype _ ->
      CEKSubtype
  | CEIntersects _ ->
      CEKIntersects
  | CESome _ ->
      CEKSome
  | CEEvery _ ->
      CEKEvery
  | CEKeyRef _ ->
      CEKKeyRef
  | CECreateTuple _ ->
      CEKCreateTuple
  | CEAccessTuple _ ->
      CEKAccessTuple
  | CEForTuple _ ->
      CEKForTuple

(************************)
(* Some rewriting rules *)
(************************)

(***************************************************)
(* Substitute Expr1 for Variable in Expr2          *)
(*     Expr2 [ Expr1 / Variable ]                  *)
(* returns (Expr2', Changed) where Changed is true *)
(* if Expr1 was substituted for Variable in Expr2  *)
(***************************************************)

let rec substitute_var_aux ce1 vname ce2 = 
  let mh = ce2.pcexpr_type in
  let eh = ce2.pcexpr_origin in
  let loc = ce2.pcexpr_loc in
  match ce2.pcexpr_desc with
  | CESort (v, cexpr1, stablekind, cexpr2, sortkind, emptysortkind) -> 
      let (cexpr1',changed1) = substitute_var_aux ce1 vname cexpr1 in
      if (vname = v)
      then
	(fmkcexpr (CESort (v, cexpr1',stablekind,cexpr2,sortkind,emptysortkind)) mh eh loc, changed1)
      else
	let (cexpr2',changed2) = substitute_var_aux ce1 vname cexpr2 in
	(fmkcexpr (CESort (v, cexpr1',stablekind,cexpr2',sortkind,emptysortkind)) mh eh loc, changed1 || changed2)

  | CEUnordered cexpr ->
      let (cexpr',changed) = substitute_var_aux ce1 vname cexpr in
      (fmkcexpr (CEUnordered cexpr') mh eh loc, changed)

  | CEFor (odt, vname', None, cexpr1, cexpr2) ->
      let (cexpr1', changed1) = (substitute_var_aux ce1 vname cexpr1) in
      if vname = vname'
      then 
        (fmkcexpr (CEFor (odt, vname', None, cexpr1', cexpr2)) mh eh loc, changed1)
      else 
	let (cexpr2', changed2) = (substitute_var_aux ce1 vname cexpr2) in 
        (fmkcexpr (CEFor(odt, vname', None, cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

  | CEFor (odt, vname', Some vname'', cexpr1, cexpr2) ->
      let (cexpr1', changed1) = (substitute_var_aux ce1 vname cexpr1) in
      if (vname = vname') || (vname = vname'')
      then 
        (fmkcexpr (CEFor (odt, vname', Some vname'', cexpr1', cexpr2)) mh eh loc, changed1)
      else 
	let (cexpr2', changed2) = (substitute_var_aux ce1 vname cexpr2) in 
        (fmkcexpr (CEFor(odt, vname', Some vname'', cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

  | CEForTuple (vname', cexpr1, cexpr2) ->
      let (cexpr1', changed1) = (substitute_var_aux ce1 vname cexpr1) in
      if vname = vname'
      then 
        (fmkcexpr (CEForTuple (vname', cexpr1', cexpr2)) mh eh loc, changed1)
      else 
	let (cexpr2', changed2) = (substitute_var_aux ce1 vname cexpr2) in 
        (fmkcexpr (CEForTuple(vname', cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

  | CELet (odt, vname', cexpr1, cexpr2)  ->
      let (cexpr1', changed1) = (substitute_var_aux ce1 vname cexpr1) in
      if vname = vname'
      then 
	(fmkcexpr (CELet (odt, vname', cexpr1', cexpr2)) mh eh loc, changed1)
      else 
	let (cexpr2', changed2) = (substitute_var_aux ce1 vname cexpr2) in 
        (fmkcexpr (CELet (odt, vname', cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

  | CEIf (cexpr1, cexpr2, cexpr3) ->
      let (cexpr1', changed1) = substitute_var_aux ce1 vname cexpr1 in
      let (cexpr2', changed2) = substitute_var_aux ce1 vname cexpr2 in
      let (cexpr3', changed3) = substitute_var_aux ce1 vname cexpr3 in
      (fmkcexpr (CEIf (cexpr1', cexpr2', cexpr3')) mh eh loc, changed1 || changed2 || changed3)
   
  | CETypeswitch (cexpr, cases) ->
      let (cexpr', changed') = (substitute_var_aux ce1 vname cexpr) in 
      let (new_cases, changed_bool) = (substitute_in_cases_var ce1 vname cases) in
      (fmkcexpr (CETypeswitch (cexpr', new_cases)) mh eh loc, changed' || changed_bool) 

  | CEVar vname' -> 
      if (vname = vname') 
      then (ce1, true)
      else (ce2, false)

  | CEScalar _ ->
      (ce2, false)

  | CEText _ -> 
      (ce2, false)

  | CETextComputed cexpr ->
      let (cexpr', changed') = substitute_var_aux ce1 vname cexpr in
      (fmkcexpr (CETextComputed cexpr') mh eh loc, changed')

  | CEPI _ -> 
      (ce2, false)

  | CEComment _ -> 
      (ce2, false)

  | CEDocument cexpr ->
      let (cexpr', changed') = substitute_var_aux ce1 vname cexpr in
      (fmkcexpr (CEDocument cexpr') mh eh loc, changed')

  | CEApp (fname, arguments) ->
      let arguments' = List.map (substitute_var_aux ce1 vname) arguments in
      let arguments1 = List.map (fun (arg, changed) -> arg) arguments' in
      let changed1 = List.map (fun (arg, changed) -> changed) arguments' in
      (fmkcexpr (CEApp (fname, arguments1)) mh eh loc, List.mem true changed1) 

  | CEKeyRef (kname, arg) ->
      let (arg1,changed1) = substitute_var_aux ce1 vname arg in
      (fmkcexpr (CEKeyRef (kname, arg1)) mh eh loc, changed1) 

  | CESeq (cexpr1, cexpr2) ->
      let (cexpr1', changed1) = substitute_var_aux ce1 vname cexpr1 in
      let (cexpr2', changed2) = substitute_var_aux ce1 vname cexpr2 in
      (fmkcexpr (CESeq(cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

  | CEEmpty -> 
      (ce2, false)

  | CEElem (relem_symbol, cexpr) ->
      let (cexpr',changed') = substitute_var_aux ce1 vname cexpr in
      (fmkcexpr (CEElem (relem_symbol, cexpr')) mh eh loc, changed')

  | CEAnyElem (cexpr1, cexpr2) ->  
      let (cexpr1', changed1) = substitute_var_aux ce1 vname cexpr1 in
      let (cexpr2', changed2) = substitute_var_aux ce1 vname cexpr2 in
      (fmkcexpr (CEAnyElem (cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

  | CEAttr (rattr_symbol, cexpr)->
      let (cexpr', changed') = substitute_var_aux ce1 vname cexpr in
      (fmkcexpr (CEAttr (rattr_symbol, cexpr')) mh eh loc, changed')

  | CEAnyAttr (cexpr1, cexpr2) ->
      let (cexpr1', changed1) = substitute_var_aux ce1 vname cexpr1 in
      let (cexpr2', changed2) = substitute_var_aux ce1 vname cexpr2 in
      (fmkcexpr (CEAnyAttr (cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

  | CEError _ ->
      (ce2, false)

  | CETreat (cexpr, model) ->
      let (cexpr', changed') = substitute_var_aux ce1 vname cexpr in
      (fmkcexpr (CETreat (cexpr', model)) mh eh loc, changed')

  | CECast (cexpr, model) ->
      let (cexpr', changed') = substitute_var_aux ce1 vname cexpr in
      (fmkcexpr (CECast (cexpr', model)) mh eh loc, changed')

  | CECastable (cexpr, model) ->
      let (cexpr', changed') = substitute_var_aux ce1 vname cexpr in
      (fmkcexpr (CECastable (cexpr', model)) mh eh loc, changed')

  | CEValidate (cexpr, model) ->
      let (cexpr', changed') = substitute_var_aux ce1 vname cexpr in
      (fmkcexpr (CEValidate (cexpr', model)) mh eh loc, changed')

  | CEForwardAxis (axis, cnode_test) -> 
      if (vname = dot)
      then
        (fmkcexpr (CELet (None, vname, ce1, ce2)) mh eh loc, true)
      else
	(ce2,false)

  | CEReverseAxis (axis, cnode_test) -> 
      if (vname = dot)
      then
        (fmkcexpr (CELet (None, vname, ce1, ce2)) mh eh loc, true)
      else
	(ce2,false)

  | CESubtype (model1, model2) -> 
      (ce2,false)

  | CEIntersects (model1, model2) -> 
      (ce2,false)

  | CESome (odt, vname', cexpr1, cexpr2) ->
      let (cexpr1', changed1) = (substitute_var_aux ce1 vname cexpr1) in
      if vname = vname'
      then 
        (fmkcexpr (CESome (odt, vname', cexpr1', cexpr2)) mh eh loc, changed1)
      else 
	let (cexpr2', changed2) = (substitute_var_aux ce1 vname cexpr2) in 
        (fmkcexpr (CESome (odt, vname', cexpr1', cexpr2')) mh eh loc, changed1 || changed2)
  | CEEvery (odt, vname', cexpr1, cexpr2) ->
      let (cexpr1', changed1) = (substitute_var_aux ce1 vname cexpr1) in
      if vname = vname'
      then 
        (fmkcexpr (CEEvery (odt, vname', cexpr1', cexpr2)) mh eh loc, changed1)
      else 
	let (cexpr2', changed2) = (substitute_var_aux ce1 vname cexpr2) in 
        (fmkcexpr (CEEvery (odt, vname', cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

  | CECreateTuple rne_list ->
      let rne_list' =
	List.map
	  (fun (rname,cexpr1) ->
	    let (cexpr1',changed) = substitute_var_aux ce1 vname cexpr1
	    in ((rname,cexpr1'),changed))
	  rne_list
      in
      let rne_list1 = List.map (fun (rne, _) -> rne) rne_list' in
      let changed1 = List.map (fun (_, changed) -> changed) rne_list' in
      (fmkcexpr (CECreateTuple rne_list1) mh eh loc, List.mem true changed1)

  | CEAccessTuple (cexpr1,rname) ->
      let (cexpr1', changed1) = (substitute_var_aux ce1 vname cexpr1) in
      (fmkcexpr (CEAccessTuple (cexpr1',rname)) mh eh loc, changed1)

and substitute_in_cases_var ce1 vname cases =
  let one_case ce1 vname (pattern, ovn, case_cexpr) = 
    let (new_case_cexpr, changed) =
      match ovn with 
      | None -> substitute_var_aux ce1 vname case_cexpr
      | Some vn ->
	  if (vn = vname) then (case_cexpr, false)
	  else substitute_var_aux ce1 vname case_cexpr
    in ((pattern, ovn, new_case_cexpr), changed)
  in
  let substituted_cases = List.map (one_case ce1 vname) cases in 
    let aggregate_case ((pattern, ovn, new_case_cexpr), changed) = (pattern, ovn, new_case_cexpr) in
    let aggregate_bool ((pattern, ovn, new_case_cexpr), changed) = changed in 
  let new_cases = List.map aggregate_case substituted_cases in 
  let new_changed_value = List.map aggregate_bool substituted_cases in
  let changed_bool_value = List.mem true new_changed_value in
  (new_cases, changed_bool_value)

and substitute_var ce1 vname ce2 =
  substitute_var_aux ce1 vname ce2
 

