(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                             XQuery Engine                           *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: error.ml,v 1.30 2003/06/13 21:51:56 simeon Exp $ *)

(* Module Error
   Description:
     This module deals with error handling in Galax.
*)

open Format

type error =
  | Datamodel of string
  | Root
  | Empty_Sequence
  | Not_A_Singleton_Sequence
  | Node_Without_Id
  | URI_Error of string
  | Lexing of Finfo.finfo * string
  | Parsing of Finfo.finfo * string
  | Type_Error of Finfo.finfo * string
  | Static_Type_Error of Finfo.finfo * string
  | Expr_Error of Finfo.finfo * string
  | Bound_Error of Finfo.finfo * string
  | Parameter_Mismatch of string
  | Unknown of string
  | Internal_Error of string
  | Wrong_Args of string
  | Prototype of string
  | Error of string
  | Automata of string
  | Undefined of string
  | Undefined_Variable of string option
  | Symbol_Already_Defined of (Finfo.finfo * string * string)
  | Mapping_Failure of string
  | Unmapped_Type of string
  | Non_Deterministic_Regexp of string
  | Validation of string
  | Load_Error of string
  | Schema of string
  | Not_An_Element
  | Not_An_Attribute
  | Not_A_Unit_Type
  | Algebra of string
  | Statistics of string
  | Toplevel_Error of string
  | Cast_Error of string
  | Key_Error of (string * string)
  | KeyRef_Error of (string * string)
  | Update_Error of Finfo.finfo * string

exception Query of error

let fprint_finfo ff finfo =
  fprintf ff " [%s:]" (Finfo.finfo_to_string finfo)

let print_error_verbose ff exn =
  match exn with
  (* Prints XQuery errors *)
  | Query (Datamodel msg) ->
      fprintf ff  "Datamodel Error: %s" msg
  | Query Root ->
      fprintf ff  "Root Error: Root node"
  | Query Empty_Sequence ->
      fprintf ff "Empty_Sequence Error: Empty forest"
  | Query Not_A_Singleton_Sequence ->
      fprintf ff "Not_A_Singleton_Sequence Error: Not a singleton forest"
  | Query Node_Without_Id ->
      fprintf ff "Node_Without_Id Error: Node without id"
  | Query (URI_Error msg) ->
      fprintf ff "URI Error: %s" msg
  | Query (Lexing (fi,msg)) ->
      fprintf ff  "Lexing Error: %s" msg;
      fprint_finfo ff fi
  | Query (Parsing (fi,msg)) ->
      fprintf ff  "Parsing Error: %s" msg;
      fprint_finfo ff fi
  | Query (Type_Error (fi,msg)) ->
      fprintf ff  "Type Error: %s" msg;
      fprint_finfo ff fi
  | Query (Static_Type_Error (fi,msg)) ->
      fprintf ff  "Static_Type Error: %s" msg;
      fprint_finfo ff fi
  | Query (Expr_Error (fi,msg)) ->
      fprintf ff  "Expr Error: %s" msg;
      fprint_finfo ff fi
  | Query (Bound_Error(fi,name)) ->
      fprintf ff  "Bound Error: %s" name;
      fprint_finfo ff fi
  | Query (Parameter_Mismatch msg) ->
      fprintf ff  "Parameter_Mismatch Error: %s" msg
  | Query (Unknown msg) ->
      fprintf ff  "Unknown Error: %s" msg
  | Query (Internal_Error msg) ->
      fprintf ff  "Internal Error: %s" msg
  | Query (Wrong_Args s) ->
      fprintf ff  "Wrong_Args Error: %s" s
  | Query (Prototype msg) -> 
      fprintf ff  "Prototype Error: %s" msg
  | Query (Error msg) ->
      fprintf ff  "User Error: 'xf:error' function raised:@\n%s" (if (String.length msg > 0) then (": "^msg) else "")
  | Query (Automata msg) ->
      fprintf ff  "Automata Error: %s" msg
  | Query (Undefined msg) ->
      fprintf ff  "Undefined Error: %s" msg
  | Query (Undefined_Variable ovn) ->
      begin
      match ovn with 
      |	None -> 
	  fprintf ff  "Context item not defined"
      |	Some vn ->
	  fprintf ff  "Variable $%s not defined" vn
      end
  | Query (Symbol_Already_Defined(fi, kind, name)) ->
      fprintf ff "Symbol_Already_Defined Error: %s %s multiply defined" kind name;
      fprint_finfo ff fi
  | Query (Mapping_Failure msg) ->
      fprintf ff  "Mapping_Failure Error: %s" msg
  | Query (Unmapped_Type msg) ->
      fprintf ff  "Unmapped_Type Error: %s" msg
  | Query (Non_Deterministic_Regexp msg) ->
      fprintf ff  "Non_Deterministic_Regexp Error: %s" msg
  | Query (Validation msg) ->
      fprintf ff "Validation Error: %s" msg
  | Query (Load_Error msg) ->
      fprintf ff "Load Error: %s" msg
  | Query (Schema msg) ->
      fprintf ff "Schema_Error: %s" msg 
  | Query Not_An_Element ->
      fprintf ff "Not_An_Element Error: Not an element"
  | Query Not_An_Attribute ->
      fprintf ff "Not_An_Attribute Error: Not an attribute"
  | Query Not_A_Unit_Type ->
      fprintf ff "Not_A_Unit_Type Error: Not a unit type"
  | Query (Algebra msg) ->
      fprintf ff "Algebra Error: %s" msg
  | Query (Statistics msg) ->
      fprintf ff "Statistics Error: %s" msg
  | Query (Toplevel_Error msg) ->
      fprintf ff "Toplevel Error: %s" msg
  | Query (Cast_Error msg) ->
      fprintf ff "Cast Error: %s" msg
  | Query (Key_Error (key,msg)) ->
      fprintf ff "Key Error: %s" msg
  | Query (KeyRef_Error (key,msg)) ->
      fprintf ff "KeyRef Error: %s" msg
  | Query (Update_Error (fi,msg)) ->
      fprintf ff "Update Error: %s" msg;
      fprint_finfo ff fi
	(* Prints Caml errors *)
  | Sys_error msg ->
      fprintf ff "System Error: %s" msg 
  | e ->
      raise e

let print_error_code ff exn =
  match exn with
  (* Prints XQuery errors *)
  | Query (Datamodel _) ->
      fprintf ff  "Datamodel Error"
  | Query Root ->
      fprintf ff  "Root Error"
  | Query Empty_Sequence ->
      fprintf ff "Empty_Sequence Error"
  | Query Not_A_Singleton_Sequence ->
      fprintf ff "Not_A_Singleton_Sequence Error"
  | Query Node_Without_Id ->
      fprintf ff "Node_Without_Id Error"
  | Query (URI_Error _) ->
      fprintf ff "URI Error"
  | Query (Lexing (fi,_)) ->
      begin
	fprintf ff "Lexing Error";
	fprint_finfo ff fi
      end
  | Query (Parsing (fi,_)) ->
      begin
	fprintf ff "Parsing Error";
	fprint_finfo ff fi
      end
  | Query (Type_Error (fi,_)) ->
      begin
	fprintf ff "Type Error: ";
	fprint_finfo ff fi
      end
  | Query (Static_Type_Error (fi,_)) ->
      begin
	fprintf ff "Static_Type Error: ";
	fprint_finfo ff fi
      end
  | Query (Expr_Error (fi,_)) ->
      begin
	fprintf ff "Expr Error: ";
	fprint_finfo ff fi
      end
  | Query (Bound_Error(fi,_)) ->
      begin
	fprintf ff "Bound Error: ";
	fprint_finfo ff fi
      end
  | Query (Parameter_Mismatch _) ->
      fprintf ff "Parameter_Mismatch Error"
  | Query (Unknown _) ->
      fprintf ff "Unknown Error"
  | Query (Internal_Error _) ->
      fprintf ff "Internal Error"
  | Query (Wrong_Args _) ->
      fprintf ff "Wrong_Args Error"
  | Query (Prototype _) -> 
      fprintf ff "Prototype Error"
  | Query (Error _) ->
      fprintf ff "User Error"
  | Query (Automata _) ->
      fprintf ff "Automata Error"
  | Query (Undefined msg) ->
      fprintf ff  "Undefined Error"
  | Query (Undefined_Variable ovn) ->
      fprintf ff "Undefined Variable or Context Item"
  | Query (Symbol_Already_Defined(fi, _, _)) ->
      begin
	fprintf ff "Symbol_Already_Defined Error: ";
	fprint_finfo ff fi
      end
  | Query (Mapping_Failure _) ->
      fprintf ff  "Mapping_Failure Error"
  | Query (Unmapped_Type _) ->
      fprintf ff  "Unmapped_Type Error"
  | Query (Non_Deterministic_Regexp _) ->
      fprintf ff  "Non_Deterministic_Regexp Error"
  | Query (Validation _) ->
      fprintf ff "Validation Error"
  | Query (Load_Error _) ->
      fprintf ff "Load Error"
  | Query (Schema _) ->
      fprintf ff "Schema Error"
  | Query Not_An_Element ->
      fprintf ff "Not_An_Element Error"
  | Query Not_An_Attribute ->
      fprintf ff "Not_An_Attribute Error"
  | Query Not_A_Unit_Type ->
      fprintf ff "Not_A_Unit_Type Error"
  | Query (Algebra _) ->
      fprintf ff "Algebra Error"
  | Query (Statistics _) ->
      fprintf ff "Statistics Error"
  | Query (Toplevel_Error _) ->
      fprintf ff "Toplevel Error"
  | Query (Cast_Error _) ->
      fprintf ff "Cast Error"
  | Query (Key_Error (_,_)) ->
      fprintf ff "Key Error"
  | Query (KeyRef_Error (key,_)) ->
      fprintf ff "KeyRef Error"
	(* Prints Caml errors *)
  | Sys_error _ ->
      fprintf ff "System Error"
  | e ->
      raise e

let print_error_dispatch ff exn =
  if !Conf.verbose_error
  then
    print_error_verbose ff exn
  else
    print_error_code ff exn

let print_error_safe_dispatch ff exn =
  try
    print_error_dispatch ff exn
  with
  | _ ->
      fprintf ff "Unknown Error Occurred"

let print_error ff exn =
  begin
    print_error_dispatch ff exn;
    pp_print_flush ff ()
  end

let print_error_safe ff exn =
  begin
    print_error_safe_dispatch ff exn;
    pp_print_flush ff ()
  end

let printf_error s exn =
  Gmisc.printf_stub s print_error exn

let eprintf_error s error =
  Gmisc.eprintf_stub s print_error error

let bprintf_error s error =
  Gmisc.bprintf_stub s print_error error

let printf_error_safe s error =
  Gmisc.printf_stub s print_error_safe error

let eprintf_error_safe s error =
  Gmisc.eprintf_stub s print_error_safe error

let bprintf_error_safe s error =
  Gmisc.bprintf_stub s print_error_safe error

let incorrect_arg_count fname actual_ct formal_ct fi = 
  (Query (Type_Error(fi,("Incorrect number of arguments ("
 			 ^ (string_of_int actual_ct)
			 ^ ") for function "
			 ^ (fname)
			 ^ " ("^ (string_of_int formal_ct) ^") "))))

let print_warning ff s =
  fprintf ff "%s" s

let printf_warning s =
  Gmisc.printf_stub "" print_warning s

let eprintf_warning s =
  Gmisc.eprintf_stub "" print_warning s

let bprintf_warning s =
  Gmisc.bprintf_stub "" print_warning s

