(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                             XQuery Engine                           *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: gmisc.ml,v 1.18 2003/06/12 21:38:06 mff Exp $ *)

(* Module Gmisc
   Description:
     This module implements some additions to the Caml standard
     library that appeared to be useful in the process of developing
     Galax.
*)

(* Note:
     All these functions raise standard Caml exceptions, rather than
     Galax exceptions in the rest of the system.
   - Jerome
 *)


(*****************)
(* I/O functions *)
(*****************)

(* Load the content of a file into a string *)

let load_file f =
  let s = ref "" in
  let ic = open_in f in
  try
    while true do
      s := !s ^ (input_line ic) ^ "\n"
    done;
    !s
  with
  | End_of_file ->
      begin
	close_in ic;
	!s
      end

(******************)
(* List functions *)
(******************)

(* an additional function to partition list,
   but only getting the first element that satisfy
   the given predicate *)

let partition_first p l =
  let rec part_first remain =
    function
      |	[] ->
	  raise Not_found
      |	x :: l ->
	  let (t,v) = (p x) in
	  if t then (v,(List.rev remain)@ l) else part_first (x :: remain) l in
  part_first [] l

(* Filter non existing elements from a list *)

let filter_non_exists g1 g2 =
  let partfun y = not(List.exists (fun x -> x = y) g2) in
  List.filter partfun g1

(* Map concat *)

let map_concat f l =
  List.concat (List.map f l)

(* Triple split *)

let rec triple_split l =
  match l with
  | [] -> ([],[],[])
  | (a,b,c)::l' ->
      let (l1,l2,l3)=triple_split l' in
      (a::l1,b::l2,c::l3)

(* Remove duplicates *)

let rec remove_duplicates ps =
  match ps with 
  | [] -> []
  | a::ps' ->
      if List.mem a ps'
      then
	remove_duplicates ps'
      else
	a::remove_duplicates ps'

let rec unwrap_option_list l =
  match l with 
  | [] -> []
  | None::l' -> unwrap_option_list l'
  | (Some v)::l' -> v :: (unwrap_option_list l')

let some_list l = 
  let len = List.length l in
  let unwrap_l = (unwrap_option_list l) in
  if (List.length unwrap_l) = len then unwrap_l
  else
    raise Not_found

(*********************)
(* Hashtbl functions *)
(*********************)

(* Creates and load a hash table *)

let create_hashtable size init =
  let tbl = Hashtbl.create size in
  List.iter (fun (key, data) -> Hashtbl.add tbl key data) init;
  tbl

let all_of_hashtable ht =
  let f a b c = (a, b) :: c in 
  Hashtbl.fold f ht []

(********************)
(* String functions *)
(********************)

let split_string_on_char s c = 
    let mid = (String.index s c) in
  let sta1 = 0 in
  let len1 = mid in
  let sta2 = mid + 1 in
  let len2 = (String.length s) - sta2 in
  (String.sub s sta1 len1, String.sub s sta2 len2)

let split_right_on_char s c =
  try 
    split_string_on_char s c 
  with
    Not_found -> ("", s)

let split_left_on_char s c =
  try 
    split_string_on_char s c 
  with
    Not_found -> (s, "")

let rec split_on_char s c =
  try
    let (s1, s2) = split_string_on_char s c
    in s1 :: split_on_char s2 c
  with
    Not_found ->
      s :: []

let remove_leading s c =
  let l = String.length s in
  let rest_first =
    let rec loop s i =
      if i >= l then l-1 else
      if s.[i] = c then loop s (i+1)
      else i
    in
    loop s 0
  in
  String.sub s rest_first (l-rest_first)
    
let remove_trailing s c =
  if (String.length s > 0) then
    let rest_last =
      let rec loop s i =
	if i = 0 then 1 else
	if s.[i] = c then loop s (i-1)
	else i+1
      in
      loop s ((String.length s) - 1)
    in
    String.sub s 0 rest_last
  else s
 
(*********************)
(* Parsing functions *)
(*********************)

let wrap_lexer f s =
  let lexbuf = Lexing.from_string s in
  f lexbuf


(**********************)
(* Printing functions *)
(**********************)

(* Print to stdout *)

let printf_stub s f x =
  Format.print_flush ();
  Format.printf "%a" f x

(* Print to stderr *)

let eprintf_stub s f x =
  Format.eprintf "%a" f x;
  Format.pp_print_flush Format.err_formatter ()

(* Print to output channel *)

let fprintf_stub c s f x =
(*  Format.pp_print_flush f (); *)
  Format.fprintf c "%a" f x

(* Print to a string buffer *)

let bprintf_stub s f x =
  let buff = Buffer.create 50 in
(*  Format.pp_print_flush f (); *)
  Format.bprintf buff "%a" f x;
  let result = Buffer.contents buff in
  Buffer.reset buff;
  result


(**********************)
(* Filename functions *)
(**********************)

(* Rename a DOS dir to a UNIX dir *)

let rename_dir wd =
  Str.global_replace (Str.regexp (Str.quote "\\")) "/" wd

