(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                           XML Query Engine                          *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: occurrence.ml,v 1.1 2002/10/27 20:02:51 simeon Exp $ *)

(* Module Occurs
   Description:
     Manipulations of occurrences, i.e., an integer or
     'unbounded'. Used for XML Schema minOccurs, maxOccurs, and XQuery
     occurrence indicators.
*)

(* The type 'occurs' corresponds to minOccurs and maxOccurs *)

type occurs =
  | UP_INT of int
  | UNBOUNDED

(* An occurrence indicator is a pair of minOccurs, maxOccurs *)

type occurrence_indicator =
    (occurs * occurs)

(* operations on occurs *)

let occurs i = UP_INT i
let unbounded = UNBOUNDED

let unbounded_operation op b1 b2 =
  match (b1,b2) with
  | (UNBOUNDED,_) -> UNBOUNDED
  | (_,UNBOUNDED) -> UNBOUNDED
  | (UP_INT(i1),UP_INT(i2)) -> UP_INT(op i1 i2)

let ub_max ma1 ma2 = unbounded_operation max ma1 ma2
let ub_min ma1 ma2 =
  match (ma1,ma2) with
  | (UNBOUNDED,UP_INT(i)) -> UP_INT(i)
  | (UP_INT(i),UNBOUNDED) -> UP_INT(i)
  | (UP_INT(i1),UP_INT(i2)) -> UP_INT(min i1 i2)
  | _ -> UNBOUNDED

let ub_add ma1 ma2 = unbounded_operation (+) ma1 ma2
let ub_mult ma1 ma2 =
  match (ma1,ma2) with
  | (_,UP_INT 0) -> UP_INT 0
  | (UP_INT 0,_) -> UP_INT 0
  | (UNBOUNDED,_) -> UNBOUNDED
  | (_,UNBOUNDED) -> UNBOUNDED
  | (UP_INT(i1),UP_INT(i2)) -> UP_INT(i1*i2)

let minus i =
  function
    | UNBOUNDED -> UNBOUNDED
    | (UP_INT i') ->
	let i'' = i'-i in
	if i'' < 0 then (UP_INT 0) else (UP_INT i'')

let mult mi ma =
  match (mi,ma) with
  | (UNBOUNDED,_) -> UNBOUNDED
  | (_,UNBOUNDED) -> UNBOUNDED
  | (UP_INT(i1),UP_INT(i2)) -> UP_INT(i1 * i2)

let b_equal mi ma =
  match (mi,ma) with
  | (UNBOUNDED,UNBOUNDED) -> true
  | (_,UNBOUNDED) -> false
  | (UNBOUNDED,_) -> false
  | (UP_INT i1,UP_INT i2) -> i1=i2

let equal mi ma =
  match (mi,ma) with
  | (UNBOUNDED,UNBOUNDED) -> true
  | (_,UNBOUNDED) -> false
  | (UNBOUNDED,_) -> false
  | (UP_INT i1,UP_INT i2) -> i1=i2

let le mi ma =
  match (mi,ma) with
  | (_,UNBOUNDED) -> true
  | (UNBOUNDED,_) -> false
  | (UP_INT i1,UP_INT i2) -> i1 <= i2

(* Prints bounds *)

let string_of_occurs ub =
  match ub with
  | UP_INT i -> (string_of_int i)
  | UNBOUNDED -> "unbounded"


(* Approximate occurrence indicators as used in XQuery *)

let one       = ((UP_INT 1), (UP_INT 1))  (* Exactly one *)
let optional  = ((UP_INT 0), (UP_INT 1))  (* Zero or one '?' *)
let star      = ((UP_INT 0), UNBOUNDED)   (* Zero or more '*' *)
let plus      = ((UP_INT 1), UNBOUNDED)   (* One or more '+' *)

(* Computes an approximate occurrence indicator *)

let approximate_occurrences (b1,b2) =
  let b1' = ub_min b1 (UP_INT 1)
  and b2' =
    if le b2 (UP_INT 1)
    then
      b2
    else
      unbounded
  in
  (b1',b2')

