(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: datamodel.ml,v 1.90 2003/06/30 02:58:19 simeon Exp $ *)

(* Module: Datamodel
   Description:
     This module implements the XQuery 1.0 and XPath 2.0 data model.
*)

open Format  (* Used for serialization *)

open Error

open Namespace
open Sym
open Datatypes

(* Note:
     The Data Model requires Schema information.
 - Jerome
 *)

open Schema

open Nodeid

(*******************************************)
(* Caml types used in the module interface *)
(*******************************************)

(* Generic Caml type for sequences *)

type 'a _Sequence = 'a list
      (* [t] in the XML Query Data Model
	 is represented as t _Sequence in Caml *)

(* Caml types for XML names *)

type _ElemName = Sym.relem_symbol   (* This is an element QName - e.g., used in element construction *)
type _AttrName = Sym.rattr_symbol   (* This is an attribute QName - e.g., used in attribute construction *)

type _ncName   = Namespace.ncname

(* Note:
     Note that there is also a type _QName in the Datatypes module, used
     to represent QName *values* as instances of the data model.
   - Jerome
 *)


(*************)
(* XML Nodes *)
(*************)

(* A tag to indicate if the node is alive or not *)

type life_status =
  | Alive
  | PassedAway

(* Polymorphic node type *)

type ('a,'b) _Node_t =
    { mutable node_id: nodeid option;
      mutable node_content: 'a;
      mutable node_parent: 'b;
      mutable node_type: _Def_T list;
      mutable node_status: life_status }

(* Specific node content *)

type _DocumentNode_t =
    { doc_uri: xs_anyURI;
      mutable doc_children: _Node _Sequence;
      doc_encoding: Glx_pxp.encoding option }

and _ElementNode_t =
    { elem_tag: _ElemName;
      elem_namespaces: _NamespaceNode _Sequence;
      mutable elem_attributes: _AttributeNode _Sequence;
      mutable elem_children: _Node _Sequence;
      mutable elem_value: _SimpleValue }

and _AttributeNode_t =
    { attr_name: _AttrName;
      mutable attr_content: xdt_untyped;
      mutable attr_value: _SimpleValue }

and _TextNode_t =
    { text_content: xdt_untyped }

and _NamespaceNode_t =
    { ns_prefix: _ncName;
      ns_absolute_uri: xs_anyURI }

and _ProcessingInstructionNode_t =
    { pi_target: xs_string;
      pi_value: xs_string }
      
and _CommentNode_t =
    { comment_content: xs_string }

(* Parent information *)

and _Parent = _Node option

(* Nodes in the data model *)

and _DocumentNode     	       = (_DocumentNode_t,              _Parent) _Node_t
and _ElementNode      	       = (_ElementNode_t,               _Parent) _Node_t
and _AttributeNode    	       = (_AttributeNode_t,             _Parent) _Node_t
and _TextNode         	       = (_TextNode_t,                  _Parent) _Node_t
and _NamespaceNode    	       = (_NamespaceNode_t,             _Parent) _Node_t
and _ProcessingInstructionNode = (_ProcessingInstructionNode_t, _Parent) _Node_t
and _CommentNode               = (_CommentNode_t,               _Parent) _Node_t

(* Generic node type *)

and _Node =
  | DocumentNode of _DocumentNode
  | ElementNode of _ElementNode
  | AttributeNode of _AttributeNode
  | TextNode of _TextNode
  | NamespaceNode of _NamespaceNode
  | ProcessingInstructionNode of _ProcessingInstructionNode
  | CommentNode of _CommentNode

type _NodeKind =
  | DocumentNodeKind
  | ElementNodeKind
  | AttributeNodeKind
  | TextNodeKind
  | NamespaceNodeKind
  | ProcessingInstructionNodeKind
  | CommentNodeKind

(**************)
(* XML Values *)
(**************)

(* Note:
     Caml type names below are based on the terminology adopted by the
     XML Query and XSLT working groups, and used in the XQuery 1.0 and
     XPath 2.0 data model.

     The data model deals with values [type _Value]. Values are
     sequences of items [type _Item _Sequence]. An item [type _Item]
     can be either a node [type _Node] or an atomic value ([type
     _AtomicValue], defined in the Datatypes module).

     Most of the evaluation functions in Galax are working on the
     [_Value] type.

   - Jerome
*)

(* Items *)

type _Item =
  | Node of _Node
  | AtomicValue of _AtomicValue

(* Values *)

type _Value =
    _Item _Sequence


(**********)
(* Tuples *)
(**********)

type _Tuple =
    (Xquery_ast.rname * _Value) list

let create_tuple t = t

let access_tuple rname t =
  try
    List.assoc rname t
  with
  | _ ->
      raise (Query (Datamodel ("Tuple component " ^ (Namespace.string_of_uqname rname) ^ " not found in tuple")))


(*******************)
(* Extended values *)
(*******************)

type _Extended_Value =
  | DM_Value of _Value
  | Tuples of _Tuple _Sequence

let _Extended_Value_of_Tuples ts =
  Tuples ts

exception Not_A_Tuple of _Extended_Value
exception Not_A_Value of _Extended_Value

let _Tuples_of_Extended_Value ev =
  match ev with
  | Tuples ts -> ts
  | DM_Value [] -> []
  | _ -> raise (Not_A_Tuple ev)
(*      raise (Query (Internal_Error "Accessing an extended value which should be a tuple but is not")) *)

let _Extended_Value_of_Value v =
  DM_Value v

let _Value_of_Extended_Value ev =
  match ev with
  | DM_Value v -> v
  | Tuples ts ->  raise (Not_A_Value ev)
(*      raise (Query (Internal_Error "Accessing an extended value which should be a value but is not")) *)

let empty_extended_value ev = 
  match ev with 
  | DM_Value [] -> true
  | Tuples [] -> true
  | _ -> false

let _Concat_Extended_Value_Sequence evs =
  (* Removes the empty sequence which is the An empty tuple arises from evaluating the empty sequence in a context in which
     a tuple should be created. *)
  let cleaned_evs = List.filter (fun ev -> not (ev = DM_Value [])) evs in
  match cleaned_evs with
  | [] -> DM_Value []
  | Tuples _ :: evs' ->
      begin
	try
	  Tuples (List.concat (List.map _Tuples_of_Extended_Value evs))
	with
	| _ ->
	    raise (Query (Internal_Error "Converting a mix of tuples and values!"))
      end
  | DM_Value _ :: evs' ->
      begin
	try
	  DM_Value (List.concat (List.map _Value_of_Extended_Value evs))
	with
	| _ ->
	    raise (Query (Internal_Error "Converting a mix of tuples and values!"))
      end

(*************************************)
(* Caml specific operations on nodes *)
(*************************************)

(* Note:
     The following functions are not part of the data model but are
     useful to manipulate the Caml representation of nodes.
   - Jerome
 *)

(* Creates a new node *)

let new_node nc tyo =
  if !Conf.monitor then
    Monitor.current_stats.Monitor.node_count <- Monitor.current_stats.Monitor.node_count + 1;
  { node_id = None;
    node_content = nc;
    node_parent = None;
    node_type = tyo;
    node_status = Alive }

(* Casts from specific node types to a generic node *)

let buildDocumentNode n =
  DocumentNode n

let buildElementNode n =
  ElementNode n

let buildAttributeNode n =
  AttributeNode n

let buildTextNode n =
  TextNode n

let buildNamespaceNode n =
  NamespaceNode n

let buildProcessingInstructionNode n =
  ProcessingInstructionNode n

let buildCommentNode n =
  CommentNode n

(* Casts from generic node to a specific node type *)

let getDocumentNode n =
  match n with
  | DocumentNode n' -> n'
  | _ -> raise (Query (Datamodel "Not a DocumentNode"))

let getElementNode n =
  match n with
  | ElementNode n' -> n'
  | _ -> raise (Query (Datamodel "Not an ElementNode "))

let getAttributeNode n =
  match n with
  | AttributeNode n' -> n'
  | _ -> raise (Query (Datamodel "Not an AttributeNode "))

let getTextNode n =
  match n with
  | TextNode n' -> n'
  | _ -> raise (Query (Datamodel "Not a TextNode"))

let getNamespaceNode n =
  match n with
  | NamespaceNode n' -> n'
  | _ -> raise (Query (Datamodel "Not an NamespaceNode"))

let getProcessingInstructionNode n =
  match n with
  | ProcessingInstructionNode n' -> n'
  | _ -> raise (Query (Datamodel "Not a ProcessingInstructionNode"))

let getCommentNode n =
  match n with
  | CommentNode n' -> n'
  | _ -> raise (Query (Datamodel "Not a CommentNode"))

(* selects nodes based on cast *)

let selectDocumentNode n =
  match n with
  | DocumentNode n' -> [n']
  | _ -> []

let selectElementNode n =
  match n with
  | ElementNode n' -> [n']
  | _ -> []

let selectAttributeNode n =
  match n with
  | AttributeNode n' -> [n']
  | _ -> []

let selectTextNode n =
  match n with
  | TextNode n' -> [n']
  | _ -> []

let selectNamespaceNode n =
  match n with
  | NamespaceNode n' -> [n']
  | _ -> []

let selectProcessingInstructionNode n =
  match n with
  | ProcessingInstructionNode n' -> [n']
  | _ -> []

let selectCommentNode n =
  match n with
  | CommentNode n' -> [n']
  | _ -> []

let selectDocumentNodes nl =
  Gmisc.map_concat selectDocumentNode nl

let selectElementNodes nl =
  Gmisc.map_concat selectElementNode nl

let selectAttributeNodes nl =
  Gmisc.map_concat selectAttributeNode nl

let selectTextNodes nl =
  Gmisc.map_concat selectTextNode nl

let selectNamespaceNodes nl =
  Gmisc.map_concat selectNamespaceNode nl

let selectProcessingInstructionNodes nl =
  Gmisc.map_concat selectProcessingInstructionNode nl

let selectCommentNodes nl =
  Gmisc.map_concat selectCommentNode nl

(* Tests for a node specific type *)

let isDocumentNode n =
  match n with
  | DocumentNode _ -> true
  | _ -> false

let isElementNode n =
  match n with
  | ElementNode _ -> true
  | _ -> false

let isAttributeNode n =
  match n with
  | AttributeNode _ -> true
  | _ -> false

let isTextNode n =
  match n with
  | TextNode _ -> true
  | _ -> false

let isNamespaceNode n =
  match n with
  | NamespaceNode _ -> true
  | _ -> false

let isProcessingInstructionNode n =
  match n with
  | ProcessingInstructionNode _ -> true
  | _ -> false

let isCommentNode n =
  match n with
  | CommentNode _ -> true
  | _ -> false


(***********************)
(* Operations on nodes *)
(***********************)

(* Generic operations on nodes *)

(* Returns the kindness of a node *)

let nodeKind n =
  match n with
  | DocumentNode _ ->
      DocumentNodeKind
  | ElementNode _ ->
      ElementNodeKind
  | AttributeNode _ ->
      AttributeNodeKind
  | TextNode _ ->
      TextNodeKind
  | NamespaceNode _ ->
      NamespaceNodeKind
  | ProcessingInstructionNode _ ->
      ProcessingInstructionNodeKind
  | CommentNode _ ->
      CommentNodeKind

(* Type operations *)

(* Returns the type(s) of a node *)

let nodeType n =
  match n with
  | DocumentNode nc ->
      raise (Query (Datamodel "No types in a DocumentNode"))
  | ElementNode nc ->
      nc.node_type
  | AttributeNode nc ->
      nc.node_type
  | TextNode nc ->
      nc.node_type
  | NamespaceNode nc ->
      nc.node_type
  | ProcessingInstructionNode nc ->
      nc.node_type
  | CommentNode nc ->
      nc.node_type

(* Adds a new type to a node *)

let nodeAddType n t =
  match n with
  | DocumentNode n ->
      n.node_type <- t :: n.node_type
  | ElementNode n ->
      n.node_type <- t :: n.node_type
  | TextNode n ->
      n.node_type <- t :: n.node_type
  | AttributeNode n ->
      n.node_type <- t :: n.node_type
  | NamespaceNode n ->
      n.node_type <- t :: n.node_type
  | ProcessingInstructionNode n ->
      n.node_type <- t :: n.node_type
  | CommentNode n ->
      n.node_type <- t :: n.node_type

(* Changes the type(s) of a node *)

let nodeChangeType n t =
  match n with
  | DocumentNode n ->
      n.node_type <- t
  | ElementNode n ->
      n.node_type <- t
  | TextNode n ->
      n.node_type <- t
  | AttributeNode n ->
      n.node_type <- t
  | NamespaceNode n ->
      n.node_type <- t
  | ProcessingInstructionNode n ->
      n.node_type <- t
  | CommentNode n ->
      n.node_type <- t

(* Changes the parent node *)

(* Note:
     Update of the parent node is not available to the
     user. Constraints on parents are enforced internally by the data
     model.
   - Jerome
 *)

let update_parent p n =
  match n with
  | DocumentNode n ->
      n.node_parent <- Some p
  | ElementNode n ->
      n.node_parent <- Some p
  | AttributeNode n ->
      n.node_parent <- Some p
  | TextNode n ->
      n.node_parent <- Some p
  | NamespaceNode n ->
      n.node_parent <- Some p
  | ProcessingInstructionNode n ->
      n.node_parent <- Some p
  | CommentNode n ->
      n.node_parent <- Some p

let update_parents p nl =
  List.iter (update_parent p) nl

(* Existence operations *)

let node_status node =
  match node with
  | DocumentNode n ->
      n.node_status
  | ElementNode n ->
      n.node_status
  | AttributeNode n ->
      n.node_status
  | TextNode n ->
      n.node_status
  | NamespaceNode n ->
      n.node_status
  | ProcessingInstructionNode n ->
      n.node_status
  | CommentNode n ->
      n.node_status

let is_alive node =
  match node_status node with
  | Alive -> true
  | PassedAway -> false

let kill_node node =
  match node with
  | DocumentNode n ->
      n.node_status <- PassedAway
  | ElementNode n ->
      n.node_status <- PassedAway
  | AttributeNode n ->
      n.node_status <- PassedAway
  | TextNode n ->
      n.node_status <- PassedAway
  | NamespaceNode n ->
      n.node_status <- PassedAway
  | ProcessingInstructionNode n ->
      n.node_status <- PassedAway
  | CommentNode n ->
      n.node_status <- PassedAway

(*************************************)
(* Internal node identity operations *)
(*************************************)

(* Accesses the node id field for a given node *)

let get_node_id n =
  match n with
  | DocumentNode n ->
      n.node_id
  | ElementNode n ->
      n.node_id
  | AttributeNode n ->
      n.node_id
  | TextNode n ->
      n.node_id
  | NamespaceNode n ->
      n.node_id
  | ProcessingInstructionNode n ->
      n.node_id
  | CommentNode n ->
      n.node_id

(* Cleans up a tree, annotating nodes with fresh identifiers, and
   fixing parent pointers *)

(* Basic node copies *)

let docNodeCopy_basic x c =
  { doc_uri = x.node_content.doc_uri;
    doc_children = c;
    doc_encoding = x.node_content.doc_encoding }

let elemNodeCopy_basic x c a ns =
  { elem_tag = x.node_content.elem_tag;
    elem_namespaces = ns;
    elem_attributes = a;
    elem_children = c;
    elem_value = x.node_content.elem_value }

let rec deep_copy_aux nid_gen node =
  match node with
  | DocumentNode specific_node ->
      let prefix = new_prefix nid_gen in
      let children = specific_node.node_content.doc_children in
      let new_children = List.map (deep_copy_aux nid_gen) children in
      let postfix = new_postfix nid_gen in
      let new_nodeid = build_nodeid nid_gen prefix postfix in
      let new_specific_node = docNodeCopy_basic specific_node new_children in
      let new_generic_node = new_node new_specific_node [] in
      begin
	new_generic_node.node_id <- (Some new_nodeid);
	let result_node = (DocumentNode new_generic_node) in
	update_parents result_node new_children;
	result_node
      end
  | ElementNode specific_node ->
      let prefix = new_prefix nid_gen in
      let nsnodes = specific_node.node_content.elem_namespaces in
      let attributes = specific_node.node_content.elem_attributes in
      let children = specific_node.node_content.elem_children in
      let new_nsnodes = List.map (deep_copy_aux nid_gen) (List.map (fun x -> NamespaceNode x) nsnodes) in
      let new_attributes = List.map (deep_copy_aux nid_gen) (List.map (fun x -> AttributeNode x) attributes) in
      let new_children = List.map (deep_copy_aux nid_gen) children in
      let postfix = new_postfix nid_gen in
      let new_nodeid = build_nodeid nid_gen prefix postfix in
      let unwrapped_nsnodes = List.map getNamespaceNode new_nsnodes in
      let unwrapped_attributes = List.map getAttributeNode new_attributes in
      let new_specific_node = elemNodeCopy_basic specific_node new_children unwrapped_attributes unwrapped_nsnodes in
      let new_generic_node = new_node new_specific_node [] in
      begin
	new_generic_node.node_id <- (Some new_nodeid);
	let result_node = (ElementNode new_generic_node) in
	update_parents result_node new_nsnodes;
	update_parents result_node new_attributes;
	update_parents result_node new_children;
	result_node
      end

  | AttributeNode specific_node ->
      let new_nodeid = new_leaf_nodeid nid_gen in
      let new_generic_node = new_node specific_node.node_content [] in
      new_generic_node.node_id <- (Some new_nodeid);
      AttributeNode new_generic_node

  | TextNode specific_node ->
      let new_nodeid = new_leaf_nodeid nid_gen in
      let new_generic_node = new_node specific_node.node_content [] in
      new_generic_node.node_id <- (Some new_nodeid);
      TextNode new_generic_node

  | NamespaceNode specific_node ->
      let new_nodeid = new_leaf_nodeid nid_gen in
      let new_generic_node = new_node specific_node.node_content [] in
      new_generic_node.node_id <- (Some new_nodeid);
      NamespaceNode new_generic_node

  | ProcessingInstructionNode specific_node ->
      let new_nodeid = new_leaf_nodeid nid_gen in
      let new_generic_node = new_node specific_node.node_content [] in
      new_generic_node.node_id <- (Some new_nodeid);
      ProcessingInstructionNode new_generic_node

  | CommentNode specific_node ->
      let new_nodeid = new_leaf_nodeid nid_gen in
      let new_generic_node = new_node specific_node.node_content [] in
      new_generic_node.node_id <- (Some new_nodeid);
      CommentNode new_generic_node

let replace_node n1 n2 =
  match (n1,n2) with
  | (DocumentNode dn1, DocumentNode dn2) ->
      dn1.node_id <- dn2.node_id;
      dn1.node_content <- dn2.node_content;
      dn1.node_parent <- dn2.node_parent;
      dn1.node_type <- dn2.node_type;
      (* Also need to fix the parent pointer... *)
      let children = dn2.node_content.doc_children in
      update_parents (DocumentNode dn1) children
  | (ElementNode dn1, ElementNode dn2) ->
      dn1.node_id <- dn2.node_id;
      dn1.node_content <- dn2.node_content;
      dn1.node_parent <- dn2.node_parent;
      dn1.node_type <- dn2.node_type;
      (* Also need to fix the parent pointer... *)
      let children = dn2.node_content.elem_children in
      update_parents (ElementNode dn1) children
  | (AttributeNode dn1, AttributeNode dn2) ->
      dn1.node_id <- dn2.node_id;
      dn1.node_content <- dn2.node_content;
      dn1.node_parent <- dn2.node_parent;
      dn1.node_type <- dn2.node_type
  | (TextNode dn1, TextNode dn2) ->
      dn1.node_id <- dn2.node_id;
      dn1.node_content <- dn2.node_content;
      dn1.node_parent <- dn2.node_parent;
      dn1.node_type <- dn2.node_type
  | (NamespaceNode dn1, NamespaceNode dn2) ->
      dn1.node_id <- dn2.node_id;
      dn1.node_content <- dn2.node_content;
      dn1.node_parent <- dn2.node_parent;
      dn1.node_type <- dn2.node_type
  | (ProcessingInstructionNode dn1, ProcessingInstructionNode dn2) ->
      dn1.node_id <- dn2.node_id;
      dn1.node_content <- dn2.node_content;
      dn1.node_parent <- dn2.node_parent;
      dn1.node_type <- dn2.node_type
  | (CommentNode dn1, CommentNode dn2) ->
      dn1.node_id <- dn2.node_id;
      dn1.node_content <- dn2.node_content;
      dn1.node_parent <- dn2.node_parent;
      dn1.node_type <- dn2.node_type
  | _ ->
      raise (Query (Internal_Error "Deep copy returned the wrong node"))

let deep_copy node =
  let nid_gen = new_nodeid_gen () in
  let new_node = deep_copy_aux nid_gen node in
  replace_node node new_node;  (* Because it is already in the right place, only its children are wrong *)
  node

let nodeDeepCopy node =
  let nid_gen = new_nodeid_gen () in
  deep_copy_aux nid_gen node

(* Get real id from a clean node *)

let get_clean_node_id node =
  let ido = get_node_id node in
  match ido with
  | None ->
      raise (Query (Internal_Error "Accessed id of a dirty node"))
  | Some id ->
      id

(* Always returns a clean node, possibly doing a copy *)

let clean_node node =
  if is_alive node then
    let ido = get_node_id node in
    match ido with
    | None ->
	deep_copy node
    | Some id ->
	node
  else
    raise (Query (Datamodel "This node has been deleted !"))

(*********************)
(* Parent operations *)
(*********************)

let get_parent node =
  match clean_node node with
  | DocumentNode nc ->
      nc.node_parent
  | ElementNode nc ->
      nc.node_parent
  | AttributeNode nc ->
      nc.node_parent
  | TextNode nc ->
      nc.node_parent
  | NamespaceNode nc ->
      nc.node_parent
  | ProcessingInstructionNode nc ->
      nc.node_parent
  | CommentNode nc ->
      nc.node_parent

(* Is there a parent ? *)

let follow pref =
  match pref with
  | None -> []
  | Some p -> [p]

(* Returns the parent node *)

let nodeParent node =
  follow (get_parent node)

(* Document Node operations *)

let docNode u cs enc =
  let nc = { doc_uri = u;
	     doc_children = cs;
	     doc_encoding = enc } in
  new_node nc []

let docNodeId id u cs enc =
  let nc = { doc_uri = u;
	     doc_children = cs;
	     doc_encoding = enc } in
  let n = new_node nc [] in
  n.node_id <- (Some id);
  update_parents (DocumentNode n) cs;
  n

let docUri x =
  x.node_content.doc_uri

let docChildren x =
  let x' = getDocumentNode (clean_node (DocumentNode x)) in
  x'.node_content.doc_children

let docRootElement x =
  let x' = getDocumentNode (clean_node (DocumentNode x)) in
  List.hd (selectElementNodes x'.node_content.doc_children)

(* Text Node operations *)

let textNode s =
  let nc = { text_content = s } in
  new_node nc []

let textNodeId id s =
  let nc = { text_content = s } in
  let n = new_node nc [] in
  n.node_id <- (Some id);
  n

let textContent t =
  t.node_content.text_content

(* Elements Node operations *)

(* 

   coalesce_text_nodes collapses consecutive text nodes into one text
   node and eliminates empty text nodes.

*)

let rec coalesce_text_nodes nl = 
  match nl with
  | [] -> []
  | (TextNode n1) :: nl1 -> 
      let nl2 = coalesce_text_nodes nl1 in
      if (textContent n1 = "") then nl2 
      else 
	begin
	  match nl2 with
	  | (TextNode n2) :: nl3 -> (buildTextNode(textNode ((textContent n1)^(textContent n2)))) :: nl3 
	  | nl3 -> (TextNode n1) :: nl3
	end
  | n :: nl1 -> n :: (coalesce_text_nodes nl1) 

let elemNode qn nsl atl nl t =
  (* first create a new element *)
  let nc = { elem_tag = qn;
	     elem_namespaces = nsl;
	     elem_attributes = atl;
	     elem_children = coalesce_text_nodes nl; 
	     elem_value = [] } in
  new_node nc t

let elemNodeId id qn nsl atl nl t =
  (* first create a new element *)
  let nc = { elem_tag = qn;
	     elem_namespaces = nsl;
	     elem_attributes = atl;
	     elem_children = coalesce_text_nodes nl; 
	     elem_value = [] } in
  let n = new_node nc t in
  n.node_id <- (Some id);
  update_parents (ElementNode n) (List.map (fun x -> NamespaceNode x) nsl);
  update_parents (ElementNode n) (List.map (fun x -> AttributeNode x) atl);
  update_parents (ElementNode n) nl;
  n

let elemNodeWithValue qn nsl atl nl t ev =
  (* first create a new element *)
  let real_nl = 
    if nl = []
    then
      (* Mary: only create non-empty text nodes *)
      let text = erase_simple_value ev in
      if (String.length text > 0) then 
	[TextNode (textNode text)]
      else []
    else
      coalesce_text_nodes nl 
  in
  let nc = { elem_tag = qn;
	     elem_namespaces = nsl;
	     elem_attributes = atl;
	     elem_children = real_nl;
	     elem_value = ev } in
  new_node nc t

let elemNodeWithValueId id qn nsl atl nl t ev =
  (* first create a new element *)
  let real_nl = 
    if nl = []
    then
      let text = erase_simple_value ev in
      [TextNode (textNode text)]
    else
      coalesce_text_nodes nl 
  in
  let nc = { elem_tag = qn;
	     elem_namespaces = nsl;
	     elem_attributes = atl;
	     elem_children = real_nl;
	     elem_value = ev } in
  let n = new_node nc t in
  n.node_id <- (Some id);
  update_parents (ElementNode n) (List.map (fun x -> NamespaceNode x) nsl);
  update_parents (ElementNode n) (List.map (fun x -> AttributeNode x) atl);
  update_parents (ElementNode n) nl;
  n

let elemNodeChangeValue e ev =
  let e' = getElementNode (clean_node (ElementNode e)) in
  e'.node_content.elem_value <- ev

let rec attributes_and_nodes il =
  match il with
  | [] -> ([], [])
  | x :: il' ->
      begin 
	match x with
	| AttributeNode a -> 
	    let (atts,others) = attributes_and_nodes il' in
	    (a :: atts, others)
	| _ ->
	    let (atts,others) = attributes_and_nodes il' in
	    (atts, x :: others)
      end

let elemNode' qn nsl nl t =
  let (atl,nl) = attributes_and_nodes nl in
  elemNode qn nsl atl nl t

let elemNodeWithValue' qn nsl nl t ev =
  let (atl,nl) = attributes_and_nodes nl in
  elemNodeWithValue qn nsl atl nl t ev


let elemName e =
  let e' = getElementNode (clean_node (ElementNode e)) in
  e'.node_content.elem_tag

let elemNamespaces e =
  let e' = getElementNode (clean_node (ElementNode e)) in
  e'.node_content.elem_namespaces

let elemAttributes e =
  let e' = getElementNode (clean_node (ElementNode e)) in
  e'.node_content.elem_attributes

let elemValue e =
  e.node_content.elem_value

let elemGetId e =
  e.node_id

let elemChildren e =
  let e' = getElementNode (clean_node (ElementNode e)) in
  e'.node_content.elem_children

let elemHasElementContent e =
  let cs = elemChildren e in
  List.for_all (fun x -> not(isTextNode x)) cs

let get_scoped_attribute e prqname result_fun doc_fun =
  let rattr_sym =
    rattr_symbol prqname
  in
  let rec get_scoped_attribute_aux e =
    let attrs =
      e.node_content.elem_attributes
    in
    try
      let ab = List.find (fun a -> rattr_equals a.node_content.attr_name rattr_sym) attrs in
      Some (result_fun ab.node_content.attr_content)
    with
    | Not_found ->
	begin
	  match get_parent (ElementNode e) with
	  | None ->
	      None
	  | Some (DocumentNode n) ->
	      doc_fun n
	  | Some (ElementNode n) ->
	      get_scoped_attribute_aux n
	  | _ ->
	      raise (Query (Internal_Error "The parent of a node is neither a document or an element node"))
	end
  in
  get_scoped_attribute_aux e

let elemUri e =
  let e' = getElementNode (clean_node (ElementNode e)) in
  get_scoped_attribute e' (xml_prefix, (xml_uri,"base")) (fun x -> AnyURI._uri_of_string x) (fun x -> Some (docUri x)) (* This is the special attribute xml:base *)

let elemLang e =
  let e' = getElementNode (clean_node (ElementNode e)) in
  get_scoped_attribute e' (xml_prefix, (xml_uri,"lang")) (fun x->x) (fun x->None) (* This is the special attribute xml:lang *)

let base_uri node =
  match nodeKind node with
  | DocumentNodeKind ->
      Some(docUri (getDocumentNode node))
  | ElementNodeKind ->
      elemUri (getElementNode node)
  | AttributeNodeKind -> None
  | TextNodeKind -> None
  | NamespaceNodeKind -> None
  | ProcessingInstructionNodeKind -> None
  | CommentNodeKind -> None
 
(* Attributes Node operations *)

let rec build_attr_content nl =
  match nl with
  | [] -> ""
  | (TextNode x) :: nl' ->
      (textContent x) ^ (build_attr_content nl')
  | _ ->
      raise (Query (Internal_Error "Content of an attribute is not text"))

let attrBuildContent nl =
  build_attr_content nl

let attrNode qn ac =
  (* create the new attribute *)
  let nc = { attr_name = qn;
	     attr_content = ac;
	     attr_value = [] } in
  new_node nc []

let attrNodeId id qn ac =
  (* create the new attribute *)
  let nc = { attr_name = qn;
	     attr_content = ac;
	     attr_value = [] } in
  let n = new_node nc [] in
  n.node_id <- (Some id);
  n

let attrNodeWithValue qn ac av =
  (* create the new attribute *)
  let real_ac = 
    if ac = ""
    then
      erase_simple_value av
    else
      ac
  in
  let nc = { attr_name = qn;
	     attr_content = real_ac;
	     attr_value = av } in
  new_node nc []

let attrNodeWithValueId id qn ac av =
  (* create the new attribute *)
  let real_ac = 
    if ac = ""
    then
      erase_simple_value av
    else
      ac
  in
  let nc = { attr_name = qn;
	     attr_content = real_ac;
	     attr_value = av } in
  let n = new_node nc [] in
  n.node_id <- (Some id);
  n

let attrChangeValue a av =
  let a' = getAttributeNode (clean_node (AttributeNode a)) in
  a'.node_content.attr_value <- av

let attrName a =
  a.node_content.attr_name

let attrContent a =
  a.node_content.attr_content

let attrValue a =
  a.node_content.attr_value

(* Namespace Node operations *)

let nsNode prf abs =
  let nc = { ns_prefix = prf;
	     ns_absolute_uri = abs } in
  new_node nc []

let nsNodeId id prf abs =
  let nc = { ns_prefix = prf;
	     ns_absolute_uri = abs } in
  let n = new_node nc [] in
  n.node_id <- (Some id);
  n

let nsPrefix n =
  n.node_content.ns_prefix

let nsUri n =
  n.node_content.ns_absolute_uri

(* Processing instructions Node operations *)

let piNode tar va =
  let nc = { pi_target = tar;
	     pi_value = va }  in
  new_node nc []

let piNodeId id tar va =
  let nc = { pi_target = tar;
	     pi_value = va }  in
  let n = new_node nc [] in
  n.node_id <- (Some id);
  n

let piTarget pi =
  pi.node_content.pi_target

let piValue pi =
  pi.node_content.pi_value

(* Comments Node operations *)

let commentNode com =
  let nc = { comment_content = com } in
  new_node nc []

let commentNodeId id com =
  let nc = { comment_content = com } in
  let n = new_node nc [] in
  n.node_id <- (Some id);
  n

let commentValue c =
  c.node_content.comment_content

(*******************************)
(* Operations on node identity *)
(*******************************)

(* Node equality *)

let nodeEqual n1 n2 =
  let n1' = clean_node n1
  and n2' = clean_node n2 in
  node_is (get_clean_node_id n1') (get_clean_node_id n2')

(* Document order *)

let nodePrecedes n1 n2 =
  let n1' = clean_node n1
  and n2' = clean_node n2 in
  node_precedes (get_clean_node_id n1') (get_clean_node_id n2')

let nodeFollows n1 n2 =
  let n1' = clean_node n1
  and n2' = clean_node n2 in
  node_follows (get_clean_node_id n1') (get_clean_node_id n2')


(****************)
(* String value *)
(****************)

let rec elemStringValue nl =
  match nl with
  | [] -> ""
  | n :: nl' ->
      let node_string =
	begin
	  match n with
	  | DocumentNode nc ->
	      elemStringValue nc.node_content.doc_children
	  | ElementNode nc ->
	      elemStringValue nc.node_content.elem_children
	  | AttributeNode nc ->
	      ""
	  | TextNode nc ->
	      textContent nc
	  | NamespaceNode nc ->
	      ""
	  | ProcessingInstructionNode nc ->
	      ""
	  | CommentNode nc ->
	      ""
	end
      in
      node_string ^ (elemStringValue nl')

let nodeStringValue n =
  match n with
  | DocumentNode nc ->
      elemStringValue nc.node_content.doc_children
  | ElementNode nc ->
      elemStringValue nc.node_content.elem_children
  | AttributeNode nc ->
      attrContent nc
  | TextNode nc ->
      textContent nc
  | NamespaceNode nc ->
      nsPrefix nc
  | ProcessingInstructionNode nc ->
      piValue nc
  | CommentNode nc ->
      commentValue nc

(***************************)
(* Operations on sequences *)
(***************************)

(* Note:
     Below are all the traditional operations on lists.

     Curently, sequences are implemented as lists. A cursor interface
     on sequences would be useful for enabling other (possibly more
     efficient) implementations of sequences in the future.

   - Jerome
 *)

(* Sequence constructors *)

let sequenceEmpty = []
let sequenceBuildSingle x = [x]
let sequenceCons u ol = u :: ol
let sequenceAppend ol1 ol2 = ol1@ol2

(* Sequence accessors *)

let sequenceIsEmpty ol = ol = []
let sequenceIsSingle ol =
  match ol with
  | [] -> false
  | _ :: [] -> true
  | _ -> false
let sequenceIsOptional ol =
  match ol with
  | [] -> true
  | _ :: [] -> true
  | _ -> false
let sequenceHead ol =
  match ol with
  | [] -> raise (Query Empty_Sequence)
  | x :: _ -> x
let sequenceGetSingle ol =
  match ol with
  | [x] -> x
  | _ -> raise (Query Not_A_Singleton_Sequence)
let sequenceTail ol =
  match ol with
  | [] -> raise (Query Empty_Sequence)
  | _ :: y -> y
let sequenceLength = List.length
let sequenceNth = List.nth

(* Sequence iterators *)

let sequenceMap = List.map
let sequenceIter = List.iter
let sequenceFilter = List.filter
let sequenceFold_left = List.fold_left
let sequenceConcat = List.concat
let sequenceConcatMap k x =
  sequenceConcat (List.map k x)

(* Casts from sequences to lists and back *)

let list_of_sequence ol =
  ol
let sequence_of_list ol =
  ol

(**************************************)
(* Caml-specific operations on values *)
(**************************************)

(* Note:
     The following functions are not part of the data model but are
     useful to manipulate the Caml representation of data model
     values.
   - Jerome
 *)

(* Casts from nodes to valuess *)

let buildAttrSequence al =
  List.map (fun x -> (Node (AttributeNode x))) al

let buildNSSequence nsl =
  List.map (fun x -> (Node (NamespaceNode x))) nsl

let buildTextSequence vl =
  List.map (fun x -> (Node (TextNode x))) vl

let buildNode n =
  Node n

let buildNodeSequence nl =
  List.map buildNode nl

let buildAtomicValue v =
  AtomicValue v

let buildAtomicValueSequence vl =
  List.map buildAtomicValue vl

(* Casts from values to nodes and atomic values *)

let getNode i =
  match i with
  | Node n -> n
  | _ -> raise (Query (Datamodel "Instance is not a node"))

let getAtomicValue i =
  match i with
  | AtomicValue v -> v
  | _ -> raise (Query (Datamodel "Instance is not a value"))

(* Filters nodes and values in an instance *)

let isNode i =
  match i with
  | Node _  -> true
  | AtomicValue _ -> false

let isAtomicValue i =
  match i with
  | Node _  -> false
  | AtomicValue _ -> true


(* Separate nodes from atomic values *)

let rec filterValue il =
  match il with
  | [] -> ([], [])
  | (Node x) :: il' ->
      let (fn,fv) = filterValue il' in
      (x :: fn, fv)
  | (AtomicValue x) :: il' ->
      let (fn,fv) = filterValue il' in
      (fn, x :: fv)

let rec filterAttributesFromSequence il =
  match il with
  | [] -> ([], [])
  | (Node x) :: il' ->
      begin 
	match x with
      |	AttributeNode a -> 
	  let (atts,others) = filterAttributesFromSequence il' in
	  (a :: atts, others)
      |	_ ->
	  let (atts,others) = filterAttributesFromSequence il' in
	  if (not(atts = [])) then 
	    raise (Query(Type_Error(Finfo.bogus, ("Attributes must precede nodes in element content."))))
	  else
	    (atts, (Node x) :: others)
      end
  | (AtomicValue x) :: il' ->
      let (atts,others) = filterAttributesFromSequence il' in
      if (not(atts = [])) then 
	raise (Query(Type_Error(Finfo.bogus, ("Attributes must precede atomic values in element content."))))
      else
	(atts, (AtomicValue x) :: others)

(***************************)
(* Operations on instances *)
(***************************)

(* Note:
     Most of these operations are used directly by the evalution
     engine. It is open which of these operations should be separated
     as part of the built-in functions and operators implementation.
   - Jerome
 *)

(* Comparison operations *)

(* typed_value of a node *)
let typed_value mynode =
  match nodeKind mynode with
  | DocumentNodeKind ->
      (* Should be string-value of node *)
      sequenceEmpty
  | ElementNodeKind ->
      let myn = getElementNode mynode in
      let myv = (elemValue myn) in 
      if (not (sequenceIsEmpty myv)) then  buildAtomicValueSequence myv
      else 
        (* 
         What we implement: 
           MFF: If element node contains simple content, return it. 
           Otherwise, return its string value.
 
         Current definition:

           If the item is an element node with type annotation
           xs:anyType, then its typed value is equal to its string
           value, as an instance of xs:untypedAtomic. The typed value
           of an element node with a type annotation other than
           xs:anyType is derived from its string value and type
           annotation in a way that is consistent with XML Schema
           validation, as described in [XQuery 1.0 and XPath 2.0 Data
           Model].  

           If the item is an element node whose type annotation
           denotes a type with complex content (i.e., a type that
           permits subelements), fn:data raises a type error.

        *)

	(* Jerome: Correction - if the input node contains the empty
	   content, then return empty. *)
	begin
	  let elem_children = (elemChildren myn) in
	  if sequenceIsEmpty elem_children then
	    sequenceEmpty
	  else
	    sequenceBuildSingle (buildAtomicValue(buildAtomicUntyped (nodeStringValue mynode)))
	end
  | AttributeNodeKind ->
      (* Should be string-value of node *)
      let myn = getAttributeNode mynode in
      let myv = (attrValue myn) in 
      if (not (sequenceIsEmpty myv)) then  buildAtomicValueSequence myv
      else 
	begin
	let textcontent = attrContent myn
	in sequenceBuildSingle (buildAtomicValue(buildAtomicUntyped textcontent))
	end
  | TextNodeKind ->
      sequenceBuildSingle (buildAtomicValue(buildAtomicUntyped (textContent (getTextNode mynode))))
  | NamespaceNodeKind ->
      (* Should be string-value of node *)
      sequenceEmpty
  | ProcessingInstructionNodeKind ->
      (* Should be string-value of node *)
      sequenceEmpty
  | CommentNodeKind ->
      (* Should be string-value of node *)
      sequenceEmpty

(* item equal *)

let rec item_eq i1 i2 =
  match (i1,i2) with
  | (Node n1, Node n2) ->
      begin
	match (n1,n2) with
	| (DocumentNode n1', DocumentNode n2') ->
	    ((docUri n1') = (docUri n2'))
	      &&
	    (value_eq (buildNodeSequence (docChildren n1')) (buildNodeSequence (docChildren n2')))
	| (ElementNode n1', ElementNode n2') ->
	    let eqtag = (equaltag (RUnitElemSymbol (elemName n1')) (RUnitElemSymbol (elemName n2'))) in
	    let eqns = (value_eq (buildNSSequence (elemNamespaces n1')) (buildNSSequence (elemNamespaces n2'))) in
	    let eqattr = (value_eq (buildAttrSequence (elemAttributes n1')) (buildAttrSequence (elemAttributes n2')) ) in
           (* MF: An element may contain children nodes, atomic values, or the empty sequence. *)
	    let hasComplexContent = (List.length (elemChildren n1') > 0 || List.length(elemChildren n2') > 0) in
            let hasSimpleContent = (List.length (elemValue n1') > 0 || List.length(elemValue n2') > 0) in
	    let eqcontent = 
	      if (hasComplexContent) then 
		(value_eq (buildNodeSequence (elemChildren n1')) (buildNodeSequence (elemChildren n2'))) 
	      else if (hasSimpleContent) then 
		(value_eq (buildAtomicValueSequence (elemValue n1')) (buildAtomicValueSequence (elemValue n2'))) 
	      else true in
	    eqtag && eqns && eqattr && eqcontent

	| (AttributeNode n1', AttributeNode n2') ->
	    (equaltag (RUnitAttrSymbol (attrName n1')) (RUnitAttrSymbol (attrName n2')))
	      &&
	    (attrContent n1' = attrContent n2')
	| (TextNode n1', TextNode n2') ->
	    (textContent n1') = (textContent n2')
	| (NamespaceNode n1', NamespaceNode n2') ->
	    ((nsPrefix n1') = (nsPrefix n2'))
	      && 
	    ((nsUri n1') = (nsUri n2'))
	| (ProcessingInstructionNode n1', ProcessingInstructionNode n2') ->
	    ((piTarget n1') = (piTarget n2'))
	      &&
	    ((piValue n1') = (piValue n2'))
	| (CommentNode n1', CommentNode n2') ->
	    (commentValue n1') = (commentValue n2')
	| _ ->
	    false
      end
  | (AtomicValue v1, AtomicValue v2) ->
      atomic_value_eq v1 v2
  | _ ->
      false

and value_eq nl1 nl2 =
  if (List.length nl1 != List.length nl2)
  then
    false
  else
    List.for_all2 item_eq nl1 nl2

(* item not equal *)

let item_neq nl1 nl2 =
  not(item_eq nl1 nl2)

let value_neq nl1 nl2 =
  not(value_eq nl1 nl2)

(* Comparison for sort by *)

let item_compare i1 i2 =
  match (i1,i2) with
  | (Node n1, Node n2) ->
      let n1' = clean_node n1
      and n2' = clean_node n2 in
      node_compare  (get_clean_node_id n1') (get_clean_node_id n2')
  | (AtomicValue v1, AtomicValue v2) ->
      atomic_value_compare v1 v2
  | _ ->
      raise (Query (Parameter_Mismatch "Comparison between a value and a node"))

let item_equal i1 i2 = 
  (item_compare i1 i2) = 0

(* item less than or equal *)

let rec item_lteq i1 i2 =
  match (i1,i2) with
  | (AtomicValue v1, AtomicValue v2) ->
      (atomic_value_lteq v1 v2)
  | (Node n1, Node n2) ->
      if (nodeEqual n1 n2)
      then
	true
      else
	nodePrecedes n1 n2
  | _ ->
      false

(* item less than *)

let rec item_lt i1 i2 =
  match (i1,i2) with
  | (AtomicValue v1, AtomicValue v2) ->
      (atomic_value_lt v1 v2)
  | (Node n1, Node n2) ->
      if (nodeEqual n1 n1)
      then
	false
      else
	(nodePrecedes n1 n2)
  | _ ->
      false

(* item greater than or equal *)

let rec item_gteq i1 i2 =
  match (i1,i2) with
  | (AtomicValue v1, AtomicValue v2) ->
      (atomic_value_gteq v1 v2)
  | (Node n1, Node n2) ->
      if (nodeEqual n1 n1)
      then
	true
      else
	(nodeFollows n1 n2)
  | _ ->
      false

(* item greater than or equal *)

let rec item_gt i1 i2 =
  match (i1,i2) with
  | (AtomicValue v1, AtomicValue v2) ->
      (atomic_value_gt v1 v2)
  | (Node n1, Node n2) ->
      if (nodeEqual n1 n1)
      then
	false
      else
	(nodeFollows n1 n2)
  | _ ->
      false

(* Some useful item constructors *)

let _true =
  AtomicValue (AtomicBoolean true)

let _false =
  AtomicValue (AtomicBoolean false)

let _int i =
  AtomicValue (AtomicInt i)

let _integer i =
  AtomicValue (AtomicInteger i)

let _decimal i =
  AtomicValue (AtomicDecimal i)

let _float i =
  AtomicValue (AtomicFloat i)

let _double i =
  AtomicValue (AtomicDouble i)

(* Convert Caml boolean into datamodel boolean *)

let boolval =
  function
    | true -> _true
    | false -> _false

(* Some useful item destructors *)

let extract_optional_item insf =
  match insf with
  | [] -> None
  | x :: [] -> Some x
  | _ ->
      raise (Query (Parameter_Mismatch "Was expecting an optional parameter"))

let extract_string_item insf =
  match insf with
  | (AtomicValue (AtomicString s)) :: [] -> s
  | _ ->
      raise (Query (Parameter_Mismatch "Was expecting a string atomic value"))

let extract_bool_item insf =
  match insf with
  | (AtomicValue (AtomicBoolean b)) :: [] -> b
  | _ ->
      raise (Query (Parameter_Mismatch "Was expecting a boolean atomic value"))

let extract_int_item insf =
  match insf with
  | (AtomicValue (AtomicInt i)) :: [] -> i
  | _ ->
      raise (Query (Parameter_Mismatch "Was expecting an int atomic value"))

let extract_float_item insf =
  match insf with
  | (AtomicValue (AtomicFloat f)) :: [] -> f
  | _ ->
      raise (Query (Parameter_Mismatch "Was expecting an float atomic value"))

(*****************************)
(* XPath specific operations *)
(*****************************)

(* Note:
     These functions are not part of the data model , but are useful
     for XPath. These are currently defined only in the XPath 1.0
     specification.
   - Jerome
 *)

(* XPath axis accessors *)

let attribute_axis mynode =
  match nodeKind mynode with
  | DocumentNodeKind ->
      sequenceEmpty
  | ElementNodeKind ->
      let myn = getElementNode mynode in
      sequenceMap (fun x -> buildAttributeNode x) (elemAttributes myn)
  | AttributeNodeKind ->
      sequenceEmpty
  | TextNodeKind ->
      sequenceEmpty
  | NamespaceNodeKind ->
      sequenceEmpty
  | ProcessingInstructionNodeKind ->
      sequenceEmpty
  | CommentNodeKind ->
      sequenceEmpty

let child_axis mynode =
  match nodeKind mynode with
  | DocumentNodeKind ->
      docChildren (getDocumentNode mynode)
  | ElementNodeKind ->
      let myn = getElementNode mynode in
      elemChildren myn
      (* NOTE: The child axis should not return attributes - Jerome
	 sequenceAppend (sequenceMap (fun x -> buildAttributeNode x) (elemAttributes myn)) (elemChildren myn) *)
  | AttributeNodeKind ->
      sequenceEmpty
  | TextNodeKind ->
      sequenceEmpty
  | NamespaceNodeKind ->
      sequenceEmpty
  | ProcessingInstructionNodeKind ->
      sequenceEmpty
  | CommentNodeKind ->
      sequenceEmpty

let rec descendant_or_self_axis node =
  let children = child_axis node in
  node :: (sequenceConcatMap descendant_or_self_axis children)

let descendant_axis node =
  let children = child_axis node in
  (sequenceConcatMap descendant_or_self_axis children)

let namespace_axis node =
  raise (Query (Prototype "namespace axis not implemented"))

let parent_axis node =
  nodeParent node

let rec ancestor_or_self_axis node =
  let parents = parent_axis node in
  node :: (sequenceConcatMap ancestor_or_self_axis parents)

let ancestor_axis node =
  let parents = parent_axis node in
  (sequenceConcatMap ancestor_or_self_axis parents)

let self_axis node =
  [ node ]

(* Computes the princinpal node kind of a given axis *)

let principal_node_kind a =
  match a with
  | Xquery_ast.Ancestor ->
      ElementNodeKind
  | Xquery_ast.Ancestor_or_self ->
      ElementNodeKind
  | Xquery_ast.Attribute ->
      AttributeNodeKind
  | Xquery_ast.Child ->
      ElementNodeKind
  | Xquery_ast.Descendant ->
      ElementNodeKind
  | Xquery_ast.Descendant_or_self ->
      ElementNodeKind
  | Xquery_ast.Namespace ->
      NamespaceNodeKind
  | Xquery_ast.Parent ->
      ElementNodeKind
  | Xquery_ast.Self ->
      ElementNodeKind


(*********************************)
(* Evaluation related operations *)
(*********************************)

(* Note:
     The status of the following functions is not clear. They are
     currently used internally by the evaluation engine.
   - Jerome
 *)

(* get 'type name' of any node *)

let get_val_name arg =
  match arg with
  | Node n ->
      begin
	match n with
	| DocumentNode n' ->
	    RUnitDocSymbol
	| ElementNode n ->
	    (RUnitElemSymbol (elemName n))
	| AttributeNode n ->
	    (RUnitAttrSymbol (attrName n))
	| TextNode _ ->
	    (RUnitTextSymbol)
	| NamespaceNode _ ->
	    (RUnitSimpleSymbol Sym.nssym)
	| ProcessingInstructionNode _ ->
	    (RUnitSimpleSymbol Sym.pisym)
	| CommentNode _ ->
	    (RUnitSimpleSymbol Sym.commentsym)
      end
  | AtomicValue v ->
      let snt =
	begin
	  match v with
	  | AtomicString _ -> Sym.stringsym
	  | AtomicBoolean _ -> Sym.booleansym
	  | AtomicDecimal _ -> Sym.decimalsym
	  | AtomicFloat _ -> Sym.floatsym
	  | AtomicDouble _ -> Sym.doublesym
	  | AtomicDuration _ -> Sym.durationsym
	  | AtomicDateTime _ -> Sym.dateTimesym
	  | AtomicTime _ -> Sym.timesym
	  | AtomicDate _ -> Sym.datesym
	  | AtomicGYearMonth _ -> Sym.gYearMonthsym
	  | AtomicGYear _ -> Sym.gYearsym
	  | AtomicGMonthDay _ -> Sym.gMonthDaysym
	  | AtomicGDay _ -> Sym.gDaysym
	  | AtomicGMonth _ -> Sym.gMonthsym
	  | AtomicHexBinary _ -> Sym.hexBinarysym
	  | AtomicBase64Binary _ -> Sym.base64Binarysym
	  | AtomicAnyURI _ -> Sym.anyURIsym
	  | AtomicQName _ -> Sym.qnamesym
	  | AtomicNOTATION _ -> Sym.notationsym
	  | AtomicInteger _ -> Sym.integersym
	  | AtomicInt _ -> Sym.intsym
	  | AtomicYearMonthDuration _ -> Sym.yearMonthDurationsym
	  | AtomicDayTimeDuration _ -> Sym.dayTimeDurationsym
	  | AtomicUntyped _ -> Sym.untypedsym
	end
      in
      (RUnitSimpleSymbol snt)

(* Function that return the name of a node *)

let getNodeName node =
  match node with
  | DocumentNode dn ->
      None
  | ElementNode en ->
      Some (Sym.relem_name (elemName en))
  | AttributeNode an ->
      Some (Sym.rattr_name (attrName an))
  | TextNode vn ->
      None
  | NamespaceNode nsn ->
      Some (NSUri "", nsPrefix nsn)
  | ProcessingInstructionNode pin ->
      Some (NSUri "", piTarget pin)
  | CommentNode cn ->
      None


(* Function that return the qname of a node *)

let getNodeQName node =
  match node with
  | DocumentNode dn ->
      ""
  | ElementNode en ->
      relem_prefix_string (elemName en)
  | AttributeNode an ->
      rattr_prefix_string (attrName an)
  | TextNode vn ->
      ""
  | NamespaceNode nsn ->
      nsPrefix nsn
  | ProcessingInstructionNode pin ->
      piTarget pin
  | CommentNode cn ->
      ""

(* Function that return the symbol representation of the name of a node *)

let getNodeSym node =
  match node with
  | DocumentNode dn ->
      None
  | ElementNode en ->
      Some (RUnitElemSymbol (elemName en))
  | TextNode vn ->
      None
  | AttributeNode an ->
      Some (RUnitAttrSymbol (attrName an))
  | NamespaceNode nsn ->
      let nsqname = (NSUri "", nsPrefix nsn) in
      Some (RUnitNSSymbol (Sym.ns_symbol nsqname))
  | ProcessingInstructionNode pin ->
      None
  | CommentNode cn ->
      None

(* 

   value_to_node_sequence converts an item sequence to a node
   sequence by applying rules in Section 3.7.1.  

*)

let rec value_to_node_sequence_aux il = 
  match il with
  | [] -> ([], [])
  | [x] ->
	if (isAtomicValue x) then ([getAtomicValue x], [])
	else ([], [getNode x])
  | x :: il' ->
	let (vals, nodes) = value_to_node_sequence_aux il' 
	in
	if (isNode x) then
	  let valstr = erase_simple_value vals in
	  ([], (getNode x) :: buildTextNode(textNode valstr) :: nodes )
	else 
	  ((getAtomicValue x) :: vals, nodes)

and value_to_node_sequence il = 
  let (vals, nodes) = value_to_node_sequence_aux il in
  match vals with
  | [] -> nodes
  | _ -> 
      let valstr = erase_simple_value vals in
      buildTextNode(textNode valstr) :: nodes

let value_to_untyped il = 
  (* First convert attribute & element nodes to their string content *)
  let item_to_value n = 
    if (isAtomicValue n) then n 
    else 
      match (getNode n) with
      | TextNode _ -> n
      | _ -> buildAtomicValue(AtomicString (nodeStringValue (getNode n)))
  in
  let nodes = value_to_node_sequence (sequenceMap item_to_value il) in
  let strings = sequenceMap nodeStringValue nodes in
  String.concat "" strings 

(*********************)
(* Update operations *)
(*********************)

(* Note:
     These functions are not part of the official XPath/XQuery data
     model. They are used as a low-level interface to the update
     language.
   - Jerome
 *)

(* Mark recursively as deleted *)

let rec kill_descendants node =
  (* Note:
       Here, be very careful to kill the children in a deep-first
       way. Otherwise, the child_axis call fails since applied on a
       deleted node!
     - Jerome
     *)
   (* First kill the children *)
  List.iter kill_descendants (child_axis node);
   (* Then kill the node itself *)
  kill_node node

(* Cleans up the children which have been deleted *)

let rec clean_killed_children node =
  let kids = child_axis node in
  let alive_kids = (List.filter is_alive kids) in
  List.iter clean_killed_children alive_kids;
  match node with
  | DocumentNode n ->
      let doc_content = n.node_content in
      let new_doc_content =
	{ doc_uri = doc_content.doc_uri;
	  doc_children = alive_kids;
	  doc_encoding = doc_content.doc_encoding } in
      n.node_content <- new_doc_content;
      n.node_status <- Alive
  | ElementNode n ->
      let elem_content = n.node_content in
      let new_elem_content =
	{ elem_tag = elem_content.elem_tag;
	  elem_namespaces = elem_content.elem_namespaces;
	  elem_attributes = elem_content.elem_attributes;
	  elem_children = alive_kids;
	  elem_value = elem_content.elem_value } in
      n.node_content <- new_elem_content;
      n.node_status <- Alive
  | TextNode n ->
      n.node_status <- Alive
  | AttributeNode n ->
      n.node_status <- Alive
  | NamespaceNode n ->
      n.node_status <- Alive
  | ProcessingInstructionNode n ->
      n.node_status <- Alive
  | CommentNode n ->
      n.node_status <- Alive

(* Deletes a node, and all of its children *)

let rec delete_node_with_children insts =
  let inst =
    try
      sequenceGetSingle insts
    with
    | _ ->
	raise (Query (Internal_Error "Delete operation: Expecting a single node but got a sequence"))
  in
  match inst with
  | Node node ->
      let node' = clean_node node in
      let opt_parent_node = get_parent node' in
      begin
	kill_descendants node';
	begin
	  match opt_parent_node with
	  | None -> ()
	  | Some parent_node ->
	      clean_killed_children parent_node
	end
      end
  | AtomicValue dv ->
      raise (Query (Datamodel "Delete operation: Expecting a node but got a value"))

(**********)
(* Insert *)
(**********)

(* Auxiliary Datamodel Primitives for insert *)

let get_first_child parent_node = 
  let node_children = child_axis parent_node in
  match node_children with
  | [] -> None
  | head :: _ -> Some head 

let rec next_node node node_list = 
  match node_list with
  | [] -> None
  | head :: tail -> 
      if (nodeEqual node head) then
	match tail with 
	| [] -> None
	| head' :: tail' -> Some head'
      else 
	next_node node tail

let get_next_sibling parent_node node = 
  let node_children = child_axis parent_node in 
  next_node node node_children

(* auxilliary primitives for insert-into *)

let rec insert_at_end node node_children = 
  match node_children with
  | [] -> [node]
  | head :: tail -> head :: (insert_at_end node tail)

let rec insert_content item_list node_children =
  match item_list with 
  | [] -> node_children
  | Node n :: tail -> 
      (* Note:
 	  Creates a clean node here. This creates node ids to make
 	  sure the nodes are kept there after inserted. However, the
 	  correspondind node-id are going to be wrong. This will have
 	  to be fixed.
        - Jerome
       *)
      let n' = nodeDeepCopy n in
      let new_node_children  = insert_at_end n' node_children in
      insert_content tail new_node_children
  | AtomicValue a :: tail -> 
      insert_content tail node_children
        (* AtomicValues should be converted to text nodes ? - Gargi *)

(* auxilliary primitives for insert-into-before *)

let rec insert_before_sibling node node_children before_node = 
  match node_children with
  | [] -> [node]
  | head :: tail ->
      if (nodeEqual head before_node) then
	node :: node_children
      else
	head :: (insert_before_sibling node tail before_node)

let rec insert_content_before item_list node_children before_node = 
  match item_list with
  | [] -> node_children
  | Node n :: tail ->
      (* Note:
 	  Creates a clean node here. This creates node ids to make
 	  sure the nodes are kept there after inserted. However, the
 	  correspondind node-id are going to be wrong. This will have
 	  to be fixed.
        - Jerome
       *)
      let n' = nodeDeepCopy n in
      let new_node_children = insert_before_sibling n' node_children before_node in
      insert_content_before tail new_node_children before_node
  | AtomicValue a :: tail ->
      insert_content_before tail node_children before_node
	(* AtomicValues should be converted to text nodes ? - Gargi *)

(* insert primitive *)

let insert v p bn =                                  
  let update_content = list_of_sequence v in
  let node_children = list_of_sequence (child_axis p) in 
  match bn with 
  | None -> 
      let new_node_children = sequence_of_list (insert_content update_content node_children) in
      begin 
	match p with 
	| DocumentNode n -> n.node_content.doc_children <- new_node_children;
	    update_parents p new_node_children
	| ElementNode n -> n.node_content.elem_children <- new_node_children;
	    update_parents p new_node_children
	| _ ->
	    raise (Query (Datamodel "TargetNode for Insert into not a Document or Element node"))
      end
       
  | Some n -> 
      let new_node_children = sequence_of_list (insert_content_before update_content node_children n) in 
      begin 
	match p with 
	| DocumentNode n -> n.node_content.doc_children <- new_node_children;
	    update_parents p new_node_children
	| ElementNode n -> n.node_content.elem_children <- new_node_children;
	    update_parents p new_node_children
	| _ ->
	    raise (Query (Datamodel "The parent of TargetNode for Insert into not a Document or Element node"))
      end


(***********)
(* Replace *)
(***********)

(* Auxilliary primitives for replace *)

(* rep_content inserts the item sequence in 'new_content' at the beginning of 'node_list' *)

let rec rep_content new_content node_list =
  match new_content with 
   | [] -> node_list
   | Node n :: tail ->
       let n' = nodeDeepCopy n in 
       n' :: (rep_content tail node_list) 
   | AtomicValue a :: tail ->                         (* Should the AtomicValue be converted into a text node ? - gargi *)
       rep_content tail node_list

(* replace_content replaces 'node' with 'new_content' in the 'node_list' or raises an error if 'node' is not found *)

let rec replace_content node new_content node_list = 
  match node_list with
  | [] ->
      raise (Query (Datamodel "Replace operation: Targetnode for replace does not exist!"))
  | head :: node_list' ->
      if (nodeEqual head node) then
	rep_content new_content node_list'
      else
	head :: (replace_content node new_content node_list')

(* replace primitive - to replace an existing node *)

let replace_node v n p = 
  match n with
  | AttributeNode n ->
      begin
	match p with         (* Attribute node will have only element node as parent ? - Gargi *)
	| ElementNode parent_node ->
	    let node_attributes = elemAttributes parent_node in
	    let node_attributes' = list_of_sequence (sequenceMap buildAttributeNode node_attributes) in
	    let new_node_attributes = replace_content (buildAttributeNode n) (list_of_sequence v) node_attributes' in
	    let new_node_attributes' = list_of_sequence (selectAttributeNodes (sequence_of_list new_node_attributes)) in 
	    parent_node.node_content.elem_attributes <- new_node_attributes'; 
	    update_parents p new_node_attributes
	| _ ->
	    raise (Query (Datamodel "Replace operation: Parent of attribute node is not an element node"))
      end 
  | _ ->
      let node_children = list_of_sequence (child_axis p) in
      let new_node_children = sequence_of_list (replace_content n (list_of_sequence v) node_children) in
      begin
	match p with
	| DocumentNode doc_node -> 
	    doc_node.node_content.doc_children <- new_node_children;
	    update_parents p new_node_children
	| ElementNode elem_node -> 
	    elem_node.node_content.elem_children <- new_node_children;
	    update_parents p new_node_children
	| _ ->
	    raise (Query (Datamodel "Replace operation: Parent of targetnode is not a Document or Element node"))
      end
    
(* replace primitive - to replace the value of an existing node *)

let replace_value v n = 
  match n with
  | DocumentNode d ->
      raise (Query (Datamodel "Replace operation: Cannot replace value of text node!"))
  | ElementNode e ->
      let (_, avs) = filterValue v in 
      let text = value_to_untyped v in
      e.node_content.elem_children <- [buildTextNode(textNode text)];
      e.node_content.elem_value <- avs
  | AttributeNode a ->
      let (_, avs) = filterValue v in 
      let ac = erase_simple_value avs in
      a.node_content.attr_content <- ac;
      a.node_content.attr_value <- avs
  | TextNode t->
      raise (Query (Datamodel "Replace operation: Cannot replace value of text node!"))
  | ProcessingInstructionNode p ->
      raise (Query (Datamodel "Replace operation: Cannot replace value of PI  node!"))
  | CommentNode c ->
      raise (Query (Datamodel "Replace operation: Cannot replace value of comment node!"))
  | NamespaceNode nsp ->
      raise (Query (Datamodel "Replace operation: Cannot replace value of namespace node!"))

