(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: load.ml,v 1.40 2003/06/09 15:28:26 simeon Exp $ *)

(* Module: Load
   Description:
     This module implements loading of a data model instance from an
     XML stream.
*)

open Error

open Namespace

open Sax_types
open Sax_typesutil

open Datamodel


(****************************)
(* Builds an attribute node *)
(****************************)

let load_sax_xml_attribute load_context a =
  let nsenv = Load_context.get_nsenv load_context in
  match a with
  | (attr_uqname, t) ->
      let rattr_sym =
	let (prefix,attr_rqname) = Resolve.resolve_attribute_qname nsenv attr_uqname in
	(Sym.rattr_symbol (prefix,attr_rqname))
      in
      let new_nodeid = Load_context.new_leaf_nodeid load_context in
      attrNodeId new_nodeid rattr_sym t


(******************************)
(* Builds a sequence of nodes *)
(******************************)

let rec load_xml_forest_from_sax load_context doc_stream =
  (* Process the next event *)
  let xml_event = Stream.next doc_stream in
  match xml_event with
  | SAX_startDocument _ ->
      raise (Query (Internal_Error ("Should not have a start document event")))

  | SAX_endDocument ->
      sequenceEmpty    (* Terminates recursion for the children of an element.
		          This gives the hand back to the calling element node creation. *)

  | SAX_startElement (elem_uqname, att_el, has_element_content) ->

        (* First things first: get the prefix for the node id *)
      let prefix = Load_context.new_prefix load_context in

        (* Then, extract the namespace attributes,
	   update the namespace environment accordingly,
	   and create local namespace nodes ?? *)
      let (ws_mode, new_nss, other_atts) = Sax_typesutil.extract_special_attributes att_el in
      let load_context' = Load_context.add_ns_bindings load_context new_nss in
      let nsenv' = Load_context.get_nsenv load_context' in

      let ns_children = sequenceEmpty (* THIS WILL NEED TO BE FIXED - Jerome *) in

        (* Then resolve the element QName wrt namespaces *)
      let relem_sym =
	let (prefix,element_rqname) = Resolve.resolve_element_qname nsenv' elem_uqname in
	(Sym.relem_symbol (prefix,element_rqname))
      in

        (* Then process children and attributes *)
        (* NOTE: attributes *must be* processed first for the node ids to be correct !! *)
      let attr_children = sequence_of_list (List.map (load_sax_xml_attribute load_context') other_atts) in
      let elem_children = load_xml_forest_from_sax load_context' doc_stream in

        (* Finally get the postfix and the node id before building the new element node *)
      let postfix = Load_context.new_postfix load_context' in
      let new_nodeid = Load_context.build_nodeid load_context' prefix postfix in

      let new_elem_node =
	buildElementNode (elemNodeId
			    new_nodeid
			    relem_sym
			    ns_children
			    attr_children
			    elem_children
			    [])
      in
      sequenceCons new_elem_node (load_xml_forest_from_sax load_context doc_stream)

  | SAX_endElement ->
      sequenceEmpty    (* Terminates recursion for the children of an element.
		         This gives the hand back to the calling element node creation. *)

  | SAX_processingInstruction (target,content) ->
      let new_nodeid = Load_context.new_leaf_nodeid load_context in
      let new_pi_node = buildProcessingInstructionNode (piNodeId new_nodeid target content) in
      sequenceCons new_pi_node (load_xml_forest_from_sax load_context doc_stream)

  | SAX_comment c ->
      let new_nodeid = Load_context.new_leaf_nodeid load_context in
      let new_comment_node = buildCommentNode (commentNodeId new_nodeid c) in
      sequenceCons new_comment_node (load_xml_forest_from_sax load_context doc_stream)

  | SAX_characters s ->
      let new_nodeid = Load_context.new_leaf_nodeid load_context in
      let new_text_node = buildTextNode (textNodeId new_nodeid s) in
      sequenceCons new_text_node (load_xml_forest_from_sax load_context doc_stream)

  | SAX_attribute a ->
      let new_attr_node =
	(AttributeNode (load_sax_xml_attribute load_context a))
      in
      sequenceCons new_attr_node (load_xml_forest_from_sax load_context doc_stream)

  | SAX_atomicValue av ->
      (* Does not support this yet *)
      raise (Query (Internal_Error "Does not support data model loading of non-node streams"))

(*********************************)
(* Load the document node itself *)
(*********************************)

let load_xml_document_from_sax root_uri doc_stream =
  let load_context = Load_context.build_load_context root_uri in

  (* Compute the prefix *)
  let prefix = Load_context.new_prefix load_context in 

  (* The first event should be a start document. *)
  let xml_decl = get_start_document doc_stream in

  let document_encoding =
    match xml_decl with
    | None -> None
    | Some (_,enc,_) -> enc
  in

  (* Then process the rest of the document. *)
  let full_root_forest = load_xml_forest_from_sax load_context doc_stream in

  (* Build the postfix. *)
  let postfix = Load_context.new_postfix load_context in 

  (* Creates the node id for that node. *)
  let new_nodeid = Load_context.build_nodeid load_context prefix postfix in

  (* Then the node itself. *)
  buildNodeSequence
    (sequenceBuildSingle
       (buildDocumentNode
	  (docNodeId new_nodeid root_uri full_root_forest document_encoding)))

