(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: nodeid.ml,v 1.11 2003/05/21 15:03:49 simeon Exp $ *)

open Gmisc
open Error


(*****************)
(* Id structures *)
(*****************)

(* Document identifier *)

type docid = int

(* document id's counter *)

let current_doc_id = Id.create 0

(* Prefix and postfix *)

type prefix = int
type postfix = int

(* The node id - unique for all nodes in all documents or fragments *)

type nodeid = (docid * prefix * postfix)

(* Node id Generator *)

(* Note:
     The following is used when doing the copy.
   - Jerome & Mary
*)

type nodeid_gen = (docid * Id.id_gen * Id.id_gen)

(****************)
(* External ids *)
(****************)

(* Note:
      The following is support for (docid,ID) <--> nodeid mappings.
   - J&M
*)

type extid = (docid * Datatypes.xs_string)

(* ID pool *)

let id_tbl : (extid, nodeid) Hashtbl.t =
  Hashtbl.create 167

(* Inverted ID pool *)

let inv_id_tbl : (nodeid, extid) Hashtbl.t =
  Hashtbl.create 167

(* Clear the id tables *)

let clear_id_tables () =
  Id.init current_doc_id 0;
  Hashtbl.clear id_tbl;
  Hashtbl.clear inv_id_tbl

(*********************************)
(* Document and node id creation *)
(*********************************)

(* create a new node id generator for a given document *)

let new_nodeid_gen () =
  (Id.next current_doc_id, Id.create 0, Id.create 0)

(* new prefix *)

let new_prefix (_,prefix_gen,_) =
  Id.next prefix_gen

(* new postfix *)

let new_postfix (_,_,postfix_gen) =
  Id.next postfix_gen

(* build node id *)

let build_nodeid (did,_,_) prefix postfix =
  (did,prefix,postfix)

(* build node id, with a given ID value *)

let build_nodeid_with_ID (did,_,_) prefix postfix _ID_value =
  let newextid = (did,_ID_value) in
  if (Hashtbl.mem id_tbl newextid)
  then
    raise (Query (Datamodel ("Node with ID : "
			     ^ (snd newextid)
			     ^ " already exists in document "
			     ^ (string_of_int (fst newextid)))))
  else
    let newnid = (did,prefix,postfix) in
    begin
      Hashtbl.add id_tbl newextid newnid;
      Hashtbl.add inv_id_tbl newnid newextid;
      newnid
    end

(* Convenience function to create ids for leaf nodes *)

let new_leaf_nodeid nid_gen =
  let prefix = new_prefix nid_gen in
  let postfix = new_postfix nid_gen in
  build_nodeid nid_gen prefix postfix


(**********************************)
(* Document and node id accessors *)
(**********************************)

(* get document id from doc_id_gen *)

let get_docid_from_nodeid_gen (did,_,_) =
  did

(* get document id from nodeid *)

let get_docid (did,_,_) =
  did

(* node equality *)

let node_is nid1 nid2 =
  nid1 = nid2

(* preceding *)

let node_precedes (did1,pref1,_) (did2,pref2,_) =
  if did1 = did2 then
    pref1 < pref2   
    (* Note: check if strict or not! - J&M *)
    (* Mary: yes, it's strictly < *)
  else
    did1 < did2

(* following *)

let node_follows (did1,pref1,_) (did2,pref2,_) =
  if did1 = did2 then
    pref1 > pref2   
    (* Note: check if strict or not! - J&M *)
    (* Mary: yes, it's strictly > *)
  else
    did1 > did2

(* node id compare function *)

(* Note:
     This is used for sorting by document order
   - Jerome *)

let node_compare (did1,pref1,_) (did2,pref2,_) =
  compare (did1,pref1) (did2,pref2)

(* accesses the node id from the document id and the IDREF *)

let node_lookup_from_IDREF docid _ID_value =
  let extid = (docid,_ID_value) in
  try
    Hashtbl.find id_tbl extid
  with
  | _ ->
      raise (Query (Datamodel ("No node with ID : "
			       ^ (snd extid)
			       ^ " in document "
			       ^ (string_of_int (fst extid)))))


(* Is there an external ID associated to that node ? *)

let lookup_external_identifier nid = 
  try
    Some (Hashtbl.find inv_id_tbl nid)
  with
  | _ ->
      None

let string_of_nodeid (docid,prefix,postfix) =
  "{ docid: "
  ^ (string_of_int docid)
  ^ "; prefix: "
  ^ (string_of_int prefix)
  ^ "; postfix: "
  ^ (string_of_int postfix)
  ^ " }"

