(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: validation-sax.ml,v 1.1 2002/05/29 03:53:23 simeon Exp $ *)

open Simple
open Xquery_ast
open Schema
open Schemautil
open Datamodel
open Error

(* Chose a step for a given token *)

let get_matching_elements frl symbol =
  let find_function (x,_,_,_) =
    Sym.subtag x (Sym.RUnitElemSymbol symbol)
  in
  List.find_all find_function frl

let get_matching_text frl =
  let find_function (x,_,_,_) =
    Sym.subtag x Sym.RUnitTextSymbol
  in
  List.find_all find_function frl

(*******************)
(* Text validation *) 
(*******************)

(* Get whole text from a content model *)

let gather_full_text_content model =
  let add_text t node =
    match node with
    | XElem _ ->
	t
    | XText t' ->
	t ^ t'
    | XPI _ ->
	t
    | XComment _ ->
	t
  in
  List.fold_left add_text "" model

(* Value extraction from text *)

let extract_value_from_text nse s t m =
  let bt = 
    let model_without_attributes =
      (apply_to_unit_types s only_non_attribute_model m)
    in
    let m' =
      Simplification.simplify_ty s (dereference_model s model_without_attributes) in
    match m'.pmodel_desc with
    | TBase bt -> 
	bt
    | _ ->
	raise (Query (Validation ("Text: " ^
				  t ^
				  " does not match a simple content model: " ^
				  (Print.bprintf_model "" m))))
  in
  let single_val =
    match bt with
    | STString ->
	DataString (string_of_text t)
    | STBoolean ->
	DataBoolean (boolean_of_text t)
    | STDecimal ->
	DataDecimal (decimal_of_text t)
    | STFloat ->
	DataFloat (float_of_text t)
    | STDouble ->
	DataDouble (double_of_text t)
    | STDuration ->
	DataDuration (duration_of_text t)
    | STDateTime ->
	DataDateTime (dateTime_of_text t)
    | STTime ->
	DataTime (time_of_text t)
    | STDate ->
	DataDate (date_of_text t)
    | STGYearMonth ->
	DataGYearMonth (gYearMonth_of_text t)
    | STGYear ->
	DataGYear (gYear_of_text t)
    | STGMonthDay ->
	DataGMonthDay (gMonthDay_of_text t)
    | STGDay ->
	DataGDay (gDay_of_text t)
    | STGMonth ->
	DataGMonth (gMonth_of_text t)
    | STHexBinary ->
	DataHexBinary (hexBinary_of_text t)
    | STBase64Binary ->
	DataBase64Binary (base64Binary_of_text t)
    | STAnyURI ->
	DataAnyURI (anyURI_of_text t)
    | STQName ->
	DataQName (qname_of_text nse t)
    | STNOTATION ->
	DataNOTATION (notation_of_text t)
    | STInteger ->
	DataInteger (integer_of_text t)
    | STInt ->
	DataInt (int_of_text t)
    | STUnknownSimpleType ->
	raise (Query (Internal_Error "Cannot extract value: UnknownSimpleType"))
  in
  oforestBuildSingle single_val

(* Text to value validation *)

let validate_text_value nse nid_gen schema text model =
  let va = extract_value_from_text nse schema text model in
  (text, va)

(* Pure text node validation without content model processing *)

let validate_text_node_skip nse nid_gen schema get_next_node_fun model text_desc doc_stream =
  let new_nodeid = Nodeid.new_leaf_nodeid nid_gen in   (* Node id creation! *)
  (buildTextNode (textNodeId new_nodeid text_desc), model)

(* Pure text node validation with content model processing *)

let validate_text_node nse nid_gen schema get_next_node_fun model text_desc doc_stream =
  let frl = first_rest_with_model_attributes schema model in
  let matching_frl = get_matching_text frl in
  let mo = 
    match matching_frl with
    | [] ->
	raise (Query (Validation ("No matching type for text node \"" ^ t ^ "\"")))
    | [(_, _, _, next_model)] ->
	next_model
    | _ ->
	raise (Query (Validation ("Non deterministic content model:"
				  ^ (Print.bprintf_model "" model))))
  in
  let new_nodeid = Nodeid.new_leaf_nodeid nid_gen in   (* Node id creation! *)
  (buildTextNode (textNodeId new_nodeid text_desc), mo)

(*****************)
(* PI validation *)
(*****************)

let validate_pi_node nse nid_gen schema get_next_node_fun model pi_desc doc_stream =
  (* Note:
       PIs are not impacting validation. I.e., they always validate
       and they do not change the current model.
     - Jerome
   *)
  let (target,content) = pi_desc in
  let new_nodeid = Nodeid.new_leaf_nodeid nid_gen in   (* Node id creation! *)
  (buildPINode (piNodeId new_nodeid target content),model)

(**********************)
(* Comment validation *)
(**********************)

let validate_comment_node nse nid_gen schema get_next_node_fun model comment_desc doc_stream =
  (* Note:
       Comments are not impacting validation. I.e., they always
       validate and they do not change the current model.
     - Jerome
   *)
  let new_nodeid = Nodeid.new_leaf_nodeid nid_gen in   (* Node id creation! *)
  (buildCommentNode (commentNodeId new_nodeid comment_desc),model)

(************************)
(* Attribute validation *)
(************************)

(* Matching for required attributes *)

let match_required_attribute ram attributes =
  match get_unit_tag_from_model ram with
  | Sym.RUnitAttrSymbol rattr_sym1 ->
      begin
	let find_fun (rattr_sym2,_) = Sym.rattr_subtag rattr_sym2 rattr_sym1 in
	let (all_matches,all_non_matches) = List.partition find_fun attributes in
	match all_matches with
	| [] ->
	    raise (Query (Validation ("No matching attribute for required attribute: " ^
				      (Sym.rattr_prefix_string rattr_sym1))))
	| [(rattr_sym,rattr_text)] ->
	    (rattr_sym,rattr_text,ram,all_non_matches)
	| _ ->
	    raise (Query (Validation ("Duplicate attribute: " ^
				      (Sym.rattr_prefix_string rattr_sym1))))
      end
  | _ ->
      raise (Query (Internal_Error ("Unresolved attribute name")))

let rec match_all_required_attributes schema rams attributes matches =
  match rams with
  | [] ->
      (matches, attributes)
  | (ram,_) :: rams' ->
      let (rattr_sym,rattr_text,ram,all_non_matches) = match_required_attribute ram attributes in
      let attribute_model = get_attribute_content_model_kind schema (snd (access_attribute schema ram)) in
      let this_match = (rattr_sym, rattr_text, attribute_model) in
      match_all_required_attributes schema rams' all_non_matches (this_match :: matches)

(* Matching for optional attributes *)

let match_optional_attribute oam attributes =
  match get_unit_tag_from_model oam with
  | Sym.RUnitAttrSymbol rattr_sym1 ->
      begin
	let find_fun (rattr_sym2,_) = Sym.rattr_subtag rattr_sym2 rattr_sym1 in
	let (all_matches,all_non_matches) = List.partition find_fun attributes in
	match all_matches with
	| [] ->
	    None
	| [(rattr_sym,rattr_text)] ->
	    Some (rattr_sym,rattr_text,oam,all_non_matches)
	| _ ->
	    raise (Query (Validation ("Duplicate attribute: " ^
				      (Sym.rattr_prefix_string rattr_sym1))))
      end
  | _ ->
      raise (Query (Internal_Error ("Unresolved attribute name")))

let rec match_all_optional_attributes schema oams attributes matches =
  match oams with
  | [] ->
      (matches,attributes)
  | (oam,_) :: oams' ->
      let result = match_optional_attribute oam attributes in
      match result with
      |	None ->
	  match_all_optional_attributes schema oams' attributes matches
      |	Some (rattr_sym,rattr_text,oam,all_non_matches) ->
	  let attribute_model = get_attribute_content_model_kind schema (snd (access_attribute schema oam)) in
	  let this_match = (rattr_sym, rattr_text, attribute_model) in
	  match_all_optional_attributes schema oams' all_non_matches (this_match :: matches)

(* Validates a single attribute *)

let validate_one_attribute nse nid_gen schema (rattr_sym, rattr_text, attribute_model) =
  match attribute_model with
  | AttributeSimpleContent m ->
      let (attribute_text, attribute_value) = validate_text_value nse nid_gen schema rattr_text m in
      let new_nodeid = Nodeid.new_leaf_nodeid nid_gen in   (* Node id creation! *)
      attrNodeWithValueId new_nodeid rattr_sym attribute_text attribute_value

(* Attribute validation main function *)

let rec validate_attribute_content nse nid_gen schema get_next_node_fun m attributes =
  let (oams,rams) = process_attribute_model schema m in

  (* First steps within each attribute to resolve namespaces *)
  let rattributes =
    List.map (step_within_xml_attribute nse) attributes
  in
  let (opt_attribute_model_matching, rest) =
    (match_all_optional_attributes schema oams rattributes [])
  in
  let (req_attribute_model_matching, rest') =
    (match_all_required_attributes schema rams rest [])
  in
  let all_matches = opt_attribute_model_matching @ req_attribute_model_matching in
  match rest' with
  | [] ->
      oforest_of_list (List.map (validate_one_attribute nse nid_gen schema) all_matches)
  | _ ->
      raise (Query (Validation "Some unmatched attribute in attribute model"))
  
(**********************)
(* Element validation *)
(**********************)

(* Element main validation function *)

(* Note:
     This function is mutually recursive with those doing validation
     of the content model below.
   - Jerome
 *)

let rec validate_element_node nse nid_gen schema get_next_node_fun model element_desc doc_stream =
  (* First things first: get the prefix for the node id *)
  let prefix = Nodeid.new_prefix nid_gen in

  let (element_tag, element_attributes) = step_within_xml_element nse element_desc in
  let frl = first_rest_with_model_attributes schema model in
  let matching_frl = get_matching_elements frl element_tag in
  let (en,mo) = 
    match matching_frl with
    | [] ->
	raise (Query (Validation ("No matching type for element: "
				  ^ (Sym.relem_prefix_string element_tag))))
    | [(_, Some unit_model, _, next_model)] ->
	begin
	  let element_node =
	    match get_element_content_model_kind schema unit_model with
	    | ElementEmptyContent m ->
		if element_children = []
		then
		  let attribute_forest =
		    validate_attribute_content nse nid_gen schema get_next_node_fun m element_attributes in

        (* Finally get the postfix and the node id before building the new element node *)
		  let postfix = Nodeid.new_postfix nid_gen in
		  let new_nodeid = Nodeid.build_nodeid nid_gen prefix postfix in

		  elemNodeId new_nodeid element_tag oforestEmpty attribute_forest oforestEmpty []
		else
		  raise (Query (Validation ("Non empty element: " ^
					    (Sym.relem_prefix_string element_tag) ^
					    " matched against empty element model")))
	    | ElementSimpleContent m ->
                (* First get the value of the node *)
		let whole_text = gather_full_text_content element_children in
		let children_value_forest = extract_value_from_text nse schema whole_text m in
		(* Then validate the content nodes *)
		let attribute_forest =
		  validate_attribute_content nse nid_gen schema get_next_node_fun m element_attributes
		in
		let children_forest =
		  validate_content_value nse nid_gen schema get_next_node_fun m element_children
		in
        (* Finally get the postfix and the node id before building the new element node *)
		let postfix = Nodeid.new_postfix nid_gen in
		let new_nodeid = Nodeid.build_nodeid nid_gen prefix postfix in
		
		elemNodeWithValueId
		  new_nodeid
		  element_tag
		  oforestEmpty
		  attribute_forest
		  children_forest
		  []
		  children_value_forest
	    | ElementComplexContent m ->
		let attribute_forest =
		  validate_attribute_content nse nid_gen schema get_next_node_fun m element_attributes in
		let children_forest =
		  validate_content_complex nse nid_gen schema get_next_node_fun m element_children in
        (* Finally get the postfix and the node id before building the new element node *)
		let postfix = Nodeid.new_postfix nid_gen in
		let new_nodeid = Nodeid.build_nodeid nid_gen prefix postfix in
		elemNodeId new_nodeid element_tag oforestEmpty attribute_forest children_forest []
	    | ElementMixedContent m ->
		let attribute_forest =
		  validate_attribute_content nse nid_gen schema get_next_node_fun m element_attributes in
		let children_forest =
		  validate_content_mixed nse nid_gen schema get_next_node_fun m element_children in
        (* Finally get the postfix and the node id before building the new element node *)
		let postfix = Nodeid.new_postfix nid_gen in
		let new_nodeid = Nodeid.build_nodeid nid_gen prefix postfix in
		elemNodeId new_nodeid element_tag oforestEmpty attribute_forest children_forest []
	  in
	  (element_node, next_model)
	end
    | _ ->
	raise (Query (Validation ("Non deterministic content model for element: "
				  ^ (Sym.relem_prefix_string element_tag))))
  in
  (buildElemNode en, mo)

(****************************)
(* Content model validation *)
(****************************)

(* Note:
     Here is the forest-walking part of the algorithm.
       - [nse] is the namespace environment
       - [schema] is the schema
       - [get_next_node_fun] is the function that returns the next
         node to process in the AST.
       - [model] is the model against which to validate the current
         node. This model should not be a group reference
       - [content] is the forest to be validated
   - Jerome
 *)

(* Empty element content *)

and validate_content_empty nse nid_gen schema get_next_node_fun model doc_stream =
  (* Is the schema allowing the end of the element content at this point ? *)
  if empty schema
      (apply_to_unit_types schema only_non_attribute_model model)
  (* If yes, build an empty forest *)
  then
    oforestEmpty
  (* Otherwise this is a validation error *)
  else
    raise (Query (Validation ("Reached end of element content on non-empty type: " ^ 
			      (Print.bprintf_model "" model))))

(* Empty document content *)

and validate_document_content_empty nse nid_gen schema get_next_node_fun model doc_stream =
  (* Is the schema allowing the end of the document content at this point ? *)
  if empty schema
      (apply_to_unit_types schema only_non_attribute_model model)
  (* If yes, build an empty forest *)
  then
    oforestEmpty
  (* Otherwise this is a validation error *)
  else
    raise (Query (Validation ("Reached end of document content on non-empty type: " ^ 
			      (Print.bprintf_model "" model))))

(* Mixed content *)

and validate_content_mixed nse nid_gen schema get_next_node_fun model doc_stream =
  (* Makes one step in the content of the element *)
  let first_result =
    match get_next_node_fun doc_stream with
    (* We reached the end of the document content *)
    | SAX_endDocument ->
	validate_document_content_empty nse nid_gen schema get_next_node_fun model doc_stream

    (* We reached the end of the element content *)
    | SAX_endElement ->
	validate_content_empty nse nid_gen schema get_next_node_fun model doc_stream

    (* We found a new element node *)
    | SAX_startElement element_desc ->
	validate_element_node nse nid_gen schema get_next_node_fun model element_desc doc_stream

    (* We found a new text node *)
    | SAX_characters text_desc ->
	validate_text_node nse nid_gen schema get_next_node_fun model text_desc doc_stream

    (* We found a new PI node *)
    | SAX_processingInstruction pi_desc ->
	validate_pi_node nse nid_gen schema get_next_node_fun model pi_desc doc_stream

    (* We found a new comment node *)
    | SAX_comment comment_desc ->
	validate_comment_node nse nid_gen schema get_next_node_fun model comment_desc doc_stream
  in
  (* Validates the rest of the mixed content *)
  let rest_result =
    validate_content_mixed nse nid_gen schema get_next_node_fun rest_model doc_stream
  in
  (* Puts the total result back together *)
  oforestCons first_result node_result

(* Complex content *)

and validate_content_complex nse nid_gen schema get_next_node_fun model doc_stream =
  (* Makes one step in the content of the element *)
  let first_result =
    match get_next_node_fun doc_stream with
    (* We reached the end of the document content *)
    | SAX_endDocument ->
	validate_document_content_empty nse nid_gen schema get_next_node_fun model doc_stream

    (* We reached the end of the element content *)
    | SAX_endElement ->
	validate_content_empty nse nid_gen schema get_next_node_fun model doc_stream

    (* We found a new element node *)
    | SAX_startElement element_desc ->
	validate_element_node nse nid_gen schema get_next_node_fun model element_desc doc_stream

    (* We found a new text node *)
    | SAX_characters _ ->
	raise (Query (Validation ("Text in a non-mixed content element of type: " ^ 
				  (Print.bprintf_model "" model))))

    (* We found a new PI node *)
    | SAX_processingInstruction pi_desc ->
	validate_pi_node nse nid_gen schema get_next_node_fun model pi_desc doc_stream

    (* We found a new comment node *)
    | SAX_comment comment_desc ->
	validate_comment_node nse nid_gen schema get_next_node_fun model comment_desc doc_stream
  in
  (* Validates the rest of the mixed content *)
  let rest_result =
    validate_content_complex nse nid_gen schema get_next_node_fun rest_model doc_stream
  in
  (* Puts the total result back together *)
  oforestCons first_result rest_result

(* Value content *)

and validate_content_value nse nid_gen schema get_next_node_fun model doc_stream =
  (* Makes one step in the content of the element *)
  let first_result =
    match get_next_node_fun doc_stream with
    (* We reached the end of the document content *)
    | SAX_endDocument ->
	oforestEmpty

    (* We reached the end of the element content *)
    | SAX_endElement ->
	oforestEmpty

    (* We found a new element node *)
    | SAX_startElement element_desc ->
	(* Note:
  	     Element in a value content element is not allowed.
	   - Jerome
	 *)
	raise (Query (Validation ("Element in a value content element of type: " ^ 
				  (Print.bprintf_model "" model))))
    (* We found a new text node *)
    | SAX_characters _ ->
	validate_text_node_skip nse nid_gen schema get_next_node_fun model text_desc doc_stream

    (* We found a new PI node *)
    | SAX_processingInstruction pi_desc ->
	validate_pi_node nse nid_gen schema get_next_node_fun model pi_desc doc_stream

    (* We found a new comment node *)
    | SAX_comment comment_desc ->
	validate_comment_node nse nid_gen schema get_next_node_fun model comment_desc doc_stream
  in
  (* Validates the rest of the mixed content *)
  let rest_result =
    validate_content_value nse nid_gen schema get_next_node_fun rest_model doc_stream
  in
  (* Puts the total result back together *)
  oforestCons first_result rest_result

(****************************)
(* Document node validation *)
(****************************)


let validate_document_node nse nid_gen schema get_next_node_fun doc_stream root_uri =
  (* Gets the schema root *)
  let schema_root =
    match get_root schema with
    | None ->
	raise (Query (Validation "Cannot validate against a schema without a root"))
    | Some r ->
	r
  in
  let root_model = Xquery_astutil.fmkmodel (TElemRef (Sym.RElemSym schema_root)) Finfo.bogus in

  (* Now starting the processing with the root node *)
  begin
    (* Opening the document node *)
    begin
      match get_next_node_fun doc_stream with
      | SAX_startDocument xml_decl ->
	    ()
      | _ ->
	  raise (Query (Validation "Cannot validate a document without a root"))
    end;
    (* Processing the content of the document node *)
    let result =
      validate_content_complex
	nse nid_gen schema get_next_node_fun root_model doc_stream
    in
    (* Building the resulting document node *)
    buildDocNode (docNode root_uri result)
  end


(*********************)
(* Schema validation *)
(*********************)

(* Note:
     Validation takes a schema and a well-formed document (in the form
     of an XML AST) and returns a validated instance of the XQuery
     data model.
   - Jerome
 *)

let validate_document nse schema root_uri doc_stream =
  (* Initialization of the node id generation for that particular document *)
  let nid_gen = Nodeid.new_nodeid_gen () in

  (* Which function should be used to process nodes ? *)
  let get_next_node_fun =
    if !Conf.whitespace
    then
      Sax.next_event_remove_whitespace
    else
      Sax.next_event
  in
      (* Note:
           Paradoxaly, if we preserve whitespace, then the content of
           non-mixed elements still might have text nodes with
           whitespace. Therefore we want to remove them at validation
           time. If we do not preserve whitespace, then the AST is cleaned
           up at lexing time, and we can use the more efficient
           get_next_node function.
         - Jerome
       *)
  let root_node = validate_document_node nse nid_gen schema get_next_node_fun doc_stream root_uri in
  buildNodeInstanceForest (oforestBuildSingle root_node)

