(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: validation.ml,v 1.25 2003/05/20 15:29:50 mff Exp $ *)

open Datatypes
open Xml_ast
open Xml_astutil
open Xquery_ast
open Schema
open Schemautil
open Datamodel
open Error

(* Chose a step for a given token *)

let get_matching_elements frl symbol =
  let find_function (x,_,_,_) =
    Sym.subtag x (Sym.RUnitElemSymbol symbol)
  in
  List.find_all find_function frl

let get_matching_text frl =
  let find_function (x,_,_,_) =
    Sym.subtag x Sym.RUnitTextSymbol
  in
  List.find_all find_function frl

(*******************)
(* Text validation *) 
(*******************)

(* Get whole text from a content model *)

let gather_full_text_content model =
  let add_text t node =
    match node with
    | XElem _ ->
	t
    | XText t' ->
	t ^ t'
    | XPI _ ->
	t
    | XComment _ ->
	t
  in
  List.fold_left add_text "" model

(* Value extraction from text *)

let extract_value_from_text nse s t m =
  let bt = 
    let model_without_attributes =
      (apply_to_unit_types s only_non_attribute_model m)
    in
    let m' =
      Simplification.simplify_ty s (dereference_model s model_without_attributes) in
    match m'.pmodel_desc with
    | TBase bt -> 
	bt
    | _ ->
	raise (Query (Validation ("Text: " ^
				  t ^
				  " does not match a simple content model: " ^
				  (Print.bprintf_model "" m))))
  in
  let single_val =
    match bt with
    | ATString ->
	AtomicString (string_of_untyped t)
    | ATBoolean ->
	AtomicBoolean (boolean_of_untyped t)
    | ATDecimal ->
	AtomicDecimal (decimal_of_untyped t)
    | ATFloat ->
	AtomicFloat (float_of_untyped t)
    | ATDouble ->
	AtomicDouble (double_of_untyped t)
    | ATDuration ->
	AtomicDuration (duration_of_untyped t)
    | ATDateTime ->
	AtomicDateTime (dateTime_of_untyped t)
    | ATTime ->
	AtomicTime (time_of_untyped t)
    | ATDate ->
	AtomicDate (date_of_untyped t)
    | ATGYearMonth ->
	AtomicGYearMonth (gYearMonth_of_untyped t)
    | ATGYear ->
	AtomicGYear (gYear_of_untyped t)
    | ATGMonthDay ->
	AtomicGMonthDay (gMonthDay_of_untyped t)
    | ATGDay ->
	AtomicGDay (gDay_of_untyped t)
    | ATGMonth ->
	AtomicGMonth (gMonth_of_untyped t)
    | ATHexBinary ->
	AtomicHexBinary (hexBinary_of_untyped t)
    | ATBase64Binary ->
	AtomicBase64Binary (base64Binary_of_untyped t)
    | ATAnyURI ->
	AtomicAnyURI (anyURI_of_untyped t)
    | ATQName ->
	AtomicQName (qname_of_untyped nse t)
    | ATNOTATION ->
	AtomicNOTATION (notation_of_untyped t)
    | ATInteger ->
	AtomicInteger (integer_of_untyped t)
    | ATInt ->
	AtomicInt (int_of_untyped t)
    | ATYearMonthDuration ->
	AtomicYearMonthDuration (yearMonthDuration_of_untyped t)
    | ATDayTimeDuration ->
	AtomicDayTimeDuration (dayTimeDuration_of_untyped t)
    | ATUntyped ->
	raise (Query (Internal_Error "Cannot extract value: untyped"))
  in
  sequenceBuildSingle single_val

(* Text to value validation *)

let validate_text_value nse nid_gen schema text model =
  let va = extract_value_from_text nse schema text model in
  (text, va)

(* Pure text node validation without content model processing *)

let validate_text_node_skip nse nid_gen schema wsmode model node =
  let t = step_within_xml_text nse node in
  let new_nodeid = Nodeid.new_leaf_nodeid nid_gen in   (* Node id creation! *)
  (buildTextNode (textNodeId new_nodeid t), model)

(* Pure text node validation with content model processing *)

let validate_text_node nse nid_gen schema wsmode model node =
  let t = step_within_xml_text nse node in
  let frl = first_rest_with_model_attributes schema model in
  let matching_frl = get_matching_text frl in
  let mo = 
    match matching_frl with
    | [] ->
	raise (Query (Validation ("No matching type for text node \"" ^ t ^ "\"")))
    | [(_, _, _, next_model)] ->
	next_model
    | _ ->
	raise (Query (Validation ("Non deterministic content model:"
				  ^ (Print.bprintf_model "" model))))
  in
  let new_nodeid = Nodeid.new_leaf_nodeid nid_gen in   (* Node id creation! *)
  (buildTextNode (textNodeId new_nodeid t), mo)

(*****************)
(* PI validation *)
(*****************)

let validate_pi_node nse nid_gen schema wsmode model node =
  (* Note:
       PIs are not impacting validation. I.e., they always validate
       and they do not change the current model.
     - Jerome
   *)
  let (target,content) = step_within_xml_pi nse node in
  let new_nodeid = Nodeid.new_leaf_nodeid nid_gen in   (* Node id creation! *)
  (buildProcessingInstructionNode (piNodeId new_nodeid target content),model)

(**********************)
(* Comment validation *)
(**********************)

let validate_comment_node nse nid_gen schema wsmode model node =
  (* Note:
       Comments are not impacting validation. I.e., they always
       validate and they do not change the current model.
     - Jerome
   *)
  let s = step_within_xml_comment nse node in
  let new_nodeid = Nodeid.new_leaf_nodeid nid_gen in   (* Node id creation! *)
  (buildCommentNode (commentNodeId new_nodeid s),model)

(************************)
(* Attribute validation *)
(************************)

(* Matching for required attributes *)

let match_required_attribute ram attributes =
  match get_unit_tag_from_model ram with
  | Sym.RUnitAttrSymbol rattr_sym1 ->
      begin
	let find_fun (rattr_sym2,_) = Sym.rattr_subtag rattr_sym2 rattr_sym1 in
	let (all_matches,all_non_matches) = List.partition find_fun attributes in
	match all_matches with
	| [] ->
	    raise (Query (Validation ("No matching attribute for required attribute: " ^
				      (Sym.rattr_prefix_string rattr_sym1))))
	| [(rattr_sym,rattr_text)] ->
	    (rattr_sym,rattr_text,ram,all_non_matches)
	| _ ->
	    raise (Query (Validation ("Duplicate attribute: " ^
				      (Sym.rattr_prefix_string rattr_sym1))))
      end
  | _ ->
      raise (Query (Internal_Error ("Unresolved attribute name")))

let rec match_all_required_attributes schema rams attributes matches =
  match rams with
  | [] ->
      (matches, attributes)
  | (ram,_) :: rams' ->
      let (rattr_sym,rattr_text,ram,all_non_matches) = match_required_attribute ram attributes in
      let attribute_model = get_attribute_content_model_kind schema (snd (access_attribute schema ram)) in
      let this_match = (rattr_sym, rattr_text, attribute_model) in
      match_all_required_attributes schema rams' all_non_matches (this_match :: matches)

(* Matching for optional attributes *)

let match_optional_attribute oam attributes =
  match get_unit_tag_from_model oam with
  | Sym.RUnitAttrSymbol rattr_sym1 ->
      begin
	let find_fun (rattr_sym2,_) = Sym.rattr_subtag rattr_sym2 rattr_sym1 in
	let (all_matches,all_non_matches) = List.partition find_fun attributes in
	match all_matches with
	| [] ->
	    None
	| [(rattr_sym,rattr_text)] ->
	    Some (rattr_sym,rattr_text,oam,all_non_matches)
	| _ ->
	    raise (Query (Validation ("Duplicate attribute: " ^
				      (Sym.rattr_prefix_string rattr_sym1))))
      end
  | _ ->
      raise (Query (Internal_Error ("Unresolved attribute name")))

let rec match_all_optional_attributes schema oams attributes matches =
  match oams with
  | [] ->
      (matches,attributes)
  | (oam,_) :: oams' ->
      let result = match_optional_attribute oam attributes in
      match result with
      |	None ->
	  match_all_optional_attributes schema oams' attributes matches
      |	Some (rattr_sym,rattr_text,oam,all_non_matches) ->
	  let attribute_model = get_attribute_content_model_kind schema (snd (access_attribute schema oam)) in
	  let this_match = (rattr_sym, rattr_text, attribute_model) in
	  match_all_optional_attributes schema oams' all_non_matches (this_match :: matches)

(* Validates a single attribute *)

let validate_one_attribute nse nid_gen schema (rattr_sym, rattr_text, attribute_model) =
  match attribute_model with
  | AttributeSimpleContent m ->
      let (attribute_text, attribute_value) = validate_text_value nse nid_gen schema rattr_text m in
      let new_nodeid = Nodeid.new_leaf_nodeid nid_gen in   (* Node id creation! *)
      attrNodeWithValueId new_nodeid rattr_sym attribute_text attribute_value

(* Attribute validation main function *)

let rec validate_attribute_content nse nid_gen schema wsmode m attributes =
  let (oams,rams) = process_attribute_model schema m in

  (* First steps within each attribute to resolve namespaces *)
  let rattributes =
    List.map (step_within_xml_attribute nse) attributes
  in
  let (opt_attribute_model_matching, rest) =
    (match_all_optional_attributes schema oams rattributes [])
  in
  let (req_attribute_model_matching, rest') =
    (match_all_required_attributes schema rams rest [])
  in
  let all_matches = opt_attribute_model_matching @ req_attribute_model_matching in
  match rest' with
  | [] ->
      sequence_of_list (List.map (validate_one_attribute nse nid_gen schema) all_matches)
  | _ ->
      raise (Query (Validation "Some unmatched attribute in attribute model"))
  
(**********************)
(* Element validation *)
(**********************)

(* Element main validation function *)

(* Note:
     This function is mutually recursive with those doing validation
     of the content model below.
   - Jerome
 *)

let rec validate_element_node nse nid_gen schema wsmode model element =
  (* First things first: get the prefix for the node id *)
  let prefix = Nodeid.new_prefix nid_gen in

  let (wsmode', new_nss, element_tag, element_attributes, element_children) = step_within_xml_element nse wsmode element in
  let nse' = Namespace.add_all_ns nse new_nss in

  let frl = first_rest_with_model_attributes schema model in
  let matching_frl = get_matching_elements frl element_tag in
  let (en,mo) = 
    match matching_frl with
    | [] ->
	raise (Query (Validation ("No matching type for element: "
				  ^ (Sym.relem_prefix_string element_tag))))
    | [(_, Some unit_model, _, next_model)] ->
	begin
	  let element_node =
	    match get_element_content_model_kind schema unit_model with
	    | ElementEmptyContent m ->
		if element_children = []
		then
		  let attribute_forest =
		    validate_attribute_content nse' nid_gen schema wsmode' m element_attributes in

        (* Finally get the postfix and the node id before building the new element node *)
		  let postfix = Nodeid.new_postfix nid_gen in
		  let new_nodeid = Nodeid.build_nodeid nid_gen prefix postfix in

		  elemNodeId new_nodeid element_tag sequenceEmpty attribute_forest sequenceEmpty []
		else
		  raise (Query (Validation ("Non empty element: " ^
					    (Sym.relem_prefix_string element_tag) ^
					    " matched against empty element model")))
	    | ElementSimpleContent m ->
                (* First get the value of the node *)
		let whole_text = gather_full_text_content element_children in
		let children_value_forest = extract_value_from_text nse' schema whole_text m in
		(* Then validate the content nodes *)
		let attribute_forest =
		  validate_attribute_content nse' nid_gen schema wsmode' m element_attributes
		in
		let children_forest =
		  validate_content_value nse' nid_gen schema wsmode' m element_children
		in
        (* Finally get the postfix and the node id before building the new element node *)
		let postfix = Nodeid.new_postfix nid_gen in
		let new_nodeid = Nodeid.build_nodeid nid_gen prefix postfix in
		
		elemNodeWithValueId
		  new_nodeid
		  element_tag
		  sequenceEmpty
		  attribute_forest
		  children_forest
		  []
		  children_value_forest
	    | ElementComplexContent m ->
		let attribute_forest =
		  validate_attribute_content nse' nid_gen schema wsmode' m element_attributes in
		let children_forest =
		  validate_content_complex nse' nid_gen schema wsmode' m element_children in
        (* Finally get the postfix and the node id before building the new element node *)
		let postfix = Nodeid.new_postfix nid_gen in
		let new_nodeid = Nodeid.build_nodeid nid_gen prefix postfix in
		elemNodeId new_nodeid element_tag sequenceEmpty attribute_forest children_forest []
	    | ElementMixedContent m ->
		let attribute_forest =
		  validate_attribute_content nse' nid_gen schema wsmode' m element_attributes in
		let children_forest =
		  validate_content_mixed nse' nid_gen schema wsmode' m element_children in
        (* Finally get the postfix and the node id before building the new element node *)
		let postfix = Nodeid.new_postfix nid_gen in
		let new_nodeid = Nodeid.build_nodeid nid_gen prefix postfix in
		elemNodeId new_nodeid element_tag sequenceEmpty attribute_forest children_forest []
	  in
	  (element_node, next_model)
	end
    | _ ->
	raise (Query (Validation ("Non deterministic content model for element: "
				  ^ (Sym.relem_prefix_string element_tag))))
  in
  (buildElementNode en, mo)

(****************************)
(* Content model validation *)
(****************************)

(* Note:
     Here is the forest-walking part of the algorithm.
       - [nse] is the namespace environment
       - [schema] is the schema
       - [wsmode] is the function that returns the next
         node to process in the AST.
       - [model] is the model against which to validate the current
         node. This model should not be a group reference
       - [content] is the forest to be validated
   - Jerome
 *)

(* Empty content *)

and validate_content_empty nse nid_gen schema wsmode model () =
  if empty schema
      (apply_to_unit_types schema only_non_attribute_model model)
  then
    sequenceEmpty
  else
    raise (Query (Validation ("Reached end of element content on non-empty type: " ^ 
			      (Print.bprintf_model "" model))))

(* Mixed content *)

and validate_content_mixed nse nid_gen schema wsmode model content =
  (* First one step in the content of the element *)
  let (onode,content') = get_next_xml_node wsmode content in
  match onode with
  | None ->
      (* We reached the end of the element content *)
      validate_content_empty nse nid_gen schema wsmode model ()
  | Some node ->
      (* Then one step in the content model / regular expression *)
        (* node_result: the result of the node validation
	   rest_model:  the remaining model with which to
	                validate the remaining forest *)
      let (node_result, rest_model) =
	match node with
	| XElem _ ->
	    validate_element_node nse nid_gen schema wsmode model node
	| XText _ ->
	    validate_text_node nse nid_gen schema wsmode model node
	| XPI _ ->
	    validate_pi_node nse nid_gen schema wsmode model node
	| XComment _ ->
	    validate_comment_node nse nid_gen schema wsmode model node
      in
      let result' = validate_content_mixed nse nid_gen schema wsmode rest_model content' in
      sequenceCons node_result result'

(* Complex content *)

and validate_content_complex nse nid_gen schema wsmode model content =
  (* First one step in the content of the element *)
  let (onode,content') = get_next_xml_node_no_whitespace content in
  match onode with
  | None ->
      (* We reached the end of the element content *)
      validate_content_empty nse nid_gen schema wsmode model ()

  | Some node ->
      (* Then one step in the content model / regular expression *)
        (* node_result: the result of the node validation
	   rest_model:  the remaining model with which to
	                validate the remaining forest *)
      let (node_result, rest_model) =
	match node with
	| XElem _ ->
	    validate_element_node nse nid_gen schema wsmode model node
	| XText _ ->
	    (* Note:
  	         Text in a non-mixed content element is not allowed.
	       - Jerome
	     *)
	    raise (Query (Validation ("Text in a non-mixed content element of type: " ^ 
				      (Print.bprintf_model "" model))))
	| XPI _ ->
	    validate_pi_node nse nid_gen schema wsmode model node
	| XComment _ ->
	    validate_comment_node nse nid_gen schema wsmode model node
      in
      let result' = validate_content_complex nse nid_gen schema wsmode rest_model content' in
      sequenceCons node_result result'

(* Value content *)

and validate_content_value nse nid_gen schema wsmode model content =
  (* One step in the content of the element *)
  let (onode,content') = get_next_xml_node wsmode content in
  match onode with
  | None ->
      (* We reached the end of the element content *)
      sequenceEmpty
  | Some node ->
      (* Then one step in the content model / regular expression *)
        (* node_result: the result of the node validation
	   rest_model:  the remaining model with which to
	                validate the remaining forest *)
      let (node_result, rest_model) =
	match node with
	| XElem _ ->
	    (* Note:
  	         Element in a value content element is not allowed.
	       - Jerome
	     *)
	    raise (Query (Validation ("Element in a value content element of type: " ^ 
				      (Print.bprintf_model "" model))))
	| XText _ ->
	    validate_text_node_skip nse nid_gen schema wsmode model node
	| XPI _ ->
	    validate_pi_node nse nid_gen schema wsmode model node
	| XComment _ ->
	    validate_comment_node nse nid_gen schema wsmode model node
      in
      let result' = validate_content_value nse nid_gen schema wsmode rest_model content' in
      sequenceCons node_result result'

(*********************)
(* Schema validation *)
(*********************)

(* Note:
     Validation takes a schema and a well-formed document (in the form
     of an XML AST) and returns a validated instance of the XQuery
     data model.
   - Jerome
 *)

let validate_document nse schema root_uri doc =
  let nid_gen = Nodeid.new_nodeid_gen () in  (* Used for node id generation *)

  let schema_root =                                (* Gets the schema root *)
    match get_root schema with
    | None ->
	raise (Query (Validation "Cannot validate against a schema without a root"))
    | Some r ->
	r
  in
  let root_node = doc.xml_root in              (* Gets the document root *)

  let document_encoding =
    match doc.xml_prolog.xml_decl with
    | None -> None
    | Some (_,enc,_) -> enc
  in

  let full_root_forest = doc.xml_prolog.xml_misc_prolog @ [root_node] @ doc.xml_misc in
  let root_model = Xquery_astutil.fmkmodel (TElemRef (Sym.RElemSym schema_root)) Finfo.bogus in
  let result =
    validate_content_complex
      nse nid_gen schema Whitespace.Default root_model full_root_forest
  in
  buildNodeSequence (sequenceBuildSingle (buildDocumentNode (docNode root_uri result document_encoding)))

