(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: datatypes.ml,v 1.6 2003/06/09 15:28:26 simeon Exp $ *)

(* This module provides support for atomic values and atomic types. *)

(* Note:
     This module needs to be extended to support XML Schema Datatype
     hierarchy
   - Jerome *)

open Namespace
open Sym

open Decimal
open AnyURI

open Error


(* This module provides support for atomic values and atomic types. *)

(*****************)
(* Atomic values *)
(*****************)

(* Note:
     Each of the following Caml types implements the *value space* of
     one the 19 XML Schema primitive atomic types.
   - Jerome
 *)

type xs_string       = string
type xs_boolean      = bool
type xs_decimal      = Decimal._decimal
type xs_float        = float
type xs_double       = float
type xs_duration     = string
type xs_dateTime     = DateTime._dateTime
type xs_time         = string
type xs_date         = string
type xs_gYearMonth   = string
type xs_gYear        = string
type xs_gMonthDay    = string
type xs_gDay         = string
type xs_gMonth       = string
type xs_hexBinary    = string
type xs_base64Binary = string
type xs_anyURI       = _uri            (* Defined in the AnyUri module *)
type xs_QName        = Sym.anon_symbol (* This is an anonymous QName - e.g, used as a data model value *)
type xs_NOTATION     = string

type xs_integer      = Decimal._integer
type xs_int          = Decimal._int


type xdt_untyped = string
type xdt_yearMonthDuration = string
type xdt_dayTimeDuration = string

(*****************)
(* Atomic values *)
(*****************)

type _AtomicValue =
  | AtomicString of xs_string
  | AtomicBoolean of xs_boolean
  | AtomicDecimal of xs_decimal
  | AtomicFloat of xs_float
  | AtomicDouble of xs_double
  | AtomicDuration of xs_duration
  | AtomicDateTime of xs_dateTime
  | AtomicTime of xs_time
  | AtomicDate of xs_date
  | AtomicGYearMonth of xs_gYearMonth
  | AtomicGYear of xs_gYear
  | AtomicGMonthDay of xs_gMonthDay
  | AtomicGDay of xs_gDay
  | AtomicGMonth of xs_gMonth
  | AtomicHexBinary of xs_hexBinary
  | AtomicBase64Binary of xs_base64Binary
  | AtomicAnyURI of xs_anyURI
  | AtomicQName of xs_QName
  | AtomicNOTATION of xs_NOTATION

      (* Values of derived built-in atomic types *) 
  | AtomicInteger of xs_integer
  | AtomicInt of xs_int

      (* Atomic types in xdt  *) 
  | AtomicYearMonthDuration of xdt_yearMonthDuration
  | AtomicDayTimeDuration of xdt_dayTimeDuration
  | AtomicUntyped of xdt_untyped

type _AtomicValueKind =
  | AtomicStringKind
  | AtomicBooleanKind
  | AtomicDecimalKind
  | AtomicFloatKind
  | AtomicDoubleKind
  | AtomicDurationKind
  | AtomicDateTimeKind
  | AtomicTimeKind
  | AtomicDateKind
  | AtomicGYearMonthKind
  | AtomicGYearKind
  | AtomicGMonthDayKind
  | AtomicGDayKind
  | AtomicGMonthKind
  | AtomicHexBinaryKind
  | AtomicBase64BinaryKind
  | AtomicAnyURIKind
  | AtomicQNameKind
  | AtomicNOTATIONKind
  | AtomicIntegerKind
  | AtomicIntKind
  | AtomicUntypedKind
  | AtomicYearMonthDurationKind
  | AtomicDayTimeDurationKind


(*****************)
(* Simple values *)
(*****************)

type _SimpleValue = _AtomicValue list


(**************************************)
(* Caml specific operations on values *)
(**************************************)

(* Note:
     The following functions are not part of the data model but are
     useful to manipulate the Caml representations of values.
   - Jerome
 *)

(* Casts from specific value types to a generic value *)

let buildAtomicString x =
  AtomicString x

let buildAtomicBoolean x =
  AtomicBoolean x

let buildAtomicDecimal x =
  AtomicDecimal x

let buildAtomicFloat x =
  AtomicFloat x

let buildAtomicDouble x =
  AtomicDouble x

let buildAtomicDuration x =
  AtomicDuration x

let buildAtomicDateTime x =
  AtomicDateTime x

let buildAtomicTime x =
  AtomicTime x

let buildAtomicDate x =
  AtomicDate x

let buildAtomicGYearMonth x =
  AtomicGYearMonth x

let buildAtomicGYear x =
  AtomicGYear x

let buildAtomicGMonthDay x =
  AtomicGMonthDay x

let buildAtomicGDay x =
  AtomicGDay x

let buildAtomicGMonth x =
  AtomicGMonth x

let buildAtomicHexBinary x =
  AtomicHexBinary x

let buildAtomicBase64Binary x =
  AtomicBase64Binary x

let buildAtomicAnyURI x =
  AtomicAnyURI x

let buildAtomicQName x =
  AtomicQName x

let buildAtomicNOTATION x =
  AtomicNOTATION x

let buildAtomicInteger x =
  AtomicInteger x

let buildAtomicInt x =
  AtomicInt x

let buildAtomicUntyped x =
  AtomicUntyped x

(* Casts from generic value to a specific value type *)

let getAtomicString x =
  match x with
  | AtomicString y -> y
  | _ -> raise (Query (Datamodel "Not a DataString"))

let getAtomicBoolean x =
  match x with
  | AtomicBoolean y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicBoolean"))

let getAtomicDecimal x =
  match x with
  | AtomicDecimal y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicDecimal"))

let getAtomicFloat x =
  match x with
  | AtomicFloat y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicFloat"))

let getAtomicDouble x =
  match x with
  | AtomicDouble y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicDouble"))

let getAtomicDuration x =
  match x with
  | AtomicDuration y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicDuration"))

let getAtomicDateTime x =
  match x with
  | AtomicDateTime y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicDateTime"))

let getAtomicTime x =
  match x with
  | AtomicTime y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicTime"))

let getAtomicDate x =
  match x with
  | AtomicDate y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicDate"))

let getAtomicGYearMonth x =
  match x with
  | AtomicGYearMonth y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicGYearMonth"))

let getAtomicGYear x =
  match x with
  | AtomicGYear y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicGYear"))

let getAtomicGMonthDay x =
  match x with
  | AtomicGMonthDay y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicGMonthDay"))

let getAtomicGDay x =
  match x with
  | AtomicGDay y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicGDay"))

let getAtomicGMonth x =
  match x with
  | AtomicGMonth y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicGMonth"))

let getAtomicHexBinary x =
  match x with
  | AtomicHexBinary y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicHexBinary"))

let getAtomicBase64Binary x =
  match x with
  | AtomicBase64Binary y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicBase64Binary"))

let getAtomicAnyURI x =
  match x with 
  | AtomicAnyURI y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicAnyURI"))

let getAtomicQName x =
  match x with
  | AtomicQName y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicQName"))

let getAtomicNOTATION x =
  match x with
  | AtomicNOTATION y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicNOTATION"))

let getAtomicInteger x =
  match x with
  | AtomicInteger y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicInteger"))

let getAtomicInt x =
  match x with
  | AtomicInt y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicInt"))

let getAtomicUntyped x =
  match x with
  | AtomicUntyped y -> y
  | _ -> raise (Query (Datamodel "Not a AtomicUntyped"))

(* Returns the kindness of a value *)

let getAtomicValueKind v =
  match v with
  | AtomicString _ ->
      AtomicStringKind
  | AtomicBoolean _ ->
      AtomicBooleanKind
  | AtomicDecimal _ ->
      AtomicDecimalKind
  | AtomicFloat _ ->
      AtomicFloatKind
  | AtomicDouble _ ->
      AtomicDoubleKind
  | AtomicDuration _ ->
      AtomicDurationKind
  | AtomicDateTime _ ->
      AtomicDateTimeKind
  | AtomicTime _ ->
      AtomicTimeKind
  | AtomicDate _ ->
      AtomicDateKind
  | AtomicGYearMonth _ ->
      AtomicGYearMonthKind
  | AtomicGYear _ ->
      AtomicGYearKind
  | AtomicGMonthDay _ ->
      AtomicGMonthDayKind
  | AtomicGDay _ ->
      AtomicGDayKind
  | AtomicGMonth _ ->
      AtomicGDayKind
  | AtomicHexBinary _ ->
      AtomicHexBinaryKind
  | AtomicBase64Binary _ ->
      AtomicBase64BinaryKind
  | AtomicAnyURI _ ->
      AtomicAnyURIKind
  | AtomicQName _ ->
      AtomicQNameKind
  | AtomicNOTATION _ ->
      AtomicNOTATIONKind
  | AtomicInteger _ ->
      AtomicIntegerKind
  | AtomicInt _ ->
      AtomicIntKind
  | AtomicYearMonthDuration _ ->
      AtomicYearMonthDurationKind
  | AtomicDayTimeDuration _ ->
      AtomicDayTimeDurationKind
  | AtomicUntyped _ ->
      AtomicUntypedKind

(************************)
(* Operations on values *)
(************************)

(* Generic operations on values *)

let serialize_float f =
  if f = 0.0 then "0" else
  if f == nan then "NaN" else
  if f = infinity then "INF" else
  if f = -.infinity then "-INF" else
  let s = Printf.sprintf "%.12e" f in
  let (mantisse, rest) =
    Gmisc.split_left_on_char s '.'
  in
  let (decimal,sign,expo) =
    try
      let (decimal,expo) = Gmisc.split_left_on_char rest 'e' in
      let (sign,expo) = (String.sub expo 0 1), (String.sub expo 1 ((String.length expo)-1)) in
      (decimal, sign, expo)
    with
    | _ ->
	(rest,"+","0")
  in
  let new_mantisse = mantisse in
  let new_decimal = Gmisc.remove_trailing decimal '0' in
  let new_sign = if sign = "+" then "" else sign in
  let new_expo = Gmisc.remove_leading expo '0' in
  new_mantisse ^ "." ^ new_decimal ^ "E" ^ new_sign ^ new_expo


(* Seralizes a value *)

let erase_atomic_value v =
  match v with
  | AtomicString x ->
      x
  | AtomicBoolean x ->
      string_of_bool x
  | AtomicFloat x ->
      serialize_float x
  | AtomicDouble x ->
      serialize_float x
  | AtomicDecimal x ->
      Decimal._string_of_decimal x
  | AtomicDuration x ->
      x
  | AtomicDateTime x ->
      DateTime.string_of_dateTime x
  | AtomicTime x ->
      x
  | AtomicDate x ->
      x
  | AtomicGYearMonth x ->
      x
  | AtomicGYear x ->
      x
  | AtomicGMonthDay x ->
      x
  | AtomicGDay x ->
      x
  | AtomicGMonth x ->
      x
  | AtomicHexBinary x ->
      x
  | AtomicBase64Binary x ->
      x
  | AtomicAnyURI x ->
      _string_of_uri x
  | AtomicQName x ->
      anon_string x
  | AtomicNOTATION x ->
      x
  | AtomicInteger x ->
      _string_of_integer x
  | AtomicInt x ->
      _string_of_int x
  | AtomicYearMonthDuration x ->
      x
  | AtomicDayTimeDuration x ->
      x
  | AtomicUntyped x ->
      x

let erase_simple_value vl =
  String.concat " " (List.map erase_atomic_value vl)

(* value equal *)

let string_equal x1 x2 =
  x1 = x2
let bool_equal x1 x2 =
  x1 = x2
let float_equal x1 x2 =
  x1 = x2
let double_equal x1 x2 =
  x1 = x2
let duration_equal x1 x2 =
  x1 = x2
let dateTime_equal x1 x2 =
  DateTime.dateTime_eq x1 x2
let time_equal x1 x2 =
  x1 = x2
let date_equal x1 x2 =
  x1 = x2
let gYearMonth_equal x1 x2 =
  x1 = x2
let gYear_equal x1 x2 =
  x1 = x2
let gMonthDay_equal x1 x2 =
  x1 = x2
let gDay_equal x1 x2 =
  x1 = x2
let gMonth_equal x1 x2 =
  x1 = x2
let hexBinary_equal x1 x2 =
  x1 = x2
let base64Binary_equal x1 x2 =
  x1 = x2
let anyURI_equal x1 x2 =
  x1 = x2
let qname_equal x1 x2 =
  equaltag (RUnitAnonSymbol x1) (RUnitAnonSymbol x2)
let notation_equal x1 x2 =
  x1 = x2

let atomic_value_eq v1 v2 =
  match (v1,v2) with
  | (AtomicString x1, AtomicString x2) ->
      string_equal x1 x2
  | (AtomicBoolean x1, AtomicBoolean x2) -> 
      bool_equal x1 x2
  | (AtomicDecimal x1,AtomicDecimal x2) -> 
      _decimal_eq x1 x2
  | (AtomicInteger x1, AtomicInteger x2) ->
      _integer_eq x1 x2
  | (AtomicInt x1, AtomicInt x2) ->
      _int_eq x1 x2
  | (AtomicFloat x1, AtomicFloat x2) -> 
      float_equal x1 x2
  | (AtomicDouble x1, AtomicDouble x2) -> 
      double_equal x1 x2
  | (AtomicDuration x1, AtomicDuration x2) -> 
      duration_equal x1 x2
  | (AtomicDateTime x1, AtomicDateTime x2) -> 
      dateTime_equal x1 x2
  | (AtomicTime x1, AtomicTime x2) -> 
      time_equal x1 x2
  | (AtomicDate x1, AtomicDate x2) -> 
      date_equal x1 x2
  | (AtomicGYearMonth x1, AtomicGYearMonth x2) -> 
      gYearMonth_equal x1 x2
  | (AtomicGYear x1, AtomicGYear x2) -> 
      gYear_equal x1 x2
  | (AtomicGMonthDay x1, AtomicGMonthDay x2) -> 
      gMonthDay_equal x1 x2
  | (AtomicGDay x1, AtomicGDay x2) -> 
      gDay_equal x1 x2
  | (AtomicGMonth x1, AtomicGMonth x2) -> 
      gMonth_equal x1 x2
  | (AtomicHexBinary x1, AtomicHexBinary x2) -> 
      hexBinary_equal x1 x2
  | (AtomicBase64Binary x1, AtomicBase64Binary x2) -> 
      base64Binary_equal x1 x2
  | (AtomicAnyURI x1, AtomicAnyURI x2) -> 
      _uri_eq x1 x2
  | (AtomicQName x1, AtomicQName x2) ->
      qname_equal x1 x2 
  | (AtomicNOTATION x1, AtomicNOTATION x2) ->
      notation_equal x1 x2
  | (AtomicUntyped x1, AtomicUntyped x2) ->
      string_equal x1 x2
  | _ ->
      false

(* value not equal *)

let atomic_value_neq v1 v2 =
  not(atomic_value_eq v1 v2)

(* value less than or equal *)

let string_lteq x1 x2 =
  x1 <= x2
let bool_lteq x1 x2 =
  x1 <= x2
let float_lteq x1 x2 =
  x1 <= x2
let double_lteq x1 x2 =
  x1 <= x2
let duration_lteq x1 x2 =
  x1 <= x2
let dateTime_lteq x1 x2 =
  x1 <= x2
let time_lteq x1 x2 =
  x1 <= x2
let date_lteq x1 x2 =
  x1 <= x2
let gYearMonth_lteq x1 x2 =
  x1 <= x2
let gYear_lteq x1 x2 =
  x1 <= x2
let gMonthDay_lteq x1 x2 =
  x1 <= x2
let gDay_lteq x1 x2 =
  x1 <= x2
let gMonth_lteq x1 x2 =
  x1 <= x2
let hexBinary_lteq x1 x2 =
  x1 <= x2
let base64Binary_lteq x1 x2 =
  x1 <= x2
let anyURI_lteq x1 x2 =
  x1 <= x2
let qname_lteq x1 x2 =
  subtag x1 x2
let notation_lteq x1 x2 =
  x1 <= x2

let atomic_value_lteq v1 v2 =
  match (v1,v2) with
  | (AtomicString x1, AtomicString x2) ->
      string_lteq x1 x2
  | (AtomicBoolean x1, AtomicBoolean x2) -> 
      bool_lteq x1 x2
  | (AtomicDecimal x1,AtomicDecimal x2) -> 
      _decimal_le x1 x2
  | (AtomicFloat x1, AtomicFloat x2) -> 
      float_lteq x1 x2
  | (AtomicDouble x1, AtomicDouble x2) -> 
      double_lteq x1 x2
  | (AtomicDuration x1, AtomicDuration x2) -> 
      duration_lteq x1 x2
  | (AtomicDateTime x1, AtomicDateTime x2) -> 
      dateTime_lteq x1 x2
  | (AtomicTime x1, AtomicTime x2) -> 
      time_lteq x1 x2
  | (AtomicDate x1, AtomicDate x2) -> 
      date_lteq x1 x2
  | (AtomicGYearMonth x1, AtomicGYearMonth x2) -> 
      gYearMonth_lteq x1 x2
  | (AtomicGYear x1, AtomicGYear x2) -> 
      gYear_lteq x1 x2
  | (AtomicGMonthDay x1, AtomicGMonthDay x2) -> 
      gMonthDay_lteq x1 x2
  | (AtomicGDay x1, AtomicGDay x2) -> 
      gDay_lteq x1 x2
  | (AtomicGMonth x1, AtomicGMonth x2) -> 
      gMonth_lteq x1 x2
  | (AtomicHexBinary x1, AtomicHexBinary x2) -> 
      hexBinary_lteq x1 x2
  | (AtomicBase64Binary x1, AtomicBase64Binary x2) -> 
      base64Binary_lteq x1 x2
  | (AtomicAnyURI x1, AtomicAnyURI x2) -> 
      anyURI_lteq x1 x2
  | (AtomicQName x1, AtomicQName x2) ->
      qname_lteq (RUnitAnonSymbol x1) (RUnitAnonSymbol x2)
  | (AtomicNOTATION x1, AtomicNOTATION x2) ->
      notation_lteq x1 x2
  | (AtomicInteger x1, AtomicInteger x2) ->
      _integer_le x1 x2
  | (AtomicInt x1, AtomicInt x2) ->
      _int_le x1 x2
  | (AtomicUntyped x1, AtomicUntyped x2) ->
      string_lteq x1 x2
  | _ ->
      false

(* value less than *)

let string_lt x1 x2 =
  x1 < x2
let bool_lt x1 x2 =
  x1 < x2
let float_lt x1 x2 =
  x1 < x2
let double_lt x1 x2 =
  x1 < x2
let duration_lt x1 x2 =
  x1 < x2
let dateTime_lt x1 x2 =
  x1 < x2
let time_lt x1 x2 =
  x1 < x2
let date_lt x1 x2 =
  x1 < x2
let gYearMonth_lt x1 x2 =
  x1 < x2
let gYear_lt x1 x2 =
  x1 < x2
let gMonthDay_lt x1 x2 =
  x1 < x2
let gDay_lt x1 x2 =
  x1 < x2
let gMonth_lt x1 x2 =
  x1 < x2
let hexBinary_lt x1 x2 =
  x1 < x2
let base64Binary_lt x1 x2 =
  x1 < x2
let anyURI_lt x1 x2 =
  x1 < x2
let qname_lt x1 x2 =
  not(equaltag x1 x2)
    &&
  (subtag x1 x2)
let notation_lt x1 x2 =
  x1 < x2

let atomic_value_lt v1 v2 =
  match (v1,v2) with
  | (AtomicString x1, AtomicString x2) ->
      string_lt x1 x2
  | (AtomicBoolean x1, AtomicBoolean x2) -> 
      bool_lt x1 x2
  | (AtomicDecimal x1,AtomicDecimal x2) -> 
      _decimal_lt x1 x2
  | (AtomicFloat x1, AtomicFloat x2) -> 
      float_lt x1 x2
  | (AtomicDouble x1, AtomicDouble x2) -> 
      double_lt x1 x2
  | (AtomicDuration x1, AtomicDuration x2) -> 
      duration_lt x1 x2
  | (AtomicDateTime x1, AtomicDateTime x2) -> 
      dateTime_lt x1 x2
  | (AtomicTime x1, AtomicTime x2) -> 
      time_lt x1 x2
  | (AtomicDate x1, AtomicDate x2) -> 
      date_lt x1 x2
  | (AtomicGYearMonth x1, AtomicGYearMonth x2) -> 
      gYearMonth_lt x1 x2
  | (AtomicGYear x1, AtomicGYear x2) -> 
      gYear_lt x1 x2
  | (AtomicGMonthDay x1, AtomicGMonthDay x2) -> 
      gMonthDay_lt x1 x2
  | (AtomicGDay x1, AtomicGDay x2) -> 
      gDay_lt x1 x2
  | (AtomicGMonth x1, AtomicGMonth x2) -> 
      gMonth_lt x1 x2
  | (AtomicHexBinary x1, AtomicHexBinary x2) -> 
      hexBinary_lt x1 x2
  | (AtomicBase64Binary x1, AtomicBase64Binary x2) -> 
      base64Binary_lt x1 x2
  | (AtomicAnyURI x1, AtomicAnyURI x2) -> 
      anyURI_lt x1 x2
  | (AtomicQName x1, AtomicQName x2) ->
      qname_lt (RUnitAnonSymbol x1) (RUnitAnonSymbol x2)
  | (AtomicNOTATION x1, AtomicNOTATION x2) ->
      notation_lt x1 x2
  | (AtomicInteger x1, AtomicInteger x2) ->
      _integer_lt x1 x2
  | (AtomicInt x1, AtomicInt x2) ->
      _int_lt x1 x2
  | (AtomicUntyped x1, AtomicUntyped x2) ->
      string_lt x1 x2
  | _ ->
      false

(* value greater than or equal *)

let string_gteq x1 x2 =
  x1 >= x2
let bool_gteq x1 x2 =
  x1 >= x2
let float_gteq x1 x2 =
  x1 >= x2
let double_gteq x1 x2 =
  x1 >= x2
let duration_gteq x1 x2 =
  x1 >= x2
let dateTime_gteq x1 x2 =
  x1 >= x2
let time_gteq x1 x2 =
  x1 >= x2
let date_gteq x1 x2 =
  x1 >= x2
let gYearMonth_gteq x1 x2 =
  x1 >= x2
let gYear_gteq x1 x2 =
  x1 >= x2
let gMonthDay_gteq x1 x2 =
  x1 >= x2
let gDay_gteq x1 x2 =
  x1 >= x2
let gMonth_gteq x1 x2 =
  x1 >= x2
let hexBinary_gteq x1 x2 =
  x1 >= x2
let base64Binary_gteq x1 x2 =
  x1 >= x2
let anyURI_gteq x1 x2 =
  x1 >= x2
let qname_gteq x1 x2 =
  subtag x2 x1
let notation_gteq x1 x2 =
  x1 >= x2

let atomic_value_gteq v1 v2 =
  match (v1,v2) with
  | (AtomicString x1, AtomicString x2) ->
      string_gteq x1 x2
  | (AtomicBoolean x1, AtomicBoolean x2) -> 
      bool_gteq x1 x2
  | (AtomicDecimal x1,AtomicDecimal x2) -> 
      _decimal_ge x1 x2
  | (AtomicFloat x1, AtomicFloat x2) -> 
      float_gteq x1 x2
  | (AtomicDouble x1, AtomicDouble x2) -> 
      double_gteq x1 x2
  | (AtomicDuration x1, AtomicDuration x2) -> 
      duration_gteq x1 x2
  | (AtomicDateTime x1, AtomicDateTime x2) -> 
      dateTime_gteq x1 x2
  | (AtomicTime x1, AtomicTime x2) -> 
      time_gteq x1 x2
  | (AtomicDate x1, AtomicDate x2) -> 
      date_gteq x1 x2
  | (AtomicGYearMonth x1, AtomicGYearMonth x2) -> 
      gYearMonth_gteq x1 x2
  | (AtomicGYear x1, AtomicGYear x2) -> 
      gYear_gteq x1 x2
  | (AtomicGMonthDay x1, AtomicGMonthDay x2) -> 
      gMonthDay_gteq x1 x2
  | (AtomicGDay x1, AtomicGDay x2) -> 
      gDay_gteq x1 x2
  | (AtomicGMonth x1, AtomicGMonth x2) -> 
      gMonth_gteq x1 x2
  | (AtomicHexBinary x1, AtomicHexBinary x2) -> 
      hexBinary_gteq x1 x2
  | (AtomicBase64Binary x1, AtomicBase64Binary x2) -> 
      base64Binary_gteq x1 x2
  | (AtomicAnyURI x1, AtomicAnyURI x2) -> 
      anyURI_gteq x1 x2
  | (AtomicQName x1, AtomicQName x2) ->
      qname_gteq (RUnitAnonSymbol x1) (RUnitAnonSymbol x2)
  | (AtomicNOTATION x1, AtomicNOTATION x2) ->
      notation_gteq x1 x2
  | (AtomicInteger x1, AtomicInteger x2) ->
      _integer_ge x1 x2
  | (AtomicInt x1, AtomicInt x2) ->
      _int_ge x1 x2
  | (AtomicUntyped x1, AtomicUntyped x2) ->
      string_gteq x1 x2
  | _ ->
      false

(* value greater than *)

let string_gt x1 x2 =
  x1 > x2
let bool_gt x1 x2 =
  x1 > x2
let float_gt x1 x2 =
  x1 > x2
let double_gt x1 x2 =
  x1 > x2
let duration_gt x1 x2 =
  x1 > x2
let dateTime_gt x1 x2 =
  x1 > x2
let time_gt x1 x2 =
  x1 > x2
let date_gt x1 x2 =
  x1 > x2
let gYearMonth_gt x1 x2 =
  x1 > x2
let gYear_gt x1 x2 =
  x1 > x2
let gMonthDay_gt x1 x2 =
  x1 > x2
let gDay_gt x1 x2 =
  x1 > x2
let gMonth_gt x1 x2 =
  x1 > x2
let hexBinary_gt x1 x2 =
  x1 > x2
let base64Binary_gt x1 x2 =
  x1 > x2
let anyURI_gt x1 x2 =
  x1 > x2
let qname_gt x1 x2 =
  not(equaltag x1 x2)
    &&
  (subtag x2 x1)
let notation_gt x1 x2 =
  x1 > x2

let atomic_value_gt v1 v2 =
  match (v1,v2) with
  | (AtomicString x1, AtomicString x2) ->
      string_gt x1 x2
  | (AtomicBoolean x1, AtomicBoolean x2) -> 
      bool_gt x1 x2
  | (AtomicDecimal x1,AtomicDecimal x2) -> 
      _decimal_gt x1 x2
  | (AtomicFloat x1, AtomicFloat x2) -> 
      float_gt x1 x2
  | (AtomicDouble x1, AtomicDouble x2) -> 
      double_gt x1 x2
  | (AtomicDuration x1, AtomicDuration x2) -> 
      duration_gt x1 x2
  | (AtomicDateTime x1, AtomicDateTime x2) -> 
      dateTime_gt x1 x2
  | (AtomicTime x1, AtomicTime x2) -> 
      time_gt x1 x2
  | (AtomicDate x1, AtomicDate x2) -> 
      date_gt x1 x2
  | (AtomicGYearMonth x1, AtomicGYearMonth x2) -> 
      gYearMonth_gt x1 x2
  | (AtomicGYear x1, AtomicGYear x2) -> 
      gYear_gt x1 x2
  | (AtomicGMonthDay x1, AtomicGMonthDay x2) -> 
      gMonthDay_gt x1 x2
  | (AtomicGDay x1, AtomicGDay x2) -> 
      gDay_gt x1 x2
  | (AtomicGMonth x1, AtomicGMonth x2) -> 
      gMonth_gt x1 x2
  | (AtomicHexBinary x1, AtomicHexBinary x2) -> 
      hexBinary_gt x1 x2
  | (AtomicBase64Binary x1, AtomicBase64Binary x2) -> 
      base64Binary_gt x1 x2
  | (AtomicAnyURI x1, AtomicAnyURI x2) -> 
      anyURI_gt x1 x2
  | (AtomicQName x1, AtomicQName x2) ->
      qname_gt (RUnitAnonSymbol x1) (RUnitAnonSymbol x2)
  | (AtomicNOTATION x1, AtomicNOTATION x2) ->
      notation_gt x1 x2
  | (AtomicInteger x1, AtomicInteger x2) ->
      _integer_gt x1 x2
  | (AtomicInt x1, AtomicInt x2) ->
      _int_gt x1 x2
  | (AtomicUntyped x1, AtomicUntyped x2) ->
      string_gt x1 x2
  | _ ->
      false

(* value compare *)

let atomic_value_compare v1 v2 =
  if not(getAtomicValueKind v1 = getAtomicValueKind v2)
  then
    raise (Query (Parameter_Mismatch "Comparison between two incompatible values"))
  else
    if atomic_value_eq v1 v2
    then
      0
    else
      if atomic_value_gt v1 v2
      then
	1
      else
	-1

(* Operations on QNames *)

let localname qn =
  match anon_name qn with
  | (_,s) -> s

let namespace qn =
  match anon_name qn with
  | (u,_) -> u

(*******************************)
(* Parsing from text to values *)
(*******************************)

let string_of_untyped t =
  t

let boolean_of_untyped t =
  try
    Gmisc.wrap_lexer Datatypes_lexer.parse_boolean t
  with
  | Failure _ ->
      raise (Query (Validation ("Text: \"" ^ t ^ "\" not a boolean value")))

let decimal_of_untyped t =
  try
    Gmisc.wrap_lexer Datatypes_lexer.parse_decimal t
  with
  | Failure _ ->
      raise (Query (Validation ("Text: \"" ^ t ^ "\" not a decimal value")))

let float_of_untyped t =
  try
    Gmisc.wrap_lexer Datatypes_lexer.parse_float t
  with
  | Failure _ ->
      raise (Query (Validation ("Text: \"" ^ t ^ "\" not a float value")))

let double_of_untyped t =
  try
    Gmisc.wrap_lexer Datatypes_lexer.parse_double t
  with
  | Failure _ ->
      raise (Query (Validation ("Text: \"" ^ t ^ "\" not a double value")))

let duration_of_untyped t = t
let yearMonthDuration_of_untyped t = t
let dayTimeDuration_of_untyped t = t
let dateTime_of_untyped t = DateTime.dateTime_of_string t
let time_of_untyped t = t
let date_of_untyped t = t
let gYearMonth_of_untyped t = t
let gYear_of_untyped t = t
let gMonthDay_of_untyped t = t
let gDay_of_untyped t = t
let gMonth_of_untyped t = t
let hexBinary_of_untyped t = t
let base64Binary_of_untyped t = t
let anyURI_of_untyped t =
  _uri_of_string t

let qname_of_untyped nse t =
  try
    let uqname = Namespace.uqname_element_of_string t in
    let (_,rqname) = Resolve.resolve_element_qname nse uqname in
    anon_symbol rqname
  with
  | Failure _ ->
      raise (Query (Validation ("Text: \"" ^ t ^ "\" not a qname value")))

let notation_of_untyped t = t
let integer_of_untyped t = 
  try
    Gmisc.wrap_lexer Datatypes_lexer.parse_integer t
  with
  | Failure _ ->
      raise (Query (Validation ("Text: \"" ^ t ^ "\" not an integer value")))

let int_of_untyped t = 
  try
    Gmisc.wrap_lexer Datatypes_lexer.parse_int t
  with
  | Failure _ ->
      raise (Query (Validation ("Text: \"" ^ t ^ "\" not an integer value")))

(****************************)
(* Support for atomic types *)
(****************************)


type atomic_type =
      (* 19 primitive XML Schema atomic types *)
  | ATString
  | ATBoolean
  | ATDecimal
  | ATFloat
  | ATDouble
  | ATDuration
  | ATDateTime
  | ATTime
  | ATDate
  | ATGYearMonth
  | ATGYear
  | ATGMonthDay
  | ATGDay
  | ATGMonth
  | ATHexBinary
  | ATBase64Binary
  | ATAnyURI
  | ATQName
  | ATNOTATION

      (* Derived atomic types *)
  | ATInteger
  | ATInt

      (* Atomic types in xdt namespace *)
  | ATYearMonthDuration
  | ATDayTimeDuration
  | ATUntyped

type _Def_T =
  | SimpleType of atomic_type
  | ComplexType of rgroup_symbol

(* Atomic type of an atomic value *)

let atomic_type_of_atomic_value bv =
  match bv with
  | AtomicString _ -> ATString
  | AtomicBoolean _ -> ATBoolean
  | AtomicFloat _ -> ATFloat
  | AtomicDouble _ -> ATDouble
  | AtomicDecimal _ -> ATDecimal
  | AtomicDuration _ -> ATDuration
  | AtomicDateTime _ -> ATDateTime
  | AtomicTime _ -> ATTime
  | AtomicDate _ -> ATDate
  | AtomicGYearMonth _ -> ATGYearMonth
  | AtomicGYear _ -> ATGYear
  | AtomicGMonthDay _ -> ATGMonthDay
  | AtomicGDay _ -> ATGDay
  | AtomicGMonth _ -> ATGMonth
  | AtomicHexBinary _ -> ATHexBinary
  | AtomicBase64Binary _ -> ATBase64Binary
  | AtomicAnyURI _ -> ATAnyURI
  | AtomicQName _ -> ATQName
  | AtomicNOTATION _ -> ATNOTATION
  | AtomicInteger _ -> ATInteger
  | AtomicInt _ -> ATInt
  | AtomicYearMonthDuration _ -> ATYearMonthDuration
  | AtomicDayTimeDuration _ -> ATDayTimeDuration
  | AtomicUntyped _ -> ATUntyped

let untyped_atomic_type = ATUntyped

(* Operations on atomic types *)

let atomic_type_subsumes st st' =
  match (st,st') with
  | (ATString, ATString) -> true
  | (ATBoolean, ATBoolean) -> true
  | (ATDecimal, ATDecimal) -> true
  | (ATFloat,ATFloat) -> true
  | (ATDouble, ATDouble) -> true
  | (ATDuration, ATDuration) -> true
  | (ATDateTime, ATDateTime) -> true
  | (ATTime, ATTime) -> true
  | (ATDate, ATDate) -> true
  | (ATGYearMonth, ATGYearMonth) -> true
  | (ATGYear, ATGYear) -> true
  | (ATGMonthDay, ATGMonthDay) -> true
  | (ATGDay, ATGDay) -> true
  | (ATGMonth, ATGMonth) -> true
  | (ATHexBinary, ATHexBinary) -> true
  | (ATBase64Binary, ATBase64Binary) -> true
  | (ATAnyURI, ATAnyURI) -> true
  | (ATQName, ATQName) -> true
  | (ATNOTATION, ATNOTATION) -> true
  | (ATInteger, ATInteger) -> true
  | (ATInt, ATInt) -> true
    (* Note:
         Untyped is in a separate part of the type hierarchy.
       - Jerome
     *)
  | (ATYearMonthDuration, ATYearMonthDuration) -> true
  | (ATDayTimeDuration, ATDayTimeDuration) -> true
  | (ATUntyped, ATUntyped) -> true
  | _ -> false

(* mappings for built-in atomic types qnames *)

let built_in_type_table = [
  ((xsd_prefix, "string"), ATString);
  ((xsd_prefix, "boolean"), ATBoolean);
  ((xsd_prefix, "decimal"), ATDecimal);
  ((xsd_prefix, "float"), ATFloat);
  ((xsd_prefix, "double"), ATDouble);
  ((xsd_prefix, "duration"), ATDuration);
  ((xsd_prefix, "dateTime"), ATDateTime);
  ((xsd_prefix, "time"), ATTime);
  ((xsd_prefix, "date"), ATDate);
  ((xsd_prefix, "gYearMonth"), ATGYearMonth);
  ((xsd_prefix, "gYear"), ATGYear);
  ((xsd_prefix, "gMonthDay"), ATGMonthDay);
  ((xsd_prefix, "gDay"), ATGDay);
  ((xsd_prefix, "gMonth"), ATGMonth);
  ((xsd_prefix, "hexBinary"), ATHexBinary);
  ((xsd_prefix, "base64Binary"), ATBase64Binary);
  ((xsd_prefix, "anyURI"), ATAnyURI);
  ((xsd_prefix, "QName"), ATQName);
  ((xsd_prefix, "NOTATION"), ATNOTATION);

  ((xsd_prefix, "integer"), ATInteger);
  ((xsd_prefix, "int"), ATInt);
  
  ((xs_prefix, "string"), ATString);
  ((xs_prefix, "boolean"), ATBoolean);
  ((xs_prefix, "decimal"), ATDecimal);
  ((xs_prefix, "float"), ATFloat);
  ((xs_prefix, "double"), ATDouble);
  ((xs_prefix, "duration"), ATDuration);
  ((xs_prefix, "dateTime"), ATDateTime);
  ((xs_prefix, "time"), ATTime);
  ((xs_prefix, "date"), ATDate);
  ((xs_prefix, "gYearMonth"), ATGYearMonth);
  ((xs_prefix, "gYear"), ATGYear);
  ((xs_prefix, "gMonthDay"), ATGMonthDay);
  ((xs_prefix, "gDay"), ATGDay);
  ((xs_prefix, "gMonth"), ATGMonth);
  ((xs_prefix, "hexBinary"), ATHexBinary);
  ((xs_prefix, "base64Binary"), ATBase64Binary);
  ((xs_prefix, "anyURI"), ATAnyURI);
  ((xs_prefix, "QName"), ATQName);
  ((xs_prefix, "NOTATION"), ATNOTATION);

  ((xs_prefix, "integer"), ATInteger);
  ((xs_prefix, "int"), ATInt);

  (* Types in xdt namespace *)

  ((xdt_prefix, "yearMonthDuration"), ATYearMonthDuration);
  ((xdt_prefix, "dayTimeDuration"), ATDayTimeDuration);
  ((xdt_prefix, "untypedAtomic"), ATUntyped);
]

let bltin_types : (uqname, atomic_type) Hashtbl.t = Hashtbl.create 121

let lookup_bltin_type name = Hashtbl.find bltin_types name
let add_bltin_type (name,t) = ignore (Hashtbl.add bltin_types name t)

let _ = List.map add_bltin_type built_in_type_table

(* mapping from atomic types to qnames *)

let uqname_of_primitive_type pt =
  match pt with
  | ATString ->
      (xsd_prefix, "string")
  | ATBoolean ->
      (xsd_prefix, "boolean")
  | ATDecimal ->
      (xsd_prefix, "decimal")
  | ATFloat ->
      (xsd_prefix, "float")
  | ATDouble ->
      (xsd_prefix, "double")
  | ATDuration ->
      (xsd_prefix, "duration")
  | ATDateTime ->
      (xsd_prefix, "dateTime")
  | ATTime ->
      (xsd_prefix, "time")
  | ATDate ->
      (xsd_prefix, "date")
  | ATGYearMonth ->
      (xsd_prefix, "gYearMonth")
  | ATGYear ->
      (xsd_prefix, "gYear")
  | ATGMonthDay ->
      (xsd_prefix, "gMonthDay")
  | ATGDay ->
      (xsd_prefix, "gDay")
  | ATGMonth ->
      (xsd_prefix, "gMonth")
  | ATHexBinary ->
      (xsd_prefix, "hexBinary")
  | ATBase64Binary ->
      (xsd_prefix, "base64Binary")
  | ATAnyURI ->
      (xsd_prefix, "anyURI")
  | ATQName ->
      (xsd_prefix, "QName")
  | ATNOTATION ->
      (xsd_prefix, "NOTATION")
  | ATInteger ->
      (xsd_prefix, "integer")
  | ATInt ->
      (xsd_prefix, "int")
  | ATYearMonthDuration ->
      (xdt_prefix, "yearMonthDuration")
  | ATDayTimeDuration ->
      (xdt_prefix, "dayTimeDuration")
  | ATUntyped ->
      (xdt_prefix, "untypedAtomic")


(* Unit tag of text node *)

let unit_symbol_of_base_type bt =
  match bt with
  | ATString ->
      stringsym
  | ATBoolean ->
      booleansym
  | ATDecimal ->
      decimalsym
  | ATFloat ->
      floatsym
  | ATDouble ->
      doublesym
  | ATDuration ->
      durationsym
  | ATDateTime ->
      dateTimesym
  | ATTime ->
      timesym
  | ATDate ->
      datesym
  | ATGYearMonth ->
      gYearMonthsym
  | ATGYear ->
      gYearsym
  | ATGMonthDay ->
      gMonthDaysym
  | ATGDay ->
      gDaysym
  | ATGMonth ->
      gMonthsym
  | ATHexBinary ->
      hexBinarysym
  | ATBase64Binary ->
      base64Binarysym
  | ATAnyURI ->
      anyURIsym
  | ATQName ->
      qnamesym
  | ATNOTATION ->
      notationsym
  | ATInteger ->
      integersym
  | ATInt ->
      intsym
  | ATUntyped ->
      untypedsym
  | ATYearMonthDuration ->
      yearMonthDurationsym
  | ATDayTimeDuration ->
      dayTimeDurationsym

let promotion_table = 
 [ ATString, [ATString]; 
   ATBoolean, [ATBoolean];
   ATInt, [ATInt; ATInteger; ATDecimal; ATFloat; ATDouble];
   ATInteger, [ATInteger; ATDecimal; ATFloat; ATDouble];
   ATDecimal, [ATDecimal; ATFloat; ATDouble];
   ATFloat, [ATFloat; ATDouble];
   ATDouble, [ATDouble];
   ATDuration, [ATDuration];
   ATDateTime, [ATDateTime];
   ATTime, [ATTime]; 
   ATDate, [ATDate];
   ATGYearMonth, [ATGYearMonth];
   ATGYear, [ATGYear];
   ATGMonthDay, [ATGMonthDay];
   ATGDay, [ATGDay];
   ATGMonth, [ATGMonth];
   ATHexBinary, [ATHexBinary];
   ATBase64Binary, [ATBase64Binary];
   ATAnyURI, [ATAnyURI];
   ATQName, [ATQName];
   ATNOTATION, [ATNOTATION];
   ATYearMonthDuration, [ATYearMonthDuration];
   ATDayTimeDuration, [ATDayTimeDuration];
   ATUntyped, [ATUntyped]]

let can_be_promoted_to at = 
  try 
    List.assoc at promotion_table 
  with
    Not_found -> raise (Query(Internal_Error("Atomic type not in promotion table.")))
