(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: decimal.ml,v 1.2 2002/12/14 21:48:36 mff Exp $ *)

open Error

open Num
open Big_int

(* Note:
     This modules implement internal support for the XML Schema
     primitive simple type decimal, and the corresponding derived types.
   - Jerome
 *)

(* Setting global parameters *)

let precision = 18

(* Note:
     18 is the minimum precision required by XML Schema.
   - Jerome
 *)

(*********************************************************)
(* Internal representation of decimals and derived types *)
(*********************************************************)

type _decimal = num
type _integer = big_int
type _int     = int

(*************************)
(* Operations on decimal *)
(*************************)

let _decimal_zero = num_of_int 0

let _decimal_of_string s =
  try
    let (s1, s2) = Gmisc.split_left_on_char s '.' in
    let s1 =          (* Fix in case there is no leading digit - Jerome *)
      match s1 with
      |	"" -> "0"
      |	"-" -> "-0"
      |	_ -> s1
    in
    let exponent = power_int_positive_int 10 (String.length s2) in
    let string_exponent = string_of_big_int exponent in
    let new_s = s1 ^ s2 ^ "/" ^ string_exponent in
    num_of_string new_s
  with
  | Not_found ->
      num_of_string s

let _string_of_decimal n =
  let s = approx_num_fix precision n in
  let (mantisse, rest) =
    Gmisc.split_left_on_char s '.'
  in
  let new_mantisse = 
    try
      if (String.get mantisse 0) = '+'
      then
	String.sub s 1 ((String.length mantisse) - 1)
      else
	mantisse
    with
    | _ ->
	mantisse
  in
  let new_rest = Gmisc.remove_trailing rest '0' in
  new_mantisse ^ "." ^ new_rest

let _decimal_eq d1 d2 =
  eq_num d1 d2

let _decimal_lt d1 d2 =
  lt_num d1 d2

let _decimal_le d1 d2 =
  le_num d1 d2

let _decimal_gt d1 d2 =
  gt_num d1 d2

let _decimal_ge d1 d2 =
  ge_num d1 d2

let _decimal_add d1 d2 =
  add_num d1 d2

let _decimal_sub d1 d2 =
  sub_num d1 d2

let _decimal_mult d1 d2 =
  mult_num d1 d2

let _decimal_div d1 d2 =
  div_num d1 d2

let _decimal_mod d1 d2 =
  mod_num d1 d2

let _decimal_unary_minus d1 =
  minus_num d1 

let _decimal_of_integer i1 =
  num_of_big_int i1

let _decimal_of_int i1 =
   num_of_int i1

(*************************)
(* Operations on integer *)
(*************************)

let _integer_zero = zero_big_int

let _integer_of_string s =
  big_int_of_string s

let _string_of_integer i =
  string_of_big_int i

let _integer_eq i1 i2 =
  eq_big_int i1 i2

let _integer_lt i1 i2 =
  lt_big_int i1 i2

let _integer_le i1 i2 =
  le_big_int i1 i2

let _integer_gt i1 i2 =
  gt_big_int i1 i2

let _integer_ge i1 i2 =
  ge_big_int i1 i2

let _integer_add i1 i2 =
  add_big_int i1 i2

let _integer_sub i1 i2 =
  sub_big_int i1 i2

let _integer_mult i1 i2 =
  mult_big_int i1 i2

let _integer_div i1 i2 =
  div_big_int i1 i2

let _integer_mod i1 i2 =
  mod_big_int i1 i2

let _integer_unary_minus i1 =
  minus_big_int i1 

let _integer_of_decimal d1 =
  if is_integer_num d1
  then
    big_int_of_num d1
  else
    raise (Query (Cast_Error ("Cannot cast decimal : " ^ (string_of_num d1) ^ " into an int")))

let _integer_of_int i1 =
  big_int_of_int i1

(*********************)
(* Operations on int *)
(*********************)

let _int_zero = 0

let _int_of_string s =
  let i = int_of_string s in
  if not(string_of_int i = s)
  then
    raise (Query (Wrong_Args ("int value: " ^ s ^ " is out of bounds")))
  else
    i

let _string_of_int i =
  string_of_int i

let _int_eq i1 i2 =
  i1 = i2

let _int_lt i1 i2 =
  i1 < i2

let _int_le i1 i2 =
  i1 <= i2

let _int_gt i1 i2 =
  i1 > i2

let _int_ge i1 i2 =
  i1 >= i2

let _int_add i1 i2 =
  i1 + i2

let _int_sub i1 i2 =
  i1 - i2

let _int_mult i1 i2 =
  i1 * i2

let _int_div i1 i2 =
  i1 / i2

let _int_mod i1 i2 =
  i1 mod i2

let _int_unary_minus i1  =
  (-(i1))

let _int_of_decimal d1 =
  if is_integer_num d1
  then
    int_of_num d1
  else
    raise (Query (Cast_Error ("Cannot cast decimal : " ^ (string_of_num d1) ^ " into an int")))
    
let _int_of_integer i1 =
  try
    int_of_big_int i1
  with
  | Failure _ ->
      raise (Query (Cast_Error ("Cannot cast integer : " ^ (string_of_big_int i1) ^ " into an int")))
