(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: dynamic_context.ml,v 1.4 2003/06/12 21:38:07 mff Exp $ *)

(* Module Dynamic_context
   Description:
     This module implements the XQuery dynamic context.
*)

open Sym
open Xquery_ast
open Datamodel
open Error


(* Dynamic (or evaluation) environment *)

(* The evaluation environment has three parts: 
   Static context + 
   Variables and their values 
   Keys and their values 
   Functions and their bodies 
 *) 

type key_table =
    (string * string, _Value) Hashtbl.t

type dynamic_context =
    { static_context : Static_context.static_context;
      var_value : (vname * _Extended_Value) list;
      keys : key_table;
      fun_bodies : ((rfname * int), cfunction_body) Hashtbl.t }

module HashedValue =
  struct
    type t = _Value
    let equal = value_eq
    let hash = Hashtbl.hash
  end

module ValueHashtbl = Hashtbl.Make(HashedValue)

(* This function raises an error if 'fn' is already defined. *)

let add_function bodies fi fd =
  let (fn,fs,fb) = fd.pcfunction_def_desc in
  let arity = List.length (fst fs) in
  try 
    let _ = (Hashtbl.find bodies (fn, arity)) 
    in raise (Query (Symbol_Already_Defined (fi, "Function ", 
					     (Namespace.string_of_rqname fn))))
  with Not_found ->
    Hashtbl.add bodies (fn,arity) fb

(* Create a new environment from a list of function declarations *)

let build_dynamic_context s funs fi =
  let ctxt = 
    { static_context = s; 
      var_value = [];
      keys = Hashtbl.create 167;
      fun_bodies  = Hashtbl.create 167 }
  in
  let _ = List.iter (add_function ctxt.fun_bodies fi) funs
  in ctxt 

let copy_dynamic_context dyn_ctxt stat_ctxt = 
  { static_context = Static_context.copy_static_context stat_ctxt;
    var_value = dyn_ctxt.var_value;
    keys = Hashtbl.copy dyn_ctxt.keys;
    fun_bodies = Hashtbl.copy dyn_ctxt.fun_bodies } 

let static_context_from_dynamic_context c = c.static_context
(* add a variable in the environment *)

let add_var_value_to_dynamic_context dyn_ctxt (v, e) =
  let ev = _Extended_Value_of_Value e in
  { static_context = dyn_ctxt.static_context;
    var_value = (v,ev) :: dyn_ctxt.var_value;
    keys = dyn_ctxt.keys;
    fun_bodies = dyn_ctxt.fun_bodies }

let add_var_exvalue_to_dynamic_context er (v, ev) =
  { static_context = er.static_context;
    var_value = (v,ev) :: er.var_value;
    keys = er.keys;
    fun_bodies = er.fun_bodies }

(* get the value of a variable in the environment *)

let var_value_from_dynamic_context er vn =
  try
    _Value_of_Extended_Value (List.assoc vn (er.var_value))
  with
  | Not_found ->
      let ovn = if (vn = Xquery_astutil.dot) then None else Some (Namespace.string_of_uqname vn) in
      raise (Query (Undefined_Variable ovn))

let var_exvalue_from_dynamic_context er vn =
  try
    (List.assoc vn (er.var_value))
  with
  | Not_found ->
      let ovn = if (vn = Xquery_astutil.dot) then None else Some (Namespace.string_of_uqname vn) in
      raise (Query (Undefined_Variable ovn))

(* get the body of a function *)

let body_from_dynamic_context er (fn,i) =
  try
    Hashtbl.find (er.fun_bodies) (fn,i) 
  with
  | Not_found ->
      raise (Query (Undefined("Function "
			      ^ (Namespace.string_of_rqname fn)
			      ^ " with arity "
			      ^ (string_of_int i)
			      ^ " not found.")))

let add_funs_to_dynamic_context dyn_ctxt funs fi = 
  let dyn_ctxt' = copy_dynamic_context dyn_ctxt dyn_ctxt.static_context in
  let _ = List.iter (add_function dyn_ctxt'.fun_bodies fi) funs in
  dyn_ctxt'

let add_key_to_dynamic_context er (kn,key) node =
  let ht = er.keys in
  begin
    if (Hashtbl.mem ht (kn,key))
    then
      raise (Query (Key_Error (key,"Key definition binding node for value " ^ key ^ " twice")))
    else
      Hashtbl.add ht (kn,key) node
  end;
  er

let key_from_dynamic_context er (kn,key) =
  let ht = er.keys in
  try
    Hashtbl.find ht (kn,key)
  with
  | _ ->
      raise (Query (KeyRef_Error (key,"Binding for key value " ^ key ^ " not found.")))

