(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: eval_expr.ml,v 1.5 2003/05/28 14:47:38 simeon Exp $ *)

(* Module Eval_expr
   Description:
     This module contains the evaluation code for XQuery expression.
*)

open Format

open Error

open Namespace
open Sym

open Datatypes
open Xquery_ast
open Datamodel

open Norm_context
open Static_context
open Dynamic_context

open Eval_util

(*************)
(* Eval loop *)
(*************)

(* Note:
     This is the main eval loop.

     This is a large recursive function which walks down the XQuery
     AST, and perform bottom-up evaluation.

     * tr is the static environment (used for type related expressions
       and function calls).
     * er is the dynamic environment (contains variable bindings).
     * nsenv is the namespace environment (needed for locally defined
       namespaces).
     * e0 is the expression being evaluated.

   - Jerome *)


let rec eval_cexpr dyn_ctxt e0 =
  let ev =
    eval_cexpr_to_extended_value dyn_ctxt e0 
  in
    _Value_of_Extended_Value ev


and eval_cexpr_to_extended_value dyn_ctxt e0 =
  let stat_ctxt = static_context_from_dynamic_context dyn_ctxt in
  let norm_ctxt = norm_context_from_stat_context stat_ctxt in
  let schema = schema_from_norm_context norm_ctxt in
  let nsenv = nsenv_from_norm_context norm_ctxt in
  (* Gets the original file information *)
  let fi = e0.pcexpr_loc in

  (* Now let's get started *)
  match e0.pcexpr_desc with
  | CESort (v, e1, stable, e2, sk, esk) ->
      (* Note:
	   [e1] is the The input_sequence to be sorted. This should
	   *always* be a sequence of tuples.

	   [e2] is the sort criteria. This should always take tuples
	   as input.
       - Jerome *)
      let input_sequence = eval_cexpr_to_extended_value dyn_ctxt e1 in
      eval_sort dyn_ctxt v input_sequence e2 stable sk esk

  (* fn:unordered is a no-op for now *)
  | CEUnordered e ->
      _Extended_Value_of_Value (eval_cexpr dyn_ctxt e)

  | CEFor (odt,v,None,e1,e2) ->
      let sequence = eval_cexpr dyn_ctxt e1 in
      let eval_for x =
	let item = dynamic_type_check stat_ctxt fi odt (sequenceBuildSingle x) in

	(* A normalized -for- expression with an -order- clause may
           include a -let- and/or -where- clause, which returns a
           tuple value or the empty sequence.  Evaluation of the
           nested -let- or -where- will yield a tuple instead of a
           value, raising an error in expr_cexpr.  The try-with
           expression below catches this error, and returns the tuple
           value. *)

	let dyn_ctxt' = add_var_value_to_dynamic_context dyn_ctxt (v,item) in 
	try
	  eval_cexpr_to_extended_value dyn_ctxt' e2
	with 
	| Not_A_Value ev -> ev
      in
      let ev =  _Concat_Extended_Value_Sequence (sequenceMap eval_for sequence) in
      (ev)

  | CEFor (odt,v,Some v',e1,e2) ->
      let forest = eval_cexpr dyn_ctxt e1 in
      let index = ref 1 in
      let eval_for x =
	let item = dynamic_type_check stat_ctxt fi odt (sequenceBuildSingle x) in
	let dyn_ctxt' = add_var_value_to_dynamic_context dyn_ctxt (v,item) in 
	let dmindex = sequenceBuildSingle (AtomicValue (AtomicInteger (Decimal._integer_of_int !index))) in
	let dyn_ctxt'' = add_var_value_to_dynamic_context dyn_ctxt' (v',dmindex) in 

        (* See comment 	'A normalized -for-...' above *)
	let result = 
	  try 
	    eval_cexpr_to_extended_value dyn_ctxt'' e2 
	  with
	  | Not_A_Value ev -> ev
	in
	begin
	  incr index;
	  result
	end
      in 
      let ev =  _Concat_Extended_Value_Sequence  (sequenceMap eval_for forest) in 
      (ev)

  | CEForTuple (v,e1,e2) ->
      let forest = _Tuples_of_Extended_Value (eval_cexpr_to_extended_value dyn_ctxt e1) 
      in
      let eval_for x =
	let item = _Extended_Value_of_Tuples (sequenceBuildSingle x) in
	let dyn_ctxt' = add_var_exvalue_to_dynamic_context dyn_ctxt (v,item)
	in (eval_cexpr dyn_ctxt' e2)
      in
      _Extended_Value_of_Value (sequenceConcatMap eval_for forest)

  | CELet (odt,var,e1,e2) ->
      let forest = dynamic_type_check stat_ctxt fi odt (eval_cexpr dyn_ctxt e1) in
      let dyn_ctxt' = add_var_value_to_dynamic_context dyn_ctxt (var,forest) in

	(* A normalized -let- expression with an -order- clause may
           include a -let-, -for- and/or -where- clause, which returns a
           tuple value or the empty sequence.  Evaluation of the
           nested -let-, -for-, or -where- will yield a tuple instead of a
           value, raising an error in expr_cexpr.  The try-with
           expression below catches this error, and returns the tuple
           value. *)
      begin
	try 
	  eval_cexpr_to_extended_value dyn_ctxt' e2 
	with
	| Not_A_Value ev -> ev
      end

  | CEIf (e1,e2,e3) ->
      let res =
	if (cast_to_bool (eval_cexpr dyn_ctxt e1))
	then eval_cexpr dyn_ctxt e2
	else eval_cexpr dyn_ctxt e3
      in
      _Extended_Value_of_Value res

  | CETypeswitch (e,cs) ->
      let forest = eval_cexpr dyn_ctxt e in
      _Extended_Value_of_Value (eval_match dyn_ctxt forest cs)

  | CEVar vn ->
      var_exvalue_from_dynamic_context dyn_ctxt vn

  | CEScalar s ->
      _Extended_Value_of_Value (sequenceBuildSingle (buildAtomicValue s))

  | CEText t ->
      _Extended_Value_of_Value (sequenceBuildSingle (buildNode (buildTextNode (textNode t))))

  | CETextComputed e1 ->
      let items = eval_cexpr dyn_ctxt e1 in
      let t =
	try
	  Datamodel.value_to_untyped items
	with
	| _ ->
	    raise (Query (Wrong_Args "Text node should contain text"))
      in
      _Extended_Value_of_Value (sequenceBuildSingle (buildNode (buildTextNode (textNode t))))

  | CEPI (target,pi_content) ->
      _Extended_Value_of_Value (sequenceBuildSingle (buildNode (buildProcessingInstructionNode (piNode target pi_content))))

  | CEComment c ->
      _Extended_Value_of_Value (sequenceBuildSingle (buildNode (buildCommentNode (commentNode c))))

  | CEDocument e1 ->
      let res =
	let items = eval_cexpr dyn_ctxt e1 in
	let (doc_children, doc_values) = filterValue items in
	if sequenceIsEmpty doc_values
	then
	  sequenceBuildSingle (buildNode
				 (buildDocumentNode (docNode (AnyURI.get_base_uri ()) doc_children None)))
	else
	  raise (Query (Wrong_Args ("Document node cannot contain values")))
      in
      _Extended_Value_of_Value res

  | CEApp (fn,args) ->

    (* 
                           statEnv |- QName => qname
                  statEnv.funcType(qname) = { FuncDecl1, ..., FuncDeclm }
               FuncDecli = define function qname(Type1, ..., Typen) as Type
                  dynEnv |- Expr1 => Value1 ... dynEnv |- Exprn => Valuen
                     statEnv |- Value1 against Type1 promotes to Value1'
                                        ...
                     statEnv |- Valuen against Typen promotes to Valuen'
           dynEnv.funcDefn(qname(Type1, ..., Typen)) = (Expr, Variable1, ... , Variablen)
           dynEnv [ varValue = ( Variable1 |-> Value1'; ...; Variablen |-> Valuen') ] |- Expr => Value
                      statEnv |- Value against Type promotes to Value'
          -------------------------------------------------------------------------------------------
                       dynEnv |- QName ( Expr1, ..., Exprn ) => Value'
    *)
      let res =
	begin
	  let eargs = List.map (fun x -> eval_cexpr dyn_ctxt x) args in
	  let arity = List.length args in
	  let signatures = Overloaded.resolve_overloaded_function norm_ctxt (fn, arity) fi in
	  (* Pick the first function whose signature matches the types
             of the function arguments, after type promotion *)
	  let (fn', eargs') = match_function_signature stat_ctxt fi fn eargs signatures in
(* print_string ("Function " ^(string_of_rqname fn)^" matches "^(string_of_rqname fn'));  *)
	  match body_from_dynamic_context dyn_ctxt (fn',arity) with
	  | CEFunctionBltIn fn ->
	      let bf = Fn_dynamic.lookup_bltin_fctn (fn,arity) in
	      (bf nsenv schema eargs')
	  | CEFunctionUser (vars, fb) ->
	      let couples =
		try
		  List.combine vars eargs'
		with
		  Invalid_argument _ ->
		    raise (Query (Wrong_Args ("Incompatible number of elements in funtion"
					      ^ (Namespace.string_of_rqname fn))))
	      in
	      let dyn_ctxt' = List.fold_right (function (v,vv) -> (function xer -> add_var_value_to_dynamic_context xer (v,vv))) couples dyn_ctxt in
              (* Should we check the return value here?  
                 Or assume that it is checked where the function is applied? *)
	      eval_cexpr dyn_ctxt' fb
	end
      in
      _Extended_Value_of_Value res

  | CEKeyRef (kn,ke) ->
      let res =
	begin
	  let kv = eval_cexpr dyn_ctxt ke in
	  let keyval =
	    try
	      if (sequenceIsEmpty kv) then 
		""
	      else
		let i = (sequenceGetSingle kv) in
		if (isNode i) then 
		  let n = getNode i in
		  nodeStringValue n
		else 
		  Datatypes.erase_atomic_value (getAtomicValue i)
	    with
	    | _ ->
		raise (Query (Internal_Error "key definition: key value not an item"))
	  in
	  key_from_dynamic_context dyn_ctxt (kn,keyval)
	end
      in
      _Extended_Value_of_Value res

  | CESeq (e1,e2) ->
      let n1 = eval_cexpr dyn_ctxt e1
      and n2 = eval_cexpr dyn_ctxt e2 in
      _Extended_Value_of_Value (sequenceAppend n1 n2)

  | CEEmpty ->
      _Extended_Value_of_Value (sequenceEmpty)

  | CEElem (sym,e1) ->
      let instances = eval_cexpr dyn_ctxt e1 in
      _Extended_Value_of_Value (element_constructor sym instances)
  | CEAnyElem (e1,e2) ->
      let sym = cast_to_elem_symbol (eval_cexpr dyn_ctxt e1) in
      let instances = eval_cexpr dyn_ctxt e2 in
      _Extended_Value_of_Value (element_constructor sym instances)
  | CEAttr (sym,e1) ->
      let items = eval_cexpr dyn_ctxt e1 in
      _Extended_Value_of_Value (attribute_constructor sym items)
  | CEAnyAttr (e1,e2) ->
      let sym = cast_to_attr_symbol(eval_cexpr dyn_ctxt e1) in
      let items = eval_cexpr dyn_ctxt e2 in
      _Extended_Value_of_Value (attribute_constructor sym items)
  | CEError e ->
      let v = eval_cexpr dyn_ctxt e in 
      if (sequenceIsEmpty v)
      then
	Fn_error.raise_empty_error ()
      else
	Fn_error.raise_item_error v

   (* JS: no need to perform validation there
      because strong typing makes sure this will succeed *)
   (* JS: on the other hand, wouldn't validation annotate the data
      with appropriate type information ? *)

  | CETreat (e,m) ->
      _Extended_Value_of_Value (eval_cexpr dyn_ctxt e)
   (* JS: no need to perform validation there
      because strong typing makes sure this will succeed *)
   (* JS: on the other hand, wouldn't validation annotate the data
      with appropriate type information ? *)

  | CEValidate (e,m) ->
      raise (Query (Prototype "XML Schema validation not implemented as an expression"))
   (* JS: no need to perform validation there
      because strong typing makes sure this will succeed *)
   (* JS: on the other hand, wouldn't validation annotate the data
      with appropriate type information ? *)

  | CECast (e,m) ->
      let e' = eval_cexpr dyn_ctxt e in
      if (sequenceIsEmpty e' && Typeutil.is_anyatomic_optional schema m) then
        _Extended_Value_of_Value sequenceEmpty
      else
	let base_value =
	    try
	      getAtomicValue(sequenceGetSingle e')
	    with
	    | _ ->
		raise (Query (Expr_Error (fi,
					  "Cast: " ^
					  (Print.bprintf_cexpr "" e0) ^ 
					  " not applied to a single atomic value")))
	in
	  let (m', _, _) = Typeutil.factor schema (Simplification.simplify_ty schema m) in
	  let base_type =
	    match m'.pmodel_desc with
	    | TBase bt -> bt
	    | _ ->
		raise (Query (Expr_Error (fi,
					  "Cast: " ^
					  (Print.bprintf_cexpr "" e0) ^ 
					  " to a non-simple type")))
	  in
	  let new_base_value = Cast_as.eval_cast nsenv base_type base_value in
	  _Extended_Value_of_Value (sequenceBuildSingle (buildAtomicValue new_base_value))

  | CECastable (e,m) ->
      begin
	try
	  let e' = eval_cexpr dyn_ctxt e in
	  if (sequenceIsEmpty e' && Typeutil.is_anyatomic_optional schema m) then
            _Extended_Value_of_Value sequenceEmpty
	  else
	    let base_value =
	      try
		getAtomicValue(sequenceGetSingle e')
	      with
	      | _ ->
		  raise (Query (Expr_Error (fi,
					    "Cast: " ^
					    (Print.bprintf_cexpr "" e0) ^ 
					    " not applied to a single atomic value")))
	    in
	    let (m', _, _) = Typeutil.factor schema (Simplification.simplify_ty schema m) in
	    let base_type =
	      match m'.pmodel_desc with
	      | TBase bt -> bt
	      | _ ->
		  raise (Query (Expr_Error (fi,
					    "Cast: " ^
					    (Print.bprintf_cexpr "" e0) ^ 
					    " to a non-simple type")))
	    in
	    let _ = Cast_as.eval_cast nsenv base_type base_value in
	    _Extended_Value_of_Value (sequenceBuildSingle (boolval (true)))
	with
	| _ ->
	    _Extended_Value_of_Value (sequenceBuildSingle (boolval (false)))
      end

  | CESubtype (m1,m2) ->
      let b = Typeutil.is_subtype_of schema m1 m2 in
      _Extended_Value_of_Value (sequenceBuildSingle (buildAtomicValue (buildAtomicBoolean b)))
	
  | CEIntersects (m1,m2) ->
      raise (Query (Prototype "Intersection between types not handled yet"))

  | CEForwardAxis (a,nt) ->
      let dot = var_value_from_dynamic_context dyn_ctxt Xquery_astutil.dot in

      (* Mary & Gargi: $dot may be bound to an atomic value *)
      let item = sequenceGetSingle dot in 
      let node_list = 
	if (isNode item) then 
          eval_axis a (getNode (sequenceGetSingle dot)) 
	else
	  (* Should we raise an error here? *)
	  []
      in
      _Extended_Value_of_Value (buildNodeSequence (sequence_of_list (List.filter (eval_node_test a nt) node_list)))

  | CEReverseAxis (a,nt) ->
      let dot = var_value_from_dynamic_context dyn_ctxt Xquery_astutil.dot in

      (* Mary & Gargi: $dot may be bound to an atomic value *)
      let item = sequenceGetSingle dot in 
      let node_list = 
	if (isNode item) then 
          eval_axis a (getNode (sequenceGetSingle dot)) 
	else
	  (* Should we raise an error here? *)
	  []
      in
      _Extended_Value_of_Value (buildNodeSequence (sequence_of_list (List.filter (eval_node_test a nt) node_list)))
  | CESome (odt,v,e1,e2) ->
      let input_list = list_of_sequence (eval_cexpr dyn_ctxt e1) in
      let eval_some x =
	let item = dynamic_type_check stat_ctxt fi odt (sequenceBuildSingle x) in
	let dyn_ctxt' = add_var_value_to_dynamic_context dyn_ctxt (v,item)
	in cast_to_bool (eval_cexpr dyn_ctxt' e2)
      in
      let b = List.exists eval_some input_list in
      _Extended_Value_of_Value (sequenceBuildSingle (boolval b))
  | CEEvery (odt,v,e1,e2) ->
      let input_list = list_of_sequence (eval_cexpr dyn_ctxt e1) in
      let eval_some x =
	let item = dynamic_type_check stat_ctxt fi odt (sequenceBuildSingle x) in
	let dyn_ctxt' = add_var_value_to_dynamic_context dyn_ctxt (v,item)
	in cast_to_bool (eval_cexpr dyn_ctxt' e2)
      in
      let b = List.for_all eval_some input_list in
      _Extended_Value_of_Value (sequenceBuildSingle (boolval b))
	
  | CECreateTuple rne_list ->
      let tuple_bindings =
	List.map (fun (rname,e1) -> (rname,eval_cexpr dyn_ctxt e1)) rne_list
      in
      _Extended_Value_of_Tuples (sequenceBuildSingle (create_tuple tuple_bindings))

  | CEAccessTuple (e1,rname) ->
      let one_tuple =
	sequenceGetSingle (_Tuples_of_Extended_Value (eval_cexpr_to_extended_value dyn_ctxt e1))
      in
      let tuple_component =
	try
	  access_tuple rname one_tuple
	with
	| _ ->
	    raise (Query (Internal_Error ("Variable '$"
					  ^ (Namespace.string_of_uqname rname)
					  ^ "' not found.")))
      in
      _Extended_Value_of_Value tuple_component


(**************)
(* Typeswitch *)
(**************)

(* Note:
     This part of the code deals with typeswitch evaluation.  It needs
     to be mutually recursive with the main eval_cexpr function.
   - Jerome *)

(* Note:
     The evaluation of match is pretty inefficient notably the use of
     the of_type function at run-time using computed types for arg
     should be taken into account.
   - Jerome *)

and eval_match dyn_ctxt arg cs =
  let stat_ctxt = static_context_from_dynamic_context dyn_ctxt in
  match cs with
  | (p,ovn,e) :: cs' ->
      begin
	match p.ppattern_desc with
	| Case m ->
	    if of_type stat_ctxt arg m
	    then
	      let dyn_ctxt' = 
		(match ovn with
		| None -> dyn_ctxt
		| Some vn -> add_var_value_to_dynamic_context dyn_ctxt (vn,arg))
	      in  eval_cexpr dyn_ctxt' e
	    else
	      eval_match dyn_ctxt arg cs'
	| Default ->
	    let dyn_ctxt' = 
	      (match ovn with
	      |	None -> dyn_ctxt
	      |	Some vn -> add_var_value_to_dynamic_context dyn_ctxt (vn,arg))
	    in
	    eval_cexpr dyn_ctxt' e 
      end
  | _ ->
      raise (Query (Internal_Error "No DEFAULT clause in TYPESWITCH expression"))

(***********)
(* Sort by *)
(***********)

and eval_sort dyn_ctxt v input_sequence e2 stable sk esk =

  (* The following function needs to be called to get the value of the
     sorting expression. *)

   (* Note:
       All of the evaluations of e2 must yield values (or nodes) of
       the same type.
     - Jerome *)

  let eval_e2 dot_value =
    (* Extends the dynamic environment *)
    let dyn_ctxt' = add_var_exvalue_to_dynamic_context dyn_ctxt (v, (_Extended_Value_of_Tuples (sequenceBuildSingle dot_value))) in
    (* Evaluates the sortspec expression *)
    eval_cexpr dyn_ctxt' e2
  in

  (* Here is the local comparison function *)

  (* Note:
       This applies both evaluation of the sorting expression, and
       actual comparison for each resulting value taking all the
       necessary parameters into account.
     - Jerome *)

  let compare_function input_value1 input_value2 =
    forest_compare sk esk (eval_e2 input_value1) (eval_e2 input_value2)
  in

  (* The sort function *)
  let sort_function =
    match stable with
    | Stable ->
	List.stable_sort
    | NonStable ->
	List.sort
  in

  (* Then sort !! *)

  _Extended_Value_of_Tuples (sequence_of_list (sort_function compare_function (list_of_sequence (_Tuples_of_Extended_Value input_sequence))))

