(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: eval_update.ml,v 1.8 2003/06/30 02:58:19 simeon Exp $ *)

(* Module Eval_update
   Description:
     This module contains the evaluation code for XML Query updates.
*)

open Format

open Error

open Namespace
open Sym

open Datatypes
open Xquery_ast
open Datamodel

open Norm_context
open Static_context
open Dynamic_context

open Eval_util
open Eval_expr

(**********)
(* Insert *)
(**********)

(* Auxilliary Insert Function for Insert Into / Insert As Last Into *)

let eval_csimple_insertaslastinto dyn_ctxt (update_content, par_node) fi =
  match par_node with
  | [] -> raise (Query (Update_Error (fi, "TargetNode for insertion is empty")))
  | [pn] -> (* to check whether the TargetLocation is a single node *)
      begin
	match pn with
	| Node parent_node -> 
	    let parent_nodekind = nodeKind parent_node in 
	    begin
	      match parent_nodekind with
	      | DocumentNodeKind -> insert (sequence_of_list update_content) parent_node None
	      | ElementNodeKind -> insert (sequence_of_list update_content) parent_node None
	      | _ -> raise (Query (Update_Error (fi, "TargetNode for Insert into not a Document or Element node")))
	    end
	| _ -> raise (Query (Update_Error (fi, "TargetNode for insert-as-last is not a node")))
      end
  | _ -> raise (Query (Update_Error (fi, "TargetNode for insertion is not a single node")))


(* Auxilliary Insert Function for Insert As First Into *)

let eval_csimple_insertasfirstinto dyn_ctxt (update_content, par_node) fi =  
  match par_node with
  | [] -> raise (Query (Update_Error (fi, "TargetNode for insertion is empty")))
  | [pn] -> (* to check whether the TargetLocation is a single node *)
      begin
	match pn with
	| Node parent_node ->
	    begin
	      let insert_location = get_first_child parent_node in 
	      let parent_nodekind = nodeKind parent_node in 
	      match parent_nodekind with
	      | DocumentNodeKind -> insert (sequence_of_list update_content) parent_node insert_location
	      | ElementNodeKind -> insert (sequence_of_list update_content) parent_node insert_location
	      | _ -> raise (Query (Update_Error (fi, "TargetNode for Insert into not a Document or Element node")))
	    end
	| _ -> raise (Query (Update_Error (fi, "TargetNode for insert-as-first is not a node")))
      end
  | _ -> raise (Query (Update_Error (fi, "TargetNode for insertion is not a single node")))


(* Auxilliary Insert Function for Insert Before *)

let eval_csimple_insertbefore dyn_ctxt (update_content, insertLocation) fi =
  match insertLocation with 
  | [] -> raise (Query (Update_Error (fi, "TargetNode for insertion is empty")))
  | [iloc] -> 
      begin
	match iloc with
	| Node insert_location ->
	    let parent_node =
	      try
		sequenceHead (nodeParent insert_location)
	      with
	      | _ ->
		  raise (Query (Update_Error (fi, "The insert location does not have any parent")))
	    in
            let insert_location_nodekind = nodeKind insert_location in 
	    begin
	      match insert_location_nodekind with
	      | ElementNodeKind -> insert (sequence_of_list update_content) parent_node (Some insert_location)
	      | CommentNodeKind -> insert (sequence_of_list update_content) parent_node (Some insert_location)
	      | ProcessingInstructionNodeKind -> insert (sequence_of_list update_content) parent_node (Some insert_location)
	      | _ -> raise (Query (Update_Error (fi, "Eval: Only element, comment and PI are valid TargetNodes in Insert before")))
	    end 
	| _ -> raise (Query (Update_Error (fi, "Eval: TargetNode for insert-before is not a node")))
      end
  | _ -> raise (Query (Update_Error (fi, "TargetNode for insertion is not a single node")))
  

(* Auxilliary Insert Function for Insert After *)

let eval_csimple_insertafter dyn_ctxt (update_content, after_node) fi =  
  match after_node with 
  | [] -> raise (Query (Update_Error (fi, "TargetNode for insertion is empty")))
  | [an] ->
      begin
	match an with
	| Node afterNode -> 
	    let parent_node =
	      try
		sequenceHead (nodeParent afterNode)
	      with
	      | _ ->
		  raise (Query (Update_Error (fi, "The insert location does not have any parent")))
	    in
	    let insert_location = get_next_sibling parent_node afterNode in
	    let after_nodekind = nodeKind afterNode in 
	    begin
	      match after_nodekind with
	      | ElementNodeKind -> insert (sequence_of_list update_content) parent_node insert_location
	      | CommentNodeKind -> insert (sequence_of_list update_content) parent_node insert_location
	      | ProcessingInstructionNodeKind -> insert (sequence_of_list update_content) parent_node insert_location
	      | _ -> raise (Query (Update_Error (fi, "Only element, comment and PI are valid TargetNodes in Insert before")))
	    end 
	| _ -> raise (Query (Update_Error (fi, "TargetNode for insert-after not a single node")))
      end
  | _ -> raise (Query (Update_Error (fi, "TargetNode for insertion is not a single node")))


(* Simple Insert *)

let eval_csimple_insert dyn_ctxt cil csi fi =
  let (ce, cil') = csi in 
  let update_content = list_of_sequence ce in 
  match cil with
  | CUInto il ->
      let par_node = list_of_sequence cil' in
      eval_csimple_insertaslastinto dyn_ctxt (update_content, par_node) fi

  | CUAsLastInto il ->
      let par_node = list_of_sequence cil' in
      eval_csimple_insertaslastinto dyn_ctxt (update_content, par_node) fi

  | CUAsFirstInto il ->
      let par_node = list_of_sequence cil' in
      eval_csimple_insertasfirstinto dyn_ctxt (update_content, par_node) fi

  | CUBefore il ->
      let (attrs, nodes) = filterAttributesFromSequence (sequence_of_list update_content) in 
      let attrs_list = list_of_sequence attrs in
      begin
        match attrs_list with 
        | [] ->                                        (* to check whether the UpdateContent has any attribute nodes - Gargi *)
            let insertLocation = list_of_sequence cil' in
	    eval_csimple_insertbefore dyn_ctxt (update_content, insertLocation) fi
        | _ -> 
	    raise (Query (Update_Error (fi, "UpdateContent in Insert before cannot have attributes")))
      end

  | CUAfter il ->
      let (attrs, nodes) = filterAttributesFromSequence (sequence_of_list update_content) in 
      let attrs_list = list_of_sequence attrs in
      begin
        match attrs_list with
        | [] ->                                         (* to check whether the UpdateContent has any attribute nodes *)
            let after_node = list_of_sequence cil' in
	    eval_csimple_insertafter dyn_ctxt (update_content, after_node) fi
        | _ -> 
	    raise (Query (Update_Error (fi, "UpdateContent in Insert after cannot have attributes")))
      end

(***********)
(* Replace *)
(***********)

  (* Steps to evaluate Replace - Gargi

     - Evaluate targetnode
     - Check if targetnode is singlenode
     - Evaluate  replacecontent
     - Depending targetnodetype, check if replacecontent type is appropriate
     - Call replace primitive: replace_node or replace_value depending on value_of_flag
   *)

(* Auxilliary functions for replace *)

(* A ContentSequence is any sequence of zero or more element, PI, comment nodes and atomic values - Gargi *)

let cstest n = 
  if (isElementNode n || isProcessingInstructionNode n || isCommentNode n)
  then true
  else false
      
let rec checkContentSequence node_list = 
  match node_list with 
  | [] -> true
  | n :: node_list' ->
	cstest n && checkContentSequence node_list'

let isContentSequence n = 
  let (nodes, avs) = filterValue n in
  checkContentSequence (list_of_sequence nodes)
	
let eval_csimple_replace_node n new_content fi = 
  let parent_node = list_of_sequence (parent_axis n) in 
  match parent_node with
  | [] ->
      raise (Query (Update_Error (fi, "Replace operation: Parent of targetnode does not exist!")))
  | [pn] -> 
      begin
	match n with
	| DocumentNode doc_node ->
	    raise (Query (Update_Error (fi, "Replace operation: Replacing document node not supported yet")))
	| ElementNode elem_node -> 
	    if (isContentSequence new_content) 
	    then replace_node new_content n pn
	    else raise (Query (Update_Error (fi, "Replace operation: UpdateContent for replace is not a ContentSequence")))
	| AttributeNode attr_node ->
	    let (attrs, nodes) = filterAttributesFromSequence new_content in
	    if (sequenceIsEmpty nodes) 
	    then replace_node new_content n pn
	    else raise (Query (Update_Error (fi, "Replace operation: Attribute node can only be replaced with other attribute nodes")))
	| TextNode text_node ->
	    if (isContentSequence new_content) 
	    then replace_node new_content n pn
	    else raise (Query (Update_Error (fi, "Replace operation: UpdateContent for replacing text node is not a ContentSequence")))
	| ProcessingInstructionNode pi_node ->
	    if (isContentSequence new_content) 
	    then replace_node new_content n pn
	    else raise (Query (Update_Error (fi, "Replace operation: UpdateContent for replacing PI node is not a ContentSequence")))
	| CommentNode comment_node ->
	    if (isContentSequence new_content) 
	    then replace_node new_content n pn
	    else raise (Query (Update_Error (fi, "Replace operation: UpdateContent for replacing comment node is not a ContentSequence")))
	| NamespaceNode ns_node ->
	    raise (Query (Update_Error (fi, "Replace operation: Cannot replace a namespace node")))
      end
  | _ -> raise (Query (Update_Error (fi, "Replace operation: Multiple parent nodes of target node")))
	      
let eval_csimple_replace_value n new_content fi = 
  match n with
  | DocumentNode doc_node ->
      raise (Query (Update_Error (fi, "Replace operation: Replace value of document node not allowed")))
  | ElementNode elem_node ->
      let (nodes, avs) = filterValue new_content in 
      if (sequenceIsEmpty nodes)
      then replace_value new_content n
      else raise (Query (Update_Error (fi, "Replace operation: UpdateContent for replace is not a ContentSequence")))
  | AttributeNode attr_node ->
      let (nodes, avs) = filterValue new_content in 
      if (sequenceIsEmpty nodes) 
      then replace_value new_content n
      else raise (Query (Update_Error (fi, "Replace operation: Attribute node value can be replaced only with sequence of atomic values")))
  | TextNode text_node ->
      raise (Query (Update_Error (fi, "Replace operation: Replace value of text node not allowed")))
  | ProcessingInstructionNode pi_node ->
      raise (Query (Update_Error (fi, "Replace operation: Replace value of PI node not allowed")))
  | CommentNode comment_node ->
      raise (Query (Update_Error (fi, "Replace operation: Replace value of comment node not allowed")))
  | NamespaceNode ns_node ->
      raise (Query (Update_Error (fi, "Replace operation: Replace value of namespace node not allowed")))


(* Simple Replace *)

let eval_csimple_replace dyn_ctxt vol tn new_content fi = 
  let target_node = list_of_sequence tn in
  match target_node with                                           (* target_node must be a single node - Gargi *)
  | [] ->
        raise (Query (Update_Error (fi, "Replace operation: Target node for replace is empty")))
  | [n] ->
      begin
	match n with 
	| Node node -> 
	  begin
	    match vol with 
	    | Normal_Replace ->
		eval_csimple_replace_node node new_content fi
	    | Value_Of_Replace ->
		eval_csimple_replace_value node new_content fi
	  end
	| AtomicValue _ ->
	    raise (Query (Update_Error (fi, "Replace operation: Target node for replace is an atomic value!")))
      end
  | n :: _ ->
        raise (Query (Update_Error (fi, "Replace operation: Target node for replace is not a single node")))


(*****************)
(* Simple Updates*)
(*****************)

(* Evaluation of Simple Updates *)

let eval_csimple_update dyn_ctxt csu =
  let fi = csu.pcsimple_update_loc in
  match csu.pcsimple_update_desc with
  | CUInsert (ce1, cil) ->
      let ce1' = eval_cexpr dyn_ctxt ce1 in 
      let cil' =
      begin
	match cil with
	| CUInto il -> eval_cexpr dyn_ctxt il
	| CUAsLastInto il -> eval_cexpr dyn_ctxt il
	| CUAsFirstInto il -> eval_cexpr dyn_ctxt il
	| CUBefore il -> eval_cexpr dyn_ctxt il
	| CUAfter il -> eval_cexpr dyn_ctxt il
      end
      in 
      eval_csimple_insert dyn_ctxt cil (ce1', cil') fi
  | CUDelete ce1 ->
      let nodes = list_of_sequence (eval_cexpr dyn_ctxt ce1) in
      begin
	(* The datamodel delete node operation only takes single nodes *)
	try
	  List.iter (fun x -> delete_node_with_children (sequenceBuildSingle x)) nodes
	with
	| (Query (Datamodel msg)) ->
	    raise (Query (Update_Error (fi,msg)))
      end
  | CUReplace (vol,ce1,ce2) ->
      let target_node = eval_cexpr dyn_ctxt ce1 in 
      let new_content = eval_cexpr dyn_ctxt ce2 in 
      eval_csimple_replace dyn_ctxt vol target_node new_content fi
  | CUEmpty ->
      (* The empty update does nothing! *)
      ()

(*******************)
(* Complex Updates *)
(*******************)

(* Evaluation of Conditional Updates *)

let eval_cond_update dyn_ctxt (ce, cs1, cs2) fi =
  if (cast_to_bool (eval_cexpr dyn_ctxt ce))
  then eval_csimple_update dyn_ctxt cs1
  else eval_csimple_update dyn_ctxt cs2


(**************)
(* PreUpdates *)
(**************)

(* DataType used in pre-update processing in FLWOR Updates - Gargi *)

type _PreUpdate =
  | PreInsert of _Value * _Value
  | PreDelete of _Value
  | PreReplace of _Value * _Value
  | PreEmpty

let preUpdate_of_PreInsert (pi1, pi2) = 
  PreInsert (pi1, pi2)

let preUpdate_of_PreDelete pd =
  PreDelete pd

let preUpdate_of_PreReplace (pr1, pr2) =
  PreReplace (pr1, pr2)

let preUpdate_of_PreEmpty () =
  PreEmpty

let preInsert_of_PreUpdate pu fi =
  match pu with
  | PreInsert (pi1, pi2) -> (pi1, pi2)
  | _ -> 
      raise (Query (Update_Error (fi, "PreInsert component not found")))

let preDelete_of_PreUpdate pu fi =    
  match pu with
  | PreDelete pd -> pd
  | _ -> 
      raise (Query (Update_Error (fi, "PreDelete component not found")))

let preReplace_of_PreUpdate pu fi =
  match pu with
  | PreReplace (pr1, pr2) -> (pr1, pr2)
  | _ -> 
      raise (Query (Update_Error (fi, "PreReplace component not found")))


(*****************)
(* FLWOR Updates *)
(*****************)

   (* Steps to evaluate FLWOR updates - Gargi

      - Create the Tuples
      - Process preupdates with respect to each tuple value
      - Evaluate each simple update in the FLWOR update
    *)

(* Function to evaluate the preupdate value of a given expression 'ce'
   for a single tuple *)

let eval_cexpr_with_tuple dyn_ctxt ce v u_tuple = 
  let item = _Extended_Value_of_Tuples (sequenceBuildSingle u_tuple) in 
  let dyn_ctxt'= add_var_exvalue_to_dynamic_context dyn_ctxt (v, item) in
  eval_cexpr dyn_ctxt' ce

(* Function to evaluate the query part of a simple update *)

let eval_preupdate dyn_ctxt simple_update v u_tuple =
  match simple_update.pcsimple_update_desc with
  | CUInsert (ce1, cil) -> 
      let v1 = eval_cexpr_with_tuple dyn_ctxt ce1 v u_tuple in 
      let v2 = 
	begin
	  match cil with
	  | CUAsLastInto ceil ->
	      eval_cexpr_with_tuple dyn_ctxt ceil v u_tuple
	  | CUAsFirstInto ceil ->
	      eval_cexpr_with_tuple dyn_ctxt ceil v u_tuple
	  | CUInto ceil ->
	      eval_cexpr_with_tuple dyn_ctxt ceil v u_tuple
	  | CUAfter ceil ->
	      eval_cexpr_with_tuple dyn_ctxt ceil v u_tuple
	  | CUBefore ceil ->
	      eval_cexpr_with_tuple dyn_ctxt ceil v u_tuple
	end
      in 
      preUpdate_of_PreInsert (v1, v2)
  | CUDelete ce1 -> 
      preUpdate_of_PreDelete (eval_cexpr_with_tuple dyn_ctxt ce1 v u_tuple)
  | CUReplace (vol,ce1,ce2) ->
      let v1 = eval_cexpr_with_tuple dyn_ctxt ce1 v u_tuple in 
      let v2 = eval_cexpr_with_tuple dyn_ctxt ce2 v u_tuple in 
      preUpdate_of_PreReplace (v1, v2)
  | CUEmpty -> 
      preUpdate_of_PreEmpty ()


(* Function to evaluate the preupdate clauses in a list of simple updates *)

let rec eval_preupdate_list dyn_ctxt csl v u_tuple = 
  match csl with
  | [] -> [] 
  | hd :: csl' -> 
      (eval_preupdate dyn_ctxt hd v u_tuple) :: eval_preupdate_list dyn_ctxt csl' v u_tuple


(* Function to evaluate a single update in a FLWOR update *)

let eval_csimple_in_flwor dyn_ctxt query_preupdate simple_update = 
  let fi = simple_update.pcsimple_update_loc in
  match simple_update.pcsimple_update_desc with 
  | CUInsert (ce, cil) ->
      let (ce1, cil1) = preInsert_of_PreUpdate query_preupdate fi in
      eval_csimple_insert dyn_ctxt cil (ce1, cil1) fi
  | CUDelete ce ->
      let nodes = list_of_sequence (preDelete_of_PreUpdate query_preupdate fi) in
      begin
	(* The datamodel delete node operation only takes single nodes *)
	try
	  List.iter (fun x -> delete_node_with_children (sequenceBuildSingle x)) nodes
	with
	| (Query (Datamodel msg)) ->
	    raise (Query (Update_Error (fi,msg)))
      end
  | CUReplace (vol,ce1,ce2) ->
      let (v1, v2) = preReplace_of_PreUpdate query_preupdate fi in 
      eval_csimple_replace dyn_ctxt vol v1 v2 fi
  | CUEmpty -> ()


(* Function to evaluate a list of simple updates in a FLWOR Update *)

let rec eval_updatelist_in_flwor dyn_ctxt query_preupdate_list csl fi = 
  match csl with
  | [] -> () (* do nothing! *)
  | uhead :: csl' ->
      begin
        match query_preupdate_list with 
        | [] ->
            raise (Query (Update_Error (fi, "Missing PreUpdate value in evaluation of FLOWR update")))
        | phead :: query_preupdate_list' -> 
            (eval_csimple_in_flwor dyn_ctxt phead uhead);
            eval_updatelist_in_flwor dyn_ctxt query_preupdate_list' csl' fi
      end


(* Evaluation of FLWOR Update *)

let eval_flwor_update dyn_ctxt (v, ce, csl) fi =
  
  (* create the tuple sequence *)
  let u_tuples = _Tuples_of_Extended_Value (eval_cexpr_to_extended_value dyn_ctxt ce)
  in 
  
  (* for each tuple in the tuple-sequence, do pre-update processing *)
  let preupdate_lists =
    let eval_tuple_preupdate u_tuple = 
      eval_preupdate_list dyn_ctxt csl v u_tuple       (* builds a list of preupdate values for one tuple *)
    in
    List.map eval_tuple_preupdate (list_of_sequence u_tuples)
  in

  (* for each tuple in the tuple-sequence, evaluate the update expression *)
  let eval_flwor u_preupdate_list = 
    (* evaluate the simple updates *)
    eval_updatelist_in_flwor dyn_ctxt u_preupdate_list csl fi
  in 
  List.iter eval_flwor preupdate_lists

(* Evaluation of Complex Updates *)

let eval_ccomplex_update dyn_ctxt ccu =
  let fi = ccu.pccomplex_update_loc in
  match ccu.pccomplex_update_desc with
  | CUForTuple (vname, ce, csl) ->
      eval_flwor_update dyn_ctxt (vname, ce, csl) fi
  | CUCond (ce, cs1, cs2) ->
      eval_cond_update dyn_ctxt (ce, cs1, cs2) fi

let eval_cupdate dyn_ctxt cu =
  match cu.pcupdate_desc with
  | CUSimple csu ->
      eval_csimple_update dyn_ctxt csu
  | CUComplex ccu ->
      eval_ccomplex_update dyn_ctxt ccu


