(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: eval_util.ml,v 1.5 2003/06/21 15:30:55 simeon Exp $ *)

(* Module Eval_util
   Description:
     This module contains some auxiliary evaluation functions, notably
     for function calls, type declarations, axis, element and
     attribute construction.
*)

open Format

open Error

open Namespace
open Sym

open Datatypes
open Xquery_ast
open Datamodel

open Norm_context
open Static_context
open Dynamic_context


(* Note:
     The following functions are used for convenience during
     evaluation. They are organized by XQuery expression.
     - Jerome
   *)

(**************)
(* Typeswitch *)
(**************)

(* Note:
     This part of the code deals contains typeswitch related
     functions.
   - Jerome *)

(* JS: this is a fake of_type operation
       just a hack to get started
       this only checks for the element name *)

(* J&M: of_type_by_name (loosely) tests whether a value is in the
   domain of a type.  For elements and attributes, it *only* compares
   for equality the element/attribute value's name with the type's
   name.  It does not recursively check the node's content.

   For simple values, it *only* checks that the type of the value and
   datatype are equal.  It should use type subsumption on simple types. 

   Group references and 'element of type' are not supported. 
*)

let of_type_by_name resolved_sym dtk =
  match (resolved_sym, dtk) with

    (* This case is handled in of_type below *)
  | (_, DTEmpty) -> 
      false 

    (* item wildcard *)
  | (_, DTItem) -> 
      true 

    (* Mary & Gargi: PIs and comments are nodes, but namespace nodes
    and atomic values are not. *)

    (* node wildcard *)
  | (RUnitSimpleSymbol sym, DTNode) ->
      (sym = Sym.pisym || sym = Sym.commentsym)

  | (_, DTNode) ->
      true

    (* attribute wildcard *)
  | (RUnitAttrSymbol _, DTAttrRef None) ->
      true
  | (_, DTAttrRef None) ->
      false

    (* element wildcard *)
  | (RUnitElemSymbol _, DTElemRef None) ->
      true
  | (_, DTElemRef None) ->
      false

    (* text node *)
  | (RUnitTextSymbol, DTText) ->
      true

  | (_, DTText) ->
      false

    (* atomic wildcard *)
  | (RUnitSimpleSymbol _, DTAtomic) -> 
      true 
  | (RUnitSimpleSymbol _, DTTypeRef anyatomic) -> 
      true 
  | (_, DTAtomic) ->
      false


    (* numeric wildcard *)
  | (RUnitSimpleSymbol sym1, DTNumeric) -> 
      (* J&M: This should be replaced by a subsumption test on numeric types, because 
         we want to heed simple-type hierarchy *)
      (sym1 = decimalsym || sym1 = integersym || sym1 = intsym || sym1 = floatsym || sym1 = doublesym)
  | (_, DTNumeric) ->
      false


    (* Simple Types *)
  | (RUnitSimpleSymbol sym1, DTBase st2) ->
      let sym2 = unit_symbol_of_base_type st2 in
      (* J&M: This should be replaced by a subsumption test on simple types, because 
         we want to heed simple-type hierarchy *)
      (* J&M: Before adding Text type, this was: *)
      (* simple_equals sym1 sym2 || ((sym1 = textsym) && sym2 = unknownsimpletypesym)) *)
      simple_equals sym1 sym2 

  | (_, DTBase _) ->
      false

    (* document of type *)
  | (RUnitDocSymbol, DTDocumentOfType group_sym2) ->
      raise (Query (Prototype "document of type not supported in SequenceType"))


  | (_, DTDocumentOfType _) ->
      false

    (* document of element *)
  | (RUnitDocSymbol, DTDocument None) ->
      true

  | (RUnitDocSymbol, DTDocument (Some (RElemSym elem_sym2))) ->
      true  (* should be replaced *)

  | (_, DTDocument _) ->
      false

    (* attribute refs *)
  | (RUnitAttrSymbol attr_sym1, DTAttrRef (Some (RAttrSym attr_sym2))) ->
      (* J&M - For complete structural subtyping, we should check that
         the attribute value is an item of the attribute type.  *)
      rattr_subtag attr_sym1 attr_sym2

  | (_, DTAttrRef _) ->
      false

    (* element refs *)
  | (RUnitElemSymbol elem_sym1, DTElemRef (Some (RElemSym elem_sym2))) ->
      relem_subtag elem_sym1 elem_sym2

  | (_, DTElemRef _) ->
      false

    (* group ref *)
  | (_, DTTypeRef gs) ->
      raise (Query (Prototype ("Type references ("^(string_of_group_symbol gs)^") not supported in SequenceType")))

    (* element of group type *)
  | (_, DTElemOfType group_sym2) ->
      raise (Query (Prototype "element of type not supported in SequenceType"))

    (* element of base type *)
  | (_, DTElemOfBase st2) ->
      raise (Query (Prototype "element of base type not supported in SequenceType"))

(* of_type checks that a value is an instance of a given datatype. *)
let of_type stat_ctxt arg dt =
  let (dtk,b1,b2) = Typeutil.factor_datatype dt in
  match dtk with
  | DTEmpty -> sequenceIsEmpty arg
  | _ -> 
      let is_of_type of_type_bool singlearg = 
	let resolved_argname = get_val_name singlearg in 
	(of_type_by_name resolved_argname dtk) && of_type_bool
      in 
      let b = Occurrence.occurs (sequenceLength arg) in
      (* Check the bounds *)
      (sequenceFold_left is_of_type true arg) && (Occurrence.le b b2) && (Occurrence.le b1 b)


(* promotes_to checks that a value can be promoted to a given
   datatype and returns the promoted value. 

  If the value matches the target type, then it is promoted to itself

   Value matches Type
   -----------------------------------
   Value against Type promotes to Value

  If the value does not match the target type, but matches a type
  which can be promoted to the target type, then the value is cast to
  the target type.

  statEnv |- Value1 matches Type1
  statEnv |- Type1 can be promoted to Type2
  statEnv |- Type1 != Type2
  cast as Type2 (Value1) => Value2 
  --------------------------------------------------
  statEnv |- Value1 against Type2 promotes to Value2

  Value promotes to Type as Value'
  ----------------------------------
  Value converts to Type as Value'

  Value : untypedAtomic
  Type <: xs:anySimpleType
  cast Value as (Type) => Value'
  ------------------------------
  Value converts to Type as Value'

*)

let promotes_to stat_ctxt (arg, (dt, m)) =
  let nsenv = Norm_context.nsenv_from_norm_context(Static_context.norm_context_from_stat_context stat_ctxt) in
  let (dtk,b1,b2) = Typeutil.factor_datatype dt in
  let b = Occurrence.occurs (sequenceLength arg) in
  match dtk with
  | DTEmpty -> if (sequenceIsEmpty arg) then (Some arg) else None
  (* For base types, promotion may require casting *)
  | DTBase bt ->
      let argl = list_of_sequence arg in
      let promote_item_to item = 
	let av = getAtomicValue item in
	let at = (atomic_type_of_atomic_value av) in

	(* untypedAtomic data is cast to the target type *)
	if (at = ATUntyped) then 
	  begin
	  try 
	    Some (buildAtomicValue (Cast_as.eval_cast nsenv bt av))
	  with
	  | Query(Cast_Error(_)) -> None
	  | Query(Validation(_)) -> None
	  end
        (* All other atomic types are promoted *)
	else
	  let atl = (can_be_promoted_to at) in
	  if (List.mem bt atl) then 
	    if (bt = at) then (Some item)
	    else (Some (buildAtomicValue (Cast_as.eval_cast nsenv bt av)))
	  else None
      in
      let argl' = Gmisc.unwrap_option_list (List.map promote_item_to argl) in
      if (List.length argl = List.length argl') && (Occurrence.le b b2) && (Occurrence.le b1 b) then 
	Some (sequence_of_list argl')
      else
	None
  (* For non-base type, promotion matches value against required type *)
  | _ -> 
      let is_of_type of_type_bool singlearg = 
	let resolved_argname = get_val_name singlearg in 
	(of_type_by_name resolved_argname dtk) && of_type_bool
      in 
      (* Check the bounds & if every item matches, return the value *)
      if (sequenceFold_left is_of_type true arg) && (Occurrence.le b b2) && (Occurrence.le b1 b) then
	Some arg
      else None

(***********)
(* Casting *)
(***********)

(* Note:
     This part of the code deals contains cast as related
     functions.
   - Jerome *)

(* cast operations *)

let cast_to_elem_symbol n =
  try 
    let v = getAtomicValue (sequenceNth n 0) in
    let qname = 
      match getAtomicValueKind v with
      | AtomicStringKind ->
	  (NSUri "", getAtomicString v)
      | AtomicQNameKind ->
	  Sym.anon_name (getAtomicQName v)
      | _ ->
	  raise (Query (Internal_Error "Computed tag not a Qname or string"))
    in Sym.relem_symbol (NSDefaultElementPrefix, qname)
  with
  | Failure listop -> raise (Query (Internal_Error "Failure in Eval_util.cast_to_elem_symbol"))

let cast_to_attr_symbol n =
  try
    let v = getAtomicValue (sequenceNth n 0) in
    let qname = 
      match getAtomicValueKind v with
      | AtomicStringKind ->
	  (NSUri "", getAtomicString v)
      | AtomicQNameKind ->
	  Sym.anon_name (getAtomicQName v)
      | _ ->
	  raise (Query (Internal_Error "Computed tag not a Qname or string"))
    in Sym.rattr_symbol (NSDefaultElementPrefix, qname)
  with
  | Failure listop -> raise (Query (Internal_Error "Failure in Eval_util.cast_to_attr_symbol"))

let cast_to_bool n =
  try
    getAtomicBoolean (getAtomicValue (sequenceNth n 0))
  with
  | Failure listop -> raise (Query (Internal_Error "Failure in Eval_util.cast_to_bool"))


(*********)
(* XPath *)
(*********)

(* Note:
     This part of the code deals contains XPath steps related
     functions.
   - Jerome *)

(* Return list of nodes along an axis *)

let eval_axis a node =
  match a with
  | Ancestor ->
      list_of_sequence (ancestor_axis node)
  | Ancestor_or_self ->
      list_of_sequence (ancestor_or_self_axis node)
  | Attribute ->
      list_of_sequence (attribute_axis node)
  | Child ->
      list_of_sequence (child_axis node)
  | Descendant ->
      list_of_sequence (descendant_axis node)
  | Descendant_or_self ->
      list_of_sequence (descendant_or_self_axis node)
  | Namespace ->
      list_of_sequence (namespace_axis node)
  | Parent ->
      list_of_sequence (parent_axis node)
  | Self ->
      list_of_sequence (self_axis node)

(* Node test checking *)

let eval_node_test a nt node =
  match nt with
  | CPNameTest qn ->
      begin
	let nodesym = getNodeSym node in
	match nodesym with 
	| None -> false
	| Some sym -> 
	    let qnsym =
	      begin
		match a with
		| Attribute -> RUnitAttrSymbol (rattr_symbol (NSDefaultElementPrefix, qn))
		| Namespace -> raise (Query (Prototype "In typing: name tests on namespaces not yet supported"))
		| _ -> RUnitElemSymbol (relem_symbol (NSDefaultElementPrefix, qn))
	      end
	    in
            (principal_node_kind a) = (nodeKind node) &&
	    Sym.subtag sym qnsym
      end
  | CPNodeKindTest node_kind ->
      begin
	match node_kind with
	| PIKind pik ->
	    nodeKind node = ProcessingInstructionNodeKind
	      &&
	    begin
	      match pik with
	      |	None -> true
	      |	Some name -> (piTarget (getProcessingInstructionNode node)) = name
	    end
	| CommentKind ->
	    nodeKind node = CommentNodeKind
	| TextKind ->
	    nodeKind node = TextNodeKind
	| AnyKind ->
	    true
      end

(********)
(* Sort *)
(********)

(* Note:
     This part of the code deals contains XPath sort related
     functions.
   - Jerome *)

(* Compare function between two arbitrary forest *)

(* Note:
     This function takes ascending/descending and
     emptygreatest/emptyleast into account.
   - Jerome *)

let forest_compare sk esk if1 if2 =
  let ascending_compare =
    (* First, extract the actual values for each forests *)
    let oi1 = extract_optional_item if1
    and oi2 = extract_optional_item if2 in
    match (oi1,oi2) with
      (* If both forests are empty, then they are equal *)
    | (None, None) ->
	0
      (* If only one of the forests is empty, then look at whether empty is greatest or least *)
    | (None, _) ->
	begin
	  match esk with
	  | EmptyGreatest ->
	      1
	  | EmptyLeast ->
	      -1
	end
    | (_, None) ->
	begin
	  match esk with
	  | EmptyGreatest ->
	      -1
	  | EmptyLeast ->
	      1
	end
      (* Otherwise, use the item_compare function from the data model *)
    | (Some i1, Some i2) ->
	item_compare i1 i2
  in
  (* Finally, deal with the sorting kind ascending or descending *)
  match sk with
  | Ascending ->
      ascending_compare
  | Descending ->
      -ascending_compare


(* 

  Dynamic semantics of function calls requires the following judgments: 

*)

(* value_matches_type checks whether a value matches a given type *)
let value_matches_type stat_ctxt (v, (dt, m)) = 
  of_type stat_ctxt v dt

(* match_function_signature finds the first function signature such
   that the given actual function arguments match the signature, after
   promotion.  
*)
let match_function_signature stat_ctxt fi fn eargs signatures =
  let nsenv = Norm_context.nsenv_from_norm_context(Static_context.norm_context_from_stat_context stat_ctxt) in
  let rec args_match_signature signatures = 
    match signatures with 
    | [] -> raise Not_found
    | (fn, (input_types, output_type))::sigs ->
	begin
	  let vts = List.combine eargs input_types in
	  let eargs' = Gmisc.unwrap_option_list(List.map (promotes_to stat_ctxt) vts) in 
	  if (List.length eargs = List.length eargs') then
	    (fn, eargs')
	  else
	    args_match_signature sigs
	end
  in 
  try
    args_match_signature signatures 
  with
    Not_found -> raise (Query (Type_Error (fi, ("Eval: Arguments to function '"
						^ (Namespace.string_of_rqname fn) 
						^ "' do not match function's signature.\n"))))

(* dynamic_type_check checks dynamically that a value matches a type *)

let dynamic_type_check stat_ctxt fi odt forest =
  match odt with
  | None -> forest
  | Some dt -> 
      try 
	if (of_type stat_ctxt forest dt) then 
	  forest 
	else raise (Query (Type_Error (fi, ("Type declaration failed.  Type of value:\n"
					    ^ (Serialization.bserialize_datamodel forest)
					    ^ "\n does not match type:\n"
					    ^ (Print.bprintf_datatype "    " dt)
					    ^ ".\n\n"))))
      with
      (* BUG: SequenceType matching (against DTTypeRef) is not fully implemented.  Instead
         of throwing an exception, we just continue silently -Mary *)
      | Query(Prototype(msg)) -> 
	  begin
	    match dt.pdatatype_desc with 
	    | (DTTypeRef gsym, _) -> forest
	    | _ -> raise (Query(Prototype(msg)))
	  end


(*********************************************)
(* Dynamic semantics of element construction *)
(*   See Typing.element_constructor for static semantics  *)
(*********************************************)

let element_constructor sym instances =
  let (attrs, items) = filterAttributesFromSequence instances in 
  let (nodes, values) = filterValue items in
  
  if (not(sequenceIsEmpty(nodes))) then
    (* If element contains both non-attribute nodes and values, 
       convert values to nodes before construction *)
    let nodes' = 
      if (not(sequenceIsEmpty(values))) then
	Datamodel.value_to_node_sequence items 
      else nodes
    in
    sequenceBuildSingle (buildNode
 	                   (buildElementNode (elemNode
						sym
						sequenceEmpty
						attrs
						nodes'
						[])))

  else
  (* Else if element contains only values, construct element with value content *)
    sequenceBuildSingle (buildNode
 	                   (buildElementNode (elemNodeWithValue
						sym
						sequenceEmpty
						attrs
                                                sequenceEmpty
						[]
					        values)))
	
(***********************************************)
(* Dynamic semantics of attribute construction *)
(***********************************************)
let attribute_constructor sym items = 
  let (nodes, values) = filterValue items in
  if (not(sequenceIsEmpty(nodes))) then
    (* If attribute contains both nodes and values, 
       convert nodes and values to string before construction *)
    let content = Datamodel.value_to_untyped items
    in sequenceBuildSingle (buildNode (buildAttributeNode
					 (attrNode
					    sym
					    content)))

  else
    sequenceBuildSingle (buildNode (buildAttributeNode
				      (attrNodeWithValue
					 sym
					 ""
					 values)))

