(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: rewrite_judge.ml,v 1.2 2003/05/28 14:47:38 simeon Exp $ *)

(* Module Rewrite_judge
   Description:
     This module implements some auxiliary judgments used during
     optimization analysis.
*)

open Error

open Datatypes
open Namespace

open Xquery_ast
open Xquery_astutil

open Typeutil


(**************************************************************************)
(*                    used_count function                                 *)
(* It determines the occurrance of variable $x in a given core expression *)
(**************************************************************************)

let rec used_count_aux variablename ce count =
  match ce.pcexpr_desc with
  | CESort (v, cexpr1, _, cexpr2, _, _) ->
      (* Expr1 sortby (Expr2 ascending) *)
     if variablename = v
     then count + (used_count_aux variablename cexpr1 count)
     else count + (used_count_aux variablename cexpr1 count) + (used_count_aux variablename cexpr2 count)

  | CEUnordered cexpr ->
      count + (used_count_aux variablename cexpr count)

  | CESome (_, vname, cexpr1, cexpr2)
  | CEEvery (_, vname, cexpr1, cexpr2)
  | CEFor (_, vname, None, cexpr1, cexpr2) ->
     if vname = variablename 
     then count + (used_count_aux variablename cexpr1 count)
     else count + (used_count_aux variablename cexpr1 count) + (used_count_aux variablename cexpr2 count)
      
  | CEFor (_, vname, Some vname', cexpr1, cexpr2) ->
     if (vname = variablename) || (vname' = variablename)
     then count + (used_count_aux variablename cexpr1 count)
     else count + (used_count_aux variablename cexpr1 count) + (used_count_aux variablename cexpr2 count)
      
  | CEForTuple (vname, cexpr1, cexpr2) ->
     if vname = variablename 
     then count + (used_count_aux variablename cexpr1 count)
     else count + (used_count_aux variablename cexpr1 count) + (used_count_aux variablename cexpr2 count)
      
  | CELet (_, vname, cexpr1, cexpr2) ->
     if vname = variablename
     then count + (used_count_aux variablename cexpr1 count)
     else count + (used_count_aux variablename cexpr1 count) + (used_count_aux variablename cexpr2 count)

  | CEIf (cexpr1, cexpr2, cexpr3) ->
      count + (used_count_aux variablename cexpr1 count) 
            + (used_count_aux variablename cexpr2 count) + (used_count_aux variablename cexpr3 count)
     
(* In XQuery:

   ----------------------------
   $x used in
   typeswitch Expr 
     case Type1 $v1 return Expr1
     case Type2 $v2 return Expr2
     ...
     default return Exprn

   We are matching against:
    CETypeswitch of (cexpr * (pattern * cexpr) list)
*)
    
  | CETypeswitch (cexpr, cases) ->
     count + (used_count_aux variablename cexpr count) + (used_count_in_cases variablename cases count)

  | CEVar vname -> 
      	      (* In XQuery: $x, $v, $fs:dot *)
     if vname = variablename 
     then count + 1 
     else count

  | CEScalar _ ->
     count
      	      (* In XQuery:
      		"Hello!"    literal string
      		1           literal integer
      		1.31231     literal float
	
      	       $v used in 1 *)
     
  | CEText _ -> 	              (* In XQuery: <a> This is text </a> <a> 1 </a> *)
      count
      
  | CETextComputed cexpr ->       	      (* In XQuery: text { Expr } *)
      count + used_count_aux variablename cexpr count
     
  | CEPI _ -> 
      count
      
  | CEComment _ -> 
      count
      
  | CEDocument cexpr ->       	      (* In XQuery: document { Expr } *)
      count + used_count_aux variablename cexpr count

  | CEApp (fname, arguments) ->
      used_count_in_arguments variablename arguments count

  | CESeq (cexpr1, cexpr2) ->
      count + (used_count_aux variablename cexpr1 count) + (used_count_aux variablename cexpr2 count)

  | CEEmpty -> 
      count
      
  | CEElem (relem_symbol, cexpr) ->
      count + (used_count_aux variablename cexpr count)
      
  | CEAnyElem (cexpr1, cexpr2) ->  
      count + (used_count_aux variablename cexpr1 count) + (used_count_aux variablename cexpr2 count)

  | CEAttr (rattr_symbol, cexpr) ->
      count + (used_count_aux variablename cexpr count)

  | CEAnyAttr (cexpr1, cexpr2) ->
      count + (used_count_aux variablename cexpr1 count) + (used_count_aux variablename cexpr2 count)	
 
  | CEError (_) ->
      count

  | CETreat (cexpr, model) ->
      count + (used_count_aux variablename cexpr count)

  | CEValidate (cexpr, model) ->
      count + (used_count_aux variablename cexpr count)

  | CECast (cexpr, model) ->
      count + (used_count_aux variablename cexpr count)

  | CECastable (cexpr, model) ->
      count + (used_count_aux variablename cexpr count)

  | CEForwardAxis (axis, cnode_test) ->
      if Xquery_astutil.dot = variablename  
      then count + 1
      else count

  | CEReverseAxis (axis, cnode_test) ->
      if Xquery_astutil.dot = variablename  
      then count + 1
      else count

  | CESubtype (model1, model2) ->
      count
  | CEIntersects (model1, model2) ->
      count
  | CEKeyRef _ ->
      count

  | CECreateTuple rne_list ->
      let arguments = snd (List.split rne_list) in
      used_count_in_arguments variablename arguments count

  | CEAccessTuple (cexpr1,rname) ->
      used_count_aux variablename cexpr1 count

(* Auxiliary function used in the used_count_aux judgment for case clauses *)
	
and used_count_in_cases variablename cases count =
  match cases with
  | [] -> count
  | (_,ovn,case_cexpr) :: other_cases ->
      count + (used_count_in_cases variablename other_cases count) +
      (match ovn with
      |	None ->
	  (used_count_aux variablename case_cexpr count) 
      |	Some vname ->
	  if (variablename = vname) then 0
	  else
	    (used_count_aux variablename case_cexpr count))

and used_count_in_arguments variablename arguments count =
  match arguments with
  | [] -> count
  | arg :: other_arguments ->
      count + (used_count_aux variablename arg count) 
	    + (used_count_in_arguments variablename other_arguments count) 

let used_count variablename ce =
  used_count_aux variablename ce 0

(******************************************************************)
(*                  is_safe_function judgement                    *)
(* Determines whether given built-in function can raise an error  *)
(******************************************************************)

let safe_functions_table =
  [ (fn_uri, "equal");
    (fn_uri, "nequal");
    (fn_uri, "nodeeq");
    (fn_uri, "precedes");
    (fn_uri, "follows");
    (fn_uri, "le");
    (fn_uri, "ge");
    (fn_uri, "gt");
    (fn_uri, "lt");
    (fn_uri, "empty");
    (fn_uri, "count");
	(* Note:
             We consider the following as true, but this is not
	     right if the user is using it!!  It works for all uses of
	     index-of generated through normalization.
           - Gargi *)
    (fn_uri, "index-of");
        (* Note:
             We remove data for now, since the semantics of data is
             still unclear and we do not know whether it might fail or
             not.
	   - Gargi and Jerome *)
    (*  (fn_uri, "data");  *)
  ]

let is_safe_function fname =
  (* Note: currently simple analysis only for built-in functions *)
  (List.exists (fun x -> x = fname) safe_functions_table)

(* Note:

   I had to change the previous pattern matching into an explicit
   lookup to deal with the namespace uri properly.

   The following functions where explicitly given as unsafe, but the
   default is unsafe.  Keep that information here for now.

     - Jerome

  (fn_uri, "int-divide")
  (fn_uri, "int-mod")
  (fn_uri, "decimal-divide")
  (fn_uri, "decimal-mod") ->
      false
*)

(*******************************************************************)
(*                    can_fail judgment                            *)
(* determines whether the given core expression can raise an error *)
(*******************************************************************)

(* MF: We do not distinguish between functions that can fail and
   "functions" that have side-effects -- "functions" with side-effects
   are lumped into the can-fail category. *)

let rec can_fail ce1 =
  match ce1.pcexpr_desc with
  | CESort (v, cexpr1, _, cexpr2, _, _) ->
      can_fail cexpr1 || can_fail cexpr2

  | CEUnordered cexpr1 ->
      can_fail cexpr1 

  | CESome (odt, vname, cexpr1, cexpr2)
  | CEEvery (odt, vname, cexpr1, cexpr2)
  | CEFor (odt, vname, _, cexpr1, cexpr2) ->
      can_fail cexpr1 || can_fail cexpr2 ||
      (match odt with
      |	None -> false
      |	Some _ ->       
	  if !Conf.typing then
	    false    (* Type declaration cannot fail if the static checking is doing its job *)
	  else
	    true)
  | CEForTuple (vname, cexpr1, cexpr2) ->
      can_fail cexpr1 || can_fail cexpr2
  | CELet (odt, vname, cexpr1, cexpr2) ->
      can_fail cexpr1 || can_fail cexpr2 ||
      (match odt with
      |	None -> false
      |	Some _ ->       
	  if !Conf.typing then
	    false    (* Type declaration cannot fail if the static checking is doing its job *)
	  else
	    true)

  | CEIf (cexpr1, cexpr2, cexpr3) ->
      can_fail cexpr1 || can_fail cexpr2 || can_fail cexpr3

  | CETypeswitch (cexpr1, cases) ->
      can_fail cexpr1 || can_fail_cases cases

  | CEVar _ ->
      false                                   (* It cannot fail if the static checking is doing its job *)
            (* Note: Even if static checking is off, we should detect unbound variables some other way. *)

  | CEScalar _ ->
      false

  | CEText _ ->
      false

  | CETextComputed cexpr ->
      can_fail cexpr

  | CEPI _ ->
      false

  | CEComment _ ->
      false

  | CEDocument cexpr ->
      can_fail cexpr

  | CEApp (fname, arguments) ->          (* LOOK for differences between BUILTIN and USER-DEFINED ONES *)
      if (is_safe_function fname) then
	begin
	  if !Conf.typing then
	    false
	  else
	    true
	end
      else
	true

  | CECreateTuple rne_tuple ->
      List.exists (fun (rname,cexpr1) -> can_fail cexpr1) rne_tuple

	(* Assumption here is that the system only generates correct
   	   access to tuples!
	  - Jerome *)
  | CEAccessTuple (cexpr1,rname) ->
      can_fail cexpr1

  | CESeq (cexpr1, cexpr2) ->
      can_fail cexpr1 || can_fail cexpr2

  | CEEmpty -> 
      false

  | CEElem (relem_symbol, cexpr) ->    (* element a { cexpr } *)
      can_fail cexpr                 (* According to the new semantics, element constructors never fail *)

  | CEAnyElem (cexpr1, cexpr2) ->
      if !Conf.typing then
	can_fail cexpr1 || can_fail cexpr2    (* It cannot fail if the static checking is doing its job *)
      else
	true (* It can fail dynamically if the expression computing
	        the name returns something else than a QName of a string.
		For instance : element { 1 } { "Hello" } *)

  | CEAttr (rattr_symbol, cexpr) -> 
      can_fail cexpr               (* According to the new semantics, attribute constructors never fail *)

  | CEAnyAttr (cexpr1, cexpr2) ->
      if !Conf.typing then
	can_fail cexpr1 || can_fail cexpr2    (* It cannot fail if the static checking is doing its job *)
      else
	true (* It can fail dynamically if the expression computing
	        the name returns something else than a QName of a string.
		For instance : attribute { 1 } { "Hello" } *)

  | CEError _ ->
      true      

  | CETreat (cexpr, _) ->
      true

  | CEValidate (cexpr, _) ->
      true

  | CECast (cexpr, _) -> 
      true

  | CECastable (cexpr, _) -> 
      true

  | CEForwardAxis (_, _) -> 
      false

  | CEReverseAxis (_, _) -> 
      false

  | CESubtype (_, _) ->
      false

  | CEIntersects (_, _) ->
      false

  | CEKeyRef _ ->
      true

and can_fail_cases cases = 
  match cases with
  | [] -> false
  | (_, _, case_cexpr) :: remaining_cases ->
      if (can_fail case_cexpr)
      then true
      else (can_fail_cases remaining_cases)

(******************************************************************)
(*                      is_ordered judgement                      *)
(* Determines whether the given core expression returns results   *)
(* in document order or not                                       *)
(******************************************************************)

let does_node_care_about_order ce =
  match ce.pcexpr_desc with
  | CESort _ ->
      true
  | CEUnordered _ ->
      true
  | CESome _ ->
      false
  | CEEvery _ ->
      false
  | CEFor _ ->
      false
  | CEForTuple _ ->
      false
  | CELet (None, _, _, _) ->
      false
  | CELet (Some _, _, _, _) ->
      false
  | CEIf _ ->
      false
  | CETypeswitch (cexpr, cases) ->
      false
  | CEVar vname -> 
      false
  | CEScalar _ ->
      false
  | CEText _ ->
      false
  | CETextComputed _ -> 
      false
  | CEPI _ -> 
      false
  | CEComment _ -> 
      false
  | CEDocument _ ->
      false
  | CEApp (fname, _) ->
      if ((fname = (glx_uri,"docorder"))
	||(fname = (glx_uri,"distinct-docorder")))
      then
	true
      else
	false
  | CESeq _ ->
      false
  | CEEmpty -> 
      false
  | CEElem _ ->
      false
  | CEAnyElem _ ->  
      false
  | CEAttr _ ->
      false
  | CEAnyAttr _ ->
      false
  | CEError _ ->
      false
  | CETreat _ ->
      false
  | CEValidate _ ->
      false
  | CECast _ ->
      false
  | CECastable _ ->
      false
  | CEForwardAxis _ ->
      false
  | CEReverseAxis _ ->
      false
  | CESubtype _ ->
      false
  | CEIntersects _ ->
      false
  | CEKeyRef _ ->
      false
  | CECreateTuple _ ->
      false
  | CEAccessTuple _ ->
      false


(* 
   Function to compute the intersection.
   Prototype: intersects tr model0 typek = true/false
*)

let intersects stat_ctxt cexpr0_model patternk_datatype =
  let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
  let (p0, m0, n0) = Typeutil.factor_with_units schema cexpr0_model in
  let (dtk1, m1, n1) = Typeutil.factor_datatype patternk_datatype in
  let t_model = Commontype.small_type stat_ctxt (p0, m0, n0) (dtk1, m1, n1) in
  match t_model.pmodel_desc with
  | TNone -> false
  |  _ -> true

