(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: rewrite_main.ml,v 1.4 2003/05/28 14:47:38 simeon Exp $ *)

(* Module Rewrite_main
   Description:
     This module contains the top-level rewriting engine operating on
     the XQuery core.
*)

open Error

open Namespace
open Datatypes

open Xquery_ast
open Xquery_astutil

open Typeutil

open Rewrite_rules


(***********************)
(* Expression rewriter *)
(***********************)

let rewrite_cexpr stat_ctxt ce = 
  let my_optim_list =
    if !Conf.typing 
    then
      all_rewrites
    else
      nontype_rewrites
  in
  Rewrite_walker.fix_point_tree_apply stat_ctxt my_optim_list ce


(*******************)
(* Update rewriter *)
(*******************)

(* Note:
    Currently, no rewriting of updates occurs!
   - Jerome
 *)

let rewrite_cinsert_location stat_ctxt cinsert_location =
  match cinsert_location with
  | CUAsLastInto ce1 ->
      CUAsLastInto (rewrite_cexpr stat_ctxt ce1)
  | CUAsFirstInto ce1 ->
      CUAsFirstInto (rewrite_cexpr stat_ctxt ce1)
  | CUInto ce1 ->
      CUInto (rewrite_cexpr stat_ctxt ce1)
  | CUAfter ce1 ->
      CUAfter (rewrite_cexpr stat_ctxt ce1)
  | CUBefore ce1 ->
      CUBefore (rewrite_cexpr stat_ctxt ce1)

let rewrite_csimple_update stat_ctxt csu =
  match csu.pcsimple_update_desc with
  | CUInsert (ce1, cinsert_location) ->
      fmkcsimple_update
	(CUInsert (rewrite_cexpr stat_ctxt ce1, rewrite_cinsert_location stat_ctxt cinsert_location))
	csu.pcsimple_update_loc
  | CUReplace (vof, ce1, ce2) ->
      fmkcsimple_update
	(CUReplace (vof, rewrite_cexpr stat_ctxt ce1, rewrite_cexpr stat_ctxt ce2))
	csu.pcsimple_update_loc
  | CUDelete ce1 ->
      fmkcsimple_update
	(CUDelete (rewrite_cexpr stat_ctxt ce1))
	csu.pcsimple_update_loc
  | CUEmpty ->
      csu

let rewrite_ccomplex_update stat_ctxt ccu =
  match ccu.pccomplex_update_desc with
  | CUForTuple (vname,ce1, csimple_updates) ->
      let new_simple_update_list =
	List.map
	  (rewrite_csimple_update stat_ctxt)
	  csimple_updates
      in
      fmkccomplex_update
	(CUForTuple (vname, (rewrite_cexpr stat_ctxt ce1),new_simple_update_list))
	ccu.pccomplex_update_loc
  | CUCond (ce1,csu1,csu2) ->
      fmkccomplex_update
	(CUCond ((rewrite_cexpr stat_ctxt ce1, rewrite_csimple_update stat_ctxt csu1, rewrite_csimple_update stat_ctxt csu2)))
	ccu.pccomplex_update_loc

let rewrite_cupdate stat_ctxt cu = 
  match cu.pcupdate_desc with
  | CUSimple csu ->
      fmkcupdate (CUSimple (rewrite_csimple_update stat_ctxt csu)) cu.pcupdate_origin cu.pcupdate_loc 
  | CUComplex ccu ->
      fmkcupdate (CUComplex (rewrite_ccomplex_update stat_ctxt ccu)) cu.pcupdate_origin cu.pcupdate_loc 

let rewrite_cstatement stat_ctxt cs = 
  match cs.pcstatement_desc with
  | CSExpression ce ->
      fmkcstatement (CSExpression (rewrite_cexpr stat_ctxt ce)) cs.pcstatement_loc
  | CSUpdate cu ->
      fmkcstatement (CSUpdate (rewrite_cupdate stat_ctxt cu)) cs.pcstatement_loc


(*************************)
(* Query prolog rewriter *)
(*************************)

(* Applying a rewriting pass to function declarations *)

let cfunction_body_simplify stat_ctxt cfunction_body =
  match cfunction_body with
  | CEFunctionBltIn _ -> cfunction_body
  | CEFunctionUser (vnl, ce) -> CEFunctionUser (vnl, rewrite_cexpr stat_ctxt ce)

let function_simplify stat_ctxt cfunction_def =
  let new_desc =
    match cfunction_def.pcfunction_def_desc with
    | (rfname, cfunction_signature, cfunction_body) ->
	(rfname, cfunction_signature, cfunction_body_simplify stat_ctxt cfunction_body)
  in
  fmkcfunction_def new_desc cfunction_def.pcfunction_loc

let var_simplify stat_ctxt cvar_decl =
  let new_desc =
    match cvar_decl.pcvar_decl_desc with
    | (vname, ce) ->
	(vname, rewrite_cexpr stat_ctxt ce)
  in
  fmkcvar_decl new_desc cvar_decl.pcvar_decl_loc

let key_simplify stat_ctxt ckey_def =
  let new_desc =
    match ckey_def.pckey_def_desc with
    | (name, ce1, ce2) ->
	(name, rewrite_cexpr stat_ctxt ce1, rewrite_cexpr stat_ctxt ce2)
  in
  fmkckey_def new_desc ckey_def.pckey_def_loc

(* Main optimizer call for the query prolog *)

let rewrite_module stat_ctxt cquery_module =
  { pcmodule_functions = List.map (function_simplify stat_ctxt) cquery_module.pcmodule_functions;
    pcmodule_vars = List.map (var_simplify stat_ctxt) cquery_module.pcmodule_vars;
    pcmodule_keys = List.map (key_simplify stat_ctxt) cquery_module.pcmodule_keys;
    pcmodule_exprs = cquery_module.pcmodule_exprs }

