(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: rewrite_rules.ml,v 1.4 2003/06/18 14:21:52 mff Exp $ *)

(* Module Rewrite_rules
   Description:
     This module contains rewriting rules on the XQuery core.
*)

open Error

open Namespace
open Datatypes

open Xquery_ast
open Xquery_astutil

open Typeutil
open Datamodel

open Rewrite_judge
open Rewrite_walker

(* One auxiliary function *)

let get_type_annotation cexpr = 
  match cexpr.pcexpr_type with 
  | Some model -> model 
  | None -> 
      Print.printf_cexpr "" cexpr;
      raise (Query (Internal_Error "Rewrite_rule : Core Expression doesn't have a type"))

(***********************************************************************)

(*********************)
(* IF SIMPLIFICATION *)
(*********************)

(* 
   if cexpr1 then cexpr2 else cexpr3
   cexpr1 := true
   ---------------------------------
   cexpr2

   if cexpr1 then cexpr2 else cexpr3
   cexpr1 := false
   ---------------------------------
   cexpr3
*)

let if_rewrite stat_ctxt (ce : cexpr) =
  match ce.pcexpr_desc with
  | CEIf (cexpr1, cexpr2, cexpr3) ->
      begin
        match cexpr1.pcexpr_desc with 
	| CEScalar(AtomicBoolean true) ->            
	    (cexpr2, true)
      	| CEScalar(AtomicBoolean false) -> 
	    (cexpr3, true)
	|  _ ->
	    raise Not_applied
      end
  | _ ->
      raise Not_applied

(***********************************************************************)

(***************************)
(* TYPESWITCH OPTIMIZATION *)
(***************************)

(*
   1. Remove empty case clauses: (This rule should be applied before 2 & 3, to each case clause)
   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~   

   typeswitch Expr_0 
   case Type_1 $v1 return Expr_1
   ...
   case Type_k $vk return Expr_k
   ...
   default $vd return Expr_n
   and 
   Expr_0 : Type_0  (i.e., Expr_0 statically known to have Type_0)
   and
   Type_k intersect Type_0 = None
   ----------------------------------------
   typeswitch Expr_0 
   case Type_1 $v1 return Expr_1
   ...
   case Type_(k-1) $v(k-1) return Expr_(k-1)
   case Type_(k+1) $v(k+1) return Expr_(k+2)
   ...
   default $vd return Expr_n

   2. Remove "never-evaluated" case clauses:
   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   (This rule should be applied until one of these subtyping relationships is true:
    Type_0 <: Type_1
    Type_0 <: (Type_1 | Type_2)
    ...
    Type_0 <: (Type_1 | ... | Type_k)
   )

   typeswitch Expr_0 
   case Type_1 $v1 return Expr_1
   ...
   case Type_k $vk return Expr_k
   ...
   case Type_(n-1) $v(n-1)return Expr_(n-1)
   default $vd return Expr_n
   and
   Type_0 <: (Type_1 | ... | Type_k)
   -----------------------------------------
   typeswitch Expr_0 
   case Type_1 $v1 return Expr_1
   ...
   case Type_(k-1) $v(k-1) return Expr_(k-1)
   default $v(k) return Expr_k

   3. Typeswitch elimination: (This rule should be applied after 2 & 3 and enables "let" simplification.)
   ~~~~~~~~~~~~~~~~~~~~~~~~~~

   typeswitch Expr_0 
   default $vd return Expr
   ----------------------------------------
   let $vd := Expr_0 return $Expr

*)

(*********************************************************)
(* Auxiliary functions to optimize typeswitch expression *)
(*********************************************************)

(* Function to remove the "empty" case clauses - Applies Rule: 1 *)

let rec remove_empty_clauses stat_ctxt cexpr_type0 cases =
  match cases with
  | [] -> ([],false)
  | (patternk, ovn, case_cexpr) :: remaining_cases ->
      begin
	match patternk.ppattern_desc with
	| Case patternk_datatype ->
	    if (intersects stat_ctxt cexpr_type0 patternk_datatype) 
	    then 
	      let (new_cases1, changed) = remove_empty_clauses stat_ctxt cexpr_type0 remaining_cases in
	      let new_cases2 = (patternk, ovn, case_cexpr) :: new_cases1 in
	      (new_cases2, changed)
	    else
	      let (new_cases1, changed) = remove_empty_clauses stat_ctxt cexpr_type0 remaining_cases in
	      (new_cases1, true)
	| Default ->
	    if remaining_cases = []
	    then
	      ([ (patternk, ovn, case_cexpr) ], false)
	    else
	      raise (Query (Internal_Error "Rewrite_rule : mal-formed typeswitch!!"))
      end

(* Function to remove the "never-evaluated" case clauses - Applies Rule: 2 *)

let rec remove_never_evaluated_clauses_aux stat_ctxt cexpr_type0 cases union_of_visited_types = 
  let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
  match cases with
  | [] -> ([],false)
  | (patternk, ovn, case_cexpr) :: remaining_cases ->
      begin
	match patternk.ppattern_desc with
	| Case patternk_datatype ->
	    let patternk_model =
	      Mapping.map_resolved_datatype_to_model patternk_datatype
	    in
	    let bigger_union_type =
	      fmkmodel (TBar (union_of_visited_types, patternk_model)) Finfo.bogus
	    in
	    if not(is_subtype_of schema cexpr_type0 bigger_union_type)
	    then 
	      let (new_cases1, changed) =
		remove_never_evaluated_clauses_aux stat_ctxt cexpr_type0 remaining_cases bigger_union_type
	      in
	      let new_cases2 = (patternk, ovn, case_cexpr) :: new_cases1 in
	      (new_cases2, changed)
	    else
	      ([ (fmkpattern Default Finfo.bogus, ovn, case_cexpr) ], true)
	| Default ->
	    if remaining_cases = []
	    then
	      ([ (patternk, ovn, case_cexpr) ], false)
	    else
	      raise (Query (Internal_Error "Rewrite_rule : mal-formed typeswitch!!"))
      end

(* Initializes the union_of_visited_types to TNone to apply Rule: 2 *)
let remove_never_evaluated_clauses stat_ctxt cexpr_type0 cases =
  remove_never_evaluated_clauses_aux stat_ctxt cexpr_type0 cases (fmkmodel TNone Finfo.bogus)

(* Function to apply typeswitch optimization rules *)

let typeswitch_rewrite stat_ctxt (ce : cexpr) =
  let mh = ce.pcexpr_type in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  match ce.pcexpr_desc with 
  | CETypeswitch (cexpr0, cases) ->
      let cexpr_type0 = get_type_annotation cexpr0 in
      let (intersecting_cases, changed1) = (remove_empty_clauses stat_ctxt cexpr_type0 cases) in 
      let (subtype_cases, changed2) = (remove_never_evaluated_clauses stat_ctxt cexpr_type0 intersecting_cases) in
      begin
	match subtype_cases with
	| (_, ovn, case_expr) :: [] ->  (* Note: Really here it's only one case so it *MUST* work *)
	    let vname = 
	      match ovn with 
	      | None -> let (v, _) = Maputil.gen_new_var_expr None Finfo.bogus in v
	      | Some vname -> vname
	    in
	    (fmkcexpr (CELet (None, vname, cexpr0, case_expr)) mh eh loc, true)
	| _ -> 
	    (fmkcexpr (CETypeswitch (cexpr0, subtype_cases)) mh eh loc, changed1 || changed2)
      end
  | _ -> raise Not_applied


(***********************************************************************)

(*********************)
(* LET OPTIMIZATIONS *)
(*********************)

(*

  Rule1: let $v := Expr1 return Expr2
         used_count $v Expr2 => 0
         ----------------------------
         Expr2
 
  Rule2: let $v := Expr1 return Expr2
         used_count $v Expr2 => 1
         ----------------------------
         Expr2 [ Expr1 / $v ]

*)

(* Note:
     The first rule uses the non-deterministic semantics allowed by
     XQuery in the sense that the optimization occurs even though the
     expression might fail.
   - Jerome *)

(***************************************************)
(* Substitute Expr1 for Variable in Expr2          *)
(*     Expr2 [ Expr1 / Variable ]                  *)
(* returns (Expr2', Changed) where Changed is true *)
(* if Expr1 was substituted for Variable in Expr2  *)
(***************************************************)

(* Note:
     Variable substitution has been moved to an ealier part of the
     system (in parsing, XQuery_astutil), so it can be used during
     normalization. It is notably used now for normalization of the
     "order by" clause.
 - Jerome
 *)

(* Function to apply let simplification *)

let let_value_rewrite (tr : Static_context.static_context) (ce : cexpr) =
  let mh = ce.pcexpr_type in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  match ce.pcexpr_desc with
  | CELet (None, vname, cexpr1, cexpr2) ->
      begin
	match used_count vname cexpr2 with
	| 0 ->
           (* MF: We do not distinguish between functions that can fail and
	      "functions" that have side-effects -- "functions" with side-effects
	      are lumped into the can-fail category. *)
	    if (not(can_fail cexpr1)) then
	      (cexpr2, true)                                        (* Rule 1 *)
	    else raise Not_applied
	| 1 ->
	    if not(Xquery_astutil.dot = vname)
	    then
	      Xquery_astutil.substitute_var cexpr1 vname cexpr2     (* Rule 2 *)
	    else
	      raise Not_applied
	| _ ->
	    raise Not_applied
      end
  | _ ->
      raise Not_applied

(***********************************************************************)

(**********************)
(* FOR SIMPLIFICATION *)
(**********************)

(* Function to apply let simplification *)

let for_position_rewrite (tr : Static_context.static_context) (ce : cexpr) =
  let mh = ce.pcexpr_type in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  match ce.pcexpr_desc with
  | CEFor (odt, vname, Some vname', cexpr1, cexpr2) ->
      begin
	match used_count vname' cexpr2 with
	| 0 ->
	    (fmkcexpr (CEFor (odt, vname, None, cexpr1, cexpr2)) mh eh loc, true)         (* Rule 1 *)
	| _ ->
	    raise Not_applied
      end
  | _ ->
      raise Not_applied

(*
   for $v in Expr1 return Expr2
   Expr1 cannot fail
   Expr1 : Type1
   Type1 <: ()
   -----------------------------
   ()
   
   OR

   for $v in Expr1 return Expr2
   Expr1 cannot fail
   Expr1 : Type1
   Type1 <: Item
   -----------------------------
   let $v := Expr1 return Expr2
*)

(* Function to apply for simplification *)

let for_type_rewrite stat_ctxt (ce : cexpr) = 
  let mh = ce.pcexpr_type in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
  match ce.pcexpr_desc with 
  | CEFor (odt, vname, None, cexpr1, cexpr2) ->
      let type_decl_can_fail = 
	(match odt with 
	| None -> false
	| Some _ -> not(!Conf.typing))
      in
      (* Note: Here cexpr1 should not fail for rule to be applicable; if typing is disabled, the type declaration may fail *)
      if (not(can_fail cexpr1) && not(type_decl_can_fail))
      then
	begin
          let cexpr1_model = get_type_annotation cexpr1  in
	  if (is_subtype_of schema cexpr1_model empty_sequence_model)
	    then (fmkcexpr (CEEmpty) mh eh loc, true)
	  else if (is_subtype_of schema cexpr1_model anyitem_model)
	    then (fmkcexpr (CELet (None, vname, cexpr1, cexpr2)) mh eh loc, true)
	  else
	    raise Not_applied
        end
      else
	raise Not_applied
  | _ -> raise Not_applied

(***********************************************************************)

(****************************************************)
(* Dynamic dispatch to static dispatch optimization *)
(****************************************************)

(*

  fun_type_rewrite replaces a dynamically dispatched function call by a
  statically dispatched function call.

  statEnv |- QName => qname
  statEnv.funcType(qname) = { FuncDecl1, ..., FuncDeclm }
  FuncDecli = define function qname'(Type1, ..., Typen) as Type
  statEnv |- Expr1 : Type1'     Type1' <: Type1''     Type1'' can be promoted to Type1
  ...
  statEnv |- Exprn : Typen'     Typen' <: Typen''     Typen'' can be promoted to Typen

  QName (Expr1, ..., Exprn) 
  ------------------------------------------------------------------------------------
  qname'(Expr1, ..., Exprn) 

*)

 let fun_type_rewrite stat_ctxt (ce : cexpr) = 
  let mh = ce.pcexpr_type in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  match ce.pcexpr_desc with
  | CEApp (fname, arguments) ->
      let argument_types = List.map get_type_annotation arguments in
      let fname' = Typing.match_function_signature_name stat_ctxt loc fname argument_types in
      if (not(fname = fname')) then 
	(fmkcexpr (CEApp(fname', arguments)) mh eh loc, true)
      else raise Not_applied
  | _ -> raise Not_applied

(***********************************************************************)

(***********************************)
(* fn:data function simplification *)
(***********************************)
(*
   fn:data(Expr)
   Expr : Type
   Type <: xs:anySimpleType
   ------------------------
   Expr
*)

 let fn_data_type_rewrite stat_ctxt (ce : cexpr) = 
  let mh = ce.pcexpr_type in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
  match ce.pcexpr_desc with
  | CEApp (fname, arguments) ->
      if (fname = (fn_uri,"data"))
      then
	begin
	  match arguments with
	  | cexpr :: [] -> 
	      begin
		(* We don't have to check for failure here, because expression is always evaluated *)
		let cexpr_model = get_type_annotation cexpr in
		if (is_anysimpletype schema cexpr_model) 
		then (cexpr, true)
		else
		  raise Not_applied
	      end
	  | _ ->
	      raise (Query (Internal_Error "Rewrite_rule : mal-formed cexpr having empty operator!!"))
	end
      else
	raise Not_applied
  | _ -> raise Not_applied

(***********************************************************************)

(***********************************)
(* fn:data function simplification *)
(***********************************)

(*
   fn:boolean(Expr)
   Expr : ()
   ------------------------
   false()

   fn:boolean(Expr)
   Expr : boolean
   ------------------------
   Expr

   fn:boolean(Expr)
   Expr : Type
   Type <: numeric
   ------------------------
   not(Expr = 0)

   fn:boolean(Expr)
   Expr : Type
   Type <: node+
   ------------------------
   true()

*)

 let fn_boolean_type_rewrite stat_ctxt (ce : cexpr) = 
  let mh = ce.pcexpr_type in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  let false_expr = fmkcexpr (CEScalar(AtomicBoolean false)) mh eh loc in
  let true_expr = fmkcexpr (CEScalar(AtomicBoolean true)) mh eh loc in
  let bool_model = fmkmodel (TBase ATBoolean) loc in
  let double_model = fmkmodel (TBase ATDouble) loc in
  let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
  match ce.pcexpr_desc with
  | CEApp (fname, arguments) ->
      if (fname = (fn_uri,"boolean"))
      then
	let ce' =
	  begin
	    match arguments with
	    | cexpr :: [] -> 
		begin
		  let cexpr_model = get_type_annotation cexpr in
		  if (is_empty_sequence schema cexpr_model) then 
		    false_expr
		  else if (is_anynumeric schema cexpr_model) then
		    let zero_expr = fmkcexpr (CEScalar(AtomicDouble 0.0)) (Some double_model) eh loc in 
		    let eq_expr = fmkcexpr (CEApp((Namespace.op_uri, "equal"), [ cexpr; zero_expr])) mh eh loc in
		    fmkcexpr (CEApp((Namespace.fn_uri, "not"), [ eq_expr ])) mh eh loc
		  else if (is_subtype_of schema cexpr_model bool_model) then 
		    cexpr
		  else if (is_anynode_plus schema cexpr_model && not(can_fail cexpr)) then
		    true_expr
		  else
		    raise Not_applied
		end
	    | _ ->
		raise (Query (Internal_Error "Rewrite_rule : mal-formed cexpr having empty operator!!"))
	  end
	in (ce', true)
      else
	raise Not_applied
  | _ -> raise Not_applied

(***********************************************************************)

(*********************************)
(* Empty operator simplification *)
(*********************************)

(*  
   fn:empty(Expr)
   Expr : ()
   ------------
   true

   OR

   fn:empty(Expr)
   Expr : Type
   Type <: Item+
   -------------
   false
*)

let empty_type_rewrite stat_ctxt (ce : cexpr) = 
  let mh = ce.pcexpr_type in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
  match ce.pcexpr_desc with
  | CEApp (fname, arguments) ->
      if (fname = (fn_uri,"empty"))
      then
	begin
	  match arguments with
	  | cexpr :: [] -> 
	      if not(can_fail cexpr)
	      then
		begin
		  let cexpr_model = get_type_annotation cexpr in
		  if (is_subtype_of schema cexpr_model empty_sequence_model)
		  then (fmkcexpr (CEScalar (buildAtomicBoolean true)) mh eh loc, true)
		  else if (is_subtype_of schema cexpr_model anyitem_model)
		  then (fmkcexpr (CEScalar (buildAtomicBoolean false)) mh eh loc, true)
		  else
		    raise Not_applied 
		end
	      else
		raise Not_applied
	  | _ ->
	      raise (Query (Internal_Error "Rewrite_rule : mal-formed cexpr having empty operator!!"))
	end
      else
	raise Not_applied
  | _ -> raise Not_applied	


(***********************************************************************)

(*******************************)
(* Not operator simplification *)
(*******************************)

(*
   fn:not(Expr)
   Expr = true
   -------------
   false

   fn:not(Expr)
   Expr = false
   -------------
   true
*)
   
let not_value_rewrite (tr : Static_context.static_context) (ce : cexpr) = 
  let mh = ce.pcexpr_type in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  match ce.pcexpr_desc with
  | CEApp (fname, arguments) ->
      if (fname = (fn_uri,"not"))
      then
	begin
	  match arguments with
	  | cexpr :: [] -> 
	      begin
		match cexpr.pcexpr_desc with 
		| CEScalar (AtomicBoolean true) -> 
		    (fmkcexpr (CEScalar (buildAtomicBoolean false)) mh eh loc, true)
		| CEScalar (AtomicBoolean false) -> 
		    (fmkcexpr (CEScalar (buildAtomicBoolean true)) mh eh loc, true)
		| _ -> 
		    raise Not_applied
	      end
	  | _ ->
 	      raise (Query (Internal_Error "Rewrite_rule : mal-formed cexpr having not operator!!"))	
	end
      else
	raise Not_applied
  | _ -> raise Not_applied	

(***********************************************************************)

(*******************************)
(* And operator simplification *)
(*******************************)

(* Implementing the 'short-circuit' semantics for and operator

   fn:and(Expr1, Expr2)
   Expr1 = false
   --------------------
   false

   fn:and(Expr1, Expr2)
   Expr2 = false
   -------------
   false
*)
   
let and_value_rewrite (tr : Static_context.static_context) (ce : cexpr) = 
  let mh = ce.pcexpr_type in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  match ce.pcexpr_desc with
  | CEApp (fname, arguments) ->
      if (fname = (fn_uri,"and"))
      then
	begin
	  match arguments with
	  | cexpr1 :: cexpr2 :: [] -> 
	      begin
		match cexpr1.pcexpr_desc with 
		| CEScalar(AtomicBoolean false) ->            
		    (fmkcexpr (CEScalar (buildAtomicBoolean false)) mh eh loc, true)
      		| CEScalar(AtomicBoolean true) -> 
		    begin
		      match cexpr2.pcexpr_desc with 
		      | CEScalar (AtomicBoolean false) ->
			  (fmkcexpr (CEScalar (buildAtomicBoolean false)) mh eh loc, true)
		      | CEScalar (AtomicBoolean true) ->
			  (fmkcexpr (CEScalar (buildAtomicBoolean true)) mh eh loc, true)
		      | _ -> 
			  raise Not_applied
		    end
		| _ -> 
		    begin
		      match cexpr2.pcexpr_desc with 
		      | CEScalar (AtomicBoolean false) ->
			  (fmkcexpr (CEScalar (buildAtomicBoolean false)) mh eh loc, true)
		      | _ -> 
			  raise Not_applied
		    end
	      end
	  | _ ->
	      raise (Query (Internal_Error "Rewrite_rule : mal-formed boolean-and-operator!!"))
	end
      else
	raise Not_applied
  | _ -> raise Not_applied	

(***********************************************************************)

(******************************)
(* Or operator simplification *)
(******************************)

(* Implementing the 'short-circuit' semantics for or operator

   fn:or(Expr1, Expr2)
   Expr1 = true
   --------------------
   true

   fn:or(Expr1, Expr2)
   Expr2 = true
   -------------
   true
*)
   
let or_value_rewrite (tr : Static_context.static_context) (ce : cexpr) = 
  let mh = ce.pcexpr_type in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  match ce.pcexpr_desc with
  | CEApp (fname, arguments) ->
      begin
	if (fname = (fn_uri,"or"))
	then 
	  begin
	    match arguments with
	    | cexpr1 :: cexpr2 :: [] -> 
		begin
		  match cexpr1.pcexpr_desc with 
		  | CEScalar(AtomicBoolean true) ->            
		      (fmkcexpr (CEScalar (buildAtomicBoolean true)) mh eh loc, true)
      		  | CEScalar(AtomicBoolean false) -> 
		      begin
			match cexpr2.pcexpr_desc with 
			| CEScalar (AtomicBoolean false) ->
			    (fmkcexpr (CEScalar (buildAtomicBoolean false)) mh eh loc, true)
			| CEScalar (AtomicBoolean true) ->
			    (fmkcexpr (CEScalar (buildAtomicBoolean true)) mh eh loc, true)
			| _ -> 
			    raise Not_applied
		      end
		  | _ -> 
		      begin
			match cexpr2.pcexpr_desc with 
			| CEScalar (AtomicBoolean true) ->
			    (fmkcexpr (CEScalar (buildAtomicBoolean true)) mh eh loc, true)
			| _ -> 
			    raise Not_applied
		      end
		end
	    | _ ->
		raise (Query (Internal_Error "Rewrite_rule : mal-formed boolean-or-operator!!"))
	  end
	else
	  raise Not_applied
      end
  | _ -> raise Not_applied	
	
(***********************************************************************)

(***************************************)
(* Type declaration simplifications    *)
(***************************************)

(*

      for $v as Type in Expr1 return Expr2
      Static Typing succeeds
      =================================
      for $v in Expr1 return Expr2

      let $v as Type := Expr1 return Expr2
      Static Typing succeeds
      =================================
      let $v := Expr1 return Expr2

      some $v as Type in Expr1 satisfies Expr2
      Static Typing succeeds
      =================================
      some $v in Expr1 satisfies Expr2

      every $v as Type in Expr1 satisfies Expr2
      Static Typing succeeds
      =================================
      every $v in Expr1 satisfies Expr2

*)

let typedecl_rewrite (tr : Static_context.static_context) (ce : cexpr) = 
  let mh = ce.pcexpr_type in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  match ce.pcexpr_desc with 
  | CEFor (Some _, vname, None, cexpr1, cexpr2) ->
      if !Conf.typing
      then (fmkcexpr (CEFor (None, vname, None, cexpr1, cexpr2)) mh eh loc, true)
      else raise Not_applied
  | CELet (Some _, vname, cexpr1, cexpr2) ->
      if !Conf.typing
      then (fmkcexpr (CELet (None, vname, cexpr1, cexpr2)) mh eh loc, true)
      else raise Not_applied
  | CESome (odt, vname, cexpr1, cexpr2) -> 
      if !Conf.typing
      then (fmkcexpr (CESome (None, vname, cexpr1, cexpr2)) mh eh loc, true)
      else raise Not_applied
  | CEEvery (odt, vname, cexpr1, cexpr2) -> 
      if !Conf.typing
      then (fmkcexpr (CEEvery (None, vname, cexpr1, cexpr2)) mh eh loc, true)
      else raise Not_applied
  | _ -> raise Not_applied

(***********************************************************************)


(*************************************************)
(* Removing existential/universal quantification *)
(*************************************************)

(*

   some $v in Expr1 return Expr2
   Expr1 cannot fail
   Expr1 : ()
   -----------------------------
   false
   
   some $v in Expr1 satisfies Expr2
   Expr1 : Type1
   Type1 <: item
   ----------------------------------
   let $v := Expr1 return Expr2

   every $v in Expr1 return Expr2
   Expr1 cannot fail
   Expr1 : ()
   -----------------------------
   true
   
   every $v in Expr1 satisfies Expr2
   Expr1 : Type1
   Type1 <: item
   ----------------------------------
   let $v := Expr1 return Expr2

*)

let existential_type_rewrite stat_ctxt (ce : cexpr) = 
  let mh = ce.pcexpr_type in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
  match ce.pcexpr_desc with 
  | CESome (odt, vname, cexpr1, cexpr2) ->
      if (!Conf.typing) then
	begin
          let cexpr1_model = get_type_annotation cexpr1 in
	  if (is_subtype_of schema cexpr1_model anyitem_model) then 
	    (fmkcexpr (CELet (odt, vname, cexpr1, cexpr2)) mh eh loc, true)
	  else if (not(can_fail cexpr1) &&
		   is_subtype_of schema cexpr1_model empty_sequence_model) then 
	    (fmkcexpr (CEScalar(AtomicBoolean false)) mh eh loc, true)
	  else raise Not_applied
	end
      else raise Not_applied
  | CEEvery (odt, vname, cexpr1, cexpr2) ->
      if (!Conf.typing) then
	begin
          let cexpr1_model = get_type_annotation cexpr1 in
	  if (is_subtype_of schema cexpr1_model anyitem_model) then 
	    (fmkcexpr (CELet (odt, vname, cexpr1, cexpr2)) mh eh loc, true)
	  else if (not(can_fail cexpr1) &&
		   is_subtype_of schema cexpr1_model empty_sequence_model) then 
	    (fmkcexpr (CEScalar(AtomicBoolean true)) mh eh loc, true)
	  else raise Not_applied
	end
      else raise Not_applied
  | _ -> raise Not_applied

(***********************************************************************)

(*******************)
(* SBDO Rewritings *)
(*******************)

(*********************************************************************)
(* This rewriting guarantees that Expr contains a sequence of nodes. *)
(*                                                                   *)
(*  glx:docorder(Expr)                                               *)
(*  order insignificant                                              *)
(*  ------------------                                               *)
(*  glx:node-sequence(Expr)                                          *)

let rec remove_distinct_doc_order stat_ctxt ce =
  let mh = ce.pcexpr_type in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  match ce.pcexpr_desc with
  | CEApp (fname, [argument]) ->
      if ((fname = (glx_uri,"docorder"))
	||(fname = (glx_uri,"distinct-docorder")))
      then
	(fmkcexpr(CEApp ((glx_uri,"node-sequence"), [argument])) mh eh loc, true)
      else
	raise Not_applied
  | CEFor (_, _, Some _, _, _) ->
      (ce,false)
  | _ ->
      raise Not_applied

let sbdo_rewrite stat_ctxt (ce : cexpr) =
  if (does_node_care_about_order ce)
  then
    children_tree_apply stat_ctxt [remove_distinct_doc_order] ce
  else
      raise Not_applied

(*********************************************************************)
(* This rewriting removes unnecessary glx:node-sequence calls.       *)
(*                                                                   *)
(*  glx:node-sequend(Expr)                                           *)
(*  Static typing enabled                                            *)
(*  ------------------                                               *)
(*         Expr                                                      *)
let remove_node_sequence (tr : Static_context.static_context) (ce : cexpr) = 
  let mh = ce.pcexpr_type in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  match ce.pcexpr_desc with 
  | CEApp (fname, [argument]) ->
      if ((fname = (glx_uri,"node-sequence")) && !Conf.typing)
      then
	(argument, true)
      else
	raise Not_applied
  | _ -> raise Not_applied

(*************************************)
(* Global sets of optimization rules *)
(*************************************)

let all_rewrites = [
  sbdo_rewrite;
  remove_node_sequence;
  if_rewrite; 
  typeswitch_rewrite;
  let_value_rewrite;
  for_type_rewrite;
  for_position_rewrite;
  existential_type_rewrite;
  fun_type_rewrite; 
  fn_data_type_rewrite; 
  fn_boolean_type_rewrite; 
  empty_type_rewrite;
  not_value_rewrite;
  and_value_rewrite;
  or_value_rewrite;
  typedecl_rewrite
] 

let nontype_rewrites = [
  sbdo_rewrite;
  if_rewrite; 
  let_value_rewrite;
  for_position_rewrite;
  not_value_rewrite;
  and_value_rewrite;
  or_value_rewrite
]

(***********************************************************************)

(* More optimization rules: *)

(* Function to simplify and boolean operator *)
(* Ask Mary/Jerome : t & _ | _ & t = _ ??? *)


(***************************************)
(* SQL-X specific simplification rules *)
(***************************************)

(* Note by Jerome / September 2002 *)

(* Descendant or self rewriting *)

(* Example:

   Input schema:
   
      define element users { element user_tuple* }
      define element user_tuple {
	element userid,
	element name,
	element rating?
      }

   Query:

      $users//user_tuple

        where $users is of type element users

   Should be rewritten as:

      $users/user_tuple/user_tuple

   $users/descendant-or-self::*/child::user_tuple

   $users,self::*     	    	     --> element users
   $users,child::* 	    	     --> element user_tuple
   $users,child::*,child::* 	     --> element userid
		       	 		 element name
		       	 		 element rating
   $users,child::*,child::*,child::* --> xsd:string

   $users/
     (child::user_tuple
     | (child::user_tuple/child::userid)
     | (child::user_tuple/child::name)
     | (child::user_tuple/child::rating)
     | (child::user_tuple/child::userid/text())
     | (child::user_tuple/child::name/text())
     | (child::user_tuple/child::rating/text()))/child::user_tuple

   two cases:
     element
     attribute
*)

