(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: rewrite_walker.ml,v 1.2 2003/05/28 14:47:38 simeon Exp $ *)

(* Module Rewrite_walker
   Description:
     This module implements the generic tree walker, recursive
     application of optimization rules, and fix-point.
*)

open Error

open Namespace

open Xquery_ast
open Xquery_astutil

open Datatypes

open Typeutil

(* Type for rewriting rules *)

type rule =
    Static_context.static_context -> Xquery_ast.cexpr -> Xquery_ast.cexpr * bool

(* Exception raised if a given rewriting rule does not apply *)

exception Not_applied


(*****************************************)
(* Generic application of the rewritings *)
(*****************************************)

(* Applies a single rewriting *)

let apply_single_rule stat_ctxt ((ce,changed) : cexpr * bool) one_rule =
  try
    let (ce',changed') = one_rule stat_ctxt ce in
    (ce', changed' || changed)
  with
    Not_applied ->
      (ce, changed)

(* Applies a list of rewritings *)

let apply_rule_generic stat_ctxt rule_list ce =
  List.fold_left (apply_single_rule stat_ctxt) (ce,false) rule_list


(***************************)
(* The generic tree-walker *)
(***************************)

let rec generic_tree_apply stat_ctxt rule_list ce = 
  let mh = ce.pcexpr_type in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  let (optimized_ce, optimized_changed) = 
  match ce.pcexpr_desc with
  | CESort (v, cexpr1, stablekind, cexpr2, sortkind, emptysortkind) -> 
      let (cexpr1',changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
      let (cexpr2',changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
      (fmkcexpr (CESort (v, cexpr1',stablekind,cexpr2',sortkind,emptysortkind)) mh eh loc, changed1 || changed2)

  | CEUnordered cexpr ->
      let (cexpr',changed) = generic_tree_apply stat_ctxt rule_list cexpr in
      (fmkcexpr (CEUnordered cexpr') mh eh loc, changed)

  | CEFor (odt, vname, ovname, cexpr1, cexpr2) ->
      let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
      let (cexpr2', changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
      (fmkcexpr (CEFor (odt, vname, ovname, cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

  | CEForTuple (vname, cexpr1, cexpr2) ->
      let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
      let (cexpr2', changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
      (fmkcexpr (CEForTuple (vname, cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

  | CELet (odt, vname, cexpr1, cexpr2) ->
      let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
      let (cexpr2', changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
      (fmkcexpr (CELet (odt, vname, cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

  | CEIf (cexpr1, cexpr2, cexpr3) ->
      let (cexpr1',changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
      let (cexpr2',changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
      let (cexpr3',changed3) = generic_tree_apply stat_ctxt rule_list cexpr3 in
      (fmkcexpr (CEIf (cexpr1', cexpr2', cexpr3')) mh eh loc, changed1 || changed2 || changed3)

  | CETypeswitch (cexpr, cases) ->
      let (cexpr',changed') = generic_tree_apply stat_ctxt rule_list cexpr in 
      let apply_one_case (pat,ovn,case_cexpr) =
	let (new_case_cexpr, changed) = generic_tree_apply stat_ctxt rule_list case_cexpr
	in ((pat,ovn,new_case_cexpr), changed)
      in
      let cases1 = List.map apply_one_case cases in 
        let aggregate_case ((pat, ovn, new_case_cexpr), changed) = (pat, ovn, new_case_cexpr) in
	let aggregate_bool ((pat, ovn, new_case_cexpr), changed) = changed in
      let cases2 = List.map aggregate_case cases1 in
      let changed2 = List.map aggregate_bool cases1 in
      let changed_bool_value = List.mem true changed2 in 
      (fmkcexpr (CETypeswitch (cexpr', cases2)) mh eh loc, changed' || changed_bool_value)

  | CEVar vname -> 
      (ce, false)

  | CEScalar _ -> 
      (ce,false)

  | CEText _ -> 
      (ce,false)

  | CETextComputed cexpr1 ->
      let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
      (fmkcexpr (CETextComputed cexpr1') mh eh loc, changed1)

  | CEPI _ ->
      (ce,false)

  | CEComment _ ->
      (ce,false)

  | CEDocument cexpr1 ->
      let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
      (fmkcexpr (CEDocument cexpr1') mh eh loc, false)

  | CEApp (fname, arguments) ->
      let arguments' = List.map (generic_tree_apply stat_ctxt rule_list) arguments in
      let arguments1 = List.map (fun (arg, changed) -> arg) arguments' in
      let changed_list1 = List.map (fun (arg, changed) -> changed) arguments' in
      (fmkcexpr (CEApp (fname, arguments1)) mh eh loc, (List.mem true changed_list1))

  | CECreateTuple rne_list ->
      let simplify_one (rname,cexpr1) =
	let (cexpr1',changed) = (generic_tree_apply stat_ctxt rule_list cexpr1) in
	((rname,cexpr1'),changed)
      in
      let rne_list' = List.map simplify_one rne_list in
      let rne_list1 = List.map (fun (arg, _) -> arg) rne_list' in
      let changed_list1 = List.map (fun (_, changed) -> changed) rne_list' in
      (fmkcexpr (CECreateTuple rne_list1) mh eh loc, (List.mem true changed_list1))

  | CEAccessTuple (cexpr1,rname) ->
      let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
      (fmkcexpr (CEAccessTuple(cexpr1', rname)) mh eh loc, changed1) 

  | CEKeyRef (kname, arg) ->
      let (arg1, changed) = generic_tree_apply stat_ctxt rule_list arg in
      (fmkcexpr (CEKeyRef (kname, arg1)) mh eh loc, changed)

  | CESeq (cexpr1, cexpr2) ->
      let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
      let (cexpr2', changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
      (fmkcexpr (CESeq(cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

  | CEEmpty -> 
      (ce,false)

  | CEElem (relem_symbol, cexpr) ->
      let (cexpr',changed') = generic_tree_apply stat_ctxt rule_list cexpr in
      (fmkcexpr (CEElem (relem_symbol, cexpr')) mh eh loc, changed')

  | CEAnyElem (cexpr1, cexpr2) ->  
      let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
      let (cexpr2', changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
      (fmkcexpr (CEAnyElem (cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

  | CEAttr (rattr_symbol, cexpr)->
      let (cexpr', changed') = generic_tree_apply stat_ctxt rule_list cexpr in
      (fmkcexpr (CEAttr (rattr_symbol, cexpr')) mh eh loc, changed')

  | CEAnyAttr (cexpr1, cexpr2) ->
      let (cexpr1',changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
      let (cexpr2', changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
      (fmkcexpr (CEAnyAttr (cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

  | CEError _ -> 
      (ce, false)

  | CETreat (cexpr, model) ->
      let (cexpr', changed') = generic_tree_apply stat_ctxt rule_list cexpr in
      (fmkcexpr (CETreat (cexpr', model)) mh eh loc, changed')

  | CECast (cexpr, model) ->
      let (cexpr', changed') = generic_tree_apply stat_ctxt rule_list cexpr in
      (fmkcexpr (CECast (cexpr', model)) mh eh loc, changed')

  | CECastable (cexpr, model) ->
      let (cexpr', changed') = generic_tree_apply stat_ctxt rule_list cexpr in
      (fmkcexpr (CECastable (cexpr', model)) mh eh loc, changed')

  | CEValidate (cexpr, model) ->
      let (cexpr', changed') = generic_tree_apply stat_ctxt  rule_list cexpr in
      (fmkcexpr (CEValidate (cexpr', model)) mh eh loc, changed')

  | CEForwardAxis (axis, cnode_test) ->
      (ce,false)

  | CEReverseAxis (axis, cnode_test) ->
      (ce,false)

  | CESubtype (model1, model2) -> 
      (ce,false)

  | CEIntersects (model1, model2) -> 
      (ce,false)

  | CESome (odt, vname, cexpr1, cexpr2) ->
      let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
      let (cexpr2', changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
      (fmkcexpr (CESome (odt, vname, cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

  | CEEvery (odt, vname, cexpr1, cexpr2) ->
      let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
      let (cexpr2', changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
      (fmkcexpr (CEEvery (odt, vname, cexpr1', cexpr2')) mh eh loc, changed1 || changed2)
  in
  (* apply local rewritings *)
  let (local_optimized_ce, local_changed) = apply_rule_generic stat_ctxt rule_list optimized_ce in
  if local_changed
  then (local_optimized_ce, true)
  else (optimized_ce, optimized_changed)

(******************)
(* Children apply *)
(******************)

(* Note:
   Applies not to the current expression, but to all of the
   sub-expressions.
   - Jerome *)

let children_tree_apply stat_ctxt rule_list ce = 
  let mh = ce.pcexpr_type in
  let eh = ce.pcexpr_origin in
  let loc = ce.pcexpr_loc in
  let (optimized_ce, optimized_changed) = 
    match ce.pcexpr_desc with
    | CESort (v, cexpr1, stablekind, cexpr2, sortkind, emptysortkind) -> 
	let (cexpr1',changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
	let (cexpr2',changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
	(fmkcexpr (CESort (v, cexpr1',stablekind,cexpr2',sortkind,emptysortkind)) mh eh loc, changed1 || changed2)
    | CEUnordered cexpr ->
	let (cexpr',changed) = generic_tree_apply stat_ctxt rule_list cexpr in
	(fmkcexpr (CEUnordered cexpr') mh eh loc, changed)

    | CEFor (odt, vname, ovname, cexpr1, cexpr2) ->
	let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
	let (cexpr2', changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
	(fmkcexpr (CEFor (odt, vname, ovname, cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

    | CEForTuple (vname, cexpr1, cexpr2) ->
	let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
	let (cexpr2', changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
	(fmkcexpr (CEForTuple (vname, cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

    | CELet (odt, vname, cexpr1, cexpr2) ->
	let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
	let (cexpr2', changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
	(fmkcexpr (CELet (odt, vname, cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

    | CEIf (cexpr1, cexpr2, cexpr3) ->
	let (cexpr1',changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
	let (cexpr2',changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
	let (cexpr3',changed3) = generic_tree_apply stat_ctxt rule_list cexpr3 in
	(fmkcexpr (CEIf (cexpr1', cexpr2', cexpr3')) mh eh loc, changed1 || changed2 || changed3)

    | CETypeswitch (cexpr, cases) ->
	let (cexpr',changed') = generic_tree_apply stat_ctxt rule_list cexpr in 
	let apply_one_case (pat,ovn,case_cexpr) =
	  let (new_case_cexpr, changed) = generic_tree_apply stat_ctxt rule_list case_cexpr
	  in ((pat,ovn,new_case_cexpr), changed)
	in
	let cases1 = List.map apply_one_case cases in 
        let aggregate_case ((pat, ovn, new_case_cexpr), changed) = (pat, ovn, new_case_cexpr) in
	let aggregate_bool ((pat, ovn, new_case_cexpr), changed) = changed in
	let cases2 = List.map aggregate_case cases1 in
	let changed2 = List.map aggregate_bool cases1 in
	let changed_bool_value = List.mem true changed2 in 
	(fmkcexpr (CETypeswitch (cexpr', cases2)) mh eh loc, changed' || changed_bool_value)

    | CEVar vname -> 
	(ce, false)

    | CEScalar _ -> 
	(ce,false)

    | CEText _ -> 
	(ce,false)

    | CETextComputed cexpr1 ->
	let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
	(fmkcexpr (CETextComputed cexpr1') mh eh loc, changed1)

    | CEPI _ ->
	(ce,false)

    | CEComment _ ->
	(ce,false)

    | CEDocument cexpr1 ->
	let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
	(fmkcexpr (CEDocument cexpr1') mh eh loc, false)

    | CEApp (fname, arguments) ->
	let arguments' = List.map (generic_tree_apply stat_ctxt rule_list) arguments in
	let arguments1 = List.map (fun (arg, changed) -> arg) arguments' in
	let changed_list1 = List.map (fun (arg, changed) -> changed) arguments' in
	(fmkcexpr (CEApp (fname, arguments1)) mh eh loc, (List.mem true changed_list1))

    | CECreateTuple rne_list ->
	let simplify_one (rname,cexpr1) =
	  let (cexpr1',changed) = (generic_tree_apply stat_ctxt rule_list cexpr1) in
	  ((rname,cexpr1'),changed)
	in
	let rne_list' = List.map simplify_one rne_list in
	let rne_list1 = List.map (fun (arg, _) -> arg) rne_list' in
	let changed_list1 = List.map (fun (_, changed) -> changed) rne_list' in
	(fmkcexpr (CECreateTuple rne_list1) mh eh loc, (List.mem true changed_list1))

    | CEAccessTuple (cexpr1,rname) ->
	let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
	(fmkcexpr (CEAccessTuple(cexpr1', rname)) mh eh loc, changed1) 

    | CEKeyRef (kname, arg) ->
	let (arg1, changed) = generic_tree_apply stat_ctxt rule_list arg in
	(fmkcexpr (CEKeyRef (kname, arg1)) mh eh loc, changed)

    | CESeq (cexpr1, cexpr2) ->
	let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
	let (cexpr2', changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
	(fmkcexpr (CESeq(cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

    | CEEmpty -> 
	(ce,false)

    | CEElem (relem_symbol, cexpr) ->
	let (cexpr',changed') = generic_tree_apply stat_ctxt rule_list cexpr in
	(fmkcexpr (CEElem (relem_symbol, cexpr')) mh eh loc, changed')

    | CEAnyElem (cexpr1, cexpr2) ->  
	let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
	let (cexpr2', changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
	(fmkcexpr (CEAnyElem (cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

    | CEAttr (rattr_symbol, cexpr)->
	let (cexpr', changed') = generic_tree_apply stat_ctxt rule_list cexpr in
	(fmkcexpr (CEAttr (rattr_symbol, cexpr')) mh eh loc, changed')

    | CEAnyAttr (cexpr1, cexpr2) ->
	let (cexpr1',changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
	let (cexpr2', changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
	(fmkcexpr (CEAnyAttr (cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

    | CEError _ -> 
	(ce, false)

    | CETreat (cexpr, model) ->
	let (cexpr', changed') = generic_tree_apply stat_ctxt rule_list cexpr in
	(fmkcexpr (CETreat (cexpr', model)) mh eh loc, changed')

    | CECast (cexpr, model) ->
	let (cexpr', changed') = generic_tree_apply stat_ctxt rule_list cexpr in
	(fmkcexpr (CECast (cexpr', model)) mh eh loc, changed')

    | CECastable (cexpr, model) ->
	let (cexpr', changed') = generic_tree_apply stat_ctxt rule_list cexpr in
	(fmkcexpr (CECastable (cexpr', model)) mh eh loc, changed')

    | CEValidate (cexpr, model) ->
	let (cexpr', changed') = generic_tree_apply stat_ctxt  rule_list cexpr in
	(fmkcexpr (CEValidate (cexpr', model)) mh eh loc, changed')

    | CEForwardAxis (axis, cnode_test) ->
	(ce,false)

    | CEReverseAxis (axis, cnode_test) ->
	(ce,false)

    | CESubtype (model1, model2) -> 
	(ce,false)

    | CEIntersects (model1, model2) -> 
	(ce,false)

    | CESome (odt, vname, cexpr1, cexpr2) ->
	let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
	let (cexpr2', changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
	(fmkcexpr (CESome (odt, vname, cexpr1', cexpr2')) mh eh loc, changed1 || changed2)

    | CEEvery (odt, vname, cexpr1, cexpr2) ->
	let (cexpr1', changed1) = generic_tree_apply stat_ctxt rule_list cexpr1 in
	let (cexpr2', changed2) = generic_tree_apply stat_ctxt rule_list cexpr2 in
	(fmkcexpr (CEEvery (odt, vname, cexpr1', cexpr2')) mh eh loc, changed1 || changed2)
  in
  (optimized_ce, optimized_changed)


(**************************************)
(* Fix-point simplify core expression *)
(**************************************)

let rec fix_point_tree_apply stat_ctxt rule_list ce =
  let (ce1, changed) = generic_tree_apply stat_ctxt rule_list ce in
  if changed 
  then fix_point_tree_apply stat_ctxt rule_list ce1
  else ce1


