(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: glx_wrap.ml,v 1.6 2003/06/27 16:52:00 mff Exp $ *)

open Error

open Galapi

(*********************)
(* Exception wrapper *)
(*********************)

(* Turns an arbitrary Galax error into a simple Caml Failure *)

(* Note:  Converts internal Query exception into generic Caml exception.  
   Attempting to raise the Query exception directly caused problems.  *)

let wrap_exception exn =
  let error_msg =
    Error.bprintf_error_safe "" exn
  in
  raise (Failure error_msg)

(***************************)
(* Data model constructors *)
(***************************)

let ffout = Format.formatter_of_out_channel stdout

let glx_xs_string s =
  try
    xs_string s
  with
  | exn ->
      wrap_exception exn
	
let glx_xs_boolean b =
  try
    xs_boolean b
  with
  | exn ->
      wrap_exception exn

let glx_xs_int i =
  try
    xs_int i
  with
  | exn ->
      wrap_exception exn

let glx_xs_integer i =
  try
    xs_integer (Decimal._integer_of_int i)
  with
  | exn ->
      wrap_exception exn

let glx_xs_decimal i =
  try
    xs_decimal (Decimal._decimal_of_int i)
  with
  | exn ->
      wrap_exception exn

let glx_xs_float d =
  try
    xs_float d
  with
  | exn ->
      wrap_exception exn

let glx_xs_double d =
  try
    xs_double d
  with
  | exn ->
      wrap_exception exn

let glx_xs_anyURI b =
  try
    xs_anyURI b
  with
  | exn ->
      wrap_exception exn

let glx_xs_untyped s =
  try
    xs_untyped s
  with
  | exn ->
      wrap_exception exn

let glx_xs_QName(c,s) =
  try
    xs_QName(c,s) 
  with
  | exn ->
      wrap_exception exn

let glx_document(s, nl) =
  try
    document(s, nl)
  with
  | exn ->
      wrap_exception exn

let glx_element(q, al, nl, t) =
  try
    element(q, al, nl, t)
  with
  | exn ->
      wrap_exception exn

let glx_attribute(q, s, t) =
  try
    attribute(q, s, t)
  with
  | exn ->
      wrap_exception exn

let glx_text s =
  try
    text s
  with
  | exn ->
      wrap_exception exn

let glx_comment s =
  try
    comment s
  with
  | exn ->
      wrap_exception exn

let glx_processing_instruction (s1,s2) =
  try
    processing_instruction(s1, s2)
  with
  | exn ->
      wrap_exception exn


(* ****************** *)
(* Accessors on items *)
(* ****************** *)

let glx_item_kind item =
  try
    item_kind item
  with
  | exn ->
      wrap_exception exn

let glx_get_node item =
  try
    get_node item
  with
  | exn ->
      wrap_exception exn

let glx_get_atomicvalue item =
  try
    get_atomicvalue item
  with
  | exn ->
      wrap_exception exn

let glx_get_element item =
  try
    get_element item
  with
  | exn ->
      wrap_exception exn

let glx_get_attribute item =
  try
    get_attribute item
  with
  | exn ->
      wrap_exception exn

let glx_string item =
  try
    string item
  with
  | exn ->
      wrap_exception exn


(* ****************** *)
(* Accessors on nodes *)
(* ****************** *)

let glx_parent node =
  try
    parent node
  with
  | exn ->
      wrap_exception exn

let glx_children node =
  try
    children node
  with
  | exn ->
      wrap_exception exn

let glx_base_uri node =
  try
    begin
      match 
	base_uri node 
      with
      | None -> []
      | Some uri -> [uri]
    end
  with
  | exn ->
      wrap_exception exn

let glx_node_kind node =
  try
    node_kind node
  with
  | exn ->
      wrap_exception exn

let glx_node_name node =
  try
    begin
      match (node_name node) with
      | None -> []
      | Some qname -> [qname]
    end
  with
  | exn ->
      wrap_exception exn

let glx_typed_value node =
  try
    typed_value node
  with
  | exn ->
      wrap_exception exn

let glx_attributes node =
  try
    attributes node
  with
  | exn ->
      wrap_exception exn

let glx_namespace node =
  try
    namespace node
  with
  | exn ->
      wrap_exception exn


(********************************)
(* Conversion functions         *)
(********************************)

let glx_string_of_atomicvalue  a =
  try
    string_of_atomicvalue a
  with
  | exn ->
      wrap_exception exn

let glx_boolean_of_xs_boolean b =
  try
    boolean_of_xs_boolean b
  with
  | exn ->
      wrap_exception exn

let glx_int_of_xs_int i =
  try
    int_of_xs_int i
  with
  | exn ->
      wrap_exception exn

let glx_integer_of_xs_integer i =
  try
    Decimal._int_of_integer(integer_of_xs_integer i)
  with
  | exn ->
      wrap_exception exn

let glx_decimal_of_xs_decimal d =
  try
    Decimal._int_of_decimal(decimal_of_xs_decimal d)
  with
  | exn ->
      wrap_exception exn

let glx_float_of_xs_float f =
  try
    float_of_xs_float f 
  with
  | exn ->
      wrap_exception exn

let glx_float_of_xs_double d =
  try
    float_of_xs_double d
  with
  | exn ->
      wrap_exception exn


(********************************)
(* Input / output of operations *)
(********************************)

let glx_load_document  string =
  try
    load_document string
  with
  | exn ->
      wrap_exception exn

let glx_load_document_validate context string =
  try
    load_document_validate context  string 
  with
  | exn ->
      wrap_exception exn

let glx_serialize_to_string v = 
  try
    serialize_to_string v
  with
  | exn ->
      wrap_exception exn

let glx_serialize_to_file string v = 
  try
    serialize_to_file string v
  with
  | exn ->
      wrap_exception exn

let glx_serialize_to_output_channel s =
  try
    serialize_to_formatter ffout s
  with
  | exn ->
      wrap_exception exn

(***********************************)
(* Operations on the query context *)
(***********************************)

let glx_default_processing_context () =
  try
    default_processing_context () 
  with
  | exn ->
      wrap_exception exn

let glx_load_standard_library ctxt =
  try
    load_standard_library ctxt
  with
  | exn ->
      wrap_exception exn

let glx_import_library_module ctxt filename =
  try
    import_library_module ctxt filename
  with
  | exn ->
      wrap_exception exn

let glx_import_main_module ctxt filename =
  try
    import_main_module ctxt filename
  with
  | exn ->
      wrap_exception exn

let glx_import_main_module_from_string ctxt str =
  try
    import_main_module ctxt str
  with
  | exn ->
      wrap_exception exn

(***********************)
(* Querying operations *)
(***********************)

let glx_build_external_context ctxt_item_list var_list val_list = 
  try
    begin
      let opt_ctxt_item = 
	match ctxt_item_list with
	|	[] -> None
	|	[i] -> Some i
	| _ -> raise (Query(Load_Error ("Context item argument to glx_build_external_context contains more than one item.")))
      in
      let var_val_list = List.combine var_list val_list in
      build_external_context opt_ctxt_item var_val_list
    end
  with
  | exn ->
      wrap_exception exn

let glx_add_external_context mod_ctxt ext_ctxt = 
  try
    add_external_context mod_ctxt ext_ctxt
  with
  | exn ->
      wrap_exception exn

let glx_eval_global_variables mod_ctxt  = 
  try
    eval_global_variables mod_ctxt
  with
  | exn ->
      wrap_exception exn

let glx_eval_query_function mod_ctxt fnname args  = 
  try
    eval_query_function mod_ctxt fnname args 
  with
  | exn ->
      wrap_exception exn

(* The C interface can't do much with values of type 'a option, so we convert None into [] *)
let glx_eval_statements (mod_ctxt,stmts) = 
  try
    begin
      List.concat 
	(List.map 
	   (fun s -> 
	     match (eval_statement mod_ctxt s) with
	     | None -> []
	     | Some v -> v)
	   stmts)
    end
  with
  | exn ->
      wrap_exception exn

(* The C interface can't do much with values of type 'a option, so we convert None into [] *)
let glx_eval_statement_from_string mod_ctxt str = 
  try
    begin
      match (eval_statement_from_string mod_ctxt str) with
      | None -> []
      | Some v -> v
    end
  with
  | exn ->
      wrap_exception exn

let glx_eval_statement_with_context_item_from_document file querystr =
  try
    begin
      match (eval_statement_with_context_item_from_document file querystr) with
      | None -> []
      | Some v -> v
    end
  with
  | exn ->
      wrap_exception exn

let glx_eval_statement_with_context_item_from_xml v querystr =
  try
    begin
      match (eval_statement_with_context_item_from_xml v querystr) with
      | None -> []
      | Some v -> v
    end
  with
  | exn ->
      wrap_exception exn

let glx_eval_statement_from_file_with_context_item_from_xml v querystr =
  try
    begin
      match (eval_statement_from_file_with_context_item_from_xml v querystr) with
      | None -> []
      | Some v -> v
    end
  with
  | exn ->
      wrap_exception exn

let glx_eval_statement_with_variables_from_xml var_list val_list querystr =
  try
    begin
      let var_val_list = List.combine var_list val_list in
      match (eval_statement_with_variables_from_xml var_val_list querystr) with
      | None -> []
      | Some v -> v
    end
  with
  | exn ->
      wrap_exception exn

let _ =
  begin
    Callback.register "glx_xs_string" glx_xs_string;
    Callback.register "glx_xs_boolean" glx_xs_boolean;
    Callback.register "glx_xs_integer" glx_xs_integer;
    Callback.register "glx_xs_int" glx_xs_int;
    Callback.register "glx_xs_decimal" glx_xs_decimal;
    Callback.register "glx_xs_float" glx_xs_float;
    Callback.register "glx_xs_double" glx_xs_double;
    Callback.register "glx_xs_anyURI" glx_xs_anyURI;
    Callback.register "glx_xs_untyped" glx_xs_untyped;
    Callback.register "glx_xs_QName" glx_xs_QName;
    
    Callback.register "glx_document" glx_document;
    Callback.register "glx_element" glx_element;
    Callback.register "glx_text" glx_text;
    Callback.register "glx_attribute" glx_attribute;
    Callback.register "glx_comment" glx_comment;
    Callback.register "glx_processing_instruction" glx_processing_instruction;
    
    Callback.register "glx_string" glx_string;
    Callback.register "glx_item_kind" glx_item_kind;
    Callback.register "glx_get_node" glx_get_node;
    Callback.register "glx_get_atomicvalue" glx_get_atomicvalue;
    Callback.register "glx_get_element" glx_get_element;
    Callback.register "glx_get_attribute" glx_get_attribute;
    
    Callback.register "glx_parent" glx_parent;
    Callback.register "glx_children" glx_children;
    
    Callback.register "glx_base_uri" glx_base_uri;
    Callback.register "glx_node_kind" glx_node_kind;
    Callback.register "glx_node_name" glx_node_name;
    Callback.register "glx_typed_value" glx_typed_value;
    Callback.register "glx_attributes" glx_attributes;
    Callback.register "glx_namespace" glx_namespace;
    
    Callback.register "glx_string_of_atomicvalue" glx_string_of_atomicvalue;
    Callback.register "glx_boolean_of_xs_boolean" glx_boolean_of_xs_boolean; 
    Callback.register "glx_int_of_xs_int" glx_int_of_xs_int; 
    Callback.register "glx_integer_of_xs_integer" glx_integer_of_xs_integer;
    Callback.register "glx_decimal_of_xs_decimal" glx_decimal_of_xs_decimal; 
    Callback.register "glx_float_of_xs_float" glx_float_of_xs_float; 
    Callback.register "glx_float_of_xs_double" glx_float_of_xs_double;
    
    Callback.register "glx_load_document" glx_load_document;
    Callback.register "glx_load_document_validate" glx_load_document_validate;
    
    Callback.register "glx_serialize_to_string" glx_serialize_to_string;
    Callback.register "glx_serialize_to_output_channel" glx_serialize_to_output_channel; 
    Callback.register "glx_serialize_to_file"  glx_serialize_to_file;
    
    Callback.register "glx_default_processing_context" glx_default_processing_context;
    Callback.register "glx_load_standard_library" glx_load_standard_library;
    Callback.register "glx_import_library_module" glx_import_library_module;
    Callback.register "glx_import_main_module" glx_import_main_module;
    
    Callback.register "glx_build_external_context" glx_build_external_context;
    Callback.register "glx_add_external_context" glx_add_external_context;
    Callback.register "glx_eval_global_variables" glx_eval_global_variables;
    
    Callback.register "glx_eval_statement_from_string" glx_eval_statement_from_string;
    Callback.register "glx_eval_query_function" glx_eval_query_function;
    Callback.register "glx_eval_statements" glx_eval_statements;
    
    Callback.register "glx_eval_statement_with_context_item_from_document" glx_eval_statement_with_context_item_from_document;
    Callback.register "glx_eval_statement_with_context_item_from_xml" glx_eval_statement_with_context_item_from_xml ;
    Callback.register "glx_eval_statement_from_file_with_context_item_from_xml" glx_eval_statement_from_file_with_context_item_from_xml ;
    Callback.register "glx_eval_statement_with_variables_from_xml" glx_eval_statement_with_variables_from_xml
  end

