(***********************************************************************)
(*                                                                     *)
(*                               Galax                                 *)
(*                            XQuery Engine                            *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: datamodel_api.ml,v 1.2 2003/06/27 16:52:00 mff Exp $ *)

open Xquery_ast
open Datamodel
open Datatypes
open Error
open Namespace 

type _node = _Node 
type _atomic = _AtomicValue
type _rqname = rqname
type _uri = AnyURI._uri
type _item = _Item


class virtual item =
object
  method virtual item_sort : unit -> string
  method virtual getAtomic : unit -> _atomic
  method virtual getNode : unit -> _node
  method virtual string_value : unit -> string
end

let _AtomicValueKind_to_string k = 
  match k with  
    | AtomicStringKind  -> "xs:string"
    | AtomicBooleanKind -> "xs:boolean"
    | AtomicDecimalKind  -> "xs:decimal"
    | AtomicFloatKind   -> "xs:float"
    | AtomicDoubleKind  -> "xs:double"
    | AtomicAnyURIKind  -> "xs:anyURI"
    | AtomicQNameKind   -> "xs:QName"
    | AtomicIntegerKind -> "xs:integer"
    | AtomicIntKind     -> "xs:int"
    | AtomicUntypedKind -> "xs:untyped"
    | _ -> raise (Query(Internal_Error("Atomic type not implemented\n")))
      
class virtual atomicvalue =
object (self)
 inherit item
 method item_sort() = _AtomicValueKind_to_string(getAtomicValueKind(self#getAtomic()))
 method string_value() = erase_atomic_value (self#getAtomic())
 method getNode () = raise (Query(Internal_Error "Not a node"))
 method virtual getAtomic : unit -> _atomic
end

let get_sym a  = 
  match a with 
  | AtomicQName sym -> sym
  | _ -> raise (Query(Internal_Error "Not a QName"))

let qname_to_anon_symbol q = Sym.anon_name(get_sym(q#getAtomic()))

class xs_QName (q) = 
object (self)
 inherit atomicvalue as super
 val a = 
   match (q) with 
   | (Some (s), None) ->
       AtomicQName(s)
(*     AtomicQName(Sym.anon_symbol(Namespace.NSUri(uri_name), local_name)) *)
   | (None, Some a) -> a 
   | _ -> raise (Query(Internal_Error("Invalid xs:QName constructor\n")))

 method getAtomic() = a
end

let uri (q : xs_QName) = 
   match Sym.anon_name(get_sym(q#getAtomic())) with 
   | (Namespace.NSUri(uri_name), local_name) -> uri_name
   | _ -> raise (Query(Internal_Error "Not a QName"))

let local_name (q : xs_QName) = 
   match Sym.anon_name(get_sym(q#getAtomic())) with 
   | (Namespace.NSUri(uri_name), local_name) -> local_name
   | _ -> raise (Query(Internal_Error "Not a QName"))

class xs_string (s', a') =
object
 inherit atomicvalue
 val a = 
   match (s', a') with 
   | (Some s, None) -> (AtomicString s) 
   | (None, Some a) -> a
   | _ -> raise (Query(Internal_Error("Invalid xs:string constructor\n")))
 method getAtomic() = a
end

class xs_boolean (b', a') =
object
 inherit atomicvalue
 val a = 
   match (b', a') with 
   | (Some b, None) -> (AtomicBoolean b) 
   | (None, Some a) -> a
   | _ -> raise (Query(Internal_Error("Invalid xs:boolean constructor\n")))
 method getAtomic() = a
end

class xs_int (i', a') =
object
 inherit atomicvalue
 val a = 
   match (i', a') with 
   | (Some i, None) -> (AtomicInt i)
   | (None, Some a) -> a
   | _ -> raise (Query(Internal_Error("Invalid xs:int constructor\n")))
 method getAtomic() = a
end

class xs_integer (i', a') =
object
 inherit atomicvalue
 val a = 
   match (i', a') with 
   | (Some i, None) -> (AtomicInteger(i))
   | (None, Some a) -> a
   | _ -> raise (Query(Internal_Error("Invalid xs:integer constructor\n")))
 method getAtomic() = a
end

class xs_decimal (i', a') =
object
 inherit atomicvalue
 val a = 
   match (i', a') with 
   | (Some d, None) -> (AtomicDecimal(d))
   | (None, Some a) -> a
   | _ -> raise (Query(Internal_Error("Invalid xs:decimal constructor\n")))
 method getAtomic() = a
end

class xs_float (d', a') =
object
 inherit atomicvalue
 val a = 
   match (d', a') with 
   | (Some d, None) -> (AtomicFloat d)
   | (None, Some a) -> a
   | _ -> raise (Query(Internal_Error("Invalid xs:float constructor\n")))
 method getAtomic() = a
end

class xs_double (d', a') =
object
 inherit atomicvalue
 val a = 
   match (d', a') with 
   | (Some d, None) -> (AtomicDouble d)
   | (None, Some a) -> a
   | _ -> raise (Query(Internal_Error("Invalid xs:double constructor\n")))
 method getAtomic() = a
end

class xs_anyURI (u', a') =
object
 inherit atomicvalue
 val a = 
   match (u', a') with 
   | (Some u, None) -> (AtomicAnyURI(AnyURI._uri_of_string u))
   | (None, Some a) -> a
   | _ -> raise (Query(Internal_Error("Invalid xs:anyURI constructor\n")))
 method getAtomic() = a
end

class xs_untyped (s', a') =
object
 inherit atomicvalue
 val a = 
   match (s', a') with 
   | (Some s, None) -> (AtomicUntyped s)
   | (None, Some a) -> a
   | _ -> raise (Query(Internal_Error("Invalid xs:untyped constructor\n")))
 method getAtomic() = a
end

class virtual node =
object(self)
 inherit item
(* method node_name : xs_QName option *)
 method virtual item_sort : unit -> string
 method virtual getNode : unit -> _node
 method getAtomic () : _atomic = raise (Query(Internal_Error "Not an atomic value"))
 method string_value () = nodeStringValue(self#getNode())
end

class document ((d': (string * node list) option), (n' : _node option)) = 
object 
  inherit node as super
  val n = 
  match (d', n') with
  | (Some (u, kids), None) -> 
       let kids' = sequence_of_list(List.map (fun n -> n#getNode()) kids) in
       buildDocumentNode(docNode (AnyURI._uri_of_string u) kids' None)
  | (None, Some n) -> n
  | _ -> raise (Query(Internal_Error("Invalid document constructor\n")))
 method getNode()      = n 
 method item_sort() = "document"
end

class attribute ((a' : (xs_QName * xs_string * xs_QName) option), (n' : _node option)) =
(* (qname : xs_QName), (content : xs_string), (typ : xs_QName)) = *)
object(self)
 inherit node as super
 val n = 
   match (a', n') with
   | (Some (qname, content, typ), None) -> 
       (*        let attrsym = (Sym.rattr_symbol(Namespace.NSWildcardPrefix, (qname_to_anon_symbol(qname)))) in *)
       let attrsym = (Sym.rattr_symbol(Namespace.NSWildcardPrefix, (Namespace.NSUri(uri(qname)), local_name(qname)))) in
       buildAttributeNode(attrNode attrsym (content#string_value()))
   | (None, Some n) -> n
   | _ -> raise (Query(Internal_Error("Invalid attribute constructor\n")))
 method getNode()      = n 
 method item_sort() = "attribute"
end

(* Mary: the type of the argument must be explicit! *)
class text ((s' : xs_string option), (n' : _node option)) =
object
 inherit node as super
 val n = 
   match (s', n') with
   | (Some s, None) -> buildTextNode(textNode(s#string_value()))
   | (None, Some n) -> n
   | _ -> raise (Query(Internal_Error("Invalid text constructor\n")))
 method getNode()      = n 
 method item_sort() = "text"
end

(* Mary: note that the (qname : xs_QName) parameter below must be parenthecized! *)

class comment ((s' : xs_string option), (n' : _node option)) = 
object
 inherit node as super
 val n = 
   match (s', n') with
   | (Some s, None) -> buildCommentNode(commentNode(s#string_value()))
   | (None, Some n) -> n
   | _ -> raise (Query(Internal_Error("Invalid comment constructor\n")))
 method getNode()      = n
 method item_sort() = "comment"
end

class processing_instruction ((p' : (xs_string * xs_string) option), (n' : _node option)) = 
(* (target : xs_string), (content : xs_string)) =  *)
object
 inherit node as super
 val n = 
   match (p', n') with
   | (Some(target, content), None) -> 
       buildProcessingInstructionNode(piNode (target#string_value()) (content#string_value()))
   | (None, Some n) -> n
   | _ -> raise (Query(Internal_Error("Invalid processing instruction constructor\n")))
 method getNode()      = n
 method item_sort() = "processing-instruction"
end


class element ((e' : (xs_QName * attribute list * node list * xs_QName) option), (n' : _node option)) =
object(self)
 inherit node as super
 val n = 
   match (e', n') with
   | (Some (qname, attrs, kids, typ), None) -> 
       let elemsym = (Sym.relem_symbol(Namespace.NSWildcardPrefix, (Namespace.NSUri(uri(qname)), local_name(qname)))) in 
       let attrs' = sequence_of_list(List.map  (fun n -> getAttributeNode(n#getNode())) attrs) in 
       let kids' = sequence_of_list(List.map (fun n -> n#getNode()) kids) in
       buildElementNode(elemNode elemsym sequenceEmpty attrs' kids' []) 
   | (None, Some n) -> n
   | _ -> raise (Query(Internal_Error("Invalid attribute constructor\n")))
 method getNode()      = n 
 method item_sort() = "element"
end

let item_to_Item i =
  if (i#item_sort() = "xs:string" ||
      i#item_sort() = "xs:boolean" ||
      i#item_sort() = "xs:decimal" ||
      i#item_sort() = "xs:float" ||
      i#item_sort() = "xs:double" ||
      i#item_sort() = "xs:anyURI" ||
      i#item_sort() = "xs:QName" ||
      i#item_sort() = "xs:integer" ||
      i#item_sort() = "xs:int" ||
      i#item_sort() = "xs:untyped") then 
    buildAtomicValue(i#getAtomic())
  else
    buildNode(i#getNode())
(*  | _ -> raise (Query(Internal_Error ("Not a node or atomic"))) *)
  
let _Atomic_to_atomic a : atomicvalue =
  let a' : atomicvalue =
    match a with  
    | AtomicString s  -> new xs_string(None, Some a)
    | AtomicBoolean b -> new xs_boolean(None, Some a)
    | AtomicDecimal d -> new xs_decimal(None, Some a)
    | AtomicFloat f   -> new xs_float(None, Some a)
    | AtomicDouble d  -> new xs_double(None, Some a)
    | AtomicAnyURI uri -> new xs_anyURI(None, Some a)
    | AtomicQName q    -> new xs_QName(None, Some a)
    | AtomicInteger i  -> new xs_integer(None, Some a)
    | AtomicInt i      -> new xs_int(None, Some a)
    | AtomicUntyped u  -> new xs_untyped(None, Some a)
    | _ -> raise (Query(Internal_Error("Atomic type not implemented\n")))
  in a'
   
let _Node_to_node n : node = 
   begin
     match n with 
    | DocumentNode d -> new document(None, Some n)  
    | ElementNode e -> new element(None, Some n)
    | AttributeNode a -> new attribute(None, Some n)
    | TextNode t -> new text(None, Some n)
(*       | NamespaceNode n' -> ([] : node list) *)
    | ProcessingInstructionNode p -> new processing_instruction(None, Some n)
    |  CommentNode c ->  new comment(None, Some n)
    | _ -> raise (Query(Internal_Error("Node not implemented\n")))
   end

let _Item_to_item i : item =
  match i with
  | AtomicValue a -> _Atomic_to_atomic a
  | Node n -> _Node_to_node n

