(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: galapi.ml,v 1.6.2.2 2003/11/07 02:46:34 simeon Exp $ *)

open Format

open Error
open Namespace

open Xquery_ast
open Print

open Datamodel
open Datamodel_api

(* Everything you wanted to know about context, but were afraid to ask *)

type _node   = _Node 
type _atomic = Datatypes._AtomicValue
type _uri    = AnyURI._uri
type _item   = _Item

type serialization_kind =
  | Serialize_As_Well_Formed
  | Serialize_As_XQuery

type processing_context = {
   (* Phases *)
   mapping : bool ref;
    (* true when mapping is on.  mapping is on by default *)
   typing : bool ref;  
    (* true when typing is on.   typing is off by default *)
   projection : bool ref;
    (* true when projection of document is on.   projection is off by default *)
   optimized_projection : bool ref;
    (* true when projection of document is on.  projection is off by default *)
   optim : bool ref;
    (* true when optimization is on. optimization is on by default *)
   eval : bool ref;
    (* true when evaluation is on.  evaluation is on by default *)

  (* Normalization and evaluation behavior *)
   do_sbdo : bool ref;
    (* true when sort-by-document-order is enabled.  on by default *)

  (* Debugging behavior *)
   debugging : bool ref;
    (* true when debugging is on. debugging is off by default *)
   verbose : bool ref;
    (* true if verbose output emitted. false by default  *)

  (* Whitespace behavior *)
   xquery_whitespace : bool ref;
     (* true if preserving whitespace in XQuery constructors
	false if not preserving non-significant whitespace *)

   xml_whitespace : bool ref;
     (* true if preserving whitespace in input XML documents
	false if not preserving non-significant whitespace *)

  (* PI and Comments behavior *)
   xml_pis_and_comments : bool ref;
    (* true if preserving PI's and comments in input XML documents
       false if not preserving PI's and comments. false by default *)

  (* Output behavior *)
   serialize : serialization_kind ref;
    (* specifies how to serialize XML output *)
    (* Serialize_As_XQuery by default *)

   serialize_xsi_type : bool ref;
    (* true if xsi attributes are serialized. false by default *)

   pretty_print        : bool ref;
     (* true if pretty printing XML output; false by default. *)
   pretty_print_output    : Pervasives.out_channel ref;
   pretty_print_formatter : Format.formatter ref;

   print_xml : bool ref;
    (* true if core expression emitted. true by default *)
   xml_output : out_channel ref ;  (* stdout by default *)
   xml_formatter : Format.formatter ref;

   print_expr : bool ref;
    (* true if expression emitted. false by default *)
   expr_output : out_channel ref;
   expr_formatter : Format.formatter ref;

   print_type : bool ref;
    (* true if type emitted. false by default *)
   type_output : out_channel ref;
   type_formatter : Format.formatter ref;

   print_core_expr : bool ref;
    (* true if core expression emitted. false by default *)
   core_expr_output : out_channel ref;
   core_expr_formatter : Format.formatter ref;

   print_typed_expr : bool ref;
    (* true if typed core expression emitted. false by default *)
   typed_expr_output : out_channel ref;
   typed_expr_formatter : Format.formatter ref;

   print_type_annotations : bool ref;
    (* true if types of expressions are printed with the core. false by default  *)

   print_optimized_expr : bool ref;
    (* true if optimized core expression emitted. false by default. *)
   optimized_expr_output : out_channel ref;
   optimized_expr_formatter : Format.formatter ref;

   print_projection : bool ref;
    (* true when printing projection of document is on.  off by default *)
   projection_output : out_channel ref;
   projection_formatter : Format.formatter ref;

   print_projected_file : bool ref;
   projected_file_output : out_channel ref;
   projected_file_formatter : Format.formatter ref;

   monitor : bool ref;
   monitor_output : out_channel ref;
   monitor_formatter : Format.formatter ref;
  } 

type external_context = 
    { opt_context_item : Datamodel_api.item option;
      external_vars : (vname * Datamodel_api.item list) list } 

type module_context = 
    { proc_context    : processing_context ; (* should be moved to base/conf somewhere *)
      core_library    : Xquery_ast.cquery_module ;
      dynamic_context : Dynamic_context.dynamic_context } 

type xquery_statement = Xquery_ast.statement


(*****************************) 
(* Atomic value constructors *)
(*****************************) 

let xs_string (s: Datatypes.xs_string) = new xs_string(Some s, None)
let xs_boolean (b:Datatypes.xs_boolean) = new xs_boolean(Some b, None)
let xs_int (i:Datatypes.xs_int) = new xs_int(Some i, None)
let xs_integer (i:Datatypes.xs_integer) = new xs_integer(Some i, None)
let xs_decimal (i:Datatypes.xs_decimal) = new xs_decimal(Some i, None)
let xs_float (f:Datatypes.xs_float) = new xs_float(Some f, None)
let xs_double (d:Datatypes.xs_double) = new xs_double(Some d, None)
let xs_anyURI (u : string) = new xs_anyURI(Some u, None)
let xs_untyped (s : string) = new xs_untyped(Some s, None)
let xs_QName ((m: module_context), (s: string)) =  
  let nsenv = Norm_context.nsenv_from_norm_context(
    Static_context.norm_context_from_stat_context(
    Dynamic_context.static_context_from_dynamic_context m.dynamic_context)) 
  in new xs_QName(Some(Datatypes.qname_of_untyped nsenv s), None)
let document (uri, kids) = new document (Some(uri, kids), None)
let element (n, al, kids, typename) = new element (Some(n, al, kids, typename), None)
let text (s:xs_string) = new text(Some s,None)
let comment (s:xs_string) = new comment(Some s,None)
let attribute (qname,s,qname1) = new attribute(Some (qname,s,qname1),None)
let processing_instruction (s,s1) = new processing_instruction(Some (s,s1),None)

(**********************) 
(* Accessors on nodes *)
(**********************) 

let item_kind (i : item) = i#item_sort()
let get_node i : item = _Node_to_node(i#getNode())
let get_atomicvalue i : item = _Atomic_to_atomic(i#getAtomic())

let get_element i : item = 
  let n = i#getNode() in
  begin
    match n with
    | ElementNode e -> new element(None, Some n)
    | _ -> raise (Query(Internal_Error("Item not an element\n")))
   end

let get_attribute i : item = 
  let n = i#getNode() in
  begin
    match n with
    | AttributeNode _ -> new attribute(None, Some n)
    | _ -> raise (Query(Internal_Error("Item not an attribute\n")))
   end

let get_text i : item = 
  let n = i#getNode() in
  begin
    match n with
    | TextNode _ -> new text(None, Some n)
    | _ -> raise (Query(Internal_Error("Item not a text node\n")))
   end

let get_comment i : item = 
  let n = i#getNode() in
  begin
    match n with
    | TextNode _ -> new comment(None, Some n)
    | _ -> raise (Query(Internal_Error("Item not a comment\n")))
   end

let get_processing_instruction i : item = 
  let n = i#getNode() in
  begin
    match n with
    | TextNode _ -> new processing_instruction(None, Some n)
    | _ -> raise (Query(Internal_Error("Item not a processing-instruction\n")))
   end

let get_xs_string i : item = 
  let a = i#getAtomic() in
  begin
    match a with
    | Datatypes.AtomicString s -> new xs_string(None, Some a)
    | _ -> raise (Query(Internal_Error("Item not a xs_string\n")))
   end

let get_xs_boolean i : item = 
  let a = i#getAtomic() in
  begin
    match a with
    | Datatypes.AtomicBoolean b -> new xs_boolean(None, Some a)
    | _ -> raise (Query(Internal_Error("Item not a xs_boolean\n")))
   end

let get_xs_int i : item = 
  let a = i#getAtomic() in
  begin
    match a with
    | Datatypes.AtomicInt b -> new xs_int(None, Some a)
    | _ -> raise (Query(Internal_Error("Item not a xs_int\n")))
   end

let get_xs_integer i : item = 
  let a = i#getAtomic() in
  begin
    match a with
    | Datatypes.AtomicInteger b -> new xs_integer(None, Some a)
    | _ -> raise (Query(Internal_Error("Item not a xs_integer\n")))
   end

let get_xs_decimal i : item = 
  let a = i#getAtomic() in
  begin
    match a with
    | Datatypes.AtomicDecimal b -> new xs_decimal(None, Some a)
    | _ -> raise (Query(Internal_Error("Item not a xs_decimal\n")))
   end

let get_xs_float i : item = 
  let a = i#getAtomic() in
  begin
    match a with
    | Datatypes.AtomicFloat b -> new xs_float(None, Some a)
    | _ -> raise (Query(Internal_Error("Item not a xs_float\n")))
   end

let get_xs_double i : item = 
  let a = i#getAtomic() in
  begin
    match a with
    | Datatypes.AtomicDouble b -> new xs_float(None, Some a)
    | _ -> raise (Query(Internal_Error("Item not a xs_double\n")))
   end

let get_xs_anyURI i : item = 
  let a = i#getAtomic() in
  begin
    match a with
    | Datatypes.AtomicAnyURI b -> new xs_anyURI(None, Some a)
    | _ -> raise (Query(Internal_Error("Item not a xs_anyURI\n")))
   end

let get_xs_QName i : item = 
  let a = i#getAtomic() in
  begin
    match a with
    | Datatypes.AtomicQName b -> new xs_QName(None, Some a)
    | _ -> raise (Query(Internal_Error("Item not a xs_QName\n")))
   end

let get_xs_untyped i : item = 
  let a = i#getAtomic() in
  begin
    match a with
    | Datatypes.AtomicUntyped b -> new xs_untyped(None, Some a)
    | _ -> raise (Query(Internal_Error("Item not a xs_untyped\n")))
   end

let parent n = 
   let ns = list_of_sequence(parent_axis(n#getNode())) in
   match ns with
   | [] -> ([] : node list)
   | [n] -> [_Node_to_node n]
   | _ -> raise (Query(Internal_Error("Node has more than one parent\n")))

let children n =
   let ns = list_of_sequence(child_axis(n#getNode())) in
   List.map _Node_to_node ns

let attributes n =
   let ns = list_of_sequence(attribute_axis(n#getNode())) in
   List.map _Node_to_node ns

let string i = i#string_value()
(* nodeStringValue(n#getNode())*)

let node_kind n = 
  match (nodeKind(n#getNode())) with 
  | DocumentNodeKind -> "document"
  | ElementNodeKind  -> "element"
  | AttributeNodeKind -> "attribute"
  | TextNodeKind -> "text" 
  | NamespaceNodeKind -> "namespace"
  | ProcessingInstructionNodeKind -> "processing instruction"
  | CommentNodeKind -> "comment"

let namespace n = raise (Query(Internal_Error "Namespace access not available."))

let base_uri n = 
  match (Datamodel.base_uri(n#getNode())) with
  | None -> None
  | Some uri -> Some(new xs_anyURI(Some(AnyURI._string_of_uri uri), None))

let node_name n = 
  match (getNodeName(n#getNode())) with
  | None -> None
  | (Some (rqname)) -> Some(new xs_QName(Some(Sym.anon_symbol(rqname)),None))

let typed_value n = 
  let v = Datamodel.typed_value(n#getNode()) in
  List.map _Item_to_item (list_of_sequence v)

(************************) 
(* Conversion functions *)
(************************) 

let string_of_atomicvalue (a : atomicvalue) = (a#string_value())

let boolean_of_xs_boolean (b : xs_boolean) = Datatypes.getAtomicBoolean(b#getAtomic())  

let int_of_xs_int (i : xs_int) = Datatypes.getAtomicInt(i#getAtomic())

let integer_of_xs_integer (i : xs_integer) = Datatypes.getAtomicInteger(i#getAtomic())

let decimal_of_xs_decimal (d : xs_decimal) = Datatypes.getAtomicDecimal(d#getAtomic())

let float_of_xs_float (f : xs_float) = Datatypes.getAtomicFloat(f#getAtomic())

let float_of_xs_double (d : xs_double) = Datatypes.getAtomicDouble(d#getAtomic())


(********************************)
(* Document I/O functions       *)
(********************************)

let print_escaped_output ff header str footer =
  if (!Conf.verbose) then pp_print_string ff header; 
  pp_print_string ff (!Conf.xml_charescape_fn str); 
  if (!Conf.verbose) then pp_print_string ff footer;
  pp_print_flush ff ()

let load_document file =
  let uri = AnyURI._uri_of_string file in

  let xml_stream = Sax_parser.open_xml_stream_from_file file in
  let v = Load.load_xml_document_from_sax uri xml_stream in
  let nl = List.map getNode (list_of_sequence v) in
  let nl' = List.map _Node_to_node nl in
  nl' 

let load_document_validate mod_ctxt file =
  raise (Query(Prototype("load_document_validate not yet implemented")))

let serialize_to_formatter fmt v =
  let v' = sequence_of_list(List.map item_to_Item v) in   
  if !Conf.print_xml then 
    begin
      let apply_serialize res =	
	let s =
	  begin
	    Serialization.fserialize_datamodel Format.str_formatter res;
	    Format.pp_print_newline Format.str_formatter ();
	    Format.flush_str_formatter()
	  end
	in print_escaped_output !Conf.xml_formatter !Conf.xml_header s !Conf.xml_footer
      in
      Monitor.wrap_monitor Monitor.Query_Serialization_Phase apply_serialize v';
    end

let serialize_to_string v =
  let v' = sequence_of_list(List.map item_to_Item v) in   
  Serialization.fserialize_datamodel Format.str_formatter v';
  Format.pp_print_newline Format.str_formatter ();
  !Conf.xml_charescape_fn(Format.flush_str_formatter ())

let serialize_to_file f v =
  let v' = sequence_of_list(List.map item_to_Item v) in   
  let oc = open_out f in
  let foc = Format.formatter_of_out_channel oc in
  begin
    let apply_serialize res =
      begin
	Serialization.fserialize_datamodel Format.str_formatter res;
	Format.pp_print_newline Format.str_formatter ();
	Format.flush_str_formatter()
      end
    in
    let s = !Conf.xml_charescape_fn (apply_serialize v') in
    pp_print_string foc s; 
    close_out oc
  end

(* Default processing context *)
let default_processing_context () = 
  { mapping = ref true; 
    typing  = ref false;

    projection           = ref false;
    optimized_projection = ref false;

    optim   = ref true;
    eval    = ref true;
  
    (* Normalization and evaluation behavior *)
    do_sbdo = ref true;
  
    (* Debugging behavior *)
    debugging = ref false;
    verbose   = ref false;
  
    (* Whitespace behavior *)
    xquery_whitespace = ref false;
    xml_whitespace    = ref false;
  
    (* PI and Comments behavior *)
    xml_pis_and_comments = ref false;
  
    (* Output behavior *)
    serialize          = ref Serialize_As_XQuery;
    serialize_xsi_type = ref false;
  
    pretty_print           = ref false;
    pretty_print_output    = ref stdout; 
    pretty_print_formatter = ref Format.std_formatter; 
  
    print_xml  	  	   = ref true;
    xml_output 	  	   = ref stdout;
    xml_formatter 	   = ref Format.std_formatter; 
  
    print_expr     = ref false;
    expr_output    = ref stdout;
    expr_formatter = ref Format.std_formatter; 

    print_type = ref  false;
    type_output = ref  stdout;
    type_formatter  = ref  Format.std_formatter; 

    print_core_expr = ref  false;
    core_expr_output = ref  stdout;
    core_expr_formatter = ref  Format.std_formatter; 
  
    print_typed_expr = ref false;
    typed_expr_output = ref  stdout;
    typed_expr_formatter = ref  Format.std_formatter; 
  
    print_optimized_expr = ref  false;
    optimized_expr_output = ref  stdout;
    optimized_expr_formatter = ref Format.std_formatter; 
  
    print_projection = ref false;
    projection_output = ref  stdout;
    projection_formatter = ref  Format.std_formatter; 
  
    print_projected_file = ref false;
    projected_file_output = ref stdout;
    projected_file_formatter = ref Format.std_formatter; 
  
    print_type_annotations = ref false;
  
    (* Monitoring *)
    monitor = ref false;
    monitor_output = ref  stdout;
    monitor_formatter = ref   Format.std_formatter; 
  }

let load_standard_library pc = 
  let q = Toputils.load_pervasive () in        (* loading stdlib *)
  let norm_ctxt = Norm_context.build_norm_context (Namespace.empty_nsenv()) (Schema.empty_schema()) [] in
  let (norm_ctxt', cqm) = Mapping.map_library_module norm_ctxt q in

  let stat_ctxt = Static_context.build_static_context norm_ctxt' [] in
  (* Type it *)
  let stat_ctxt' =
    if !(Conf.typing)
    then Typing.type_module stat_ctxt cqm
    else stat_ctxt
  in
  (* Rewrite it *)
  let cqm' = 
    if !(Conf.optim)
    then Rewrite_main.rewrite_module stat_ctxt' cqm
    else cqm
  in
  { proc_context = pc;
    core_library = cqm';
    dynamic_context = Dynamic_context.build_dynamic_context stat_ctxt' cqm'.pcmodule_functions Finfo.bogus } 

(* Apply projection if requested ! *)

let build_projection  mod_ctxt cstmt =
  let core_query_module =
    Xquery_astutil.fmkcquery_main_module_from_library_module (mod_ctxt.core_library) [cstmt]
  in
  begin
    if !Conf.projection
    then
      begin
	let apply_projection core_query_module =
	  let final_projection = Path_analysis.query_path_analysis core_query_module in
	  if !Conf.print_projection
	  then
	    Path_analysis.print_full_analysis !Conf.projection_formatter final_projection;
	  Store.add_projections final_projection
	in
	Monitor.wrap_monitor Monitor.Query_Path_Analysis_Phase apply_projection core_query_module
      end
  end;
      
  begin
    if !Conf.optimized_projection
    then
      begin
	let apply_optimized_projection core_query_module =
	  let final_projection = Path_analysis.query_path_analysis_optim core_query_module in
	  if !Conf.print_projection
	  then
	    Path_analysis.print_full_analysis !Conf.projection_formatter final_projection;
	  Store.add_projections final_projection
	in
	Monitor.wrap_monitor Monitor.Query_Path_Analysis_Phase apply_optimized_projection core_query_module
      end
  end

let aux_import_library_module mod_ctxt qm =
  (* Parse library module *)
  let dyn_ctxt = mod_ctxt.dynamic_context in 
  let stat_ctxt = (Dynamic_context.static_context_from_dynamic_context mod_ctxt.dynamic_context) in
  let norm_ctxt = (Static_context.norm_context_from_stat_context stat_ctxt) in

  (* Normalize library module using existing norm context *)
  let (norm_ctxt', cqm) =
    Mapping.map_library_module norm_ctxt qm
  in

  (* Build a new static context using existing static context *)
  let stat_ctxt' =
    Static_context.build_static_context norm_ctxt' (Static_context.vars_from_static_context stat_ctxt) 
  in

  (* Type it *)
  let stat_ctxt'' =
    if !(Conf.typing)
    then Typing.type_module stat_ctxt' cqm
    else stat_ctxt'
  in

  (* Rewrite it *)
  let cqm' = 
    if !(Conf.optim)
    then Rewrite_main.rewrite_module stat_ctxt'' cqm
    else cqm
  in
  (* Build a new dynamic context merged with existing dynamic context *)
  (* Add new function definitions to existing function definitions *)
  let xx = (Dynamic_context.copy_dynamic_context dyn_ctxt stat_ctxt'')  in
  let dyn_ctxt' = Dynamic_context.add_funs_to_dynamic_context xx (cqm'.pcmodule_functions) Finfo.bogus in

  (* Merge new library's var & key expressions with existing var & key expressions *)
  let cqm'' =
    { pcmodule_functions = [];
      pcmodule_vars  	 = mod_ctxt.core_library.pcmodule_vars @ cqm'.pcmodule_vars;
      pcmodule_keys  	 = mod_ctxt.core_library.pcmodule_keys @ cqm'.pcmodule_keys;
      pcmodule_exprs 	 = [] }
  in
  let mod_ctxt' =
    { proc_context    = mod_ctxt.proc_context;
      core_library    = cqm'';
      dynamic_context = dyn_ctxt' }
  in mod_ctxt'

let import_library_module mod_ctxt filename =
  (* Parse library module *)
  let qm = (Parseutil.parseFXQuery filename) in
  match qm.pmodule_exprs with 
  | [] -> aux_import_library_module mod_ctxt qm
  | _ -> raise (Query (Internal_Error("Library module contains top-level statements.")))

let norm_type_optim_expr mod_ctxt expr = 
  let dyn_ctxt = mod_ctxt.dynamic_context in 
  let stat_ctxt = (Dynamic_context.static_context_from_dynamic_context dyn_ctxt) in
  let norm_ctxt = (Static_context.norm_context_from_stat_context stat_ctxt) in

  if (!Conf.print_expr) then
    begin 
      let s = bprintf_expr "" expr in
      print_escaped_output !Conf.expr_formatter !Conf.expr_header s !Conf.expr_footer; 
      (* Used for testing the pretty-printer *)
      if !Conf.pretty_print then
	begin
	  print_escaped_output !Conf.pretty_print_formatter "" s "";
	  fprintf !Conf.pretty_print_formatter "%s@\n@\n" ";"
	end;
    end;
  
  let ce = Mapping.map_expr norm_ctxt expr in
  (* The Query *)
  
  (* The expanded query *)
  if (!Conf.print_core_expr) then
    begin
      let s = bprintf_cexpr "" ce in
      print_escaped_output !Conf.core_expr_formatter !Conf.core_expr_header s !Conf.core_expr_footer
    end;

  (* Static type checking phase *)

  if !Conf.typing
  then
    begin
      let st =
	let apply_typing ce =
	  let t =  Typing.type_cexpr stat_ctxt ce in
	  (Simplification.simplify_ty (Norm_context.schema_from_norm_context norm_ctxt) t)
	in
	Monitor.wrap_monitor Monitor.Query_Static_Typing_Phase apply_typing ce
      in

      (* The type *)
      if (!Conf.print_type) then 
	begin
	  let s = bprintf_result_type "" st in
	  print_escaped_output !Conf.type_formatter !Conf.type_header s !Conf.type_footer
	end;
    end;
  
  if (!Conf.print_typed_expr) then
    begin
      let s = bprintf_cexpr "" ce in
      print_escaped_output !Conf.typed_expr_formatter !Conf.typed_expr_header s !Conf.typed_expr_footer
    end;

  (* Rewriting phase *)

  let ce =
    if !Conf.optim
    then
      let local_ce =
	Monitor.wrap_monitor Monitor.Query_Optimization_Phase (Rewrite_main.rewrite_cexpr stat_ctxt) ce
      in
      begin
	if (!Conf.print_optimized_expr) then
	  begin
	    let s = bprintf_cexpr "" local_ce in
	    print_escaped_output !Conf.optimized_expr_formatter !Conf.optimized_expr_header s !Conf.optimized_expr_footer
	  end;
	local_ce
      end
    else
      ce
  in ce

let norm_type_optim_update mod_ctxt update =
  let dyn_ctxt = mod_ctxt.dynamic_context in 
  let stat_ctxt = (Dynamic_context.static_context_from_dynamic_context dyn_ctxt) in
  let norm_ctxt = (Static_context.norm_context_from_stat_context stat_ctxt) in

  let cu = Mapping.map_update norm_ctxt update in
  (* The Query *)
  if (!Conf.print_expr) then
    begin 
      let update = 
	match cu.pcupdate_origin with
	| Some u -> u
	| None ->
	    raise (Query (Internal_Error "Could not find original update"));
      in
      if !Conf.verbose then Printf.fprintf (!Conf.expr_output) "%s" !Conf.expr_header; 
      (* Used for testing the pretty-printer *)
      pp_print_flush !Conf.expr_formatter ();
      if !Conf.pretty_print
      then
	begin
	  let formatter = !Conf.pretty_print_formatter in
	  begin
	    fprintf_update formatter "" update;
	    fprintf formatter "%s@\n@\n" ";"
	  end;
	end;
      fprintf_update !Conf.expr_formatter "" update;
      if !Conf.verbose then pp_print_string !Conf.expr_formatter !Conf.expr_footer;
      pp_print_flush !Conf.expr_formatter ();
    end;
  
  (* The expanded query *)
  if (!Conf.print_core_expr) then
    begin
      if (!Conf.verbose) then Printf.fprintf !Conf.core_expr_output "%s" !Conf.core_expr_header;
      pp_print_flush !Conf.core_expr_formatter ();
      fprintf_cupdate !Conf.core_expr_formatter "" cu;
      if (!Conf.verbose) then Printf.fprintf !Conf.core_expr_output "%s" !Conf.core_expr_footer;
      pp_print_flush !Conf.core_expr_formatter ();
    end;

  (* Static type checking phase *)

  if !Conf.typing
  then
    begin
      let st =
	let apply_typing cu =
	  let t =  Typing.type_cupdate stat_ctxt cu in
	  (Simplification.simplify_ty (Norm_context.schema_from_norm_context norm_ctxt) t)
	in
	Monitor.wrap_monitor Monitor.Query_Static_Typing_Phase apply_typing cu
      in

      (* The type *)
      if (!Conf.print_type) then 
	begin
	  if (!Conf.verbose) then Printf.fprintf !Conf.type_output "%s" !Conf.type_header; 
	  pp_print_flush !Conf.type_formatter (); 
	  fprintf_result_type !Conf.type_formatter "" st;
	  if (!Conf.verbose) then Printf.fprintf !Conf.type_output "%s" !Conf.type_footer;
	  pp_print_flush !Conf.type_formatter (); 
	end;
    end;
  
  if (!Conf.print_typed_expr) then
    begin
      if (!Conf.verbose) then Printf.fprintf !Conf.typed_expr_output "%s" !Conf.typed_expr_header;
      pp_print_flush !Conf.typed_expr_formatter ();
      fprintf_cupdate !Conf.typed_expr_formatter "" cu;
      if (!Conf.verbose) then Printf.fprintf !Conf.typed_expr_output "%s" !Conf.typed_expr_footer;
      pp_print_flush !Conf.typed_expr_formatter ();
    end;

  (* Rewriting phase *)

  let cu =
    if !Conf.optim
    then
      let local_cu =
	Monitor.wrap_monitor Monitor.Query_Optimization_Phase (Rewrite_main.rewrite_cupdate stat_ctxt) cu
      in
      begin
	if (!Conf.print_optimized_expr) then
	  begin
	    if (!Conf.verbose) then Printf.fprintf !Conf.optimized_expr_output "%s" !Conf.optimized_expr_header;
	    pp_print_flush !Conf.optimized_expr_formatter ();
	    fprintf_cupdate !Conf.optimized_expr_formatter "" local_cu;
	    if (!Conf.verbose) then Printf.fprintf !Conf.optimized_expr_output "%s" !Conf.optimized_expr_footer;
	    pp_print_flush !Conf.optimized_expr_formatter ();
	  end;
	local_cu
      end
    else
      cu
  in cu

let norm_type_optim_stmt mod_ctxt stmt = 
  let dyn_ctxt = mod_ctxt.dynamic_context in 
  let stat_ctxt = (Dynamic_context.static_context_from_dynamic_context dyn_ctxt) in
  let norm_ctxt = (Static_context.norm_context_from_stat_context stat_ctxt) in
  let fi = stmt.pstatement_loc in
  if (!Conf.mapping) then 
    match stmt.pstatement_desc with 
    | SExpression e -> Xquery_astutil.fmkcstatement (CSExpression (norm_type_optim_expr mod_ctxt e)) fi
    | SUpdate u -> Xquery_astutil.fmkcstatement (CSUpdate (norm_type_optim_update mod_ctxt u)) fi
  else Xquery_astutil.fmkcstatement (CSExpression (Xquery_astutil.fmkcexpr CEEmpty None None fi)) fi

let import_main_module mod_ctxt filename = 
  (* Parse main module *)
  let qm = (Parseutil.parseFXQuery filename) in
  if !Conf.print_module then
    begin
      fprintf_module std_formatter "" qm;
      flush stdout
    end;
  let stmts = qm.pmodule_exprs in 
  let qm' = 
    { pmodule_sorts = qm.pmodule_sorts;
      pmodule_functions = qm.pmodule_functions;
      pmodule_contexts = qm.pmodule_contexts;
      pmodule_vars = qm.pmodule_vars;
      pmodule_keys = qm.pmodule_keys;
      pmodule_exprs =  [] }
  in

  let mod_ctxt' = aux_import_library_module mod_ctxt qm' in
  (mod_ctxt', stmts)

let import_main_module_from_string mod_ctxt query_string = 
  (* Parse main module *)
  let qm = (Parseutil.parseStringFXQuery query_string) in
  let stmts = qm.pmodule_exprs in 
  let qm' = 
    { pmodule_sorts = qm.pmodule_sorts;
      pmodule_functions = qm.pmodule_functions;
      pmodule_contexts = qm.pmodule_contexts;
      pmodule_vars = qm.pmodule_vars;
      pmodule_keys = qm.pmodule_keys;
      pmodule_exprs =  [] }
  in

  let mod_ctxt' = aux_import_library_module mod_ctxt qm' in 
  (mod_ctxt', stmts)

(*
  let cstmts = List.map (norm_type_optim_stmt mod_ctxt') stmts in 
  (mod_ctxt', cstmts)
*)

let build_external_context opt_ctxt_item var_val_list = 
  { opt_context_item = opt_ctxt_item; 
    external_vars =  List.map (fun (vn, vl) -> (Namespace.uqname_element_of_string vn, vl)) var_val_list }

let add_external_context mod_ctxt ext_ctxt = 
  let dyn_ctxt = mod_ctxt.dynamic_context in
  let dyn_ctxt' = 
    match ext_ctxt.opt_context_item with
    | None -> dyn_ctxt
    | Some v -> 
	let v' = sequence_of_list([item_to_Item v]) in   
	Dynamic_context.add_var_value_to_dynamic_context dyn_ctxt (Xquery_astutil.dot, v') 
  in
  let dyn_ctxt'' = 
    List.fold_left (fun dyn_ctxt (vn, v) -> Dynamic_context.add_var_value_to_dynamic_context dyn_ctxt (vn, sequence_of_list(List.map item_to_Item v))) dyn_ctxt' ext_ctxt.external_vars
  in 
  { proc_context = mod_ctxt.proc_context;
    core_library = mod_ctxt.core_library;
    dynamic_context = dyn_ctxt'' }

let eval_global_variables mod_ctxt = 
  let toplevel_vars = mod_ctxt.core_library.pcmodule_vars in
  let eval_var dyn_ctxt var_decl = 
    let (v, ce) = var_decl.pcvar_decl_desc in
    let vval = Eval_expr.eval_cexpr dyn_ctxt ce in
    Dynamic_context.add_var_value_to_dynamic_context dyn_ctxt (v, vval)
  in
  let dyn_ctxt' = Eval_top.eval_toplevel mod_ctxt.dynamic_context mod_ctxt.core_library in
  begin
    { proc_context = mod_ctxt.proc_context ;
      core_library = mod_ctxt.core_library;
      dynamic_context = dyn_ctxt' } 
  end

let eval_statement mod_ctxt s = 
  let cs = norm_type_optim_stmt mod_ctxt s in 
  begin
    (* Build projection paths here! - Jerome *)
    build_projection mod_ctxt cs;
    let dyn_ctxt = mod_ctxt.dynamic_context in 
    let opt_res = 
      if !Conf.eval then
	match cs.pcstatement_desc with
	| CSExpression ce ->
	    begin
	      let res =
		Monitor.wrap_monitor Monitor.Query_Evaluation_Phase (Eval_expr.eval_cexpr dyn_ctxt) ce
	      in 
	      Some res
	    end
	| CSUpdate cu -> 
	    begin
	      Monitor.wrap_monitor Monitor.Query_Evaluation_Phase (Eval_update.eval_cupdate dyn_ctxt) cu;
	      None
	    end
      else 
	None
    in
    match opt_res with
    | None -> None
    | Some res -> Some (List.map _Item_to_item (list_of_sequence res))
  end

let eval_statement_internal mod_ctxt qs = 
  let dyn_ctxt = mod_ctxt.dynamic_context in 
  let stat_ctxt = (Dynamic_context.static_context_from_dynamic_context mod_ctxt.dynamic_context) in
  let norm_ctxt = (Static_context.norm_context_from_stat_context stat_ctxt) in

  eval_statement mod_ctxt qs

let eval_statement_from_string mod_ctxt query_string = 
  (* Parse statement *)
  let qs =  Parseutil.parse_statement_from_string query_string in
  eval_statement_internal mod_ctxt qs 

let eval_statement_from_file mod_ctxt query_file = 
  (* Parse statement from a given file *)
  let qs =  Parseutil.parse_statement_from_file query_file in
  eval_statement_internal mod_ctxt qs 

let eval_query_function mod_ctxt funname params = 

  let dyn_ctxt = mod_ctxt.dynamic_context in 
  let stat_ctxt = (Dynamic_context.static_context_from_dynamic_context mod_ctxt.dynamic_context) in
  let norm_ctxt = (Static_context.norm_context_from_stat_context stat_ctxt) in

  (* Pair each parameter value with a dummy variable and variable expression *)
  let idg = Id.create 0 in

  let (var_args, var_cexprs) = 
    List.split 
      (List.map (fun param -> 
	let var_name = "_root_var" ^ (string_of_int (Id.next idg)) in
	let var_expr = Xquery_astutil.fmkcexpr (CEVar (NSDefaultElementPrefix, var_name)) None None Finfo.bogus in
	((var_name, param), var_expr))
       params)
  in 
  let ext_ctxt = build_external_context None var_args in
  let mod_ctxt' = add_external_context mod_ctxt ext_ctxt in
  let dyn_ctxt' = mod_ctxt'.dynamic_context in 

  let fname = Namespace.uqname_function_of_string funname in
  let (_,rfname) = Resolve.resolve_function_qname (Norm_context.nsenv_from_norm_context norm_ctxt) fname in
  let ce = Maputil.map_function_app norm_ctxt rfname var_cexprs None Finfo.bogus  in

  (* Type it! Note that typing has side-effect on cexpr *)
  if !Conf.typing then
    let cet = Monitor.wrap_monitor Monitor.Query_Static_Typing_Phase (Typing.type_cexpr stat_ctxt) ce
    in ()
  else ();
  let ce' = 
  (* Rewrite it! *)
    if !Conf.optim then
      Monitor.wrap_monitor Monitor.Query_Optimization_Phase (Rewrite_main.rewrite_cexpr stat_ctxt) ce
    else ce
  in
  let v = Monitor.wrap_monitor Monitor.Query_Evaluation_Phase (Eval_expr.eval_cexpr dyn_ctxt') ce' in
  List.map _Item_to_item (list_of_sequence v)

(* Quick start evaluation functions *)

let eval_statement_with_context_item_from_document file querystr = 
  let proc_ctxt = default_processing_context() in
  let mod_ctxt = load_standard_library(proc_ctxt) in
  let xmldoc = load_document file in
  match xmldoc with 
  | [n] -> 
      let ext_ctxt = build_external_context (Some n) [] in
      let mod_ctxt' = add_external_context mod_ctxt ext_ctxt in
      eval_statement_from_string mod_ctxt' querystr
  | _ -> raise (Query(Load_Error ("Input document "^file^" contains more than one root-level node.")))

let eval_statement_with_context_item_from_xml ctxt_item querystr = 
  let proc_ctxt = default_processing_context() in
  let mod_ctxt = load_standard_library(proc_ctxt) in
  let ext_ctxt = build_external_context (Some ctxt_item) [] in
  let mod_ctxt' = add_external_context mod_ctxt ext_ctxt in
  eval_statement_from_string mod_ctxt' querystr

let eval_statement_from_file_with_context_item_from_xml ctxt_item queryfile = 
  let proc_ctxt = default_processing_context() in
  let mod_ctxt = load_standard_library(proc_ctxt) in
  let ext_ctxt = build_external_context (Some ctxt_item) [] in
  let mod_ctxt' = add_external_context mod_ctxt ext_ctxt in
  eval_statement_from_file mod_ctxt' queryfile

let eval_statement_with_variables_from_xml var_val_list querystr = 
  let proc_ctxt = default_processing_context() in
  let mod_ctxt = load_standard_library(proc_ctxt) in
  let ext_ctxt = build_external_context None var_val_list in
  let mod_ctxt' = add_external_context mod_ctxt ext_ctxt in
  eval_statement_from_string mod_ctxt' querystr
