/***********************************************************************/
/*                                                                     */
/*                                 GALAX                               */
/*                              XQuery Engine                          */
/*                                                                     */
/*  Copyright 2001-2003, Lucent Technologies.                          */
/*  Distributed only by permission.                                    */
/*                                                                     */
/***********************************************************************/


/* $Id: Galax.java,v 1.2 2003/06/12 02:30:44 simeon Exp $ */

/**
 * @(#)Galax.java
 *
 */

package galapi;


/**
 * The <code>Galax</code> class contains methods needed to perform 
 * queries and to manipulate data under different representations.
 *
 */

public class Galax
{
    static ProcessingContext defPC;
    static {    
	try {
	    System.out.flush ();
	    System.loadLibrary("glxoptj");
	} catch (Exception e) {
	    System.err.println ("cannot load native glx library");
	    System.err.println ("trying with bytecode version");
	    System.err.flush ();
            try {
  	       System.loadLibrary("glxj");
            } catch (Exception e2) {
	       System.err.println ("cannot load bytecode glx library");
	       System.err.println ("Aborting!");
 	       System.exit (1);
            }
	}
    }



    /**
     *
     * Does the initialization stuff, that is it passes program arguments 
     * to the OCaml Sys module.
     *
     */
    public static void init ()
	throws GalapiException
    {

	nativeInitialize ();

	int i = nativeDefaultProcessingContext ();
	defPC = new ProcessingContext (i);
	
    }// init()
    

    /*
     * initialize
     */
    private static native void nativeInitialize ();


    /**************************/
    /* Document I/O functions */
    /**************************/

    /**
     * Loads an XML document into the Galax data model.
     *
     * @param filename The name of the file containing the XML document
     * @return a list of nodes giving the Galax representation of the document
     *
     */

    public static NodeList loadDocument (String filename)
	throws GalapiException
    {
	//	int i = -1;

	int i = nativeLoadDocument (filename);

	return new NodeList (i);
    }
    
    
    /**
     * Loads an XML document and validates it with respect to a 
     * {@link galapi.ModuleContext ModuleContext} which may contain 
     * imported schemas.
     *
     * @param mc the context used for validation
     * @param filename the name of the file containing the XML document
     */

    public static NodeList loadDocumentValidate (ModuleContext mc, String filename)
	throws GalapiException
    {

	//	int i = -1;
	int i = nativeLoadDocumentValidate (mc.getNativeItem(), filename);

	return new NodeList (i);
    }
    
    /**
     * 
     * Serializes an XQuery data model value to a string
     *
     * @param xml a collection of items representing an XML value
     *
     * @return a string representation of the collection of items
     *  
     */

    public static String serializeToString (ItemList xml)
	throws GalapiException
    {

	String s = nativeSerializeToString (xml.getNativeItemList());

	return s;
    }


    /**
     * 
     * Serializes an XQuery data model value to standard output
     *
     * @param xml a collection of items representing an XML value
     *
     */

    public static void serializeToOutputChannel (ItemList xml)
	throws GalapiException
    {

	nativeSerializeToOutputChannel (xml.getNativeItemList());

    }

    /**
     * 
     * Serializes an XQuery data model value to a file
     *
     * @param filename the name of the output file to be produced
     * @param xml a collection of items representing an XML value
     *
     */

    public static void serializeToFile (String filename, ItemList xml)
	throws GalapiException
    {

	nativeSerializeToFile (filename, xml.getNativeItemList());

    }


    /**********************/
    /* Evaluation context */
    /**********************/

    /**
     *
     * Returns the default {@link galapi.ProcessingContext ProcessingContext} 
     * which just contains flags for controlling debugging, printing, and 
     * the processing phases.
     *
     */

    public static ProcessingContext defaultProcessingContext()
	throws GalapiException
    {
	return defPC;
    }


    /**
     * Load the standard Galax library, which contains the built-in
     * types, namespaces, and functions.
     *
     */

    public static ModuleContext loadStandardLibrary (ProcessingContext pc)
	throws GalapiException
    {
	//int i = -1;
	
	int i = nativeLoadStandardLibrary (pc.getNativeItem());

	return new ModuleContext (i);
    }
    
    /**
     * Used to import other library modules.
     * 
     * @return the {@link galapi.ModuleContext ModuleContext} argument extended 
     * with the module in the string filename argument.
     *
     */

    public static ModuleContext importLibraryModule (ModuleContext mod_ctxt, String filename)
	throws GalapiException
    {
	//	int i = -1;

	int i = nativeImportLibraryModule (mod_ctxt.getNativeItem(), filename);
	    
	return new ModuleContext (i);
    }

    /**
     * Used to import a main module defined in a file.
     *
     * @return the {@link galapi.ModuleContext ModuleContext} argument 
     * extended with the main module in the string filename argument
     * and a list of statements to evaluate.
     * 
     * @see QueryContext
     *
     */

    public static QueryContext importMainModule (ModuleContext mod_ctxt, String filename)
	throws GalapiException
    {

	//	int i = -1;

	int i = nativeImportMainModule (mod_ctxt.getNativeItem(), filename);

	return new QueryContext (i);
    }
    
    /**
     * Same as {@link #importMainModule importMainModule}, but module 
     * definition is provided in a string instead of a file.
     *
     */

    public static QueryContext importMainModuleFromString (ModuleContext mod_ctxt, String str)
	throws GalapiException
    {

	//int i = -1;

	int i = importMainModuleFromString (mod_ctxt.getNativeItem(), str);

	return new QueryContext (i);
    }


    /**
     * Extends the given module context with the external context.
     * 
     * @param mod_ctxt a given {@link galapi.ModuleContext ModuleContext}
     * @param ext_ctxt an {@link galapi.ExternalContext ExternalContext} 
     * used for the extension of the context 
     * 
     * @return a new <code>ModuleContext</code> obtained by extending
     * the a given environment with an external environment.
     */

    public static ModuleContext addExternalContext (ModuleContext mod_ctxt,
						    ExternalContext ext_ctxt)
	throws GalapiException
    {
	
	//int i = -1;
	
	int i = nativeAddExternalContext (mod_ctxt.getNativeItem(), 
				      ext_ctxt.getNativeItem());
	    
	return new ModuleContext (i);
    }


    /**
     * Evaluates the expressions for all - possibly
     * mutually dependent - global variables. 
     * <p>It must be called before the other evalXXX functions.
     *
     * @param mod_ctxt a context obtained with one of the methods from
     * this class
     * 
     * @return a new context in which the expressions are evaluated
     */

    public static ModuleContext evalGlobalVariables (ModuleContext mod_ctxt)
	throws GalapiException
    {

	//int i = -1;

	int i = nativeEvalGlobalVariables (mod_ctxt.getNativeItem());

	return new ModuleContext (i);	
    }


    /********************/
    /* Query evaluation */
    /********************/

    /**
     * Given the <code>ModuleContext</code>, evaluates the XQuery
     *  statement  in  the string argument.  
     * 
     * @return if the statement is an XQuery expression,
     *  returns a list of items, otherwise if the statement is an
     *  XQuery update, returns an empty list because update statements have
     *  side effects on the data model store, but do not return values.
     */

    public static ItemList evalStatementFromString (ModuleContext c, String querystring)
	throws GalapiException
    {

	//int il = -1;

	int il = nativeEvalStatementFromString (c.getNativeItem(), querystring);

	return new ItemList (il);
    }

    /**
     * Given the <code>ModuleContext</code>, evaluates the function
     * whose name is given by <code>funname</code> applied to the
     * array of <code>ItemListitem<code> arguments.
     *
     * <p>NOTE: Each actual function argument is bound to one item list.
     *
     */

    public static ItemList evalQueryFunction (ModuleContext c, String funname,
					      ItemList params[])
	throws GalapiException
    {

	int p[] = new int[params.length];
	for (int i=0; i<p.length; i++)
	    p[i] = params[i].getNativeItemList();
	// int il = -1;

	int il = nativeEvalQueryFunction (c.getNativeItem(), 
					  funname,
					  p);

	return new ItemList (il);
    }


    /**
     * Given the {@link galapi.QueryContext QueryContext}, 
     * evaluates the statements containded in the query.
     * 
     * @return an XSquery sequence representing the result of this evaluation
     */

    public static ItemList evalStatements (QueryContext q)
	throws GalapiException
    {

	//int il = -1;
	int il = nativeEvalStatements (q.getNativeItem());

	return new ItemList (il);
    }


    /**
     * Bind the context item (the XPath "." expression) to the XML
     * document in the resource named by the first string argument,
     * and evaluate the XQuery statement in the second string argument. 
     *
     * @return the result of evaluating the query in the aforementioned context
     * 
     */
    public static  ItemList
	evalStatementWithContextItemFromDocument (String filename, String querystr)
	throws GalapiException
    {
	
	//int il = -1;

	int il = nativeEvalStatementWithContextItemFromDocument 
	    (filename, querystr);

	return new ItemList (il);
    }

    /**
     * Bind the context item (the XPath "." expression) to the XML value
     * in the item list argument and evaluate the XQuery statement in
     * the string argument. 
     *
     * @return the result of evaluating the query in the aforementioned context
     * 
     */

    public static  ItemList
	evalStatementWithContextItemFromXML (Item ctxt, String querystr)
	throws GalapiException
    {
	
	//int il = -1;

	int il = nativeEvalStatementWithContextItemFromXML 
	    (ctxt.getNativeItem(), querystr);

	return new ItemList (il);
    }

    /**
     * Bind the context item (the XPath "." expression) to the XML value
     * in the item list argument and evaluate the XQuery statement in
     * the file with name in the string argument. 
     *
     * @return the result of evaluating the query in the aforementioned context
     * 
     */

    public static  ItemList
	evalStatementFromFileWithContextItemFromXML (Item ctxt, String querystr)
	throws GalapiException
    {
	
	//int il = -1;

	int il = nativeEvalStatementFromFileWithContextItemFromXML 
	    (ctxt.getNativeItem(), querystr);

	return new ItemList (il);
    }

    /**
     * The first argument is a list of variable name and XML value pairs.
     * Bind each variable to the corresponding XML value and evaluate the
     * XQuery statement in the string argument.
     *
     * @return the result of evaluating the query in the aforementioned context
     * 
     */

    public static  ItemList
	evalStatementWithVariablesFromXML (NameValuePair vars[], String querystr)
	throws GalapiException
    {
	
	String varnames[] = new String[vars.length];
	int varvals[] = new int[vars.length];
	for (int i=0; i<vars.length; i++) {
	    varnames[i] = vars[i].getName();
	    varvals[i] = vars[i].getValue().getNativeItemList();
	}

	//int il = -1;

	int il = nativeEvalStatementWithVariablesFromXML 
	    (varnames, varvals, querystr);

	return new ItemList (il);
    }


    
    /*
     * native methods
     *
     */
    protected static native int nativeLoadDocument (String filename);
    protected static native int nativeLoadDocumentValidate (int mc, String filename);
    protected static native String nativeSerializeToString (int xml);
    protected static native void nativeSerializeToOutputChannel (int xml);
    protected static native void nativeSerializeToFile (String filename, int xml);

    protected static native int nativeDefaultProcessingContext ();
    protected static native int nativeLoadStandardLibrary (int pc);
    protected static native int nativeImportLibraryModule (int mod_ctxt, String filename);
    protected static native int nativeImportMainModule (int mod_ctxt, String filename);
    protected static native int importMainModuleFromString (int mod_ctxt, String str);

    protected static native int nativeAddExternalContext (int mod_ctxt, int ext_ctxt);
    protected static native int nativeEvalGlobalVariables (int mod_ctxt);

    protected static native int nativeEvalStatementFromString (int c, String querystring);
    protected static native int nativeEvalQueryFunction (int c, 
							 String funname,
							 int params[]);
    protected static native int nativeEvalStatements (int q);

    protected static native int nativeEvalStatementWithContextItemFromDocument 
	(String filename, String querystr);
    
    protected static native int nativeEvalStatementWithContextItemFromXML 
	(int it, String querystr);

    protected static native int nativeEvalStatementFromFileWithContextItemFromXML 
	(int it, String filename);

    protected static native int nativeEvalStatementWithVariablesFromXML 
	(String varnames[], int varvals[], String querystr);

}// class Galax

