(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: lexutil.ml,v 1.3 2003/06/27 00:05:07 simeon Exp $ *)

(* Module Lexutil
   Description:
     This module contains some basic functions used during lexing.
*)

open Error
open Xquery_ast


(****************************)
(* Lexing global parameters *)
(****************************)

(* Character encoding *)

type encoding_kind =
  | Utf8
  | Iso88591


(* Language currently parsed *)

type language_kind =
  | XML_1_0
  | XPath_2_0
  | XQuery_1_0


(* Lexical state *)

type lex_state =
      (* XQuery states *)
  | SchemaState
  | DefaultState
      (* XML states *)
  | OpeningTagState
  | ClosingTagState
  | TextState
  | AttributeTextDoubleState      (* Note: There are three kinds of attribute value lexing. - Jerome *)
  | AttributeTextSingleState
  | EntityIncludedInLiteralState
  | PIState
  | CDATAState
  | CommentState
      (* Top-level document state *)
  | XMLDeclState
  | XMLDocumentState
      (* DTD states *)
  | DocTypeState
  | DTDDeclState
  | EntityValueState

(********************)
(* Comment handling *)
(********************)

(* Used to store the comment level *)

let depth = ref 0


(****************************)
(* String and text handling *)
(****************************)

(* Used to store string or text content *)

let initial_string_buffer () = String.create 256
let string_buffer = ref (initial_string_buffer ())
let string_end = ref 0

(* Manipulation of the string buffer *)

let reset_string () =
  string_buffer := initial_string_buffer ();
  string_end := 0

let add_char_to_string ch =
  if !string_end >= String.length (!string_buffer)
  then
    begin
      let newBuffer = String.create (String.length (!string_buffer) * 2) in
      String.blit (!string_buffer) 0 newBuffer 0 (String.length (!string_buffer));
      string_buffer := newBuffer
    end;
  String.unsafe_set (!string_buffer) (!string_end) ch;
  incr string_end

let add_string_to_string str =
  String.iter add_char_to_string str

let get_string () =
  let s = String.sub (!string_buffer) 0 (!string_end) in
  string_buffer := initial_string_buffer ();
  s


(**************************)
(* Lexical state handling *)
(**************************)

let current_lex_state =
  Stack.create ()

let get_current_lex_state () =
  Stack.top current_lex_state

let push_schema () =
  Stack.push SchemaState current_lex_state

let push_default () =
  reset_string ();
  Stack.push DefaultState current_lex_state

let push_text () =
  reset_string ();
  Stack.push TextState current_lex_state

let push_attribute_text_single () =
  reset_string ();
  Stack.push AttributeTextSingleState current_lex_state

let push_attribute_text_double () =
  reset_string ();
  Stack.push AttributeTextDoubleState current_lex_state

let push_entity_included_in_literal () =
  reset_string ();
  Stack.push EntityIncludedInLiteralState current_lex_state

let push_entity_included () =
  reset_string ();
  Stack.push TextState current_lex_state

let push_entity_value () =
  reset_string ();
  Stack.push EntityValueState current_lex_state

let push_opening_tag () =
  Stack.push OpeningTagState current_lex_state

let push_closing_tag () =
  Stack.push ClosingTagState current_lex_state

let push_processing_instruction () =
  reset_string ();
  Stack.push PIState current_lex_state

let push_cdata () =
  Stack.push CDATAState current_lex_state

let push_comment () =
  reset_string ();
  Stack.push CommentState current_lex_state

let push_xmldecl () =
  reset_string ();
  Stack.push XMLDeclState current_lex_state

let push_xmldocument () =
  reset_string ();
  Stack.push XMLDocumentState current_lex_state

let push_doctype () =
  reset_string ();
  Stack.push DocTypeState current_lex_state

let push_dtddecl () =
  reset_string ();
  Stack.push DTDDeclState current_lex_state

let pop_state () =
  reset_string ();
  ignore(Stack.pop current_lex_state)

let pop_state_keep_buffer () =
  ignore(Stack.pop current_lex_state)

let current_parsing_kind =
  ref XQuery_1_0

let get_parsing_kind () =
  !current_parsing_kind

type attribute_quote_kind =
  | SingleQuoteKind
  | DoubleQuoteKind
  | EntityIncludedInLiteralKind

let attribute_quote_kind () =
  match get_current_lex_state () with
  | AttributeTextSingleState -> SingleQuoteKind
  | AttributeTextDoubleState -> DoubleQuoteKind
  | EntityIncludedInLiteralState -> EntityIncludedInLiteralKind
  | _ ->
      raise (Query (Internal_Error "Requiring quote kind outside of an attribute"))

let init_xquery_lexing () =
  Stack.clear current_lex_state;
  current_parsing_kind := XQuery_1_0;
  reset_string ();
  Stack.push DefaultState current_lex_state

let init_xml_lexing () =
  Stack.clear current_lex_state;
  current_parsing_kind := XML_1_0;
  reset_string ();
  Stack.push XMLDocumentState current_lex_state


(********************)
(* Token processing *)
(********************)

(* Extracts an axis name *)

let resolve_axis =
  function
    | "attribute" -> Attribute
    | "child" -> Child
    | "descendant" -> Descendant
    | "descendant-or-self" -> Descendant_or_self
    | "parent" -> Parent
    | "self" -> Self
    | "ancestor" -> Ancestor
    | "ancestor-or-self" -> Ancestor_or_self
    | s -> raise (Query (Parsing (Finfo.parsing_locinfo (),"Unkown XQuery axis " ^ s)))

(* The following axis are not in XQuery - Jerome
    | "namespace" -> Namespace
    | "following" -> Following
    | "following-sibling" -> Following_sibling
    | "preceding" -> Preceding
    | "preceding-sibling" -> Preceding_sibling
*)

let get_axis var_s =
  let mid =
    try
      String.index var_s ' '
    with
      Not_found ->
	(String.index var_s ':') in
  let sta = 0 in
  let len = mid in
  resolve_axis (String.sub var_s sta len)

(* Extracts the type name in a type definition *)

let get_type_defn var_s =
  let mid1 = String.rindex var_s ' ' in
  let var_s' = Whitespace.remove_whitespace (String.sub var_s 0 mid1) in
  let sta = (String.rindex var_s' ' ') + 1 in
  let len = (String.length var_s') - sta in
  Namespace.uqname_element_of_string (String.sub var_s' sta len)

(* Extracts the element name in an element declaration *)

let get_element_defn var_s =
  let mid1 = String.rindex var_s ' ' in
  let var_s' = Whitespace.remove_whitespace (String.sub var_s 0 mid1) in
  let sta = (String.rindex var_s' ' ') + 1 in
  let len = (String.length var_s') - sta in
  Namespace.uqname_element_of_string (String.sub var_s' sta len)

(* Extracts the attribute name in an attribute declaration *)

let get_attribute_defn var_s =
  let mid1 = String.rindex var_s ' ' in
  let var_s' = Whitespace.remove_whitespace (String.sub var_s 0 mid1) in
  let sta = (String.rindex var_s' ' ')+1 in
  let len = (String.length var_s') - sta in
  Namespace.uqname_element_of_string (String.sub var_s' sta len)

(* Extracts escaped name *)

let get_escaped_name var_s =
  let len = (String.length var_s) - 1 in
  (String.sub var_s 1 len)

(* Extracts the function name in a function definition *)

let get_function_name var_s =
  let mid =
    try
      String.index var_s ' '
    with
      Not_found ->
	(String.index var_s '(') in
  let sta = 0 in
  let len = mid in
  Namespace.uqname_function_of_string (String.sub var_s sta len)

(* Extracts the PI target *)

let get_target_pi var_s =
  let mid = String.rindex var_s '?' in
  let mid2 =
    try
      String.index var_s ' '
    with
    | _ -> String.length var_s
  in
  let sta = mid+1 in
  let len = mid2 - sta in
  let pi_target = (String.sub var_s sta len) in
  if (String.lowercase pi_target = "xml")
  then
    raise (Query (Parsing (Finfo.parsing_locinfo (),"Process instruction target cannot match ['X' 'x']['M' 'm']['L' 'l']")))
  else
    pi_target

(* Extracts the character encoding *)

let get_xml_encoding_quotes var_s =
  let mid = String.index var_s '"' in
  let sta = mid+1 in
  let len = (String.length var_s) - sta - 1 in
  String.sub var_s sta len

let get_xml_encoding_single_quotes var_s =
  let mid = String.index var_s '\'' in
  let sta = mid+1 in
  let len = (String.length var_s) - sta - 1 in
  String.sub var_s sta len

(* Extracts character and entity references *)

let get_char_ref       l =
  let len = String.length l in
  String.sub l 2 (len-3)

let get_hexchar_ref    l =
  let len = String.length l in
  String.sub l 3 (len-4)

let get_entity_ref l =
  let len = String.length l in
  String.sub l 1 (len-2)

