(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: type_lexer.mll,v 1.1 2003/05/10 18:49:58 simeon Exp $ *)

(* Module Type_lexer
   Description:
     This module contains the lexer for type declarations.
*)

{

open Lexing

open Xquery_ast
open Finfo
open Error
open Xquery_parser   (* tokens definition *)
open Lexutil         (* lexical state handling *)

}

(*****************************)
(* Lexer regular expressions *)
(*****************************)

(* Whitespace *)

let space = ['\032' '\009' '\013' '\010']
let whitespace = space*
let whitespaceplus = space+

(* XML character sets *)

(* Note:
     Some of these Char codes have been taken from the PXP ISO
     parser. I am not sure that they are doing the right thing and
     should be double checked for conformance.
   - Jerome
 *)

let base_char    = ['A'-'Z' 'a'-'z']
let ideographics = ['\192'-'\214' '\216'-'\246' '\248'-'\255']
let extender     = '\183'
let digit        = ['0'-'9']

(* XML names *)

let letter   = base_char (* | ideographics *)
let nmstart  = letter | '_'
let nmchar   = letter | extender | digit | '.' | '-' | '_'
 
let ncname   = nmstart nmchar*

let qname    = ncname? (':' ncname)?

(****************)
(* Lexing rules *)
(****************)

rule schema = parse
  | whitespace                  (* for all others, skip blanks *)
      { schema lexbuf }
  | "{--" 
      { depth := 1; xquery_comment lexbuf; schema lexbuf }
  | "#"
      { xquery_comment_line lexbuf; schema lexbuf }
  | "{" { push_schema (); LCURLY }
  | "}" { pop_state (); RCURLY }
      (* out of a global variable *)
  | ":="
      { pop_state (); COLONEQUALS }
  | "document"
      { DOCUMENT }
  | "element"
      { ELEMENT }
  | "attribute"
      { ATTRIBUTE }
  | "type"
      { TYPE }
  | "none"
      { NONE }
  | "text"
      { TEXTNODE }
  | ":" ncname
      { QNAME (Namespace.NSDefaultElementPrefix, get_escaped_name (lexeme lexbuf)) }
  | qname
      { let var_s = (lexeme lexbuf) in
        QNAME (Namespace.uqname_element_of_string var_s) }
  | "|"
      { BAR }
  | "*"
      { STAR }
  | "+"
      { PLUS }
  | ","
      { COMMA }
  | "&"
      { AMPERSAND }
  | "?"
      { QUESTION }
(* Parenthesis *)
  | "("
      { LPAR }
  | ")" 
      { RPAR }
(* Unknown caracter *)
  | _
      { raise (Query (Lexing(lexing_locinfo lexbuf, "Illegal character in schema declaration"))) }

and xquery_comment = parse
  | "{--"
      { depth := succ !depth; xquery_comment lexbuf }
  | "--}"
      { depth := pred !depth; if !depth > 0 then xquery_comment lexbuf }
  | eof
      { raise (Query (Lexing(lexing_locinfo lexbuf, "Comment not terminated"))) }
  | [^ '\n']
      { xquery_comment lexbuf }
  | "\n"
      { xquery_comment lexbuf }

and xquery_comment_line = parse
  | eof
      { () }
  | "\n"
      { () }
  | [^ '\n']
      { xquery_comment_line lexbuf }

