(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: xquery_lexer.mll,v 1.3 2003/06/09 15:28:26 simeon Exp $ *)

(* Module XQuery_lexer
   Description:
     This module contains the main XQuery lexer.
*)

{

open Lexing

open Xquery_ast
open Finfo
open Error
open Xquery_parser   (* tokens definition *)
open Lexutil         (* lexical state handling *)

}

(*****************************)
(* Lexer regular expressions *)
(*****************************)

(* Whitespace *)

let space = ['\032' '\009' '\013' '\010']
let whitespace = space*
let whitespaceplus = space+

(* XML character sets *)

(* Note:
     Some of these Char codes have been taken from the PXP ISO
     parser. I am not sure that they are doing the right thing and
     should be double checked for conformance.
   - Jerome
 *)

let base_char    = ['A'-'Z' 'a'-'z']
let ideographics = ['\192'-'\214' '\216'-'\246' '\248'-'\255']
let extender     = '\183'
let digit        = ['0'-'9']

(* XML names *)

let letter   = base_char (* | ideographics *)
let nmstart  = letter | '_'
let nmchar   = letter | extender | digit | '.' | '-' | '_'
 
let ncname   = nmstart nmchar*

let qname    = ncname (':' ncname)?

(* Numerics *)

let sign            = '-' | '+'
let exponent        = 'e' | 'E'

let integer_numeric = digit+
let decimal_numeric = (('.' digit+) | (digit+ '.' (digit+)?))
let double_numeric  = (('.' digit+) | (digit+ ('.' (digit+)?)?)) exponent sign? digit+

(****************)
(* Lexing rules *)
(****************)

rule token = parse

(* Arithmetic operators *)
  | "-"
      { MINUS }
  | "+"
      { PLUS }
  | whitespaceplus "*"
      { MULT }

(* Whitespace *)
  | space                  (* for all others, skip blanks *)
      { token lexbuf }

(* Comments *)
  | "--}"
      { raise (Query (Lexing(lexing_locinfo lexbuf, "Unmatched end of comment"))) }
  | "{--" 
      { depth := 1; xquery_comment lexbuf; token lexbuf }
  | "#"
      { xquery_comment_line lexbuf; token lexbuf }

(* Reserved words *)
  | "or"
      { OR }
  | "and"
      { AND }
  | "div"
      { DIV }
  | "idiv"
      { IDIV }
  | "mod"
      { MOD }
  | "*"
      { STAR }
  | "in"
      { IN }
  | "satisfies"
      { SATISFIES }
  | "return"
      { RETURN }
  | "then"
      { THEN }
  | "end"
      { END }
  | "else"
      { ELSE }
  | "to"
      { TO }
  | "where"
      { WHERE }
  | "intersect"
      { INTERSECT }
  | "union"
      { UNION }
  | "except"
      { EXCEPT }
  | "<<"
      { PRECEDES }
  | ">>"
      { FOLLOWS }
  | "case"
      { CASE }
  | "stable"
      { STABLE }
  | "ascending"
      { ASCENDING }
  | "descending"
      { DESCENDING }
  | "empty" whitespaceplus "greatest"
      { EMPTYGREATEST }
  | "empty" whitespaceplus "least"
      { EMPTYLEAST }
  | "default"
      { DEFAULT }

(* Composed reserved words *)
  | "instance" whitespaceplus "of"
      { INSTANCEOF }
  | "assert" whitespaceplus "as"
      { ASSERTAS }
  | "cast" whitespaceplus "as"
      { CASTAS }
  | "castable" whitespaceplus "as"
      { CASTABLEAS }
  | "treat" whitespaceplus "as"
      { TREATAS }
  | "validate" whitespaceplus "as"
      { VALIDATEAS }

(* Axis names *)
  | ncname "::"  (* Should normally allow for whitespace,
		    but there seems to be a bug in OCaml's lexer - JS *)
      { let var_s = (lexeme lexbuf) in
        AXIS (get_axis var_s) }

(* Declaration keywords *)

  | "declare" whitespaceplus "namespace"
      { DECLARENAMESPACE }

  | "define" whitespaceplus "key"
      { DEFINEKEY }

  | "namespace"
      { NAMESPACE }

  | "default" whitespaceplus "element"
      { DEFAULTELEMENT }

  | "default" whitespaceplus "function"
      { DEFAULTFUNCTION }

  | "define" whitespaceplus "type" whitespaceplus qname whitespaceplus "{"
      { let var_s = (lexeme lexbuf) in
        let sn = (get_type_defn var_s) in
        push_schema ();
        DEFINETYPE (sn) }

  | "define" whitespaceplus "element" whitespaceplus qname whitespaceplus "{"
      { let var_s = (lexeme lexbuf) in
        let sn = (get_element_defn var_s) in
        push_schema ();
        DEFINEELEMENT (sn) }

  | "define" whitespaceplus "attribute" whitespaceplus qname whitespaceplus "{"
      { let var_s = (lexeme lexbuf) in
        let sn = (get_attribute_defn var_s) in
        push_schema ();
        DEFINEATTRIBUTE (sn) }

  | "define" whitespaceplus "function"
      { DEFINEFUNCTION }

  (* Added by Byron. The XQuery grammar does not support function signature declaration *)
  (* Note: The below rule is subjected to change. *)
  | "define" whitespaceplus "function" whitespaceplus "signature"
      { DEFINESIGNATURE }

  | "define" whitespaceplus "global" whitespaceplus "$"
      { DEFINEGLOBAL }

  | "import" whitespaceplus "schema"
      { IMPORTSCHEMA }

  | "element"
      { ELEMENT }

  | "element" whitespaceplus "of" whitespaceplus "type"
      { ELEMENTOF }

  | "attribute"
      { ATTRIBUTE }

  | "document" whitespaceplus "of" whitespaceplus "type"
      { DOCUMENTOF }

  | "type"
      { TYPE }

  | "node"
      { NODE }

  | "item"
      { ITEM }

  | "atomic"
      { ATOMIC }

  | "numeric"
      { NUMERIC }

  | "document" 
      { DOCUMENT }

  | "text"
      { TEXTNODE }

  | "empty"
      { EMPTY }

  | "as"
      { AS }

  | "at"
      { AT }

(* Keywords followed by variables *)

  | "some"
      { SOME }

  | "every"
      { EVERY }

  | "for"
      { FOR }

  | "let"
      { LET }

(* Keywords followed by a parenthesis *)

  | "comment" whitespace "("
      { COMMENTLPAR }

  | "text" whitespace "("
      { TEXTLPAR }

  | "node" whitespace "("
      { NODELPAR }

  | "processing-instruction" whitespace "("
      { PROCESSINGINSTRUCTIONLPAR }

  | "if"
      { IF }

  | "typeswitch"
      { TYPESWITCH }

  | "sort" whitespaceplus "by"
      { SORTBY }

  | "stable" whitespace "sort" whitespaceplus "by"
      { STABLESORTBY }

  | "order" whitespaceplus "by"
      { ORDERBY }

(* Update tokens *)

  | "insert"
      { INSERT }
  | "replace"
      { REPLACE }
  | "delete"
      { DELETE }
  | "as" whitespaceplus "last"
      { ASLAST }
  | "value" whitespaceplus "of"
      { VALUEOF }
  | "as" whitespaceplus "first"
      { ASFIRST }
  | "into"
      { INTO }
  | "after"
      { AFTER }
  | "before"
      { BEFORE }
  | "update"
      { UPDATE }
  | "with"
      { WITH }

(* Type thingies *)
  | ":"
      { COLON }
  | "<:"
      { LTCOLON }
  | "/\\"
      { BACKSLASHSLASH }

(* Slash and SlashSlash *)
  | "/"
      { SLASH }
  | "//"
      { SLASHSLASH }

(* Equals and EqualsEquals *)
  | "eq"
      { EQ }
  | "="
      { EQUALS }

(* NotEquals and NotEqualsEquals *)
  | "ne"
      { NE }
  | "!="
      { NOTEQUALS }

(* Comparators *)
  | "<" whitespaceplus
      { LT }
  | ">"
      { GT }

  | "lt" 
      { LTOP }
  | "gt"
      { GTOP }

  | "le"
      { LTE }
  | "ge"
      { GTE }

  | "<="
      { LTEQUALS }
  | ">="
      { GTEQUALS }

  | "is"
      { IS }

  | "isnot"
      { ISNOT }

(* Assignment *)
  | ":="
      { COLONEQUALS }

(* XML tags *)
  | "<?" ncname             (* Processing instruction *)
      { let var_s = (lexeme lexbuf) in
        push_processing_instruction () ; OPENINGPI (get_target_pi var_s) }
  | "<!--"                  (* Opening comment *)
      { push_comment () ; LOPENINGCOMMENT }

  | "<"                     (* Opening XML tag *)
      { push_opening_tag () ; LOPENINGTAG }

(* Choice *)
  | "|"
      { BAR }

(* Parenthesis *)
  | "("
      { LPAR }
  | ")" 
      { RPAR }
  | "[" 
      { LBRACK }
  | "]" 
      { RBRACK }
  | "{"
      { push_default ();
	LCURLY }
  | "}"
      { try
	  pop_state (); RCURLY
        with
          Stack.Empty ->
	    raise (Query (Lexing(lexing_locinfo lexbuf, "Unmatched end of enclosed expression"))) }

(* Misc symbols *)
  | "@"
      { ATSIGN }
  | ","
      { COMMA }
  | "?"
      { QUESTION }
  | ";"
      { SEMICOLON }
  | "."
      { DOT }
  | ".."
      { DOTDOT }

(* Variables *)
  | "$"
      { DOLLAR }

(* Function calls *)
  | qname "("   (* Should normally allow for whitespace,
		   but there seems to be a bug in OCaml's lexer - JS *)
      { let var_s = (lexeme lexbuf) in
        FUNCTIONNAMELPAR (get_function_name var_s) }

(* XML Names *)
  | ncname
      { NCNAME (lexeme lexbuf) }
  | ":" ncname
      { NCNAME (get_escaped_name (lexeme lexbuf)) }
  | qname
      { let var_s = (lexeme lexbuf) in
        QNAME (Namespace.uqname_element_of_string var_s) }

(********************)
(* Numeric literals *)
(********************)

(* Integers *)
  | integer_numeric
      { let v = Decimal._integer_of_string (lexeme lexbuf) in
        (INT v) }

(* Decimals *)
  | decimal_numeric
      { let v = Decimal._decimal_of_string (lexeme lexbuf) in
        (DECIMAL v) }

(* Double *)
  | double_numeric
      { let v = float_of_string (lexeme lexbuf) in
        (DOUBLE v) }

(*******************)
(* String literals *)
(*******************)

  | "\""
      { reset_string(); string lexbuf }
  | "'"
      { reset_string(); string_single lexbuf }

(***************)
(* End of file *)
(***************)

  | eof
      { EOF }

(********************)
(* Unknown caracter *)
(********************)

(* Note:
     If the lexer arrives here, this is an error.
   - Jerome
*)

  | _
      { raise (Query (Lexing(lexing_locinfo lexbuf, ("Illegal character '" ^ (lexeme lexbuf) ^ "' in expression")))) }

and xquery_comment = parse
  | "{--"
      { depth := succ !depth; xquery_comment lexbuf }
  | "--}"
      { depth := pred !depth; if !depth > 0 then xquery_comment lexbuf }
  | eof
      { raise (Query (Lexing(lexing_locinfo lexbuf, "Comment not terminated"))) }
  | [^ '\n']
      { xquery_comment lexbuf }
  | "\n"
      { xquery_comment lexbuf }

and xquery_comment_line = parse
  | eof
      { () }
  | "\n"
      { () }
  | [^ '\n']
      { xquery_comment_line lexbuf }

(****************)
(* String lexer *)
(****************)

and string = parse
  | "\"\""  { add_char_to_string '"'; string lexbuf }          (* Escaped quote *)
  | '"'     { STRING (get_string()) }                  (* End of string *)
  | eof     { raise (Query (Lexing(lexing_locinfo lexbuf, "String not terminated"))) }
  | _       { add_char_to_string (lexeme_char lexbuf 0); string lexbuf }

and string_single = parse
  | "''"  { add_char_to_string '\''; string_single lexbuf }    (* Escaped ' *)
  | '\''  { STRING (get_string()) }                    (* End of string *)
  | eof   { raise (Query (Lexing(lexing_locinfo lexbuf, "String not terminated"))) }
  | _     { add_char_to_string (lexeme_char lexbuf 0); string_single lexbuf }

