(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                           XML Query Engine                          *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: monitor.ml,v 1.7 2002/11/15 05:29:23 simeon Exp $ *)

open Format
open Xquery_astutil

(* This module is used for monitoring the activity of Galax. *)

(* Type that represents various processing phases within the system *)

type phase =
  | Query_Toplevel_Phase
  | Query_Processing_Context
  | Query_Processing_Expr of Xquery_ast.expr
  | Query_Parsing_Phase
  | Query_Normalization_Phase
  | Query_Static_Typing_Phase
  | Query_Optimization_Phase
  | Query_Path_Analysis_Phase
  | Query_Evaluation_Phase
  | Query_Serialization_Phase
  | Document_Parsing_Phase of AnyURI._uri
  | Document_Loading_Phase of AnyURI._uri
  | Document_ParsingLoading_Phase of AnyURI._uri
  | Expression_Phase of Xquery_astutil.cexpr_kind


(*************************)
(* Data model statistics *)
(*************************)

type dm_stats =
    { mutable node_count : int }

let current_stats =
  { node_count = 0 }

(************************)
(* Statistics functions *)
(************************)

type stats_item =
    { glx_total_memory : float;
      glx_live_memory : float;
      glx_time : float }

type stats_point =
    { relative_stats : stats_item;
      absolute_stats : stats_item }

let empty_stats_point =
  { relative_stats =
    { glx_total_memory = 0.0;
      glx_live_memory = 0.0;
      glx_time = 0.0 };
    absolute_stats =
    { glx_total_memory = 0.0;
      glx_live_memory = 0.0;
      glx_time = 0.0 } }

let get_glx_time () =
  let times = Unix.times () in
  times.Unix.tms_utime +. times.Unix.tms_stime

let latest_stats_point =
  ref empty_stats_point

let new_stats_point () =
  let st =
    Gc.compact();
    Gc.stat () in
  let glx_absolute_time = get_glx_time () in
  let glx_absolute_total_memory = float_of_int (st.Gc.heap_words) *. float_of_int (Sys.word_size / 8) in
  let glx_absolute_live_memory = float_of_int (st.Gc.live_words) *. float_of_int (Sys.word_size / 8) in

  let glx_relative_time = glx_absolute_time -. !latest_stats_point.absolute_stats.glx_time in
  let glx_relative_total_memory = glx_absolute_total_memory -. !latest_stats_point.absolute_stats.glx_total_memory in
  let glx_relative_live_memory = glx_absolute_live_memory -. !latest_stats_point.absolute_stats.glx_live_memory in
  let new_stat_point =
    { relative_stats =
	{ glx_total_memory = glx_relative_total_memory;
	  glx_live_memory = glx_relative_live_memory;
	  glx_time = glx_relative_time };
	absolute_stats =
	{ glx_total_memory = glx_absolute_total_memory;
	  glx_live_memory = glx_absolute_live_memory;
	  glx_time = glx_absolute_time } }
  in
  begin
    latest_stats_point := new_stat_point;
    new_stat_point
  end

(*
type expr_stat_summary =
    {
   }
*)

type phase_stat_summary =
    { mutable query_toplevel_phase : stats_point;
      mutable query_processing_context : stats_point;
      mutable query_processing_expr : stats_point;
      mutable query_parsing_phase : stats_point;
      mutable query_normalization_phase : stats_point;
      mutable query_static_typing_phase : stats_point;
      mutable query_optimization_phase : stats_point;
      mutable query_path_analysis_phase : stats_point;
      mutable query_evaluation_phase : stats_point;
      mutable query_serialization_phase : stats_point;
      mutable document_parsing_phase : (AnyURI._uri, stats_point) Hashtbl.t;
      mutable document_loading_phase : (AnyURI._uri, stats_point) Hashtbl.t;
      mutable document_parsingloading_phase : (AnyURI._uri, stats_point) Hashtbl.t;
      mutable expression_phase : stats_point }

let empty_summary () =
  { query_toplevel_phase = empty_stats_point;
    query_processing_context = empty_stats_point;
    query_processing_expr = empty_stats_point;
    query_parsing_phase = empty_stats_point;
    query_normalization_phase = empty_stats_point;
    query_static_typing_phase = empty_stats_point;
    query_optimization_phase = empty_stats_point;
    query_path_analysis_phase = empty_stats_point;
    query_evaluation_phase = empty_stats_point;
    query_serialization_phase = empty_stats_point;
    document_parsing_phase = Hashtbl.create 17;
    document_loading_phase = Hashtbl.create 17;
    document_parsingloading_phase = Hashtbl.create 17;
    expression_phase  = empty_stats_point }

let current_summary =
  ref (empty_summary ())

let add_relative_stats_to_stats_point stats_point stat_item =
  { relative_stats =
    { glx_total_memory = stats_point.relative_stats.glx_total_memory +. stat_item.glx_total_memory;
      glx_live_memory = stats_point.relative_stats.glx_live_memory +. stat_item.glx_live_memory;
      glx_time = stats_point.relative_stats.glx_time +. stat_item.glx_time };
    absolute_stats =
    { glx_total_memory = stats_point.absolute_stats.glx_total_memory;
      glx_live_memory = stats_point.absolute_stats.glx_live_memory;
      glx_time = stats_point.absolute_stats.glx_time } }

let set_absolute_stats_in_stats_point stats_point stat_item =
  { relative_stats =
    { glx_total_memory = stats_point.relative_stats.glx_total_memory;
      glx_live_memory = stats_point.relative_stats.glx_live_memory;
      glx_time = stats_point.relative_stats.glx_time };
    absolute_stats =
    { glx_total_memory = stat_item.glx_total_memory;
      glx_live_memory = stat_item.glx_live_memory;
      glx_time = stat_item.glx_time } }

let add_relative_stats_to_summary phase stat_item =
  match phase with
  | Query_Toplevel_Phase ->
      !current_summary.query_toplevel_phase <-
	add_relative_stats_to_stats_point !current_summary.query_toplevel_phase stat_item;
  | Query_Processing_Context ->
      !current_summary.query_processing_context <-
	add_relative_stats_to_stats_point !current_summary.query_processing_context stat_item;
  | Query_Processing_Expr expr ->
      !current_summary.query_processing_expr <-
	add_relative_stats_to_stats_point !current_summary.query_processing_expr stat_item;
  | Query_Parsing_Phase ->
      !current_summary.query_parsing_phase <-
	add_relative_stats_to_stats_point !current_summary.query_parsing_phase stat_item;
  | Query_Normalization_Phase ->
      !current_summary.query_normalization_phase <-
	add_relative_stats_to_stats_point !current_summary.query_normalization_phase stat_item;
  | Query_Static_Typing_Phase ->
      !current_summary.query_static_typing_phase <-
	add_relative_stats_to_stats_point !current_summary.query_static_typing_phase stat_item;
  | Query_Optimization_Phase ->
      !current_summary.query_optimization_phase <-
	add_relative_stats_to_stats_point !current_summary.query_optimization_phase stat_item;
  | Query_Path_Analysis_Phase ->
      !current_summary.query_path_analysis_phase <-
	add_relative_stats_to_stats_point !current_summary.query_path_analysis_phase stat_item;
  | Query_Evaluation_Phase ->
      !current_summary.query_evaluation_phase <-
	add_relative_stats_to_stats_point !current_summary.query_evaluation_phase stat_item;
  | Query_Serialization_Phase ->
      !current_summary.query_serialization_phase <-
	add_relative_stats_to_stats_point !current_summary.query_serialization_phase stat_item;
  | Document_Parsing_Phase uri ->
      let doc_table = !current_summary.document_parsing_phase in
      if not(Hashtbl.mem doc_table uri)
      then
	Hashtbl.add doc_table uri empty_stats_point;
      let previous_stats = Hashtbl.find doc_table uri in
      begin
	let new_stats = add_relative_stats_to_stats_point previous_stats stat_item in
	Hashtbl.replace doc_table uri new_stats
      end
  | Document_Loading_Phase uri ->
      let doc_table = !current_summary.document_loading_phase in
      if not(Hashtbl.mem doc_table uri)
      then
	Hashtbl.add doc_table uri empty_stats_point;
      let previous_stats = Hashtbl.find doc_table uri in
      begin
	let new_stats = add_relative_stats_to_stats_point previous_stats stat_item in
	Hashtbl.replace doc_table uri new_stats
      end
  | Document_ParsingLoading_Phase uri ->
      let doc_table = !current_summary.document_parsingloading_phase in
      if not(Hashtbl.mem doc_table uri)
      then
	Hashtbl.add doc_table uri empty_stats_point;
      let previous_stats = Hashtbl.find doc_table uri in
      begin
	let new_stats = add_relative_stats_to_stats_point previous_stats stat_item in
	Hashtbl.replace doc_table uri new_stats
      end
  | Expression_Phase _ ->
      ()

let set_absolute_stats_in_summary phase stat_item =
  match phase with
  | Query_Toplevel_Phase ->
      !current_summary.query_toplevel_phase <-
	set_absolute_stats_in_stats_point !current_summary.query_toplevel_phase stat_item;
  | Query_Processing_Context ->
      !current_summary.query_processing_context <-
	set_absolute_stats_in_stats_point !current_summary.query_processing_context stat_item;
  | Query_Processing_Expr expr ->
      !current_summary.query_processing_expr <-
	set_absolute_stats_in_stats_point !current_summary.query_processing_expr stat_item;
  | Query_Parsing_Phase ->
      !current_summary.query_parsing_phase <-
	set_absolute_stats_in_stats_point !current_summary.query_parsing_phase stat_item;
  | Query_Normalization_Phase ->
      !current_summary.query_normalization_phase <-
	set_absolute_stats_in_stats_point !current_summary.query_normalization_phase stat_item;
  | Query_Static_Typing_Phase ->
      !current_summary.query_static_typing_phase <-
	set_absolute_stats_in_stats_point !current_summary.query_static_typing_phase stat_item;
  | Query_Optimization_Phase ->
      !current_summary.query_optimization_phase <-
	set_absolute_stats_in_stats_point !current_summary.query_optimization_phase stat_item;
  | Query_Path_Analysis_Phase ->
      !current_summary.query_path_analysis_phase <-
	set_absolute_stats_in_stats_point !current_summary.query_path_analysis_phase stat_item;
  | Query_Evaluation_Phase ->
      !current_summary.query_evaluation_phase <-
	set_absolute_stats_in_stats_point !current_summary.query_evaluation_phase stat_item;
  | Query_Serialization_Phase ->
      !current_summary.query_serialization_phase <-
	set_absolute_stats_in_stats_point !current_summary.query_serialization_phase stat_item;
  | Document_Parsing_Phase uri ->
      let doc_table = !current_summary.document_parsing_phase in
      if not(Hashtbl.mem doc_table uri)
      then
	Hashtbl.add doc_table uri empty_stats_point;
      let previous_stats = Hashtbl.find doc_table uri in
      begin
	let new_stats = set_absolute_stats_in_stats_point previous_stats stat_item in
	Hashtbl.replace doc_table uri new_stats
      end
  | Document_Loading_Phase uri ->
      let doc_table = !current_summary.document_loading_phase in
      if not(Hashtbl.mem doc_table uri)
      then
	Hashtbl.add doc_table uri empty_stats_point;
      let previous_stats = Hashtbl.find doc_table uri in
      begin
	let new_stats = set_absolute_stats_in_stats_point previous_stats stat_item in
	Hashtbl.replace doc_table uri new_stats
      end
  | Document_ParsingLoading_Phase uri ->
      let doc_table = !current_summary.document_parsingloading_phase in
      if not(Hashtbl.mem doc_table uri)
      then
	Hashtbl.add doc_table uri empty_stats_point;
      let previous_stats = Hashtbl.find doc_table uri in
      begin
	let new_stats = set_absolute_stats_in_stats_point previous_stats stat_item in
	Hashtbl.replace doc_table uri new_stats
      end
  | Expression_Phase _ ->
      ()

let get_phase_stats phase =
  match phase with
  | Query_Toplevel_Phase ->
      Some !current_summary.query_toplevel_phase
  | Query_Processing_Context ->
      Some !current_summary.query_processing_context
  | Query_Processing_Expr expr ->
      Some !current_summary.query_processing_expr
  | Query_Parsing_Phase ->
      Some !current_summary.query_parsing_phase
  | Query_Normalization_Phase ->
      Some !current_summary.query_normalization_phase
  | Query_Static_Typing_Phase ->
      Some !current_summary.query_static_typing_phase
  | Query_Optimization_Phase ->
      Some !current_summary.query_optimization_phase
  | Query_Path_Analysis_Phase ->
      Some !current_summary.query_path_analysis_phase
  | Query_Evaluation_Phase ->
      Some !current_summary.query_evaluation_phase
  | Query_Serialization_Phase ->
      Some !current_summary.query_serialization_phase
  | Document_Parsing_Phase uri ->
      begin
	let doc_table = !current_summary.document_parsing_phase in
	try
	  Some (Hashtbl.find doc_table uri)
	with
	| Not_found -> None
      end
  | Document_Loading_Phase uri ->
      begin
	let doc_table = !current_summary.document_loading_phase in
	try
	  Some (Hashtbl.find doc_table uri)
	with
	| Not_found -> None
      end
  | Document_ParsingLoading_Phase uri ->
      begin
	let doc_table = !current_summary.document_parsingloading_phase in
	try
	  Some (Hashtbl.find doc_table uri)
	with
	| Not_found -> None
      end
  | Expression_Phase _ ->
      None

let get_marshal_stats uri =
  let name =  uri ^ ".dm.marshal" in
  try
    let ic = open_in_bin name in
    let size = in_channel_length ic in
    begin
      close_in ic;
      Sys.remove name;
      Some (float_of_int size)
    end
  with
  | _ -> None

(**********************)
(* Printing functions *)
(**********************)

let separator =
  "-----------------------------------------------------------------"

let fprintf_time ff time =
  let total_seconds = time in
  let total_minutes = floor (total_seconds /. 60.0) in
  let total_hours = floor (total_minutes /. 60.0) in
  let relative_seconds_float = mod_float total_seconds 60.0 in
  let relative_minutes_float = mod_float total_minutes 60.0 in
  let seconds = relative_seconds_float in
  let minutes = int_of_float relative_minutes_float in
  let hours = int_of_float total_hours in
  if hours = 0 then
    if minutes = 0 then
      fprintf ff "%.3f\t(%.3fs)" total_seconds seconds
    else
      fprintf ff "%.3f\t(%im%.3fs)" total_seconds minutes seconds
  else
    fprintf ff "%.3f\t(%ih%im%.3fs)" total_seconds hours minutes seconds

let fprintf_memory ff memory =
  let bytes = memory in
  let kilo_bytes = (bytes /. 1024.0) in
  let mega_bytes = (kilo_bytes /. 1024.0) in
  fprintf ff "%.3f\t(%.3f M)" kilo_bytes mega_bytes

let print_stats phase_name phase =
  match get_phase_stats phase with
  | None -> ()
  | Some phase_stats ->
      begin
	fprintf !Conf.monitor_formatter "@,Elapsed time during \"%s\":\t\t\t%a" phase_name fprintf_time phase_stats.relative_stats.glx_time;
	fprintf !Conf.monitor_formatter "@,Live memory allocated during \"%s\":\t%a" phase_name fprintf_memory phase_stats.relative_stats.glx_live_memory;
	fprintf !Conf.monitor_formatter "@,Total memory allocated during \"%s\":\t%a" phase_name fprintf_memory phase_stats.relative_stats.glx_total_memory;
	fprintf !Conf.monitor_formatter "@,Total elapsed time after \"%s\":\t\t%a" phase_name fprintf_time phase_stats.absolute_stats.glx_time;
	fprintf !Conf.monitor_formatter "@,Live memory allocated after \"%s\":\t%a" phase_name fprintf_memory phase_stats.absolute_stats.glx_live_memory;
	fprintf !Conf.monitor_formatter "@;<0 -2>| Total memory allocated after \"%s\":\t%a" phase_name fprintf_memory phase_stats.absolute_stats.glx_total_memory
      end

let print_marshal_stats uri =
  match get_marshal_stats uri with
  | None -> ()
  | Some memory ->
      fprintf !Conf.monitor_formatter "@;<0 -2>| Size of marshalled document \"%s\":\t\t\t%a" uri fprintf_memory memory

let fprintf_time_fraction ff (total_time, partial_time, total_string) =
  let percentage = partial_time /. total_time *. 100.0 in
  fprintf ff "%.3f%% of total %s" percentage total_string

let fprintf_phase_opening ff phase =
  match phase with
  | Query_Toplevel_Phase ->
      fprintf !Conf.monitor_formatter "%s@\n@[<v 2>| " separator;
      fprintf ff "Starting toplevel@\n@\n"
  | Query_Processing_Context ->
      fprintf !Conf.monitor_formatter "%s@\n@[<v 2>| " separator;
      fprintf ff "Processing XQuery context@\n@\n"
  | Query_Processing_Expr expr ->
      fprintf !Conf.monitor_formatter "%s@\n@[<v 2>| " separator;
      fprintf ff "@[<v 2>XQuery expression:@\n[@%a@]@]@\n@\n" Print.print_expr_body expr
  | Query_Parsing_Phase ->
      fprintf !Conf.monitor_formatter "%s@\n@[<v 2>| " separator; 
      fprintf ff "XQuery parsing@\n@\n"
  | Query_Normalization_Phase ->
      fprintf !Conf.monitor_formatter "%s@\n@[<v 2>| " separator;
      fprintf ff "Normalization phase@\n@\n"
  | Query_Static_Typing_Phase ->
      fprintf !Conf.monitor_formatter "%s@\n@[<v 2>| " separator;
      fprintf ff "Static typing phase@\n@\n"
  | Query_Optimization_Phase ->
      fprintf !Conf.monitor_formatter "%s@\n@[<v 2>| " separator;
      fprintf ff "Optimization phase@\n@\n"
  | Query_Path_Analysis_Phase ->
      fprintf !Conf.monitor_formatter "%s@\n@[<v 2>| " separator;
      fprintf ff "Path projection analysis phase@\n@\n"
  | Query_Evaluation_Phase ->
      fprintf !Conf.monitor_formatter "%s@\n@[<v 2>| " separator;
      fprintf ff "Query evaluation phase@\n@\n"
  | Query_Serialization_Phase ->
      fprintf !Conf.monitor_formatter "%s@\n@[<v 2>| " separator;
      fprintf ff "Serialization phase@\n@\n"
  | Document_Parsing_Phase uri ->
      fprintf !Conf.monitor_formatter "%s@\n@[<v 2>| " separator;
      fprintf ff "Parsing document \"%s\"@\n@\n" (AnyURI._string_of_uri uri)
  | Document_Loading_Phase uri ->
      fprintf !Conf.monitor_formatter "%s@\n@[<v 2>| " separator;
      fprintf ff "Loading document \"%s\" in the data model@\n@\n" (AnyURI._string_of_uri uri)
  | Document_ParsingLoading_Phase uri ->
      fprintf !Conf.monitor_formatter "%s@\n@[<v 2>| " separator;
      fprintf ff "SAX parsing/loading document \"%s\"@\n@\n" (AnyURI._string_of_uri uri)
  | Expression_Phase _ ->
      ()

let fprintf_phase_closing ff phase =
  match phase with
  | Query_Toplevel_Phase ->
      let phase_name = "toplevel" in
      fprintf ff "Summary:";
      print_stats phase_name phase;
      fprintf !Conf.monitor_formatter "@]@\n%s@\n@\n" separator
  | Query_Processing_Context ->
      let phase_name = "context" in
      fprintf ff "Summary:";
      print_stats phase_name phase;
      fprintf !Conf.monitor_formatter "@]@\n%s@\n@\n" separator
  | Query_Processing_Expr expr ->
      let phase_name = "expression" in
      fprintf ff "Summary:";
      print_stats phase_name phase;
      fprintf !Conf.monitor_formatter "@]@\n%s@\n@\n" separator
  | Query_Parsing_Phase ->
      let phase_name = "parsing" in
      fprintf ff "Summary:";
      print_stats phase_name phase;
      fprintf !Conf.monitor_formatter "@]@\n%s@\n@\n" separator
  | Query_Normalization_Phase ->
      let phase_name = "normalization" in
      fprintf ff "Summary:";
      print_stats phase_name phase;
      fprintf !Conf.monitor_formatter "@]@\n%s@\n@\n" separator
  | Query_Static_Typing_Phase ->
      let phase_name = "typing" in
      fprintf ff "Summary:";
      print_stats phase_name phase;
      fprintf !Conf.monitor_formatter "@]@\n%s@\n@\n" separator
  | Query_Optimization_Phase ->
      let phase_name = "optimization" in
      fprintf ff "Summary:";
      print_stats phase_name phase;
      fprintf !Conf.monitor_formatter "@]@\n%s@\n@\n" separator
  | Query_Path_Analysis_Phase ->
      let phase_name = "analysis" in
      fprintf ff "Summary:";
      print_stats phase_name phase;
      fprintf !Conf.monitor_formatter "@]@\n%s@\n@\n" separator
  | Query_Evaluation_Phase ->
      let phase_name = "evaluation" in
      fprintf ff "Summary:"; 
      print_stats phase_name phase;
      fprintf !Conf.monitor_formatter "@]@\n%s@\n@\n" separator
  | Query_Serialization_Phase ->
      let phase_name = "serialization" in
      fprintf ff "Summary:";
      print_stats phase_name phase;
      fprintf !Conf.monitor_formatter "@]@\n%s@\n@\n" separator
  | Document_Parsing_Phase uri ->
      let phase_name = "parse-document" in
      fprintf ff "Summary:";
      print_stats phase_name phase;
      fprintf !Conf.monitor_formatter "@]@\n%s@\n@\n" separator
  | Document_Loading_Phase uri ->
      let phase_name = "load-document" in
      fprintf ff "Summary:";
      print_stats phase_name phase;
      print_marshal_stats (AnyURI._string_of_uri uri);
      fprintf !Conf.monitor_formatter "@]@\n%s@\n@\n" separator
  | Document_ParsingLoading_Phase uri ->
      let phase_name = "SAX parse-load-document" in
      fprintf ff "Summary:";
      print_stats phase_name phase;
      print_marshal_stats (AnyURI._string_of_uri uri);
      fprintf !Conf.monitor_formatter "@]@\n%s@\n@\n" separator
  | Expression_Phase _ ->
      ()

let fprintf_prolog ff () =
  fprintf ff "***********************@\n";
  fprintf ff "* Run-time monitoring *@\n";
  fprintf ff "***********************@\n@\n@\n"

let fprintf_time_summary ff (total_time,partial_time,total_string) =
  fprintf ff "\t%a\t(%a)" fprintf_time partial_time fprintf_time_fraction (total_time,partial_time,total_string)

let fprintf_document_time_summary ff total_time uri stats =
  let partial_time = stats.absolute_stats.glx_time in
  fprintf ff "@,Document \"%s\": %a" (AnyURI._string_of_uri uri) fprintf_time_summary (total_time,partial_time,"evaluation time")

let fprintf_documents_time_summary ff (total_time,docs) =
  Hashtbl.iter (fprintf_document_time_summary ff total_time) docs

let fprintf_evaluation_time_summary ff total_time =
  begin
    fprintf ff "@,@[<v 3>Parsing documents time:%a@]"
      fprintf_documents_time_summary (total_time,!current_summary.document_parsing_phase);
    fprintf ff "@,@[<v 3>Loading document time:%a@]"
      fprintf_documents_time_summary (total_time,!current_summary.document_loading_phase);
    fprintf ff "@,@[<v 3>SAX parsing/loading document time:%a@]"
      fprintf_documents_time_summary (total_time,!current_summary.document_parsingloading_phase)
  end
  
let fprintf_generate_summary ff () =
  let total_time = !latest_stats_point.absolute_stats.glx_time in
  begin
    fprintf ff "@,Total execution time:\t\t\t\t%a@," fprintf_time total_time;
    fprintf ff "@,XQuery context processing time:\t%a"
      fprintf_time_summary (total_time,!current_summary.query_toplevel_phase.relative_stats.glx_time,"execution time");
    fprintf ff "@,XQuery parsing time:\t\t%a"
      fprintf_time_summary (total_time,!current_summary.query_parsing_phase.relative_stats.glx_time,"execution time");
    fprintf ff "@,XQuery normalization time:\t%a"
      fprintf_time_summary (total_time,!current_summary.query_normalization_phase.relative_stats.glx_time,"execution time");
    fprintf ff "@,Static typing time:\t\t%a"
      fprintf_time_summary (total_time,!current_summary.query_static_typing_phase.relative_stats.glx_time,"execution time");
    fprintf ff "@,Optimization time:\t\t%a"
      fprintf_time_summary (total_time,!current_summary.query_optimization_phase.relative_stats.glx_time,"execution time");
    fprintf ff "@,Path projection analysis time:\t%a"
      fprintf_time_summary (total_time,!current_summary.query_path_analysis_phase.relative_stats.glx_time,"execution time");
    fprintf ff "@,@[<v 3>Query evaluation time:\t\t%a%a@]"
      fprintf_time_summary (total_time,!current_summary.query_evaluation_phase.relative_stats.glx_time,"execution time")
      fprintf_evaluation_time_summary total_time;
    fprintf ff "@,Serialization time:\t\t%a"
      fprintf_time_summary (total_time,!current_summary.query_serialization_phase.relative_stats.glx_time,"execution time")
  end
  

let fprintf_generate_one_pool_summary ff pool_name (sc,hash,array) =
  begin
    Format.fprintf ff "@,@,Summary for pool: %s:" pool_name;
    Format.fprintf ff "@,\tSymbol counter:\t%i" sc;
    Format.fprintf ff "@,\tHashtbl size:\t%i" hash;
    Format.fprintf ff "@,\tArray size:\t%i" array
  end

let fprintf_generate_pool_summary ff () =
  let sym_stats = Sym.stats () in
  fprintf_generate_one_pool_summary ff "Unresolved element names" sym_stats.Sym.unresolved_element_names;
  fprintf_generate_one_pool_summary ff "Unresolved attribute names" sym_stats.Sym.unresolved_attribute_names;
  fprintf_generate_one_pool_summary ff "Unresolved type names" sym_stats.Sym.unresolved_type_names;
  fprintf_generate_one_pool_summary ff "Resolved anon names" sym_stats.Sym.resolved_anon_names;
  fprintf_generate_one_pool_summary ff "Resolved element names" sym_stats.Sym.resolved_element_names;
  fprintf_generate_one_pool_summary ff "Resolved attribute names" sym_stats.Sym.resolved_attribute_names;
  fprintf_generate_one_pool_summary ff "Resolved type names" sym_stats.Sym.resolved_type_names

let fprintf_pool_summary () =
  let ff = !Conf.monitor_formatter in
  Format.fprintf ff "@\n@\n";
  Format.fprintf ff "**********************@\n";
  Format.fprintf ff "* Data model summary *@\n";
  Format.fprintf ff "**********************@\n\n";
  Format.fprintf ff "Total number of created nodes:\t%i@\n\n" current_stats.node_count;
  Format.fprintf ff "String pools@\n";
  Format.fprintf ff "------------@\n";
  Format.fprintf ff "@[<v 0>%a@]@\n@\n" fprintf_generate_pool_summary ()


let fprintf_summary ff () =
  fprintf ff "@\n@\n";
  fprintf ff "********************@\n";
  fprintf ff "* Activity summary *@\n";
  fprintf ff "********************@\n@\n";
  fprintf ff "@[<v 0>%a@]@\n@\n" fprintf_generate_summary ();
  fprintf_pool_summary ()



(*****************************)
(* Phase push/pop discipline *)
(*****************************)

let phase_stack =
  Stack.create ()

let process_starting_phase phase =
  let stats_point = new_stats_point () in
  begin
    try
      let previous_phase = Stack.top phase_stack in
      add_relative_stats_to_summary previous_phase stats_point.relative_stats
    with
    | Stack.Empty ->
	()
  end;
  Stack.push phase phase_stack

let process_closing_phase () =
  let stats_point = new_stats_point () in
  let phase = Stack.pop phase_stack in
  begin
    add_relative_stats_to_summary phase stats_point.relative_stats;
    set_absolute_stats_in_summary phase stats_point.absolute_stats;
    phase
  end

let push_phase phase =
  begin
    process_starting_phase phase;
    fprintf_phase_opening !Conf.monitor_formatter phase
  end

let pop_phase () =
  let phase = process_closing_phase () in
  fprintf_phase_closing !Conf.monitor_formatter phase

let init_monitor () =
  if !Conf.monitor then
    begin
      current_summary := empty_summary ();
      latest_stats_point := empty_stats_point;
      fprintf_prolog !Conf.monitor_formatter ();
      push_phase Query_Toplevel_Phase
    end

let close_monitor () =
  if !Conf.monitor then
    begin
      pop_phase ();
      fprintf_summary !Conf.monitor_formatter ()
    end

(*
val push_expr_kind : expr_kind -> unit

val pop_expr_kind : unit -> expr_kind

*)

(***********************)
(* Monitoring wrappers *)
(***********************)

let wrap_monitor phase f x =
  if !Conf.monitor then
    begin
      push_phase phase;
      let y = f x in
      pop_phase ();
      y
    end
  else
    f x

