(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2001, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: namespace.ml,v 1.5 2003/05/20 15:29:50 mff Exp $ *)

open Error

(***********************************************)
(* This module contains support for namespaces *)
(***********************************************)

(**************)
(* Name types *)
(**************)

(* Local name *)

type ncname = string

(* Uri *)

type uri =
  | NSUri of string
  | NSWildcardUri

let string_of_uri uri =
  match uri with
  | NSWildcardUri ->
      "\"*\""
  | NSUri uri ->
      "\"" ^ uri ^ "\""

(* Namespace prefix *)

type prefix =
  | NSDefaultElementPrefix
  | NSDefaultFunctionPrefix
  | NSWildcardPrefix
  | NSPrefix of ncname


(***********************)
(* Built-in namespaces *)
(***********************)

(* XML Schema namespaces *)

let xml_prefix   = NSPrefix "xml"
let xmlns_prefix = NSPrefix "xmlns"
let xs_prefix    = NSPrefix "xs"
let xsd_prefix   = NSPrefix "xsd"
let xsi_prefix   = NSPrefix "xsi"
let fn_prefix    = NSPrefix "fn"
let op_prefix    = NSPrefix "op"
let xdt_prefix   = NSPrefix "xdt"
let xq_prefix    = NSPrefix "xq"
let glx_prefix   = NSPrefix "glx"

let empty_uri 	 = NSUri (Conf.emptyns)

let xml_uri    	 = NSUri (Conf.xmlns)
let xmlns_uri    = NSUri (Conf.xmlnsns)
let xs_uri    	 = NSUri (Conf.xsns)
let xsd_uri   	 = NSUri (Conf.xsdns)
let xsi_uri   	 = NSUri (Conf.xsins)
let fn_uri    	 = NSUri (Conf.fnns)
let op_uri    	 = NSUri (Conf.opns)
let xdt_uri    	 = NSUri (Conf.xdtns)
let xq_uri    	 = NSUri (Conf.xqns)
let glx_uri    	 = NSUri (Conf.glxns)

(* Wildcard namespace *)

(* Note:
     Is that a hack or is that smart ?
   - J&B
*)

let wild_ns             = (NSWildcardPrefix, NSWildcardUri)

(* Default namespaces *)

let default_element_ns  = (NSDefaultElementPrefix, empty_uri)
let default_function_ns = (NSDefaultFunctionPrefix, fn_uri)

(* Table of default namespaces *)

let default_namespaces =
  [ default_element_ns;
    default_function_ns ]

(* Table of fixed, built-in namespaces *)

let fixed_built_in_namespaces =
  [ (xml_prefix, xml_uri);
    (xmlns_prefix, xmlns_uri) ]

(* Table of built-in namespaces *)

let built_in_namespaces =
  default_namespaces @
  fixed_built_in_namespaces @
  [ (xs_prefix, xs_uri);
    (xsd_prefix, xsd_uri);
    (xsi_prefix, xsi_uri);
    (fn_prefix, fn_uri);
    (op_prefix, op_uri);
    (xdt_prefix, xdt_uri);
    (xq_prefix, xq_uri);
    (glx_prefix, glx_uri); ]

let is_built_in_namespace prefix_uri =
  List.exists (fun x -> x = prefix_uri) built_in_namespaces

let is_right_fixed_namespace (prefix,uri) =
  (not(prefix = xml_prefix) &&
   not(prefix = xmlns_prefix) &&
   not(uri = xml_uri) &&
   not(uri = xmlns_uri)) ||
   (List.exists (fun x -> x = (prefix,uri)) fixed_built_in_namespaces)


(*******************)
(* Qualified named *)
(*******************)

(* Unresolved qualified named *)

type uqname = prefix * ncname

(* special unresolved qualified names *)

let anyuqname = (NSWildcardPrefix, "*")

(* takes a string value containing a ':' and extract prefix and local name *)

let uqname_element_of_string s =
  match (Gmisc.wrap_lexer Qname_lexer.parse_qname s) with
  | (None,s2) ->
      (NSDefaultElementPrefix,s2)
  | (Some s1,s2) ->
      (NSPrefix s1,s2)

let uqname_function_of_string s =
  match (Gmisc.wrap_lexer Qname_lexer.parse_qname s) with
  | (None,s2) ->
      (NSDefaultFunctionPrefix,s2)
  | (Some s1,s2) ->
      (NSPrefix s1,s2)

(* takes a string value containing a ':' and extract prefix and local name *)

let wildcard_uqname_of_string s =
  match (Gmisc.wrap_lexer Qname_lexer.parse_qname s) with
  | (None,s2) ->
      (NSDefaultElementPrefix,s2)
  | (Some "*", s2) ->
      (NSWildcardPrefix, s2)
  | (Some s1, s2) ->
      (NSPrefix s1,s2)

(* takes a string value containing a ':' and extract prefix and local name *)

let escaped_uqname_of_string s =
  if (String.get s 0) = ':'
  then
    uqname_element_of_string (String.sub s 1 ((String.length s)-1))
  else
    uqname_element_of_string s

(* string value for a resolved qname *)

let string_of_uqname qn =
  match qn with
  | (NSDefaultElementPrefix, nc) -> nc
  | (NSDefaultFunctionPrefix, nc) -> nc
  | (NSWildcardPrefix, nc) -> "*:" ^ nc
  | (NSPrefix ns, nc) -> ns ^ ":" ^ nc

(* Resolved qualified named *)

type rqname = uri * ncname

(* special resolved qualified names *)

let anyrqname = (NSWildcardUri, "*")

(* Subtagging between resolved qualified names *)

(* Note:
    - Subtagging does not make sense on unresolved qnames
    - Eventually subtagging will have to go into a specific module for
      wildcards.
   - B&J *)

let subrqname q1 q2 =
  let (uri1,nc1) = q1
  and (uri2,nc2) = q2 in
  if uri2 = NSWildcardUri
  then
    if nc2 = "*"
    then
      true
    else
      nc1 = nc2
  else
    if nc2 = "*"
    then
      uri1 = uri2
    else
      (uri1 = uri2) && (nc1 = nc2)

(* string value for a resolved qname *)

let string_of_rqname qn =
  let (uri,nc) = qn in
  (string_of_uri uri) ^ ":" ^ nc

(**********************************)
(* Operations on prefixes symbols *)
(**********************************)

(* From resolved symbols to original prefixes *)

let uri_prefix_table : (uri, prefix) Hashtbl.t = Hashtbl.create 167
let prefix_uri_table : (prefix, uri) Hashtbl.t = Hashtbl.create 167

let unsafe_add_uri uri prefix =
  Hashtbl.add uri_prefix_table uri prefix;
  Hashtbl.add prefix_uri_table prefix uri

(* Adding the xml built-in prefixes *)

let _ =
  unsafe_add_uri xml_uri xml_prefix;
  unsafe_add_uri xmlns_uri xmlns_prefix

(* Creates a new prefix *)

let prefix_counter = Id.create 0

let new_prefix prefix =
  let new_nc =
    match prefix with
    | NSDefaultElementPrefix ->
	"xq"
    | NSDefaultFunctionPrefix ->
	"fn"
    | NSWildcardPrefix ->
	"*"
    | (NSPrefix nc) ->
	nc
  in
  (NSPrefix (new_nc ^ "__" ^ (string_of_int (Id.next prefix_counter))))

(* Adds hookup from uri to prefix *)

let rec add_uri_prefix uri prefix =
  if is_right_fixed_namespace (prefix,uri)
  then
    if not(Hashtbl.mem uri_prefix_table uri)
    then
      try
	if uri != Hashtbl.find prefix_uri_table prefix 
	then
	  unsafe_add_uri uri (new_prefix prefix)
      with
      | Not_found ->
	  unsafe_add_uri uri prefix
    else
      ()
  else
    begin
      match (uri,prefix) with
      | (NSUri uri, NSPrefix prefix) -> 
	  raise (Query (URI_Error ("Cannot bind built-in prefix " ^ prefix ^ " to URI " ^ uri)))
      | _ ->
	  ()
    end

(* Retrieves hookup from symbol to prefix *)

let get_uri_prefix uri =
  try
    Hashtbl.find uri_prefix_table uri
  with
  | Not_found ->
      raise (Query (Internal_Error ("Uri not '"^ (string_of_uri uri) ^"'registered in prefix table")))

(*************************)
(* Namespace environment *)
(*************************)

(* type for namespace environments *)

type nsenv =
    { all_namespaces: (prefix * uri) list;
      schema_namespaces: uri list }

(* creates an empty environment *)

let empty_nsenv () =
  { all_namespaces = wild_ns :: (default_namespaces @ fixed_built_in_namespaces);
    schema_namespaces = [] }

(* adds a new prefix,uri binding in a namespace environment *)

let add_ns nsr pref uri =
  begin
    match pref with
    | NSWildcardPrefix ->
	raise (Query (Internal_Error ("Cannot redefine the wildcard prefix")))
    | NSDefaultElementPrefix ->
	add_uri_prefix uri pref
    | NSDefaultFunctionPrefix ->
	add_uri_prefix uri pref
    | NSPrefix _ ->
	(* WARNING: Here I test updating the URI/prefix table at the
  	   namespace env level - Jerome *)
	add_uri_prefix uri pref
  end;
  { all_namespaces = (pref, uri) :: nsr.all_namespaces;
    schema_namespaces = nsr.schema_namespaces }

let add_all_ns nsr pul =
  let local_add_ns nsr (x,y) = add_ns nsr x y in
  List.fold_left local_add_ns nsr pul

(* gets the namespace of a given prefix *)

let get_ns_of_prefix nsr pref =
  try
    List.assoc pref nsr.all_namespaces
  with
  | Not_found ->
      begin
	match pref with
	| NSDefaultElementPrefix ->
	    raise (Query (Internal_Error ("Default element namespace not found.")))
	| NSDefaultFunctionPrefix ->
	    raise (Query (Internal_Error ("Default function namespace not found.")))
	| NSWildcardPrefix ->
	    raise (Query (Internal_Error ("Wildcard namespace not found.")))
        | NSPrefix nc ->
	    raise (Query (Internal_Error ("Namespace prefix "
					  ^ nc
					  ^ " not found.")))
      end

(* adds a known schema target uri in the environment *)

let add_schema_ns nsr uri =
  { all_namespaces = nsr.all_namespaces;
    schema_namespaces = uri :: nsr.schema_namespaces }

(* is a namespace bound to a schema ? *)

let is_schema_ns nsr uri =
  List.mem uri nsr.schema_namespaces

(* build a new environment from a list of namespaces and schema uris *)

let build_ns ns_bindings schema_ns =
  { all_namespaces = ns_bindings;
    schema_namespaces = schema_ns }

(* Append two namespace environments *)

let append_nsenv nsr1 nsr2 =
  { all_namespaces = nsr1.all_namespaces @ nsr2.all_namespaces;
    schema_namespaces = nsr1.schema_namespaces @ nsr2.schema_namespaces }

