(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                           XML Query Engine                          *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id $ *)

(*********************************************************)
(* This module implement namespace resolution for qnames *)
(*********************************************************)

open Namespace
open Sym
open Error

(********************************************)
(* Functions to check consistency of qnames *)
(********************************************)

(* Checks that the resolved qname exists in case the namespace is
   bound to a schema *)

(* Note:
     f is a boolean function
 - J&B *)

let is_schema_consistent nsr rqname f =
  let (uri,nc) = rqname in
  if (is_schema_ns nsr uri)                    (* if the resolved qname is bound to a schema *)
  then
    f rqname                                   (* then check function f *)
  else
    true                                       (* otherwise it's just ok *)

let is_group_defined nsr rgroup_sym =
  let rqname = rgroup_name rgroup_sym in
  if (is_schema_consistent nsr rqname Sym.rgroup_exists)
  then
    ()
  else
    raise (Query (Prototype ("Group name is undefined in namespace ")))

let is_elem_defined nsr relem_sym =
  let rqname = relem_name relem_sym in
  if (is_schema_consistent nsr rqname Sym.relem_exists)
  then
    ()
  else
    raise (Query (Prototype ("Element name is undefined in namespace ")))

let is_attr_defined nsr rattr_sym =
  let rqname = rattr_name rattr_sym in
  if (is_schema_consistent nsr rqname Sym.rattr_exists)
  then
    ()
  else
    raise (Query (Prototype ("Element name is undefined in namespace ")))

(**********************************************)
(* Resolving QNames to Expanded QNames        *)
(* Expanded QNames are represented by symbols *)
(**********************************************)

(* Note:
     Given namespace environment an unresolved symbol, returns the
     resolved symbol for the resolved qname
   - J&B
*)

let resolve_element_qname nsr uqname =
  match uqname with
  | (NSWildcardPrefix, localname) ->
      begin
	try
	  let uri = get_ns_of_prefix nsr NSWildcardPrefix in
	  (NSWildcardPrefix, (uri, localname))
	with
	| _ -> raise (Query (Undefined ("Cannot resolve default element namespace prefix ")))
      end
  | (NSDefaultElementPrefix, localname) ->
      begin
	try
	  let uri = get_ns_of_prefix nsr NSDefaultElementPrefix in
	  (NSDefaultElementPrefix, (uri, localname))
	with
	| _ -> raise (Query (Undefined ("Cannot resolve default element namespace prefix ")))
      end
  | (NSPrefix nc, localname) ->
      begin
	try
	  let uri = get_ns_of_prefix nsr (NSPrefix nc) in
	  ((NSPrefix nc), (uri, localname))
	with
	| _ -> raise (Query (Undefined ("Cannot resolve namespace prefix " ^ nc)))
      end
  | (NSDefaultFunctionPrefix, localname) ->
      raise (Query (Internal_Error "Trying to resolve an element name as a function name"))

let resolve_function_qname nsr uqname =
  match uqname with
  | (NSWildcardPrefix, localname) ->
      raise (Query (Internal_Error ("Resolution of wildcard namespace not supported!")))
  | (NSDefaultElementPrefix, localname) ->
      raise (Query (Internal_Error "Trying to resolve a function name as an element name"))
  | (NSPrefix nc, localname) ->
      begin
	try
	  let uri = get_ns_of_prefix nsr (NSPrefix nc) in
	  ((NSPrefix nc), (uri, localname))
	with
	| _ -> raise (Query (Undefined ("Cannot resolve namespace prefix " ^ nc)))
      end
  | (NSDefaultFunctionPrefix, localname) ->
      begin
	try
	  let uri = get_ns_of_prefix nsr NSDefaultFunctionPrefix in
	  (NSDefaultFunctionPrefix, (uri, localname))
	with
	| _ -> raise (Query (Undefined ("Cannot resolve default element namespace prefix ")))
      end

let resolve_group_qname nsr uqname =
  match uqname with
  | (NSWildcardPrefix, localname) ->
      begin
	try
	  let uri = get_ns_of_prefix nsr NSWildcardPrefix in
	  (NSWildcardPrefix, (uri, localname))
	with
	| _ -> raise (Query (Undefined ("Cannot resolve default element namespace prefix ")))
      end
  | (NSDefaultElementPrefix, localname) ->
      begin
	try
	  let uri = get_ns_of_prefix nsr NSDefaultElementPrefix in
	  (NSDefaultElementPrefix, (uri, localname))
	with
	| _ -> raise (Query (Undefined ("Cannot resolve default element namespace prefix ")))
      end
  | (NSPrefix nc, localname) ->
      begin
	try
	  let uri = get_ns_of_prefix nsr (NSPrefix nc) in
	  ((NSPrefix nc), (uri, localname))
	with
	| _ -> raise (Query (Undefined ("Cannot resolve namespace prefix " ^ nc)))
      end
  | (NSDefaultFunctionPrefix, localname) ->
      raise (Query (Internal_Error "Trying to resolve an element name as a function name"))

let resolve_attribute_qname nsr uqname =
  match uqname with 
  | (NSWildcardPrefix, localname) ->
      begin
	try
	  let uri = get_ns_of_prefix nsr NSWildcardPrefix in
	  (NSWildcardPrefix, (uri, localname))
	with
	| _ -> raise (Query (Undefined ("Cannot resolve default element namespace prefix ")))
      end
  | (NSDefaultElementPrefix, localname) ->
      (NSDefaultElementPrefix,(NSUri "", localname))
  | (NSPrefix prefix, localname) -> 
      begin
	try
	  let uri = get_ns_of_prefix nsr (NSPrefix prefix) in 
	  let rqname = (uri, localname) in
	  (NSPrefix prefix,rqname)
	with
	| _ ->
	    raise (Query (Undefined ("Namespace prefix " ^ prefix ^ " in attribute name")))
      end
  | (NSDefaultFunctionPrefix, _) ->
      raise (Query (Internal_Error "Trying to resolve an attribute name as a function name"))

let resolve_ugroup_qname nsr ugroup_sym = 
  let uqname = ugroup_name ugroup_sym in 
  rgroup_symbol (resolve_group_qname nsr uqname)

let resolve_uelem_qname nsr uelem_sym = 
  let uqname = uelem_name uelem_sym in 
  relem_symbol (resolve_element_qname nsr uqname)

(* Note:
     Resolving QNames of groups and elements is identical, attributes
     are treated differently, because they are never in default
     namespace. this is unfortunate.
*)

let resolve_uattr_qname nsr uattr_sym = 
  let uqname = uattr_name uattr_sym in 
  rattr_symbol (resolve_attribute_qname nsr uqname)

let resolve_group_symbol nsenv group_sym =
  match group_sym with
  | UGroupSym us ->
      resolve_ugroup_qname nsenv us
  | RGroupSym rs ->
      rs

let resolve_elem_symbol nsenv elem_sym =
  match elem_sym with
  | UElemSym us ->
      resolve_uelem_qname nsenv us
  | RElemSym rs ->
      rs

let resolve_attr_symbol nsenv attr_sym =
  match attr_sym with
  | UAttrSym us ->
      resolve_uattr_qname nsenv us
  | RAttrSym rs ->
      rs

