(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                           XML Query Engine                          *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: sym.ml,v 1.4 2003/05/20 15:29:50 mff Exp $ *)

open Error
open Pool
open Namespace

(************************************************)
(* This module implements qname symbols support *)
(************************************************)

(***************)
(* QName pools *)
(***************)

(* Note:
     The string pools are working on both unresolved and resolved
     qualified names.
   - B & J *)

(* Namepool for unresolved qualified names *)

module UnresolvedQName =
  struct
    type t = uqname
    let equal = (=)
    let hash = Hashtbl.hash
  end

module UnresolvedQNamePool = MakeNamePool (UnresolvedQName)

(* Namepool for resolved qualified names *)

module ResolvedQName =
  struct
    type t = rqname
    let equal = (=)
    let hash = Hashtbl.hash
  end

module ResolvedQNamePool = MakeNamePool (ResolvedQName)

(* Namepool for resolved qualified names with a prefix *)

module Prefix =
  struct
    type t = prefix
    let equal = (=)
    let hash = Hashtbl.hash
  end

module PrefixPool = MakeNamePool (Prefix)

(* Pool for unresolved qnames *)

(* Note:
     It might well be that we could get rid of the string pool for
     unresolved qnames if we are smart during parsing. But trying to
     do that is probably non trivial so we keep it like that for now.
   - B & J *)

let uelem_pool  = UnresolvedQNamePool.create_pool ()
let uattr_pool  = UnresolvedQNamePool.create_pool ()
let ugroup_pool = UnresolvedQNamePool.create_pool ()

(* Pool for resolved qnames with an original prefix *)

let rname_pool  = ResolvedQNamePool.create_pool ()

(* Pool for resolved qnames with an original prefix *)

let relem_pool  = ResolvedQNamePool.create_pool ()
let rattr_pool  = ResolvedQNamePool.create_pool ()
let rgroup_pool = ResolvedQNamePool.create_pool ()

(*********************************)
(* Operations on element symbols *)
(*********************************)

(* Operations on unresolved element symbols *)

type uelem_symbol = UnresolvedQNamePool.symbol

(* Creates/accesses symbols *)

let uelem_symbol nm = UnresolvedQNamePool.add_name uelem_pool nm
let uelem_name  sym = UnresolvedQNamePool.get_name uelem_pool sym

(* Checks if a name is already defined *)

let uelem_exists nm =
  UnresolvedQNamePool.exists_name uelem_pool nm

(* Converts a symbol to a string *)

let uelem_string s =
  string_of_uqname (uelem_name s)

(* Operations on resolved element symbols *)

type relem_symbol = ResolvedQNamePool.symbol

(* Creates/accesses symbols *)

let relem_name sym   = ResolvedQNamePool.get_name relem_pool sym

let relem_symbol (p,nm) =
  let (uri,_) = nm in
  add_uri_prefix uri p;
  ResolvedQNamePool.add_name relem_pool nm

let relem_prefix sym =
  let (uri,_) = relem_name sym in
  (get_uri_prefix uri)

(* Checks if a name is already defined *)

let relem_exists nm =
  ResolvedQNamePool.exists_name relem_pool nm

(* Equality between symbols *)

let relem_equals =
  (ResolvedQNamePool.symbol_equals relem_pool)

(* Subtagging between symbols *)

let relem_subtag t t' =
  let (uri1, lcname1) = relem_name t and
      (uri2, lcname2) = relem_name t' in
  (((uri1 = uri2) && (lcname1 = lcname2))
 || ((uri1 = uri2) && (lcname2 = "*"))
 || ((uri2 = NSWildcardUri) && (lcname1 = lcname2))
 || ((uri2 = NSWildcardUri) && (lcname2 = "*")))
  
(* Converts a symbol to a string *)

let relem_string s =
  let rs = relem_name s in
  string_of_rqname rs

let relem_prefix_string s =
  let (_,nc) = relem_name s
  and prefix = relem_prefix s in
  string_of_uqname (prefix,nc)

(* Special symbols *)

let anyrelem = relem_symbol (NSWildcardPrefix,anyrqname)

(************************************)
(* Operations on attributes symbols *)
(************************************)

(* Operations on unresolved attribute symbols *)

type uattr_symbol = UnresolvedQNamePool.symbol

(* Creates/accesses symbols *)

let uattr_symbol nm = UnresolvedQNamePool.add_name uattr_pool nm
let uattr_name  sym = UnresolvedQNamePool.get_name uattr_pool sym

(* Checks if a name is already defined *)

let uattr_exists nm =
  UnresolvedQNamePool.exists_name uattr_pool nm

(* Converts a symbol to a string *)

let uattr_string s =
  string_of_uqname (uattr_name s)

(* Operations on resolved attribute symbols *)

type rattr_symbol = ResolvedQNamePool.symbol

(* Creates/accesses symbols *)

let rattr_name sym   = ResolvedQNamePool.get_name rattr_pool sym

let rattr_symbol (p,nm)  =
  let (uri,_) = nm in
  add_uri_prefix uri p;
  ResolvedQNamePool.add_name rattr_pool nm

let rattr_prefix sym =
  let (uri,_) = rattr_name sym in
  let prefix = get_uri_prefix uri in
  match uri with
  | NSUri "" ->
      NSDefaultElementPrefix
  | _ ->
      prefix

(* Checks if a name is already defined *)

let rattr_exists nm =
  ResolvedQNamePool.exists_name rattr_pool nm

(* Equality between symbols *)

let rattr_equals =
  (ResolvedQNamePool.symbol_equals rattr_pool)

(* Subtagging between symbols *)

let rattr_subtag t t' =
  let (uri1, lcname1) = rattr_name t and
      (uri2, lcname2) = rattr_name t' in
  (((uri1 = uri2) && (lcname1 = lcname2)) 
 || ((uri1 = uri2) && (lcname2 = "*"))
 || ((uri2 = NSWildcardUri) && (lcname1 = lcname2))
 || ((uri2 = NSWildcardUri) && (lcname2 = "*")))

(* Converts a symbol to a string *)

let rattr_string s =
  let rs = rattr_name s in
  string_of_rqname rs

let rattr_prefix_string s =
  let (_,nc) = rattr_name s
  and prefix = rattr_prefix s in
  string_of_uqname (prefix,nc)

(* Special symbols *)

let anyrattr = rattr_symbol (NSWildcardPrefix,anyrqname)

(*******************************)
(* Operations on group symbols *)
(*******************************)

(* Operations on unresolved group symbols *)

type ugroup_symbol = UnresolvedQNamePool.symbol

(* Creates/accesses symbols *)

let ugroup_symbol nm = UnresolvedQNamePool.add_name ugroup_pool nm
let ugroup_name  sym = UnresolvedQNamePool.get_name ugroup_pool sym

(* Checks if a name is already defined *)

let ugroup_exists nm =
  UnresolvedQNamePool.exists_name ugroup_pool nm

(* Converts a symbol to a string *)

let ugroup_string s =
  string_of_uqname (ugroup_name s)

(* Operations on resolved group symbols *)

type rgroup_symbol = ResolvedQNamePool.symbol

(* Creates/accesses symbols *)

let rgroup_name sym   = ResolvedQNamePool.get_name rgroup_pool sym

let rgroup_symbol (p,nm)  =
  let (uri,_) = nm in
  add_uri_prefix uri p;
  ResolvedQNamePool.add_name rgroup_pool nm

let rgroup_prefix sym =
  let (uri,_) = rgroup_name sym in
  (get_uri_prefix uri)

(* Checks if a name is already defined *)

let rgroup_exists nm =
  ResolvedQNamePool.exists_name rgroup_pool nm

(* Equality between symbols *)

let rgroup_equals =
  (ResolvedQNamePool.symbol_equals rgroup_pool)

(* Converts a symbol to a string *)

let rgroup_string s =
  let rs = rgroup_name s in
  string_of_rqname rs

let rgroup_prefix_string s =
  let (_,nc) = rgroup_name s
  and prefix = rgroup_prefix s in
  string_of_uqname (prefix,nc)

(* Special symbols *)

let anynumeric           = rgroup_symbol (NSPrefix "xs", (Namespace.xs_uri, "AnyNumeric"))
let anyatomic            = rgroup_symbol (NSPrefix "xdt", (Namespace.xdt_uri, "anyAtomicType"))
let anysimpletype        = rgroup_symbol (NSPrefix "xs", (Namespace.xs_uri, "AnySimpleType"))
let anyattribute         = rgroup_symbol (NSPrefix "xs", (Namespace.xs_uri, "AnyAttribute"))
let anyelement           = rgroup_symbol (NSPrefix "xs", (Namespace.xs_uri, "AnyElement"))
let anydoc               = rgroup_symbol (NSPrefix "xs", (Namespace.xs_uri, "AnyDocument"))
let anytype              = rgroup_symbol (NSPrefix "xs", (Namespace.xs_uri, "AnyType"))
let anynode              = rgroup_symbol (NSPrefix "xs", (Namespace.xs_uri, "AnyNode"))
let anyitem              = rgroup_symbol (NSPrefix "xs", (Namespace.xs_uri, "AnyItem"))

let roottypesym          = rgroup_symbol (NSPrefix "xs", (Namespace.xs_uri, "__root_type"))

(*************************************)
(* Operations on simple type symbols *)
(*************************************)

(* Operations on resolved simple symbols *)

type simple_symbol = rgroup_symbol

(* Creates/accesses symbols *)

let simple_name sym   = ResolvedQNamePool.get_name rgroup_pool sym

let simple_symbol (p,nm)  =
  let (uri,_) = nm in
  add_uri_prefix uri p;
  ResolvedQNamePool.add_name rgroup_pool nm

let simple_prefix sym =
  let (uri,_) = relem_name sym in
  (get_uri_prefix uri)

(* Checks if a name is already defined *)

let simple_exists nm =
  ResolvedQNamePool.exists_name rgroup_pool nm

(* Equality between symbols *)

let simple_equals =
  (ResolvedQNamePool.symbol_equals rgroup_pool)

(* Converts a symbol to a string *)

let simple_string s =
  let rs = simple_name s in
  string_of_rqname rs

let simple_prefix_string s =
  let (_,nc) = rgroup_name s
  and prefix = rgroup_prefix s in
  string_of_uqname (prefix,nc)

(* Special symbols *)

let anyatomicsym         = simple_symbol (NSPrefix "xdt", (Namespace.xdt_uri, "anyAtomicType"))

let stringsym            = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "string"))
let booleansym           = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "boolean"))
let decimalsym           = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "decimal"))
let floatsym             = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "float"))
let doublesym            = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "double"))
let durationsym          = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "duration"))
let dateTimesym          = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "dateTime"))
let timesym              = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "time"))
let datesym              = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "date"))
let gYearMonthsym        = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "gYearMonth"))
let gYearsym             = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "gYear"))
let gMonthDaysym         = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "gMonthDay"))
let gDaysym              = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "gDay"))
let gMonthsym            = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "gMonth"))
let hexBinarysym         = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "hexBinary"))
let base64Binarysym      = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "base64Binary"))
let anyURIsym            = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "anyURI"))
let qnamesym             = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "qname"))
let notationsym          = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "notation"))

let integersym           = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "integer"))
let intsym               = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "int"))

(* Predefined types in xdt namespace: *)
let untypedsym           = simple_symbol (NSPrefix "xdt", (Namespace.xdt_uri, "untypedAtomic"))
let dayTimeDurationsym           = simple_symbol (NSPrefix "xdt", (Namespace.xdt_uri, "dayTimeDuration"))
let yearMonthDurationsym           = simple_symbol (NSPrefix "xdt", (Namespace.xdt_uri, "yearMonthDuration"))

let nssym                = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "ns"))
let pisym                = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "pi"))
let commentsym           = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "comment"))
let cdatasym             = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "cdata"))
let refsym               = simple_symbol (NSPrefix "xsd", (Namespace.xsd_uri, "ref"))

(***********************************)
(* Operations on namespace symbols *)
(***********************************)

(* Operations on resolved namespace symbols *)

type ns_symbol = ResolvedQNamePool.symbol

(* Creates/accesses symbols *)

let ns_symbol nm  = ResolvedQNamePool.add_name rname_pool nm
let ns_name sym   = ResolvedQNamePool.get_name rname_pool sym

(* Checks if a name is already defined *)

let ns_exists nm =
  ResolvedQNamePool.exists_name rname_pool nm

(* Equality between symbols *)

let ns_equals =
  (ResolvedQNamePool.symbol_equals rname_pool)

(* Converts a symbol to a string *)

let ns_string s =
  string_of_rqname (ns_name s)

(***********************************)
(* Operations on anonymous symbols *)
(***********************************)

(* Operations on resolved anonymous symbols *)

type anon_symbol = ResolvedQNamePool.symbol

(* Creates/accesses symbols *)

let anon_symbol nm  = ResolvedQNamePool.add_name rname_pool nm
let anon_name sym   = ResolvedQNamePool.get_name rname_pool sym

let anon_prefix sym =
  let (uri,_) = anon_name sym in
  (get_uri_prefix uri)

(* Checks if a name is already defined *)

let anon_exists nm =
  ResolvedQNamePool.exists_name rname_pool nm

(* Equality between symbols *)

let anon_equals =
  (ResolvedQNamePool.symbol_equals rname_pool)

(* Converts a symbol to a string *)

let anon_string s =
  string_of_rqname (anon_name s)

let anon_prefix_string s =
  let (_,nc) = anon_name s
  and prefix = anon_prefix s in
  string_of_uqname (prefix,nc)

(************************)
(* Generic symbol types *)
(************************)

(* Generic element symbols *)

type elem_symbol =
  | UElemSym of uelem_symbol
  | RElemSym of relem_symbol

let relem_symbol_of_elem_symbol elems =
  match elems with
  | UElemSym _ ->
      raise (Query (Internal_Error "Casting an unresolved element symbol to a resolved element symbol"))
  | RElemSym res ->
      res

let string_of_elem_symbol elems =
  match elems with
  | UElemSym ues ->
      uelem_string ues
  | RElemSym res ->
      relem_string res

let string_prefix_of_elem_symbol elems =
  match elems with
  | UElemSym ues ->
      uelem_string ues
  | RElemSym res ->
      relem_prefix_string res

(* Generic attribute symbols *)

type attr_symbol =
  | UAttrSym of uattr_symbol
  | RAttrSym of rattr_symbol

let rattr_symbol_of_attr_symbol attrs =
  match attrs with
  | UAttrSym _ ->
      raise (Query (Internal_Error "Casting an unresolved attribute symbol to a resolved attribute symbol"))
  | RAttrSym ras ->
      ras

let string_of_attr_symbol attrs =
  match attrs with
  | UAttrSym uas ->
      uattr_string uas
  | RAttrSym ras ->
      rattr_string ras

let string_prefix_of_attr_symbol attrs =
  match attrs with
  | UAttrSym uas ->
      uattr_string uas
  | RAttrSym ras ->
      rattr_prefix_string ras

(* Generic group symbols *)

type group_symbol =
  | UGroupSym of ugroup_symbol
  | RGroupSym of rgroup_symbol

let rgroup_symbol_of_group_symbol groups =
  match groups with
  | UGroupSym _ ->
      raise (Query (Internal_Error "Casting an unresolved group symbol to a resolved group symbol"))
  | RGroupSym rgs ->
      rgs

let string_of_group_symbol groups =
  match groups with
  | UGroupSym ugs ->
      ugroup_string ugs
  | RGroupSym rgs ->
      rgroup_string rgs

let string_prefix_of_group_symbol groups =
  match groups with
  | UGroupSym ugs ->
      ugroup_string ugs
  | RGroupSym rgs ->
      rgroup_prefix_string rgs

(* Generic resolved symbols *)

type resolved_unit_symbol =
  | RUnitTupleSymbol                      (* for tuples *)
  | RUnitDocSymbol                        (* for documents *)
  | RUnitElemSymbol   of relem_symbol     (* for elements *)
  | RUnitAttrSymbol   of rattr_symbol     (* for attributes *)
  | RUnitTextSymbol                       (* for text nodes *)
  | RUnitSimpleSymbol of simple_symbol    (* for simple types *)
  | RUnitNSSymbol     of ns_symbol        (* for namespaces *)
  | RUnitAnonSymbol   of anon_symbol      (* for anonymous symbols *)

let equaltag s1 s2 =
  match (s1,s2) with
  | (RUnitDocSymbol, RUnitDocSymbol) ->
      true
  | (RUnitElemSymbol es1, RUnitElemSymbol es2) ->
      relem_equals es1 es2
  | (RUnitAttrSymbol as1, RUnitAttrSymbol as2) ->
      rattr_equals as1 as2
  | (RUnitSimpleSymbol ss1, RUnitSimpleSymbol ss2) ->
      simple_equals ss1 ss2
  | (RUnitNSSymbol ss1, RUnitNSSymbol ss2) ->
      ns_equals ss1 ss2
  | (RUnitAnonSymbol ss1, RUnitAnonSymbol ss2) ->
      anon_equals ss1 ss2
  | (RUnitTextSymbol, RUnitTextSymbol) ->
      true
  | _ ->
      false

let subtag s1 s2 =
  match (s1,s2) with
  | (RUnitDocSymbol, RUnitDocSymbol) ->
      true
  | (RUnitElemSymbol es1, RUnitElemSymbol es2) ->
      relem_subtag es1 es2
  | (RUnitAttrSymbol as1, RUnitAttrSymbol as2) ->
      rattr_subtag as1 as2
  | (RUnitTextSymbol, RUnitTextSymbol) ->
      true
  | (RUnitSimpleSymbol ss1, RUnitSimpleSymbol ss2) ->
      (simple_equals ss1 ss2) || (simple_equals ss2 anyatomicsym)
  | (RUnitNSSymbol ss1, RUnitNSSymbol ss2) ->
      ns_equals ss1 ss2
  | (RUnitAnonSymbol ss1, RUnitAnonSymbol ss2) ->
      anon_equals ss1 ss2
  | _ ->
      false

let label_string sym =
  match sym with
  | RUnitTupleSymbol ->
      "tuple"
  | RUnitDocSymbol ->
      "document"
  | RUnitElemSymbol esym ->
      "element " ^ relem_string esym
  | RUnitAttrSymbol asym ->
      "attribute " ^ rattr_string asym
  | RUnitTextSymbol -> 
      "text"
  | RUnitSimpleSymbol ssym ->
      simple_string ssym
  | RUnitNSSymbol ssym ->
      ns_string ssym
  | RUnitAnonSymbol ssym ->
      anon_string ssym

let relem_symbol_of_unit_symbol rus   =
  match rus with
  | RUnitElemSymbol s -> s
  | _ -> raise (Query(Internal_Error "Casting an non element symbol to element symbol"))

let rattr_symbol_of_unit_symbol rus   = 
  match rus with
  | RUnitAttrSymbol s -> s
  | _ -> raise (Query(Internal_Error "Casting an non attribute symbol to attribute symbol"))

let rsimple_symbol_of_unit_symbol rus = 
  match rus with
  | RUnitSimpleSymbol s -> s
  | _ -> raise (Query(Internal_Error "Casting an non simple symbol to simple symbol"))

let rns_symbol_of_unit_symbol rus     = 
  match rus with
  | RUnitNSSymbol s -> s
  | _ -> raise (Query(Internal_Error "Casting an non namespace symbol to namespace symbol"))

let ranon_symbol_of_unit_symbol rus   = 
  match rus with
  | RUnitAnonSymbol s -> s
  | _ -> raise (Query(Internal_Error "Casting an non anonymous symbol to anonymous symbol"))


(**************)
(* Statistics *)
(**************)

type sym_stats =
    { unresolved_element_names : (int * int * int);
      unresolved_attribute_names : (int * int * int);
      unresolved_type_names : (int * int * int);
      resolved_anon_names : (int * int * int);
      resolved_element_names : (int * int * int);
      resolved_attribute_names : (int * int * int);
      resolved_type_names : (int * int * int) }

let  stats () =
  { unresolved_element_names = UnresolvedQNamePool.pool_size uelem_pool;
    unresolved_attribute_names = UnresolvedQNamePool.pool_size uattr_pool;
    unresolved_type_names = UnresolvedQNamePool.pool_size ugroup_pool;
    resolved_anon_names = ResolvedQNamePool.pool_size rname_pool;
    resolved_element_names = ResolvedQNamePool.pool_size relem_pool;
    resolved_attribute_names = ResolvedQNamePool.pool_size rattr_pool;
    resolved_type_names = ResolvedQNamePool.pool_size rgroup_pool }
  








