(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: depend.ml,v 1.1 2002/06/27 17:40:26 gsur Exp $ *)

open Error
open Xasyn
open Typeutil

(* This module contain post-parsing checking on the AST.
   Constraints on declarations:
   - all used functions are defined
   - all variables are bound

   Constraints on expressions:
   - appropriate variable dependencies in for/let expressions

   Constraints on types:
   - all group variables are defined
   - recursive groups appear in guarded position
*)

(* Functions to check reachable components within a group *)

let rec reachable_names_aux sds m names =
  match m.pmodel_desc with
  | TAttrRef _ ->
      names
  | TAttrLocal (_,tm) ->
      begin
	match tm.ptyped_model_desc with
	| TAnonymousType m' ->
	    reachable_names_aux sds m' names
      end
  | TElemRef _ ->
      names
  | TElemLocal (_,_,tm) ->
      begin
	match tm.ptyped_model_desc with
	| TAnonymousType m' ->
	    reachable_names_aux sds m' names
      end
  | TGroupRef tn ->
      if List.exists (fun x -> x = tn) names then names else tn :: names
  | TBound (m',_,_) ->
      reachable_names_aux sds m' names
  | TComma (m1,m2) ->
      (reachable_names_aux sds m1 names) @ (reachable_names_aux sds m2 names)
  | TEmpty ->
      []
  | TBar (m1,m2) ->
      (reachable_names_aux sds m1 names) @ (reachable_names_aux sds m2 names)
  | TNone ->
      []
  | TAmp (m1,m2) ->
      (reachable_names_aux sds m1 names) @ (reachable_names_aux sds m2 names)
  | TBase _ ->
      []

let reachable_names sds m =
  reachable_names_aux sds m []

(* accesses local components *)

let rec reachable_local_names_aux sds m names =
  match m.pmodel_desc with
  | TAttrRef _ ->
      names
  | TAttrLocal _ ->
      names
  | TElemRef _ ->
      names
  | TElemLocal _ ->
      names
  | TGroupRef tn ->
      if List.exists (fun x -> x = tn) names then names else tn :: names
  | TBound (m',_,_) ->
      reachable_local_names_aux sds m' names
  | TComma (m1,m2) ->
      (reachable_local_names_aux sds m1 names) @ (reachable_local_names_aux sds m2 names)
  | TEmpty ->
      []
  | TBar (m1,m2) ->
      (reachable_local_names_aux sds m1 names) @ (reachable_local_names_aux sds m2 names)
  | TNone ->
      []
  | TAmp (m1,m2) ->
      (reachable_local_names_aux sds m1 names) @ (reachable_local_names_aux sds m2 names)
  | TBase _ ->
      []

let reachable_local_names sds m =
  reachable_local_names_aux sds m []

(* Check recursion criterias *)

let is_recursive_group sds tn =
  let m = get_model_in_group sds tn in
  let rn = reachable_names sds m in
  List.exists (fun x -> x = tn) rn

let is_guarded_group sds tn =
  let m = get_model_in_group sds tn in
  let lrn = reachable_local_names sds m in
  not(List.exists (fun x -> x = tn) lrn)
