(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2001, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mapping.ml,v 1.54 2003/06/30 02:58:19 simeon Exp $ *)

open Namespace
open Sym
open Resolve
open Datatypes
open Xquery_ast
open Xquery_astutil
open Maputil
open Error
open Norm_context

open Format

(********************)
(* Helper functions *)
(********************)

(* Special text nodes operations *)

(* Note:
     Well.. now that text is different than string, we need to have
     text as part of the core!
   - Jerome
 *)
(* Type models for XPath *)

let mkvar var eh fi =
  fmkcexpr (CEVar var) None (Some eh) fi

let mkvar_none var fi =
  fmkcexpr (CEVar var) None None fi

(* Extract all variable bindings from a FLWR expressions *)
(* Note: This is used to build the tuple in a FLWR with order by clause.
   - Jerome *)

let rec all_flwr_bindings fl_expr_list =
  match fl_expr_list with
  | [] -> []
  | fl_expr ::  fl_expr_list' ->
      begin
	match fl_expr.pfl_desc with
	| ELet (_, v1, _) ->
	    let prev_bindings = all_flwr_bindings fl_expr_list' in
	    if (List.mem v1 prev_bindings)
	    then prev_bindings
	    else v1 :: prev_bindings
	| EFor (_, v1, None, _) ->
	    let prev_bindings = all_flwr_bindings fl_expr_list' in
	    if (List.mem v1 prev_bindings)
	    then prev_bindings
	    else v1 :: prev_bindings
	| EFor (_, v1, (Some v2), _) ->
	    let prev_bindings = all_flwr_bindings fl_expr_list' in
	    let prev_bindings' =
	      if (List.mem v1 prev_bindings)
	      then prev_bindings
	      else v1 :: prev_bindings
	    in
	    if (List.mem v2 prev_bindings')
	    then prev_bindings'
	    else v2 :: prev_bindings'
      end

(**********************)
(* Some useful macros *)
(**********************)

let cexpr_true =
  fmkcexpr (CEApp ((fn_uri,"true"),[])) None None Finfo.bogus
let cexpr_false =
  fmkcexpr (CEApp ((fn_uri,"false"),[])) None None Finfo.bogus

(*********************************************************)
(* Mapping functions                                     *)
(* They encode the mapping rules appears in mapping.mli  *)
(*********************************************************)

(*******************************)
(* Mapping context of a module *)
(*******************************)

(* Note:
     map_context takes a context and an expression map the expression
     and return the updated context
   - Jerome *)

let map_context_decl (nsenv, impt_types) c =
  match c.pcontext_decl_desc with 
  | ENamespaceDecl (ncn, uri) -> 
      let nsenv' = Namespace.add_ns nsenv (NSPrefix ncn) uri in
      (nsenv', impt_types)
  | EDefaultElementNamespaceDecl uri -> 
      let nsenv' = Namespace.add_ns nsenv NSDefaultElementPrefix uri in
      (nsenv', impt_types)
  | EDefaultFunctionNamespaceDecl uri -> 
      let nsenv' = Namespace.add_ns nsenv NSDefaultFunctionPrefix uri in
      (nsenv', impt_types)
  | ESchemaDecl (Some prefix, uri, _) ->
      let xml_schema = Parseutil.parseXMLSchema uri in
      let (ncn, uri, types_in_schema) = Schemamapping.map_xml_schema (NSPrefix prefix) xml_schema in
      let nsenv' = Namespace.add_schema_ns (Namespace.add_ns nsenv ncn uri) uri in
      (nsenv', types_in_schema @ impt_types)
  | ESchemaDecl (None, uri, _) ->
      raise (Query (Prototype "Schema import without a namespace prefix not implemented yet."))
  | EDefaultSchemaDecl uri ->
      let xml_schema = Parseutil.parseXMLSchema uri in
      let (ncn, uri, types_in_schema) = Schemamapping.map_xml_schema NSDefaultElementPrefix xml_schema in
      let nsenv' = Namespace.add_schema_ns (Namespace.add_ns nsenv ncn uri) uri in
      (nsenv', types_in_schema @ impt_types)

let map_context cdl =
  List.fold_left map_context_decl (Namespace.empty_nsenv (), []) cdl

let map_context_with_env cdl nsenv =
  List.fold_left map_context_decl (nsenv, []) cdl

(**************************************)
(* Mapping inline global declarations *)
(**************************************)

let map_sort nsenv sds =
  let s = Schema.create_schema nsenv None None sds in
   (* Note:
        This test if just for consistency checking for now but
        compiled schema should replace schemas in the environment.
      - Jerome *)
  ignore(Schema.compile_schema s); s

let map_sort_with_env nsenv schema sds =
  let s = Schema.create_schema_with_env nsenv schema None None sds in
  ignore(Schema.compile_schema s);
  (* Note:
     This test if just for consistency checking for now but
     compiled schema should replace schemas in the environment.
     - Jerome *)
  s

(* map_model, map_typed_model, and map_sort_decl resolve
   all QNames that occur in types to expanded QNames *)

let rec map_model norm_context m = 
  let nsr = nsenv_from_norm_context norm_context in
  let schema = schema_from_norm_context norm_context in 
  match m.pmodel_desc with
  | TAttrRef attr_sym ->
      let rattr_sym = resolve_attr_symbol nsr attr_sym in
      Schema.attribute_exists schema rattr_sym;  (* Just checks schema consistency here - Jerome *)
      fmkmodel (TAttrRef (RAttrSym rattr_sym)) m.pmodel_loc
  | TAttrLocal (attr_sym, tm) ->
      let rattr_sym = resolve_attr_symbol nsr attr_sym in
      fmkmodel (TAttrLocal (RAttrSym rattr_sym, map_typed_model norm_context tm)) m.pmodel_loc
  | TElemRef elem_sym ->
      let relem_sym = resolve_elem_symbol nsr elem_sym in
      Schema.element_exists schema relem_sym;  (* Just checks schema consistency here - Jerome *)
      fmkmodel (TElemRef (RElemSym relem_sym)) m.pmodel_loc
  | TElemLocal (elem_sym, deriv, tm) ->
      let relem_sym = resolve_elem_symbol nsr elem_sym in
      fmkmodel (TElemLocal (RElemSym relem_sym, deriv, map_typed_model norm_context tm)) m.pmodel_loc
  | TGroupRef group_sym ->
      let rgroup_sym = resolve_group_symbol nsr group_sym in
      Schema.group_exists schema rgroup_sym;  (* Just checks schema consistency here - Jerome *)
      fmkmodel (TGroupRef (RGroupSym rgroup_sym)) m.pmodel_loc
  | TBound (m, b1, b2) ->
      fmkmodel (TBound(map_model norm_context m, b1, b2)) m.pmodel_loc
  | TComma (m1, m2) ->
      fmkmodel (TComma(map_model norm_context m1, map_model norm_context m2)) m.pmodel_loc
  | TBar (m1, m2) ->
      fmkmodel (TBar(map_model norm_context m1, map_model norm_context m2)) m.pmodel_loc
  | TAmp (m1, m2) ->
      fmkmodel (TAmp(map_model norm_context m1, map_model norm_context m2)) m.pmodel_loc
  | TEmpty ->
      m
  | TNone ->
      m
  | TBase b ->
      m
  | TText ->
      m
  | TDocument m ->
      fmkmodel (TDocument (map_model norm_context m)) m.pmodel_loc
  | TTuple t ->
      let new_tuple_row (rname1,m1) = (rname1,(map_model norm_context m1)) in
      fmkmodel (TTuple (List.map new_tuple_row t)) m.pmodel_loc

and map_typed_model norm_context tm = 
  match tm.ptyped_model_desc with 
  | TAnonymousType m ->
      fmktyped_model (TAnonymousType (map_model norm_context m)) tm.ptyped_model_loc

(*************************)
(* Mapping for Datatypes *)
(*************************)

(* Resolving name information in datatypes *)

let resolve_datatype nsr schema dt =
  match dt.pdatatype_desc with
  | (dtk,occ) ->
      let mapped_dtk =
	match dtk with
	| DTAttrRef None ->
	    dtk
	| DTAttrRef (Some attr_sym) ->
	    let rattr_sym = resolve_attr_symbol nsr attr_sym in
	    Schema.attribute_exists schema rattr_sym;  (* Just checks schema consistency here - Jerome *)
	    DTAttrRef (Some (RAttrSym rattr_sym))
	| DTElemRef None ->
	    dtk
	| DTElemRef (Some elem_sym) ->
	    let relem_sym = resolve_elem_symbol nsr elem_sym in
	    Schema.element_exists schema relem_sym;  (* Just checks schema consistency here - Jerome *)
	    DTElemRef (Some (RElemSym relem_sym))
	| DTTypeRef group_sym ->
	    let rgroup_sym = resolve_group_symbol nsr group_sym in
	    Schema.group_exists schema rgroup_sym;  (* Just checks schema consistency here - Jerome *)
	    DTTypeRef (RGroupSym rgroup_sym)
	| DTNode ->
	    dtk
	| DTItem ->
	    dtk
	| DTNumeric ->
	    dtk
	| DTAtomic ->
	    dtk
	| DTText ->
	    dtk
	| DTEmpty ->
	    dtk
	| DTBase st ->
	    dtk
	| DTDocument None ->
            dtk
	| DTDocument (Some elem_sym) ->
	    let relem_sym = resolve_elem_symbol nsr elem_sym in
	    Schema.element_exists schema relem_sym;  (* Just checks schema consistency here - Jerome *)
	    DTDocument (Some (RElemSym relem_sym))
	| DTDocumentOfType (group_sym) ->
	    let rgroup_sym = resolve_group_symbol nsr group_sym in
	    Schema.group_exists schema rgroup_sym;  (* Just checks schema consistency here - Jerome *)
	    DTDocumentOfType (RGroupSym rgroup_sym)
	| DTElemOfType group_sym ->
	    let rgroup_sym = resolve_group_symbol nsr group_sym in
	    Schema.group_exists schema rgroup_sym;  (* Just checks schema consistency here - Jerome *)
	    DTElemOfType (RGroupSym rgroup_sym)
	| DTElemOfBase st ->
	    dtk
      in
      fmkdatatype (mapped_dtk,occ) dt.pdatatype_loc

let resolve_optional_datatype norm_context odt = 
  let nsenv = nsenv_from_norm_context norm_context in
  let schema = schema_from_norm_context norm_context in
  match odt with
  | None -> None
  | Some dt -> Some (resolve_datatype nsenv schema dt)

(* Mapping function on already resolved datatypes *)

let map_resolved_datatype_to_model dt =
  (* First some wildcard short-cuts *)
  let elem_wild_qname = anyrelem in
  (* Then map the datatype kind *)
  match dt.pdatatype_desc with
  | (dtk,occ) ->
      let mapped_dtk =
	match dtk with
	| DTAttrRef None ->
	    Typeutil.anyattr_model
	| DTAttrRef (Some attr_sym) ->
	    fmkmodel (TAttrRef attr_sym) dt.pdatatype_loc
	| DTElemRef None ->
	    Typeutil.anyelem_model
	| DTElemRef (Some elem_sym) ->
	    fmkmodel (TElemRef elem_sym) dt.pdatatype_loc
	| DTTypeRef group_sym ->
	    fmkmodel (TGroupRef group_sym) dt.pdatatype_loc
	| DTNode ->
	    Typeutil.anynode_model
	| DTItem ->
	    Typeutil.anyitem_model
	| DTAtomic ->
	    Typeutil.anyatomic_model
	| DTNumeric ->
	    Typeutil.anynumeric_model
	| DTText ->
	    fmkmodel (TText) dt.pdatatype_loc
	| DTEmpty ->
	    fmkmodel (TEmpty) Finfo.bogus 
	| DTBase st ->
	    fmkmodel (TBase st) dt.pdatatype_loc
	| DTDocument None ->
	    let group_model = Typeutil.anyelem_model in
	    fmkmodel (TDocument group_model) dt.pdatatype_loc 
	| DTDocument (Some elem_sym) ->
	    let elem_model = fmkmodel (TElemRef elem_sym) dt.pdatatype_loc in
	    fmkmodel (TDocument elem_model) dt.pdatatype_loc 
	| DTDocumentOfType group_sym ->
	    let relem_sym = elem_wild_qname in
	    let group_model = fmkmodel (TGroupRef group_sym) dt.pdatatype_loc in
	    let elem_model = fmkmodel (TElemLocal (RElemSym relem_sym,None, (fmktyped_model (TAnonymousType group_model) Finfo.bogus))) dt.pdatatype_loc in
            fmkmodel (TDocument elem_model) dt.pdatatype_loc
	| DTElemOfType group_sym ->
	    let relem_sym = elem_wild_qname in
	    let group_model = fmkmodel (TGroupRef group_sym) dt.pdatatype_loc in
	    fmkmodel (TElemLocal (RElemSym relem_sym,None, (fmktyped_model (TAnonymousType group_model) Finfo.bogus))) dt.pdatatype_loc
	| DTElemOfBase st ->
	    let relem_sym = elem_wild_qname in
            let base_model  = fmkmodel (TBase st) dt.pdatatype_loc in
	    fmkmodel (TElemLocal (RElemSym relem_sym, None, (fmktyped_model (TAnonymousType base_model) Finfo.bogus))) dt.pdatatype_loc
      in
      (* Then map the occurence indicator *)
      match occ with
      |	None ->
	  mapped_dtk
      |	Some (m,n) ->
	  fmkmodel (TBound (mapped_dtk,m,n))dt.pdatatype_loc

let map_datatype_to_model nsenv schema dt =
  map_resolved_datatype_to_model (resolve_datatype nsenv schema dt)

let map_optdatatype_to_model nsenv schema odt =
  match odt with 
  | None -> None
  | Some dt -> Some (map_resolved_datatype_to_model (resolve_datatype nsenv schema dt))

(**********************)
(* Mapping expressions *)
(**********************)

(***********************************************************************************)
(* Goal: transversing the Full XQuery expression AST                               *)
(* map_expr:                                                                       *)
(* - match the Full expression                                                     *)
(* - project its parts                                                             *)
(* - apply map_expr on each parts -> obtaining parts'                              *)
(* - construct intermediate mapping structure, defined in maputil.mli, from parts' *)
(* - apply the mapping function                                                    *)
(***********************************************************************************)

let remove_boundary_whitespace ce = 
  if (!Conf.xquery_whitespace = false) then
    match ce.pcexpr_desc with 
    | (CEText s) -> 
	if (Whitespace.whitespace_only s) then (fmkcexpr (CEText "") None None ce.pcexpr_loc)
	else ce
    | _ -> ce 
  else ce

(***************************************************************)
(* Remove empty text nodes at head and tail of expression list *)
(***************************************************************)

let remove_empty_textnodes celist = 
  let celist' = 
    match celist with
    | [] -> []
    | cl -> 
	let first = List.nth cl 0 in
	begin
	  match first.pcexpr_desc with 
	  | CEText s -> if (s = "") then (List.tl cl) else cl
	  | _ -> cl
	end
  in
  let revcelist' = List.rev celist' in
  let revcelist'' = 
    match revcelist' with 
    | [] -> []
    | cl -> 
	let first = List.nth cl 0 in
	begin
	  match first.pcexpr_desc with 
	  | CEText s -> if (s = "") then (List.tl cl) else cl
	  | _ -> cl
	end
  in List.rev revcelist''

(******************************)
(* aux functions for map_expr *)
(******************************)

(* Note:
     Several 'special' functions are mapped to core expressions or
     special variables.
   - Jerome *)

(* Note: [2002/09/21]
     Function normalization used to be done based on prefixes, which
     is wrong. Namespace resolution needs to happen before.

     Hopefully, this is a local change, and leaves the rest of the
     normalization untouched.
 - Jerome *)

let rec map_eapp norm_context frfname celist e fi =
  (* Gets the prefix, uri and local name of the function *)
  let (prefix, rfname) = frfname in
  let (uri, lname) = rfname in
  let cempty = fmkcexpr CEEmpty None None fi in
  let arity = List.length celist in
  (* Is the function in the fn: namespace? *)
  if (uri = fn_uri)
  then
    begin
      match (lname, arity) with
	  (* fn:error(Expr) ==
             fn:error(fn:string(Expr)) 
           *)
	(* HACK!!! - Jerome and Amelie *)
      | ("document", 1) ->
	  fmkcexpr (CEApp ((fn_uri, "document"),celist)) None (Some e) fi
      | ("error", 0) ->
	  fmkcexpr (CEError cempty) None (Some e) fi
      | ("error", 1) ->
	  fmkcexpr (CEError (List.hd celist)) None (Some e) fi
      | ("context-item", 0) ->
	  mkvar dot e fi
      | ("position", 0) ->
	  mkvar position e fi
      | ("last", 0) ->
	  mkvar last e fi
      |	("concat", _) -> 
        (* fn:concat is variadic; convert to binary applications of fn:concat here *)
	let fn_concat ce1 ce2 = 
	  map_function_app norm_context (fn_uri, "concat") [ce1; ce2] (Some e) fi 
	in
        List.fold_left fn_concat (fmkcexpr (CEScalar (AtomicString "")) None (Some e) fi) (celist) 
      |	("unordered", 1) ->
        fmkcexpr (CEUnordered (List.hd celist)) None (Some e) fi 
      | ("not", 1) ->
	  begin
	    match celist with
	    | [ce] ->
		let ce' = map_effective_boolean_value ce (Some e) fi in
		fmkcexpr (CEApp ((fn_uri, "not"), [ce'])) None (Some e) fi
	    | _ ->
		raise (Error.incorrect_arg_count (string_of_rqname rfname) (List.length celist) 1 fi)
	  end
      |	(lname, arity) ->
	  let celist' = 
	    begin
	      match (lname, arity) with 
	      |	("string", 0) ->
 	  (* fn:string() is normalized to fn:string($dot) *)
		  [fmkcexpr (CEVar dot) None (Some e) fi]
	      |	("string", 1) ->
		  [List.hd celist]
	      |	("name",0) ->
	  (* fn:name() is normalized to fn:name($dot) *)
		  [fmkcexpr (CEVar dot) None (Some e) fi]
	      |	("lang",1) ->
	  (* fn:lang($textlang) is normalized to fn:name($dot,$textlang) *)
		  (fmkcexpr (CEVar dot) None (Some e) fi) :: celist
	      |	("local-name",0) ->
	  (* fn:local-name() is normalized to fn:local-name($dot) *)
		  [fmkcexpr (CEVar dot) None (Some e) fi]
	      |	("namespace-uri",0) ->
	  (* fn:namespace-uri() is normalized to fn:namespace-uri($dot) *)
		  [fmkcexpr (CEVar dot) None (Some e) fi]
	      | _ ->
		  celist
	    end
	  in map_function_app norm_context rfname celist' (Some e) fi
    end
      (* Is the function in the op: namespace? *)
  else if (uri = glx_uri)
  then
    begin
      match (lname,arity) with
	(* HACK!!! - Jerome and Amelie *)
      | ("document-sax", 1) ->
	  fmkcexpr (CEApp ((glx_uri, "document-sax"),celist)) None (Some e) fi
      | ("keyref",2) ->
	  let (ce1,ce2) =
	    match celist with
	    | [ce1;ce2] -> (ce1,ce2)
	    | _ ->
		raise (Query (Parameter_Mismatch "glx:key-ref function should have two parameters"))
	  in
	  let kname =
	    match ce1.pcexpr_desc with
	    | CEScalar (AtomicString kname) ->
		kname
	    | _ ->
		raise (Query (Parameter_Mismatch "glx:key-ref first parameter must be a string literal"))
	  in
	  fmkcexpr (CEKeyRef (kname, ce2)) None (Some e) fi
      | _ ->
	  map_function_app norm_context rfname celist (Some e) fi
    end
  else if (uri = op_uri)
  then
    begin
      match (lname,arity) with
      | ("concatenate",2) ->
	  begin
	    match celist with
	    | ce1 :: ce2 :: [] ->
		fmkcexpr (CESeq (ce1, ce2)) None (Some e) fi
	    | _ ->
		raise (Query (Parameter_Mismatch "concatenate function should have two parameters"))
	  end
      | _ ->
	  map_function_app norm_context rfname celist (Some e) fi
    end
  (* Is the function in the xs: or xsd: namespace? *)
  else if ((uri = xs_uri) || (uri = xsd_uri))
  then
    begin
      if (List.length celist) != 1 then
	raise (Error.incorrect_arg_count (string_of_rqname rfname) (List.length celist) 1 fi)
      else
	let simplet =
	  try
	    Datatypes.lookup_bltin_type (xsd_prefix,lname)
	  with
	  | _ ->
	      raise (Query (Mapping_Failure ((string_of_rqname rfname) ^ " used in constructor is not a built-in simple type")))
	in
	map_to_cast (List.nth celist 0) (fmkmodel (TBase simplet) fi) (Some e) fi
    end
  else
    map_function_app norm_context rfname celist (Some e) fi

(* 
  [ attribute QName { Expr } ]
  ==
  if QName is globally defined attribute then
    let attribute QName $v := [ Expr ] return $v
  else [ Expr ]
*)
and map_eattfixed norm_context rs celist e fi = 
  let nsenv = nsenv_from_norm_context norm_context in
  let schema = schema_from_norm_context norm_context in
  let (v, var) = gen_new_var_expr (Some e) fi in
  let (uri, _) = rattr_name rs in

  (* We remove empty text nodes at beginning and end of attribute content.
     Any empty text nodes within the sequence are significant.  *)

  let celist' = remove_empty_textnodes celist in
  let ceseq = map_to_sequence (Some e) fi celist' in
  let attr_expr = fmkcexpr (CEAttr (rs, ceseq)) None (Some e) fi in
  if ((Namespace.is_schema_ns nsenv uri) && Schema.is_unique_global_attribute schema rs) then
    fmkcexpr (CELet(Some(fmkdatatype (DTAttrRef (Some (RAttrSym rs)), None) fi), v, attr_expr, var)) None (Some e) fi
  else
    attr_expr

(* 
  [ element QName { Expr } ]
  ==
  if QName is globally defined element then
    let element QName $v := [ Expr ] return $v
  else [ Expr ]
*)
and map_eelemfixed norm_context rs cattlist cchildlist e fi =
  let nsenv = nsenv_from_norm_context norm_context in
  let schema = schema_from_norm_context norm_context in

  let (v, var) = gen_new_var_expr (Some e) fi in
  let (uri, _) = relem_name rs in

  (* We first convert boundary whitespace into empty text nodes, then
     remove empty text nodes at beginning and end of element content.
     Any empty text nodes within the sequence are significant.  *)

  let cchildlist' = List.map remove_boundary_whitespace cchildlist in
  let cchildlist'' = remove_empty_textnodes cchildlist' in
  let elem_expr = fmkcexpr (CEElem (rs, map_to_sequence (Some e) fi (cattlist @ cchildlist''))) None (Some e) fi in
  if (Namespace.is_schema_ns nsenv uri && Schema.is_unique_global_element schema rs) then
    fmkcexpr (CELet(Some(fmkdatatype (DTElemRef (Some (RElemSym rs)), None) fi), v, elem_expr, var)) None (Some e) fi
  else
    elem_expr

(* 

Here's what the spec says:

  3.8.3 Order By and Return Clauses

  The process of evaluating and comparing the orderspecs is based on
  the following rules:

  * Atomization is applied to the result of the expression in each
    orderspec. If the result of atomization is neither a single atomic
    value nor an empty sequence, a type error is raised.

  * If the value of an orderspec has the type xdt:untypedAtomic, it is
    cast to the type xs:string.

  * Each orderspec must return values of the same type for all tuples
    in the tuple stream, and this type must be a (possibly optional)
    atomic type for which the gt operator is defined--otherwise, a
    dynamic error is raised.

Here's what we implement:

  * Atomization is applied to the result of the expression in each
    orderspec. If the result of atomization is neither a single atomic
    value nor an empty sequence, a type error is raised.

  * Currently, do not cast xdt:untypedAtomic to xs:string

*)

and map_sortkind norm_context v replace_fun fi stable ce (key, kind, empty_kind) = 
  let key' = replace_fun (map_expr norm_context key) in 
  let key'' = atomization key' (Some key) fi in
  (fmkcexpr (CESort(v, ce, stable, key'', kind, empty_kind)) None (Some key) fi)

and map_pattern norm_context p = 
  let nsenv = nsenv_from_norm_context norm_context in
  let schema = schema_from_norm_context norm_context in
  match p.ppattern_desc with
  | Case dt ->
      fmkpattern (Case (resolve_datatype nsenv schema dt)) p.ppattern_loc
  | Default ->
      p 

and map_pat_expr norm_context pat_expr = 
  let (p, ovn, expr) = pat_expr in
  let p' = map_pattern norm_context p in 
  let ce = map_expr norm_context expr in
  (p', ovn, ce)

(************)
(* map_expr *)
(************)

and map_expr norm_context e =
  let desc = e.pexpr_desc
  and fi = e.pexpr_loc in
  let cempty = (fmkcexpr CEEmpty None None fi) in
  let posvar = mkvar position e fi in
  let nsenv = nsenv_from_norm_context norm_context in 
  let schema = schema_from_norm_context norm_context in 
  let cexpr_result =
    begin
      match desc with 
      | EApp (fn, elist) ->
	  (* First resolves the function name - Jerome *)
	  let frfname = Resolve.resolve_function_qname nsenv fn in
	  let celist = map_expr_list norm_context elist in
	  map_eapp norm_context frfname celist e fi
      | EAttrFixed (attr_symbol, elist) ->
	  let celist = map_expr_list norm_context elist in
	  let rs = resolve_attr_symbol nsenv attr_symbol in
	  map_eattfixed norm_context rs celist e fi
      | EAttrComputed (e, elist) ->
	  let (v, var) = gen_new_var_expr (Some e) fi in
	  let celist = map_expr_list norm_context elist in
	  let ce = map_expr norm_context  e in
	  let ceseq = map_to_sequence (Some e) fi celist in
	  (* Attribute name is cast to a QName during query evaluation
	  -- should probably happen during normalization *)
	  fmkcexpr (CEAnyAttr (ce, ceseq)) None (Some e) fi

(*	  let attr_expr = fmkcexpr (CEAnyAttr (var, ceseq)) None (Some e) fi in 
	  fmkcexpr (CELet(Some(fmkdatatype (DTBase ATQName, None) fi), v, ce, attr_expr)) None (Some e) fi *)
      | ETreat (e, dt) ->
	  let ce = map_expr norm_context e in
	  let m' = map_datatype_to_model nsenv schema dt in 
          (* TODO Mary -- normalize treat as typeswitch *)
	  fmkcexpr (CETreat(ce,m')) None (Some e) fi
      | EValidate (e, dt) ->
	  let ce = map_expr norm_context e in
	  let m' = map_datatype_to_model nsenv schema dt in 
	  fmkcexpr (CEValidate(ce,m')) None (Some e) fi
      | ECast (e, dt) ->
          (* Argument to cast can be a node sequence, which is mapped to a sequence of atomic values *)
	  let ce = map_expr norm_context e in
	  let m' = map_datatype_to_model nsenv schema dt in 
	  if (not (Typeutil.is_anyatomic_optional schema m')) then
	    raise (Query (Expr_Error (fi, ("Type in cast expression is not an atomic or empty type"))))
	  else
	    map_to_cast ce m' (Some e) fi
      | ECastable (e, dt) ->
          (* Argument to cast can be a node sequence, which is mapped to a sequence of atomic values *)
	  let ce = map_expr norm_context e in
	  let m' = map_datatype_to_model nsenv schema dt in 
	  if (not (Typeutil.is_anyatomic_optional schema m')) then
	    raise (Query (Expr_Error (fi, ("Type in castable expression is not an atomic or empty type"))))
	  else
	    map_to_castable ce m' (Some e) fi
      | ESubtype (dt1,dt2) ->
	  let m1' = map_datatype_to_model nsenv schema dt1 in 
	  let m2' = map_datatype_to_model nsenv schema dt2 in 
	  fmkcexpr (CESubtype(m1',m2')) None (Some e) fi
      | EIntersects (dt1,dt2) ->
	  let m1' = map_datatype_to_model nsenv schema dt1 in 
	  let m2' = map_datatype_to_model nsenv schema dt2 in 
	  fmkcexpr (CEIntersects(m1',m2')) None (Some e) fi
      | EBinaryOp (e1, bop, e2) ->
	  let ce1 = map_expr norm_context e1 in
	  let ce2 = map_expr norm_context e2 in
	  map_binary_operator bop ce1 ce2 (Some e) fi
      | EElemFixed (elem_symbol, attlist, childlist) -> 
	  let (new_nss, other_atts) = get_ns_attributes attlist in
	  let nsenv' = Namespace.add_all_ns nsenv new_nss in
	  let cattlist =  map_expr_list norm_context other_atts in
	  let cchildlist = map_expr_list norm_context childlist in
	  let rs = resolve_elem_symbol nsenv' elem_symbol in
	  map_eelemfixed norm_context rs cattlist cchildlist e fi
      | EElemComputed (e, childlist) ->
	  let (v, var) = gen_new_var_expr (Some e) fi in
	  let cchildlist = map_expr_list norm_context childlist in
	  let ce = map_expr norm_context e in
	  (* Element name is cast to a QName during query evaluation
	  -- should probably happen during normalization *)
	  fmkcexpr (CEAnyElem (ce, (map_to_sequence (Some e) fi cchildlist))) None (Some e) fi 
(*
	  let elem_expr = fmkcexpr (CEAnyElem (var, (map_to_sequence (Some e) fi cchildlist))) None (Some e) fi in
	  fmkcexpr (CELet(Some(fmkdatatype (DTBase ATQName, None) fi), v, ce, elem_expr)) None (Some e) fi
*)
	    (**********************************************)
	    (* Changed here to support order by! - Jerome *)
	    (**********************************************)

	    (* Note:
	         There are two very distinct normalization rules: one
	         for FLWR without order by (None), and one for FLWR
	         with order by (Some order)
	     - Jerome *)

      | EFLWR (fl_expr_list, where_expr_opt, None, ret_expr) ->
	  (* look inside the for or let expression list, apply map_expr recursively *)
	  let flwr_expr_list' = map_fl_list norm_context fl_expr_list in
	  let where_expr_opt' = 
	    begin
	      match where_expr_opt with
	      | None -> None
	      | Some e' -> let ce = (map_expr norm_context e') in Some ce
	    end 
	  in
	  let ret_expr' = map_expr norm_context ret_expr in
	  map_to_flwr flwr_expr_list' where_expr_opt' None ret_expr' (Some e) fi

        (*
        
          Normalization of FLWR expressions with order-by clause:
        
          [ for $v_1 in Expr_1, ..., $v_k := Expr_k
            where Expr_where
            order by Expr_order_1, ..., Expr_order_j
            return Expr_return 
          ]
             ===
        
           1. for-tuple $tuple_var in (
           2.   sort $tuple_var in (
           3.     ...
           4.     sort $tuple_var in (
           5.       sort $tuple_var in (
           6.         (for $v_1 in [ Expr_1 ] return
           7.           ...
           8.           for $v_k in [ Expr_k ] return 
           9.             if ( [ Expr_where ] ) then 
          10.               {-- Create a tuple: CECreateTuple --}
          11.               ((#v_1, $v_1) , ..., (#v_k, $v_k))
          12.              else ())
          13.       ) by [ Expr_order_j ] subst [$tuple_var#v_1/$v_1, ... $tuple_var#v_k/$v_k]
          14.     ) by [ Expr_order_{j-1} ] subst [$tuple_var#v_1/$v_1, ... $tuple_var#v_k/$v_k]
          15.    ...
          16.   ) by [ Expr_order_1 ] subst [$tuple_var#v_1/$v_1, ... $tuple_var#v_k/$v_k]
          17. ) 
          18. return 
          19.   [ Expr_return ] subst [$tuple_var#v_1/$v_1, ... $tuple_var#v_k/$v_k])
            
         
         Notes: 
           1. Tuples are really associative arrays, which map field
              names to field values.

              The expression $tuple_var#v_1 accesses the value
              associated with the field #v_1 in the tuple variable
              $tuple_var.

           2. Lines 6-12 above evalaute the original 'for' and 'where'
              expressions, and construct a sequence of tuples.

           3. Lines 2-5, 13-16 sort the sequence of tuples by the 'order-by'
              clause, using the internal sort-by expression.  Note
              that because the sort operator is stable, we first sort
              by the j-ary order expression (Expr_order_j), up to the
              primary order expression (Expr_order_1). 

              Obviously, we might want a sort-operator in the query
              execution plan that could handle multiple keys.

           4. The 'subst' function on Lines 13-19 (called replace_fun
              in the code) applies variable substitution.  Now that
              the inner for-expression (on lines 6-12) yields tuples,
              we have to replace all references to the original 'for'
              variables by expressions that access the corresponding
              field in the tuple variable.  For example:

              [ $x * 2 ] subst [$tuple_var#v_1/$x]
              === 
              $tuple_var#v_1 * 2


        *)
      | EFLWR (fl_expr_list, where_expr_opt, Some order, ret_expr) ->
	  (* look inside the for or let expression list, apply map_expr recursively *)
	  let flwr_expr_list' = map_fl_list norm_context fl_expr_list in
	  let tuple_vars = all_flwr_bindings fl_expr_list in
	  let build_tuple_field vname =
	    let field_expr = fmkcexpr (CEVar vname) None (Some e) fi
	    in (vname, field_expr)
	  in
	  let tuple_expr = 
	    fmkcexpr (CECreateTuple (List.map build_tuple_field tuple_vars)) None (Some e) fi
	  in
	  let where_expr_opt' = 
	    begin
	      match where_expr_opt with
	      | None -> None
	      | Some e' -> let ce = (map_expr norm_context e') in Some ce
	    end 
	  in
	  (* The intermediate table with tuple creation *)
	  let materialize_tuple_expr = map_to_flwr 
             flwr_expr_list' where_expr_opt' (Some order) tuple_expr (Some e) fi in
          (* Generate a new tuple variable *)
	  let (tuple_vname, tuple_var) = gen_new_var_expr (Some e) fi in 
	  let replace_variable_expression_fun expr vname =
	    let local_tuple_access_expr =
	      fmkcexpr (CEAccessTuple (tuple_var, vname)) None (Some e) fi
	    in
	    fst (Xquery_astutil.substitute_var local_tuple_access_expr vname expr)
	  in
	  let replace_fun expr =
	    List.fold_left replace_variable_expression_fun expr tuple_vars
	  in

	  (* Creating the sort by *)
	  let sorted_materialize_tuple_expr =
	    let (stable, sortspecs) = order in
	    let keys = List.rev sortspecs in
	    List.fold_left (map_sortkind norm_context tuple_vname replace_fun fi stable) materialize_tuple_expr keys
	  in

	  let intermediate_ret_expr' = map_expr norm_context ret_expr in
	  let real_return_expr =
	    List.fold_left replace_variable_expression_fun intermediate_ret_expr' tuple_vars
	  in
	  fmkcexpr (CEForTuple (tuple_vname, sorted_materialize_tuple_expr, real_return_expr)) None (Some e) fi
 

      | EIf (cond, e1, e2) ->
	  let cond' = map_expr norm_context cond in
	  let e1' = map_expr norm_context e1 in
	  let e2' = map_expr norm_context e2 in
	  map_to_CEIf cond' e1' e2' (Some e) fi
      | EList elist ->
	  let elist' = map_expr_list norm_context elist in
	  map_to_sequence (Some e) fi elist' 
      | EPath p -> 
	  map_path_expr norm_context e p
      | EScalar bv ->
	  fmkcexpr (CEScalar bv) None (Some e) fi
      | EPI (target,pi_content) ->
	  fmkcexpr (CEPI (target,pi_content)) None (Some e) fi
      | EText t ->
	  fmkcexpr (CEText t) None (Some e) fi
      | ETextComputed elist ->
	  let elist' = map_expr_list norm_context elist in
	  let e' = map_to_sequence (Some e) fi elist' in
	  fmkcexpr (CETextComputed e') None (Some e) fi
      | EDocument elist ->
	  let elist' = map_expr_list norm_context elist in
	  let e' = map_to_sequence (Some e) fi elist' in
	  fmkcexpr (CEDocument e') None (Some e) fi
      | ESelf ->
	  fmkcexpr (CEVar dot) None (Some e) fi
      | EParent ->
	  let ce = 
	    CEReverseAxis(Parent, CPNodeKindTest AnyKind)
	  in fmkcexpr ce None (Some e) fi
      | ESome (bdl, e2) ->
	  let e2' = map_expr norm_context e2 in 
	  let rec map_bindings bdl =
	    match bdl with
	    | [] -> e2'
	    | (odt,v,e1) :: obdl ->
		let result_expr = map_bindings obdl in
		let e1' = map_expr norm_context e1 in 
		let rodt = (resolve_optional_datatype norm_context odt) in
		map_to_CESome rodt v e1' result_expr (Some e) fi
	  in
	  map_bindings bdl

(* Version for some with a single binding - Jerome
	  let e1' = map_expr norm_context e1 in 
	  let e2' = map_expr norm_context e2 in 
	  let rodt = (resolve_optional_datatype norm_context odt) in
	  map_to_CESome rodt v e1' e2' (Some e) fi *)

      | EEvery (bdl, e2) ->
	  let e2' = map_expr norm_context e2 in 
	  let rec map_bindings bdl =
	    match bdl with
	    | [] -> e2'
	    | (odt,v,e1) :: obdl ->
		let result_expr = map_bindings obdl in
		let e1' = map_expr norm_context e1 in 
		let rodt = (resolve_optional_datatype norm_context odt) in
		map_to_CEEvery rodt v e1' result_expr (Some e) fi
	  in
	  map_bindings bdl

      | ESort (e1, order_by) ->
	  let for_e1 = fmkfl_expr (EFor (None,dot,None,e1)) fi in
	  let var_e1 = fmkexpr (EVar dot) fi in
	  let flwr_e1 = fmkexpr (EFLWR ([for_e1],None,Some order_by,var_e1)) fi in
	  map_expr norm_context flwr_e1

      | ETypeswitch (e1, pat_expr_list) ->
	  let e1' = map_expr norm_context e1 in
	  let pat_cexpr_list = List.map (map_pat_expr norm_context) pat_expr_list in
	  fmkcexpr (CETypeswitch (e1', pat_cexpr_list)) None (Some e) fi
      | EUnaryOp (uop, e1) ->
	  let ce = map_expr norm_context e1 in 
	  begin
	    match uop with 
          (* +e1 *)
	    | UEPlus ->
		map_unary_operator (op_uri, "numeric-unary-plus") ce (Some e) fi 
          (* -e1 *) 
	    | UEMinus ->
		map_unary_operator (op_uri, "numeric-unary-minus") ce (Some e) fi 
	  end	    
      | EUnordered e -> 
	  fmkcexpr (CEUnordered (map_expr norm_context e )) None (Some e) fi
      | EVar vn ->
	  mkvar vn e fi
      | EComment c ->
	  fmkcexpr (CEComment c) None (Some e) fi
      | ERoot ->
	  let frfname = (NSPrefix "fn",(fn_uri,"root")) in
	  let celist = [mkvar dot e fi] in
	  map_eapp norm_context frfname celist e fi

      | EInstanceOf (e,dt) ->
	  let e' = map_expr norm_context e in
	  let pat_cexpr_list =
	    [ (fmkpattern (Case (resolve_datatype nsenv schema dt)) Finfo.bogus, None, cexpr_true );
	      (fmkpattern (Default)                                 Finfo.bogus, None, cexpr_false) ] 
	  in
	  fmkcexpr (CETypeswitch (e', pat_cexpr_list)) None (Some e) fi
      | ERange (e1,e2) ->
	  (* Do like if it was a direct call to the built-in function op:to - Jerome *)
	  let frfname = (NSPrefix "op",(op_uri,"to")) in
	  let elist = [e1;e2] in
	  let celist = map_expr_list norm_context elist in
	  map_eapp norm_context frfname celist e fi
    end
  in
  begin
    cexpr_result.pcexpr_origin <- Some e;
    cexpr_result
  end

(*********************************)
(* helper functions for map_expr *)
(*********************************)
and map_fl_expr norm_context e  =
  let cempty = fmkcexpr CEEmpty None None Finfo.bogus in
  let schema = schema_from_norm_context norm_context in 
  match e.pfl_desc with
  | ELet (odt, v, e') ->
      let ce = (map_expr norm_context e') in
      let rodt = resolve_optional_datatype norm_context odt in
      fmkcexpr (CELet (rodt, v, ce, cempty)) None None e.pfl_loc 
  | EFor (odt, v, ovn, e') ->
      let ce = (map_expr norm_context e') in
      let rodt = resolve_optional_datatype norm_context odt in
      fmkcexpr (CEFor (rodt, v, ovn, ce, cempty)) None None e.pfl_loc 

and map_expr_list norm_context es = 
  List.map (map_expr norm_context) es

and map_fl_list norm_context es = 
  let f e  cexprs = 
    let ce' =  map_fl_expr norm_context e
    in ce' :: cexprs 
  in  List.fold_right f es [] 


(*************************)
(* Path Expressions      *)
(*************************)

and map_path_expr norm_context e p  =
  let desc = e.pexpr_desc
  and fi = e.pexpr_loc in
  let dotvar = mkvar dot e fi in
  let posvar = mkvar position e fi in
  let lastvar = mkvar last e fi in
  let seqvar = mkvar sequence e fi in
  let nsenv = nsenv_from_norm_context norm_context in 
  match p with
  (* 

   [StepExpr1 "/" StepExpr2]Expr   ==
     fs:distinct-doc-order (
       let $fs:sequence := fs:distinct-doc-order( [StepExpr1]Expr ) return
       let $fs:last := fn:count($fs:sequence) return
       for $fs:dot at $fs:position in $fs:sequence return
         [StepExpr2]Expr
    )

   *)
  | PSlash (e1, e2) -> 
      let ce1 = map_expr norm_context e1 in
      let ce2 = map_expr norm_context e2 in
      let count_expr = fmkcexpr (CEApp((fn_uri, "count"), [seqvar])) None (Some e) fi in 
      let for_expr = fmkcexpr (CEFor(None, dot, Some position, seqvar, ce2)) None (Some e) fi in
      let let_expr2 = fmkcexpr (CELet(None, last, count_expr, for_expr)) None (Some e) fi in
     (* following if-statement checks if the commandline 
      * options say something about duplicate elimination
      * or sorting by document order
      *)
      if (!Conf.do_sbdo) then  (* ddo meta-instruction inserted -- default *)
      	let docorder_expr = fmkcexpr (CEApp((glx_uri, "distinct-docorder"), [ce1])) None (Some e) fi in 
      	let let_expr1 = fmkcexpr (CELet(None, sequence, docorder_expr, let_expr2)) None (Some e) fi in 
      	fmkcexpr (CEApp((glx_uri, "distinct-docorder"), [let_expr1])) None (Some e) fi 
      else                     (* ddo meta-instruction NOT inserted *)
      	fmkcexpr (CELet(None, sequence, ce1, let_expr2)) None (Some e) fi
   (* 

     [ StepExpr1 // StepExpr2 ]Expr ==  [StepExpr1 / descendant-or-self::node() / StepExpr2]Expr
   *)

  | PSlashSlash (e1, e2) ->
      let axis = fmkexpr (EPath(PAxis(Descendant_or_self, PNodeKindTest(AnyKind)))) fi in
      let e' = fmkexpr (EPath(PSlash(axis, e2))) fi in
      map_path_expr norm_context e (PSlash(e1, e'))
  | PAxis (a, nt) ->
      let ce = 
	begin
	  match nt with 
	  | PNodeKindTest nk -> 
	      if (forward_axis a) then CEForwardAxis(a, CPNodeKindTest nk) else CEReverseAxis(a, CPNodeKindTest nk) 
	  | PNameTest qname ->
	      (* J&M : resolve_attr_symbol & resolve_elem_symbol have side effects: 
                 For the empty namespace, they update the global namespace table. 
                 Revisit this code when we clean up namespace resolution *)
	      let rqname = 
		if (a = Attribute) then 
		  rattr_name(resolve_attr_symbol nsenv (UAttrSym (uattr_symbol qname)))
		else 
		  relem_name(resolve_elem_symbol nsenv (UElemSym (uelem_symbol qname)))
	      in
	      if (forward_axis a) then CEForwardAxis(a, CPNameTest rqname) else CEReverseAxis(a, CPNameTest rqname)
	end 
      in fmkcexpr ce None (Some e) fi

  | PStepQualifiers (e, sqs) ->
      let ce = map_expr norm_context e in
      List.fold_left (map_step_qualifier ce norm_context) ce sqs

(*

   Step qualifiers can be applied to three types of step expressions:
   forward axes, reverse axes, and primary expressions.

*)
and map_step_qualifier ce norm_context qual_expr sq = 
  match sq.pstep_qualifier_desc with 
  | PredicateQualifier e -> 
      let desc = e.pexpr_desc
      and fi = e.pexpr_loc in
      let cempty = fmkcexpr CEEmpty None None fi in
      let dotvar = mkvar dot e fi in
      let seqvar = mkvar sequence e fi in
      let posvar = mkvar position e fi in 
      let lastvar = mkvar last e fi in 
      let cond = map_expr norm_context e in
      let cond' = map_predicate_truth_value cond posvar (Some e) fi in 
      let body = fmkcexpr (CEIf(cond', dotvar, cempty))  None (Some e) fi in 
      let count_expr = fmkcexpr (CEApp((fn_uri, "count"), [seqvar])) None (Some e) fi in 
      let let_expr = 
	begin
         (* Determine whether StepExpr is a forward or reverse axis or a primary expression *)
	  match ce.pcexpr_desc with
	  | CEForwardAxis _ -> 
           (*  ForwardAxis ::= child | descendant | attribute | self | descendant-or-self 
              | following-sibling | following | namespace
              [ForwardStep Predicates "[" Expr "]"]Expr
              ==
              let $fs:sequence := fs:distinct-doc-order( [ForwardStep Predicates]Expr ) return
                let $fs:last := fn:count($fs:sequence) return
                  for $fs:dot at $fs:position in $fs:sequence return
                    if [Expr]Predicates then $fs:dot else ()
           *)
	      begin
		let for_expr = fmkcexpr (CEFor(None, dot, Some position, seqvar, body)) None (Some e) fi in
		let let_expr2 = fmkcexpr (CELet(None, last, count_expr, for_expr)) None (Some e) fi in
		if (!Conf.do_sbdo) then  (* ddo meta-instruction inserted -- default *)
			let docorder_expr = fmkcexpr (CEApp((glx_uri, "distinct-docorder"), [qual_expr])) None (Some e) fi in 
			fmkcexpr (CELet(None, sequence, docorder_expr, let_expr2)) None (Some e) fi
		else
			fmkcexpr (CELet(None, sequence, qual_expr, let_expr2)) None (Some e) fi
	      end

	  | CEReverseAxis _ -> 
           (* ReverseAxis    ::= parent | ancestor | preceding-sibling | preceding | ancestor-or-self
              [ReverseStep Predicates "[" Expr "]"]Expr
              ==
              let $fs:sequence := fs:distinct-doc-order( [ReverseStep Predicates]Expr ) return
                let $fs:last := fn:count($fs:sequence) return
                  for $fs:dot at $fs:new in $fs:sequence return
                    let $fs:position := $fs:last - $fs:new + 1 return
                      if [Expr]Predicates then $fs:dot else ()
            *)
   	      begin
		let (v, var) = gen_new_var_expr (Some e) fi in 
		let one_int_expr = fmkcexpr (CEScalar (buildAtomicInteger (Decimal._integer_of_int 1))) None (Some e) fi in
		let arith_expr1 = fmkcexpr (CEApp((op_uri,"integer-add"), [var; one_int_expr])) None (Some e) fi in
		let arith_expr = fmkcexpr (CEApp((op_uri,"integer-subtract"), [lastvar; arith_expr1])) None (Some e) fi in
		let body' = fmkcexpr (CELet(None, position, arith_expr, body)) None (Some e) fi in
		let for_expr = fmkcexpr (CEFor(None, dot, Some v, seqvar, body')) None (Some e) fi in
		let let_expr2 = fmkcexpr (CELet(None, last, count_expr, for_expr)) None (Some e) fi in
		if (!Conf.do_sbdo) then  (* ddo meta-instruction inserted -- default *)
			let docorder_expr = fmkcexpr (CEApp((glx_uri, "distinct-docorder"), [qual_expr])) None (Some e) fi in 
			fmkcexpr (CELet(None, sequence, docorder_expr, let_expr2)) None (Some e) fi
		else	
			fmkcexpr (CELet(None, sequence, qual_expr, let_expr2)) None (Some e) fi
	      end
	  | _ ->
           (*  [PrimaryExpr Predicates "[" Expr "]"]Expr   
               == 
               let $fs:sequence := [PrimaryExpr Predicates]Expr return
                 let $fs:last := fn:count($fs:sequence) return
                   for $fs:dot at $fs:position in $fs:sequence return
                     if [Expr]Predicates then $fs:dot else ()
            *)
	      begin
		let for_expr = fmkcexpr (CEFor(None, dot, Some position, seqvar, body)) None (Some e) fi in
		let let_expr2 = fmkcexpr (CELet(None, last, count_expr, for_expr)) None (Some e) fi in
		fmkcexpr (CELet(None, sequence, qual_expr, let_expr2)) None (Some e) fi 
	      end
	end
      in let_expr
and var_in_predicate v e =
  match e.pcexpr_desc with 
  | CEVar vname -> v = vname
  | CEApp (_, ces) -> (List.mem true (List.map (var_in_predicate v) ces))
  | CEFor (_, _, _, e1, e2) -> (var_in_predicate v e1) || (var_in_predicate v e2)
  | CELet (_, _, e1, e2) -> (var_in_predicate v e1) || (var_in_predicate v e2)
  | CEIf (e1, e2, e3) -> (var_in_predicate v e1) || (var_in_predicate v e2) || (var_in_predicate v e3)
  | CESeq (e1, e2) -> (var_in_predicate v e1) || (var_in_predicate v e2)
  | CECast (e,_) -> (var_in_predicate v e)
  | _ -> false


(****************************************************)
(* functions to map each components in query_module *)
(****************************************************)

(* Global variable declaration *)

let map_var norm_context v =
  let var_desc =
    match v.pvar_decl_desc with
    | (vn,e) ->
	let ce = map_expr norm_context e in
	(vn, ce)
  in
  fmkcvar_decl var_desc v.pvar_decl_loc

(* Key definition *)

let map_key norm_context v =
  let key_def =
    match v.pkey_def_desc with
    | (kn,e1,e2) ->
	let ce1 = map_expr norm_context e1 in
	let ce2 = map_expr norm_context e2 in
	(kn, ce1, ce2)
  in
  fmkckey_def key_def v.pkey_def_loc

(* Function Body *)

let map_body norm_context b output_dt = 
  let nsenv = nsenv_from_norm_context norm_context in 
  match b with
  | EFunctionBltIn fname ->
      let (_,rfname) = resolve_function_qname nsenv fname in
      CEFunctionBltIn rfname
  | EFunctionUser (vnames, expr) ->
      (* Type assertion checks that result of evaluating function body matches declared return type *)
      let ce = (map_expr norm_context expr) in
      let fi = (expr.pexpr_loc) in
      let (v1, var1) = gen_new_var_expr (Some expr) fi in
      let let_expr = fmkcexpr (CELet(Some output_dt, v1, ce, var1)) None (Some expr) fi in
      CEFunctionUser(vnames, let_expr)

(* Function Definition *)

let map_function_signature nsenv schema f = 
  let (fname, (input_types, return_type), _) = f.pfunction_def_desc in
  let (_,rfname) = Resolve.resolve_function_qname nsenv fname in
  let csign = (List.map (fun dt -> (resolve_datatype nsenv schema dt, map_datatype_to_model nsenv schema dt)) input_types, 
              (resolve_datatype nsenv schema return_type, map_datatype_to_model nsenv schema return_type)) in
  let arity = List.length (fst csign) in
  ((rfname, arity), (rfname, csign))

let map_function_body norm_context f = 
  let nsenv = nsenv_from_norm_context norm_context in
  let (fname, sign, body) = f.pfunction_def_desc in
  let (input_datattypes, output_datatype) = sign in
  let (_,rfname) = Resolve.resolve_function_qname nsenv fname in
  let cbody = map_body norm_context body output_datatype in
  let arity = List.length (fst sign) in
  let csign = one_sig_from_norm_context norm_context (rfname, arity) f.pfunction_loc in
  fmkcfunction_def (rfname, csign, cbody) f.pfunction_loc 

(***********)
(* Updates *)
(***********)

let map_insert_location norm_context replace_fun il =
  match il with
  | UAsLastInto e ->
      CUAsLastInto (replace_fun (map_expr norm_context e))
  | UAsFirstInto e ->
      CUAsFirstInto (replace_fun (map_expr norm_context e))
  | UInto e ->
      CUInto (replace_fun (map_expr norm_context e))
  | UAfter e ->
      CUAfter (replace_fun (map_expr norm_context e))
  | UBefore e ->
      CUBefore (replace_fun (map_expr norm_context e))

let map_simple_update norm_context replace_fun su =
  let sudesc = su.psimple_update_desc
  and fi = su.psimple_update_loc in
  let new_csudesc =
    match sudesc with
    | UInsert (e,il) ->
	CUInsert (replace_fun (map_expr norm_context e), map_insert_location norm_context replace_fun il)
    | UReplace (vo,e1,e2) ->
	CUReplace (vo,replace_fun (map_expr norm_context e1),replace_fun (map_expr norm_context e2))
    | UDelete e ->
	CUDelete (replace_fun (map_expr norm_context e))
    | UEmpty ->
	CUEmpty
  in
  fmkcsimple_update new_csudesc fi

let map_complex_update norm_context cu =
  let cudesc = cu.pcomplex_update_desc
  and fi = cu.pcomplex_update_loc in
  let new_csudesc =
    match cudesc with
    | UCond (cond,su1,su2) ->
	let cond' = map_expr norm_context cond in
	let cond'' = map_effective_boolean_value cond' None fi in
	CUCond (cond'', map_simple_update norm_context (fun x -> x) su1, map_simple_update norm_context (fun x -> x) su2)

    | UFLWR (fl_expr_list, where_expr_opt, order_opt, update_expr_list) ->

	(* look inside the for or let expression list, apply map_expr recursively *)
	let flwr_expr_list' = map_fl_list norm_context fl_expr_list in
	let tuple_vars = all_flwr_bindings fl_expr_list in
	let build_tuple_field vname =
	  let field_expr = fmkcexpr (CEVar vname) None None fi
	  in (vname, field_expr)
	in
	let tuple_expr = 
	  fmkcexpr (CECreateTuple (List.map build_tuple_field tuple_vars)) None None fi
	in
	let where_expr_opt' = 
	  begin
	    match where_expr_opt with
	    | None -> None
	    | Some e' -> let ce = (map_expr norm_context e') in Some ce
	  end 
	in
	(* The intermediate table with tuple creation *)
	let materialize_tuple_expr =
	  map_to_flwr flwr_expr_list' where_expr_opt' order_opt tuple_expr None fi
	in
	
	let replace_variable_expression_fun expr vname =
	  let local_tuple_access_expr =
	    fmkcexpr (CEAccessTuple (mkvar_none dot fi, vname)) None None fi
	  in
	  fst (Xquery_astutil.substitute_var local_tuple_access_expr vname expr)
	in
	let replace_fun expr =
	  List.fold_left replace_variable_expression_fun expr tuple_vars
	in
	
	(* Creating the sort by *)
	let sorted_materialize_tuple_expr =
	match order_opt with
	| None ->
	    materialize_tuple_expr
	| Some order ->
	    let (stable, sortspecs) = order in
	    let keys = List.rev sortspecs in
	    List.fold_left (map_sortkind norm_context dot replace_fun fi stable) materialize_tuple_expr sortspecs
	in
	let real_update_expr =
	  List.map (map_simple_update norm_context replace_fun) update_expr_list
	in
	CUForTuple (dot,sorted_materialize_tuple_expr,real_update_expr)
  in
  fmkccomplex_update new_csudesc fi

let map_update norm_context u =
  let udesc = u.pupdate_desc
  and fi = u.pupdate_loc in
  match udesc with
  | USimple su ->
      fmkcupdate (CUSimple (map_simple_update norm_context (fun x -> x) su)) (Some u) fi
  | UComplex cu ->
      fmkcupdate (CUComplex (map_complex_update norm_context cu)) (Some u) fi


(**************)
(* Statements *)
(**************)

let map_statement norm_context statement =
  let fi = statement.pstatement_loc in
  match statement.pstatement_desc with
  | SExpression e ->
      Xquery_astutil.fmkcstatement (CSExpression (map_expr norm_context e)) fi
  | SUpdate u ->
      Xquery_astutil.fmkcstatement (CSUpdate (map_update norm_context u)) fi

let map_statement_list norm_context ss = 
  List.map (map_statement norm_context) ss


(****************************************************************)
(* Toplevel functions to map the Full XQuery to the core XQuery *)
(****************************************************************)

(* Separate function to map prolog information *)

let map_library_module norm_context q = 
  let nsenv = nsenv_from_norm_context norm_context in
  let schema = schema_from_norm_context norm_context in
  let { pmodule_sorts = inlined_tys;
	pmodule_functions = fs;
	pmodule_contexts = ns;
	pmodule_vars = vs;
	pmodule_keys = ks;
	pmodule_exprs =  es } = q in

  match es with
  | [] ->
      begin
	let (nsenv', impt_types') = map_context_with_env ns nsenv in
	let all_sort_decl' = impt_types' @ inlined_tys in
	let schema' = map_sort_with_env nsenv' schema all_sort_decl' in
	let sigs' = (List.map (map_function_signature nsenv' schema) fs) in
        let norm_context' = add_sigs_to_norm_context (copy_norm_context_with_sigs norm_context nsenv' schema') sigs' in

        (* Second, map all function bodies: *)  
	let cmodule = { pcmodule_functions = List.map (map_function_body norm_context') fs;
			pcmodule_vars = List.map (map_var norm_context') vs;
			pcmodule_keys = List.map (map_key norm_context') ks;
			pcmodule_exprs = [] } in
	
	(norm_context', cmodule)
      end
  | _ ->
      raise (Query (Mapping_Failure ("Context contains some expressions...")))

(* Separate function to map prolog information *)

let map_main_module norm_context q = 
  let nsenv = nsenv_from_norm_context norm_context in
  let schema = schema_from_norm_context norm_context in
  let { pmodule_sorts = inlined_tys;
	pmodule_functions = fs;
	pmodule_contexts = ns;
	pmodule_vars = vs;
	pmodule_keys = ks;
	pmodule_exprs =  es } = q in
  begin
    let (nsenv', impt_types') = map_context_with_env ns nsenv in
    let all_sort_decl' = impt_types' @ inlined_tys in
    let schema' = map_sort_with_env nsenv' schema all_sort_decl' in
    let sigs' = (List.map (map_function_signature nsenv' schema') fs)  in
    let norm_context' = add_sigs_to_norm_context (copy_norm_context_with_sigs norm_context nsenv' schema') sigs' in

    (* Second, map all function bodies: *)  
    let cmodule = { pcmodule_functions = List.map (map_function_body norm_context') fs;
		    pcmodule_vars = List.map (map_var norm_context') vs;
		    pcmodule_keys = List.map (map_key norm_context') ks;
		    pcmodule_exprs = map_statement_list norm_context' es } in
    
    (norm_context', cmodule)
  end

