(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2001, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: maputil.ml,v 1.33 2003/06/23 00:53:09 simeon Exp $ *)

open Namespace
open Sym
open Finfo
open Datatypes
open Xquery_ast
open Xquery_astutil
open Error

let forward_axis axis =
  match axis with 
  | Parent -> false 
  | Ancestor -> false 
  | Ancestor_or_self -> false 
  | Attribute -> true
  | Child -> true
  | Descendant -> true
  | Descendant_or_self -> true
  | Namespace -> true
  | Self -> true

(* Generate new variables *)

let tempvar = ref 0

let gen_new_var () = 
  let _ = tempvar := !tempvar + 1 in 
  "v"^(string_of_int !tempvar)

let gen_new_var_expr eh fi = 
  let v = (NSPrefix "glx", gen_new_var()) in 
  (v, fmkcexpr (CEVar v) None eh fi)

(* helper functions for mapping functions *)
(* Adding 'parenthesises right associatively: l = [e1; e2; e3 ...; en] -> (e1 (e2 (e3 (... en)...) *)

let rec map_to_sequence eh fi exprlist =
  let cempty = fmkcexpr CEEmpty None None bogus in
  begin
    match exprlist with
    | [] -> cempty
    | [ expr ] -> expr
    | expr::exprlist' ->
	fmkcexpr (CESeq (expr, (map_to_sequence eh fi exprlist'))) expr.pcexpr_type eh fi 
  end

(* map_node_comp_operator

3.5.3 Node Comparisons

   The result of a node comparison is defined by applying the
   following rules, in order:

   1. Each operand must be either a single node or an empty sequence;
      otherwise a type error is raised.

   2. If either operand is an empty sequence, the result of the
      comparison is an empty sequence.

   3. A comparison with the is operator is true if the two operands
      are nodes that have the same identity; otherwise it is false. A
      comparison with the isnot operator is true if the two operands
      are nodes that have different identities; otherwise it is
      false. See [XQuery 1.0 and XPath 2.0 Data Model] for a
      discussion of node identity.

3.5.4 Order Comparisons

   The result of an order comparison is defined by applying the
   following rules, in order:

   1. Both operands must be either a single node or an empty sequence;
      otherwise a type error is raised.

   2. If either operand is an empty sequence, the result of the
      comparison is an empty sequence.

   3. A comparison with the << operator returns true if the first
      operand node is earlier than the second operand node in document
      order; otherwise it returns false.

   4.  A comparison with the >> operator returns true if the first
       operand node is later than the second operand node in document
       order; otherwise it returns false.

  [ Expr1 NodeOp Expr2 ] ==

  typeswitch [Expr1] 
  case () return ()
  default $v1 return 
    typeswitch [Expr2] 
    case () return ()
    default $v2 op:NodeOp($v1, $v2)

*)

let map_node_comp_operator fn e1 e2 eh fi = 
  let (x, varx) = gen_new_var_expr eh fi in
  let (v1, var1) = gen_new_var_expr eh fi in
  let (v2, var2) = gen_new_var_expr eh fi in

  let error_expr = fmkcexpr (CEError (fmkcexpr (CEScalar (AtomicString "Invalid node comparison argument")) None eh fi)) None eh fi in
  let node_datatype = fmkdatatype (DTNode, None) fi in
  let empty_expr = fmkcexpr (CEEmpty) None eh fi in
  let empty_datatype = fmkdatatype (DTEmpty, None) fi in

  let comp_fn = fmkcexpr (CEApp (fn, [ var1; var2 ])) None eh fi in 
  let patlist2 = [ 
    (fmkpattern (Case empty_datatype) fi, (Some x), empty_expr);
    (fmkpattern Default fi, (Some v2), comp_fn)
  ] in
  let typs_expr2 = fmkcexpr (CETypeswitch (e2, patlist2)) None eh fi in

  let patlist = [ 
    (fmkpattern (Case empty_datatype) fi, (Some x), empty_expr);
    (fmkpattern Default fi, (Some v1), typs_expr2)
  ] in
  fmkcexpr (CETypeswitch (e1, patlist)) None eh fi


(* atomization takes an item sequence and returns an optional atomic
   value: used to implicitly convert any sequence to its atomic
   content.

    [Expr]Atomization == 
      let atomic value? $v := fn:data([Expr])
      return $v 

   This semantics raises a static type error if:
   1. fn:data() yields more than one atomic value

*)

let atomization e1 eh fi = 
  let (v, var) = gen_new_var_expr eh fi in
  let opt_atomic_datatype = fmkdatatype (DTAtomic, Some(Occurrence.occurs 0, Occurrence.occurs 1)) fi in
  let dataex = fmkcexpr (CEApp ((fn_uri, "data"), [ e1 ])) None eh fi in 
  fmkcexpr (CELet (Some opt_atomic_datatype, v, dataex, var)) None eh fi

let fn_data e1 eh fi = 
  fmkcexpr (CEApp ((fn_uri, "data"), [ e1 ])) None eh fi

(* map_to_cast does atomization; cast_to_simpletype does not *)
let map_to_cast e1 m eh fi = 
  let (e1v, e1var) = gen_new_var_expr eh fi in
  let cast_expr = fmkcexpr (CECast (e1var, m)) None eh fi in
  let let_expr = fmkcexpr (CELet (None, e1v, (atomization e1 eh fi), cast_expr)) None eh fi in
  let_expr 

(* map_to_castable does atomization; cast_to_simpletype does not *)
let map_to_castable e1 m eh fi = 
  let (e1v, e1var) = gen_new_var_expr eh fi in
  let cast_expr = fmkcexpr (CECastable (e1var, m)) None eh fi in
  let let_expr = fmkcexpr (CELet (None, e1v, (atomization e1 eh fi), cast_expr)) None eh fi in
  let_expr 

(* cast_to_simpletype takes an expression and a simple type, and constructs a CAST express *)
let cast_to_simpletype e1 simple_type eh fi = 
  let cast_expr = fmkcexpr (CECast (e1, (fmkmodel (TBase simple_type) fi))) None eh fi in
  cast_expr 


(* map_effective_boolean_value takes an expression and returns a
   boolean value.

   -- see also: map_predicate_truth_value

  The EBV of an expression is false if the expression is:
  * The empty sequence
  * The singleton boolean value false
  * The singleton string value ""
  * A singleton numeric value that is numerically equal to zero or a float or double equal to NaN
  Otherwise, it is true.

  We have not implemented comparison ($v != NaN)
*)

let map_effective_boolean_value e1 eh fi = 
  fmkcexpr (CEApp ((fn_uri, "boolean"), [ e1 ])) None eh fi

let map_to_CEIf cond e1 e2 eh fi = 
  let cond' = map_effective_boolean_value cond eh fi in 
  fmkcexpr (CEIf (cond', e1, e2)) None eh fi

let map_to_CESome rodt v e1 e2 eh fi = 
  let e2' = map_effective_boolean_value e2 eh fi in 
  fmkcexpr (CESome (rodt, v, e1, e2')) None eh fi

let map_to_CEEvery rodt v e1 e2 eh fi = 
  let e2' = map_effective_boolean_value e2 eh fi in 
  fmkcexpr (CEEvery (rodt, v, e1, e2')) None eh fi

(*
  map_predicate_truth_value implements the semantics for StepQualifiers.
  -- see also: map_effective_boolean_value

  The predicate truth value is derived by applying the following
  rules, in order:

    1. If the value of the predicate expression is one simple value of
      a numeric type, the value is rounded to an integer using the
      rules of the round function, and the predicate truth value is
      true if and only if the resulting integer is equal to the value
      of the context position.

    2. Otherwise, the predicate truth value is the effective boolean
       value of the predicate expression.

  Formally: 

    [Expr]_PredicateTruthValue = 
    typeswitch ([Expr]) 
    case integer $v return position() = $v
    case numeric $v return xsd:double(position()) = round(cast as xsd:double $v)
    default $v return [$v]_EffectiveBooleanValue

*)

let map_predicate_truth_value e1 pos_var eh fi = 
  let (v, var) = gen_new_var_expr eh fi in
  let (v1, var1) = gen_new_var_expr eh fi in
  let (v2, var2) = gen_new_var_expr eh fi in

  let integer_type = fmkdatatype (DTBase ATInteger, None) fi in
  let numeric_datatype = fmkdatatype (DTNumeric, None) fi in




  let pos_integer_expr =  fmkcexpr (CEApp ((op_uri,"equal"), [pos_var; var])) None eh fi in

  let cast_expr = cast_to_simpletype var1 ATDouble eh fi in
  let round_expr = fmkcexpr (CEApp ((fn_uri,"round"), [cast_expr])) None eh fi  in 
  let cast_expr' = cast_to_simpletype pos_var ATDouble eh fi in 
  let pos_expr =  fmkcexpr (CEApp ((op_uri,"equal"), [cast_expr'; round_expr])) None eh fi in

  let default_expr = map_effective_boolean_value var2 eh fi in 
  let patlist = [
    (fmkpattern (Case integer_type) fi, Some v, pos_integer_expr);
    (fmkpattern (Case numeric_datatype) fi, Some v1, pos_expr);
    (fmkpattern Default fi, Some v2, default_expr)
 ]
  in
  let typs_expr = fmkcexpr (CETypeswitch (e1, patlist)) None eh fi in
  typs_expr

(*

What the spec says:

  3.5.2 General Comparisons

  Each of the operands is first atomized. The result of the comparison
  is true if and only if there is a pair of atomic values, one
  belonging to the result of the atomization of the first operand and
  the other belonging to the result of the atomization of the second
  operand, that have the required magnitude relationship. Otherwise
  the result of the comparison is false.

  The magnitude relationship between two atomic values is determined as follows:
 
  1. If either atomic value has the most specific dynamic type
     untypedAtomic, that value is cast to a required type, which is
     determined as follows:

  a. If the type of the other atomic value is numeric, the required type is xs:double.

  b. If the most specific dynamic type of the other atomic value is
     untypedAtomic, the required type is xs:string.

  c. Otherwise, the required type is the type of the other atomic value.

  If the cast fails, a dynamic error is raised.

What we implement:
(If both args are untypedAtomic, we cast both to string.
 If one arg is untypedAtomic and the other is numeric, we cast untypedAtomic to double.
 If one arg is untypedAtomic and the other is string, we cast untypedAtomic to string.
 Type promotion during function application does the rest of the work for us.)

  [Expr1 GeneralOp Expr2] ==  
  some $e1v in fn:data([Expr1]) satisfies
    some $e2v in fn:data([Expr2]) satisfies
      (* type of e21 depends on e2v *)
      let $v1 := 
        typeswitch ($e1v)
        case $e1v' as untypedAtomic return
          (typeswitch ($e2v)     
           case untypedAtomic return cast as xs:string($e1v')
           case string return cast as xs:string($e1v')
           case numeric return cast as xs:double($e1v')
           default return $e1v')
        default return $e1v
      return
      (* type of e2v depends on e1v *)
      let $v2 := 
        typeswitch ($e2v)
        case $e2v' untypedAtomic return
          (typeswitch ($e1v)     
           case untypedAtomic return cast as xs:string($e2v')
           case string return cast as xs:string($e2v')
           case numeric return cast as xs:double($e2v')
           default return $e2v')
        default return $e2v
      return
        op:ValueEqOp($v1, $v2)

*)

let map_general_comp_operator fn e1 e2 eh fi = 
  let (v1, var1) = gen_new_var_expr eh fi in
  let (v2, var2) = gen_new_var_expr eh fi in

  let (e1v, e1var) = gen_new_var_expr eh fi in
  let (e1v', e1var') = gen_new_var_expr eh fi in

  let (e2v, e2var) = gen_new_var_expr eh fi in
  let (e2v', e2var') = gen_new_var_expr eh fi in

  let numeric_datatype = fmkdatatype (DTNumeric, None) fi in
  let string_datatype = fmkdatatype ((DTBase ATString), None) fi in
  let untyped_atomic_datatype = fmkdatatype ((DTBase ATUntyped), None) fi in

  let comp_fn = fmkcexpr (CEApp (fn, [ var1; var2 ])) None eh fi in 

  let patlist2 = [ 
    (fmkpattern (Case untyped_atomic_datatype) fi, None, (cast_to_simpletype e2var' ATString eh fi));
    (fmkpattern (Case string_datatype) fi, None, (cast_to_simpletype e2var' ATString eh fi));
    (fmkpattern (Case numeric_datatype) fi, None, (cast_to_simpletype e2var' ATDouble eh fi));
    (fmkpattern Default fi, None, e2var)
  ] in
  let typs_expr4 = fmkcexpr (CETypeswitch (e1var, patlist2)) None eh fi in

  let patlist = [ 
    (fmkpattern (Case untyped_atomic_datatype) fi, (Some e2v'), typs_expr4);
    (fmkpattern Default fi, None, e2var)
  ] in
  let typs_expr3 = fmkcexpr (CETypeswitch (e2var, patlist)) None eh fi in


  let patlist2 = [ 
    (fmkpattern (Case untyped_atomic_datatype) fi, None, (cast_to_simpletype e1var' ATString eh fi));
    (fmkpattern (Case string_datatype) fi, None, (cast_to_simpletype e1var' ATString eh fi));
    (fmkpattern (Case numeric_datatype) fi, None, (cast_to_simpletype e1var' ATDouble eh fi));
    (fmkpattern Default fi, None, e1var')
  ] in
  let typs_expr2 = fmkcexpr (CETypeswitch (e2var, patlist2)) None eh fi in
  let patlist = [ 
    (fmkpattern (Case untyped_atomic_datatype) fi, (Some e1v'), typs_expr2);
    (fmkpattern Default fi, None, e1var)
  ] in
  let typs_expr = fmkcexpr (CETypeswitch (e1var, patlist)) None eh fi in

  let let_expr2 = fmkcexpr (CELet (None, v2, typs_expr3, comp_fn)) None eh fi in 
  let let_expr = fmkcexpr (CELet (None, v1, typs_expr, let_expr2)) None eh fi in 
  let some_expr2 = map_to_CESome None e2v (fn_data e2 eh fi) let_expr eh fi in
  let some_expr1 = map_to_CESome None e1v (fn_data e1 eh fi) some_expr2 eh fi in
  some_expr1

(*
  Predicate(Expr1, Expr2) ==
  some $v1 in [Expr1] satisfies
    some $v2 in [Expr2] satisfies
      Predicate($v1,$v2)
*)
let map_existential predicate_expr e1 e2 eh fi =
  let (v1, var1) = gen_new_var_expr eh fi in
  let (v2, var2) = gen_new_var_expr eh fi in
  let pred_expr = predicate_expr var1 var2 eh fi in
  let some_expr2 = map_to_CESome None v2 e2 pred_expr eh fi in
  let some_expr1 = map_to_CESome None v1 e1 some_expr2 eh fi in
  some_expr1

(*  map_value_comp_operator

  
  3.5.1 Value Comparisons

  Value comparisons are intended for comparing single values. The
  result of a value comparison is defined by applying the following
  rules, in order:
  
  1. Atomization is applied to each operand. If the result, called an
     atomized operand, is not a single atomic value, a type error is
     raised.

  2. If either atomized operand is an empty sequence, a type error is raised. 

  3. If either atomized operand has type xdt:untypedAtomic, that
     atomized operand (those atomized operands) is (are) cast to
     xs:string.

  4. The result of the comparison is true if the first atomized
     operand is (equal, not equal, less than, less than or equal,
     greater than, greater than or equal) to the second atomized
     operand; otherwise the result of the comparison is false. B.2
     Operator Mapping describes which combinations of atomic types are
     comparable, and how comparisons are performed on values of
     various types. If the first atomized operand is not comparable
     with the second atomized operand, a type error is raised.

  ==========================================

    [Expr1 ValueEqOp Expr2] 
                                        ==
    let  $e1 := 
           typeswitch ([ Expr1 ]_Atomization)
           case $v1 as xdt:untypedAtomic return cast as string($v1)
           default $v2 return $v2
    return 
    let $e2 := 
           typeswitch ([ Expr2 ]_Atomization)
           case $v1 as xdt:untypedAtomic return cast as string($v1)
           default $v2 return $v2
    return
      op:ValueEqOp($e1, $e2)
*)

let map_value_comp_operator fn e1 e2 eh fi = 

  let (e1v, e1var) = gen_new_var_expr eh fi in
  let (e2v, e2var) = gen_new_var_expr eh fi in

  let (v1, var1) = gen_new_var_expr eh fi in
  let (v2, var2) = gen_new_var_expr eh fi in
  let (v3, var3) = gen_new_var_expr eh fi in
  let (v4, var4) = gen_new_var_expr eh fi in

  let untyped_atomic_datatype = fmkdatatype ((DTBase ATUntyped), None) fi in

  let comp_fn = fmkcexpr (CEApp (fn, [ e1var; e2var ])) None eh fi in 

  let patlist2 = [ 
    (fmkpattern (Case untyped_atomic_datatype) fi, (Some v2), (cast_to_simpletype var2 ATString eh fi));
    (fmkpattern Default fi, (Some v3), var3)
  ] in
  let typs_expr2 = fmkcexpr (CETypeswitch ((fn_data e2 eh fi), patlist2)) None eh fi in

  let patlist = [ 
    (fmkpattern (Case untyped_atomic_datatype) fi, (Some v1), (cast_to_simpletype var1 ATString eh fi));
    (fmkpattern Default fi, (Some v4), var4)
  ] in
  let typs_expr = fmkcexpr (CETypeswitch ((fn_data e1 eh fi), patlist)) None eh fi in

  let let_expr2 = fmkcexpr (CELet (None, e2v, typs_expr2, comp_fn)) None eh fi in
  let let_expr = fmkcexpr (CELet (None, e1v, typs_expr, let_expr2)) None eh fi in
  let_expr  

(* 
  map_unary_operator 

    [UnaryOp Expr] 
         ==
    let $e := 
      typeswitch ([ $core ]_Atomization)
      case $v1 as xsd:untypedAtomic return  cast as xsd:double($v1)
      default $v1 return $v1
    return op:numeric-UnaryOp($e)
*)
let map_unary_operator fn e1 eh fi = 
  let (x, varx) = gen_new_var_expr eh fi in
  let (v1, var1) = gen_new_var_expr eh fi in
  let (v2, var2) = gen_new_var_expr eh fi in
  let (e1v, e1var) = gen_new_var_expr eh fi in

  let untyped_atomic_datatype = fmkdatatype ((DTBase ATUntyped), None) fi in
  let numeric_fn = fmkcexpr (CEApp (fn, [ e1var ])) None eh fi in 
  let patlist3 = [ 
    (fmkpattern (Case untyped_atomic_datatype) fi, (Some v1), (cast_to_simpletype var1 ATDouble eh fi));
    (fmkpattern Default fi, (Some v2), var2)
  ] in
  let typs_expr3 = fmkcexpr (CETypeswitch ((fn_data e1 eh fi), patlist3)) None eh fi in

  let let_expr = fmkcexpr (CELet (None, e1v, typs_expr3, numeric_fn)) None eh fi in
  let_expr  

(*  
  3.4 Arithmetic Expressions

  An arithmetic expression is evaluated by applying the following
  rules, in order, until an error is raised or a value is computed:

  1. Atomization is applied to each operand. If the resulting value is
     not an empty sequence or a single atomic value, then a type error
     is raised.

  2. If either operand is an empty sequence, the result of the
     operation is an empty sequence.

  3. If an operand has type xdt:untypedAtomic, it is
     cast to xs:double. If the cast fails, a type error is raised.

  4. If the operand type(s) are valid for the given operator, the
     operator is applied to the operand(s), resulting in an atomic
     value or a dynamic error (for example, an error might result from
     dividing by zero.)  If the operand type(s) are not valid for the
     given operator, a type error is raised.

  map_arithmetic_operator
      
      [Expr1 ArithOp Expr2] 
                ==
      let $e1 := 
        typeswitch ([ Expr1 ]_Atomization)
        case $v1 as xdt:untypedAtomic return cast as xsd:double($v1)
        default $v1' return $v1'
      return 
      let $e2 := 
        typeswitch ([ Expr2 ]_Atomization)
        case $v2 as xdt:untypedAtomic return cast as xsd:double($v2)
        default $v2' return $v2'
      return
        op:numeric-ArithOp($e1, $e2)
*)

let map_arithmetic_operator fn e1 e2 eh fi = 
  let (v1, var1) = gen_new_var_expr eh fi in
  let (v2, var2) = gen_new_var_expr eh fi in
  let (v3, var3) = gen_new_var_expr eh fi in
  let (v4, var4) = gen_new_var_expr eh fi in

  let (e1v, e1var) = gen_new_var_expr eh fi in
  let (e2v, e2var) = gen_new_var_expr eh fi in

  let untyped_atomic_datatype = fmkdatatype ((DTBase ATUntyped), None) fi in
  let numeric_fn = fmkcexpr (CEApp (fn, [ e1var; e2var ])) None eh fi in

  let patlist4 = [ 
    (fmkpattern (Case untyped_atomic_datatype) fi, (Some v3), (cast_to_simpletype var3 ATDouble eh fi));
    (fmkpattern Default fi, (Some v4), var4)
  ] in
  let typs_expr4 = fmkcexpr (CETypeswitch ((fn_data e2 eh fi), patlist4)) None eh fi in

  let patlist3 = [ 
    (fmkpattern (Case untyped_atomic_datatype) fi, (Some v1), (cast_to_simpletype var1 ATDouble eh fi));
    (fmkpattern Default fi, (Some v2), var2)
  ] in
  let typs_expr3 = fmkcexpr (CETypeswitch ((fn_data e1 eh fi), patlist3)) None eh fi in

  let let_expr2 = fmkcexpr (CELet (None, e2v, typs_expr4, numeric_fn)) None eh fi in
  let let_expr = fmkcexpr (CELet (None, e1v, typs_expr3, let_expr2)) None eh fi in
  let_expr  

(* 
   map_idiv_operator

      [Expr1 idiv Expr2] 
                ==
      let $e1 := 
        typeswitch ([ Expr1 ]_Atomization)
        case $v1 as xdt:untypedAtomic return cast as xsd:integer($v1)
        default $v1' return $v1'
      return 
      let $e2 := 
        typeswitch ([ Expr2 ]_Atomization)
        case $v2 as xdt:untypedAtomic return cast as xsd:integer($v2)
        default $v2' return $v2'
   return op:numeric-idivide($e1, $e2)

*)
let map_idiv_operator e1 e2 eh fi = 

  let (v1, var1) = gen_new_var_expr eh fi in
  let (v2, var2) = gen_new_var_expr eh fi in
  let (v3, var3) = gen_new_var_expr eh fi in
  let (v4, var4) = gen_new_var_expr eh fi in

  let (e1v, e1var) = gen_new_var_expr eh fi in
  let (e2v, e2var) = gen_new_var_expr eh fi in

  let untyped_atomic_datatype = fmkdatatype ((DTBase ATUntyped), None) fi in
  let numeric_fn = fmkcexpr (CEApp ((op_uri, "numeric-idivide"), [ e1var; e2var ])) None eh fi in 
  let patlist4 = [ 
    (fmkpattern (Case untyped_atomic_datatype) fi, (Some v3), (cast_to_simpletype var3 ATInteger eh fi));
    (fmkpattern Default fi, (Some v4), var4)
  ] in
  let typs_expr4 = fmkcexpr (CETypeswitch ((fn_data e2 eh fi), patlist4)) None eh fi in

  let patlist3 = [ 
    (fmkpattern (Case untyped_atomic_datatype) fi, (Some v1), (cast_to_simpletype var1 ATInteger eh fi));
    (fmkpattern Default fi, (Some v2), var2)
  ] in
  let typs_expr3 = fmkcexpr (CETypeswitch ((fn_data e1 eh fi), patlist3)) None eh fi in

  let let_expr2 = fmkcexpr (CELet (None, e2v, typs_expr4, numeric_fn)) None eh fi in
  let let_expr = fmkcexpr (CELet (None, e1v, typs_expr3, let_expr2)) None eh fi in
  let_expr  


let map_binary_operator b e1 e2 eh fi = 
  begin
    match b with
          (* Sequence operators *)
          (* e1 intersect e2 *)
    | BEIntersect ->
	fmkcexpr (CEApp ((op_uri, "intersect"), [e1;e2])) None eh fi   
          (* e1 union e2 *)
    | BEUnion ->
	fmkcexpr (CEApp ((op_uri, "union"), [e1;e2])) None eh fi
          (* e1 except e2 *)
    | BEExcept ->
	fmkcexpr (CEApp ((op_uri, "except"), [e1;e2])) None eh fi         
          (* e1 | e2 *)
    | BEBar ->
	fmkcexpr (CEApp ((op_uri, "union"), [e1;e2])) None eh fi               
          (* Logical operators *)
          (* e1 and e2 *)
    | BEAnd ->
	let e1' = map_effective_boolean_value e1 eh fi in 
        let e2' = map_effective_boolean_value e2 eh fi in 
	fmkcexpr (CEApp ((op_uri, "and"), [e1';e2'])) None eh fi             
          (* e1 or e2 *)
    | BEOr -> 
	let e1' = map_effective_boolean_value e1 eh fi in 
        let e2' = map_effective_boolean_value e2 eh fi in 
	fmkcexpr (CEApp ((op_uri, "or"), [e1';e2'])) None eh fi          
	  
          (* Order operators *)
          (* e1 << e2 *)
    | BEPrecedes -> 
	map_node_comp_operator (op_uri,"node-before") e1 e2 eh fi 
          (* e1 >> e2 *)
    | BEFollows -> 
	map_node_comp_operator (op_uri,"node-after") e1 e2 eh fi 
          (* ValueComp operators *)
          (* e1 eq e2 *)
    | BEEq -> 
	map_value_comp_operator (op_uri,"equal") e1 e2 eh fi
           (* e1 neq e2 *)
    | BENEq -> 
	map_value_comp_operator (op_uri,"nequal") e1 e2 eh fi
          (* e1 lt e2 *)
    | BELtOp -> 
	map_value_comp_operator (op_uri,"lt") e1 e2 eh fi                     
          (* e1 gt e2 *)
    | BEGtOp -> 
	map_value_comp_operator (op_uri,"gt") e1 e2 eh fi                     
          (* e1 le e2 *)
    | BELte -> 
	map_value_comp_operator (op_uri,"le") e1 e2 eh fi                     
          (* e1 ge e2 *)
    | BEGte -> 
	map_value_comp_operator (op_uri,"ge") e1 e2 eh fi                     
	  
          (* GeneralComp operators *)
          (* e1 = e2 *)
    | BEEqual -> 
        map_general_comp_operator (op_uri,"equal") e1 e2 eh fi 
          (* e1 != e2 *)
    | BENEqual ->  
	map_general_comp_operator (op_uri,"nequal") e1 e2 eh fi 
          (* e1 < e2 *)
    | BELt -> 
	map_general_comp_operator (op_uri,"lt") e1 e2 eh fi 
          (* e1 > e2 *)
    | BEGt -> 
	map_general_comp_operator (op_uri,"gt") e1 e2 eh fi 
          (* e1 <= e2 *)
    | BELteq -> 
	map_general_comp_operator (op_uri,"le") e1 e2 eh fi 
          (* e1 >= e2 *)
    | BEGteq -> 
	map_general_comp_operator (op_uri,"ge") e1 e2 eh fi 
          (* NodeComp operators *)
          (* e1 == e2 *)
    | BEIs -> 
	map_node_comp_operator (op_uri,"node-equal") e1 e2 eh fi 
          (* e1 !== e2 *)
    | BEIsNot -> 
	let ce' = map_node_comp_operator (op_uri,"node-equal") e1 e2 eh fi in
	fmkcexpr (CEApp ((fn_uri, "not"), [ce'])) None eh fi
          (* Arithmetic operators *)
          (* e1 * e2 *)
    | BEMult ->
        map_arithmetic_operator (op_uri, "numeric-multiply") e1 e2 eh fi
          (* e1 + e2 *)
    | BEPlus ->
        map_arithmetic_operator (op_uri, "numeric-add") e1 e2 eh fi
          (* e1 - e2 *)
    | BEMinus ->
        map_arithmetic_operator (op_uri, "numeric-subtract") e1 e2 eh fi
          (* e1 div e2 *)
    | BEDiv ->
        map_arithmetic_operator (op_uri, "numeric-divide") e1 e2 eh fi
          (* e1 idiv e2 *)
    | BEIDiv ->
        map_idiv_operator e1 e2 eh fi
          (* e1 mod e2 *)
    | BEMod ->
        map_arithmetic_operator (op_uri, "numeric-mod") e1 e2 eh fi
  end

(*

3.1.4 Function Calls

  The function conversion rules are used to convert an argument value
  or a return value to its expected type; that is, to the declared
  type of the function parameter or return. The expected type is
  expressed as a SequenceType. The function conversion rules are
  applied to a given value as follows:

  * If the expected type is a sequence of an atomic type (possibly
    with an occurrence indicator *, +, or ?), the following
    conversions are applied:

    1. Atomization is applied to the given value, resulting in a
       sequence of atomic values.

    2. Each item in the atomic sequence that is of type
       xs:anySimpleType is cast to the required atomic type.

    3. For each numeric item in the atomic sequence that can be
       promoted to the expected atomic type using the promotion rules
       in B.1 Type Promotion, the promotion is done.

  * If, after the above conversions, the resulting value does not
    match the expected type according to the rules for SequenceType
    Matching , a type error is raised. Note that the rules for
    SequenceType Matching permit a value of a derived type to be
    substituted for a value of its base type.

     let $var_1 := [ Expr_1 ]_AtomizeArg return
     ...
     let $var_n := [ Expr_n ]_AtomizeArg return
     let $var_ret := FuncName($var_i, ..., $var_n) 
     return $var_ret
*)

let map_function_app norm_context rfname elist eh fi  = 
  let schema = Norm_context.schema_from_norm_context norm_context in
  let (uri, lname) = rfname in
  let map_atomize_arg act_expr act_model = 
    if (Typeutil.is_anysimpletype schema act_model) then
      fn_data act_expr eh fi 
    else act_expr
  in

  (* Dynamic type check is only necessary if static analysis is off. 
     We should generate dynamic checks conditionally, but will wait to do 
     this until context issue is resolved. - Mary *)

  let map_actual_arg ((v, var), (act_expr, (act_dt, act_model))) cexpr = 
    fmkcexpr (CELet (None, v, map_atomize_arg act_expr act_model, cexpr)) None eh fi 

  in
      (* Arity is used for dispatch now.
	 Previous error is now raised from the static environment.
	 - Jerome *)
  let arity = List.length elist in
  let (input_types, output_datatype) = Norm_context.one_sig_from_norm_context norm_context (rfname,arity) fi in
  let expr_types = List.combine elist input_types in
  let vars = List.map (fun _ -> gen_new_var_expr eh fi) expr_types in
  let var_expr_types = List.combine vars expr_types in
  let varlist = List.map snd vars in
  let funapp_expr = fmkcexpr (CEApp (rfname, varlist)) None eh fi in
  let (ret_v, ret_var) = gen_new_var_expr eh fi  in
  let ret_expr = funapp_expr 
  in  List.fold_right map_actual_arg var_expr_types ret_expr

let rec aux_map_to_CEFLWR fl_list ret_expr eh fi =
  match fl_list with
  | [] ->
      raise (Query (Internal_Error ("FLWR expression without FOR or LET")))
  | [fl] -> 
      begin
	match fl.pcexpr_desc with
	| CELet (odt, vn, ce, _) ->
	    fmkcexpr (CELet (odt, vn, ce, ret_expr)) None fl.pcexpr_origin fl.pcexpr_loc
	| CEFor (odt, vn, ovn, ce, _) ->
	    fmkcexpr (CEFor (odt, vn, ovn, ce, ret_expr)) None fl.pcexpr_origin fl.pcexpr_loc
	| _ -> 
	    raise (Query (Internal_Error ("FLWR expression contains invalid clause "^(Print.bprintf_cexpr "" fl))))
      end
  | h::tl ->
      begin
	match h.pcexpr_desc with
	| CELet (odt, vn, ce, _) ->
	    fmkcexpr (CELet (odt, vn, ce, (aux_map_to_CEFLWR tl ret_expr eh fi))) None h.pcexpr_origin h.pcexpr_loc
	| CEFor (odt, vn, ovn, ce, _) ->
	    fmkcexpr (CEFor (odt, vn, ovn, ce, (aux_map_to_CEFLWR tl ret_expr eh fi))) None h.pcexpr_origin h.pcexpr_loc
	| _ -> 
	    raise (Query (Internal_Error ("FLWR expression contains invalid clause "^(Print.bprintf_cexpr "" h))))
      end

let map_to_flwr fl_list where_opt order_opt ret_expr eh fi = 
  let ret_expr' =
    match where_opt with
    | None -> ret_expr
    | Some ce -> 
	begin
	  (* If the order-by clause is present, return an empty tuple, otherwise an empty value *)
          let empty_expr =
	    match order_opt with 
	    | None -> fmkcexpr CEEmpty ce.pcexpr_type ce.pcexpr_origin ce.pcexpr_loc 
(*	    | Some _ -> fmkcexpr (CECreateTuple []) None ce.pcexpr_origin ce.pcexpr_loc *)
	    | Some _ -> fmkcexpr CEEmpty ce.pcexpr_type ce.pcexpr_origin ce.pcexpr_loc 
	  in map_to_CEIf ce ret_expr empty_expr eh fi
	end
  in
(* NB: |fl_list| > 1 *)
  aux_map_to_CEFLWR fl_list ret_expr' eh fi

