(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: parseutil.ml,v 1.35 2003/06/27 00:05:08 simeon Exp $ *)

open Finfo
open Error
open Conf
open Encoding

open Xquery_ast
open Lexutil

(******************************)
(* post-parsing verifications *)
(******************************)

let local_fun_equals f1 f2 =
  match (f1.pcfunction_def_desc, f2.pcfunction_def_desc) with
  | ((fn1,_,_),(fn2,_,_)) ->
      not(fn1 = fn2)

let rec check_functions fds =
  match fds with
  | [] ->
      None
  | fd :: fds' ->
      if
	List.for_all (local_fun_equals fd) fds'
      then
	(check_functions fds')
      else
	Some fd

let local_var_equals v1 v2 =
  match (v1.pcvar_decl_desc, v2.pcvar_decl_desc) with
  | ((vn1,_),(vn2,_)) ->
      not(vn1 = vn2)

let rec check_variables vds =
  match vds with
  | [] ->
      None
  | vd :: vds' ->
      if
	List.for_all (local_var_equals vd) vds'
      then
	(check_variables vds')
      else
	Some vd

let check_multi_def_functions q =
  match (check_functions (q.pcmodule_functions)) with
  | None ->
      ()
  | Some fd ->
      begin
	match fd.pcfunction_def_desc with
	| (fn,_,_) ->
	    raise (Query (Symbol_Already_Defined (fd.pcfunction_loc, "Function ", (Namespace.string_of_rqname fn))))
      end

let check_multi_def_variables q =
  match (check_variables (q.pcmodule_vars)) with
  | None ->
      ()
  | Some vd ->
      begin
	match vd.pcvar_decl_desc with
	| (vn,_) ->
	    raise (Query (Symbol_Already_Defined (vd.pcvar_decl_loc, "Variable ", (Namespace.string_of_uqname vn))))
      end

let check_multi_def q =
  check_multi_def_functions q;
  check_multi_def_variables q

(*******************)
(* Lexers handling *)
(*******************)

type xml_lexers =
    { encoding             : Glx_pxp.encoding;
      opening_tag_state    : Lexing.lexbuf -> Xquery_parser.token;
      closing_tag_state    : Lexing.lexbuf -> Xquery_parser.token;
      text_state           : Lexing.lexbuf -> Xquery_parser.token;
      attribute_text_state : Lexing.lexbuf -> Xquery_parser.token;
      pi_state             : Lexing.lexbuf -> Xquery_parser.token;
      cdata_state          : Lexing.lexbuf -> Xquery_parser.token;
      comment_state        : Lexing.lexbuf -> Xquery_parser.token;
      xml_decl_state       : Lexing.lexbuf -> Xquery_parser.token;
      xml_document_state   : Lexing.lexbuf -> Xquery_parser.token;
      doctype_state        : Lexing.lexbuf -> Xquery_parser.token;
      dtddecl_state        : Lexing.lexbuf -> Xquery_parser.token;
      entity_value_state   : Lexing.lexbuf -> Xquery_parser.token }


type language_lexers =
    { language             : language_kind;
      schema_state         : Lexing.lexbuf -> Xquery_parser.token;
      default_state        : Lexing.lexbuf -> Xquery_parser.token }

let lexers_by_encoding = Hashtbl.create 5;;

let dummy_lexer x =
  raise (Query (Undefined ("No lexical support compiled! You must have at least one character encoding selected before compilation!!")))

let current_xml_lexer = ref
    { encoding = `Enc_utf8;
      opening_tag_state    = dummy_lexer;
      closing_tag_state    = dummy_lexer;
      text_state           = dummy_lexer;
      attribute_text_state = dummy_lexer;
      pi_state             = dummy_lexer;
      cdata_state          = dummy_lexer;
      comment_state        = dummy_lexer;
      xml_decl_state       = dummy_lexer;
      xml_document_state   = dummy_lexer;
      doctype_state        = dummy_lexer;
      dtddecl_state        = dummy_lexer;
      entity_value_state   = dummy_lexer }

let current_xquery_lexer = ref
    { language             = XQuery_1_0;
      schema_state         = Type_lexer.schema;
      default_state        = Xquery_lexer.token }

let register_lexer lx =
  let encoding = lx.encoding in
  begin
    match encoding with
    | `Enc_utf8 ->
	begin
	  set_internal_encoding `Enc_utf8;
	  current_xml_lexer := lx
	end
    | `Enc_iso88591 ->
	if Hashtbl.mem lexers_by_encoding `Enc_utf8
	then ()
	else
	  begin
	    set_internal_encoding `Enc_iso88591;
	    current_xml_lexer := lx
	  end
    | _ ->
	raise (Query (Undefined "Internal encoding should be either UTF8 or ISO-8859-1"))
  end;
  Hashtbl.add lexers_by_encoding lx.encoding lx

let set_lexer enc =
  match enc with
  | `Enc_utf8 ->
      let lx =
	try
	  Hashtbl.find lexers_by_encoding `Enc_utf8
	with
	| _ ->
	    raise (Query (Undefined ("Support for UTF8 internal encoding node compiled! You must select UTF8 at configuration time.")))
      in
      set_internal_encoding `Enc_utf8;
      current_xml_lexer := lx
  | `Enc_iso88591 ->
      let lx =
	try
	  Hashtbl.find lexers_by_encoding `Enc_iso88591
	with
	| _ ->
	    raise (Query (Undefined ("Support for ISO-8859-1 internal encoding node compiled! You must select ISO-8859-1 at configuration time.")))
      in
      set_internal_encoding `Enc_iso88591;
      current_xml_lexer := lx

(*******************)
(* Lexing wrappers *)
(*******************)

(* Note:
     These lexing functions switch according to the lexical states.
 - Jerome
 *)

(* XQuery lexer *)

let xquery_lexfun x =
  match get_current_lex_state () with
  | SchemaState
    -> !current_xquery_lexer.schema_state x
  | DefaultState
    -> !current_xquery_lexer.default_state x
  | OpeningTagState
    -> !current_xml_lexer.opening_tag_state x
  | ClosingTagState
    -> !current_xml_lexer.closing_tag_state x
  | TextState
    -> !current_xml_lexer.text_state x
  | AttributeTextDoubleState
    -> !current_xml_lexer.attribute_text_state x
  | AttributeTextSingleState
    -> !current_xml_lexer.attribute_text_state x
  | EntityIncludedInLiteralState
    -> !current_xml_lexer.attribute_text_state x
  | PIState
    -> !current_xml_lexer.pi_state x
  | CDATAState
    -> !current_xml_lexer.cdata_state x
  | CommentState
    -> !current_xml_lexer.comment_state x
  | XMLDeclState
    -> raise (Query (Internal_Error "Cannot be in XMLDecl lexical state during XQuery parsing"))
  | XMLDocumentState
    -> raise (Query (Internal_Error "Cannot be in XMLDocument lexical state during XQuery parsing"))
  | DocTypeState
    -> raise (Query (Internal_Error "Cannot be in DocType lexical state during XQuery parsing"))
  | DTDDeclState
    -> raise (Query (Internal_Error "Cannot be in DTDDecl lexical state during XQuery parsing"))
  | EntityValueState
    -> raise (Query (Internal_Error "Cannot be in EntityValue lexical state during XQuery parsing"))

(* Lexing for XML *)

let xml_lexfun x =
  match get_current_lex_state () with
  | OpeningTagState
    -> !current_xml_lexer.opening_tag_state x
  | ClosingTagState
    -> !current_xml_lexer.closing_tag_state x
  | TextState
    -> !current_xml_lexer.text_state x
  | AttributeTextDoubleState
    -> !current_xml_lexer.attribute_text_state x
  | AttributeTextSingleState
    -> !current_xml_lexer.attribute_text_state x
  | EntityIncludedInLiteralState
    -> !current_xml_lexer.attribute_text_state x
  | PIState
    -> !current_xml_lexer.pi_state x
  | CDATAState
    -> !current_xml_lexer.cdata_state x
  | CommentState
    -> !current_xml_lexer.comment_state x
  | SchemaState
    -> raise (Query (Internal_Error "Cannot be in Schema lexical state during XML parsing"))
  | DefaultState
    -> raise (Query (Internal_Error "Cannot be in Default lexical state during XML parsing"))
  | XMLDeclState
    -> !current_xml_lexer.xml_decl_state x
  | XMLDocumentState
    -> !current_xml_lexer.xml_document_state x
  | DocTypeState
    -> !current_xml_lexer.doctype_state x
  | DTDDeclState
    -> !current_xml_lexer.dtddecl_state x
  | EntityValueState
    -> !current_xml_lexer.entity_value_state x


(********************)
(* parsing wrappers *)
(********************)

(* Parser xquery *)

let rec parseFXQuery fname =
  init_xquery_lexing();
  Finfo.current_file := fname;
  let inch = open_in fname in
  try
    let lexbuf = Lexing.from_channel inch in
    try
      let ast = Xquery_parser.querymodule xquery_lexfun lexbuf in
      begin
	close_in inch;
	Finfo.current_file := "";
	ast
      end
    with Parsing.Parse_error -> 
      raise (Query(Parsing (make_finfo (Lexing.lexeme_start lexbuf)
			      (Lexing.lexeme_end lexbuf), ("Parse error in XQuery file " ^ fname))))
  with e ->
    close_in inch;
    Finfo.current_file := "";
    raise e

let rec parseStringFXQuery xquerytext =
  init_xquery_lexing();
  Finfo.current_file := "";
  let lexbuf = Lexing.from_string xquerytext in
  try
    Xquery_parser.querymodule xquery_lexfun lexbuf
  with Parsing.Parse_error -> 
    raise (Query(Parsing (make_finfo (Lexing.lexeme_start lexbuf)
			    (Lexing.lexeme_end lexbuf), ("Parse error in XQuery string")))) 

(* parse XQuery statements *)

let rec parse_statement_from_file fname =
  init_xquery_lexing();
  Finfo.current_file := fname; 
  let inch = open_in fname in
  try
    let lexbuf = Lexing.from_channel inch in
    try
      let ast = Xquery_parser.statement xquery_lexfun lexbuf in
      begin
	close_in inch;
	Finfo.current_file := "";
	ast
      end
    with Parsing.Parse_error -> 
      raise (Query(Parsing (make_finfo (Lexing.lexeme_start lexbuf)
			      (Lexing.lexeme_end lexbuf), ("Parse error in XQuery file " ^ fname))))
  with e ->
    close_in inch; 
    Finfo.current_file := "";
    raise e

let rec parse_statement_from_string xquerytext =
  init_xquery_lexing();
  Finfo.current_file := "";
  let lexbuf = Lexing.from_string xquerytext in
  try
    Xquery_parser.statement xquery_lexfun lexbuf
  with Parsing.Parse_error -> 
    raise (Query(Parsing (make_finfo (Lexing.lexeme_start lexbuf)
			    (Lexing.lexeme_end lexbuf), ("Parse error in XQuery string")))) 

(* Lexer testing for XQuery *)

let rec lexerFXQuery fname =
  init_xquery_lexing();
  Finfo.current_file := fname;
  let inch = open_in fname in
  try
    let lexbuf = Lexing.from_channel inch in
    try
      let ast = Xquery_parser.test xquery_lexfun lexbuf in
      begin
	close_in inch;
	Finfo.current_file := "";
	ast
      end
    with Parsing.Parse_error -> 
      raise (Query(Parsing (make_finfo (Lexing.lexeme_start lexbuf)
			      (Lexing.lexeme_end lexbuf), ("Lexing error in XQuery file " ^ fname))))
  with e ->
    close_in inch; 
    Finfo.current_file := "";
    raise e

let rec lexerStringFXQuery xquerytext =
  init_xquery_lexing();
  Finfo.current_file := "";
  let lexbuf = Lexing.from_string xquerytext in
  try
    Xquery_parser.test xquery_lexfun lexbuf
  with Parsing.Parse_error -> 
    raise (Query(Parsing (make_finfo (Lexing.lexeme_start lexbuf)
			    (Lexing.lexeme_end lexbuf), ("Lexing error in XQuery string")))) 

(***********************)
(* Parse XML documents *)
(***********************)

let rec parseXML fname =
  init_xml_lexing ();
  Finfo.current_file := fname;
  let inch = open_in fname in
  try
    let lexbuf = Lexing.from_channel inch in
    try
      let ast = Xquery_parser.xml_document xml_lexfun lexbuf in
      begin
	close_in inch;
	Finfo.current_file := "";
	ast
      end
    with Parsing.Parse_error -> 
      raise (Query(Parsing (make_finfo (Lexing.lexeme_start lexbuf)
			      (Lexing.lexeme_end lexbuf), ("Parse error in XML document " ^ fname))))
  with e ->
    close_in inch; 
    Finfo.current_file := "";
    raise e

let rec parseXMLString xmltext =
  init_xml_lexing ();
  Finfo.current_file := "";
  let lexbuf = Lexing.from_string xmltext in
  try
    Xquery_parser.xml_document xml_lexfun lexbuf
  with Parsing.Parse_error ->
    raise (Query(Parsing (make_finfo (Lexing.lexeme_start lexbuf)
			    (Lexing.lexeme_end lexbuf), ("Parse error in XML string")))) 

(* parser XML schemas *)

let parseXMLSchema = parseXML

let parseXMLSchemaString = parseXMLString

(*****************************)
(* Module related operations *)
(*****************************)

let mergeFXQuery q1 q2 =
  { pmodule_sorts = List.append q1.pmodule_sorts q2.pmodule_sorts;
    pmodule_functions = List.append q1.pmodule_functions q2.pmodule_functions;
    pmodule_contexts = List.append q1.pmodule_contexts q2.pmodule_contexts;
    pmodule_vars = List.append q1.pmodule_vars q2.pmodule_vars;
    pmodule_keys = List.append q1.pmodule_keys q2.pmodule_keys;
    pmodule_exprs = List.append q1.pmodule_exprs q2.pmodule_exprs }

let mergeCFXQuery cq1 cq2 =
  { pcmodule_functions = List.append cq1.pcmodule_functions cq2.pcmodule_functions;
    pcmodule_vars = List.append cq1.pcmodule_vars cq2.pcmodule_vars;
    pcmodule_keys = List.append cq1.pcmodule_keys cq2.pcmodule_keys;
    pcmodule_exprs = List.append cq1.pcmodule_exprs cq2.pcmodule_exprs }

