/***********************************************************************/
/*                                                                     */
/*                                 GALAX                               */
/*                              XQuery Engine                          */
/*                                                                     */
/*  Copyright 2001-2003, Lucent Technologies.                          */
/*  Distributed only by permission.                                    */
/*                                                                     */
/***********************************************************************/

/* $Id: xquery_parser.mly,v 1.86 2003/06/30 02:58:19 simeon Exp $ */

/*  
 *  XQuery Parser
 */

%{

open Namespace
open Sym
open Datatypes
open Xquery_ast
open Xquery_astutil

open Xml_ast
open Xml_astutil

open Tokens
open Error
open Finfo

open Format

(* Checks consitency between opening and closing tags *)

let same_tag os cs =
  if (os = cs)
  then ()
  else
    raise (Query (Parsing (Finfo.parsing_locinfo (),
			   "Opening tag: "
			   ^ (Namespace.string_of_uqname os)
			   ^ " and closing tag: "
			   ^ (Namespace.string_of_uqname cs)
			   ^ " mismatched")))

(* Checks that all attributes within an element tag are distinct *)

let distinct_attributes qname previous_attributes =
  let same_symbol x y =
    match y.pexpr_desc with
    | (EAttrFixed (UAttrSym sym,_)) ->
	x = uattr_name sym
    | _ ->
	raise (Query (Internal_Error ("Attributes do not contain attributes?")))
  in
  if (List.exists (same_symbol qname) previous_attributes)
  then
    raise (Query (Parsing (Finfo.parsing_locinfo (),
			   "Opening tag contains duplicate attribute "
			   ^ (Namespace.string_of_uqname qname))))
  else
    ()

(* Checks that all attributes within an element tag are distinct *)

let distinct_xattributes qname previous_attributes =
  let same_symbol x y =
    match y with
    | (UAttrSym sym,_) ->
	x = uattr_name sym
    | _ ->
	raise (Query (Internal_Error ("Attributes do not contain attributes?")))
  in
  if (List.exists (same_symbol qname) previous_attributes)
  then
    raise (Query (Parsing (Finfo.parsing_locinfo (),
			   "Opening tag contains duplicate attribute "
			   ^ (Namespace.string_of_uqname qname))))
  else
    ()

(* Function used to resolve built-in types *)

let wrap_group_ref qn =
  try
    lookup_bltin_type qn
  with
  | Not_found ->
      raise (Query (Parsing (Finfo.parsing_locinfo (),
			     "Built-in type: "
			     ^ (Namespace.string_of_uqname qn)
			     ^ " not found")))

(* Deals with entity references in XQuery *)

let xquery_parse_context =
  Parse_context.build_parse_context ()

%}

/**********/
/* tokens */
/**********/

%token <Namespace.uqname> XQNAME
%token <Namespace.ncname> XNCNAME
%token <Datatypes.xs_string> XSTRING

%token XMLPI XVERSION XSDDECL XSINGLEQUOTE XDOUBLEQUOTE
%token <string> XDECLNAME
%token <string> XENCODING

%token DOCTYPE DOCTYPECLOSE
%token DOCTYPE_SYSTEM DOCTYPE_PUBLIC
%token DTDDECLOPEN DTDDECLCLOSE

/* XML whitespace */

%token S

/* XML markup tokens */

%token LOPENINGTAG LCLOSINGTAG
%token ROPENINGTAG RCLOSINGTAG
%token REMPTYELEMENT
%token <Datatypes.xdt_untyped> TEXTLCLOSINGTAG
%token <Datatypes.xdt_untyped> TEXTLOPENINGTAG
%token <Datatypes.xdt_untyped> TEXTLCURLY

%token <Namespace.ncname> OPENINGPI
%token CLOSINGPI
%token LOPENINGCOMMENT
%token <Datatypes.xdt_untyped> TEXTCLOSINGPI
%token <Datatypes.xdt_untyped> TEXTRCLOSINGCOMMENT
%token <Datatypes.xdt_untyped * Namespace.ncname> TEXTOPENINGPI
%token <Datatypes.xdt_untyped> TEXTLOPENINGCOMMENT

%token <Datatypes.xdt_untyped> ATTRIBUTETEXT ATTRIBUTETEXTLCURLY

%token <Datatypes.xdt_untyped * int> TEXTCHARREF
%token <Datatypes.xdt_untyped * string> TEXTENTITYREF

/* DTD tokens */

%token <Namespace.ncname> PEREF
%token PERCENT
%token <Namespace.ncname> NMTOKEN
%token REQUIRED IMPLIED FIXED PCDATA
%token ELEMENTDECL ATTLISTDECL ENTITYDECL NOTATIONDECL
%token BEGINDECL ENDDECL BODYDECL

/* Query prolog tokens */

%token NAMESPACE DECLARENAMESPACE DEFAULTELEMENT DEFAULTFUNCTION
%token DEFINEKEY
%token DEFINEFUNCTION DEFINESIGNATURE
%token DEFINEGLOBAL
%token IMPORTSCHEMA

/* Type declaration tokens */

%token <Namespace.uqname> DEFINETYPE
%token <Namespace.uqname> DEFINEELEMENT
%token <Namespace.uqname> DEFINEATTRIBUTE
%token DOCUMENTOF ELEMENT ELEMENTOF ATTRIBUTE DOCUMENT EMPTY
%token TYPE NONE QUESTION NODE ITEM TEXTNODE ATOMIC NUMERIC

/* XQuery words */

%token STABLE ASCENDING DESCENDING EMPTYGREATEST EMPTYLEAST
%token OR AND
%token STAR MULT DIV IDIV MOD MINUS PLUS
%token IN SATISFIES RETURN THEN ELSE TO WHERE END
%token INTERSECT UNION EXCEPT
%token PRECEDES FOLLOWS
%token CASE INSTANCEOF DEFAULT
%token IF TYPESWITCH STABLESORTBY SORTBY ORDERBY
%token ASSERTAS CASTAS CASTABLEAS TREATAS VALIDATEAS
%token SOME EVERY FOR LET
%token AS AT

/* Update tokens */

%token INSERT REPLACE DELETE ASFIRST ASLAST VALUEOF INTO WITH BEFORE AFTER UPDATE

/* Names */

%token <Namespace.ncname> NCNAME
%token <Namespace.uqname> QNAME

/* XPath tokens */

%token <Xquery_ast.axis> AXIS
%token PROCESSINGINSTRUCTIONLPAR COMMENTLPAR TEXTLPAR NODELPAR

/* XQuery symbols */

%token COLON LTCOLON BACKSLASHSLASH
%token SLASH SLASHSLASH DOT DOTDOT
%token EQ EQUALS
%token NE NOTEQUALS
%token IS ISNOT
%token LTE GTE LTEQUALS GTEQUALS
%token COLONEQUALS
%token LTOP GTOP LT GT
%token PLUS MINUS
%token BAR ATSIGN COMMA AMPERSAND SEMICOLON
%token LPAR RPAR RPARLPAR LBRACK RBRACK LCURLY RCURLY
%token DOLLAR
%token <Namespace.uqname> FUNCTIONNAMELPAR

/* XQuery literals */

%token <Num.num> DECIMAL
%token <Decimal._integer> INT
%token <float> DOUBLE
%token <string> STRING

/* End of buffer/file */

%token EOF


/**************/
/* precedence */
/**************/

%left SORTBY STABLESORTBY
%left OR
%left AND
%left PRECEDES FOLLOWS
%left BAR
%left COMMA
%left AMPERSAND
%nonassoc STAR
%left FOR LET WHERE RETURN
%left IF THEN ELSE
%left SOME SATISFIES
%left EVERY
%left CASTABLEAS
%left EQ EQUALS NE NOTEQUALS IS ISNOT
%left LTOP GTOP LTE GTE LT GT LTEQUALS GTEQUALS
%left INSTANCEOF
%left TO
%left PLUS MINUS
%left DIV IDIV MOD MULT
%nonassoc QUESTION
%left UNION EBAR
%left INTERSECT EXCEPT
%left LTCOLON BACKSLASHSLASH
%nonassoc UPLUS
%nonassoc UMINUS


/****************/
/* Entry points */
/****************/


%start xml_document
%type <Xml_ast.xml_document> xml_document

%start test
%type <Tokens.tokens> test

%start statement
%type <Xquery_ast.statement> statement

%start querymodule
%type <Xquery_ast.query_module> querymodule

%%


/*****************/
/* XML documents */
/*****************/

xml_document:
  | prolog xml_root_element miscs
      { let (xmldecl,miscs1) = $1
        and root_element = $2
        and miscs2 = $3 in
        let prolog =
	  { xml_decl = xmldecl;
	    xml_misc_prolog = miscs1 }
	in
        { xml_prolog = prolog;
	  xml_root = root_element;
	  xml_misc = miscs2 } }
;

prolog:
  | xmldecl miscs_or_doctype
      { ($1,$2) }
;

xmldecl:
  | /* No XML Declaration */
      { None }
  | XMLPI version closingxmlpi
      { Some ($2,None,None) }
  | XMLPI version encoding closingxmlpi
      { Some ($2,Some $3,None) }
  | XMLPI version sddecl closingxmlpi
      { Some ($2,None, Some $3) }
  | XMLPI version encoding sddecl closingxmlpi
      { Some ($2,Some $3, Some $4) }
;

closingxmlpi:
  | CLOSINGPI
      { () }
  | S CLOSINGPI
      { () }
;

version:
  | XVERSION XSINGLEQUOTE XDECLNAME XSINGLEQUOTE
      { $3 }
  | XVERSION XDOUBLEQUOTE XDECLNAME XDOUBLEQUOTE
      { $3 }
  | S XVERSION XSINGLEQUOTE XDECLNAME XSINGLEQUOTE
      { $4 }
  | S XVERSION XDOUBLEQUOTE XDECLNAME XDOUBLEQUOTE
      { $4 }
;

encoding:
  | XENCODING
      { Netconversion.encoding_of_string $1 }
  | S XENCODING
      { Netconversion.encoding_of_string $2 }
;

sddecl:
  | XSDDECL XSINGLEQUOTE XDECLNAME XSINGLEQUOTE
      { let sddecl = $3 in
        if (sddecl = "yes") || (sddecl = "no")
        then
          sddecl
	else
          raise (Query (Parsing (Finfo.parsing_locinfo (), "Wrong encoding declaration"))) }
  | XSDDECL XDOUBLEQUOTE XDECLNAME XDOUBLEQUOTE
      { let sddecl = $3 in
        if (sddecl = "yes") || (sddecl = "no")
        then
          sddecl
	else
          raise (Query (Parsing (Finfo.parsing_locinfo (), "Wrong encoding declaration"))) }
  | S XSDDECL XSINGLEQUOTE XDECLNAME XSINGLEQUOTE
      { let sddecl = $4 in
        if (sddecl = "yes") || (sddecl = "no")
        then
          sddecl
	else
          raise (Query (Parsing (Finfo.parsing_locinfo (), "Wrong encoding declaration"))) }
  | S XSDDECL XDOUBLEQUOTE XDECLNAME XDOUBLEQUOTE
      { let sddecl = $4 in
        if (sddecl = "yes") || (sddecl = "no")
        then
          sddecl
	else
          raise (Query (Parsing (Finfo.parsing_locinfo (), "Wrong encoding declaration"))) }
;

miscs_or_doctype:
  | S miscs_or_doctype  /* Skip whitespace here */
      { $2 }
  | OPENINGPI TEXTCLOSINGPI miscs_or_doctype
      { let target_pi = $1 in
        if !Conf.xml_pis_and_comments then
          (XPI (target_pi, $2)) :: $3
        else
          $3 }
  | LOPENINGCOMMENT TEXTRCLOSINGCOMMENT miscs_or_doctype
      { if !Conf.xml_pis_and_comments then
          (XComment $2) :: $3
        else
          $3 }
  | DOCTYPE S QNAME opt_dtddecls DOCTYPECLOSE miscs
      { $6 }
  | DOCTYPE S QNAME S opt_dtddecls DOCTYPECLOSE miscs
      { $7 }
  | DOCTYPE S QNAME S extid opt_dtddecls DOCTYPECLOSE miscs
      { $8 }
  | DOCTYPE S QNAME S extid S opt_dtddecls DOCTYPECLOSE miscs
      { $9 }
  | /* empty */
      { [] }
;

extid:
  | DOCTYPE_SYSTEM S STRING
      { () }
  | DOCTYPE_PUBLIC S STRING S STRING
      { () }
;

opt_dtddecls:
  | /* nothing here */
      { () }
  | DTDDECLOPEN DTDDECLCLOSE
      { () }
  | DTDDECLOPEN DTDDECLCLOSE S
      { () }
;

miscs:
  | S miscs /* Skip whitespace here */
      { $2 }
  | OPENINGPI TEXTCLOSINGPI miscs
      { let target_pi = $1 in
        if !Conf.xml_pis_and_comments then
          (XPI (target_pi, $2)) :: $3
        else
          $3 }
  | LOPENINGCOMMENT TEXTRCLOSINGCOMMENT miscs
      { if !Conf.xml_pis_and_comments then
          (XComment $2) :: $3
        else
          $3 }
  | /* empty */
      { [] }
;

xml_root_element:
    /* Empty element */
  | TEXTLOPENINGTAG qname xattrlist REMPTYELEMENT
      { (XElem (UElemSym (uelem_symbol $2), $3, [])) }

    /* Full element */
  | TEXTLOPENINGTAG qname xattrlist ROPENINGTAG element_content qname opt_whitespace RCLOSINGTAG
      { let os = $2
	and cs = $6 in
        same_tag os cs;
        (XElem (UElemSym (uelem_symbol $2), $3, $5)) }

      /* Note:
           Since we are parsing the document root, the text content
           before the opening tag will always be empty.
         - Jerome */
;

opt_whitespace:
  |
      { () }
  | S
      { () }
;

element_content:
  | TEXTLCLOSINGTAG
      { (XText $1) :: [] }
  | LCLOSINGTAG
      { [] }

  | TEXTLOPENINGTAG qname xattrlist REMPTYELEMENT element_content
      { (XText $1) :: (XElem (UElemSym (uelem_symbol $2), $3, [])) :: $5 }
  | LOPENINGTAG qname xattrlist REMPTYELEMENT element_content
      { (XElem (UElemSym (uelem_symbol $2), $3, [])) :: $5 }

  | TEXTLOPENINGTAG qname xattrlist ROPENINGTAG element_content qname opt_whitespace RCLOSINGTAG element_content
      { let os = $2
	and cs = $6 in
        same_tag os cs;
        (XText $1) :: (XElem (UElemSym (uelem_symbol $2),$3, $5)) :: $9 }
  | LOPENINGTAG qname xattrlist ROPENINGTAG element_content qname opt_whitespace RCLOSINGTAG element_content
      { let os = $2
	and cs = $6 in
        same_tag os cs;
        (XElem (UElemSym (uelem_symbol $2),$3, $5)) :: $9 }

    /* Processing instruction */
  | TEXTOPENINGPI TEXTCLOSINGPI element_content
      { let (text, target_pi) = $1 in
        if !Conf.xml_pis_and_comments then
          (XText text) :: (XPI (target_pi,$2)) :: $3
	else
          (XText text) :: $3 }
  | OPENINGPI TEXTCLOSINGPI element_content
      { let target_pi = $1 in
        if !Conf.xml_pis_and_comments then
          (XPI (target_pi, $2)) :: $3
        else
          $3 }

    /* Comment */
  | TEXTLOPENINGCOMMENT TEXTRCLOSINGCOMMENT element_content
      { if !Conf.xml_pis_and_comments then
          (XText $1) :: (XComment $2) :: $3
        else
          (XText $1) :: $3 }
  | LOPENINGCOMMENT TEXTRCLOSINGCOMMENT element_content
      { if !Conf.xml_pis_and_comments then
          (XComment $2) :: $3
        else
          $3 }

    /* Character references */
  | TEXTCHARREF element_content
      { let (text,charref) = $1 in
        let charref_text = Glx_pxp.character (Encoding.get_internal_encoding ()) charref in
        (XText text) :: (XText charref_text) :: $2 }

    /* Entity references */
  | TEXTENTITYREF element_content
      { let (text,entityref) = $1 in
        let entity_ref_text = Parse_context.get_general_entity xquery_parse_context (parsing_locinfo()) entityref in
        (XText text) :: (XText entity_ref_text) :: $2 }
;

/* Attributes inside element tags subrules */

xattrlist:
  | /* empty */
      { [] }
  | S
      { [] }
  | S qname eq xattributecontent xattrlist
      { let previous_attributes = $5 in
        let qname = $2 in
	begin
	  distinct_xattributes qname previous_attributes;
	  (UAttrSym (uattr_symbol qname), $4) :: previous_attributes
	end }
;

xattributecontent:
  | ATTRIBUTETEXT
      { $1 }

    /* Char references */
  | TEXTCHARREF xattributecontent
      { let (text,charref) = $1 in
        let charref_text = Glx_pxp.character (Encoding.get_internal_encoding ()) charref in
        text ^ charref_text ^ $2 }

    /* Entity references */
  | TEXTENTITYREF xattributecontent
      { let (text,entityref) = $1 in
        let entity_ref_text = Parse_context.get_general_entity xquery_parse_context (Finfo.parsing_locinfo ()) entityref in
        text ^ entity_ref_text ^ $2 }
;

eq:
  | EQUALS { () }
  | S EQUALS { () }
  | EQUALS S { () }
  | S EQUALS S { () }
;


/**********/
/* XQuery */
/**********/

querymodule:
  | contextdecllist functiondefnlist statements
      { let (nsd,td,vd,kd) = $1
        and fd = $2
        and statements = $3 in
        { pmodule_sorts = td;
	  pmodule_functions = fd;
	  pmodule_contexts = nsd;
	  pmodule_vars = vd;
	  pmodule_keys = kd;
	  pmodule_exprs = statements } }
;

statements:
  | /* no statement */
      { [] }
  | statement
      { $1 :: [] }
  | statement SEMICOLON statements
      { $1 :: $3 }
;

statement:
  | exprsequence
      { let expr = mkexpr (EList $1) in
        mkstatement (SExpression expr) }
  | update
      { let update = $1 in
        mkstatement (SUpdate update) }
;

contextdecllist:
  |   /* empty */
      { ([],[],[],[]) }

  /* namespace declaration */
  | DECLARENAMESPACE NCNAME EQUALS STRING contextdecllist
      { let (nsd,td,vd,kd) = $5 in
        (mkcontext_decl (ENamespaceDecl ($2,NSUri $4)) :: nsd, td, vd,kd) }

  | DEFAULTELEMENT NAMESPACE EQUALS STRING contextdecllist
      { let (nsd,td,vd,kd) = $5 in
        (mkcontext_decl (EDefaultElementNamespaceDecl (NSUri $4)) :: nsd, td, vd,kd) }
 
  | DEFAULTFUNCTION NAMESPACE EQUALS STRING contextdecllist
      { let (nsd,td,vd,kd) = $5 in
        (mkcontext_decl (EDefaultFunctionNamespaceDecl (NSUri $4)) :: nsd, td, vd,kd) }
 
 /* Schema declaration */
  | IMPORTSCHEMA STRING contextdecllist
      { let (nsd,td,vd,kd) = $3 in
        (mkcontext_decl (ESchemaDecl (None,$2,None)) :: nsd, td, vd,kd) }

  | IMPORTSCHEMA NAMESPACE NCNAME EQUALS STRING contextdecllist
      { let (nsd,td,vd,kd) = $6 in
        (mkcontext_decl (ESchemaDecl (Some $3,$5,None)) :: nsd, td, vd,kd) }

  | IMPORTSCHEMA DEFAULTELEMENT NAMESPACE EQUALS STRING contextdecllist
      { let (nsd,td,vd,kd) = $6 in
        ((mkcontext_decl (EDefaultSchemaDecl $5)) :: nsd, td, vd,kd) }

  | typedefinition contextdecllist
      { let (nsd,td,vd,kd) = $2 in
        let moretd = $1 in
        (nsd, moretd :: td, vd, kd) }

  /* Extension to XQuery */
  | DEFINEGLOBAL qname LCURLY exprsequence RCURLY contextdecllist
      { let (nsd,td,vd,kd) = $6 in
        let varname = $2 in
        let expr = mkexpr (EList $4) in
        let newvd = mkvar_decl (varname,expr) in
        (nsd, td, newvd :: vd, kd) }

  | DEFINEKEY STRING LCURLY exprsequence RCURLY LCURLY exprsequence RCURLY contextdecllist
      { let (nsd,td,vd,kd) = $9 in
        let keyname = $2 in
        let expr1 = mkexpr (EList $4) in
        let expr2 = mkexpr (EList $7) in
        let newkd = mkkey_def (keyname,expr1,expr2) in
        (nsd, td, vd, newkd :: kd) }
;

functiondefnlist:
  | /* Empty */ 
      { [] }
      /* Function definitions */
  | DEFINEFUNCTION FUNCTIONNAMELPAR RPAR LCURLY expr RCURLY functiondefnlist
      { let fd = $7 in
        let anytype = mkdatatype (DTTypeRef (RGroupSym anyitem), Some(Occurrence.occurs 0, Occurrence.unbounded)) in
        mkfunction_def ($2,([],anytype),EFunctionUser ([],$5)) :: fd }
  | DEFINEFUNCTION FUNCTIONNAMELPAR paramlist RPAR LCURLY expr RCURLY functiondefnlist
      { let fd = $8 in
        let (input_types,input_vars) = $3 in
        let anytype = mkdatatype (DTTypeRef (RGroupSym anyitem), Some(Occurrence.occurs 0, Occurrence.unbounded)) in
        mkfunction_def ($2,(input_types,anytype),EFunctionUser (input_vars,$6)) :: fd }
  | DEFINEFUNCTION FUNCTIONNAMELPAR RPAR AS datatype LCURLY expr RCURLY functiondefnlist
      { let fd = $9 in
        mkfunction_def ($2,([],$5),EFunctionUser ([],$7)) :: fd }
  | DEFINEFUNCTION FUNCTIONNAMELPAR paramlist RPAR AS datatype LCURLY expr RCURLY functiondefnlist
      { let fd = $10 in
        let (input_types,input_vars) = $3 in
        mkfunction_def ($2,(input_types,$6),EFunctionUser (input_vars,$8)) :: fd }

      /* Function signatures */
  | DEFINESIGNATURE FUNCTIONNAMELPAR RPAR AS return_param functiondefnlist
      { let fd = $6 in
        mkfunction_def ($2,([],$5), EFunctionBltIn ($2)) :: fd }

  | DEFINESIGNATURE FUNCTIONNAMELPAR sig_paramlist RPAR AS return_param functiondefnlist
      { let fd = $7 in
        let input_types = $3 in
        mkfunction_def ($2,(input_types,$6), EFunctionBltIn ($2)) :: fd }

  | DEFINESIGNATURE LPAR op_name RPAR LPAR sig_paramlist RPAR AS return_param functiondefnlist
      { let fd = $10 in
        let input_types = $6 in
        mkfunction_def ($3,(input_types,$9), EFunctionBltIn ($3)) :: fd }

;

paramlist:
  | param
      { let (t,v) = $1 in
        (t :: [], v :: []) }
  | param COMMA paramlist
      { let (t,v) = $1
        and (pt,pv) = $3 in
        (t :: pt, v :: pv) }
;

param:
  | variable AS datatype
      { ($3,$1) }
  | variable
      { let anytype = mkdatatype (DTItem, Some(Occurrence.occurs 0, Occurrence.unbounded)) in
        (anytype,$1) }
;

sig_paramlist:
  | datatype
      { let t = $1 in
        (t :: []) }
  | datatype COMMA sig_paramlist
      { let t = $1
        and pt = $3 in
        (t :: pt) }
;

return_param:
/*| limitedmodel
      { $1 } */
  | LPAR datatype RPAR
      { $2 }
;

op_name:
  | MINUS { (NSDefaultFunctionPrefix, "(-)") }
  | PLUS  { (NSDefaultFunctionPrefix, "(+)") }
  | STAR  { (NSDefaultFunctionPrefix, "(*)") }
  | DIV   { (NSDefaultFunctionPrefix, "(div )") }
  | IDIV  { (NSDefaultFunctionPrefix, "(idiv )") }
  | MOD   { (NSDefaultFunctionPrefix, "(mod )") }
;

/**********************/
/* XQuery Expressions */
/**********************/

expr:
      /* Sort expression */
  | expr SORTBY LPAR sortspeclist RPAR
      { mkexpr (ESort($1, (NonStable, $4))) }

  | expr STABLESORTBY LPAR sortspeclist RPAR
      { mkexpr (ESort($1, (Stable, $4))) }

      /* Or expression */
  | expr OR expr
      { mkexpr (EBinaryOp($1,BEOr,$3)) }

      /* And expression */
  | expr AND expr
      { mkexpr (EBinaryOp($1,BEAnd,$3)) }

      /* Precedes Follows expressions */
  | expr PRECEDES expr
      { mkexpr (EBinaryOp($1,BEPrecedes,$3)) }
  | expr FOLLOWS expr
      { mkexpr (EBinaryOp($1,BEFollows,$3)) }

      /* FLWR expression */
  | for_let_clause where_order_clause return_clause
      { let fl = $1
        and (where,order) = $2
        and return = $3 in
        mkexpr (EFLWR (fl,where,order,return)) }

      /* If expression */
  | IF LPAR expr RPAR THEN expr ELSE expr
      { mkexpr (EIf ($3, $6, $8)) }

      /* Some expression */
  | SOME bindings SATISFIES expr
      { mkexpr (ESome ($2, $4)) }

      /* Every expression */
  | EVERY bindings SATISFIES expr
      { mkexpr (EEvery ($2, $4)) }

      /* Typeswitch expression */
  | TYPESWITCH LPAR expr RPAR case_clause_list
      { mkexpr (ETypeswitch ($3, $5)) }

      /* Equality expression */
  | expr EQ expr
      { mkexpr (EBinaryOp($1,BEEq,$3)) }
  | expr EQUALS expr
      { mkexpr (EBinaryOp($1,BEEqual,$3)) }
  | expr IS expr
      { mkexpr (EBinaryOp($1,BEIs,$3)) }
  | expr NE expr
      { mkexpr (EBinaryOp($1,BENEq,$3)) }
  | expr NOTEQUALS expr
      { mkexpr (EBinaryOp($1,BENEqual,$3)) }
  | expr ISNOT expr
      { mkexpr (EBinaryOp($1,BEIsNot,$3)) }

      /* Relational expression */
  | expr LT expr
      { mkexpr (EBinaryOp($1,BELt,$3)) }
  | expr GT expr
      { mkexpr (EBinaryOp($1,BEGt,$3)) }
  | expr LTOP expr
      { mkexpr (EBinaryOp($1,BELtOp,$3)) }
  | expr GTOP expr
      { mkexpr (EBinaryOp($1,BEGtOp,$3)) }

  | expr LTE expr
      { mkexpr (EBinaryOp($1,BELte,$3)) }
  | expr GTE expr
      { mkexpr (EBinaryOp($1,BEGte,$3)) }

  | expr LTEQUALS expr
      { mkexpr (EBinaryOp($1,BELteq,$3)) }
  | expr GTEQUALS expr
      { mkexpr (EBinaryOp($1,BEGteq,$3)) }

      /* Instance Of expression */
  | expr INSTANCEOF limitedmodel
      { let dt = mkdatatype ($3, None)
	in mkexpr (EInstanceOf ($1,dt)) }
  | expr INSTANCEOF limitedmodel STAR
      { let dt = mkdatatype ($3, Some(Occurrence.occurs 0, Occurrence.unbounded))
	in mkexpr (EInstanceOf ($1,dt)) }
  | expr INSTANCEOF limitedmodel PLUS  %prec UPLUS
      { let dt = mkdatatype ($3, Some(Occurrence.occurs 1, Occurrence.unbounded))
	in mkexpr (EInstanceOf ($1,dt)) }
  | expr INSTANCEOF limitedmodel QUESTION
      { let dt = mkdatatype ($3, Some(Occurrence.occurs 0, Occurrence.occurs 1))
	in mkexpr (EInstanceOf ($1,dt)) }

      /* Range expression */
  | expr TO expr
      { mkexpr (ERange ($1,$3)) }

      /* Additive expression */
  | expr PLUS expr
      { mkexpr (EBinaryOp($1,BEPlus,$3)) }
  | expr MINUS expr
      { mkexpr (EBinaryOp($1,BEMinus,$3)) }

      /* Multiplicative expression */
  | expr MULT expr
      { mkexpr (EBinaryOp($1,BEMult,$3)) }
  | expr DIV expr
      { mkexpr (EBinaryOp($1,BEDiv,$3)) }
  | expr IDIV expr
      { mkexpr (EBinaryOp($1,BEIDiv,$3)) }
  | expr MOD expr
      { mkexpr (EBinaryOp($1,BEMod,$3)) }

      /* Unary expression */
  | MINUS expr  %prec UMINUS
      { mkexpr (EUnaryOp(UEMinus,$2)) }
  | PLUS expr   %prec UMINUS
      { mkexpr (EUnaryOp(UEPlus,$2)) }

      /* Union expression */
  | expr UNION expr
      { mkexpr (EBinaryOp($1,BEUnion,$3)) }
  | expr BAR expr   %prec EBAR
      { mkexpr (EBinaryOp($1,BEBar,$3)) }

      /* Intersect expression */
  | expr INTERSECT expr
      { mkexpr (EBinaryOp($1,BEIntersect,$3)) }
  | expr EXCEPT expr
      { mkexpr (EBinaryOp($1,BEExcept,$3)) }

      /* Castable expression */
  | expr CASTABLEAS simpletype
      { mkexpr (ECastable ($1,$3)) }

  | pathexpr
      { $1 }
;


sortspeclist:
  | sortspec
      { $1 :: [] }
  | sortspec COMMA sortspeclist
      { $1 :: $3 }
;

sortspec:
  | expr
      { ($1,Ascending,EmptyGreatest) }
  | expr ASCENDING
      { ($1,Ascending,EmptyGreatest) }
  | expr DESCENDING
      { ($1,Descending,EmptyGreatest) }
  | expr EMPTYGREATEST
      { ($1,Ascending,EmptyGreatest) }
  | expr ASCENDING EMPTYGREATEST
      { ($1,Ascending,EmptyGreatest) }
  | expr DESCENDING EMPTYGREATEST
      { ($1,Descending,EmptyGreatest) }
  | expr EMPTYLEAST
      { ($1,Ascending,EmptyLeast) }
  | expr ASCENDING EMPTYLEAST
      { ($1,Ascending,EmptyLeast) }
  | expr DESCENDING EMPTYLEAST
      { ($1,Descending,EmptyLeast) }
;

bindings:
  | variable opt_datatype IN expr
      {  [($2, $1, $4)] }
  | variable opt_datatype IN expr COMMA bindings
      {  ($2, $1, $4) :: $6 }
;

opt_at:
  | /* No at variable */
      { None }
  | AT variable
      { Some $2 }
;

for_let_clause:
  | FOR variable opt_datatype opt_at IN expr for_clause
      { (mkfl_expr (EFor ($3, $2, $4, $6))) :: $7 }
  | LET variable opt_datatype COLONEQUALS expr let_clause
      { (mkfl_expr (ELet ($3, $2, $5))) :: $6 }
;

for_clause:
  | /* empty */
      { [] }
  | COMMA variable opt_datatype opt_at IN expr for_clause
      { (mkfl_expr (EFor ($3, $2, $4, $6))) :: $7 }
  | for_let_clause
      { $1 }
;

let_clause:
  | /* empty */
      { [] }
  | COMMA variable opt_datatype COLONEQUALS expr let_clause
      { (mkfl_expr (ELet ($3, $2, $5))) :: $6 }
  | for_let_clause
      { $1 }
;

where_order_clause:
  | /* empty */
      { (None,None) }
  | where_clause
      { (Some $1,None) }
  | order_clause
      { (None, Some $1) }
  | where_clause order_clause
      { (Some $1, Some $2) }
;

where_clause:
  | WHERE expr
      { $2 }
;

order_clause:
  | ORDERBY sortspeclist
      { (NonStable, $2) }
  | STABLE ORDERBY sortspeclist
      { (Stable, $3) }
;

return_clause:
  | RETURN expr
      { $2 }
;

opt_variable:
  | /* empty */
      { None }
  | variable
      { Some $1 }
;

opt_variable_as:
  | /* empty */
      { None }
  | variable AS
      { Some $1 }
;

case_clause_list:
  | CASE opt_variable_as datatype RETURN expr DEFAULT opt_variable RETURN expr
      { (mkpattern (Case ($3)), $2, $5) :: (mkpattern (Default), $7, $9) :: [] }
  | CASE opt_variable_as datatype RETURN expr case_clause_list
      { (mkpattern (Case ($3)), $2, $5) :: $6 }
;

pathexpr:
  | relativepathexpr
      { $1 }
  | SLASH
      { mkexpr ERoot }
  | SLASH relativepathexpr
      { mkexpr (EPath (PSlash(mkexpr ERoot, $2))) }
  | SLASHSLASH relativepathexpr
      { mkexpr (EPath (PSlashSlash(mkexpr ERoot, $2))) }
;

relativepathexpr:
  | stepexpr
      { $1 }
  | stepexpr SLASH relativepathexpr
      { mkexpr (EPath (PSlash($1, $3))) }
  | stepexpr SLASHSLASH relativepathexpr
      { mkexpr (EPath (PSlashSlash($1, $3))) }
;

stepexpr:
  | axisstepexpr
      { $1 }
  | otherstepexpr
      { $1 }
;

axisstepexpr:
  | AXIS nodetest
      { let axis = $1 in
        mkexpr (EPath(PAxis (axis,$2))) }
  | AXIS nodetest stepqualifiers
      { let axis = $1 in
        let axis_expr = mkexpr (EPath(PAxis (axis,$2))) in 
        mkexpr (EPath(PStepQualifiers (axis_expr,$3))) }
  | ATSIGN nodetest
      { mkexpr (EPath(PAxis (Attribute,$2))) }
  | ATSIGN nodetest stepqualifiers
      { let axis_expr = mkexpr (EPath(PAxis (Attribute,$2))) in 
        mkexpr (EPath(PStepQualifiers (axis_expr,$3))) }
;

otherstepexpr:
  | primaryexpr
      { $1 }
  | primaryexpr stepqualifiers
      { mkexpr (EPath (PStepQualifiers ($1,$2))) }
;

stepqualifiers:
  | LBRACK expr RBRACK
      { (mkstep_qualifier (PredicateQualifier $2)) :: [] }
  | LBRACK expr RBRACK stepqualifiers
      { (mkstep_qualifier (PredicateQualifier $2)) :: $4 }
;

primaryexpr:
      /* Self */
  | DOT
      { mkexpr ESelf }

      /* Parent */
  | DOTDOT
      { mkexpr EParent }

      /* Node test */
  | nodetest
      { mkexpr (EPath(PAxis (Child,$1))) }  /* Jerome: I am still not sure if adding the Child axis at parsing level here is safe */

      /* Variable */
  | variable
      { mkexpr (EVar $1) }

      /* Literal */
  | literal
      { $1 }

      /* Function call */
  | FUNCTIONNAMELPAR RPAR   /* Warning: this one is not in the grammar right now - JS */
      { mkexpr (EApp ($1,[])) }
  | FUNCTIONNAMELPAR exprsequence RPAR
      { mkexpr (EApp ($1,$2)) }

      /* Parenthesized expression */
  | LPAR RPAR
      { mkexpr (EList []) }

  | LPAR exprsequence RPAR
      { mkexpr (EList $2) }

      /* Cast expression */
  | CASTAS datatype LPAR expr RPAR
      { mkexpr (ECast ($4,$2)) }

  | TREATAS datatype LPAR expr RPAR
      { mkexpr (ETreat ($4,$2)) }

  | VALIDATEAS datatype LPAR expr RPAR
      { mkexpr (EValidate ($4,$2)) }

      /* Document node constructors */
  | DOCUMENT LCURLY exprsequence RCURLY
      { mkexpr (EDocument $3) }

      /* Text node constructors */
  | TEXTNODE LCURLY exprsequence RCURLY
      { mkexpr (ETextComputed $3) }

      /* Element Constructors */
  | ELEMENT qname LCURLY exprsequence RCURLY
      { mkexpr (EElemFixed (UElemSym (uelem_symbol $2),[],$4)) }

  | ATTRIBUTE qname LCURLY exprsequence RCURLY
      { mkexpr (EAttrFixed (UAttrSym (uattr_symbol $2),$4)) }

      /* Computed element constructors */
  | ELEMENT LCURLY expr RCURLY LCURLY exprsequence RCURLY
      { mkexpr (EElemComputed ($3, $6)) }

  | ATTRIBUTE LCURLY expr RCURLY LCURLY exprsequence RCURLY
      { mkexpr (EAttrComputed ($3, $6)) }

      /* Escapes into XML */
    /* Empty element */
  | LOPENINGTAG qname attrlist REMPTYELEMENT
      { mkexpr (EElemFixed (UElemSym (uelem_symbol $2), $3, [])) }

    /* Element having a content inside */
  | LOPENINGTAG qname attrlist ROPENINGTAG elementcontent qname opt_whitespace RCLOSINGTAG
      { let os = $2
	and cs = $6 in
        same_tag os cs;
        mkexpr (EElemFixed ((UElemSym (uelem_symbol os)), $3, $5)) }

    /* Processing instruction */
  | OPENINGPI TEXTCLOSINGPI
      { mkexpr (EPI ($1,$2)) }

    /* Comment */
  | LOPENINGCOMMENT TEXTRCLOSINGCOMMENT
      { mkexpr (EComment $2) }

    /* Subtype assertion: extension to XQuery */
  |  ASSERTAS LPAR datatype LTCOLON datatype RPAR
      { mkexpr (ESubtype($3, $5)) }

    /* Intersection assertion: extension to XQuery */
  |  ASSERTAS LPAR datatype BACKSLASHSLASH datatype RPAR
      { mkexpr (EIntersects($3, $5)) }
;

/* Expression sequence */

exprsequence:
  | expr
      { $1 :: [] }
  | expr COMMA exprsequence
      { $1 :: $3 }
;

/* Variable subrule */

variable:
  | DOLLAR qnamekeyword
      { $2 }
;

/* Literal subrules */

literal:
  | numericliteral
      { $1 }
  | stringliteral
      { $1 }
;

numericliteral:
  | integer
      { mkexpr (EScalar (AtomicInteger $1)) }
  | decimal
      { mkexpr (EScalar (AtomicDecimal $1)) }
  | DOUBLE
      { mkexpr (EScalar (AtomicDouble $1)) }
;

stringliteral:
  | STRING
      { mkexpr (EScalar (AtomicString $1)) }
;

/* Subrules for node tests */

nodetest:
  | nametest
      { (PNameTest $1) }
  | kindtest
      { (PNodeKindTest ($1)) }
;

nametest:
  | qname
      { $1 }
  | wildcard
      { $1 }
;

qname:
  | NCNAME
      { (NSDefaultElementPrefix,$1) }
  | QNAME
      { $1 }
;

qnamekeyword:
  | ITEM
      { (NSDefaultElementPrefix,"item") }
  | NCNAME
      { (NSDefaultElementPrefix,$1) }
  | QNAME
      { $1 }
;

wildcard:
  | STAR
      { (NSWildcardPrefix,"*") }
  | STAR COLON STAR
      { (NSWildcardPrefix,"*") }
  | NCNAME COLON STAR
      { (NSPrefix $1,"*") }
  | STAR COLON NCNAME
      { (NSWildcardPrefix, $3) }
;

kindtest:
  | PROCESSINGINSTRUCTIONLPAR RPAR
      { (PIKind None) }
  | PROCESSINGINSTRUCTIONLPAR STRING RPAR
      { (PIKind (Some $2)) }
  | COMMENTLPAR RPAR
      { CommentKind }
  | TEXTLPAR RPAR
      { TextKind }
  | NODELPAR RPAR
      { AnyKind }
;

/* Attributes inside element tags subrules */

attrlist:
  | /* empty */
      { [] }
  | S
      { [] }
  | S qname eq attributecontent attrlist
      { let previous_attributes = $5 in
        let qname = $2 in
        let real_att_content =
          let att_content = $4 in
	  if att_content = []
          then [(mkexpr (EText ""))]
	  else att_content
        in
	begin
	  distinct_attributes qname previous_attributes;
          (mkexpr (EAttrFixed ((UAttrSym (uattr_symbol qname)),real_att_content))) :: previous_attributes
	end }
;

attributecontent:
  | ATTRIBUTETEXT
      { let text = $1 in
        if text = "" then
	  []
	else
	  (mkexpr (EText $1)) :: [] }
  | ATTRIBUTETEXTLCURLY exprsequence RCURLY attributecontent
      { if $1 = "" then
	  if ($4 = []) then 
	    $2
	  else
	    $2 @ ((mkexpr(EText "")) :: $4)
	else
	  if ($4 = []) then 
	    (mkexpr (EText $1)) :: $2
	  else
	    ((mkexpr (EText $1)) :: $2) @ ((mkexpr(EText "")) :: $4)
      }

    /* Character references */
  | TEXTCHARREF attributecontent
      { let (text,charref) = $1 in
        let charref_text = Glx_pxp.character (Encoding.get_internal_encoding ()) charref in
        (mkexpr (EText text)) :: (mkexpr (EText charref_text)) :: $2 }

    /* Entity references */
  | TEXTENTITYREF attributecontent
      { let (text,entityref) = $1 in
        let entity_ref_text = Parse_context.get_general_entity xquery_parse_context (Finfo.parsing_locinfo ()) entityref in
        (mkexpr (EText text)) :: (mkexpr (EText entity_ref_text)) :: $2 }
;

elementcontent:
  | TEXTLCLOSINGTAG
      { (mkexpr (EText $1)) :: [] }
  | LCLOSINGTAG
      { [] }

  | TEXTLCURLY exprsequence RCURLY elementcontent
      { (mkexpr (EText $1)) :: (mkexpr (EList $2)) :: $4 }

  | TEXTLOPENINGTAG qname attrlist REMPTYELEMENT elementcontent
      { (mkexpr (EText $1)) :: (mkexpr (EElemFixed ((UElemSym (uelem_symbol $2)), $3, []))) :: $5 }

  | TEXTLOPENINGTAG qname attrlist ROPENINGTAG elementcontent qname opt_whitespace RCLOSINGTAG elementcontent
      { let os = $2
	and cs = $6 in
        same_tag os cs;
        (mkexpr (EText $1)) :: (mkexpr (EElemFixed ((UElemSym (uelem_symbol $2)),$3, $5))) :: $9 }

    /* Processing instruction */
  | TEXTOPENINGPI TEXTCLOSINGPI elementcontent
      { let (text,target_pi) = $1
	in (mkexpr (EText text)) :: (mkexpr (EPI (target_pi, $2))) :: $3 }

    /* Comment */
  | TEXTLOPENINGCOMMENT TEXTRCLOSINGCOMMENT elementcontent
      { (mkexpr (EText $1)) :: (mkexpr (EComment ($2))) :: $3 }

    /* Character references */
  | TEXTCHARREF elementcontent
      { let (text,charref) = $1 in
        let charref_text = Glx_pxp.character (Encoding.get_internal_encoding ()) charref in
        (mkexpr (EText text)) :: (mkexpr (EText charref_text)) :: $2 }

    /* Entity references */
  | TEXTENTITYREF elementcontent
      { let (text,entityref) = $1 in
        let entity_ref_text = Parse_context.get_general_entity xquery_parse_context (Finfo.parsing_locinfo ()) entityref in
        (mkexpr (EText text)) :: (mkexpr (EText entity_ref_text)) :: $2 }
;

/**********************/
/* Update expressions */
/**********************/

update:
  | simple_update
      { mkupdate (USimple $1) }
  | UPDATE complex_update
      { mkupdate (UComplex $2) }
;

target_node:
  | expr
      { $1 }
;

insert_location:
  | INTO target_node
      { UInto $2 }
  | ASLAST INTO target_node
      { UAsLastInto $3 }
  | ASFIRST INTO target_node
      { UAsFirstInto $3 }
  | AFTER target_node
      { UAfter $2 }
  | BEFORE target_node
      { UBefore $2 }
;

update_content:
  | expr
      { $1 }
;

simple_update:
  | DELETE expr
      { mksimple_update (UDelete $2) }
  | INSERT update_content insert_location
      { mksimple_update (UInsert ($2,$3)) }
  | REPLACE expr WITH expr
      { mksimple_update (UReplace (Normal_Replace,$2,$4)) }
  | REPLACE VALUEOF expr WITH expr
      { mksimple_update (UReplace (Value_Of_Replace,$3,$5)) }
;

simple_update_or_empty:
  | DELETE expr
      { mksimple_update (UDelete $2) }
  | INSERT update_content insert_location
      { mksimple_update (UInsert ($2,$3)) }
  | REPLACE expr WITH expr
      { mksimple_update (UReplace (Normal_Replace,$2,$4)) }
  | REPLACE VALUEOF expr WITH expr
      { mksimple_update (UReplace (Value_Of_Replace,$3,$5)) }
  | LPAR RPAR
    { mksimple_update (UEmpty) }
;

block_update:
  | simple_update
      { $1 :: [] }
  | simple_update block_update
      { $1 :: $2 }
;

complex_update:
      /* Conditional update */
  | IF LPAR expr RPAR THEN simple_update_or_empty ELSE simple_update_or_empty
      { mkcomplex_update (UCond ($3, $6, $8)) }
      /* FLWR update */
  | for_let_clause where_order_clause block_update
      { let (where,order) = $2 in
        mkcomplex_update (UFLWR ($1,where,order,$3)) }
;


/*********/
/* Types */
/*********/

/* Syntax for types */
/* Note: This is compliant with the latest consensus grammar - Jerome */

model:
    /* Single qualified name */
  /*- Note: Only a reference to a Group here - Jerome*/
  | qname
      { mkmodel (TBase (wrap_group_ref $1)) }

    /* Type reference */
  | TYPE qname
      { mkmodel (TGroupRef (UGroupSym (ugroup_symbol $2))) }

    /* Root document */
  | DOCUMENT LCURLY model RCURLY
      { mkmodel (TDocument $3) }

  | TEXTNODE
      { mkmodel (TText) }

    /* Local Element */
  | ELEMENT nameclass LCURLY typedmodel RCURLY
      { mkmodel (TElemLocal (UElemSym (uelem_symbol $2), None, $4)) } 
  | ELEMENT qname
      { mkmodel (TElemRef (UElemSym (uelem_symbol $2))) } 

    /* Local Attribute */
  | ATTRIBUTE nameclass LCURLY typedmodel RCURLY
      { mkmodel (TAttrLocal (UAttrSym (uattr_symbol $2),$4)) } 
  | ATTRIBUTE qname
      { mkmodel (TAttrRef (UAttrSym (uattr_symbol $2))) } 

    /* Sequence */
  | model COMMA model
      { mkmodel (TComma ($1,$3)) }
    /* Empty sequence */
  | LPAR RPAR
      { mkmodel TEmpty }

    /* Choice */
  | model BAR model
      { mkmodel (TBar ($1,$3)) }
    /* Empty choice */
  | NONE
      { mkmodel TNone }

    /* Repetitions */
  | model STAR
      { mkmodel (TBound ($1,Occurrence.occurs 0, Occurrence.unbounded)) }
  | model PLUS
      { mkmodel (TBound ($1,Occurrence.occurs 1, Occurrence.unbounded)) }

    /* Optional */
  | model QUESTION
      { mkmodel (TBound ($1,Occurrence.occurs 0, Occurrence.occurs 1)) }

    /* All group */
  | model AMPERSAND model
      { mkmodel (TAmp ($1,$3)) }

    /* Parenthesis for precedence */
  | LPAR model RPAR
      { $2 }
;

datatype:
  | limitedmodel
      { mkdatatype($1, None) }
  | limitedmodel MULT
      { mkdatatype ($1, Some(Occurrence.occurs 0, Occurrence.unbounded)) }
  | limitedmodel STAR
      { mkdatatype ($1, Some(Occurrence.occurs 0, Occurrence.unbounded)) }
  | limitedmodel PLUS   %prec UPLUS
      { mkdatatype ($1, Some(Occurrence.occurs 1, Occurrence.unbounded)) }
  | limitedmodel QUESTION
      { mkdatatype ($1, Some(Occurrence.occurs 0, Occurrence.occurs 1)) }
;

simpletype:
  | qname
      { let st = DTBase (wrap_group_ref $1) in 
        mkdatatype(st, None) }
  | qname QUESTION
      { let st = DTBase (wrap_group_ref $1) in 
        mkdatatype(st, Some(Occurrence.occurs 0, Occurrence.occurs 1)) }
;

opt_datatype:
  | /* Empty */
      { None }
  | AS datatype
      { Some $2 }
;

optelementref:
  | /* empty */
    { None }
  | qname
      { (Some (UElemSym (uelem_symbol $1))) }
;

optattributeref:
  | /* empty */
    { None }
  | qname
      { (Some (UAttrSym (uattr_symbol $1))) }
;

limitedmodel:
  | EMPTY
      { DTEmpty }
  | qname
      { DTBase (wrap_group_ref $1) }
  | ELEMENT optelementref
      { DTElemRef $2 }
  | ELEMENTOF qname
      { let qname = $2 in 
	try DTElemOfBase (lookup_bltin_type qname)
	with 
	| Not_found -> DTElemOfType (UGroupSym (ugroup_symbol qname)) }
  | ATTRIBUTE optattributeref
      { DTAttrRef $2 }
  | TYPE qname
      { DTTypeRef (UGroupSym (ugroup_symbol $2)) }
  | NODE
      { DTNode }
  | ITEM
      { DTItem }
  | ATOMIC
      { DTAtomic }
  | NUMERIC
      { DTNumeric }
  | TEXTNODE
      { DTText }
  | DOCUMENT 
      { DTDocument (None) }
  | DOCUMENT qname 
      { DTDocument (Some(UElemSym (uelem_symbol $2))) }
  | DOCUMENTOF qname 
      { DTDocumentOfType (UGroupSym (ugroup_symbol $2)) }
;

typedmodel:
  | model
      { mktyped_model (TAnonymousType ($1)) }
;

typedefinition:
  | DEFINETYPE model RCURLY
      { let sort_decl = mksort_decl (TypeDecl (UGroupSym (ugroup_symbol $1),$2)) in 
        sort_decl }
  | DEFINEATTRIBUTE typedmodel RCURLY
      { mksort_decl (AttrDecl (UAttrSym (uattr_symbol $1),$2)) }
  | DEFINEELEMENT typedmodel RCURLY
      { mksort_decl (ElemDecl (UElemSym (uelem_symbol $1),$2)) }
;

nameclass:
  | qname
      { $1 }
  | wildcard
      { $1 }
;

integer:
  | INT { $1 }
;

decimal:
  | DECIMAL { $1 }
;

/**************************************/
/* Test production to check the lexer */
/**************************************/

test:
  | /* empty */
      { [] }
  | OR test
      { Or :: $2 }
  | AND test
      { And :: $2 }
  | DIV test
      { Div :: $2 }
  | IDIV test
      { IDiv :: $2 }
  | MOD test
      { Mod :: $2 }
  | MULT test
      { Mult :: $2 }
  | PLUS test
      { Plus :: $2 }
  | MINUS test
      { Minus :: $2 }
  | STAR test
      { Star :: $2 }
  | IN test
      { In :: $2 }
  | SATISFIES test
      { Satisfies :: $2 }
  | RETURN test
      { Return :: $2 }
  | THEN test
      { Then :: $2 }
  | ELSE test
      { CElse :: $2 }
  | TO test
      { To :: $2 }
  | WHERE test
      { Where :: $2 }
  | INTERSECT test
      { Intersect :: $2 }
  | UNION test
      { Union :: $2 }
  | EXCEPT test
      { Except :: $2 }
  | PRECEDES test
      { Before :: $2 }
  | FOLLOWS test
      { After :: $2 }
  | CASE test
      { CCase :: $2 }
  | INSTANCEOF test
      { Instanceof :: $2 }
  | SORTBY test
      { SortBy :: $2 }
  | NCNAME test
      { (NCName $1) :: $2 }
  | QNAME test
      { (QName $1) :: $2 }
  | AXIS test
      { (Axis $1) :: $2 }
  | COLON test
      { Colon :: $2 }
  | SLASH test
      { Slash :: $2 }
  | SLASHSLASH test
      { SlashSlash :: $2 }
  | EQUALS test
      { Equals :: $2 }
  | IS test
      { EqualsEquals :: $2 }
  | NOTEQUALS test
      { NotEquals :: $2 }
  | ISNOT test
      { NotEqualsEquals :: $2 }
  | COLONEQUALS test
      { ColonEquals :: $2 }
  | LT test
      { Lt :: $2 }
  | GT test
      { Gt :: $2 }
  | LOPENINGTAG test
      { LOpeningTag :: $2 }
  | LCLOSINGTAG test
      { LClosingTag :: $2 }
  | REMPTYELEMENT test
      { REmptyElement :: $2 }
  | ROPENINGTAG test
      { ROpeningTag :: $2 }
  | RCLOSINGTAG test
      { RClosingTag :: $2 }
  | TEXTLOPENINGTAG test
      { (TextLOpeningTag $1) :: $2 }
  | TEXTLCLOSINGTAG test
      { (TextLClosingTag $1) :: $2 }
  | TEXTLCURLY test
      { (TextLCurly $1) :: $2 }
  | BAR test
      { Bar :: $2 }
  | ATSIGN test
      { At :: $2 }
  | LPAR test
      { Lpar :: $2 }
  | RPAR test
      { Rpar :: $2 }
  | LBRACK test
      { Lbrack :: $2 }
  | RBRACK test
      { Rbrack :: $2 }
  | LCURLY test
      { Lcurly :: $2 }
  | RCURLY test
      { Rcurly :: $2 }
  | DOLLAR test
      { Dollar :: $2 }
  | SOME test
      { CSome :: $2 }
  | EVERY test
      { CEvery :: $2 }
  | FOR test
      { CFor :: $2 }
  | LET test
      { CLet :: $2 }
  | COMMENTLPAR test
      { CommentLpar :: $2 }
  | TEXTLPAR test
      { TextLpar :: $2 }
  | NODELPAR test
      { NodeLpar :: $2 }
  | PROCESSINGINSTRUCTIONLPAR test
      { ProcessingInstructionLpar :: $2 }
  | FUNCTIONNAMELPAR test
      { (FunctionNameLpar $1) :: $2 }
  | DEFINEFUNCTION test
      { DefineFunction :: $2 }
  | DEFINESIGNATURE test
      { DefineSignature :: $2 }
  | IF test
      { If :: $2 }
  | CASTAS test
      { CastAs :: $2 }
  | TREATAS test
      { TreatAs :: $2 }
  | ASSERTAS test
      { AssertAs :: $2 }
  | INT test
      { (Integer $1) :: $2 }
  | STRING test
      { (String $1) :: $2 }
  | DESCENDING test
      { DescendingT :: $2 }
  | ASCENDING test
      { AscendingT :: $2 }
  | ELEMENT test
      { TElement :: $2 }
  | ATTRIBUTE test
      { TAttribute :: $2 }
  | TYPE test
      { TType :: $2 }
  | DEFINETYPE test
      { (TTypeDefn $1) :: $2 }
  | DEFINEELEMENT test
      { (TElementDefn $1) :: $2 }
  | DEFINEATTRIBUTE test
      { (TAttributeDefn $1) :: $2 }
  | DEFINEGLOBAL test
      { DefineGlobal :: $2 }
  | ATTRIBUTETEXT test
      { (RTextAttribute $1) :: $2 }
  | ATTRIBUTETEXTLCURLY test
      { (RTextAttributeLCurly $1) :: $2 }
;

