(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: function_analysis.ml,v 1.1 2003/05/21 15:03:49 simeon Exp $ *)

(* Module: Function_analysis
   Description:
     This module implements the path analysis for built-in functions.
*)

open Error

open Namespace

open Xquery_ast


(* Contains functions that only need parameters used paths *)

let fun_used_only_table =
  [ 
    (* Section 5: Functions on numbers *)
    (op_uri,"int-add");
    (op_uri,"int-subtract");
    (op_uri,"int-multiply");
    (op_uri,"int-divide");
    (op_uri,"int-mod");
    (op_uri,"int-unary-plus");
    (op_uri,"int-unary-minus");
    (op_uri,"integer-add");
    (op_uri,"integer-subtract");
    (op_uri,"integer-multiply");
    (op_uri,"integer-divide");
    (op_uri,"integer-mod");
    (op_uri,"integer-unary-plus");
    (op_uri,"integer-unary-minus");
    (op_uri,"decimal-add");
    (op_uri,"decimal-subtract");
    (op_uri,"decimal-multiply");
    (op_uri,"decimal-divide");
    (op_uri,"decimal-mod");
    (op_uri,"decimal-unary-plus");
    (op_uri,"decimal-unary-minus");
    (op_uri,"float-add");
    (op_uri,"float-subtract");
    (op_uri,"float-multiply");
    (op_uri,"float-divide");
    (op_uri,"float-mod");
    (op_uri,"float-unary-plus");
    (op_uri,"float-unary-minus");
    (op_uri,"double-add");
    (op_uri,"double-subtract");
    (op_uri,"double-multiply");
    (op_uri,"double-divide");
    (op_uri,"double-mod");
    (op_uri,"double-unary-plus");
    (op_uri,"double-unary-minus");
    (fn_uri,"floor");
    (fn_uri,"ceiling");
    (fn_uri,"round");

    (* Section 6. Functions on Strings *)
    (fn_uri,"concat");
    (fn_uri,"string-join");
    (fn_uri,"starts-with");
    (fn_uri,"ends-with");
    (fn_uri,"contains");
    (fn_uri,"substring");
    (fn_uri,"string-length");
    (fn_uri,"substring-before");
    (fn_uri,"substring-after");
    (fn_uri,"normalize-space");
    (fn_uri,"upper-case");
    (fn_uri,"lower-case");
    (fn_uri,"translate");
    (fn_uri,"string-pad");

    (* Section 7. Functions and Operators on Booleans *)
    (fn_uri,"true");
    (fn_uri,"false");
    (fn_uri,"not");
    (op_uri,"or");
    (op_uri,"and");

    (* Section 9. Functions on QNames *)
    (fn_uri,"expanded-QName");
    (fn_uri,"get-local-name-from-QName");
    (fn_uri,"get-namespace-from-QName");

    (* Section 10. Functions and Operators for anyURI *)
    (fn_uri,"resolve-uri");

    (* Section 14. Functions and operators on sequences *)
    (op_uri,"to");
  ] 


(* Contains functions that use used and return paths *)

let fun_used_return_simple_table =
  [
   (* Section 2. Accessors *)
   (fn_uri,"node-kind");
   (fn_uri,"node-name");
   (fn_uri,"base-uri");

   (* Section 13. Functions and Operators on Nodes *)
   (fn_uri,"name");
   (fn_uri,"local-name");
   (fn_uri,"namespace-uri");
   (op_uri,"node-equal");
   (op_uri,"node-before");
   (op_uri,"node-after");

   (* Section 14. Functions and operators on sequences *)
   (fn_uri,"index-of");
   (fn_uri,"empty");
   (fn_uri,"exists");
   (fn_uri,"distinct-values");
   (fn_uri,"count");
   (fn_uri,"sequence-node-equal");

 ] 


(* Contains functions that use used and return (with subtree) paths *)

let fun_used_return_subtree_table =
  [
   (* Section 2. Accessors *)
   (fn_uri,"string");
   (fn_uri,"data");

   (* Section 13. Functions and Operators on Nodes *)
   (fn_uri,"root");
   (fn_uri,"deep-equal");
   (fn_uri,"copy");
   (fn_uri,"insert");

   (* Section 14. Functions and operators on sequences *)
   (fn_uri,"sequence-deep-equal");
   (op_uri,"equal");
   (op_uri,"nequal");
   (op_uri,"le");
   (op_uri,"ge");
   (op_uri,"gt");
   (op_uri,"lt");
 ] 


(* Contains functions that return paths *)

let fun_returns_paths_table =
  [ (glx_uri,"distinct-docorder") ] 


(*******************************************************************************)

(* The following functions are not treated. - Amelie

  Section 2. Accessors 
  (fn_uri,"unique-ID")

   Section14.
   (fn_uri,"item-at")
   (fn_uri,"distinct-nodes")
   (fn_uri,"remove")
   (fn_uri,"subsequence")
   (fn_uri,"subsequence")
   (op_uri,"except") 
   (op_uri,"intersect")
   (op_uri,"union")

   
  ((fn_uri,"avg-int"), 1), _fn_avg_int;
  ((fn_uri,"max-int"), 1), _fn_max_int;
  ((fn_uri,"min-int"), 1), _fn_min_int;
  ((fn_uri,"sum-int"), 1), _fn_sum_int;

  ((fn_uri,"avg-decimal"), 1), _fn_avg_decimal;
  ((fn_uri,"max-decimal"), 1), _fn_max_decimal;
  ((fn_uri,"min-decimal"), 1), _fn_min_decimal;
  ((fn_uri,"sum-decimal"), 1), _fn_sum_decimal;

  ((fn_uri,"avg-float"), 1), _fn_avg_float;
  ((fn_uri,"max-float"), 1), _fn_max_float;
  ((fn_uri,"min-float"), 1), _fn_min_float;
  ((fn_uri,"sum-float"), 1), _fn_sum_float;

  ((fn_uri,"avg-double"), 1), _fn_avg_double;
  ((fn_uri,"max-double"), 1), _fn_max_double;
  ((fn_uri,"min-double"), 1), _fn_min_double;
  ((fn_uri,"sum-double"), 1), _fn_sum_double;

  ((fn_uri,"document"), 1), _fn_document;
  ((glx_uri,"document-validate"), 2), _glx_document_validate;

  ((glx_uri,"unique"), 1),             _glx_unique;
  ((glx_uri,"optional"), 1),           _glx_optional;

  ((glx_uri,"union-values"), 2), _glx_union_values;
  ((glx_uri,"intersect-values"), 2), _glx_intersect_values;
  ((glx_uri,"except-values"), 2), _glx_except_values;

*)



(* Function to get XQuery function type for analysis                         *)

type function_kind =
  | UsedOnly
  | UsedReturnSimple
  | UsedReturnSubtree
  | ReturnsPaths
  | ReturnsDefault

let get_fun_analysis_type fname =
  if (List.exists (fun x -> x = fname) fun_used_only_table)
  then UsedOnly
  else
    if (List.exists (fun x -> x = fname) fun_used_return_simple_table)
    then UsedReturnSimple
    else
      if (List.exists (fun x -> x = fname) fun_used_return_subtree_table)
      then UsedReturnSubtree
      else
	if (List.exists (fun x -> x = fname) fun_returns_paths_table)
	then ReturnsPaths
	else
	  ReturnsDefault

