(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                              XQuery Engine                          *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: path_structutil.ml,v 1.2 2003/06/09 15:28:26 simeon Exp $ *)

(* Module: Path_structutil
   Description:
     This module contains basic operations on path structures.
*)

open Format

open Error

open Namespace
open Sym

open Xquery_ast

open Sax_types
open Sax_parser

open Path_struct


(* Turns all subtree annotations on *)

let imposes_subtree path_seq =
  let subtree_fun (did, (path, subtree)) =
    (did,(path,Subtree))
  in
  List.map subtree_fun path_seq

(* Actions and related operations *)

type action =
  | GetSubtree
  | PreserveNode
  | SkipNode
  | KeepMovingPreserveNode of path_fragment_sequence
  | KeepMovingSkipNode of path_fragment_sequence

let merge_actions a1 a2 =
  match (a1,a2) with
  | (GetSubtree,_) -> GetSubtree
  | (_,GetSubtree) -> GetSubtree

  | (KeepMovingSkipNode p1,KeepMovingSkipNode p2)     	  -> KeepMovingSkipNode (p1@p2)
  | (KeepMovingSkipNode p1,KeepMovingPreserveNode p2) 	  -> KeepMovingPreserveNode (p1@p2)
  | (KeepMovingPreserveNode p1,KeepMovingSkipNode p2) 	  -> KeepMovingPreserveNode (p1@p2)
  | (KeepMovingPreserveNode p1,KeepMovingPreserveNode p2) -> KeepMovingPreserveNode (p1@p2)

  | (KeepMovingSkipNode p1, PreserveNode)     -> KeepMovingPreserveNode p1
  | (PreserveNode, KeepMovingPreserveNode p2) -> KeepMovingPreserveNode p2
  | (KeepMovingPreserveNode p1, PreserveNode) -> KeepMovingPreserveNode p1
  | (PreserveNode, KeepMovingSkipNode p2)     -> KeepMovingPreserveNode p2
  | (PreserveNode,PreserveNode)               -> PreserveNode
  
  | (KeepMovingSkipNode p1, SkipNode) 	  -> KeepMovingSkipNode p1
  | (SkipNode, KeepMovingSkipNode p2) 	  -> KeepMovingSkipNode p2
  | (SkipNode, KeepMovingPreserveNode p2) -> KeepMovingPreserveNode p2
  | (KeepMovingPreserveNode p1, SkipNode) -> KeepMovingPreserveNode p1

  | (PreserveNode,SkipNode) -> PreserveNode
  | (SkipNode,PreserveNode) -> PreserveNode
  
  | (SkipNode,SkipNode) -> SkipNode

(* Navigation functions *)

let node_sym_of_sax_event nsenv sax_event =
  match sax_event with
  | SAX_startDocument (sax_xml_decl,_) ->
      None
  | SAX_endDocument ->
      None
  | SAX_startElement (element_uqname,_,_) ->
      let (prefix,element_rqname) = Resolve.resolve_element_qname nsenv element_uqname in
      Some (RUnitElemSymbol (Sym.relem_symbol (prefix,element_rqname)))
  | SAX_endElement ->
      None
  | SAX_processingInstruction pi_desc ->
      None
  | SAX_comment comment_desc ->
      None
  | SAX_characters text_desc ->
      None
  | SAX_attribute (attribute_qname,_) ->
      let (prefix,attr_rqname) = Resolve.resolve_attribute_qname nsenv attribute_qname in
      Some (RUnitAttrSymbol (Sym.rattr_symbol (prefix,attr_rqname)))
  | SAX_atomicValue _ ->
      raise (Query (Internal_Error "Should not apply projection operation on an atomic value!"))

let pi_sym_of_sax_event sax_event =
  match sax_event with
  | SAX_startDocument _ ->
      None
  | SAX_endDocument ->
      None
  | SAX_startElement _ ->
      None
  | SAX_endElement ->
      None
  | SAX_processingInstruction (pi_name,_) ->
      Some pi_name
  | SAX_comment _ ->
      None
  | SAX_characters _ ->
      None
  | SAX_attribute _ ->
      None
  | SAX_atomicValue _ ->
      raise (Query (Internal_Error "Should not apply projection operation on an atomic value!"))

type _NodeKind =
  | DocNodeKind
  | ElemNodeKind
  | AttrNodeKind
  | TextNodeKind
  | NSNodeKind
  | PINodeKind
  | CommentNodeKind

let node_kind_of_sax_event sax_event =
  match sax_event with
  | SAX_startDocument _ ->
      DocNodeKind
  | SAX_endDocument ->
      DocNodeKind
  | SAX_startElement _ ->
      ElemNodeKind
  | SAX_endElement ->
      ElemNodeKind
  | SAX_processingInstruction _ ->
      PINodeKind
  | SAX_comment _ ->
      CommentNodeKind
  | SAX_characters _ ->
      TextNodeKind
  | SAX_attribute _ ->
      AttrNodeKind
  | SAX_atomicValue _ ->
      raise (Query (Internal_Error "Should not apply projection operation on an atomic value!"))

(* Computes the princinpal node kind of a given axis *)

let principal_node_kind a =
  match a with
  | Xquery_ast.Ancestor ->
      ElemNodeKind
  | Xquery_ast.Ancestor_or_self ->
      ElemNodeKind
  | Xquery_ast.Attribute ->
      AttrNodeKind
  | Xquery_ast.Child ->
      ElemNodeKind
  | Xquery_ast.Descendant ->
      ElemNodeKind
  | Xquery_ast.Descendant_or_self ->
      ElemNodeKind
  | Xquery_ast.Namespace ->
      NSNodeKind
  | Xquery_ast.Parent ->
      ElemNodeKind
  | Xquery_ast.Self ->
      ElemNodeKind

(* Whether a node test holds on a given sax event under a particular axis or not *)

let eval_node_test nsenv a nt sax_event =
  let nodesym1 = node_sym_of_sax_event nsenv sax_event in
  let nodekind1 = node_kind_of_sax_event sax_event in
  match nt with
  | CPNameTest qn ->
      begin
	match nodesym1 with 
	| None -> false
	| Some sym -> 
	    let qnsym =
	      begin
		match a with
		| Attribute -> RUnitAttrSymbol (rattr_symbol (NSDefaultElementPrefix, qn))
		| Namespace -> raise (Query (Prototype "In typing: name tests on namespaces not yet supported"))
		| _ -> RUnitElemSymbol (relem_symbol (NSDefaultElementPrefix, qn))
	      end
	    in
            (principal_node_kind a) = nodekind1 &&
	    Sym.subtag sym qnsym
      end
  | CPNodeKindTest node_kind ->
      begin
	match node_kind with
	| PIKind pik ->
	    nodekind1 = PINodeKind
	      &&
	    begin
	      match pik with
	      |	None -> true
	      |	Some name ->
		  let pi_targ = pi_sym_of_sax_event sax_event in
		  begin
		    match pi_targ with
		    | None -> false
		    | Some pitargreal -> pitargreal = name
		  end
	    end
	| CommentKind ->
	    nodekind1 = CommentNodeKind
	| TextKind ->
	    nodekind1 = TextNodeKind
	| AnyKind ->
	    true
      end

(* Actual movement in the path structure *)

let rec one_step_one_path nsenv sax_event (path, subtree) =
  match path with
  | [] -> (SkipNode,[])
  | (axis1,nt1) :: [] ->
      begin
	match axis1 with
	| Ancestor ->
	    raise (Query (Prototype "Cannot perform projection with ancestor axis involved"))
	| Ancestor_or_self ->
	    raise (Query (Prototype "Cannot perform projection with ancestor-or-self axis involved"))
	| Attribute ->
	    (* Don't care about attributes in the projection *)
	    (SkipNode,[])
	| Child ->
	    begin
	      if eval_node_test nsenv axis1 nt1 sax_event
	      then
		match subtree with
		| Subtree -> (GetSubtree,[])
		| NoSubtree -> (PreserveNode,[])
	      else
		(SkipNode,[])
	    end
	| Descendant ->
	    begin
	      if eval_node_test nsenv axis1 nt1 sax_event
	      then
		match subtree with
		| Subtree -> (GetSubtree,[])
		| NoSubtree -> (PreserveNode,[([(Descendant,nt1)],subtree)])
	      else
		(SkipNode,[([(Descendant,nt1)],subtree)])
	    end
	| Descendant_or_self ->
	    begin
	      if eval_node_test nsenv axis1 nt1 sax_event
	      then
		match subtree with
		| Subtree -> (GetSubtree,[])
		| NoSubtree -> (PreserveNode,[([(Descendant,nt1)],subtree)])
	      else
		(SkipNode,[([(Descendant,nt1)],subtree)])
	    end
	| Namespace ->
	    (* Don't care about namespace nodes in the projection *)
	    (SkipNode,[])
	| Parent ->
	    raise (Query (Prototype "Cannot perform projection with parent axis involved"))
	| Self ->
	    (SkipNode,[])
      end
  | (axis1,nt1) :: path' ->
      begin
	match axis1 with
	| Ancestor ->
	    raise (Query (Prototype "Cannot perform projection with ancestor axis involved"))
	| Ancestor_or_self ->
	    raise (Query (Prototype "Cannot perform projection with ancestor-or-self axis involved"))
	| Attribute ->
	    (* Don't care about attributes in the projection *)
	    (SkipNode,[])
	| Child ->
	    begin
	      if eval_node_test nsenv axis1 nt1 sax_event
	      then
		(KeepMovingSkipNode [(path',subtree)],[])
	      else
		(SkipNode,[])
	    end
	| Descendant ->
	    begin
	      if eval_node_test nsenv axis1 nt1 sax_event
	      then
		(KeepMovingSkipNode [(path',subtree)],[((Descendant,nt1)::path',subtree)])
	      else
		(SkipNode,[((Descendant,nt1)::path',subtree)])
	    end
	| Descendant_or_self ->
	    begin
	      if eval_node_test nsenv axis1 nt1 sax_event
	      then
		(KeepMovingSkipNode [(path',subtree)],[((Descendant,nt1)::path',subtree)])
	      else
		(SkipNode,[((Descendant,nt1)::path',subtree)])
	    end
	| Namespace ->
	    (* Don't care about namespace nodes in the projection *)
	    (SkipNode,[])
	| Parent ->
	    raise (Query (Prototype "Cannot perform projection with parent axis involved"))
	| Self ->
	    (SkipNode,[])
      end

let rec one_step_remove_self nsenv sax_event (path, subtree) previous_action plus1 =
  match path with
  | [] -> (previous_action,plus1)
  | (axis1,nt1) :: [] ->
      begin
	match axis1 with
	| Ancestor ->
	    (previous_action,plus1)
	| Ancestor_or_self ->
	    (previous_action,plus1)
	| Attribute ->
	    (previous_action,plus1)
	| Child ->
	    (previous_action,plus1)
	| Descendant ->
	    (previous_action,plus1)
	| Descendant_or_self ->
	    begin
	      let (new_action,plus) =
		if eval_node_test nsenv axis1 nt1 sax_event
		then
		  match subtree with
		  | Subtree -> (GetSubtree,plus1)
		  | NoSubtree -> (PreserveNode, [([(Descendant,nt1)],subtree)])
		else
		  (SkipNode, [([(Descendant,nt1)],subtree)])
	      in
	      (new_action,plus1@plus)
	    end
	| Namespace ->
	    (previous_action,plus1)
	| Parent ->
	    (previous_action,plus1)
	| Self ->
	    begin
	      let new_action =
		if eval_node_test nsenv axis1 nt1 sax_event
		then
		  match subtree with
		  | Subtree -> GetSubtree
		  | NoSubtree -> PreserveNode
		else
		  SkipNode
	      in
	      (new_action, plus1)
	    end
      end
  | (axis1,nt1) :: path' ->
      begin
	match axis1 with
	| Ancestor ->
	    (previous_action,plus1)
	| Ancestor_or_self ->
	    (previous_action,plus1)
	| Attribute ->
	    (previous_action,plus1)
	| Child ->
	    (previous_action,plus1)
	| Descendant ->
	    (previous_action,plus1)
	| Descendant_or_self ->
	    begin
	      let (new_action,plus) =
		if eval_node_test nsenv axis1 nt1 sax_event
		then
		  

		  (KeepMovingSkipNode [(path',subtree)], [((Descendant,nt1)::path',subtree)])
		else
		  (SkipNode, [((Descendant,nt1)::path',subtree)])
	      in
	      one_step_remove_self nsenv sax_event (path', subtree) new_action (plus1@plus)
	    end
	| Namespace ->
	    (previous_action,plus1)
	| Parent ->
	    (previous_action,plus1)
	| Self ->
	    begin
	      let new_action =
		if eval_node_test nsenv axis1 nt1 sax_event
		then
		  KeepMovingSkipNode [(path',subtree)]
		else
		  SkipNode
	      in
	      one_step_remove_self nsenv sax_event (path', subtree) new_action plus1
	    end
      end

let compose_steps nsenv sax_event (path, subtree) =
  let (action1,plus1) = one_step_one_path nsenv sax_event (path, subtree) in
  match action1 with
  | GetSubtree -> (action1,plus1)
  | PreserveNode -> (action1,plus1)
  | SkipNode -> (action1,plus1)
  | KeepMovingPreserveNode path_fragment_sequence ->
      let (npath,nsubtree) = List.nth path_fragment_sequence 0 in
      one_step_remove_self nsenv sax_event (npath,nsubtree) action1 plus1
  | KeepMovingSkipNode path_fragment_sequence ->
      let (npath,nsubtree) = List.nth path_fragment_sequence 0 in
      one_step_remove_self nsenv sax_event (npath,nsubtree) action1 plus1

let one_step nsenv sax_event (path_fragment_seq : path_fragment_sequence) =
  let stage1 = List.map (compose_steps nsenv sax_event) path_fragment_seq in
  let (stage1_actions, stage1_plus) = List.split stage1 in
  let stage2_actions = List.fold_left merge_actions SkipNode stage1_actions in
  let stage3_actions =
    let all_add_paths = List.concat stage1_plus in
    match all_add_paths with
    | [] -> SkipNode
    | _ -> KeepMovingSkipNode all_add_paths
  in
  merge_actions stage2_actions stage3_actions

(* attributes *)

let rec one_step_document_remove_self nsenv sax_event (path, subtree) =
  match path with
  | [] -> [(path, subtree)]
  | (axis1,nt1) :: path' ->
      begin
	match axis1 with
	| Ancestor ->
	    [(path, subtree)]
	| Ancestor_or_self ->
	    [(path, subtree)]
	| Attribute ->
	    [(path, subtree)]
	| Child ->
	    [(path, subtree)]
	| Descendant ->
	    [(path, subtree)]
	| Descendant_or_self ->
	    begin
	      if eval_node_test nsenv axis1 nt1 sax_event
	      then
		(one_step_document_remove_self nsenv sax_event (path', subtree)) @ [((Descendant,nt1)::path',subtree)]
	      else
		[((Descendant,nt1)::path',subtree)]
	    end
	| Namespace ->
	    [(path, subtree)]
	| Parent ->
	    [(path, subtree)]
	| Self ->
	    begin
	      if eval_node_test nsenv axis1 nt1 sax_event
	      then
		(one_step_document_remove_self nsenv sax_event (path', subtree))
	      else
		[]
	    end
      end

let compose_document_steps nsenv sax_event (path, subtree) =
  one_step_document_remove_self nsenv sax_event (path,subtree)

let one_document_step nsenv sax_event (path_fragment_seq : path_fragment_sequence) =
  List.concat (List.map (compose_document_steps nsenv sax_event) path_fragment_seq)

let get_start_pfs path_seq docid =
  let rec select_path path_seq =
    match path_seq with
    | [] -> []
    | (docid1, (path1, subtree1)) :: path_seq' ->
	if (docid1 = docid)
	then
	  (path1, subtree1) :: (select_path path_seq')
	else
	  (select_path path_seq')
  in
  select_path path_seq

let inside_document nsenv sax_event path_seq docid =
  let start_pfs = get_start_pfs path_seq docid in
  one_document_step nsenv sax_event start_pfs

(* attributes *)

let one_step_attribute_check nsenv uqn (path, subtree) =
  match path with
  | [] -> false
  | (axis1,nt1) :: _ ->
      begin
      	match axis1 with
	| Attribute ->
	    true
	| _ ->
	    false
      end

let one_step_attribute nsenv pfl (uqn,_) =
  List.exists (one_step_attribute_check nsenv uqn) pfl
    

(* -----------------------------------------------------------------------------
   Print functions
----------------------------------------------------------------------------- *)

let string_node n =
  match n with
  | CPNameTest (prefix,name) ->
      name
  | CPNodeKindTest kind ->
      match kind with
      |	PIKind st ->
	  begin
	    match st with
	    | None -> 
	        "pi()"
	    | Some x ->
		"pi(\"" ^ x ^ "\")"
	  end
      |	CommentKind ->
	  "comment()"
      |	TextKind ->
	  "text()"
      |	AnyKind ->
	  "node()"

let rec string_path p = 
  let path_string =
    match p with
    | [] -> " "
    | (axis,node)::p_rest ->
	let axis_string =
	  match axis with
	  | Ancestor ->
	      "/ancestor::"
	  | Ancestor_or_self ->
	      "/ancestor_or_self::"
	  | Attribute ->
	      "/attribute::"
	  | Child ->
	      "/"
	  | Descendant ->
	      "//"
	  | Descendant_or_self ->
	      "/descendant_or_self::"
	  | Namespace ->
	      "/namespace::"
	  | Parent ->
	      "/parent::"
	  | Self ->
	      "/self::"
	  in axis_string ^ string_node node ^ string_path p_rest
  in path_string
 
let rec string_path_sequence ps =
  match ps with
  | [] -> ""
  | (docid,(path,subtree))::ps_rest ->
      let string_subtree subtree =
	match subtree with
	| Subtree -> "# \n"
	| NoSubtree ->  "\n"
      in docid ^ string_path path ^ string_subtree subtree ^ string_path_sequence ps_rest


let print_path_sequence ff ps = 
  let string_ps = string_path_sequence ps in
  fprintf ff "%s" string_ps

