(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: stream_project.ml,v 1.2 2003/06/09 15:28:27 simeon Exp $ *)

(* Module: Stream_project
   Description:
     This module implements document projection on an XML stream.
*)

open Error

open Sax_types

open Project_context

(* Projects attributes *)

let project_attributes nsenv other_atts pfs' =
  List.filter (Path_structutil.one_step_attribute nsenv pfs') other_atts

let rec project_next_projection pfs project_context =
  let nsenv = Project_context.get_nsenv project_context in

  (* Get the next event *)
  let xml_event = Project_context.get_next_xml_event project_context in

  (* And refill the buffer accordingly *)

  match xml_event with
  | SAX_startDocument _ ->
      raise (Query (Internal_Error ("Should not have a start document event")))

  | SAX_endDocument ->
      Project_context.pop_project_context project_context [xml_event]

  | SAX_startElement (elem_uqname, attributes, has_element_content) ->

      (* Identify which kind of action is required based on the path structure *)
      let action = Path_structutil.one_step nsenv xml_event pfs in
      begin
	match action with
	| Path_structutil.GetSubtree ->

            (* SUBTREE ACTION --

               In the case the required action is to return the whole
               subtree, flip to non-projection parsing using None in
               the path structure stack. - Jerome *)

	    Project_context.push_project_context_get_subtree project_context xml_event
	      
	| Path_structutil.KeepMovingSkipNode pfs' ->
	    
            (* KEEP MOVING SKIP NODE ACTION --
	       
               In that case, we keep talking down the tree until we
               can decide whether we should return the node or not. - Jerome *)
	    
            (* Extract the namespace attributes and update the
	       namespace environment accordingly. *)
	    
	    let (_, new_nss, _) = Sax_typesutil.extract_special_attributes attributes in
	    let nsenv' = Namespace.add_all_ns nsenv new_nss in
	    
	    let projected_attributes = project_attributes nsenv' attributes pfs' in
	    
	    let new_xml_event = SAX_startElement (elem_uqname, projected_attributes, has_element_content) in
	    
	    if (projected_attributes = []) then
	      begin
		Project_context.push_project_context_keep_moving_skip_node
		  project_context
		  nsenv'
		  new_xml_event
		  pfs'
	      end
	    else
	      begin
		Project_context.push_project_context_keep_moving_preserve_node
		  project_context
		  nsenv'
		  new_xml_event
		  pfs'
	      end
		
	| Path_structutil.KeepMovingPreserveNode pfs' ->
	    
            (* KEEP MOVING PRESERVE NODE ACTION --
	       
               In that case, we return the current event and keep
               talking down the tree. - Jerome *)
	    
            (* Extract the namespace attributes and update the
	       namespace environment accordingly. *)
	    
	    let (_, new_nss, _) = Sax_typesutil.extract_special_attributes attributes in
	    let nsenv' = Namespace.add_all_ns nsenv new_nss in
	    
	    let projected_attributes = project_attributes nsenv' attributes pfs' in
	    
	    let new_xml_event = SAX_startElement (elem_uqname, projected_attributes, has_element_content) in
	    
	    begin
	      Project_context.push_project_context_keep_moving_preserve_node
		project_context
		nsenv'
		new_xml_event
		pfs'
	    end
	      
	| Path_structutil.PreserveNode ->
	    
            (* PRESERVE NODE ACTION --
	       
               In that case, discard the XML stream for the subtree,
               but keep the events for the current node. - Jerome *)
	    
            (* Extract the namespace attributes and update the
	       namespace environment accordingly. *)
	    
	    let (_, new_nss, _) = Sax_typesutil.extract_special_attributes attributes in
	    let nsenv' = Namespace.add_all_ns nsenv new_nss in
	    
	    let refill_local_buffer = [
	      SAX_startElement (elem_uqname, attributes, has_element_content);
	      SAX_endElement
	    ]
	    in
	    Project_context.push_project_context_preserve_node project_context refill_local_buffer
	      
	| Path_structutil.SkipNode ->
	    
            (* SKIP NODE ACTION --
	       
               In that case, discard the XML stream. - Jerome *)
	    
            (* Extract the namespace attributes and update the
	       namespace environment accordingly. *)
	    
	    Project_context.push_project_context_skip_node project_context
      end
	
  | SAX_endElement ->
      Project_context.pop_project_context project_context [xml_event]

  | SAX_processingInstruction (target,content) ->
      let action = Path_structutil.one_step nsenv xml_event pfs in
      begin
	match action with
	| Path_structutil.GetSubtree
	| Path_structutil.KeepMovingPreserveNode _
	| Path_structutil.PreserveNode ->
	    Project_context.refill_local_buffer project_context [xml_event]
	| Path_structutil.KeepMovingSkipNode _
	| Path_structutil.SkipNode ->
	    project_next_projection pfs project_context
      end
	
  | SAX_comment c ->
      let action = Path_structutil.one_step nsenv xml_event pfs in
      begin
	match action with
	| Path_structutil.GetSubtree
	| Path_structutil.KeepMovingPreserveNode _
	| Path_structutil.PreserveNode ->
	    Project_context.refill_local_buffer project_context [xml_event]
	| Path_structutil.KeepMovingSkipNode _
	| Path_structutil.SkipNode ->
	    project_next_projection pfs project_context
      end
	
  | SAX_characters s ->
      let action = Path_structutil.one_step nsenv xml_event pfs in
      begin
	match action with
	| Path_structutil.GetSubtree
	| Path_structutil.KeepMovingPreserveNode _
	| Path_structutil.PreserveNode ->
	    Project_context.refill_local_buffer project_context [xml_event]
	| Path_structutil.KeepMovingSkipNode _
	| Path_structutil.SkipNode ->
	    project_next_projection pfs project_context
      end

  | SAX_attribute a ->
      let action = Path_structutil.one_step nsenv xml_event pfs in
      begin
	match action with
	| Path_structutil.GetSubtree
	| Path_structutil.KeepMovingPreserveNode _
	| Path_structutil.PreserveNode ->
	    Project_context.refill_local_buffer project_context [xml_event]
	| Path_structutil.KeepMovingSkipNode _
	| Path_structutil.SkipNode ->
	    project_next_projection pfs project_context
      end	      

  | SAX_atomicValue _ ->
      (* A path expression never applies to an atomic value *)
      project_next_projection pfs project_context
	
let rec project_next_get_subtree project_context =
  (* Get the next event *)
  let xml_event = Project_context.get_next_xml_event project_context in
  match xml_event with
  | SAX_startDocument _
  | SAX_startElement _ ->
      Project_context.push_project_context_get_subtree project_context xml_event
  | SAX_endDocument
  | SAX_endElement ->
      Project_context.pop_project_context project_context [xml_event]
  | SAX_processingInstruction _
  | SAX_comment _
  | SAX_characters _
  | SAX_attribute _
  | SAX_atomicValue _ ->
      Project_context.refill_local_buffer project_context [xml_event]

let rec project_next project_context =
  match get_pfs project_context with
  | None ->
      project_next_get_subtree project_context
  | Some pfs ->
      project_next_projection pfs project_context

(* Stream wrapping function *)

let rec next_project_event_internal project_context =
  match Project_context.get_next_buffered_sax_event project_context with
  | None ->
      begin
	project_next project_context;
	next_project_event_internal project_context
      end
  | Some event ->
      Some event

let next_project_event project_context n =
  next_project_event_internal project_context

(* Top level stream operation *)

let project_xml_stream nsenv root_uri path_seq xml_stream =

  let first_event = Stream.next xml_stream in
  begin
    match first_event with
    | SAX_startDocument _ ->
	()
    | _ ->
	raise (Query (Internal_Error ("Was expecting a start document event")))
  end;

  let pfs = Path_structutil.inside_document nsenv first_event path_seq (AnyURI._string_of_uri root_uri) in

  let project_context =
    Project_context.build_project_context xml_stream pfs [first_event]
  in

  (Stream.from (next_project_event project_context))

