(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2001, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: glushkov.ml,v 1.16 2002/10/27 20:02:52 simeon Exp $ *)

open Occurrence
open Error
open Sym
open Xquery_ast

module OrderedTypeNames =
  struct
    type t = Sym.rgroup_symbol 

    let compare tn1 tn2 =
      let (uri1, ncn1) = (Sym.rgroup_name tn1) and
	  (uri2, ncn2) = (Sym.rgroup_name tn2) in
      let x = compare uri1 uri2 in
      if (x=0) then
	compare ncn1 ncn2
      else
	x
  end

module TypeName_Glushkov = Regexp.MakeGlushkovType(OrderedTypeNames)

open TypeName_Glushkov
open Regexp

(******************************************)
(* Mapping Type AST to Regular expression *)
(******************************************)
(* Normalizes a model into a 'pure' regular expression *)

(* Note: this function also initializes positions.
   - Jerome *)

let rec normalize_regexp_aux pc m =
  match m.pmodel_desc with
  | TGroupRef (Sym.RGroupSym tn) ->
      RSym (tn, Id.next pc)
  | TBound (m1,min,max) ->
      normalize_bound pc m1 min max 
  | TComma (m1,m2) ->
      RSeq (normalize_regexp_aux pc m1, normalize_regexp_aux pc m2)
  | TEmpty ->
      REmpty
  | TBar (m1,m2) ->
      RChoice (normalize_regexp_aux pc m1, normalize_regexp_aux pc m2)
  | TNone ->
      RNone         (* will raise exception. None is undefined in automata. 0 <: Everything, anyway *)
  | TAmp (m1,m2) ->
      raise (Query (Prototype "Subsumption: Ampersand not supported yet"))
  | _ ->
      raise (Query (Internal_Error "Normalizing non renamed schema"))

(* Note: Bound normalization proceeds in two steps.

     1. normalize lower bound decrements the lower bound until it
        reaches zero using the following law:

        if m > 0 then t{m,n} = t,t{m-1,n-1}

     2. normalize upper bound takes a bounded type with lower bound
        zero and normalizes the upper bound with either the following
        laws:

        if n = unbounded then t{0,unbounded} = t*
        if n < unbounded then t{0,n} = (() | t,t{0,n-1})

        remark that there are many ways t{0,n} could be normalized:

        t{0,n} = t?,t?....,t?
        t{0,n} = () | (t) | (t,t) | .... | (t,...,t)
        t{0,n} = () | (t, (() | (t, (..... (() | t)))))

        the law above generates the third normalization, which is the
        only one determninstic.

   Finally, there is a determinism problem when normalizing nested
   types. For instance:

   assuming:
   t = a{0,2}

   N(t)          --> (E|(a,(E|a)))

   N(t{0,2})     --> (E|(N(t),(E|N(t))))

                 --> (E|(a,(E|a))) , (E|(E|(a,(E|a))))

   - Jerome *)

(* BChoi: Modified from Jerome's email
                 (I)      t{0,*} == t*
                 (II)     t{1,*} == t+
                 (III)    t{m,*} == t, t{m-1, *}

                 (IV)     t{0,1} == ()|t
                 (IV')    t{0,n} == ()|t,t{0,n-1}
                 (V)      if m > 1 then t{m,n} == t,t{m-1,n-1}

*)

and normalize_bound pc m1 min max =
  match (min, max) with
  (* the common cases *)
  | (UP_INT 0, UNBOUNDED) -> RStar (normalize_regexp_aux pc m1)                                                                           (* (I) *)
  | (UP_INT 1, UNBOUNDED) -> RPlus (normalize_regexp_aux pc m1)                                                                           (* (II) *)
  | (UP_INT min', UNBOUNDED) -> RSeq (normalize_regexp_aux pc m1, normalize_bound pc m1 (UP_INT (min'-1)) UNBOUNDED)                      (* (III) *)

  | (UP_INT 0, UP_INT 1) -> RChoice (REmpty, normalize_regexp_aux pc m1)                                                                  (* (IV) *)
  | (UP_INT 0, UP_INT max') -> RChoice (REmpty, RSeq(normalize_regexp_aux pc m1, normalize_bound pc m1 (UP_INT 0) (UP_INT (max'-1))))     (* (IV') *)
  | (UP_INT min', UP_INT max') -> RSeq (normalize_regexp_aux pc m1, normalize_bound pc m1 (UP_INT (min'-1)) (UP_INT (max'-1)))            (* (V) *)
  | _ -> raise (Query (Internal_Error "Error during normaliznig bound type: Invalid bounds"))

(* and normalize_upper_bound *)
let normalize_regexp m =
  let position_counter = Id.create 1 in
  normalize_regexp_aux position_counter m


(* good for debugging *)

(*
let print_label x = print_string (Sym.rgroup_string x)

let print_glushkov me = TypeName_NFA.print_automata me print_label (Printf.printf "%i")
let print_dtm_glushkov me = TypeName_DFA.print_automata me print_label (Printf.printf "%i")
let print_intersection_glushkov me = TypeName_NFA_Pair.print_automata me print_label (fun (x,y) -> Printf.printf "(%i,%i)" x y)
*)

(* example *)
(* let me = build_glushkov (RSeq((RStar(RSym("a",1))),(RStar(RSym("b",2)))));;  *)
(* print_glushkov me;; *)

(************)
(* includes *)
(************)

(* BChoi: Assumption: the bigModel is deterministic. The *)
(* code does not statically/dynamically check this       *)
(* L(M1) is included in L(M2) iff                        *)
(* L(M1 intersects (not M2)) = Empty Set                 *)

let includes smallModel bigModel = 
  if !Conf.debugging
  then
    begin
      Print.printf_model "\nSmall model:" smallModel;
      Print.printf_model "\nBig model:" bigModel;
      flush stdout
    end;
  let mesmall = (build_glushkov (normalize_regexp smallModel)) and
      mebig = (build_dtm_glushkov (normalize_regexp bigModel)) in
  let all_letters = NFA.Alphabet.elements (NFA.get_nfa_alphabet mesmall) in
  let new_alph = List.fold_right DFA.Alphabet.add all_letters DFA.Alphabet.empty in
  let not_mebig = negates_in_alphabet_space mebig new_alph in
  not(NFAPair.reachable (intersects_dtm mesmall not_mebig))

(**************)
(* intersects *)
(**************)

(* Jerome: Assumption: the bigModel is deterministic. The *)
(* code does not statically/dynamically check this        *)
(* L(M1) intersects L(M2) iff                             *)
(* L(M1 intersects M2) != Empty Set                       *)

let intersects smallModel bigModel = 
  if !Conf.debugging
  then
    begin
      Print.printf_model "\nIntersection Small model:" smallModel;
      Print.printf_model "nIntersection Big model:" bigModel;
      flush stdout
    end;
  let mesmall = (build_glushkov (normalize_regexp smallModel)) and
      mebig = (build_dtm_glushkov (normalize_regexp bigModel)) in
  NFAPair.reachable (intersects_dtm mesmall mebig)

