(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: intersects.ml,v 1.1 2002/07/09 02:50:55 simeon Exp $ *)

open Namespace
open Xquery_ast
open Xquery_astutil
open Schema
open Schemautil
open Error

(***********************************)
(* Regular expression intersection *)
(***********************************)

(* Note: this part of the code deals with regular expressions intersection.
   Given an intersection mapping, and two models:
     1. the intersection mapping is applied on the model which comes from inference
     2. the two models are normalized as pure regular expressions (e.g., no t{m,n})
     3. the two corresponding automatas are constructed 
     4. the intersection algorithm is called
   - Jerome *)

(* Applies the intersection mapping on a given model *)
	
let rec apply_intersection_mapping s m th =
  match m.pmodel_desc with
  | TGroupRef (Sym.RGroupSym tn) ->
      if (is_unit_type s tn) then
	fmkmodel (TGroupRef (Sym.RGroupSym (Theta.find_if tn th))) Finfo.bogus
      else
	let m1 = get_group s tn in
	apply_intersection_mapping s m1 th
  | TBound (m1,min,max) ->
      fmkmodel (TBound (apply_intersection_mapping s m1 th, min, max)) Finfo.bogus
  | TComma (m1,m2) ->
      fmkmodel (TComma (apply_intersection_mapping s m1 th, apply_intersection_mapping s m2 th)) Finfo.bogus
  | TEmpty ->
      fmkmodel (TEmpty) Finfo.bogus
  | TBar (m1,m2) ->
      fmkmodel (TBar (apply_intersection_mapping s m1 th, apply_intersection_mapping s m2 th)) Finfo.bogus
  | TNone ->
      fmkmodel (TNone) Finfo.bogus
  | TAmp (m1,m2) ->
      fmkmodel (TAmp (apply_intersection_mapping s m1 th, apply_intersection_mapping s m2 th)) Finfo.bogus
  | _ ->
      raise (Query (Internal_Error "Applying subsumption mapping on non renamed schema"))

(* subsumption between two regular expression models *)

let regexp_intersects s s' m m' th =
  let theta_of_m = apply_intersection_mapping s m th in         (* applies the subsumption mapping to the small model *)
  Glushkov.intersects theta_of_m m'

(************************)
(* Intersection mapping *)
(************************)

(* Note: This part of the code deals with guessing a intersection
   mapping between two models. this is achieved by walking through the
   regular expressions using the derivatives. i.e., first/last
   functions
   - Jerome *)

(* add new mappings in sub_theta *)
(* fails in case of conflicts between mappings *)

let rec guess_intersection_map l l' =
  function (th,rest) ->
    match l with
    | [] ->
	(th,rest)
    | (lab,x,r) :: l1 ->
	try
	  match (List.find (fun (lab',_,_) -> Sym.subtag lab lab') l') with
	  | (_,x',r') ->
	      let (th1,rest1) = guess_intersection_map l1 l' (th,rest)
	      in (Theta.add x x' th1, (r,r') :: rest1)
	with
	| Not_found ->
	    guess_intersection_map l1 l' (th,rest)

(* Note: This part of the code performs one walk
   m and m' are the two models to walk through
   s and s' are the current schemas
   th is the current intersection mapping
   - Jerome *)

let one_step_guess_intersection_th s s' m m' th =
  let l  =
    try
      first_rest s m                     (* the first tokens and following regular expressions *)
    with
      _ ->
	[]
  and l' =
    first_rest_deterministic s' m' in
  guess_intersection_map l l' (th,[])    (* computes the theta mapping and the remainding types *)

(* Note: this function walks in the regular expression to guess a
   intersection mapping.
      s s' are the current schemas
      th is the current intersection mapping
      heap just keep track of visited types to terminate recursion
      ms is thw list of pairs of models to map together
   - Jerome *)

let rec guess_intersection_th_aux s s' th heap ms =
  match ms with
  | [] -> th                                              (* all types visited - we are done *)
  | (m,m') :: ms' ->                                      (* takes the first pair of models *)
      let (th', others) =                                 (* gets more mappings and pairs of models *)
	one_step_guess_intersection_th s s' m m' th in    (* by proceeding one step further in both models *)
      let new_ms =                                        (* filters all the new types to pair *)
	Gmisc.filter_non_exists others heap in
      let new_heap = heap @ new_ms in                     (* update the visited types *)
      guess_intersection_th_aux s s' th' new_heap (new_ms @ ms')  (* keep moving *)

let guess_intersection_th s s' m m' th =
  guess_intersection_th_aux s s' th [] [(m,m')]

(*****************************)
(* Tree Grammar intersection *)
(*****************************)

(* Note: this part of the code deals with the tree grammar parts of the intersection,
   using a bunch of (mutually recursive ?) functions that walk down the tree structure
   of types.
   - Jerome *)

(* Intersection between two regular expressions *)

let model_intersects s s' m m' known_th =
  (* Note: I kept the TNone safety check here for now. This should
     logically go away - Jerome *)
  let m1 = model_normalize s m
  and m1' = model_normalize s' m' in
  match (m1.pmodel_desc, m1'.pmodel_desc) with
  | (TNone,_) ->
      (* 0 /\ t is false *)
      raise (Query (Mapping_failure "Infered type is none"))
  | (_,TNone) ->
      (* t /\ 0 is also false! *)
      raise (Query (Mapping_failure "Given type is none"))
  | _ ->
      let new_th = guess_intersection_th s s' m1 m1' known_th in
      if (regexp_intersects s s' m1 m1' new_th)
      then
	new_th
      else
	raise (Query (Mapping_failure ("Could not map models during intersection: "
				       ^ (Print.bprintf_model "" m)
				       ^ " to: "
				       ^ (Print.bprintf_model "" m'))))

(* Intersection between two unit types *)

let unit_type_intersects s s' x x' th =
  let m  = (dereference_elem_attr s (dereference_rgroup s  x ))        (* gets the meaning m  of x  *)
  and m' = (dereference_elem_attr s (dereference_rgroup s' x')) in     (* gets the meaning m' of x' *)

  match (m.pmodel_desc,m'.pmodel_desc) with
  | (TElemLocal (Sym.RElemSym l, _, t), TElemLocal (Sym.RElemSym l',_,t')) ->
      if (Sym.relem_subtag l l')
      then
	begin
	  match (t.ptyped_model_desc,t'.ptyped_model_desc) with
	  | (TAnonymousType m1, TAnonymousType m1') ->
	      model_intersects s s' m1 m1' th
	end
      else
	raise (Query (Mapping_failure "Element tags unmatched"))
  | (TAttrLocal (Sym.RAttrSym l,t), TAttrLocal (Sym.RAttrSym l',t')) ->
      if (Sym.rattr_subtag l l')
      then
	begin
	  match (t.ptyped_model_desc,t'.ptyped_model_desc) with
	  | (TAnonymousType m1, TAnonymousType m1') ->
	      model_intersects s s' m1 m1' th
	end
      else
	raise (Query (Mapping_failure "Attribute tags unmatched"))
  | (TBase bt, TBase bt') ->
      if Simple.simple_type_subsumes bt bt'
      then Theta.empty
      else raise (Query (Mapping_failure ("Mapping incompatible base type")))
  | (TText, TText) ->
      Theta.empty
  | _ ->
      raise (Query (Internal_Error ("Checking subsumption on non incompatible unit types")))

(* Performs subsumption between a list of pairs or unit types (or goals) *)

let rec units_subsumes s s' goals known_th =
  match goals with
  | [] ->
      known_th                                               (* Ok, we are done *)
  | (x,x') :: goals' ->
      begin
	check_unit_type s x;                                 (* x  should be a unit type  *)
	check_unit_type s' x';                               (* x' should be a unit type  *)
	
	let new_th = unit_type_intersects s s' x x' known_th in(* subsumption on the corresp. unit types *)
	let new_goals =                                      (* what are the new goals ? *)
	  Gmisc.filter_non_exists
	    (Theta.content new_th)
	    (Theta.content known_th) in
	
	let global_th = Theta.merge new_th known_th          (* the new subsumption mapping *)
	and global_goals = new_goals @ goals' in             (* the new set of goals *)
	
	units_subsumes s s' global_goals global_th           (* keeps going ... *)
      end
	(* s and s' are the schemas for the first and second types respectively
	   goals is a list of pairs of two type names for which you want to check subsumption
	   known_th is the current subsumption mapping, i.e., current subtyping assumptions *)

(*********************************)
(* Global subsumptions functions *)
(*********************************)

(* Note: this part of the code provides just wrappers of the above subsumption
   algorithm for basic type components. All global subsumption functions return
   a theta mapping or fail.
   - Jerome *)

(* Subsumption between two unit types *)

let unit_subsumption s s' x x' =
  let init_th = Theta.empty                          (* starts with an empty mapping *)
  and init_goals = [(x,x')] in                       (* starts with a pair of unit types to subsume *)
  units_subsumes s s' init_goals init_th             (* applies unit types subsumption *)
    (* checks whether x_s <: x'_s' *)

(* Subsumption between two arbitrary types *)

let model_subsumption s s' m m' =
  let init_th = Theta.empty in                       (* starts with an empty mapping *)
  let new_th = model_intersects s s' m m' init_th in   (* applies regular expressions subsumption *)
  let init_goals = Theta.content new_th in           (* gets all the original goals *)
  units_subsumes s s' init_goals new_th              (* applies subsumption on each unit type *)
    (* checks whether t_s <: t_s' *)

(* Subsumption between two schemas *)

let schema_subsumption s s' =
  let rxo = get_root s
  and rxo' = get_root s' in
  match (rxo,rxo') with
  | (None,_) -> raise (Query (Mapping_failure "Schema has no root type"))
  | (_,None) -> raise (Query (Mapping_failure "Schema has no root type"))
  | (Some rx, Some rx') ->
      let m  = (fmkmodel (TElemRef (Sym.RElemSym rx)) Finfo.bogus)
      and m' = (fmkmodel (TElemRef (Sym.RElemSym rx')) Finfo.bogus) in
      model_subsumption s s' m m'
	(* checks whether root_s <: root'_s'
	   i.e., subsumption between the roots of two schemas *)

(*************)
(* Subtyping *)
(*************)

(* Note: this part of the code is the main subtyping function used by the type
   checker.
   - Jerome *)

(* Subtyping function used by the type-checker *)

let isSubtype s m1 m2 =
  try
    let s' = compile_schema s in
    let g_sym1 = (Sym.rgroup_symbol (NSDefaultPrefix, (NSUri "", "_root1")))
    and g_sym2 = (Sym.rgroup_symbol (NSDefaultPrefix, (NSUri "", "_root2"))) in
    let (m1',sdl1) = compile_model s g_sym1 m1
    and (m2',sdl2) = compile_model s g_sym2 m2 in
    let (s1,s2) =
      match (sdl1,sdl2) with
      | ([],[]) ->
	  (s',s')
      | _ ->
	  let s'' = copy_schema s' in
	  insert_bulk s'' sdl1;
	  insert_bulk s'' sdl2;
	  (s'',s'')
    in
    ignore(model_subsumption s1 s2 m1' m2');
    true
  with
  | Query (Non_Deterministic_Regexp _) as ex ->
      raise ex
  | _ ->
      false

