(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: schema.ml,v 1.32 2002/12/03 00:01:43 simeon Exp $ *)

open Namespace
open Resolve
open Xquery_ast
open Xquery_astutil
open Print
open Error

(**********************************)
(* schema internal representation *)
(**********************************)

(* Whether a global element or attribute has a local counterpart *)

type conflict = bool

(* Schema type *)

type schema =
    { mutable schema_name : sname option;                                              (* Schema name *)
      mutable schema_root : Sym.relem_symbol option;                                  (* Root element of the schema *)
      schema_element_names : (Sym.relem_symbol, unit) Hashtbl.t;                       (* All element names *)
      schema_attribute_names : (Sym.rattr_symbol, unit) Hashtbl.t;                     (* All attribute names *)
      schema_groups : (Sym.rgroup_symbol, model) Hashtbl.t;                            (* Groups *)
      schema_elements : (Sym.relem_symbol, (conflict ref * typed_model)) Hashtbl.t;    (* Global elements *)
      schema_attributes : (Sym.rattr_symbol, (conflict ref * typed_model)) Hashtbl.t } (* Global attributes *)

(* Global variable storing all current schemas *)

let schema_table : (string, schema) Hashtbl.t = Hashtbl.create 137
let store_schema sn s = Hashtbl.add schema_table sn s

(***********************************************)
(* Namespace resolution and conflict detection *)
(***********************************************)

(* Adds names in the schema tables *)

let add_elem_name s ren =
  Hashtbl.add s.schema_element_names ren ();
  try
    let (c,_) = Hashtbl.find s.schema_elements ren in
    c := false
  with
  | Not_found -> ()

let add_attr_name s ran =
  Hashtbl.add s.schema_attribute_names ran ();
  try
    let (c,_) = Hashtbl.find s.schema_attributes ran in
    c := false
  with
  | Not_found -> ()

(* Model resolution *)

(* Note:
     resolve_model: walk through the model tree and resolve the qname
     on its way. There is no need to alter the structure fo the tree
   - J&B
*)

(* Resolves namespaces within a model *)

let rec resolve_model nsr s m =
  match m.pmodel_desc with 
  | TAttrRef attr_sym ->
      let rattr_sym = resolve_attr_symbol nsr attr_sym in
      add_attr_name s rattr_sym;
      fmkmodel (TAttrRef (Sym.RAttrSym rattr_sym)) m.pmodel_loc
  | TAttrLocal (attr_sym, tm) -> 
      let rattr_sym = resolve_attr_symbol nsr attr_sym in 
      add_attr_name s rattr_sym;
      fmkmodel (TAttrLocal (Sym.RAttrSym rattr_sym, resolve_typed_model nsr s tm)) m.pmodel_loc
  | TElemRef elem_sym -> 
      let relem_sym = resolve_elem_symbol nsr elem_sym in 
      add_elem_name s relem_sym;
      fmkmodel (TElemRef (Sym.RElemSym relem_sym)) m.pmodel_loc
  | TElemLocal (elem_sym, deriv, tm) -> 
      let relem_sym = resolve_elem_symbol nsr elem_sym in 
      add_elem_name s relem_sym;
      fmkmodel (TElemLocal (Sym.RElemSym relem_sym, deriv, resolve_typed_model nsr s tm)) m.pmodel_loc
  | TGroupRef group_sym ->
      let rgroup_sym = resolve_group_symbol nsr group_sym in
      fmkmodel (TGroupRef (Sym.RGroupSym rgroup_sym)) m.pmodel_loc
  | TBound (m, b1, b2) ->
      fmkmodel (TBound (resolve_model nsr s m, b1, b2)) m.pmodel_loc
  | TComma (m1, m2) ->
      fmkmodel (TComma (resolve_model nsr s m1, resolve_model nsr s m2)) m.pmodel_loc
  | TBar (m1, m2) ->
      fmkmodel (TBar (resolve_model nsr s m1, resolve_model nsr s m2)) m.pmodel_loc
  | TAmp (m1, m2) ->
      fmkmodel (TAmp (resolve_model nsr s m1, resolve_model nsr s m2)) m.pmodel_loc
  | TEmpty ->
      m
  | TNone ->
      m
    (* MF & BC : TBase isn't correct, because we could refer to base types in 
       some namespace -- see comment in xquery_parser.mly  *)
  | TBase b ->
      m
  | TDocument m ->
      fmkmodel (TDocument (resolve_model nsr s m)) m.pmodel_loc
  | TText ->
      m
  | TTuple t ->
      m

and resolve_typed_model nsr s tm = 
  match tm.ptyped_model_desc with 
  | TAnonymousType m ->
      fmktyped_model (TAnonymousType (resolve_model nsr s m)) tm.ptyped_model_loc

(* Checks if a new global element conflicts with an existing local element *)

let is_elem_conflict s relem_sym =
  not(Hashtbl.mem s.schema_element_names relem_sym)

(* Checks if a new global attribute conflicts with an existing local attribute *)

let is_attr_conflict s rattr_sym =
  not(Hashtbl.mem s.schema_attribute_names rattr_sym)

(************************************)
(* Elementary operations on schemas *)
(************************************)

(* Update an existing schema *)

(* Adds a type in a schema *)

let insert_and_resolve_group nsr s ugroup_sym m =
  let m' = resolve_model nsr s m in
  let rgroup_sym = resolve_group_symbol nsr ugroup_sym in
  Hashtbl.add s.schema_groups rgroup_sym m'

(* Adds an element in a schema *)

let insert_and_resolve_element nsr s uelem_sym tm =
  let tm' = resolve_typed_model nsr s tm in
  let relem_sym = resolve_elem_symbol nsr uelem_sym in
  let conflict = is_elem_conflict s relem_sym in
  Hashtbl.add s.schema_elements relem_sym ((ref conflict),tm')

(* Adds an attribute in a schema *)

let insert_and_resolve_attribute nsr s uattr_sym tm =
  let tm' = resolve_typed_model nsr s tm in
  let rattr_sym = resolve_attr_symbol nsr uattr_sym in
  let conflict = is_attr_conflict s rattr_sym in
  Hashtbl.add s.schema_attributes rattr_sym ((ref conflict),tm')

(* Adds a toplevel sort declaration in a schema *)

let insert_and_resolve_tnt nsr s sd =
  match sd.psort_decl_desc with
  | TypeDecl  (ugs,m) -> insert_and_resolve_group     nsr s ugs m
  | ElemDecl  (ues,m) -> insert_and_resolve_element   nsr s ues m
  | AttrDecl  (uas,m) -> insert_and_resolve_attribute nsr s uas m

(* Adds a list of sort declarations in a schema *)

let insert_and_resolve_bulk nsr s sds =
  List.iter (insert_and_resolve_tnt nsr s) sds

(* Adds a type in a schema *)

let insert_group s rgroup_sym m =
  Hashtbl.add s.schema_groups rgroup_sym m

(* Adds an element in a schema *)

let insert_element s relem_sym ctm =
  Hashtbl.add s.schema_elements relem_sym ctm

(* Adds an attribute in a schema *)

let insert_attribute s rattr_sym ctm =
  Hashtbl.add s.schema_attributes rattr_sym ctm

(* Adds a toplevel sort declaration in a schema *)

let insert_tnt s (c,sd) =
  match sd.psort_decl_desc with
  | TypeDecl (Sym.RGroupSym rgs,m) -> insert_group     s rgs m
  | TypeDecl (Sym.UGroupSym _,_)   ->
      raise (Query (Internal_Error ("Dereferencing an unresolved group")))
  | ElemDecl  (Sym.RElemSym res,tm) -> insert_element   s res (c,tm)
  | ElemDecl  (Sym.UElemSym _,_) ->
      raise (Query (Internal_Error ("Dereferencing an unresolved element")))
  | AttrDecl  (Sym.RAttrSym ras,tm) -> insert_attribute s ras (c,tm)
  | AttrDecl  (Sym.UAttrSym _,_) ->
      raise (Query (Internal_Error ("Dereferencing an unresolved attribute")))

(* Adds a list of sort declarations in a schema *)

let insert_bulk s sds =
  List.iter (insert_tnt s) sds

(* Creates the schema root *)

let create_and_resolve_root nsr s uro =
  match uro with
  | None -> ()
  | Some ur ->
      let rr = resolve_uelem_qname nsr ur in
      s.schema_root <- (Some rr)

let create_root s rro =
  s.schema_root <- rro

(* Creates the schema name *)

let create_schema_name s sno =
  s.schema_name <- sno

(* Accesses schema information *)

(* Gets all the group names *)

let get_all_groups s =
  let gl = ref [] in
  Hashtbl.iter (fun x y -> (gl := x :: !gl)) s.schema_groups;
  !gl

(* Gets the root of the schema *)

let get_root s =
  s.schema_root

(* Gets the name of the schema *)

let get_name s =
  s.schema_name

(* get a toplevel group *)

let get_group s rgroup_sym =
  try
    Hashtbl.find s.schema_groups rgroup_sym
  with
  | Not_found ->
      raise (Query (Unknown ("Unknown group " ^ (Sym.rgroup_string rgroup_sym))))

(* Get a toplevel element *)

let get_element s relem_sym =
  try
    let (_,e) = Hashtbl.find s.schema_elements relem_sym in e
  with
  | Not_found ->
      raise (Query (Unknown ("Unknown element " ^ (Sym.relem_string relem_sym))))

(* Get a toplevel attribute *)

let get_attribute s rattr_sym =
  try
    let (_,a) = Hashtbl.find s.schema_attributes rattr_sym in a
  with
  | Not_found ->
      raise (Query (Unknown ("Unknown attribute " ^ (Sym.rattr_string rattr_sym))))

(* Checks if a given element reference exists in the schema *)

let element_exists s relem_sym =
  ignore(get_element s relem_sym)

let attribute_exists s rattr_sym =
  ignore(get_attribute s rattr_sym)

let group_exists s rgroup_sym =
  ignore(get_group s rgroup_sym)

(* Checks if a global elements exists and does not conflicts with a local element *)

let is_unique_global_element s relem_sym =
  try
    let (c,_) = Hashtbl.find s.schema_elements relem_sym in !c
  with
  | Not_found ->
      false

(* Checks if a global attribute exists and does not conflicts with a local attribute *)

let is_unique_global_attribute s rattr_sym =
  try
    let (c,_) = Hashtbl.find s.schema_attributes rattr_sym in !c
  with
  | Not_found ->
      false

(* Resolves a group reference *)

let dereference_rgroup s rgroup_sym =
  let rec dereference_rgroup_aux s rgroup_sym' visited_rgroup_syms =
    let t = get_group s rgroup_sym' in
    match t.pmodel_desc with
    | TGroupRef (Sym.RGroupSym rgroup_sym'') ->
	begin
	  if List.exists (fun x -> Sym.rgroup_equals x rgroup_sym'') visited_rgroup_syms
	  then
	    raise
	      (Query (Wrong_Args ("Cannot resolve group "
				  ^ (Sym.rgroup_string rgroup_sym))))
	  else
	    dereference_rgroup_aux s rgroup_sym'' (rgroup_sym'' :: visited_rgroup_syms)
	end
    | TGroupRef (Sym.UGroupSym _) ->
	raise (Query (Internal_Error ("Dereferencing an unresolved group")))
    | _ ->
	t
  in
  dereference_rgroup_aux s rgroup_sym [rgroup_sym]

(* Resolves a group reference - keep tracks of visited names *)

let dereference_rgroup_stats s rgroup_sym =
  let rec dereference_rgroup_stat_aux s rgroup_sym' visited_rgroup_syms =
    let t = get_group s rgroup_sym' in
    match t.pmodel_desc with
    | TGroupRef (Sym.RGroupSym rgroup_sym'') ->
	begin
	  if List.exists (fun x -> Sym.rgroup_equals x rgroup_sym'') visited_rgroup_syms
	  then
	    raise
	      (Query (Wrong_Args ("Cannot resolve group "
				  ^ (Sym.rgroup_string rgroup_sym))))
	  else
	    dereference_rgroup_stat_aux s rgroup_sym'' (rgroup_sym'' :: visited_rgroup_syms)
	end
    | TGroupRef (Sym.UGroupSym _) ->
	raise (Query (Internal_Error ("Dereferencing an unresolved group")))
    | _ ->
	(* Note:
	     The list below is returned reversed wrt the order in
	     which the groups are visited. This need to be taken into
	     accound during validation with statistics generation.
	   - Jerome
	 *)
	(t,visited_rgroup_syms)
  in
  dereference_rgroup_stat_aux s rgroup_sym [rgroup_sym]

(* Check recursion criterias *)

let rec dereference_group s gsym =
  let rgsym =
    match gsym with
    | Sym.RGroupSym rgsym -> rgsym
    | Sym.UGroupSym _ ->
	raise (Query (Internal_Error "Depend : cannot access a non resolved group"))
  in
  dereference_rgroup s rgsym

(* Resolves an element/attribute reference *)

let dereference_model s m =
  match m.pmodel_desc with
  | TGroupRef gsym -> dereference_group s gsym
  | _ -> m

let dereference_elem_attr s m =
  match m.pmodel_desc with
  | TElemRef (Sym.RElemSym relem_sym) ->
      fmkmodel (TElemLocal (Sym.RElemSym relem_sym, None, get_element s relem_sym)) Finfo.bogus
  | TElemRef (Sym.UElemSym _) ->
      raise (Query (Internal_Error ("Dereferencing an unresolved element")))
  | TAttrRef (Sym.RAttrSym rattr_sym) ->
      fmkmodel (TAttrLocal (Sym.RAttrSym rattr_sym, get_attribute s rattr_sym)) Finfo.bogus
  | TAttrRef (Sym.UAttrSym _) ->
      raise (Query (Internal_Error ("Dereferencing an unresolved attribute")))
  | _ ->
      m

(*********************************)
(* Operation on compiled schemas *)
(*********************************)

(* Creates an empty schema *)

let empty_schema () =
  { schema_name = None;
    schema_root = None;
    schema_element_names = Hashtbl.create 167;
    schema_attribute_names = Hashtbl.create 167;
    schema_groups = Hashtbl.create 167;
    schema_elements = Hashtbl.create 167;
    schema_attributes = Hashtbl.create 167 }

(* Creates a new schema *)

(* Note:
     Create schema takes a schema name, a root name (group) and a list
     of sort declarations and compiles into a schema in which all
     names are resolved
   - J&B *)

let create_schema nsr sn sr sds =
  let s = empty_schema () in
  List.iter (insert_and_resolve_tnt nsr s) sds;
  create_schema_name s sn;
  create_and_resolve_root nsr s sr;
  s

let create_schema_with_env nsr schema sn sr sds =
  let s = schema in
  List.iter (insert_and_resolve_tnt nsr s) sds;
  create_schema_name s sn;
  create_and_resolve_root nsr s sr;
  s

(* Copy a schema *)

let copy_schema s1 =
  let s2 = empty_schema () in
  Hashtbl.iter (insert_group s2) s1.schema_groups;
  Hashtbl.iter (insert_element s2) s1.schema_elements;
  Hashtbl.iter (insert_attribute s2) s1.schema_attributes;
  create_schema_name s2 (get_name s1);
  create_root s2 (get_root s1);
  s2

(* Exports a schema *)

let dump_schema s =
  let sds = ref [] in
  Hashtbl.iter
    (fun tn ty ->
      let nt = (fmksort_decl (TypeDecl (Sym.RGroupSym tn,ty)) Finfo.bogus) in
      sds := nt :: !sds) s.schema_groups;
  Hashtbl.iter
    (fun l (_,tm) ->
      let nt = (fmksort_decl (ElemDecl (Sym.RElemSym l,tm)) Finfo.bogus) in
      sds := nt :: !sds) s.schema_elements;
  Hashtbl.iter
    (fun l (_,tm) ->
      let nt = (fmksort_decl (AttrDecl (Sym.RAttrSym l,tm)) Finfo.bogus) in
      sds := nt :: !sds) s.schema_attributes;
  !sds

(**********************)
(* Printing functions *)
(**********************)

(* Printing a schema *)

let print_schema s =
  let sds = dump_schema s in
  List.iter (printf_sort_decl "") sds

(************************)
(* Schema normalization *)
(************************)

(* converts from user schema to internal schema *)

(* This function recursively looks up type/group names until it 
   finds a unique symbol. *)

let rec create_new_group_name s counter rgroup_sym =
  let (uri,ln) = Sym.rgroup_name rgroup_sym in
  let prefix = Sym.rgroup_prefix rgroup_sym in
  let new_ln = "_" ^ ln ^ "_" ^ (string_of_int (Id.next counter)) in
  let new_rgroup_sym = Sym.rgroup_symbol (prefix, (uri, new_ln)) in
  try 
    let dupt = get_group s new_rgroup_sym
    in create_new_group_name s counter new_rgroup_sym
  with
  | _ ->
      new_rgroup_sym

(* Note:
     rconv returns a pair:
     - first component of the pair is the newly created type variable
       (group reference) that needs to replace the element where it
       was found
     - second component of the pair is the list of sort_decl that
       needs to be added in the schema
   - Jerome *)

let rec rconv s counter rgroup_sym m =
  match m.pmodel_desc with
  | TGroupRef group_sym ->
      group_exists s (Sym.rgroup_symbol_of_group_symbol group_sym);  (* Just checks schema consistency here - Jerome *)
      (m,[])
  | TElemRef elem_sym ->
      element_exists s (Sym.relem_symbol_of_elem_symbol elem_sym);  (* Just checks schema consistency here - Jerome *)
      let rgroup_sym' = create_new_group_name s counter rgroup_sym in
      let new_group_decl =
	fmksort_decl (TypeDecl (Sym.RGroupSym rgroup_sym',m)) m.pmodel_loc
      in
      (fmkmodel (TGroupRef (Sym.RGroupSym rgroup_sym')) m.pmodel_loc,
       [(ref true,new_group_decl)])
  | TElemLocal (l', d', tm) ->
      begin
	match tm.ptyped_model_desc with 
	| TAnonymousType m' -> 
	    let rgroup_sym' = create_new_group_name s counter rgroup_sym in
	    let (m'',addt) = rconv s counter rgroup_sym m' in
	    let new_group_decl =
	      fmksort_decl (TypeDecl (Sym.RGroupSym rgroup_sym',
				       fmkmodel (TElemLocal (l',
							     d',
							     (fmktyped_model (TAnonymousType m'') m''.pmodel_loc))) m'.pmodel_loc)) Finfo.bogus
	    in
	    (fmkmodel (TGroupRef (Sym.RGroupSym rgroup_sym')) m.pmodel_loc, 
	     (ref true,new_group_decl) :: addt)
      end
  | TAttrRef attr_sym ->
      attribute_exists s (Sym.rattr_symbol_of_attr_symbol attr_sym);  (* Just checks schema consistency here - Jerome *)
      let rgroup_sym' = create_new_group_name s counter rgroup_sym in
      let new_group_decl =
	fmksort_decl (TypeDecl (Sym.RGroupSym rgroup_sym',m)) m.pmodel_loc
      in
      (fmkmodel (TGroupRef (Sym.RGroupSym rgroup_sym')) m.pmodel_loc,
       (ref true,new_group_decl) :: [])
  | TAttrLocal (l', tm) ->
      begin
	match tm.ptyped_model_desc with 
	| TAnonymousType m' -> 
	    let rgroup_sym' = create_new_group_name s counter rgroup_sym in
	    let (m'',addt) = rconv s counter rgroup_sym m' in
	    let new_group_decl =
	      fmksort_decl (TypeDecl (Sym.RGroupSym rgroup_sym',
				       fmkmodel (TAttrLocal (l',
							     (fmktyped_model (TAnonymousType m'') m''.pmodel_loc))) m'.pmodel_loc)) Finfo.bogus
	    in
	    (fmkmodel (TGroupRef (Sym.RGroupSym rgroup_sym')) m.pmodel_loc, 
	     (ref true,new_group_decl) :: addt)
      end
  |  TBase _ ->
      let rgroup_sym' = create_new_group_name s counter rgroup_sym in
      let new_group_decl =
	fmksort_decl (TypeDecl (Sym.RGroupSym rgroup_sym', m)) Finfo.bogus
      in
      (fmkmodel (TGroupRef (Sym.RGroupSym rgroup_sym')) m.pmodel_loc,
       (ref true,new_group_decl) :: [])
  |  TText ->
      let rgroup_sym' = create_new_group_name s counter rgroup_sym in
      let new_group_decl =
	fmksort_decl (TypeDecl (Sym.RGroupSym rgroup_sym', m)) Finfo.bogus
      in
      (fmkmodel (TGroupRef (Sym.RGroupSym rgroup_sym')) m.pmodel_loc,
       (ref true,new_group_decl) :: [])
  | TComma (m1,m2) ->
      let (m1',addt1) = rconv s counter rgroup_sym m1
      and (m2',addt2) = rconv s counter rgroup_sym m2 in
      (fmkmodel (TComma (m1',m2')) m.pmodel_loc, addt1@addt2)
  | TBar (m1,m2) ->
      let (m1',addt1) = rconv s counter rgroup_sym m1
      and (m2',addt2) = rconv s counter rgroup_sym m2 in
      (fmkmodel (TBar (m1',m2')) m.pmodel_loc, addt1@addt2)
  | TBound (m1,i,j) ->
      let (m1',addt1) = rconv s counter rgroup_sym m1 in
      (fmkmodel (TBound (m1',i,j)) m.pmodel_loc, addt1)
  | TDocument m ->
      let rgroup_sym' = create_new_group_name s counter rgroup_sym in
      let (m',addt) = rconv s counter rgroup_sym m in
      let new_group_decl =
	fmksort_decl (TypeDecl (Sym.RGroupSym rgroup_sym',
				fmkmodel (TDocument m') m.pmodel_loc)) Finfo.bogus
      in
      (fmkmodel (TGroupRef (Sym.RGroupSym rgroup_sym')) m.pmodel_loc, 
       (ref true,new_group_decl) :: addt)
  | TAmp _ ->
      (m,[])
  | TNone ->
      (m,[])
  | TEmpty ->
      (m,[])
  | TTuple _ ->
      let rgroup_sym' = create_new_group_name s counter rgroup_sym in
      let new_group_decl =
	fmksort_decl (TypeDecl (Sym.RGroupSym rgroup_sym', m)) Finfo.bogus
      in
      (fmkmodel (TGroupRef (Sym.RGroupSym rgroup_sym')) m.pmodel_loc,
       (ref true,new_group_decl) :: [])

(* Toplevel group conversion
   - Always return a group declaration *)

let group_conv s counter rgroup_sym m =
  match m.pmodel_desc with
  | TElemLocal (l', d', tm) ->
      begin
	match tm.ptyped_model_desc with 
	| TAnonymousType m' -> 
	    let (m'', addt) = rconv s counter rgroup_sym m' in
	    let new_group_decl =
	      fmksort_decl (TypeDecl (Sym.RGroupSym rgroup_sym,
				       fmkmodel (TElemLocal (l',
							     d',
							     fmktyped_model (TAnonymousType m'') m''.pmodel_loc)) m'.pmodel_loc)) Finfo.bogus
	    in
	    (ref true,new_group_decl) :: addt
      end
  | TElemRef _ ->
      let new_group_decl =
	fmksort_decl (TypeDecl(Sym.RGroupSym rgroup_sym,
				m)) Finfo.bogus
      in
      (ref true,new_group_decl) :: []
  | TAttrLocal (l', tm) ->
      begin
	match tm.ptyped_model_desc with 
	| TAnonymousType m' -> 
	    let (m'', addt) = rconv s counter rgroup_sym m' in
	    let new_group_decl =
	      fmksort_decl (TypeDecl (Sym.RGroupSym rgroup_sym,
				      fmkmodel (TAttrLocal(l',
							   fmktyped_model (TAnonymousType m'') m''.pmodel_loc)) m'.pmodel_loc)) Finfo.bogus
	    in
	    (ref true,new_group_decl) :: addt
      end
  | TBase _ ->
      let new_group_decl =
	fmksort_decl (TypeDecl (Sym.RGroupSym rgroup_sym,
				m)) Finfo.bogus
      in
      (ref true,new_group_decl) :: []
  | TDocument m ->
      let (m', addt) = rconv s counter rgroup_sym m in
      let new_group_decl =
	fmksort_decl (TypeDecl (Sym.RGroupSym rgroup_sym,
				fmkmodel (TDocument m') m.pmodel_loc)) Finfo.bogus
      in
      (ref true,new_group_decl) :: addt
  | TText ->
      let new_group_decl =
	fmksort_decl (TypeDecl (Sym.RGroupSym rgroup_sym,
				m)) Finfo.bogus
      in
      (ref true,new_group_decl) :: []
  | _ ->
      let (m', addt) = rconv s counter rgroup_sym m in
      let new_group_decl =
	fmksort_decl (TypeDecl (Sym.RGroupSym rgroup_sym, m')) Finfo.bogus
      in
      (ref true,new_group_decl) :: addt

(* Toplevel element conversion
   - Always return a element declaration *)

let elem_conv s counter rgroup_sym l c tm =
  match tm.ptyped_model_desc with
  | TAnonymousType m' -> 
      let (m'', addt) = rconv s counter rgroup_sym m' in
      let new_elem_decl =
	fmksort_decl (ElemDecl(l,
			       fmktyped_model (TAnonymousType m'') m''.pmodel_loc)) m'.pmodel_loc
      in
      (c,new_elem_decl) :: addt

let attr_conv s counter rgroup_sym l c tm =
  match tm.ptyped_model_desc with 
  | TAnonymousType m' -> 
      let (m'', addt) = rconv s counter rgroup_sym m' in
      let new_attr_decl =
	fmksort_decl (AttrDecl (l,
				fmktyped_model (TAnonymousType m'') m''.pmodel_loc)) m'.pmodel_loc
      in (c,new_attr_decl) :: addt

(* Convert a group declaration *)

let convert_group ns os tn m =
  let idg = Id.create 0 in
  insert_bulk ns (group_conv os idg tn m)

(* Convert an element declaration *)

let convert_element ns os l (c,tm) =
  let idg = Id.create 0 in
  let (uri, ln) = Sym.relem_name l in
  let prefix = Sym.relem_prefix l in
  let rgroup_sym = Sym.rgroup_symbol (prefix, (uri , "_element_" ^ ln)) in
  insert_bulk ns (elem_conv os idg rgroup_sym (Sym.RElemSym l) c tm)

(* Convert an attribute declaration *)

let convert_attribute ns os l (c,tm) =
  let idg = Id.create 0 in
  let (uri, ln) = Sym.rattr_name l in
  let prefix = Sym.rattr_prefix l in
  let rgroup_sym = Sym.rgroup_symbol (prefix, (uri, "_attribute_" ^ ln)) in
  insert_bulk ns (attr_conv os idg rgroup_sym (Sym.RAttrSym l) c tm)

(* Normalizes a schema *)

let compile_schema s1 =
  let s2 = empty_schema () in
  Hashtbl.iter (convert_group s2 s1) s1.schema_groups;
  Hashtbl.iter (convert_element s2 s1) s1.schema_elements;
  Hashtbl.iter (convert_attribute s2 s1) s1.schema_attributes;
  create_schema_name s2 (get_name s1);
  create_root s2 (get_root s1);
  s2

(* Convert a toplevel model *)

(* Note:
     This function is used for subsumption
   - Jerome
 *)

let compile_model s group_sym m =
  let idg = Id.create 0 in
  rconv s idg group_sym m

