(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

open Namespace
open Sym
open Datatypes

open Xml_ast            (* For XML schema AST *)
open Xquery_ast         (* For XQuery types *)
open Xquery_astutil     (* To help build XQuery types *)

open Error

(************************************************)
(* Mapping XML Schema to the XQuery type system *)
(************************************************)

(* Note:
     This module provides functions to map XML Schema to the sort
     declarations.
   - Byron
 *)

(********************)
(* helper functions *)
(********************)

(* helper functions for element/attribute names *)

(* a special comparator for checking element tag  *)
(* nname1 and nname2 format: {namespace}localname or namespace:localname *)

(* Note:
     This part is completely bogus:
       - compareLocal should be done after namespace resolution and on
         global names.
       - compareLocal should be case sensitive.
   - Jerome
 *)

let compareLocal nname1 nname2 = 
  let localname1 = 
    try 
      let colon1 = String.index nname1 '}' and
	  l1 = String.length nname1 in
      (String.sub nname1 (colon1+1) (l1 - (colon1+1)))
    with
    | Not_found -> 
	begin
	  try 
	    let colon1' = String.index nname1 ':' and
		l1' = String.length nname1 in
	    (String.sub nname1 (colon1'+1) (l1' - (colon1'+1)))
	  with
	  | Not_found ->
	      nname1
	end
  in
  let localname2 = 
    try 
      let colon2 = String.index nname2 '}' and
	  l2 = String.length nname2 in
      (String.sub nname2 (colon2+1) (l2 - (colon2+1)))
    with
    | Not_found -> 
	begin
	  try 
	    let colon2' = String.index nname2 ':' and
		l2' = String.length nname2 in
	    (String.sub nname2 (colon2'+1) (l2' - (colon2'+1)))
	  with
	  | Not_found ->
	      nname2
	end
  in
  (String.lowercase localname1) = (String.lowercase localname2)

let getLocalName_aux nname =
  try 
    let colon = String.index nname '}' and
	l = String.length nname in
    (String.sub nname (colon+1) (l - (colon+1)))
  with
  | Not_found -> 
      begin
	try 
	  let colon' = String.index nname ':' and
	      l' = String.length nname in
	  (String.sub nname (colon'+1) (l' - (colon'+1)))
	with
	| Not_found ->
	    nname
      end

let getLocalName e =
  match e with
  | XElem (UElemSym n, _, _) ->
      getLocalName_aux (string_of_uqname (uelem_name n))
  | _ ->
      raise (Query (Internal_Error "Getting local name of unexpected expression"))

(* helper functions for attributes *)

let isAttrWithName name attr =
  match attr with
  | (UAttrSym n, _) ->
      compareLocal name (string_of_uqname (uattr_name n))
  | _ ->
      false

let getAttrString attr =
  match attr with
  | (UAttrSym n, l) -> l
  | _ -> raise (Query (Internal_Error "Getting attribute's value from a non-attribute"))

let findFromAttrlist attrlist name =
  try let attr = List.find (isAttrWithName name) attrlist in true
  with
  | Not_found -> false

let getFromAttrlist attrlist name =
  List.find (isAttrWithName name) attrlist

let probeAttr attrlist str =
  let bn = findFromAttrlist attrlist str in
  let str' = 
    if bn then 
      Some (getAttrString (getFromAttrlist attrlist str)) 
    else 
      None
  in
  str'

(* helper functions for base type *)

let isbasety tn =
  let tn' = String.lowercase tn in
  compareLocal tn' "string" || compareLocal tn' "integer" || compareLocal tn' "float" || compareLocal tn' "boolean" || compareLocal tn' "decimal" || compareLocal tn' "double" || compareLocal tn' "int"

let tobasety tn =
  if (compareLocal tn "string") then ATString
  else if (compareLocal tn "integer") then ATDecimal
  else if (compareLocal tn "float") then ATFloat
  else if (compareLocal tn "boolean") then ATBoolean
  else if (compareLocal tn "decimal") then ATDecimal
  else if (compareLocal tn "int") then ATInt
  else if (compareLocal tn "double") then ATDouble
  else raise (Query (Internal_Error "Not a basetype"))

(* helper functions for converting data *)

let qname_of_string s =
  try
    let colon = String.index s ':'
    and	l = String.length s in
    let prefix = String.sub s 0 colon
    and	local = String.sub s (colon+1) (l -(colon+1))
    in
    (NSPrefix prefix, local)
  with 
  | Not_found ->
      (NSDefaultElementPrefix, s)

let bound_of_string s =
  try 
    Occurrence.occurs (int_of_string s)
  with
  | _ -> 
      if (String.lowercase s = "unbounded") 
      then 
	Occurrence.unbounded
      else
	raise (Query (Internal_Error "Not a valid schema: Bound"))

let rec string_list_of_string delim str l =
  try 
    let i = String.index str delim in
    let token = String.sub str 0 i in
    let str' = String.sub str (i+1) ((String.length str) - i) in
    string_list_of_string delim str' (token::l)
  with
  | Not_found -> l 
  | _ -> raise (Query (Internal_Error "string_list_of_string"))
	
let rec seq_of_list modelList =
  match modelList with
  | [] -> fmkmodel TEmpty Finfo.bogus
  | h::[] -> h
  | h::t  ->
      fmkmodel (TComma (h, seq_of_list t)) Finfo.bogus 

let choice_of_list modelList =
  match modelList with
  | [] -> fmkmodel TNone Finfo.bogus
  | h::[] -> h
  | h::t  ->
      fmkmodel (TBar (h, seq_of_list t)) Finfo.bogus 

let filter_nonelem elist =
  let f e =
    match e with
    | XElem (UElemSym n,attrlist, elemlist)  -> 
      let s = String.lowercase (getLocalName e) in
      begin
      match s with 
      |	"annotation" -> Error.eprintf_warning ("Annotation is ignored."); false
      | "anyattribute" -> Error.eprintf_warning ("Anyattribute is ignored."); false
      | _ -> true
      end
    | _ -> false
  in
  List.filter f elist

(*********************************)
(* Common auxilary map functions *)
(*********************************)

let map_name p uqname =
  match uqname with
  | (NSDefaultElementPrefix, s) ->
      begin
	match p with
	| NSDefaultElementPrefix ->
	    (NSDefaultElementPrefix, s)
	| _ ->
	    (p, s)
      end
  | _ ->
      uqname

let map_type prefix tstr =
  if (isbasety tstr) then
    fmkmodel (TBase (tobasety tstr)) Finfo.bogus
  else
    begin
      let ugroup_name = map_name prefix (qname_of_string tstr) in
      fmkmodel (TGroupRef (UGroupSym (ugroup_symbol(ugroup_name)))) Finfo.bogus
    end

let map_bound model (min, max) =
  try
    begin
      match (min, max) with
      |	(None, None) -> model
      | (Some "0", None) -> fmkmodel (TBound (model, Occurrence.occurs 0, Occurrence.occurs 1)) Finfo.bogus
      | (Some "1", None) -> model
      | (Some _, None) -> raise (Query (Internal_Error "Invalid schema: Invalid bounds"))
      | (None, Some s) -> 
	  begin
	    match (String.lowercase s) with
	    | "unbounded" ->
		fmkmodel (TBound (model, Occurrence.occurs 1, Occurrence.unbounded)) Finfo.bogus
	    | _ ->
                (* Loosing precision here - Jerome *)
		let upper_bound = (int_of_string s) in
		if upper_bound = 1
		then
		  fmkmodel (TBound (model, Occurrence.occurs 1, Occurrence.occurs 1)) Finfo.bogus
		else
		  fmkmodel (TBound (model, Occurrence.occurs 1, Occurrence.unbounded)) Finfo.bogus
	  end
      | (Some m, Some n) -> 
	  begin
	    match ((String.lowercase m), (String.lowercase n)) with
	    | ("unbounded", "unbounded") ->
		fmkmodel (TBound (model, Occurrence.unbounded, Occurrence.unbounded)) Finfo.bogus
	    | ("unbounded", _) ->
		raise (Query (Internal_Error "Invalid schema: Invalid bounds"))
            (* Loosing precision here - Jerome *)
	    | (strm, "unbounded") ->
		let lower_bound = (int_of_string strm) in
		if lower_bound = 0
		then
		  fmkmodel (TBound (model, Occurrence.occurs 0, Occurrence.unbounded)) Finfo.bogus
		else
		  fmkmodel (TBound (model, Occurrence.occurs 1, Occurrence.unbounded)) Finfo.bogus
	    | (strm, strn) ->
		let lower_bound = (int_of_string strm)
		and upper_bound = (int_of_string strn) in
		begin
                  (* Loosing precision here - Jerome *)
		  match (lower_bound,upper_bound) with
		  | (0,0) ->
		      fmkmodel TEmpty Finfo.bogus
		  | (0,1) ->
		      fmkmodel (TBound (model, Occurrence.occurs 0, Occurrence.occurs 1)) Finfo.bogus
		  | (0,_) ->
		      fmkmodel (TBound (model, Occurrence.occurs 0, Occurrence.unbounded)) Finfo.bogus
		  | (1,1) ->
		      model
		  | _ ->
		      fmkmodel (TBound (model, Occurrence.occurs 1, Occurrence.unbounded)) Finfo.bogus
		end
	  end
    end
  with
  | Failure s -> raise (Query (Internal_Error ("Invalid schema: Invalid bounds: " ^ s)))

let map_use model use_opt =
  match use_opt with
  | None -> fmkmodel (TBound (model, Occurrence.occurs 0, Occurrence.occurs 1)) Finfo.bogus
  | Some "optional" -> fmkmodel (TBound (model, Occurrence.occurs 0, Occurrence.occurs 1)) Finfo.bogus
  | Some "required" -> model
  | Some "prohibited" -> raise (Query (Prototype "'use' attribute value 'prohibited' is not supported"))
  | _ -> raise (Query (Internal_Error "Invalid schema: Invalid value of 'use' attribute"))

(**************************************)
(* helper functions for mapping group *)
(**************************************)

let rec map_group_of_complexType prefix elemlist attrlist =
  let complexTypeContent' = List.map (map_group prefix) elemlist in
(* Not clear what to do with attributes here
  let attributes' = List.map (map_group prefix) attrlist in
  seq_of_list (complexTypeContent' @ attributes')
   - Jerome *)
  seq_of_list complexTypeContent'

and map_group_of_group prefix elemlist attrlist =
  let content' = List.map (map_group prefix) elemlist in
  seq_of_list content'

and map_group_of_choice prefix elemlist attrlist =
  let content' = List.map (map_group prefix) elemlist in
  let bound = (probeAttr attrlist "minoccurs", probeAttr attrlist "maxoccurs") in
  map_bound (choice_of_list content') bound

and map_group_of_sequence prefix elemlist attrlist =
  let content' = List.map (map_group prefix) elemlist in
  let bound = (probeAttr attrlist "minoccurs", probeAttr attrlist "maxoccurs") in
  map_bound (seq_of_list content') bound

and map_group_of_element prefix elemlist attrlist =
  let pref = probeAttr attrlist "ref" and
      pname = probeAttr attrlist "name" and
      ptype = probeAttr attrlist "type" and
      bound = (probeAttr attrlist "minoccurs", probeAttr attrlist "maxoccurs") in
  match (pref, pname, ptype) with
  | (Some r, None, None) ->
      let uelem_qname = map_name prefix (qname_of_string r) in
      let sy = uelem_symbol uelem_qname in
      map_bound (fmkmodel (TElemRef (UElemSym sy)) Finfo.bogus) bound
  | (None, Some n, Some t) ->
      let uelem_qname = map_name prefix (qname_of_string n) in
      let rncn = uelem_symbol uelem_qname and
	  model = map_type prefix t
      in
      map_bound (fmkmodel (TElemLocal (UElemSym rncn, None, fmktyped_model (TAnonymousType model) Finfo.bogus)) Finfo.bogus) bound
  | (None, Some n, None) ->
      let uelem_qname = map_name prefix (qname_of_string n) in
      let rncn = uelem_symbol uelem_qname and
	  content' = List.map (map_group prefix) elemlist in
      map_bound (fmkmodel (TElemLocal (UElemSym rncn, None, fmktyped_model (TAnonymousType (seq_of_list content')) Finfo.bogus)) Finfo.bogus) bound
  |	_ -> raise (Query (Internal_Error "Invalid schema: Invalid attributes of 'element' element"))

and map_group_of_attribute prefix elemlist attrlist =
  let pref = probeAttr attrlist "ref" and
      pname = probeAttr attrlist "name" and
      ptype = probeAttr attrlist "type" and 
      puse = probeAttr attrlist "use" in
  match (pref, pname, ptype, puse) with
  | (Some r, None, None, use) ->
      let uattr_qname = map_name prefix (qname_of_string r) in
      let rsy = uattr_symbol uattr_qname in
      map_use (fmkmodel (TAttrRef (UAttrSym rsy)) Finfo.bogus) use
  | (None, Some n, Some t, use) ->
      let uattr_qname = map_name prefix (qname_of_string n) in
      let rncn =  uattr_symbol uattr_qname and
	  model = map_type prefix t
      in
      map_use (fmkmodel (TAttrLocal (UAttrSym rncn,fmktyped_model (TAnonymousType model) Finfo.bogus)) Finfo.bogus) use
  | (None, Some n, None, use) ->
      let uattr_qname = map_name prefix (qname_of_string n) in
      let rncn = uattr_symbol uattr_qname and
		      content' = List.map (map_group prefix) elemlist in
      map_use (fmkmodel (TAttrLocal (UAttrSym rncn, fmktyped_model (TAnonymousType (seq_of_list content')) Finfo.bogus)) Finfo.bogus) use
  | _ -> raise (Query (Internal_Error "Invalid schema: Invalid attributes of 'element' element"))

and map_group_of_restriction prefix elemlist attrlist =
  let ptname = (probeAttr attrlist "base") in
  match ptname with
  | None -> raise (Query (Internal_Error "Invalid schema: 'restriction' element without a 'base' attribute"))
  | Some base -> 
      map_type prefix base

and map_group_of_complexContext  prefix elemlist attrlist =
  let content' = List.map (map_group prefix) elemlist in
  seq_of_list content'

and map_group_of_simpleType prefix elemlist attrlist =
  let content' = List.map (map_group prefix) elemlist in
  seq_of_list content'

and map_group_of_list prefix elemlist attrlist =
  let ptname = (probeAttr attrlist "itemtype") in
  match ptname with
  | None -> raise (Query (Internal_Error "Invalid schema, 'list' element without a 'itemType' attribute"))
  | Some itemType -> 
      fmkmodel (TBound ((map_type prefix itemType), Occurrence.occurs 0, Occurrence.unbounded)) Finfo.bogus

and map_group_of_union prefix elemlist attrlist =
  let ptnamelist = (probeAttr attrlist "membertypes") in
  match ptnamelist with
  | None -> raise (Query (Internal_Error "Invalid schema, 'union' element without a 'memberTypes' attribute"))
  | Some memberTypes ->
      let tnamelist = string_list_of_string ' ' memberTypes [] in
      let qnamelist = List.map (map_type prefix) tnamelist in
      let content' = List.map (map_group prefix) elemlist in
      choice_of_list (qnamelist @ content')

(**********************)
(* map_group function *)
(**********************)
and map_group prefix e =
  match e with
  | XElem (UElemSym n,attrlist, elemlist) ->
      let elemlist = filter_nonelem elemlist in
      let s = String.lowercase (getLocalName e) in
      begin
	match s with
	| "complextype" ->
	    map_group_of_complexType prefix elemlist attrlist
	| "group" ->
	    map_group_of_group prefix elemlist attrlist
	| "choice" ->
	    map_group_of_choice prefix elemlist attrlist
	| "sequence" ->
	    map_group_of_sequence prefix elemlist attrlist
	| "element" ->
	    map_group_of_element prefix elemlist attrlist
	| "attribute" ->
	    map_group_of_attribute prefix elemlist attrlist
	| "complexContent" ->
	    map_group_of_complexContext  prefix elemlist attrlist
	| "restriction" ->
	    map_group_of_restriction prefix elemlist attrlist
	| "extension" ->
	    raise (Query (Prototype "extension attribute value 'prohibited' is not supported"))
	| "simpletype" ->
	    map_group_of_simpleType prefix elemlist attrlist
	| "list" ->
	    map_group_of_list prefix elemlist attrlist
	| "union" ->
	    map_group_of_union prefix elemlist attrlist

	| str -> raise (Query (Internal_Error ("Invalid schema: " ^ str ^ " element is found")))
      end
  | str -> raise (Query (Internal_Error "Internal error in map_group"))

(******************************************)
(* helper functions for mapping sort decl *)
(******************************************)
let rec map_sort_decl_of_complexType prefix elemlist attrlist =
  let pname = (probeAttr attrlist "name") in
  let pmix = (probeAttr attrlist "mixed") in
  (* ignore mixed for now *)
  match (pname, pmix) with
  | (None, _) -> raise (Query (Internal_Error "Not a valid schema, toplevel complexType without name attribute."))
  | (Some str, _) ->
      let ugroup_qname = map_name prefix (qname_of_string str) in
      [ fmksort_decl (TypeDecl (UGroupSym (ugroup_symbol(ugroup_qname)), seq_of_list (List.map (map_group prefix) elemlist))) Finfo.bogus ]

and map_sort_decl_of_group prefix elemlist attrlist =
  let pname = (probeAttr attrlist "name") in
  match pname with
  | None -> raise (Query (Internal_Error "Not a valid schema, toplevel group without a name attribute"))
  | Some str ->
      let ugroup_qname = map_name prefix (qname_of_string str) in
      [ fmksort_decl (TypeDecl (UGroupSym (ugroup_symbol(ugroup_qname)), seq_of_list (List.map (map_group prefix) elemlist))) Finfo.bogus ]

and map_sort_decl_of_simpleType prefix elemlist attrlist =
  match (probeAttr attrlist "name") with
  | None -> raise (Query (Internal_Error "Not a valid schem, toplevel simpleType without name attribute."))
  | Some str ->
      let ugroup_qname = map_name prefix (qname_of_string str) in
      [ fmksort_decl (TypeDecl (UGroupSym (ugroup_symbol(ugroup_qname)), seq_of_list (List.map (map_group prefix) elemlist))) Finfo.bogus ]

and map_sort_decl_of_element prefix elemlist attrlist =
  let pname = (probeAttr attrlist "name") in
  let ptype = (probeAttr attrlist "type") in
  match (pname, ptype) with
  | (Some n, Some t) ->
      let uelem_qname = map_name prefix (qname_of_string n) in
      [ fmksort_decl (ElemDecl (UElemSym (uelem_symbol(uelem_qname)), fmktyped_model (TAnonymousType (map_type prefix t)) Finfo.bogus)) Finfo.bogus ]
  | (Some n, None) ->
      let uelem_qname = map_name prefix (qname_of_string n) in
      [ fmksort_decl (ElemDecl (UElemSym (uelem_symbol(uelem_qname)), fmktyped_model (TAnonymousType (seq_of_list (List.map (map_group prefix) elemlist))) Finfo.bogus)) Finfo.bogus ]
  | _ -> raise (Query (Internal_Error "Not a valid schema: Element declaration"))

and map_sort_decl_of_attribute prefix elemlist attrlist =
  let pname = (probeAttr attrlist "name") and
      pref = (probeAttr attrlist "ref") and
      ptype = (probeAttr attrlist "type") and
      puse = (probeAttr attrlist "use") in
  
  match (pname, ptype, puse, pref) with
  |(None, None, None, _) -> 
      raise (Query (Prototype "Not support globle attribute referrence"))
  | (Some n, Some t, None, None) ->
      let uattr_qname = map_name prefix (qname_of_string n) in
      [ fmksort_decl (AttrDecl (UAttrSym (uattr_symbol(uattr_qname)), fmktyped_model (TAnonymousType (map_type prefix t)) Finfo.bogus)) Finfo.bogus ]
  | (Some n, None, None, None) ->
      let uattr_qname = map_name prefix (qname_of_string n) in
      [ fmksort_decl (AttrDecl (UAttrSym (uattr_symbol(uattr_qname)), fmktyped_model (TAnonymousType (seq_of_list (List.map (map_group prefix) elemlist))) Finfo.bogus)) Finfo.bogus ]
  | _ -> raise (Query (Internal_Error "Not a valid schema: Type"))

and map_sort_decl_of_attributeGroup prefix elemlist attrlist =
  let pname = probeAttr attrlist "name" in
  match pname with
  | Some n ->
      let uattr_grp_qname = map_name prefix (qname_of_string n) in
      [ fmksort_decl (TypeDecl (UGroupSym (ugroup_symbol(uattr_grp_qname)), seq_of_list (List.map (map_group prefix) elemlist))) Finfo.bogus ]
  | _ ->
      raise (Query (Internal_Error "Not a valid schema: AttributGroup"))

(**************************)
(* map_sort_decl function *)
(**************************)
(* map the XML Schema to sort_decl, *)
(* side effect: add the conflicting (global vs local) symbols to the conflict tables *)
and map_sort_decl prefix e =
  match e with
  | XElem (n, attrlist, elemlist) ->
      let elemlist = filter_nonelem elemlist in
      let s = String.lowercase (getLocalName e) in
      begin
	match s with
	| "complextype" ->
	    map_sort_decl_of_complexType prefix elemlist attrlist
	| "group" ->
	    map_sort_decl_of_group prefix elemlist attrlist
	| "simpletype" ->
	    map_sort_decl_of_simpleType prefix elemlist attrlist
	| "element" ->
	    map_sort_decl_of_element prefix elemlist attrlist
	| "attribute" -> 
	    map_sort_decl_of_attribute prefix elemlist attrlist
	| "attributegroup" ->
	    map_sort_decl_of_attributeGroup prefix elemlist attrlist
	| "anyattribute" -> Error.eprintf_warning "Anyattribute is ignored." ; []
	| "annotation" -> Error.eprintf_warning "Annotation is ignored."; []
	| _ -> raise (Query (Prototype ("Element name, " ^ s ^", is not yet supported")))
      end
  | _ -> raise (Query (Internal_Error "Mapping of non element expression is not supported yet"))


(*******************************)
(* Mapping Schema to SortDecls *)
(*******************************)

(* val map_sort_decls : expr list -> sort_decl list *)

let map_schema prefix e =
  match e with
  | XElem (n, attrlist, elemlist) ->
      let elemlist = filter_nonelem elemlist in
      let s = String.lowercase (getLocalName e) in
      begin
	match s with
        (**********)
        (* schema *)
        (**********)
	| "schema" ->
	    let targetNS = (probeAttr attrlist "targetNamespace") in
	    begin
	      match targetNS with
	      | None ->
		  raise (Query (Internal_Error "Target namespace is missing."))
	      | Some tns ->
		  (tns, List.flatten (List.map (map_sort_decl prefix) elemlist))
	    end
       	| _ ->
	    raise (Query (Internal_Error "Not a valid schema: schema element is not found at the toplevel."))
      end
  | _ ->
      raise (Query (Internal_Error "Not a valid schema: Mapping non element expression"))

(*****************************)
(* the toplevel map function *)
(*****************************)

let map_xml_schema p xml_schema =
  let root_schema_element = xml_schema.xml_root in
  let (uri, usdl) = map_schema p root_schema_element in
  (p, NSUri uri, usdl)

