(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: schemautil.ml,v 1.30 2003/06/12 21:38:07 mff Exp $ *)

(**********************************************************************)
(* This module implements a number of general-purpose type operations *)
(**********************************************************************)

open Error
open Occurrence
open Namespace
open Datatypes
open Xquery_ast
open Xquery_astutil
open Schema
open Print
open Format

(***********************************************)
(* Some constructors/destructors on model ASTs *)
(***********************************************)

(* Test some type structures *)

let is_string_type m =
  match m.pmodel_desc with
  | TBase ATString -> true
  | _ -> false

let is_bool_type m =
  match m.pmodel_desc with
  | TBase ATBoolean -> true
  | _ -> false

let is_empty_choice_type m =
  match m.pmodel_desc with
  | TNone -> true
  | _ -> false

let is_empty_seq_type m =
  match m.pmodel_desc with
  | TEmpty -> true
  | _ -> false

let is_none m =
  match m.pmodel_desc with
  | TNone -> false
  | _ -> true

(* Checks if a model is an element *)

let is_element s tn =
  let t = dereference_rgroup s tn in
  match t.pmodel_desc with
  | TElemLocal _ -> true
  | TElemRef _ -> true
  | _ -> false

(* Checks if a model is an attribute *)

let is_attribute s tn =
  let t = dereference_rgroup s tn in
  match t.pmodel_desc with
  | TAttrLocal _ -> true
  | TAttrRef _ -> true
  | _ -> false

let is_attribute_unit s m =
  match m.pmodel_desc with
  | TAttrLocal _ -> true
  | TAttrRef _ -> true
  | _ -> false

(* Checks if a model is a unit type *)

let is_unit_type s tn =
  let t = dereference_rgroup s tn in
  match t.pmodel_desc with
  | TElemLocal _ -> true
  | TElemRef _ -> true
  | TAttrLocal _ -> true
  | TAttrRef _ -> true
  | TBase _ -> true
  | TText -> true
  | TDocument _ -> true
  | _ -> false  

(* Checks if a model is an XML Schema simple type *)

let is_simple s m =
  match m.pmodel_desc with
  | TBase _ -> true
  | _ -> false

(* Checks if a model is an XML Schema node type *)

let is_node s m =
  match m.pmodel_desc with
  | TElemLocal _ -> true
  | TElemRef _ -> true
  | TAttrLocal _ -> true
  | TAttrRef _ -> true
  | TDocument _ -> true
  | TText -> true
  | _ -> false  

let is_complex_node s m =
  match m.pmodel_desc with
  | TElemLocal _ -> true
  | TElemRef _ -> true
  | TAttrLocal _ -> true
  | TAttrRef _ -> true
  | _ -> false  

let check_unit_type s x =
  if not(is_unit_type s x)
  then
    raise (Query (Internal_Error ("Group "
				  ^ (Sym.rgroup_string x)
				  ^ " not a unit type")))
  else
    ()

(* Returns the element name *)

let get_element_tag s tn =
  let t = dereference_rgroup s tn in
  match t.pmodel_desc with
  | TElemLocal(l,_,_) -> l
  | TElemRef l -> l
  | _ ->
      raise (Query (Not_An_Element))

(* Returns the attribute name *)

let get_attribute_tag s tn =
  let t = dereference_rgroup s tn in
  match t.pmodel_desc with
  | TAttrLocal(l,_) -> l
  | TAttrRef l -> l
  | _ ->
      raise (Query Not_An_Attribute)

(* Returns the name of a unit type *)

let rec get_unit_tag_from_unresolved_model nse m =
  match m.pmodel_desc with
  | TElemLocal (l,_,_) -> Sym.RUnitElemSymbol (Resolve.resolve_elem_symbol nse l)
  | TElemRef l -> Sym.RUnitElemSymbol (Resolve.resolve_elem_symbol nse l)
  | TAttrLocal (l,_) -> Sym.RUnitAttrSymbol (Resolve.resolve_attr_symbol nse l)
  | TAttrRef l -> Sym.RUnitAttrSymbol (Resolve.resolve_attr_symbol nse l)
  | TBase bt -> Sym.RUnitSimpleSymbol (Datatypes.unit_symbol_of_base_type bt)
  | TText -> Sym.RUnitTextSymbol
  | TDocument m -> Sym.RUnitDocSymbol
  | _ -> raise (Query Not_A_Unit_Type)

let rec get_unit_tag_from_model m =
  match m.pmodel_desc with
  | TElemLocal (Sym.RElemSym l,_,_) -> Sym.RUnitElemSymbol l
  | TElemRef (Sym.RElemSym l) -> Sym.RUnitElemSymbol l
  | TAttrLocal (Sym.RAttrSym l,_) -> Sym.RUnitAttrSymbol l
  | TAttrRef (Sym.RAttrSym l) -> Sym.RUnitAttrSymbol l
  | TBase bt -> Sym.RUnitSimpleSymbol (Datatypes.unit_symbol_of_base_type bt)
  | TText -> Sym.RUnitTextSymbol 
  | TDocument m -> Sym.RUnitDocSymbol
  | TTuple m -> Sym.RUnitTupleSymbol
  | _ -> raise (Query Not_A_Unit_Type)

let get_unit_tag s tn =
  let m = dereference_rgroup s tn in
  get_unit_tag_from_model m

let get_unit_tag_unresolved s nse tn =
  let m = dereference_rgroup s tn in
  get_unit_tag_from_unresolved_model nse m

(* Accesses element and attribute models *)

let access_attribute s am =
  match (dereference_elem_attr s am).pmodel_desc with
  | TAttrLocal (Sym.RAttrSym sym,attr_model) ->
      (sym,attr_model)
  | _ ->
      raise (Query (Internal_Error "Cannot access attribute model"))

let access_element s em =
  match (dereference_elem_attr s em).pmodel_desc with
  | TElemLocal (Sym.RElemSym sym,_,elem_model) ->
      (sym,elem_model)
  | _ ->
      raise (Query (Internal_Error "Cannot access element model"))

(* Union normalization *)

let rec union_types ms =
  let ms' = List.filter is_none ms in
  match ms' with
  | [] ->
      fmkmodel TNone Finfo.bogus
  | m :: ms'' ->
      fmkmodel (TBar (m,union_types ms'')) Finfo.bogus

let rec weakunion_types ms =
  let ms' = List.filter is_none ms in
  match ms' with
  | [m] -> m
  | m :: ms'' ->
      fmkmodel (TBar (m,weakunion_types ms'')) Finfo.bogus
  | [] ->
      fmkmodel TNone Finfo.bogus

(* Applies a function to each unit of a given type *)

let rec apply_to_unit_types tenv f t =
  match t.pmodel_desc with
  | TGroupRef v' ->
      let t' = dereference_group tenv v' in
      apply_to_unit_types tenv f t'
  | TBase b ->
      f t
  | TAttrRef _ ->
      let t' = dereference_elem_attr tenv t in
      f t'
  | TAttrLocal _ ->
      f t
  | TElemRef _ ->
      let t' = dereference_elem_attr tenv t in
      f t'
  | TElemLocal _ ->
      f t
  | TComma (t1, t2) ->
      let t1' = apply_to_unit_types tenv f t1
      and t2' = apply_to_unit_types tenv f t2 in
      fmkmodel (TComma(t1',t2')) Finfo.bogus
  | TEmpty ->
      fmkmodel (TEmpty) Finfo.bogus
  | TBar (t1, t2) ->
      let t1' = apply_to_unit_types tenv f t1
      and t2' = apply_to_unit_types tenv f t2 in
      fmkmodel (TBar(t1', t2')) Finfo.bogus
  | TAmp _ ->
      raise (Query (Prototype "Ampersand not supported yet"))
  | TNone ->
      fmkmodel (TNone) Finfo.bogus
  | TBound (t1, m, n) ->
      let t1' = apply_to_unit_types tenv f t1 in
      fmkmodel (TBound(t1', m, n)) Finfo.bogus
  | TDocument _ ->
      f t
  | TText ->
      f t
  | TTuple _ ->
      f t

(* Universal quantification over unit types ! *)

let rec forall_unit_types tenv f t =
  match t.pmodel_desc with
  | TGroupRef v' ->
      let t' = dereference_group tenv v' in
      forall_unit_types tenv f t'
  | TBase b ->
      f t
  | TAttrRef _ ->
      let t' = dereference_elem_attr tenv t in
      f t'
  | TAttrLocal _ ->
      f t
  | TElemRef _ ->
      let t' = dereference_elem_attr tenv t in
      f t'
  | TElemLocal _ ->
      f t
  | TComma (t1, t2) ->
      let b1 = forall_unit_types tenv f t1
      and b2 = forall_unit_types tenv f t2 in
      (b1 && b2)
  | TEmpty ->
      true
  | TBar (t1, t2) ->
      let b1 = forall_unit_types tenv f t1
      and b2 = forall_unit_types tenv f t2 in
      (b1 && b2)
  | TAmp (t1,t2) ->
      let b1 = forall_unit_types tenv f t1
      and b2 = forall_unit_types tenv f t2 in
      (b1 && b2)
  | TNone ->
      true
  | TBound (t1, m, n) ->
      forall_unit_types tenv f t1
  | TDocument _ ->
      f t
  | TText  ->
      f t
  | TTuple _ ->
      f t

(* Checks for emptiness of a type *)

let rec empty_type tenv t =
  match t.pmodel_desc with
  | TGroupRef v' ->
      let t' = dereference_group tenv v' in
      empty_type tenv t'
  | TBase b ->
      false
  | TAttrRef _ ->
      false
  | TAttrLocal _ ->
      false
  | TElemRef _ ->
      false
  | TElemLocal _ ->
      false
  | TComma (t1, t2) ->
      (empty_type tenv t1) && (empty_type tenv t2)
  | TEmpty ->
      true
  | TBar (t1, t2) ->
      (empty_type tenv t1) && (empty_type tenv t2)
  | TAmp (t1,t2) ->
      (empty_type tenv t1) && (empty_type tenv t2)
  | TNone ->
      false
  | TBound (t1, m, n) ->
      empty_type tenv t1
  | TDocument _ ->
      false
  | TText ->
      false
  | TTuple _ ->
      false

(* Checks for noneness of a type *)

let rec none_type tenv t =
  match t.pmodel_desc with
  | TGroupRef v' ->
      let t' = dereference_group tenv v' in
      none_type tenv t'
  | TBase b ->
      false
  | TAttrRef _ ->
      false
  | TAttrLocal _ ->
      false
  | TElemRef _ ->
      false
  | TElemLocal _ ->
      false
  | TComma (t1, t2) ->
      (none_type tenv t1)||(none_type tenv t2)
  | TEmpty ->
      false
  | TBar (t1, t2) ->
      (none_type tenv t1)&&(none_type tenv t2)
  | TAmp (t1,t2) ->
      (none_type tenv t1)||(none_type tenv t2)
  | TNone ->
      true
  | TBound (t1, UNBOUNDED, n) ->
      true
  | TBound (t1, UP_INT m, n) ->
      (none_type tenv t1) && m > 0
  | TDocument _ ->
      false
  | TText ->
      false
  | TTuple _ ->
      false

(********************************)
(* Operations on specific types *)
(********************************)

(* Pair type *)

let pair_type m1 m2 =
  let tm1' = fmktyped_model (TAnonymousType m1) Finfo.bogus
  and tm2' = fmktyped_model (TAnonymousType m2) Finfo.bogus
  in
  let m1' = fmkmodel (TElemLocal (Sym.RElemSym (Sym.relem_symbol (NSDefaultElementPrefix, (NSUri "","fst"))),
				  None,
				  tm1')) Finfo.bogus
  and m2' = fmkmodel (TElemLocal (Sym.RElemSym (Sym.relem_symbol (NSDefaultElementPrefix, (NSUri "","snd"))),
				  None,
				  tm2')) Finfo.bogus
  in
  fmkmodel (TElemLocal (Sym.RElemSym (Sym.relem_symbol (NSDefaultElementPrefix, (NSUri "","pair"))),
			None,
			fmktyped_model
			  (TAnonymousType (fmkmodel (TComma (m1',m2')) Finfo.bogus))
			  Finfo.bogus)) Finfo.bogus

let unpair_type m =
  match m.pmodel_desc with
  | TElemLocal(_,_,tm') ->
      begin
	match tm'.ptyped_model_desc with
	| TAnonymousType m' ->
	    begin
	      match m'.pmodel_desc with
	      | TComma(m1,m2) ->
		  begin
		    match (m1.pmodel_desc,m2.pmodel_desc) with
		    | (TElemLocal(_,_,tm1'),TElemLocal(_,_,tm2')) ->
			begin
			  match (tm1'.ptyped_model_desc, tm2'.ptyped_model_desc) with
			  | (TAnonymousType m1', TAnonymousType m2') ->
			      (m1',m2')
			end
		    | _ ->
			raise (Query (Type_Error (m.pmodel_loc,
						  "unpair_type: not a pair type")))
		  end
	      | _ ->
		  raise (Query (Type_Error (m.pmodel_loc,
					    "unpair_type: not a pair type")))
	    end
      end
  | _ ->
      raise (Query (Type_Error (m.pmodel_loc,
				"unpair_type: not a pair type")))

(*************)
(* Factoring *)
(*************)

let rec factor_aux sch m =
  match m.pmodel_desc with
  | TAttrRef s ->
      factor_aux sch (dereference_elem_attr sch m)
  | TAttrLocal(l,_) ->
      (m :: [],occurs 1,occurs 1)
  | TElemLocal(l,_,_) ->
      (m :: [],occurs 1,occurs 1)
  | TElemRef s ->
      factor_aux sch (dereference_elem_attr sch m)
  | TGroupRef tn ->
      let m' = dereference_group sch tn in
      factor_aux sch m'
  | TBound (m',i,j) ->
      let (m'',i',j') = factor_aux sch m' in
      (m'', ub_mult i' i, ub_mult j' j)
  | TComma (m1, m2) ->
      let (m1', i1, j1) = factor_aux sch m1 in
      let (m2', i2, j2) = factor_aux sch m2 in
      ((m1' @ m2'), ub_add i1 i2, ub_add j1 j2)
  | TEmpty ->
      ((fmkmodel TNone Finfo.bogus) :: [],occurs 0,occurs 0)
  | TBar (m1, m2) ->
      let (m1', i1, j1) = factor_aux sch m1 in
      let (m2', i2, j2) = factor_aux sch m2 in
      ((m1' @ m2'), ub_min i1 i2, ub_max j1 j2)
  | TNone ->
      ((fmkmodel TNone Finfo.bogus) :: [],UNBOUNDED, occurs 0)
  | TAmp _ ->
      raise (Query (Type_Error (m.pmodel_loc,
				"factor: ampersand '&' not supported yet")))
  | TBase _ ->
      (m :: [], occurs 1, occurs 1)
  | TDocument _ ->
      (m :: [], occurs 1, occurs 1)
  | TText ->
      (m :: [], occurs 1, occurs 1)
  | TTuple _ ->
      (m :: [], occurs 1, occurs 1)

(* Recursive factoring *)

let rec recfactor_aux sds m previous_groups previous_elements = 
  match m.pmodel_desc with
  | TElemLocal(nc,_,tm) ->
      let m' = 
	begin
	  match tm.ptyped_model_desc with
	  | TAnonymousType m' -> m'
	end
      in
      let (m1',i1',j1')  = recfactor_aux sds m' previous_groups previous_elements in 
      ((m :: m1'), ub_add (occurs 1) i1', ub_add (occurs 1) j1')

  | TElemRef s ->
      let m' = dereference_elem_attr sds m in
      if (List.exists ((=) s) previous_elements)
      then
	([m], occurs 0, UNBOUNDED)
      else
	let (m1',i1',j1')  = recfactor_aux sds m' previous_groups (s :: previous_elements) in 
	((m :: m1'), ub_add (occurs 1) i1', ub_add (occurs 1) j1')

	(* Attribute models should not be recursives *)
  | TAttrRef s ->
	([fmkmodel TNone Finfo.bogus], occurs 1, occurs 1)
  | TAttrLocal(nc,_) ->
	([fmkmodel TNone Finfo.bogus], occurs 1, occurs 1)

  | TGroupRef tn ->
      let m' = dereference_group sds tn in
      if (List.exists ((=) tn) previous_groups) 
      then 
	let (m'', _, _) = factor_aux sds m' in 
	(m'', occurs 0, UNBOUNDED)

        (* this is because of recursive type *)
	(* this should probably be checked somewhere before *)
      else
	begin
	  try 
            recfactor_aux sds m' (tn :: previous_groups) previous_elements
	  with
	  | Not_found ->
	      raise (Query (Type_Error (m.pmodel_loc, ("Group "
						       ^ (Sym.string_of_group_symbol tn)
						       ^ " not in scope"))))
	end
  | TBound (m',i,j) ->
      let (m'',i',j') = recfactor_aux sds m' previous_groups previous_elements in
      (m'', ub_mult i' i, ub_mult j' j)
  | TComma (m1, m2) ->
      let (m1', i1, j1) = recfactor_aux sds m1 previous_groups previous_elements in
      let (m2', i2, j2) = recfactor_aux sds m2 previous_groups previous_elements in
      ((m1' @ m2'), ub_add i1 i2, ub_add j1 j2)
  | TEmpty ->
      ([fmkmodel TNone Finfo.bogus],occurs 0,occurs 0)
  | TBar (m1, m2) ->
      let (m1', i1, j1) = recfactor_aux sds m1 previous_groups previous_elements in
      let (m2', i2, j2) = recfactor_aux sds m2 previous_groups previous_elements in
      ((m1' @ m2'), ub_min i1 i2, ub_max j1 j2)
  | TNone ->
      ([fmkmodel TNone Finfo.bogus], UNBOUNDED, occurs 0)
  | TAmp _ ->
      raise (Query (Type_Error (m.pmodel_loc,
				"recfactor: ampersand '&' not supported yet")))
  | TBase _ ->
      ([fmkmodel TNone Finfo.bogus],occurs 0,occurs 0)
  | TDocument m1 ->
      let (m',i',j')  = recfactor_aux sds m1 previous_groups previous_elements in 
      ((m :: m'), ub_add (occurs 1) i', ub_add (occurs 1) j')
  | TText ->
      ([m],occurs 1, occurs 1)
  | TTuple _ ->
      ([m],occurs 1, occurs 1)

(************************************)
(* operations on regular expresions *)
(************************************)

(* Accesses the content of a model as a regular expression over type names *)

let rec model_normalize s m =
  match m.pmodel_desc with
  | TGroupRef (Sym.RGroupSym tn) ->
      if (is_unit_type s tn)
      then m
      else
	model_normalize s (dereference_rgroup s tn)
  | TBound (m',b1,b2) ->
      fmkmodel (TBound ((model_normalize s m'),b1,b2)) Finfo.bogus
  | TComma (m1,m2) ->
      fmkmodel (TComma (model_normalize s m1, model_normalize s m2)) Finfo.bogus
  | TEmpty ->
      fmkmodel TEmpty Finfo.bogus
  | TBar (m1,m2) ->
      fmkmodel (TBar (model_normalize s m1, model_normalize s m2)) Finfo.bogus
  | TNone ->
      fmkmodel TNone Finfo.bogus
  | TAmp (m1,m2) ->
      fmkmodel (TAmp (model_normalize s m1, model_normalize s m2)) Finfo.bogus
  | _ ->
      raise (Query (Internal_Error "model normalization of a non renamed model"))

(* Does the model contain empty ? *)

let rec empty s m =
  match m.pmodel_desc with
  | TGroupRef (Sym.RGroupSym x) ->
      if (is_unit_type s x)
      then false
      else
	let m' = get_group s x
	in empty s m'
  | TGroupRef (Sym.UGroupSym _) ->
      raise (Query (Internal_Error ("Dereferencing an unresolved group")))
  | TBase _ ->
      false
  | TText ->
      false
  | TEmpty ->
      true
  | TElemLocal _ ->
      false
  | TAttrLocal _ ->
      false
  | TComma (m1, m2) ->
      if (empty s m1)
      then (empty s m2)
      else false
  | TBar (m1, m2) ->
      ((empty s m1)||(empty s m2))
  | TNone ->
      false
  | TBound (t',m,n) ->
      if (m = (UP_INT 0))
      then true
      else (empty s t')
  | TAttrRef _ ->
      false
  | TElemRef _ ->
      false
  | TAmp _ ->
      raise (Query (Prototype ("TAmp is not supported")))
  | TDocument _ ->
      false
  | TTuple _ ->
      false

(* Does the model contain empty ? *)

(* Note:
     This function treats attributes as empty!
     This is used for validation when content models have a mix of
     element and attribute declarations.
   - Jerome
 *)

let rec empty_or_attribute s m =
  match m.pmodel_desc with
  | TGroupRef (Sym.RGroupSym x) ->
      if (is_unit_type s x) && not(is_attribute s x)
      then false
      else
	let m' = get_group s x
	in empty_or_attribute s m'
  | TGroupRef (Sym.UGroupSym _) ->
      raise (Query (Internal_Error ("Dereferencing an unresolved group")))
  | TBase _ ->
      false
  | TEmpty ->
      true
  | TElemLocal _ ->
      false
  | TAttrLocal _ ->
      true
  | TComma (m1, m2) ->
      if (empty_or_attribute s m1)
      then (empty_or_attribute s m2)
      else false
  | TBar (m1, m2) ->
      ((empty_or_attribute s m1)||(empty_or_attribute s m2))
  | TNone ->
      false
  | TBound (t',m,n) ->
      if (m = (UP_INT 0))
      then true
      else (empty_or_attribute s t')
  | TAttrRef _ ->
      true
  | TElemRef _ ->
      false
  | TAmp _ ->
      raise (Query (Prototype ("TAmp is not supported")))
  | TDocument _ ->
      false
  | TText ->
      false
  | TTuple _ ->
      false

(* Derivatives of a model *)

(* first_rest computes all the type names that can be a first token,
   plus the rest of the regular expression *)

let rec first_rest s m =
  match m.pmodel_desc with
  | TGroupRef (Sym.RGroupSym x) ->
      if (is_unit_type s x)
      then [(get_unit_tag s x, x,(fmkmodel TEmpty (Finfo.bogus)))]
      else
	let t' = (get_group s x)
	in first_rest s t'
  | TGroupRef (Sym.UGroupSym _) ->
      raise (Query (Internal_Error ("Dereferencing an unresolved group")))
  | TBase _ ->
      raise (Query (Internal_Error ("Regexp access over non-named unit type:\n"
				    ^ (bprintf_model "" m))))
  | TEmpty ->
      []
  | TElemRef _ ->
      raise (Query (Internal_Error ("Regexp access over non-named unit type:\n"
				    ^ (bprintf_model "" m))))
  | TElemLocal _ ->
      raise (Query (Internal_Error ("Regexp access over non-named unit type:\n"
				    ^ (bprintf_model "" m))))
  | TAttrRef _ ->
      raise (Query (Internal_Error ("Regexp access over non-named unit type:\n"
				    ^ (bprintf_model "" m))))
  | TAttrLocal _ ->
      raise (Query (Internal_Error ("Regexp access over non-named unit type:\n"
				    ^ (bprintf_model "" m))))
  | TComma (t1,t2) ->
      let li1 = (first_rest s t1) in
      let li1' = List.map (fun (x,y,r) -> (x,y,(fmkmodel (TComma (r, t2)) (Finfo.bogus)))) li1 in
      if not(empty s t1)
      then li1'
      else li1'@(first_rest s t2)
  | TBar (t1,t2) ->
      (first_rest s t1)@(first_rest s t2)
  | TNone ->
      []  (* Jerome: that looks really wrong to treat TNone as TEmpty here *)
  | TBound (t',m,n) ->
      let concat_types t1 t2 =
	match t1.pmodel_desc with
	| TEmpty -> t2
	| _ -> (fmkmodel (TComma (t1, t2)) Finfo.bogus) in
      let nm = minus 1 m
      and nn = minus 1 n in
      let nt =
	if nn = (UP_INT 0)
	then (fmkmodel TEmpty Finfo.bogus)
	else (fmkmodel (TBound (t',nm,nn)) Finfo.bogus) in
      let li = (first_rest s t')
      in List.map (fun (x,y,r) -> (x,y,concat_types r nt)) li
  | TAmp _ ->
      raise (Query (Prototype ("TAmp is not supported")))
  | TDocument _ ->
      raise (Query (Internal_Error ("Regexp access over non-named unit type:\n"
				    ^ (bprintf_model "" m))))
  | TText ->
      raise (Query (Internal_Error ("Regexp access over non-named unit type:\n"
				    ^ (bprintf_model "" m))))
  | TTuple _ ->
      raise (Query (Internal_Error ("Regexp access over non-named unit type:\n"
				    ^ (bprintf_model "" m))))

(* Checks for duplicates in the result of a first_rest *)

let rec check_dup xl =
  match xl with
  | [] -> false
  | [xi] -> false
  | (x,_,_) :: xl' ->
      if (List.exists (fun (x',_,_) -> (Sym.subtag x x') || (Sym.subtag x' x)) xl')
      then true
      else check_dup xl'

let first_rest_deterministic s m =
  let fr = first_rest s m in
  if (check_dup fr)                      (* Duplicates correspond to non deterministic expression *)
  then
    raise (Query (Non_Deterministic_Regexp ("Model "
					    ^ (Print.bprintf_model "" m)
					    ^ " is non deterministic")));
  fr

(* first_rest_stats computes all the tags for possible first token,
   plus the rest of the regular expression but also keeps the
   intermediate type names *)

let rec first_rest_with_model_aux s visited_names m =
  match m.pmodel_desc with
  | TGroupRef (Sym.RGroupSym x) ->
      let t' = (get_group s x)
      in first_rest_with_model_aux s (x::visited_names) t'
  | TGroupRef (Sym.UGroupSym _) ->
      raise (Query (Internal_Error ("Dereferencing an unresolved group")))
  | TBase _ ->
      [(get_unit_tag_from_model m, None, visited_names, (fmkmodel TEmpty (Finfo.bogus)))]
  | TEmpty ->
      []
  | TElemRef (Sym.RElemSym relem_sym) ->
      [(get_unit_tag_from_model m, Some (get_element s relem_sym), visited_names, (fmkmodel TEmpty (Finfo.bogus)))]
  | TElemRef (Sym.UElemSym _) ->
      raise (Query (Internal_Error ("Dereferencing an unresolved element")))
  | TElemLocal (_,_,tm) ->
      [(get_unit_tag_from_model m, Some tm, visited_names, (fmkmodel TEmpty (Finfo.bogus)))]
  | TAttrRef (Sym.RAttrSym rattr_sym) ->
      [(get_unit_tag_from_model m, Some (get_attribute s rattr_sym), visited_names, (fmkmodel TEmpty (Finfo.bogus)))]
  | TAttrRef (Sym.UAttrSym _) ->
      raise (Query (Internal_Error ("Dereferencing an unresolved attribute")))
  | TAttrLocal (_,tm) ->
      [(get_unit_tag_from_model m, Some tm, visited_names, (fmkmodel TEmpty (Finfo.bogus)))]
  | TComma (t1,t2) ->
      let li1 = (first_rest_with_model_aux s visited_names t1) in
      let li1' = List.map (fun (x,y,z,r) -> (x,y,z,(fmkmodel (TComma (r, t2)) (Finfo.bogus)))) li1 in
      if not(empty s t1)
      then li1'
      else li1' @ (first_rest_with_model_aux s visited_names t2)
  | TBar (t1,t2) ->
      (first_rest_with_model_aux s visited_names t1) @ (first_rest_with_model_aux s visited_names t2)
  | TNone ->
      []  (* Jerome: that looks really wrong to treat TNone as TEmpty here *)
  | TBound (t',m,n) ->
      let concat_types t1 t2 =
	match t1.pmodel_desc with
	| TEmpty -> t2
	| _ -> (fmkmodel (TComma (t1, t2)) Finfo.bogus) in
      let nm = minus 1 m
      and nn = minus 1 n in
      let nt =
	if nn = (UP_INT 0)
	then (fmkmodel TEmpty Finfo.bogus)
	else (fmkmodel (TBound (t',nm,nn)) Finfo.bogus) in
      let li = (first_rest_with_model_aux s visited_names t')
      in List.map (fun (x,y,z,r) -> (x,y,z,concat_types r nt)) li
  | TAmp _ ->
      raise (Query (Prototype ("TAmp is not supported")))
  | TDocument _ ->
      raise (Query (Internal_Error ("Regexp access over non-named unit type:\n"
				    ^ (bprintf_model "" m))))
  | TText ->
      [(get_unit_tag_from_model m, None, visited_names, (fmkmodel TEmpty (Finfo.bogus)))]
  | TTuple _ ->
      [(get_unit_tag_from_model m, None, visited_names, (fmkmodel TEmpty (Finfo.bogus)))]

let first_rest_with_model s m =
  first_rest_with_model_aux s [] m

(* Again the same function, but treat attributes as empty *)

let rec first_rest_with_model_attributes_aux s visited_names m =
  match m.pmodel_desc with
  | TGroupRef (Sym.RGroupSym x) ->
      let t' = (get_group s x)
      in first_rest_with_model_attributes_aux s (x::visited_names) t'
  | TGroupRef (Sym.UGroupSym _) ->
      raise (Query (Internal_Error ("Dereferencing an unresolved group")))
  | TBase _ ->
      [(get_unit_tag_from_model m, None, visited_names, (fmkmodel TEmpty (Finfo.bogus)))]
  | TEmpty ->
      []
  | TElemRef (Sym.RElemSym relem_sym) ->
      [(get_unit_tag_from_model m, Some (get_element s relem_sym), visited_names, (fmkmodel TEmpty (Finfo.bogus)))]
  | TElemRef (Sym.UElemSym _) ->
      raise (Query (Internal_Error ("Dereferencing an unresolved element")))
  | TElemLocal (_,_,tm) ->
      [(get_unit_tag_from_model m, Some tm, visited_names, (fmkmodel TEmpty (Finfo.bogus)))]
  | TAttrRef (Sym.RAttrSym rattr_sym) ->
      []
  | TAttrRef (Sym.UAttrSym _) ->
      raise (Query (Internal_Error ("Dereferencing an unresolved attribute")))
  | TAttrLocal (_,tm) ->
      []
  | TComma (t1,t2) ->
      let li1 = (first_rest_with_model_attributes_aux s visited_names t1) in
      let li1' = List.map (fun (x,y,z,r) -> (x,y,z,(fmkmodel (TComma (r, t2)) (Finfo.bogus)))) li1 in
      if not(empty_or_attribute s t1)
      then li1'
      else li1' @ (first_rest_with_model_attributes_aux s visited_names t2)
  | TBar (t1,t2) ->
      (first_rest_with_model_attributes_aux s visited_names t1) @ (first_rest_with_model_attributes_aux s visited_names t2)
  | TNone ->
      []  (* Jerome: that looks really wrong to treat TNone as TEmpty here *)
  | TBound (t',m,n) ->
      let concat_types t1 t2 =
	match t1.pmodel_desc with
	| TEmpty -> t2
	| _ -> (fmkmodel (TComma (t1, t2)) Finfo.bogus) in
      let nm = minus 1 m
      and nn = minus 1 n in
      let nt =
	if nn = (UP_INT 0)
	then (fmkmodel TEmpty Finfo.bogus)
	else (fmkmodel (TBound (t',nm,nn)) Finfo.bogus) in
      let li = (first_rest_with_model_attributes_aux s visited_names t')
      in List.map (fun (x,y,z,r) -> (x,y,z,concat_types r nt)) li
  | TAmp _ ->
      raise (Query (Prototype ("TAmp is not supported")))
  | TDocument _ ->
      raise (Query (Internal_Error ("Regexp access over non-named unit type:\n"
				    ^ (bprintf_model "" m))))
  | TText ->
      [(get_unit_tag_from_model m, None, visited_names, (fmkmodel TEmpty (Finfo.bogus)))]
  | TTuple _ ->
      [(get_unit_tag_from_model m, None, visited_names, (fmkmodel TEmpty (Finfo.bogus)))]

let first_rest_with_model_attributes s m =
  first_rest_with_model_attributes_aux s [] m

(* Now, one function that applies first_rest_with_model recursively while gather attributes *)
    
let is_attribute_fr (x,_,_,_) =
  match x with
  | Sym.RUnitAttrSymbol _ -> true
  | _ -> false

type attribute_stack =
    (Sym.resolved_unit_symbol * typed_model option * Sym.rgroup_symbol list) list

(* NOTE: Here We ignore whether attributes are required, optional or
not even well-formed - Jerome & Juliana *)

let rec first_rest_with_model_discard_attributes s m =
  let frl = first_rest_with_model s m in
  let (attributes_partition, element_partition) = List.partition is_attribute_fr frl in

  let further_process (attr_sym, attr_typed_model, attr_rgroup_list, attr_model) =
    first_rest_with_model_discard_attributes s attr_model
  in

  let result_attribute_partition =
    List.concat (List.map further_process attributes_partition)
  in
  
  result_attribute_partition @ element_partition

(****************************)
(* Content model operations *)
(****************************)

(* Is the type an attribute? *)

let attribute_model m = 
  match m.pmodel_desc with
  | TAttrRef _ -> true
  | TAttrLocal _ -> true
  | _ -> false

(* Is the type an element? *)

let element_model m = 
  match m.pmodel_desc with
  | TElemRef _ -> true
  | TElemLocal _ -> true
  | _ -> false

let rec model_contains sds modelfn m =
  match m.pmodel_desc with
  | TGroupRef tn -> 
      begin
	try
          let m' = dereference_group sds tn in
          model_contains sds modelfn m'
	with
	| Not_found ->
	    raise (Query (Type_Error (m.pmodel_loc, ("Type "
						     ^ (Sym.string_of_group_symbol tn)
						     ^ " not in scope"))))
      end
  | TBound (m',i',j') -> 
      (model_contains sds modelfn m')
  | TComma (m1, m2) ->
      (model_contains sds modelfn m1) || (model_contains sds modelfn m2)
  | TBar (m1, m2) ->
      (model_contains sds modelfn m1) || (model_contains sds modelfn m2)
  | TAmp _ ->
      raise (Query (Type_Error (m.pmodel_loc,
				"project: ampersand '&' not supported yet")))
  | _ -> (modelfn m)

(* split_content_model takes a content model of the form t1, t2
   where t1 <: xs:AnyAttribute* and
         t2 <: (xs:AnyElement | xs:AnySimpleType)*
   and returns the pair (t1, t2)
*)

let rec split_content_model sds m = 
  if !Conf.debugging then
    begin
      printf "\nsplit_content_model: ";
      printf_model "" m
    end;
  let mempty = (fmkmodel TEmpty m.pmodel_loc) in
  match m.pmodel_desc with
  | TGroupRef tn ->
      begin
	try
          let m' = dereference_group sds tn in
          split_content_model sds m'
	with
	| Not_found ->
	    raise (Query (Type_Error (m.pmodel_loc, ("Type "
						     ^ (Sym.string_of_group_symbol tn)
						     ^ " not in scope"))))
      end
  | TBound (m',i',j') ->
      if (model_contains sds attribute_model m') then (m, mempty)
      else (mempty, m)
  | TComma (m1, m2) ->
      (* If both c1 and a2 are non-empty, we have a problem *)
      let (a1, c1) = split_content_model sds m1 in 
      let (a2, c2) = split_content_model sds m2 in 
      let invalid_content_model = 
	begin
	  match c1.pmodel_desc with
	  | TEmpty -> false
	  | _ -> 
	      begin 
		match a2.pmodel_desc with
		| TEmpty -> false
		| _ -> true
	      end
	end
      in 
      if (invalid_content_model) then 
	raise (Query (Type_Error (m.pmodel_loc, ("Invalid content model " ^ (bprintf_model "" m)))))
      else 
	begin
	  match c1.pmodel_desc with
	  | TEmpty -> ((fmkmodel (TComma(a1, a2)) m.pmodel_loc), c2)
	  | _ -> (a1, (fmkmodel (TComma(c1, c2)) m.pmodel_loc))
	end
  | TBar (m1, m2) ->
      if ((model_contains sds attribute_model m1) || (model_contains sds attribute_model m2)) then 
	(m, mempty)
      else (mempty, m)
  | TAmp _ ->
      raise (Query (Type_Error (m.pmodel_loc,
				"project: ampersand '&' not supported yet")))
  | _ ->
      if (model_contains sds attribute_model m) then 
	(m, mempty)
      else (mempty, m) 

(* Extract the attribute model from within an element content model *)

let only_non_attribute_model m =
  match m.pmodel_desc with
  | TAttrLocal _ -> fmkmodel TEmpty Finfo.bogus
  | TAttrRef _ -> fmkmodel TEmpty Finfo.bogus
  | _ -> m

let extract_content_model s m = 
  try
    apply_to_unit_types s only_non_attribute_model m
  with
  | _ ->
      raise (Query (Internal_Error ("Error processing content model: " ^
				    (Print.bprintf_model "" m))))

type optional_attribute_model =
    (model * Sym.rgroup_symbol list)

type required_attribute_model =
    (model * Sym.rgroup_symbol list)

type attribute_model =
    optional_attribute_model list * required_attribute_model list

let make_optional_tag minmax =
  match minmax with
  | None ->
      false
  | Some (_,UNBOUNDED) ->
      raise (Query (Internal_Error "Cannot have maxOccurs = unbounded in attribute model"))
  | Some (UNBOUNDED,_) ->
      raise (Query (Internal_Error "Cannot have maxOccurs = unbounded in attribute model"))
  | Some (UP_INT min, UP_INT max) ->
      if max > 1
      then
	raise (Query (Internal_Error "Cannot have maxOccurs > 1 in attribute model"))
      else
	if min = 0
	then
	  if max = 1
	  then
	    true
	  else
	    raise (Query (Internal_Error "Cannot have maxOccurs < 1 in attribute model"))
	else
	  if min = 1
	  then
	    false
	  else
	    raise (Query (Internal_Error "Cannot have minOccurs > 1 in attribute model"))

let rec process_attribute_model_aux s visited_names m optional_tag =
  match m.pmodel_desc with
  | TGroupRef (Sym.RGroupSym x) ->
      let t' = (get_group s x)
      in process_attribute_model_aux s (x::visited_names) t' optional_tag
  | TGroupRef (Sym.UGroupSym _) ->
      raise (Query (Internal_Error ("Dereferencing an unresolved group")))
  | TAttrRef s ->
      if (make_optional_tag optional_tag)
      then ((m,visited_names) :: [], [])
      else ([], (m,visited_names) :: [])
  | TAttrLocal(l,_) ->
      if (make_optional_tag optional_tag)
      then ((m,visited_names) :: [], [])
      else ([], (m,visited_names) :: [])
  | TBase _ ->
      ([],[])
  | TDocument _ ->
      ([],[])
  | TText ->
      ([],[])
  | TTuple _ ->
      ([],[])
  | TElemRef _ ->
      ([],[])
  | TElemLocal(_,_,_) ->
      ([],[])
  | TBound (m', min, max) ->
      process_attribute_model_aux s visited_names m' (Some (min,max))
  | TComma (m1, m2) ->
      (* Should be disallowed to the benefit of & *)
      let (om1, rm1) = process_attribute_model_aux s visited_names m1 optional_tag in
      let (om2, rm2) = process_attribute_model_aux s visited_names m2 optional_tag in
      (om1 @ om2, rm1 @ rm2)
  | TEmpty ->
      ([],[])
  | TBar (m1, m2) ->
      let (om1, rm1) = process_attribute_model_aux s visited_names m1 optional_tag in
      let (om2, rm2) = process_attribute_model_aux s visited_names m2 optional_tag in
      if (om1 != []) &&  (rm1 != []) &&  (om2 != []) &&  (rm2 != [])
      then
        raise (Query (Internal_Error "Cannot have choice group in attribute model"))
      else
	([],[])
  | TNone ->
      ([],[])
  | TAmp (m1,m2) ->
      let (om1, rm1) = process_attribute_model_aux s visited_names m1 optional_tag in
      let (om2, rm2) = process_attribute_model_aux s visited_names m2 optional_tag in
      (om1 @ om2, rm1 @ rm2)

let process_attribute_model s model =
  process_attribute_model_aux s [] model None

let extract_attribute_model s m = 
  let only_attribute_model m =
    match m.pmodel_desc with
    | TAttrLocal _ -> m
    | TAttrRef _ -> m
    | _ -> fmkmodel TEmpty Finfo.bogus
  and only_non_attribute_model m =
    match m.pmodel_desc with
    | TAttrLocal _ -> fmkmodel TEmpty Finfo.bogus
    | TAttrRef _ -> fmkmodel TEmpty Finfo.bogus
    | _ -> m
  in
  let full_attribute_model = apply_to_unit_types s only_attribute_model m in
  try
    let attr_model_list = process_attribute_model s full_attribute_model in
    (attr_model_list, apply_to_unit_types s only_non_attribute_model m)
  with
  | _ ->
      raise (Query (Internal_Error ("Error processing attribute model: " ^
				    (Print.bprintf_model "" full_attribute_model) ^
				    " derived from original model: " ^
				    (Print.bprintf_model "" m))))

(**********************)
(* Content model kind *)
(**********************)

(* Note:
     These functions are used to classify content model for elements
     according to the XML Schema specification. There are three kinds
     of content models: simple content, complex content and mixed
     content.
   - Jerome
 *)

type element_content_model_kind =
  | ElementEmptyContent   of model
  | ElementSimpleContent  of model
  | ElementComplexContent of model
  | ElementMixedContent   of model      (* cannot be exercised yet... *)

let get_element_content_model_kind s tm =
  match tm.ptyped_model_desc with
  | TAnonymousType m ->
      let content_model = extract_content_model s m in
      if (empty_type s content_model)
      then
	ElementEmptyContent m
      else if (forall_unit_types s (is_simple s) content_model)
      then
	ElementSimpleContent m
      else if (forall_unit_types s (is_complex_node s) content_model)
      then
	ElementComplexContent m
      else if (forall_unit_types s (is_node s) content_model)
      then
	ElementMixedContent m
      else
	raise (Query (Internal_Error ("Unrecognized element content model" ^
				      (bprintf_model "" m))))

type attribute_content_model_kind =
  | AttributeSimpleContent of model

let get_attribute_content_model_kind s tm =
  match tm.ptyped_model_desc with
  | TAnonymousType m ->
      if (forall_unit_types s (is_simple s) m)
      then
	AttributeSimpleContent m
      else
	raise (Query (Internal_Error "Unrecognized attribute content model"))

