(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: simplification.ml,v 1.6 2002/10/27 20:02:53 simeon Exp $ *)

open Occurrence
open Xquery_ast
open Xquery_astutil
open Schema
open Error
open Print
open Format

(* Union normalization *)

let rec deunion_types s m =
  match m.pmodel_desc with
  | TGroupRef (Sym.RGroupSym x) ->
      let m' = dereference_rgroup s x in
      deunion_types s m'
  | TBar (m1,m2) ->
      (deunion_types s m1) @ (deunion_types s m2)
  | _ -> [m]

(* type normalization *)

let rec elim_subtype_duplicates tr tl stack =
  match tl with
  | [] -> []
  | t :: tl' ->
      begin
	if (List.length(tl') > 0) && !Conf.debugging then
	  begin
	    printf "in elim_subtype : non-empty list";
	    List.iter (printf_model "\n") tl'
	  end;
	if (List.exists (Subsumption.isSubtype tr t) (stack @ tl'))
	then
	  elim_subtype_duplicates tr tl' stack
	else
	  t :: (elim_subtype_duplicates tr tl' (t :: stack))
      end

(* Note: rec_simplify__ty is applied recursively until a fixed point
   is reached.  simplify_one_ty returns a pair; the second value is a
   boolean indicating whether the type expression was changed and
   therefore it is necessary to recurse.
   - Jerome *)

let rec simplify_one_ty tr t =
  match t.pmodel_desc with
  | TAttrLocal (l, tm) ->
      begin
	match tm.ptyped_model_desc with 
	| TAnonymousType(m') -> 
	    let (m'',b) = simplify_one_ty tr m' in 
            (fmkmodel (TAttrLocal (l,
				   fmktyped_model (TAnonymousType(m'')) tm.ptyped_model_loc)) t.pmodel_loc,
	     b)
      end
  | TElemLocal (l, d, tm) ->
      begin
	match tm.ptyped_model_desc with 
	| TAnonymousType(m') -> 
	    let (m'',b) = simplify_one_ty tr m' in 
            (fmkmodel (TElemLocal(l,
				  d,
				  fmktyped_model (TAnonymousType(m'')) tm.ptyped_model_loc)) t.pmodel_loc,
	     b)
      end
  | TComma (m, m') ->
      begin
	match m.pmodel_desc with 
	(* ((e1,e2),e3) == (e1,(e2,e3)) *)
	| TComma(m1, m2) -> 
            let (t1',_) = simplify_one_ty tr m1
            and (t2',_) = simplify_one_ty tr m2
            and (t3',_) = simplify_one_ty tr m' in 
            let seq = fmkmodel (TComma(t2', t3')) t2'.pmodel_loc in
	    (fmkmodel (TComma(t1', seq)) t.pmodel_loc, true)
	(* (0,e) == 0 *)
	| TNone ->
	    (fmkmodel TNone t.pmodel_loc,true)
	(* ((),e) == e *)
	| TEmpty ->
	    let (m'',_) = simplify_one_ty tr m' in
	    (m'',true)
	(* Otherwise, match on RHS of TComma *)
	| _ ->
	    begin
              match m'.pmodel_desc with 
	(* (e,0) == 0 *)
              | TNone ->
		  (fmkmodel TNone t.pmodel_loc,true)
	(* (e,()) == e *)
	      | TEmpty ->
		  let (m',_) = simplify_one_ty tr m in
		  (m',true)
        (* e1, e2 == simplify(e1), simplify(e2) *)
	      | _ -> 
		  let (m1',b1) = simplify_one_ty tr m
		  and (m2',b2) = simplify_one_ty tr m' in
		  (fmkmodel (TComma (m1', m2')) t.pmodel_loc, b1||b2)
	    end
      end

  | TBar (_, _) ->
      let m' = simplify_union tr t in
      begin
	match m'.pmodel_desc with
	| TBar (m1,m2) ->
	    begin
	      match m1.pmodel_desc with
		(* t|0 = t *)
	      |	TNone ->
		  (m2,true)
		(* t|() = t{0,1} *)
	      |	TEmpty ->
		  ((fmkmodel (TBound (m2,Occurrence.occurs 0,Occurrence.occurs 1)) m2.pmodel_loc), true)
	      |	_ ->
		  begin
		    match m2.pmodel_desc with
		(* 0|t = t *)
		    | TNone ->
			(m1,true)
		(* ()|t = t{0,1} *)
		    | TEmpty ->
			((fmkmodel (TBound (m1,Occurrence.occurs 0,Occurrence.occurs 1)) m1.pmodel_loc), true)
		    |	_ ->
			(m',false)
		  end
 	    end
	| _ ->
	    (m',true)
      end
        (* t{1,1} == t *)
  | TBound (m, UP_INT 1, UP_INT 1) -> 
      let (m',b) = simplify_one_ty tr m in
      (fmkmodel m'.pmodel_desc t.pmodel_loc, true)

  | TBound (m, UP_INT 0, n) -> 
     begin
       match m.pmodel_desc with  
       | TEmpty ->
	   (fmkmodel TEmpty m.pmodel_loc,true)
       | TBar(m', m'') ->
	(* (() | e){0,n} == e{0,n} *)
	   begin
	     match m'.pmodel_desc with 
	     | TEmpty -> 
		 let (m''',_) = simplify_one_ty tr m'' in
		 (fmkmodel (TBound (m''', UP_INT 0, n)) m.pmodel_loc, true)
	     | _ ->
		 begin
		   match m''.pmodel_desc with 
                   (* (e | ()){0,n} == e{0,n} *)
		   | TEmpty -> 
		       let (m''',_) = simplify_one_ty tr m' in
		       (fmkmodel (TBound (m''', UP_INT 0, n)) m.pmodel_loc, true)
		   | _ -> 
		       let (m''', b) = simplify_one_ty tr m in
		       (fmkmodel (TBound (m''', UP_INT 0, n)) m.pmodel_loc, b)
		 end
	   end
       | TBound (m', _, n') -> 
	   (fmkmodel (TBound (m', UP_INT 0, mult n n')) m.pmodel_loc, true)
       | _ -> 
	   let (m',b) = simplify_one_ty tr m in
	   (fmkmodel (TBound(m', UP_INT 0, n)) m.pmodel_loc, b)
     end

(* M&B : This case and the one above need to be coalesced *)

  | TBound (m, i, j) -> 
	(* (e{n,m}){n',m'} == e{n*n',m*m'} *)
      begin
	match m.pmodel_desc with 
	|  TBound(m', i', j') ->
	    let (m'',_) = simplify_one_ty tr m' in
	    (fmkmodel (TBound (m'', mult i i', mult j j')) m.pmodel_loc, true)

	(* (){n,m} == () *)
	| TEmpty ->
	    (fmkmodel TEmpty m.pmodel_loc,true)
	| _ -> 
	    let (m',b) = simplify_one_ty tr m in
	    (fmkmodel (TBound (m', i, j)) m.pmodel_loc, b)
      end

  | TAmp (m1, m2) ->
      raise (Query (Prototype("Ampersand not supported")))
  | _ ->
      (t,false)

and simplify_union tr t =
  let all_types = deunion_types tr t in
  if !Conf.debugging then 
    begin
      printf "\nAfter deunion: \n";
      List.iter (printf_model "\n") all_types
    end;
  let all_simplified_types = List.map (rec_simplify_ty tr) all_types in
  if !Conf.debugging then
    begin
      printf "\nAfter simplify_ty: \n";
      List.iter (printf_model "\n") all_simplified_types
    end;
  let no_nones = List.filter Schemautil.is_none all_simplified_types in 
  if !Conf.debugging then
    begin
      printf "\nAfter filter TNone: \n";
      List.iter (printf_model "\n") no_nones
    end;
  let min_types = (elim_subtype_duplicates tr no_nones []) in
  if !Conf.debugging then
    begin
      printf "\nAfter elim_subtype_duplicates: \n";
      List.iter (printf_model "\n") min_types
    end;
  let ty = Schemautil.weakunion_types min_types in 
  if !Conf.debugging then
    begin
      printf "\nAfter weakunion: \n";
      printf_model "\n" ty
    end;
  ty

(* rec_simplify__ty is applied recursively until a fixed point is
   reached.  simplify_one_ty returns a pair; the second value is a
   boolean indicating whether the type expression was changed and
   therefore it is necessary to recurse.  
*)

and rec_simplify_ty tr t =
  let (t',b) = simplify_one_ty tr t in
  if b
  then rec_simplify_ty tr t'
  else t'

let simplify_ty tr t =
  rec_simplify_ty tr t

