(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: typeutil.ml,v 1.15 2003/05/20 15:29:50 mff Exp $ *)

open Sym
open Occurrence

open Xquery_ast
open Xquery_astutil

open Schemautil
open Simplification

(******************)
(* Type shortcuts *)
(******************)

(* Empty types *)

let empty_sequence_model = fmkmodel TEmpty Finfo.bogus
let empty_choice_model = fmkmodel TNone Finfo.bogus

(* Simple types *)

let anyatomic_model     = fmkmodel (TGroupRef (RGroupSym anyatomic)) Finfo.bogus 
let anynumeric_model    = fmkmodel (TGroupRef (RGroupSym anynumeric)) Finfo.bogus
let anysimpletype_model = fmkmodel (TGroupRef (RGroupSym anysimpletype)) Finfo.bogus
let integer_model       = fmkmodel (TBase Datatypes.ATInteger) Finfo.bogus
let boolean_model       = fmkmodel (TBase Datatypes.ATBoolean) Finfo.bogus

(* Sequences of simple types *)

let anyatomic_optional_model = fmkmodel (TBound (anyatomic_model, Occurrence.occurs 0, Occurrence.occurs 1)) Finfo.bogus
let anyatomic_sequence_model = fmkmodel (TBound (anyatomic_model, Occurrence.occurs 1, Occurrence.unbounded)) Finfo.bogus

(* Complex types *)

let anyelem_model = fmkmodel (TGroupRef (RGroupSym anyelement)) Finfo.bogus
let anydoc_model  = fmkmodel (TGroupRef (RGroupSym anydoc)) Finfo.bogus
let anyattr_model = fmkmodel (TGroupRef (RGroupSym anyattribute)) Finfo.bogus
let anynode_model = fmkmodel (TGroupRef (RGroupSym anynode)) Finfo.bogus
let anyitem_model = fmkmodel (TGroupRef (RGroupSym anyitem)) Finfo.bogus 

let anyparent_model = fmkmodel (TBar (anydoc_model,anyelem_model))  Finfo.bogus
let anyancestor_model = fmkmodel (TBound (anyparent_model, Occurrence.occurs 1, Occurrence.unbounded)) Finfo.bogus

(* The sort type should be an optional item *)

let sorttype_model =
  fmkmodel (TBound (anyitem_model, Occurrence.occurs 0, Occurrence.occurs 1)) Finfo.bogus

let anynode_sequence_model =
  fmkmodel (TBound (anynode_model, Occurrence.occurs 0, Occurrence.unbounded)) Finfo.bogus

let anynode_plus_model =
  fmkmodel (TBound (anynode_model, Occurrence.occurs 1, Occurrence.unbounded)) Finfo.bogus

(************************)
(* Subtyping operations *)
(************************)

(* Main subtyping function *)

let is_subtype_of schema m1 m2 =
  let m1' = simplify_ty schema m1
  and m2' = simplify_ty schema m2 in
  Subsumption.isSubtype schema m1' m2'

(* Specific subtyping functions *)

let is_anysimpletype schema model =
  is_subtype_of schema model anysimpletype_model

let is_anynumeric schema model =
  is_subtype_of schema model anynumeric_model

let is_anyatomic schema model =
  is_subtype_of schema model anyatomic_model

let is_anyatomic_optional schema model =
  is_subtype_of schema model anyatomic_optional_model

let is_anyatomic_sequence schema model =
  is_subtype_of schema model anyatomic_sequence_model

let is_empty_sequence schema t =
  is_subtype_of schema t empty_sequence_model

let is_empty_choice schema t =
  is_subtype_of schema t empty_choice_model

let is_anynode schema model =
  is_subtype_of schema model anynode_model

let is_anynode_sequence schema model =
  is_subtype_of schema model anynode_sequence_model

let is_anynode_plus schema model =
  is_subtype_of schema model anynode_plus_model

let is_sorttype schema model =
  is_subtype_of schema model sorttype_model

let is_element schema model =
  is_subtype_of schema model anyelem_model

(*************)
(* Factoring *)
(*************)

(* Factor *)

let factor_with_units sds m =
  let (mts,i,j) = factor_aux sds m in
  let (i',j') = approximate_occurrences (i,j) in
  (mts, i', j')

(*

  factor_type_in_typeswicth factors the type of the argument
  expression.  For the empty type TEmpty, factor_with_units (used to
  type for and sort expressions) returns (TNone, 0, 0), which is not
  applicable to typeswitch.  In this case, the factored type is
  converted back to TEmpty.

*)

let factor_with_units sds m =
  if (is_empty_sequence sds m)
  then
    ([fmkmodel TEmpty Finfo.bogus],  Occurrence.UP_INT 0,  Occurrence.UP_INT 0)
  else
    let (mts,i,j) = factor_aux sds m in
    let (i',j') = approximate_occurrences (i,j) in
    (mts, i', j')

let factor sds m =
  if (is_empty_sequence sds m)
  then
    (fmkmodel TEmpty Finfo.bogus,  Occurrence.UP_INT 0,  Occurrence.UP_INT 0)
  else
    let (mts,i,j) = factor_aux sds m in
    let (i',j') = approximate_occurrences (i,j) in
    (weakunion_types mts, i', j')

(* Turns a factored type back into a type *)

let defactor sds (m,i,j) = 
  fmkmodel (TBound (m, i, j)) m.pmodel_loc

(* Recursive factoring *)

let recfactor sds m = 
  if (is_empty_sequence sds m)
  then
    fmkmodel TEmpty Finfo.bogus
  else
    let (mts, i, j) =  recfactor_aux sds m [] [] in
    let (i',j') = approximate_occurrences (i,j) in
    defactor sds (weakunion_types mts, i', j')

(* Factoring of sequence types *)

let factor_datatype dt =
  match dt.pdatatype_desc with
  | (DTEmpty, _) ->
      (DTEmpty, occurs 0, occurs 0)
  | (dtk, None) ->
      (dtk,occurs 1,occurs 1)
  | (dtk,(Some (m,n))) ->
      (dtk,m,n)

(***************************)
(* Intersection operations *)
(***************************)

(* Main intersects function *)

(*
let has_some_intersection_with schema m1 m2 =
  let m1' = simplify_ty schema m1
  and m2' = simplify_ty schema m2 in
  Intersects.hasIntersection schema m1' m2'

*)


