(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: cast_as.ml,v 1.2 2003/05/20 15:29:50 mff Exp $ *)

(* Module Cast_as
   Description:
     This module implements casting of atomic values to a given atomic
     type.
*)

open Error

open Namespace
open Sym

open Datatypes
open Decimal


(************************)
(* Cast evaluation code *)
(************************)

(* Note:
     This is a fairly large match-case which looks at each pairs of
     two atomic types. See [F&O Section 16. Casting Functions].
   - Jerome
 *)

let eval_cast nse bt bv =
  match bt with
  | ATString ->
      let result_value = 
	begin
	  match bv with
	  | AtomicString s ->
	      s
	  | AtomicBoolean b ->
	      if b then "true" else "false"
	  | AtomicDecimal d ->
	      (* Note: This should return the XML Schema normalized value *)
	      Decimal._string_of_decimal d
	  | AtomicFloat f ->
	      (* Note: This should return the XML Schema normalized value *)
	      string_of_float f
	  | AtomicDouble d ->
	      (* Note: This should return the XML Schema normalized value *)
	      string_of_float d
	  | AtomicDuration d ->
	      d
	  | AtomicDateTime d ->
	      DateTime.string_of_dateTime d
	  | AtomicTime t ->
	      t
	  | AtomicDate d ->
	      d
	  | AtomicGYearMonth g ->
	      g
	  | AtomicGYear g ->
	      g
	  | AtomicGMonthDay g ->
	      g
	  | AtomicGDay g ->
	      g
	  | AtomicGMonth g ->
	      g
	  | AtomicHexBinary h ->
	      h
	  | AtomicBase64Binary b ->
	      b
	  | AtomicAnyURI a ->
	      AnyURI._string_of_uri a
	  | AtomicQName q ->
	      anon_prefix_string q
	  | AtomicNOTATION n ->
	      raise (Query (Cast_Error ("Cannot cast NOTATION value: \"" ^ n ^ "\" to a string")))
	  | AtomicInteger i ->
	      _string_of_integer i
	  | AtomicInt i ->
	      _string_of_int i
	  | AtomicYearMonthDuration t ->
	      t
	  | AtomicDayTimeDuration t ->
	      t
	  | AtomicUntyped t ->
	      t
      end
      in AtomicString result_value
  | ATBoolean ->
      let result_value = 
	begin
	  match bv with
	  | AtomicString s ->
	      let us = String.uppercase s in
	      begin
		match us with
		| "TRUE" -> true
		| "1" -> true
		| "FALSE" -> false
		| "0" -> false
		|  _ ->
		    raise (Query (Cast_Error ("Cannot cast String value: \"" ^ s ^ "\" to a Boolean")))
	      end
	  | AtomicBoolean b ->
	      b
	  | AtomicDecimal d ->
	      if Decimal._decimal_eq d _decimal_zero then false else true
	  | AtomicFloat f ->
	      if f = 0.0 or f = nan then false else true
	  | AtomicDouble d ->
	      if d = 0.0 or d = nan then false else true
	  | AtomicDuration d ->
	      raise (Query (Cast_Error "Cannot cast a Duration value to a Boolean"))
	  | AtomicDateTime d ->
	      raise (Query (Cast_Error "Cannot cast a DateTime value to a Boolean"))
	  | AtomicTime t ->
	      raise (Query (Cast_Error "Cannot cast a Time value to a Boolean"))
	  | AtomicDate d ->
	      raise (Query (Cast_Error "Cannot cast a Date value to a Boolean"))
	  | AtomicGYearMonth g ->
	      raise (Query (Cast_Error "Cannot cast a GYearMonth value to a Boolean"))
	  | AtomicGYear g ->
	      raise (Query (Cast_Error "Cannot cast a GYear value to a Boolean"))
	  | AtomicGMonthDay g ->
	      raise (Query (Cast_Error "Cannot cast a GMonthDay value to a Boolean"))
	  | AtomicGDay g ->
	      raise (Query (Cast_Error "Cannot cast a GMonthGDay value to a Boolean"))
	  | AtomicGMonth g ->
	      raise (Query (Cast_Error "Cannot cast a GMonth value to a Boolean"))
	  | AtomicHexBinary h ->
	      begin
		match h with
		| "0" -> false
		| "1" -> true
		| _ ->
		    raise (Query (Cast_Error "Cannot cast a HexBinary value to a Boolean"))
	      end
	  | AtomicBase64Binary b ->
	      begin
		match b with
		| "0" -> false
		| "1" -> true
		| _ ->
		    raise (Query (Cast_Error "Cannot cast a Base64Binary value to a Boolean"))
	      end
	  | AtomicAnyURI a ->
	      raise (Query (Cast_Error "Cannot cast a URI value to a Boolean"))
	  | AtomicQName q ->
	      raise (Query (Cast_Error "Cannot cast a QName value to a Boolean"))
	  | AtomicNOTATION n ->
	      raise (Query (Cast_Error "Cannot cast a NOTATION value to a Boolean"))
	  | AtomicInteger i ->
	      if Decimal._integer_eq i _integer_zero then false else true
	  | AtomicInt i ->
	      if Decimal._int_eq i _int_zero then false else true
	  | AtomicYearMonthDuration n ->
	      raise (Query (Cast_Error "Cannot cast a yearMonthDuration value to a Boolean"))
	  | AtomicDayTimeDuration n ->
	      raise (Query (Cast_Error "Cannot cast a dayTimeDuration value to a Boolean"))
	  | AtomicUntyped t ->
	      let us = String.uppercase t in
	      begin
		match us with
		| "TRUE" -> true
		| "1" -> true
		| "FALSE" -> false
		| "0" -> false
		| _ ->
		    raise (Query (Cast_Error ("Cannot cast String value: ^ \"" ^ t ^ "\" to a Decimal")))
	      end
	end
      in AtomicBoolean result_value
  | ATDecimal ->
      let result_value = 
	begin
	  match bv with
	  | AtomicString s ->
	      (* Note: Should we make a difference between string and text ?  - Jerome *)
	      decimal_of_untyped s
	  | AtomicBoolean b ->
	      raise (Query (Cast_Error "Cannot cast a Boolean value to a Decimal"))
	  | AtomicDecimal d ->
	      d
	  | AtomicFloat f ->
	      raise (Query (Cast_Error "Cannot cast a Float value to a Decimal"))
	  | AtomicDouble d ->
	      raise (Query (Cast_Error "Cannot cast a Double value to a Decimal"))
	  | AtomicDuration d ->
	      raise (Query (Cast_Error "Cannot cast a Duration value to a Decimal"))
	  | AtomicDateTime d ->
	      raise (Query (Cast_Error "Cannot cast a DateTime value to a Decimal"))
	  | AtomicTime t ->
	      raise (Query (Cast_Error "Cannot cast a Time value to a Decimal"))
	  | AtomicDate d ->
	      raise (Query (Cast_Error "Cannot cast a Date value to a Decimal"))
	  | AtomicGYearMonth g ->
	      raise (Query (Cast_Error "Cannot cast a GYearMonth value to a Decimal"))
	  | AtomicGYear g ->
	      raise (Query (Cast_Error "Cannot cast a GYear value to a Decimal"))
	  | AtomicGMonthDay g ->
	      raise (Query (Cast_Error "Cannot cast a GMonthDay value to a Decimal"))
	  | AtomicGDay g ->
	      raise (Query (Cast_Error "Cannot cast a GMonthGDay value to a Decimal"))
	  | AtomicGMonth g ->
	      raise (Query (Cast_Error "Cannot cast a GMonth value to a Decimal"))
	  | AtomicHexBinary h ->
	      raise (Query (Cast_Error "Cannot cast a HexBinary value to a Decimal"))
	  | AtomicBase64Binary b ->
	      raise (Query (Cast_Error "Cannot cast a Base64Binary value to a Decimal"))
	  | AtomicAnyURI a ->
	      raise (Query (Cast_Error "Cannot cast a URI value to a Decimal"))
	  | AtomicQName q ->
	      raise (Query (Cast_Error "Cannot cast a QName value to a Decimal"))
	  | AtomicNOTATION n ->
	      raise (Query (Cast_Error "Cannot cast a NOTATION value to a Decimal"))
	  | AtomicInteger i ->
	      Decimal._decimal_of_integer i
	  | AtomicInt i ->
	      Decimal._decimal_of_int i
	  | AtomicYearMonthDuration t ->
	      raise (Query (Cast_Error "Cannot cast a yearMonthDuration value to a Decimal"))
	  | AtomicDayTimeDuration t ->
	      raise (Query (Cast_Error "Cannot cast a dayTimeDuration value to a Decimal"))
	  | AtomicUntyped t ->
	      (* Note: Should we make a difference between string and text ?  - Jerome *)
	      decimal_of_untyped t
	end
      in AtomicDecimal result_value
  | ATFloat ->
      let result_value = 
	begin
	  match bv with
	  | AtomicString s ->
	      (* Note: Should we make a difference between string and text ?  - Jerome *)
	      float_of_untyped s
	  | AtomicBoolean b ->
	      raise (Query (Cast_Error "Cannot cast a Boolean value to a Float"))
	  | AtomicDecimal d ->
	      float_of_untyped (Decimal._string_of_decimal d)
	  | AtomicFloat f ->
	      f
	  | AtomicDouble d ->
	      d
	  | AtomicDuration d ->
	      raise (Query (Cast_Error "Cannot cast a Duration value to a Float"))
	  | AtomicDateTime d ->
	      raise (Query (Cast_Error "Cannot cast a DateTime value to a Float"))
	  | AtomicTime t ->
	      raise (Query (Cast_Error "Cannot cast a Time value to a Float"))
	  | AtomicDate d ->
	      raise (Query (Cast_Error "Cannot cast a Date value to a Float"))
	  | AtomicGYearMonth g ->
	      raise (Query (Cast_Error "Cannot cast a GYearMonth value to a Float"))
	  | AtomicGYear g ->
	      raise (Query (Cast_Error "Cannot cast a GYear value to a Float"))
	  | AtomicGMonthDay g ->
	      raise (Query (Cast_Error "Cannot cast a GMonthDay value to a Float"))
	  | AtomicGDay g ->
	      raise (Query (Cast_Error "Cannot cast a GMonthGDay value to a Float"))
	  | AtomicGMonth g ->
	      raise (Query (Cast_Error "Cannot cast a GMonth value to a Float"))
	  | AtomicHexBinary h ->
	      raise (Query (Cast_Error "Cannot cast a HexBinary value to a Float"))
	  | AtomicBase64Binary b ->
	      raise (Query (Cast_Error "Cannot cast a Base64Binary value to a Float"))
	  | AtomicAnyURI a ->
	      raise (Query (Cast_Error "Cannot cast a URI value to a Float"))
	  | AtomicQName q ->
	      raise (Query (Cast_Error "Cannot cast a QName value to a Float"))
	  | AtomicNOTATION n ->
	      raise (Query (Cast_Error "Cannot cast a NOTATION value to a Float"))
	  | AtomicInteger i ->
	      float_of_untyped (Decimal._string_of_integer i)
	  | AtomicInt i ->
	      float_of_untyped (Decimal._string_of_int i)
	  | AtomicYearMonthDuration t ->
	      raise (Query (Cast_Error "Cannot cast a yearMonthDuration value to a Float"))
	  | AtomicDayTimeDuration t ->
	      raise (Query (Cast_Error "Cannot cast a dayTimeDuration value to a Float"))
	  | AtomicUntyped t ->
	      float_of_untyped t
	end
      in AtomicFloat result_value
  | ATDouble ->
      let result_value = 
	begin
	  match bv with
	  | AtomicString s ->
	      (* Note: Should we make a difference between string and text ?  - Jerome *)
	      double_of_untyped s
	  | AtomicBoolean b ->
	      raise (Query (Cast_Error "Cannot cast a Boolean value to a Double"))
	  | AtomicDecimal d ->
	      double_of_untyped (Decimal._string_of_decimal d)
	  | AtomicFloat f ->
	      f
	  | AtomicDouble d ->
	      d
	  | AtomicDuration d ->
	      raise (Query (Cast_Error "Cannot cast a Duration value to a Double"))
	  | AtomicDateTime d ->
	      raise (Query (Cast_Error "Cannot cast a DateTime value to a Double"))
	  | AtomicTime t ->
	      raise (Query (Cast_Error "Cannot cast a Time value to a Double"))
	  | AtomicDate d ->
	      raise (Query (Cast_Error "Cannot cast a Date value to a Double"))
	  | AtomicGYearMonth g ->
	      raise (Query (Cast_Error "Cannot cast a GYearMonth value to a Double"))
	  | AtomicGYear g ->
	      raise (Query (Cast_Error "Cannot cast a GYear value to a Double"))
	  | AtomicGMonthDay g ->
	      raise (Query (Cast_Error "Cannot cast a GMonthDay value to a Double"))
	  | AtomicGDay g ->
	      raise (Query (Cast_Error "Cannot cast a GMonthGDay value to a Double"))
	  | AtomicGMonth g ->
	      raise (Query (Cast_Error "Cannot cast a GMonth value to a Double"))
	  | AtomicHexBinary h ->
	      raise (Query (Cast_Error "Cannot cast a HexBinary value to a Double"))
	  | AtomicBase64Binary b ->
	      raise (Query (Cast_Error "Cannot cast a Base64Binary value to a Double"))
	  | AtomicAnyURI a ->
	      raise (Query (Cast_Error "Cannot cast a URI value to a Double"))
	  | AtomicQName q ->
	      raise (Query (Cast_Error "Cannot cast a QName value to a Double"))
	  | AtomicNOTATION n ->
	      raise (Query (Cast_Error "Cannot cast a NOTATION value to a Double"))
	  | AtomicInteger i ->
	      double_of_untyped (Decimal._string_of_integer i)
	  | AtomicInt i ->
	      double_of_untyped (Decimal._string_of_int i)
	  | AtomicYearMonthDuration t ->
	      raise (Query (Cast_Error "Cannot cast a yearMonthDuration value to a Double"))
	  | AtomicDayTimeDuration t ->
	      raise (Query (Cast_Error "Cannot cast a dayTimeDuration value to a Double"))
	  | AtomicUntyped t ->
	      double_of_untyped t
	end
      in AtomicDouble result_value
  | ATDuration ->
      raise (Query (Prototype "Cast to a Duration not supported"))
  | ATDateTime ->
      let result_value = 
	begin
	  match bv with
	  | AtomicString s ->
	      DateTime.dateTime_of_string s
	  | AtomicBoolean b ->
	      raise (Query (Cast_Error "Cannot cast a Boolean value to a DateTime"))
	  | AtomicDecimal d ->
	      raise (Query (Cast_Error "Cannot cast a Decimal value to a DateTime"))
	  | AtomicFloat f ->
	      raise (Query (Cast_Error "Cannot cast a Float value to a DateTime"))
	  | AtomicDouble d ->
	      raise (Query (Cast_Error "Cannot cast a Double value to a DateTime"))
	  | AtomicDuration d ->
	      raise (Query (Prototype "Cannot cast a Duration value to a DateTime"))
	  | AtomicDateTime d ->
	      raise (Query (Prototype "Cannot cast a DateTime value to a DateTime"))
	  | AtomicTime t ->
	      raise (Query (Prototype "Cannot cast a Time value to a DateTime"))
	  | AtomicDate d ->
	      raise (Query (Prototype "Cannot cast a Date value to a DateTime"))
	  | AtomicGYearMonth g ->
	      raise (Query (Prototype "Cannot cast a GYearMonth value to a DateTime"))
	  | AtomicGYear g ->
	      raise (Query (Prototype "Cannot cast a GYear value to a DateTime"))
	  | AtomicGMonthDay g ->
	      raise (Query (Prototype "Cannot cast a GMonthDay value to a DateTime"))
	  | AtomicGDay g ->
	      raise (Query (Prototype "Cannot cast a GMonthGDay value to a DateTime"))
	  | AtomicGMonth g ->
	      raise (Query (Prototype "Cannot cast a GMonth value to a DateTime"))
	  | AtomicHexBinary h ->
	      raise (Query (Cast_Error "Cannot cast a HexBinary value to a DateTime"))
	  | AtomicBase64Binary b ->
	      raise (Query (Cast_Error "Cannot cast a Base64Binary value to a DateTime"))
	  | AtomicAnyURI a ->
	      raise (Query (Cast_Error "Cannot cast a URI value to a DateTime"))
	  | AtomicQName q ->
	      raise (Query (Cast_Error "Cannot cast a QName value to a DateTime"))
	  | AtomicNOTATION n ->
	      raise (Query (Cast_Error "Cannot cast a NOTATION value to a DateTime"))
	  | AtomicInteger i ->
	      raise (Query (Cast_Error "Cannot cast a Integer value to a DateTime"))
	  | AtomicInt i ->
	      raise (Query (Cast_Error "Cannot cast a Int value to a DateTime"))
	  | AtomicYearMonthDuration t ->
	      raise (Query (Cast_Error "Cannot cast a yearMonthDuration value to a DateTeime"))
	  | AtomicDayTimeDuration t ->
	      raise (Query (Cast_Error "Cannot cast a dayTimeDuration value to a DateTime"))
	  | AtomicUntyped t ->
	      DateTime.dateTime_of_string t
	end
      in AtomicDateTime result_value
  | ATTime ->
      raise (Query (Prototype "Cast to a Time not supported"))
  | ATDate ->
      raise (Query (Prototype "Cast to a Date not supported"))
  | ATGYearMonth ->
      raise (Query (Prototype "Cast to a GYearMonth not supported"))
  | ATGYear ->
      raise (Query (Prototype "Cast to a GYear not supported"))
  | ATGMonthDay ->
      raise (Query (Prototype "Cast to a GMonthDay not supported"))
  | ATGDay ->
      raise (Query (Prototype "Cast to a GDay not supported"))
  | ATGMonth ->
      raise (Query (Prototype "Cast to a GMonth not supported"))
  | ATHexBinary ->
      raise (Query (Cast_Error "Cannot cast to a HexBinary"))
  | ATBase64Binary ->
      raise (Query (Cast_Error "Cannot cast to a Base64Binary"))
  | ATAnyURI ->
      let result_value = 
	begin
	  match bv with
	  | AtomicString s ->
	      (* Note: Should we make a difference between string and text ?  - Jerome *)
	      anyURI_of_untyped s
	  | AtomicBoolean b ->
	      raise (Query (Cast_Error "Cannot cast a Boolean value to a AnyURI"))
	  | AtomicDecimal d ->
	      raise (Query (Cast_Error "Cannot cast a Decimal value to a AnyURI"))
	  | AtomicFloat f ->
	      raise (Query (Cast_Error "Cannot cast a Float value to a AnyURI"))
	  | AtomicDouble d ->
	      raise (Query (Cast_Error "Cannot cast a Double value to a AnyURI"))
	  | AtomicDuration d ->
	      raise (Query (Cast_Error "Cannot cast a Duration value to a AnyURI"))
	  | AtomicDateTime d ->
	      raise (Query (Cast_Error "Cannot cast a DateTime value to a AnyURI"))
	  | AtomicTime t ->
	      raise (Query (Cast_Error "Cannot cast a Time value to a AnyURI"))
	  | AtomicDate d ->
	      raise (Query (Cast_Error "Cannot cast a Date value to a AnyURI"))
	  | AtomicGYearMonth g ->
	      raise (Query (Cast_Error "Cannot cast a GYearMonth value to a AnyURI"))
	  | AtomicGYear g ->
	      raise (Query (Cast_Error "Cannot cast a GYear value to a AnyURI"))
	  | AtomicGMonthDay g ->
	      raise (Query (Cast_Error "Cannot cast a GMonthDay value to a AnyURI"))
	  | AtomicGDay g ->
	      raise (Query (Cast_Error "Cannot cast a GMonthGDay value to a AnyURI"))
	  | AtomicGMonth g ->
	      raise (Query (Cast_Error "Cannot cast a GMonth value to a AnyURI"))
	  | AtomicHexBinary h ->
	      raise (Query (Cast_Error "Cannot cast a HexBinary value to a AnyURI"))
	  | AtomicBase64Binary b ->
	      raise (Query (Cast_Error "Cannot cast a Base64Binary value to a AnyURI"))
	  | AtomicAnyURI a ->
	      a
	  | AtomicQName q ->
	      raise (Query (Cast_Error "Cannot cast a QName value to a AnyURI"))
	  | AtomicNOTATION n ->
	      raise (Query (Cast_Error "Cannot cast a NOTATION value to a AnyURI"))
	  | AtomicInteger i ->
	      raise (Query (Cast_Error "Cannot cast a Integer value to a AnyURI"))
	  | AtomicInt i ->
	      raise (Query (Cast_Error "Cannot cast a Int value to a AnyURI"))
	  | AtomicYearMonthDuration t ->
	      raise (Query (Cast_Error "Cannot cast a yearMonthDuration value to a AnyURI"))
	  | AtomicDayTimeDuration t ->
	      raise (Query (Cast_Error "Cannot cast a dayTimeDuration value to a AnyURI"))
	  | AtomicUntyped t ->
	      anyURI_of_untyped t
	end
      in AtomicAnyURI result_value
  | ATQName ->
      let result_value = 
	begin
	  match bv with
	  | AtomicString s ->
	      (* Note: Should we make a difference between string and text ?  - Jerome *)
	      qname_of_untyped nse s
	  | AtomicBoolean b ->
	      raise (Query (Cast_Error "Cannot cast a Boolean value to a QName"))
	  | AtomicDecimal d ->
	      raise (Query (Cast_Error "Cannot cast a Decimal value to a QName"))
	  | AtomicFloat f ->
	      raise (Query (Cast_Error "Cannot cast a Float value to a QName"))
	  | AtomicDouble d ->
	      raise (Query (Cast_Error "Cannot cast a Double value to a QName"))
	  | AtomicDuration d ->
	      raise (Query (Cast_Error "Cannot cast a Duration value to a QName"))
	  | AtomicDateTime d ->
	      raise (Query (Cast_Error "Cannot cast a DateTime value to a QName"))
	  | AtomicTime t ->
	      raise (Query (Cast_Error "Cannot cast a Time value to a QName"))
	  | AtomicDate d ->
	      raise (Query (Cast_Error "Cannot cast a Date value to a QName"))
	  | AtomicGYearMonth g ->
	      raise (Query (Cast_Error "Cannot cast a GYearMonth value to a QName"))
	  | AtomicGYear g ->
	      raise (Query (Cast_Error "Cannot cast a GYear value to a QName"))
	  | AtomicGMonthDay g ->
	      raise (Query (Cast_Error "Cannot cast a GMonthDay value to a QName"))
	  | AtomicGDay g ->
	      raise (Query (Cast_Error "Cannot cast a GMonthGDay value to a QName"))
	  | AtomicGMonth g ->
	      raise (Query (Cast_Error "Cannot cast a GMonth value to a QName"))
	  | AtomicHexBinary h ->
	      raise (Query (Cast_Error "Cannot cast a HexBinary value to a QName"))
	  | AtomicBase64Binary b ->
	      raise (Query (Cast_Error "Cannot cast a Base64Binary value to a QName"))
	  | AtomicAnyURI a ->	
	      raise (Query (Cast_Error "Cannot cast a AnyURI value to a QName"))
	  | AtomicQName q ->
	      q
	  | AtomicNOTATION n ->
	      raise (Query (Cast_Error "Cannot cast a NOTATION value to a QName"))
	  | AtomicInteger i ->
	      raise (Query (Cast_Error "Cannot cast a Integer value to a QName"))
	  | AtomicInt i ->
	      raise (Query (Cast_Error "Cannot cast a Int value to a QName"))
	  | AtomicYearMonthDuration t ->
	      raise (Query (Cast_Error "Cannot cast a yearMonthDuration value to a QName"))
	  | AtomicDayTimeDuration t ->
	      raise (Query (Cast_Error "Cannot cast a dayTimeDuration value to a QName"))
	  | AtomicUntyped t ->
	      qname_of_untyped nse t
	end
      in AtomicQName result_value
  | ATNOTATION ->
      raise (Query (Prototype "Cast to a NOTATION not supported"))
  | ATInteger ->
      let result_value = 
	begin
	  match bv with
	  | AtomicString s ->
	      (* Note: Should we make a difference between string and text ?  - Jerome *)
	      integer_of_untyped s
	  | AtomicBoolean b ->
	      raise (Query (Cast_Error "Cannot cast a Boolean value to an Integer"))
	  | AtomicDecimal d ->
	      Decimal._integer_of_decimal d
	  | AtomicFloat f ->
	      raise (Query (Cast_Error "Cannot cast a Float value to an Integer"))
	  | AtomicDouble d ->
	      raise (Query (Cast_Error "Cannot cast a Double value to an Integer"))
	  | AtomicDuration d ->
	      raise (Query (Cast_Error "Cannot cast a Duration value to an Integer"))
	  | AtomicDateTime d ->
	      raise (Query (Cast_Error "Cannot cast a DateTime value to an Integer"))
	  | AtomicTime t ->
	      raise (Query (Cast_Error "Cannot cast a Time value to an Integer"))
	  | AtomicDate d ->
	      raise (Query (Cast_Error "Cannot cast a Date value to an Integer"))
	  | AtomicGYearMonth g ->
	      raise (Query (Cast_Error "Cannot cast a GYearMonth value to an Integer"))
	  | AtomicGYear g ->
	      raise (Query (Cast_Error "Cannot cast a GYear value to an Integer"))
	  | AtomicGMonthDay g ->
	      raise (Query (Cast_Error "Cannot cast a GMonthDay value to an Integer"))
	  | AtomicGDay g ->
	      raise (Query (Cast_Error "Cannot cast a GMonthGDay value to an Integer"))
	  | AtomicGMonth g ->
	      raise (Query (Cast_Error "Cannot cast a GMonth value to an Integer"))
	  | AtomicHexBinary h ->
	      raise (Query (Cast_Error "Cannot cast a HexBinary value to an Integer"))
	  | AtomicBase64Binary b ->
	      raise (Query (Cast_Error "Cannot cast a Base64Binary value to an Integer"))
	  | AtomicAnyURI a ->
	      raise (Query (Cast_Error "Cannot cast a URI value to an Integer"))
	  | AtomicQName q ->
	      raise (Query (Cast_Error "Cannot cast a QName value to an Integer"))
	  | AtomicNOTATION n ->
	      raise (Query (Cast_Error "Cannot cast a NOTATION value to an Integer"))
	  | AtomicInteger i ->
	      i
	  | AtomicInt i ->
	      Decimal._integer_of_int i
	  | AtomicYearMonthDuration t ->
	      raise (Query (Cast_Error "Cannot cast a yearMonthDuration value to a Integer"))
	  | AtomicDayTimeDuration t ->
	      raise (Query (Cast_Error "Cannot cast a dayTimeDuration value to a Integer"))
	  | AtomicUntyped t ->
	      (* Note: Should we make a difference between string and text ?  - Jerome *)
	      integer_of_untyped t
	end
      in AtomicInteger result_value
  | ATInt ->
      let result_value = 
	begin
	  match bv with
	  | AtomicString s ->
	      (* Note: Should we make a difference between string and text ?  - Jerome *)
	      int_of_untyped s
	  | AtomicBoolean b ->
	      raise (Query (Cast_Error "Cannot cast a Boolean value to an Int"))
	  | AtomicDecimal d ->
	      Decimal._int_of_decimal d
	  | AtomicFloat f ->
	      raise (Query (Cast_Error "Cannot cast a Float value to an Int"))
	  | AtomicDouble d ->
	      raise (Query (Cast_Error "Cannot cast a Double value to an Int"))
	  | AtomicDuration d ->
	      raise (Query (Cast_Error "Cannot cast a Duration value to an Int"))
	  | AtomicDateTime d ->
	      raise (Query (Cast_Error "Cannot cast a DateTime value to an Int"))
	  | AtomicTime t ->
	      raise (Query (Cast_Error "Cannot cast a Time value to an Int"))
	  | AtomicDate d ->
	      raise (Query (Cast_Error "Cannot cast a Date value to an Int"))
	  | AtomicGYearMonth g ->
	      raise (Query (Cast_Error "Cannot cast a GYearMonth value to an Int"))
	  | AtomicGYear g ->
	      raise (Query (Cast_Error "Cannot cast a GYear value to an Int"))
	  | AtomicGMonthDay g ->
	      raise (Query (Cast_Error "Cannot cast a GMonthDay value to an Int"))
	  | AtomicGDay g ->
	      raise (Query (Cast_Error "Cannot cast a GMonthGDay value to an Int"))
	  | AtomicGMonth g ->
	      raise (Query (Cast_Error "Cannot cast a GMonth value to an Int"))
	  | AtomicHexBinary h ->
	      raise (Query (Cast_Error "Cannot cast a HexBinary value to an Int"))
	  | AtomicBase64Binary b ->
	      raise (Query (Cast_Error "Cannot cast a Base64Binary value to an Int"))
	  | AtomicAnyURI a ->
	      raise (Query (Cast_Error "Cannot cast a URI value to an Int"))
	  | AtomicQName q ->
	      raise (Query (Cast_Error "Cannot cast a QName value to an Int"))
	  | AtomicNOTATION n ->
	      raise (Query (Cast_Error "Cannot cast a NOTATION value to an Int"))
	  | AtomicInteger i ->
	      Decimal._int_of_integer i
	  | AtomicInt i ->
	      i
	  | AtomicYearMonthDuration t ->
	      raise (Query (Cast_Error "Cannot cast a yearMonthDuration value to a Int"))
	  | AtomicDayTimeDuration t ->
	      raise (Query (Cast_Error "Cannot cast a dayTimeDuration value to a Int"))
	  | AtomicUntyped t ->
	      (* Note: Should we make a difference between string and text ?  - Jerome *)
	      int_of_untyped t
	end
      in AtomicInt result_value
  | ATYearMonthDuration ->
      raise (Query (Prototype "Cast to a yearMonthDuration not supported"))
  | ATDayTimeDuration ->
      raise (Query (Prototype "Cast to a dateTimeDuration not supported"))
  | ATUntyped ->
      let result_value = 
	begin
	  match bv with
	  | AtomicString s ->
	      s
	  | AtomicBoolean b ->
	      if b then "true" else "false"
	  | AtomicDecimal d ->
	      (* Note: This should return the XML Schema normalized value *)
	      Decimal._string_of_decimal d
	  | AtomicFloat f ->
	      (* Note: This should return the XML Schema normalized value *)
	      string_of_float f
	  | AtomicDouble d ->
	      (* Note: This should return the XML Schema normalized value *)
	      string_of_float d
	  | AtomicDuration d ->
	      d
	  | AtomicDateTime d ->
	      DateTime.string_of_dateTime d
	  | AtomicTime t ->
	      t
	  | AtomicDate d ->
	      d
	  | AtomicGYearMonth g ->
	      g
	  | AtomicGYear g ->
	      g
	  | AtomicGMonthDay g ->
	      g
	  | AtomicGDay g ->
	      g
	  | AtomicGMonth g ->
	      g
	  | AtomicHexBinary h ->
	      h
	  | AtomicBase64Binary b ->
	      b
	  | AtomicAnyURI a ->
	      AnyURI._string_of_uri a
	  | AtomicQName q ->
	      anon_prefix_string q
	  | AtomicNOTATION n ->
              n
	  | AtomicInteger i ->
	      _string_of_integer i
	  | AtomicInt i ->
	      _string_of_int i
	  | AtomicYearMonthDuration t ->
	      t
	  | AtomicDayTimeDuration t ->
	      t
	  | AtomicUntyped t ->
	      t
      end
      in AtomicUntyped result_value
