(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: fn_dynamic.ml,v 1.12 2003/06/30 20:47:38 simeon Exp $ *)

(* Module Fn_dynamic
   Description:
     This module implements evaluation of F&O functions.
*)

open Namespace
open Datatypes
open Xquery_ast
open Datamodel
open Error
open Decimal

(***************************************)
(* Signature of all built-in functions *)
(***************************************)

type bltin = nsenv -> Schema.schema -> _Value list -> _Value


(******************)
(* Help functions *)
(******************)

(* Convert a data model string to a Caml string *)

let get_string i =
  try
    getAtomicString (getAtomicValue i)
  with
  | _ ->
      raise (Query (Wrong_Args "Function parameter is not a string"))

let get_int n =
  try
    getAtomicInt (getAtomicValue n)
  with
  | _ ->
      raise (Query (Wrong_Args "Function parameter is not an int"))

let get_integer n =
  try
    getAtomicInteger (getAtomicValue n)
  with
  | _ ->
      raise (Query (Wrong_Args "Function parameter is an integer"))

let get_decimal n =
  try
    getAtomicDecimal (getAtomicValue n)
  with
  | _ ->
      raise (Query (Wrong_Args "Function parameter is not a decimal"))

let get_float n =
  try
    getAtomicFloat (getAtomicValue n)
  with
  | _ ->
      raise (Query (Wrong_Args "Function parameter is not a float"))

let get_double n =
  try
    getAtomicDouble (getAtomicValue n)
  with
  | _ ->
      raise (Query (Wrong_Args "Function parameter is not a double"))

let get_anyURI n =
  try
    getAtomicAnyURI (getAtomicValue n)
  with
  | _ ->
      raise (Query (Wrong_Args "Function parameter is not a URI"))

let get_QName n =
  try
    getAtomicQName (getAtomicValue n)
  with
  | _ ->
      raise (Query (Wrong_Args "Function parameter is not a QName"))

let get_boolean n =
  try
    getAtomicBoolean (getAtomicValue n)
  with
  | _ ->
      raise (Query (Wrong_Args "Function parameter is not a boolean"))


(***************************)
(* Dummy built-in function *)
(***************************)

let _DUMMY fn nse schema n =
  raise (Query (Prototype ("Function " ^ fn ^ " not supported in "
			   ^ Conf.system ^ " version " ^ Conf.version)))

(****************************)
(* F&O Section 2. Accessors *)
(****************************)

(* F&O Section 2.1 fn:node-kind *)

let _fn_node_kind nse schema n =
  let p1 = Args.get_param1 n in
  let node =
    try
      getNode (sequenceGetSingle p1)
    with
    | _ ->
	raise (Query (Wrong_Args "node-kind built-in function: not a node"))
  in
  let node_kind =
    match nodeKind node with
    | DocumentNodeKind -> "document"
    | ElementNodeKind -> "element"
    | AttributeNodeKind -> "attribute"
    | TextNodeKind -> "text"
    | NamespaceNodeKind -> "namespace"
    | ProcessingInstructionNodeKind -> "processing-instruction"
    | CommentNodeKind -> "comment"
  in
  sequenceBuildSingle (buildAtomicValue (buildAtomicString node_kind))


(* F&O Section 2.2 fn:node-name *)

let _fn_node_name nse schema n =
  let p1 = Args.get_param1 n in
  let node =
    try
      getNode (sequenceGetSingle p1)
    with
    | _ ->
	raise (Query (Wrong_Args "node-name built-in function: not a node"))
  in
  match getNodeName node with 
  | Some qn ->
      let anon_sym = Sym.anon_symbol qn in
      sequenceBuildSingle (buildAtomicValue (buildAtomicQName anon_sym))
  | None ->
      sequenceEmpty


(* F&O Section 2.3 fn:string *)

let _fn_string nse schema n =
  let p1 = Args.get_param1 n in
  try
    if (sequenceIsEmpty p1) then 
      sequenceBuildSingle (buildAtomicValue (buildAtomicString ""))
    else
      let i = (sequenceGetSingle p1) in
      if (isNode i) then 
	let n = getNode i in
	let string_value = nodeStringValue n in
	sequenceBuildSingle (buildAtomicValue (buildAtomicString string_value))
      else 
	sequenceBuildSingle (buildAtomicValue (buildAtomicString (erase_atomic_value (getAtomicValue i))))
  with
  | _ ->
      raise (Query (Wrong_Args "fn:string function: argument not an item"))

(* F&O Section 2.4 fn:data *)
(*
let _fn_data nse schema n 
*)

let _fn_data nse schema n =
  let p1 = Args.get_param1 n in
  let get_atomic_content item = 
    if (isNode item) then Datamodel.typed_value (getNode item)
    else sequenceBuildSingle item
  in (sequenceConcatMap get_atomic_content p1)

(* F&O Section 2.5 fn:base-uri *)

let _fn_base_uri_zero nse schema n =
  let _ = Args.get_param0 n in
  begin
    let uri = AnyURI.get_base_uri ()
    in sequenceBuildSingle (buildAtomicValue (buildAtomicAnyURI uri))
  end

let _fn_base_uri_one nse schema n =
  let p1 = Args.get_param1 n in
  let node =
    try
      getNode (sequenceGetSingle p1)
    with
    | _ ->
	raise (Query (Wrong_Args "base-uri built-in function: not a node"))
  in
  match (Datamodel.base_uri node) with
  | None -> sequenceEmpty
  | Some uri -> sequenceBuildSingle (buildAtomicValue (buildAtomicAnyURI uri))

(************************)
(* F&O Section 4. Trace *)
(************************)

let _fn_trace nse schema n =
  let (il, s) = Args.get_param2 n in
  let s' = get_string (sequenceGetSingle s) in
  let _ =
    begin
      print_string s';
      Serialization.fserialize_datamodel Format.std_formatter il;
      flush stdout
    end
  in
  il

(*****************************************************)
(* F&O Section 5. Functions and Operators on Numbers *)
(*****************************************************)

(* Note: arithmetic functions are monomorphic in Galax for now... *)
(* xsd:int *)
let _fn_int_add nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_int (sequenceGetSingle p1), get_int (sequenceGetSingle p2)) in
  sequenceBuildSingle (_int (_int_add i1 i2))

let _fn_int_sub nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_int (sequenceGetSingle p1), get_int (sequenceGetSingle p2)) in
  sequenceBuildSingle (_int (_int_sub i1 i2))

let _fn_int_mult nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_int (sequenceGetSingle p1), get_int (sequenceGetSingle p2)) in
  sequenceBuildSingle (_int (_int_mult i1 i2))

let _fn_int_div nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_int (sequenceGetSingle p1), get_int (sequenceGetSingle p2)) in
  sequenceBuildSingle (_int (_int_div i1 i2))

let _fn_int_mod nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_int (sequenceGetSingle p1), get_int (sequenceGetSingle p2)) in
  sequenceBuildSingle (_int (_int_mod i1 i2))

let _fn_int_equal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_int (sequenceGetSingle p1), get_int (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (_int_eq i1 i2))

let _fn_int_nequal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_int (sequenceGetSingle p1), get_int (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (not(_int_eq i1 i2)))

let _fn_int_lt nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_int (sequenceGetSingle p1), get_int (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (_int_lt i1 i2))

let _fn_int_gt nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_int (sequenceGetSingle p1), get_int (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (_int_gt i1 i2))

let _fn_int_le nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_int (sequenceGetSingle p1), get_int (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (_int_le i1 i2))

let _fn_int_ge nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_int (sequenceGetSingle p1), get_int (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (_int_ge i1 i2))


let _fn_int_unary_minus nse schema n =
  let p1 = Args.get_param1 n in
  if (sequenceIsEmpty p1) then sequenceEmpty
  else
  let i1 = (get_int (sequenceGetSingle p1)) in
  sequenceBuildSingle (_int (_int_unary_minus i1))

(* unary plus is a no-op *)
let _fn_int_unary_plus nse schema n =
  let p1 = Args.get_param1 n in
  p1 

(* xsd:integer *)
let _fn_integer_add nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_integer (sequenceGetSingle p1), get_integer (sequenceGetSingle p2)) in
  sequenceBuildSingle (_integer (_integer_add i1 i2))

let _fn_integer_sub nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_integer (sequenceGetSingle p1), get_integer (sequenceGetSingle p2)) in
  sequenceBuildSingle (_integer (_integer_sub i1 i2))

let _fn_integer_mult nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_integer (sequenceGetSingle p1), get_integer (sequenceGetSingle p2)) in
  sequenceBuildSingle (_integer (_integer_mult i1 i2))

let _fn_integer_div nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_integer (sequenceGetSingle p1), get_integer (sequenceGetSingle p2)) in
  sequenceBuildSingle (_integer (_integer_div i1 i2))

let _fn_integer_mod nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_integer (sequenceGetSingle p1), get_integer (sequenceGetSingle p2)) in
  sequenceBuildSingle (_integer (_integer_mod i1 i2))

let _fn_integer_unary_minus nse schema n =
  let p1 = Args.get_param1 n in
  if (sequenceIsEmpty p1) then sequenceEmpty
  else
  let i1 = (get_integer (sequenceGetSingle p1)) in
  sequenceBuildSingle (_integer (_integer_unary_minus i1))

let _fn_integer_equal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_integer (sequenceGetSingle p1), get_integer (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (_integer_eq i1 i2))

let _fn_integer_nequal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_integer (sequenceGetSingle p1), get_integer (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (not(_integer_eq i1 i2)))

let _fn_integer_lt nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_integer (sequenceGetSingle p1), get_integer (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (_integer_lt i1 i2))

let _fn_integer_gt nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_integer (sequenceGetSingle p1), get_integer (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (_integer_gt i1 i2))

let _fn_integer_le nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_integer (sequenceGetSingle p1), get_integer (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (_integer_le i1 i2))

let _fn_integer_ge nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_integer (sequenceGetSingle p1), get_integer (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (_integer_ge i1 i2))

(* unary plus is a no-op *)
let _fn_integer_unary_plus nse schema n =
  let p1 = Args.get_param1 n in
  p1 

(* xsd:decimal *)
let _fn_decimal_add nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_decimal (sequenceGetSingle p1), get_decimal (sequenceGetSingle p2)) in
  sequenceBuildSingle (_decimal(_decimal_add i1 i2))

let _fn_decimal_sub nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1, i2) = (get_decimal (sequenceGetSingle p1), get_decimal (sequenceGetSingle p2)) in
  sequenceBuildSingle (_decimal(_decimal_sub i1 i2))

let _fn_decimal_mult nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_decimal (sequenceGetSingle p1), get_decimal (sequenceGetSingle p2)) in
  sequenceBuildSingle (_decimal(_decimal_mult i1 i2))

let _fn_decimal_div nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_decimal (sequenceGetSingle p1), get_decimal (sequenceGetSingle p2)) in
  sequenceBuildSingle (_decimal(_decimal_div i1 i2))

let _fn_decimal_mod nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_decimal (sequenceGetSingle p1), get_decimal (sequenceGetSingle p2)) in
  sequenceBuildSingle (_decimal(_decimal_mod i1 i2))

let _fn_decimal_unary_minus nse schema n =
  let p1 = Args.get_param1 n in
  if (sequenceIsEmpty p1) then sequenceEmpty
  else
  let i1 = (get_decimal (sequenceGetSingle p1)) in
  sequenceBuildSingle (_decimal(_decimal_unary_minus i1))

let _fn_decimal_equal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_decimal (sequenceGetSingle p1), get_decimal (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (_decimal_eq i1 i2))

let _fn_decimal_nequal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_decimal (sequenceGetSingle p1), get_decimal (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (not(_decimal_eq i1 i2)))

let _fn_decimal_lt nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_decimal (sequenceGetSingle p1), get_decimal (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (_decimal_lt i1 i2))

let _fn_decimal_gt nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_decimal (sequenceGetSingle p1), get_decimal (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (_decimal_gt i1 i2))

let _fn_decimal_le nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_decimal (sequenceGetSingle p1), get_decimal (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (_decimal_le i1 i2))

let _fn_decimal_ge nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_decimal (sequenceGetSingle p1), get_decimal (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (_decimal_ge i1 i2))

(* unary plus is a no-op *)
let _fn_decimal_unary_plus nse schema n =
  let p1 = Args.get_param1 n in
  p1 

(* xsd:float *)
let _fn_float_add nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_float (sequenceGetSingle p1), get_float (sequenceGetSingle p2)) in
  sequenceBuildSingle (_float(i1 +. i2))

let _fn_float_sub nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_float (sequenceGetSingle p1), get_float (sequenceGetSingle p2)) in
  sequenceBuildSingle (_float(i1 -. i2))

let _fn_float_mult nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_float (sequenceGetSingle p1), get_float (sequenceGetSingle p2)) in
  sequenceBuildSingle (_float(i1 *. i2))

let _fn_float_div nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_float (sequenceGetSingle p1), get_float (sequenceGetSingle p2)) in
  sequenceBuildSingle (_float(i1 /. i2))

let _fn_float_mod nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_float (sequenceGetSingle p1), get_float (sequenceGetSingle p2)) in
  sequenceBuildSingle (_float(mod_float i1 i2))

let _fn_float_unary_minus nse schema n =
  let p1 = Args.get_param1 n in
  if (sequenceIsEmpty p1) then sequenceEmpty
  else
  let i1 = (get_float (sequenceGetSingle p1)) in
  sequenceBuildSingle (_float(-.(i1)))

let _fn_float_equal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_float (sequenceGetSingle p1), get_float (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (i1 = i2))

(* NB! Must use not(i1 = i2).  i1 != i2 checks for physical equivalence. *)
let _fn_float_nequal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_float (sequenceGetSingle p1), get_float (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (not(i1 = i2)))

let _fn_float_lt nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_float (sequenceGetSingle p1), get_float (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (i1 < i2))

let _fn_float_gt nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_float (sequenceGetSingle p1), get_float (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (i1 > i2))

let _fn_float_le nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_float (sequenceGetSingle p1), get_float (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (i1 <= i2))

let _fn_float_ge nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_float (sequenceGetSingle p1), get_float (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (i1 >= i2))

(* unary plus is a no-op *)
let _fn_float_unary_plus nse schema n =
  let p1 = Args.get_param1 n in
  p1 

(* xsd:double *)
let _fn_double_add nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_double (sequenceGetSingle p1), get_double (sequenceGetSingle p2)) in
  sequenceBuildSingle (_double(i1 +. i2))

let _fn_double_sub nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_double (sequenceGetSingle p1), get_double (sequenceGetSingle p2)) in
  sequenceBuildSingle (_double(i1 -. i2))

let _fn_double_mult nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_double (sequenceGetSingle p1), get_double (sequenceGetSingle p2)) in
  sequenceBuildSingle (_double(i1 *. i2))

let _fn_double_div nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_double (sequenceGetSingle p1), get_double (sequenceGetSingle p2)) in
  sequenceBuildSingle (_double(i1 /. i2))

let _fn_double_mod nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else
  let (i1,i2) = (get_double (sequenceGetSingle p1), get_double (sequenceGetSingle p2)) in
  sequenceBuildSingle (_double(mod_float i1 i2))

let _fn_double_unary_minus nse schema n =
  let p1 = Args.get_param1 n in
  if (sequenceIsEmpty p1) then sequenceEmpty  
  else
  let i1 = (get_double (sequenceGetSingle p1)) in
  sequenceBuildSingle (_double(-.(i1)))

let _fn_double_equal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_double (sequenceGetSingle p1), get_double (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (i1 = i2))

(* NB! Must use not(i1 = i2).  i1 != i2 checks for physical equivalence. *)
let _fn_double_nequal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_double (sequenceGetSingle p1), get_double (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval(not(i1 = i2)))

let _fn_double_lt nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_double (sequenceGetSingle p1), get_double (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (i1 < i2))

let _fn_double_gt nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_double (sequenceGetSingle p1), get_double (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (i1 > i2))

let _fn_double_le nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_double (sequenceGetSingle p1), get_double (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (i1 <= i2))

let _fn_double_ge nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (i1,i2) = (get_double (sequenceGetSingle p1), get_double (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval  (i1 >= i2))

(* unary plus is a no-op *)
let _fn_double_unary_plus nse schema n =
  let p1 = Args.get_param1 n in
  p1 


(* F&O Section 5.4.1 fn:floor *)

let _fn_floor nse schema n =
  let s  = Args.get_param1 n in
  let f  = get_double (sequenceGetSingle s) in
  let f' = floor f in
  sequenceBuildSingle (_double f')


(* F&O Section 5.4.2 fn:ceiling *)

let _fn_ceiling nse schema n = 
  let s  = Args.get_param1 n in
  let f  = get_double (sequenceGetSingle s) in
  let f' = ceil f in
  sequenceBuildSingle (_double f')


(* F&O Section 5.4.3 fn:round *)

let _fn_round nse schema n = 
  let s  = Args.get_param1 n in
  let f  = get_double (sequenceGetSingle s) in
  let f' = floor (f +. 0.5) in
  sequenceBuildSingle (_double f')


(***************************************)
(* F&O Section 6. Functions on Strings *)
(***************************************)

(* F&O Section 6.3.1 fn:concat *)

let _fn_concat nse schema n =
  let (s1,s2) = Args.get_param2 n in
  let s1' = get_string (sequenceGetSingle s1) in
  let s2' = get_string (sequenceGetSingle s2) in
  sequenceBuildSingle (AtomicValue (buildAtomicString (s1' ^ s2')))

let _fn_string_join nse schema n =
  let (s1, s2) = (Args.get_param2 n) in
  let sl = list_of_sequence s1 in
  let concatenated_string = String.concat (get_string (sequenceGetSingle s2)) (List.map get_string sl) in
  sequenceBuildSingle (AtomicValue (buildAtomicString concatenated_string))

(* F&O Section 6.3.2 fn:starts-with *)

let _fn_starts_with nse schema n = 
  let (s1, s2) = Args.get_param2 n in
  let s1' = get_string (sequenceGetSingle s1) in
  let s2' = get_string (sequenceGetSingle s2) in
  let lens1' = String.length s1' in
  let lens2' = String.length s2' in 
  let b = 
    try 
      let s' = String.sub s1' 0 lens2' in (s' = s2')
    with _ -> false
  in
  (sequenceBuildSingle (boolval b))


(* F&O Section 6.3.3 fn:ends-with *)

let _fn_ends_with nse schema n = 
  let (s1, s2) = Args.get_param2 n in
  let s1' = get_string (sequenceGetSingle s1) in
  let s2' = get_string (sequenceGetSingle s2) in
  let lens1' = String.length s1' in
  let lens2' = String.length s2' in 
  let pos = lens1' - lens2' in 
  let b = 
    try 
      let s' = String.sub s1' pos lens2' in (s' = s2')
    with _ -> false
  in
  (sequenceBuildSingle (boolval b))


(* F&O Section 6.3.4 fn:contains *)

(* Note [MFF]:
     This is very expensive! 
     Should implement a more rational sub-string search ! 
     This code should be using the Str. Caml library 
 *)

(* Note:
     This function is also needed in finding substring-before.
 *)

let rec _aux_contains_at s1 s2 pos =
   let lens1 = String.length s1 in 
   let lens2 = String.length s2 in 
   if (lens2 > lens1)
   then (pos + lens1)
   else 
     let s = String.sub s1 0 lens2 in 
     begin
       if (s = s2)
       then pos
       else _aux_contains_at (String.sub s1 1 (lens1-1)) s2 (pos+1)
     end

let _aux_contains s1 s2 = 
  let pos = _aux_contains_at s1 s2 0 in
  let len1 = String.length s1 in 
  (pos < len1)

let _fn_contains nse schema n = 
  let (s1, s2) = Args.get_param2 n in
  let s1' = get_string (sequenceGetSingle s1) in
  let s2' = get_string (sequenceGetSingle s2) in
  let b = _aux_contains s1' s2' in 
  (sequenceBuildSingle (boolval b))


(* F&O Section 6.3.5 fn:substring *)

let _fn_substring nse schema n =
  let (str, start, len) = Args.get_param3 n in
  let str'    = get_string (sequenceGetSingle str)   in
  let start'  = Decimal._int_of_integer(get_integer (sequenceGetSingle start)) in
  (* Note:
       substring indexes start at position 1 in XQuery, but 0 in Caml.
     - Jerome
   *)
  let start'' = start' - 1 in
  let len'    = Decimal._int_of_integer(get_integer (sequenceGetSingle len))  in
  let sub     =
    try
      String.sub str' start'' len'
    with
    | Invalid_argument _ ->
	raise (Query (Wrong_Args "fn:substring index out of range"))
  in
  sequenceBuildSingle (AtomicValue (buildAtomicString sub))


(* F&O Section 6.3.6 fn:string-length *)

let _fn_string_length nse schema n =
  let s  = Args.get_param1 n in
  let s' = get_string (sequenceGetSingle s) in
  let i = String.length s' in
  sequenceBuildSingle (_integer (Decimal._integer_of_int i))


(* F&O Section 6.3.7 fn:substring-before *)

let _substring_before_s str substr =
  let pos     = _aux_contains_at str substr 0 in
  let len     = String.length str in 
  if (pos < len) then String.sub str 0 pos
  else ""


let _fn_substring_before nse schema n =
  let (str, substr) = Args.get_param2 n in
  let str'    = get_string (sequenceGetSingle str)   in
  let substr' = get_string (sequenceGetSingle substr) in
  let sub     = _substring_before_s str' substr' in
  sequenceBuildSingle (AtomicValue (buildAtomicString sub))


(* F&O Section 6.3.8 fn:substring-after *)

let _substring_after_s str substr = 
  let pos     = _aux_contains_at str substr 0 in
  let len     = String.length str in 
  let len2    = String.length substr in 
  if ((pos+len2) < len) then String.sub str (pos+len2) (len-pos-len2)
  else ""

let _fn_substring_after nse schema n =
  let (str, substr) = Args.get_param2 n in
  let str'    = get_string (sequenceGetSingle str)   in
  let substr' = get_string (sequenceGetSingle substr) in
  let sub     = _substring_after_s str' substr' in
  sequenceBuildSingle (AtomicValue (buildAtomicString sub))

let _op_string_equal nse schema n =
  let (p1, p2) = Args.get_param2 n in
  let s1 = get_string (sequenceGetSingle p1)   in
  let s2 = get_string (sequenceGetSingle p2) in
  sequenceBuildSingle (boolval(string_equal s1 s2))

let _op_string_nequal nse schema n =
  let (p1, p2) = Args.get_param2 n in
  let s1 = get_string (sequenceGetSingle p1)   in
  let s2 = get_string (sequenceGetSingle p2) in
  sequenceBuildSingle (boolval(not(string_equal s1 s2)))

let _op_string_lt nse schema n =
  let (p1, p2) = Args.get_param2 n in
  let s1 = get_string (sequenceGetSingle p1)   in
  let s2 = get_string (sequenceGetSingle p2) in
  sequenceBuildSingle (boolval(string_lt s1 s2))

let _op_string_le nse schema n =
  let (p1, p2) = Args.get_param2 n in
  let s1 = get_string (sequenceGetSingle p1)   in
  let s2 = get_string (sequenceGetSingle p2) in
  sequenceBuildSingle (boolval(string_lteq s1 s2))

let _op_string_gt nse schema n =
  let (p1, p2) = Args.get_param2 n in
  let s1 = get_string (sequenceGetSingle p1)   in
  let s2 = get_string (sequenceGetSingle p2) in
  sequenceBuildSingle (boolval(string_gt s1 s2))

let _op_string_ge nse schema n =
  let (p1, p2) = Args.get_param2 n in
  let s1 = get_string (sequenceGetSingle p1)   in
  let s2 = get_string (sequenceGetSingle p2) in
  sequenceBuildSingle (boolval(string_gteq s1 s2))

(* F&O Section 6.3.9 fn:normalize-space *)

let cut_leading_and_trailing_spaces s =
  let rec cut_leading_spaces str =
  begin
    try
      if ((String.get str 0) = ' ') then
	let len = String.length str in
	let s' = String.sub str 1 (len-1) in
	cut_leading_spaces s'
      else str
    with
    | Invalid_argument msg -> raise (Query (Wrong_Args (msg ^ ": Invalid argument found while cutting the leading spaces.")))
  end
  in
  let rec cut_trailing_spaces str =
  begin
    try
      let len = String.length str in
      if ((String.get str (len-1)) = ' ') then
	let len' = len -1 in
	let s' = String.sub str 0 len' in
	cut_trailing_spaces s'
      else str
    with
    | Invalid_argument msg -> raise (Query (Wrong_Args (msg ^ ": Invalid argument found while cutting the trailing spaces.")))
  end
  in
  let s'  = cut_leading_spaces  s  in
  let s'' = cut_trailing_spaces s' in
  s''

let rec remove_duplicate_spaces_aux str i =
  let len = String.length str in
  if (i+2) > len then str
  else
    let ch1 = String.get str i in
    let ch2 = String.get str (i+1) in
    begin
      if((ch1 = ' ') && (ch2 = ' ')) 
      then
	begin
	  remove_duplicate_spaces_aux ((String.sub str 0 i) ^ (String.sub str (i+1) (len - (i+1)))) i
	end
      else 
	remove_duplicate_spaces_aux str (i+1)
    end

let remove_duplicate_spaces str = 
  remove_duplicate_spaces_aux str 0

let _fn_normalize_space nse schema n =
  let s  = Args.get_param1 n in
  let s'   = get_string (sequenceGetSingle s) in
  let s''  = cut_leading_and_trailing_spaces s' in
  let str =  remove_duplicate_spaces s'' in 
  sequenceBuildSingle (AtomicValue (buildAtomicString str))


(* F&O Section 6.3.11 fn:upper-case *)

let _fn_upper_case nse schema n =
  let s   = Args.get_param1 n in
  let s'  = get_string (sequenceGetSingle s) in
  let s'' = String.uppercase s' in
  sequenceBuildSingle (AtomicValue (buildAtomicString s''))


(* F&O Section 6.3.12 fn:lower-case *)

let _fn_lower_case nse schema n =
  let s   = Args.get_param1 n in
  let s'  = get_string (sequenceGetSingle s) in
  let s'' = String.lowercase s' in
  sequenceBuildSingle (AtomicValue (buildAtomicString s''))


(* F&O Section 6.3.13 fn:translate *)

let _translate_aux str mapStr transStr =
  let char_stream = Stream.of_string str in
  let buffer = Buffer.create 0 in
  let new_char c = Buffer.add_char buffer c in
  let translate_char c =
    try
      let index = String.index mapStr c in
      begin
	try
	  new_char (String.get transStr index)
	with
	| Invalid_argument _ ->
	    ()
      end
    with
    | Not_found ->
	new_char c
  in
  while not(Stream.peek char_stream = None) do
    translate_char (Stream.next char_stream)
  done;
  Buffer.contents buffer
    

let _fn_translate nse schema n =
  let (str, mapStr, transStr) = Args.get_param3 n in
  let str'       = get_string (sequenceGetSingle str)      in
  let mapStr'    = get_string (sequenceGetSingle mapStr)   in
  let transStr'  = get_string (sequenceGetSingle transStr) in
  let result     = _translate_aux str' mapStr' transStr'  in
  sequenceBuildSingle (AtomicValue (buildAtomicString result))


(* F&O Section 6.3.14 fn:string-pad *)

let _fn_string_pad nse schema n  =
  let (str, d, str1) = Args.get_param3 n in
  let str'   = get_string (sequenceGetSingle str)  in
  let d'     = _int_of_integer(get_integer (sequenceGetSingle d)) in
  let str1'  = get_string (sequenceGetSingle str1) in
  if (String.length str1' != 1) then
    raise (Query (Wrong_Args "The 3rd argument is not a character"))
  else
    let str'' = 
      begin
	  (String.make d' (String.get str1' 0)) ^ str'
(* F&O only supports pad at beginning
	else
 	  str' ^ (String.make d' (String.get str1' 0)) 
*)
      end
    in
    sequenceBuildSingle (AtomicValue (buildAtomicString str''))


(******************************************************)
(* F&O Section 7. Functions and Operators on Booleans *)
(******************************************************)

(* F&O Section 7.1.1 fn:true *)

let _fn_true nse schema n =
  sequenceBuildSingle (boolval (true))


(* F&O Section 7.1.3 fn:false *)

let _fn_false nse schema n =
  sequenceBuildSingle (boolval (false))
 

(* F&O Section 7.3.1 fn:not *)

let _fn_not nse schema n =
  let p1 = Args.get_param1 n in
  let b1 = get_boolean (sequenceGetSingle p1) in
  sequenceBuildSingle (boolval (not(b1)))

(* Note:
     The two following are missing from F&O [August 16th draft].
   - Jerome
 *)

let _op_or nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (b1,b2) = (get_boolean (sequenceGetSingle p1),
		get_boolean (sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (b1 or b2))

let _op_and nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (b1,b2) = (get_boolean (sequenceGetSingle p1),
		 get_boolean(sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (b1 & b2))

let _op_boolean_equal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (b1,b2) = (get_boolean(sequenceGetSingle p1),
		 get_boolean(sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (b1 = b2))

let _op_boolean_nequal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (b1,b2) = (get_boolean(sequenceGetSingle p1),
		 get_boolean(sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (b1 != b2))

let _op_boolean_lt nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (b1,b2) = (get_boolean (sequenceGetSingle p1),
		 get_boolean(sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (b1 < b2))

let _op_boolean_gt nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (b1,b2) = (get_boolean(sequenceGetSingle p1),
		 get_boolean(sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (b1 > b2))

let _op_boolean_le nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (b1,b2) = (get_boolean(sequenceGetSingle p1),
		 get_boolean(sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (b1 <= b2))

let _op_boolean_ge nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (b1,b2) = (get_boolean(sequenceGetSingle p1),
		 get_boolean(sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (b1 >= b2))

(**************************************)
(* F&O Section 9. Functions on QNames *)
(**************************************)

(* F&O Section 9.1.1 fn:expanded-QName *)

let _fn_expanded_qname nse schema n =
  let (str1, str2) = Args.get_param2 n in
  let uri = get_string (sequenceGetSingle str1)
  and localname = get_string (sequenceGetSingle str2) in
  let qname = Sym.anon_symbol (NSUri uri, localname) in
  sequenceBuildSingle (buildAtomicValue (buildAtomicQName qname))


(* F&O Section 9.2.2 fn:get-local-name-from-QName *)

let _fn_get_local_name_from_qname nse schema n =
  let qname1 = Args.get_param1 n in
  let qname = get_QName (sequenceGetSingle qname1) in
  let (_,localname) = Sym.anon_name qname in
  sequenceBuildSingle (buildAtomicValue (buildAtomicString localname))


(* F&O Section 9.2.2 fn:get-namespace-from-QName *)

let _fn_get_namespace_from_qname nse schema n =
  let qname1 = Args.get_param1 n in
  let qname = get_QName (sequenceGetSingle qname1) in
  let (uri,_) = Sym.anon_name qname in
  match uri with
  | NSUri urit ->
(*      let anyURI = anyURI_of_untyped urit in *)
      sequenceBuildSingle (buildAtomicValue (buildAtomicString urit))
  | _ ->
      raise (Query (Wrong_Args "Qname contains a wildcard URI"))

let _op_QName_equal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (q1,q2) = (get_QName(sequenceGetSingle p1),
		 get_QName(sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (Datatypes.qname_equal q1 q2))

let _op_QName_nequal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (q1,q2) = (get_QName(sequenceGetSingle p1),
		 get_QName(sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (not(Datatypes.qname_equal q1 q2)))


(******************************************************)
(* F&O Section 10. Functions and Operators for anyURI *)
(******************************************************)

(* F&O Section 10.1.1 fn:resolve-uri *)

let _fn_resolve_uri nse schema n = 
  let (uri1, uri2) = Args.get_param2 n in
  let base_uri = get_anyURI (sequenceGetSingle uri1)
  and relative_uri = get_anyURI (sequenceGetSingle uri2) in
  let resolved_uri = AnyURI._uri_resolve base_uri relative_uri in
  sequenceBuildSingle (buildAtomicValue (buildAtomicAnyURI resolved_uri))

let _op_anyURI_equal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (q1,q2) = (get_anyURI(sequenceGetSingle p1),
		 get_anyURI(sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (Datatypes.anyURI_equal q1 q2))

let _op_anyURI_nequal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let (q1,q2) = (get_anyURI(sequenceGetSingle p1),
		 get_anyURI(sequenceGetSingle p2)) in
  sequenceBuildSingle (boolval (not(Datatypes.anyURI_equal q1 q2)))

(****************************************************)
(* F&O Section 13. Functions and Operators on Nodes *)
(****************************************************)

(* F&O Section 13.1.1 fn:name *)

let _fn_name nse schema n =
  let p1 = Args.get_param1 n in
  try
    let name_string =
      if (sequenceIsEmpty p1) then 
	""
      else
	let node = getNode (sequenceGetSingle p1) in
	getNodeQName node
    in
    sequenceBuildSingle (buildAtomicValue (buildAtomicString name_string))
  with
  | _ ->
      raise (Query (Wrong_Args "fn:name function: argument not a node"))


(* F&O Section 13.1.2 fn:local-name *)

let _fn_local_name nse schema n =
  let p1 = Args.get_param1 n in
  try
    let local_name_string =
      if (sequenceIsEmpty p1) then 
	""
      else
	let node =
	  getNode (sequenceGetSingle p1)
	in
	match getNodeName node with 
	| Some (uri,local_name_string) ->
	    local_name_string
	| None ->
	    ""
  in
  sequenceBuildSingle (buildAtomicValue (buildAtomicString local_name_string))
  with
  | _ ->
      raise (Query (Wrong_Args "fn:local-name function: argument not a node"))


(* F&O Section 13.1.3 fn:namespace-uri *)

let _fn_namespace_uri nse schema n =
  let p1 = Args.get_param1 n in
  try
    let local_name_string =
      if (sequenceIsEmpty p1) then 
	""
      else
	let node =
	  getNode (sequenceGetSingle p1)
	in
	match getNodeName node with 
	| Some (NSUri uri,local_name_string) ->
	    uri
	| Some (NSWildcardUri,local_name_string) ->
	    raise (Query (Wrong_Args "Wildcard namespace should not appear in node"))
	| None ->
	    ""
  in
  sequenceBuildSingle (buildAtomicValue (buildAtomicString local_name_string))
  with
  | _ ->
      raise (Query (Wrong_Args "fn:namespace-uri function: argument not a node"))

(* F&O Section 13.1.3 fn:lang *)

let get_lang node =
  match nodeKind node with
  | DocumentNodeKind ->
      None
  | ElementNodeKind ->
      elemLang (getElementNode node)
  | AttributeNodeKind ->
      elemLang (getElementNode (sequenceGetSingle (nodeParent node)))
  | TextNodeKind ->
      elemLang (getElementNode (sequenceGetSingle (nodeParent node)))
  | NamespaceNodeKind ->
      elemLang (getElementNode (sequenceGetSingle (nodeParent node)))
  | ProcessingInstructionNodeKind ->
      elemLang (getElementNode (sequenceGetSingle (nodeParent node)))
  | CommentNodeKind ->
      elemLang (getElementNode (sequenceGetSingle (nodeParent node)))


let _fn_lang nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let node =
    try
      getNode (sequenceGetSingle p1)
    with
    | _ ->
	raise (Query (Wrong_Args "base-uri built-in function: not a node"))
  in
  let textlang = get_string (sequenceGetSingle p2) in
  let oactuallang = get_lang node in
  let result_bool =
    match oactuallang with
    | None -> false
    | Some actuallang ->
	if (String.lowercase actuallang) = (String.lowercase textlang)
	then
	  true
	else
	  let comparelang =
	    try
	      fst (Gmisc.split_left_on_char actuallang '-')
	    with
	    | _ -> actuallang
	  in
	  (String.lowercase comparelang) = (String.lowercase textlang)
  in
  (sequenceBuildSingle (boolval result_bool))

(* F&O Section 13.1.6 op:node-equal *)

let _op_node_equal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2)
  then
    sequenceEmpty
  else
  sequenceBuildSingle (boolval (nodeEqual (getNode (sequenceGetSingle p1)) (getNode (sequenceGetSingle p2))))


(* F&O Section 13.1.8 op:node-before *)

let _fn_node_before nse schema n =
  let (p1,p2) = Args.get_param2 n in
  sequenceBuildSingle (boolval (nodePrecedes (getNode (sequenceGetSingle p1)) (getNode (sequenceGetSingle p2))))


(* F&O Section 13.1.9 op:node-after *)

let _fn_node_after nse schema n =
  let (p1,p2) = Args.get_param2 n in
  sequenceBuildSingle (boolval (nodeFollows (getNode (sequenceGetSingle p1)) (getNode (sequenceGetSingle p2))))


(* F&O Section 13.1.11 fn:root *)

let rec get_root n =
  let parent = nodeParent n in
  if (sequenceIsEmpty parent)
  then
    n
  else
    get_root (sequenceGetSingle parent)

let _fn_root nse schema n =
  let p1 = Args.get_param1 n in
  let node =
    try
      getNode (sequenceGetSingle p1)
    with
    | _ ->
	raise (Query (Wrong_Args "fn:deep-copy: not a node"))
  in
  sequenceBuildSingle (buildNode (get_root node))


(********************************************************)
(* F&O Section 14. Functions and Operators on Sequences *)
(********************************************************)

(* F&O Section 14.1.1 op:to *)

let rec build_integer_sequence i1 i2 =
  if Decimal._integer_eq i1 i2
  then
    _integer i1 :: []
  else if Decimal._integer_lt i1 i2 then
    (_integer i1) :: (build_integer_sequence (Decimal._integer_add i1 (_integer_of_int 1)) i2)
  else
    (_integer i1) :: (build_integer_sequence (Decimal._integer_sub i1 (_integer_of_int 1)) i2)

let _op_to nse schema n =
  let (di1,di2) = Args.get_param2 n in
  let i1 = get_integer (sequenceGetSingle di1)
  and i2 = get_integer (sequenceGetSingle di2) in
  sequence_of_list (build_integer_sequence i1 i2)


(* F&O Section 14.2.1 fn:boolean 

  fn:boolean returns false if its argument is:
  * The empty sequence
  * The singleton boolean value false
  * The singleton string value ""
  * A singleton numeric value that is numerically equal to zero or a float or double equal to NaN
  Otherwise, it is true.

*)

let _fn_boolean nse schema n =
  let p1 = Args.get_param1 n in
  let bv = 
    if (sequenceIsEmpty p1) then 
      (boolval false)
    else if (sequenceIsSingle p1) then 
      let item = sequenceGetSingle p1 in
      if (isAtomicValue item) then
	let a = (getAtomicValue item) in
	match a with
	| AtomicBoolean _ -> item
	| AtomicString _  -> boolval (not(atomic_value_eq a (AtomicString "")))
	| AtomicInt     _ -> boolval (not(atomic_value_eq a (AtomicInt 0)))
	| AtomicInteger _ -> boolval (not(atomic_value_eq a (AtomicInteger _integer_zero)))
	| AtomicDecimal _ -> boolval (not(atomic_value_eq a (AtomicDecimal _decimal_zero)))
	| AtomicFloat _   -> boolval (not(atomic_value_eq a (AtomicFloat 0.0)))
	| AtomicDouble _  -> boolval (not(atomic_value_eq a (AtomicDouble 0.0)))
	| _ -> (boolval true)
      else 
	(boolval true)
    else
      (boolval true)
  in sequenceBuildSingle bv

(* F&O Section 14.2.3 fn:item-at *)

let _fn_item_at nse schema n =
  let (ds,di) = Args.get_param2 n in
  let s = list_of_sequence ds
  and i = _int_of_integer(get_integer (sequenceGetSingle di)) in
  (* If s is empty, then return empty *)
  if s = []
  then
    sequenceEmpty
  else
    let ns =
      try
	(List.nth s (i-1))
      with
      | _ ->
	  raise (Query (Wrong_Args ("Invalid position " ^ (string_of_int i) ^ " in sequence")))
    in
    sequenceBuildSingle ns


(* F&O Section 14.2.4 fn:index-of *)

(* Note:
     The function index-of atomizes its arguments and applies value
     equality to identify index of the second atomic-value argument in
     the first atomic-value-sequence argument.
   -mff *)

let _fn_index_of nse schema n = 
  let (p1,p2) = Args.get_param2 n in
  let seq = list_of_sequence(sequenceMap getAtomicValue p1) in
  let item = getAtomicValue (sequenceGetSingle p2) in
  let eq_fn (idx, idxlist) av = 
    if (atomic_value_eq item av) then (idx+1, idxlist@[idx])
    else (idx+1, idxlist)
  in 
  let (idx', idxlist) = List.fold_left eq_fn (1, []) seq 
  in sequence_of_list(List.map (fun i -> buildAtomicValue(buildAtomicInteger (_integer_of_int i))) idxlist)

(* F&O Section 14.2.5 fn:empty *)

let _fn_empty nse schema n =
  let p1 = Args.get_param1 n in
  sequenceBuildSingle (boolval (sequenceIsEmpty p1))


(* F&O Section 14.2.6 fn:exists *)

let _fn_exists nse schema n =
  let p1 = Args.get_param1 n in
  sequenceBuildSingle (boolval (not(sequenceIsEmpty p1)))


(* F&O Section 14.2.7 fn:distinct-nodes *)

let remove x eq l =
  sequenceFilter (fun y -> not(eq x y)) l

let rec remove_dup eq l =
  if sequenceIsEmpty l
  then sequenceEmpty
  else
    let x = sequenceHead l
    and l' = sequenceTail l in
    sequenceCons x (remove_dup eq (remove x eq l')) 

let _fn_distinct_nodes nse schema n =
  let p1 = Args.get_param1 n in
  let node_forest = sequenceMap getNode p1 in
  let res = remove_dup nodeEqual node_forest
  in sequenceMap buildNode res


(* F&O Section 14.2.8 fn:distinct-values *)

let remove x eq l =
  sequenceFilter (fun y -> not(eq x y)) l

let rec remove_dup eq l =
  if sequenceIsEmpty l
  then sequenceEmpty
  else
    let x = sequenceHead l
    and l' = sequenceTail l in
    sequenceCons x (remove_dup eq (remove x eq l')) 

let _fn_distinct_value nse schema n =
  let p1 = Args.get_param1 n in
  let res = remove_dup item_eq p1
  in res


(* F&O Section 14.2.9 fn:insert *)

let rec apply_insert n target inserts =
  if n <= 0
  then
    sequenceAppend inserts (sequence_of_list target)
  else
    match target with
    | [] ->
	inserts
    | node :: target' ->
	sequenceCons node (apply_insert (n-1) target' inserts)

let _fn_insert nse schema n =
  let (p1,p2,p3) = Args.get_param3 n in
  let i2 = _int_of_integer(get_integer (sequenceGetSingle p2)) in
  let target_list = list_of_sequence p1 in
  apply_insert i2 target_list p3


(* F&O Section 14.2.10 fn:remove *)

let rec apply_remove n target =
  if n <= 0
  then
    sequence_of_list target
  else
    match target with
    | [] ->
	sequenceEmpty
    | node :: target' ->
	if n = 1
	then
	  sequence_of_list target'
	else
	  sequenceCons node (apply_remove (n-1) target')

let _fn_remove nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let i2 = _int_of_integer(get_integer (sequenceGetSingle p2)) in
  let target_list = list_of_sequence p1 in
  apply_remove i2 target_list


(* F&O Section 14.2.11 fn:subsequence *)

let rec keep_subsequence l target =
  if l <= 0 then
    sequenceEmpty
  else
    match target with
    | [] ->
	sequenceEmpty
    | node :: target' ->
	sequenceCons node (keep_subsequence (l-1) target')

let rec apply_subsequence n l target =
  if n <= 1
  then
    keep_subsequence l target
  else
    match target with
    | [] ->
	sequenceEmpty
    | node :: target' ->
	apply_subsequence (n-1) l target'

let _fn_subsequence_2 nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let i2 = _int_of_integer(get_integer (sequenceGetSingle p2)) in
  let target_list = list_of_sequence p1 in
  let total_length = List.length target_list in
  apply_subsequence i2 total_length target_list

let _fn_subsequence_3 nse schema n =
  let (p1,p2,p3) = Args.get_param3 n in
  let i2 = _int_of_integer(get_integer (sequenceGetSingle p2)) in
  let i3 = _int_of_integer(get_integer (sequenceGetSingle p3)) in
  let target_list = list_of_sequence p1 in
  apply_subsequence i2 i3 target_list


(* F&O Section 14.3.1 fn:sequence-deep-equal *)

let _fn_sequence_deep_equal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  sequenceBuildSingle (boolval (value_eq p1 p2))


(* F&O Section 14.3.2 fn:sequence-node-equal *)

let _fn_sequence_node_equal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let f1 = list_of_sequence (sequenceMap getNode p1) in
  let f2 = list_of_sequence (sequenceMap getNode p2) in
  if (List.length f1 != List.length f2)
  then
    sequenceBuildSingle (boolval false)
  else
    let pair_list = List.combine f1 f2 in
    sequenceBuildSingle (boolval (List.for_all (fun (x,y) -> nodeEqual x y) pair_list))

(* F&O Section 14.3.3 op:union *)

(* Note:
     Those operators do not perform sorting by document order yet.
   - Jerome *)

let _op_union nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let nf1 = sequenceMap getNode p1 in
  let nf2 = sequenceMap getNode p2 in
  let big_seq = sequenceAppend nf1 nf2 in
  (* following if-statement checks if the commandline 
   * options say something about duplicate elimination
   * or sorting by document order
   *)
  if (!Conf.do_sbdo) then  (* ddo turned ON -- default *)
    let unordered_res = remove_dup nodeEqual big_seq in
    sequenceMap buildNode unordered_res
  else                      (* ddo turned off *)
    let unordered_res = big_seq in
    sequenceMap buildNode unordered_res

(* F&O Section 14.3.4 op:intersect *)

let compare eq x y =
  eq x y

let rec apply_intersect eq nf1 nf2 =
  match nf1 with
  | [] -> sequenceEmpty
  | node :: nf1' ->
      let rest_intersect = apply_intersect eq nf1' nf2 in
      if (List.exists (compare eq node) nf2)
      then
	sequenceCons node rest_intersect
      else
	rest_intersect

let _op_intersect nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let nf1 = list_of_sequence (sequenceMap getNode p1) in
  let nf2 = list_of_sequence (sequenceMap getNode p2) in
  let intersect_forest = apply_intersect nodeEqual nf1 nf2 in
  sequenceMap buildNode intersect_forest


(* F&O Section 14.3.5 op:except *)

let rec apply_except eq nf1 nf2 =
  match nf1 with
  | [] -> sequenceEmpty
  | node :: nf1' ->
      let rest_except = apply_except eq nf1' nf2 in
      if (List.exists (compare eq node) nf2)
      then
	rest_except
      else
	sequenceCons node rest_except

let _op_except nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let nf1 = list_of_sequence (sequenceMap getNode p1) in
  let nf2 = list_of_sequence (sequenceMap getNode p2) in
  let except_forest = apply_except nodeEqual nf1 nf2 in
  sequenceMap buildNode except_forest


(* F&O Section 14.4.1 fn:count *)

let _fn_count nse schema n =
  let p1 = Args.get_param1 n in
  sequenceBuildSingle (_integer (Decimal._integer_of_int(sequenceLength p1)))

(* xsd:int *)
let _fn_avg_int nse schema n =
  let p1 = Args.get_param1 n in
  let allints = sequenceMap get_int p1 in
  if (sequenceIsEmpty allints)
  then
    sequenceEmpty
  else
    let sum = sequenceFold_left (fun x0 x1 -> x0 + x1) 0 allints in
    sequenceBuildSingle (_int (sum/(sequenceLength allints)))

let _fn_max_int nse schema n =
  let p1 = Args.get_param1 n in
  let allints = sequenceMap get_int p1 in
  if (sequenceIsEmpty allints)
  then
    sequenceEmpty
  else
    let start_max = sequenceHead allints in
    let max =
      sequenceFold_left (fun x0 x1 -> if x0 <= x1 then x1 else x0) start_max allints
    in
    sequenceBuildSingle (_int max)

let _fn_min_int nse schema n =
  let p1 = Args.get_param1 n in
  let allints = sequenceMap get_int p1 in
  if (sequenceIsEmpty allints)
  then
    sequenceEmpty
  else
    let start_min = sequenceHead allints in
    let min =
      sequenceFold_left (fun x0 x1 -> if x0 < x1 then x0 else x1) start_min allints
    in
    sequenceBuildSingle (_int min)

let _fn_sum_int nse schema n =
  let p1 = Args.get_param1 n in
  let allints = sequenceMap get_int p1 in
  if (sequenceIsEmpty allints)
  then
    sequenceBuildSingle(_int _int_zero)
  else
    let sum = sequenceFold_left (fun x0 x1 -> x0 + x1) 0 allints in
    sequenceBuildSingle (_int sum)

(* xsd:integer *)
let _fn_avg_integer nse schema n =
  let p1 = Args.get_param1 n in
  let allintegers = sequenceMap get_integer p1 in
  if (sequenceIsEmpty allintegers)
  then
    sequenceEmpty 
  else
    let sum = sequenceFold_left (fun x0 x1 -> _integer_add x0 x1) _integer_zero allintegers in
    sequenceBuildSingle (_integer (_integer_div sum (_integer_of_int (sequenceLength allintegers))))

let _fn_max_integer nse schema n =
  let p1 = Args.get_param1 n in
  let allintegers = sequenceMap get_integer p1 in
  if (sequenceIsEmpty allintegers)
  then
    sequenceEmpty
  else
    let start_max = sequenceHead allintegers in
    let max =
      sequenceFold_left (fun x0 x1 -> if _integer_le x0 x1 then x1 else x0) start_max allintegers
    in
    sequenceBuildSingle (_integer max)

let _fn_min_integer nse schema n =
  let p1 = Args.get_param1 n in
  let allintegers = sequenceMap get_integer p1 in
  if (sequenceIsEmpty allintegers)
  then
    sequenceEmpty
  else
    let start_min = sequenceHead allintegers in
    let min =
      sequenceFold_left (fun x0 x1 -> if _integer_lt x0 x1 then x0 else x1) start_min allintegers
    in
    sequenceBuildSingle (_integer min)

let _fn_sum_integer nse schema n =
  let p1 = Args.get_param1 n in
  let allintegers = sequenceMap get_integer p1 in
  if (sequenceIsEmpty allintegers)
  then
    sequenceBuildSingle(_integer _integer_zero)
  else
    let sum = sequenceFold_left (fun x0 x1 -> _integer_add x0 x1) _integer_zero allintegers in
    sequenceBuildSingle (_integer sum)

(* xsd:decimal *)
let _fn_avg_decimal nse schema n =
  let p1 = Args.get_param1 n in
  let alldecimals = sequenceMap get_decimal p1 in
  if (sequenceIsEmpty alldecimals)
  then
    sequenceEmpty
  else
  let sum = sequenceFold_left (fun x0 x1 -> _decimal_add x0 x1) _decimal_zero alldecimals in
  sequenceBuildSingle (_decimal (_decimal_div sum (_decimal_of_int (sequenceLength alldecimals))))

let _fn_max_decimal nse schema n =
  let p1 = Args.get_param1 n in
  let alldecimals = sequenceMap get_decimal p1 in
  if (sequenceIsEmpty alldecimals)
  then
    sequenceEmpty
  else
    let start_max = sequenceHead alldecimals in
    let max =
      sequenceFold_left (fun x0 x1 -> if _decimal_le x0 x1 then x1 else x0) start_max alldecimals
    in
    sequenceBuildSingle (_decimal max)

let _fn_min_decimal nse schema n =
  let p1 = Args.get_param1 n in
  let alldecimals = sequenceMap get_decimal p1 in
  if (sequenceIsEmpty alldecimals)
  then
    sequenceEmpty
  else
    let start_min = sequenceHead alldecimals in
    let min =
      sequenceFold_left (fun x0 x1 -> if _decimal_lt x0 x1 then x0 else x1) start_min alldecimals
    in
    sequenceBuildSingle (_decimal min)

let _fn_sum_decimal nse schema n =
  let p1 = Args.get_param1 n in
  let alldecimals = sequenceMap get_decimal p1 in
  if (sequenceIsEmpty alldecimals)
  then
    sequenceBuildSingle(_decimal _decimal_zero)
  else
  let sum = sequenceFold_left (fun x0 x1 -> _decimal_add x0 x1) _decimal_zero alldecimals in
  sequenceBuildSingle (_decimal sum)

(* xsd:float *)
let _fn_avg_float nse schema n =
  let p1 = Args.get_param1 n in
  let allfloats = sequenceMap get_float p1 in
  if (sequenceIsEmpty allfloats)
  then
    sequenceEmpty
  else
    let sum = sequenceFold_left (fun x0 x1 -> x0 +. x1) 0.0 allfloats in
    sequenceBuildSingle (_float (sum/.(float_of_int (sequenceLength allfloats))))

let _fn_max_float nse schema n =
  let p1 = Args.get_param1 n in
  let allfloats = sequenceMap get_float p1 in
  if (sequenceIsEmpty allfloats)
  then
    sequenceEmpty
  else
    let start_max = sequenceHead allfloats in
    let max = sequenceFold_left (fun x0 x1 -> if x0 <= x1 then x1 else x0) start_max allfloats in
    sequenceBuildSingle (_float max)

let _fn_min_float nse schema n =
  let p1 = Args.get_param1 n in
  let allfloats = sequenceMap get_float p1 in
  if (sequenceIsEmpty allfloats)
  then
    sequenceEmpty
  else
    let start_min = sequenceHead allfloats in
    let min = sequenceFold_left (fun x0 x1 -> if x0 < x1 then x0 else x1) start_min allfloats in
    sequenceBuildSingle (_float min)

let _fn_sum_float nse schema n =
  let p1 = Args.get_param1 n in
  let allfloats = sequenceMap get_float p1 in
  if (sequenceIsEmpty allfloats)
  then
    sequenceBuildSingle (_float 0.0)
  else
    let sum = sequenceFold_left (fun x0 x1 -> x0 +. x1) 0.0 allfloats in
    sequenceBuildSingle (_float sum)

(* xsd:double *)
let _fn_avg_double nse schema n =
  let p1 = Args.get_param1 n in
  let alldoubles = sequenceMap get_double p1 in
  if (sequenceIsEmpty alldoubles)
  then
    sequenceEmpty
  else
    let sum = sequenceFold_left (fun x0 x1 -> x0 +. x1) 0.0 alldoubles in
    sequenceBuildSingle (_double (sum/.(float_of_int (sequenceLength alldoubles))))

let _fn_max_double nse schema n =
  let p1 = Args.get_param1 n in
  let alldoubles = sequenceMap get_double p1 in
  if (sequenceIsEmpty alldoubles)
  then
    sequenceEmpty
  else
    let start_max = sequenceHead alldoubles in
    let max = sequenceFold_left (fun x0 x1 -> if x0 <= x1 then x1 else x0) start_max alldoubles in
    sequenceBuildSingle (_double max)

let _fn_min_double nse schema n =
  let p1 = Args.get_param1 n in
  let alldoubles = sequenceMap get_double p1 in
  if (sequenceIsEmpty alldoubles)
  then
    sequenceEmpty
  else
    let start_min = sequenceHead alldoubles in
    let min = sequenceFold_left (fun x0 x1 -> if x0 < x1 then x0 else x1) start_min alldoubles in
    sequenceBuildSingle (_double min)

let _fn_sum_double nse schema n =
  let p1 = Args.get_param1 n in
  let alldoubles = sequenceMap get_double p1 in
  if (sequenceIsEmpty alldoubles)
  then
    sequenceBuildSingle (_double 0.0)
  else
    let sum = sequenceFold_left (fun x0 x1 -> x0 +. x1) 0.0 alldoubles in
    sequenceBuildSingle (_double sum)

(* F&O Section 14.5.3 fn:document *)

(* This function builds an XML stream differently for various kind of URIs.
   - Jerome *)
let local_open_xml_stream uri_string =
  match Glx_http.glx_decode_url uri_string with
  | Glx_http.Http (host,port,local) ->
      let string_buffer = Glx_http.glx_http_get host port local in
      Sax_parser.open_xml_stream_from_string string_buffer

  | Glx_http.File local ->
      Sax_parser.open_xml_stream_from_file local

let _fn_document nse schema n =
  let p1 = Args.get_param1 n in
  let uri_string = get_string (sequenceGetSingle p1) in
  let uri = AnyURI._uri_of_string uri_string in

  (* Debugging code. - Jerome *)
  begin
    if !Conf.debugging
    then
      begin
	Format.fprintf !Conf.xml_formatter "Loading document %s:\n" uri_string;
	let store = Store.get_projections () in
	if store = []
	then
	  Format.fprintf !Conf.xml_formatter "Store-projection: empty!!\n"
	else
	  if !Conf.debugging
	  then
	    Format.fprintf !Conf.xml_formatter "Store-projection: not empty...\n"
      end
  end;

  if not(!Conf.projection || !Conf.optimized_projection) || ((Store.get_projections ()) = [])
  then
    (* Default loading function *)
    let maybe_stored_document = Store.lookup_document uri_string in
    match maybe_stored_document with
    | None ->
	let apply_load_document uri_string =
	  let xml_stream = local_open_xml_stream uri_string in
	  let xml_datamodel = Load.load_xml_document_from_sax uri xml_stream in
	  begin
	    Store.add_document uri_string xml_datamodel;
	    xml_datamodel
	  end
	in
	Monitor.wrap_monitor (Monitor.Document_ParsingLoading_Phase uri) apply_load_document uri_string

    | Some document ->
	document

  else
    begin
    (* Projecting loading function *)
    let maybe_stored_projected_document = Store.lookup_projected_document uri_string in
    match maybe_stored_projected_document with
    | None ->
	let apply_load_document uri_string =
	  if !Conf.debugging
	  then
	    Format.fprintf !Conf.xml_formatter "Doing projection over document %s:\n" uri_string;
	  let ps = Store.get_projections () in
	  let xml_stream = local_open_xml_stream uri_string in
	  let projected_xml_stream =
	    Stream_project.project_xml_stream
	      (Namespace.empty_nsenv ())
	      uri
	      ps
	      xml_stream
	  in
	  let xml_datamodel = Load.load_xml_document_from_sax uri projected_xml_stream in
	  begin
	    if !Conf.print_projected_file
	    then
	      begin
		Format.pp_print_flush !Conf.projected_file_formatter ();  
		Serialization.fserialize_datamodel !Conf.projected_file_formatter xml_datamodel;
		Format.pp_print_flush !Conf.projected_file_formatter ();  
	      end;
	    Store.add_projected_document uri_string xml_datamodel;
	    xml_datamodel
	  end
	in
	Monitor.wrap_monitor (Monitor.Document_ParsingLoading_Phase uri) apply_load_document uri_string

    | Some projected_document ->
	if !Conf.debugging
	then
	  Format.fprintf !Conf.xml_formatter "Lookup failed for document:\n";

	projected_document
    end

(* Note:
    The above function does the lexing + parsing + loading of the document
    This is where we should measure memory/CPU time. 
  - Jerome and Amelie *)

let _glx_docorder nse schema n =
  let p1 = Args.get_param1 n in 
  sequence_of_list(List.sort item_compare (list_of_sequence p1))

let _glx_distinct_docorder nse schema n =
  let p1 = Args.get_param1 n in 
  let unordered_seq = remove_dup item_equal p1 in 
  sequence_of_list(List.sort item_compare (list_of_sequence unordered_seq))

let _glx_node_sequence nse schema n =
  let p1 = Args.get_param1 n in p1

let _glx_document_validate nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let uri_string = get_string (sequenceGetSingle p1) in
  let uri = AnyURI._uri_of_string uri_string in
  let maybe_stored_document = Store.lookup_document uri_string in
  match maybe_stored_document with
  | None ->
      let schema_root_string = get_string (sequenceGetSingle p2) in
      let schema_root =
	let uqname = (Namespace.uqname_element_of_string schema_root_string) in
	let (_,rqname) = Resolve.resolve_element_qname nse uqname in
	Sym.relem_symbol (fst uqname, rqname)
      in
      begin
	try
	  Schema.element_exists schema schema_root
	with
	| _ ->
	    raise (Query (Wrong_Args ("Unknown schema element in validate:" ^ schema_root_string)))
      end;
      Schema.create_root schema (Some schema_root);
      let xml_datamodel =  (* Root URI is empty for now *)
	Validation.validate_document nse schema uri (Parseutil.parseXML uri_string)
      in
      begin
	Store.add_document uri_string xml_datamodel;
	xml_datamodel
      end
  | Some document ->
      document

(************************)
(* Comparison functions *)
(************************)

(* comparison operations *)

let _op_equal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2)
  then
    sequenceEmpty
  else
  sequenceBuildSingle (boolval (value_eq p1 p2))

let _op_notequal nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2)
  then
    sequenceEmpty
  else
  sequenceBuildSingle (boolval (value_neq p1 p2))

let _op_lteq nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2)
  then
    sequenceEmpty
  else
  sequenceBuildSingle (boolval (item_lteq (sequenceGetSingle p1) (sequenceGetSingle p2)))

let _op_gteq nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2)
  then
    sequenceEmpty
  else
  sequenceBuildSingle (boolval (item_gteq (sequenceGetSingle p1) (sequenceGetSingle p2)))

let _op_lt nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2)
  then
    sequenceEmpty
  else
  sequenceBuildSingle (boolval (item_lt (sequenceGetSingle p1) (sequenceGetSingle p2)))

let _op_gt nse schema n =
  let (p1,p2) = Args.get_param2 n in
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2)
  then
    sequenceEmpty
  else
  sequenceBuildSingle (boolval (item_gt (sequenceGetSingle p1) (sequenceGetSingle p2)))


(*******************)
(* Galax functions *)
(*******************)

let _glx_unique nse schema n =
  let items = Args.get_param1 n in
  if sequenceIsSingle items
  then items
  else 
   raise (Query (Wrong_Args ("Function [fn:exactly-one]: not a singleton item")))

let _glx_optional nse schema n =
  let items = Args.get_param1 n in
  if sequenceIsOptional items
  then items
  else 
   raise (Query (Wrong_Args ("Function [fn:zero-or-one]: not a singleton item or ()")))


let _glx_one_plus nse schema n =
  let items = Args.get_param1 n in
  if sequenceIsEmpty items
  then 
   raise (Query (Wrong_Args ("Function [fn:one-or-more]: not a singleton item or more")))
  else 
    items


(* Galax function glx:union-values *)

let _glx_union_values nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let big_seq = sequenceAppend p1 p2 in
  remove_dup item_eq big_seq

(* Galax function glx:intersect-values *)

let _glx_intersect_values nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let nf1 = list_of_sequence p1 in
  let nf2 = list_of_sequence p2 in
  apply_intersect item_eq nf1 nf2

(* Galax function glx:except-values *)

let _glx_except_values nse schema n =
  let (p1,p2) = Args.get_param2 n in
  let nf1 = list_of_sequence p1 in
  let nf2 = list_of_sequence p2 in
  apply_except item_eq nf1 nf2

(* Galax function glx:print-string *)

let _glx_print_string nse schema n =
  let s1 = Args.get_param1 n in
  let s1' = get_string (sequenceGetSingle s1) in
  let _ =
    begin
      print_string s1';
      flush stdout
    end
  in
  sequenceEmpty

(* Galax function glx:print-item *)

let _glx_print_item nse schema n =
  let i1 = Args.get_param1 n in
  begin
    Serialization.fserialize_datamodel Format.std_formatter i1;
    flush stdout;
    sequenceEmpty
  end

(* Galax function glx:save-document *)

let _glx_save_document nse schema n =
  let (output_file_param, document_param) = Args.get_param2 n in
  let output_file_name = get_string (sequenceGetSingle output_file_param) in
  let output_channel = open_out output_file_name in
  let output_formatter = Format.formatter_of_out_channel output_channel in
  begin
    Serialization.fserialize_datamodel output_formatter document_param;
    close_out output_channel;
    sequenceEmpty
  end

(* Galax function glx:print-string-err *)

let _glx_print_string_err nse schema n =
  let s1 = Args.get_param1 n in
  let s1' = get_string (sequenceGetSingle s1) in
  let _ =
(*    match !Conf.glx_stderr with
    | None -> ()
    | Some oc ->
*)
	begin
	  output_string !Conf.glx_stderr s1';
	  flush !Conf.glx_stderr
	end
  in
  sequenceEmpty

(* Galax function glx:print-item-err *)

let _glx_print_item_err nse schema n =
  let i1 = Args.get_param1 n in
  let _ =
(*    match !Conf.glx_stderr with
    | None -> ()
    | Some oc ->
*)
	begin
	  let formatter = Format.formatter_of_out_channel !Conf.glx_stderr
	  in
	  Serialization.fserialize_datamodel formatter i1;
	  flush !Conf.glx_stderr
	end
  in
  sequenceEmpty

(* Galax function glx:get-lang *)

let _glx_get_lang nse schema n =
  let p1 = Args.get_param1 n in
  let node =
    try
      getNode (sequenceGetSingle p1)
    with
    | _ ->
	raise (Query (Wrong_Args "base-uri built-in function: not a node"))
  in
  let olang = get_lang node in
  match olang with
  | None -> sequenceEmpty
  | Some lang -> sequenceBuildSingle (buildAtomicValue (buildAtomicString lang))

(* Galax function glx:exponent *)

let _glx_exponent nse schema n =
  let (p1,p2) = Args.get_param2 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else *)
  let (i1,i2) = (get_double (sequenceGetSingle p1), get_double (sequenceGetSingle p2)) in
  sequenceBuildSingle (_double(i1 ** i2))

(* Galax function glx:sqrt *)

let _glx_sqrt nse schema n =
  let p1 = Args.get_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else *)
  let i1 = get_double (sequenceGetSingle p1) in
  sequenceBuildSingle (_double(sqrt i1))

(* Galax function glx:exp *)

let _glx_exp nse schema n =
  let (p1,p2) = Args.get_param2 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else *)
  let i1 = get_double (sequenceGetSingle p1) in
  sequenceBuildSingle (_double(exp i1))

(* Galax function glx:log *)

let _glx_log nse schema n =
  let p1 = Args.get_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else *)
  let i1 = get_double (sequenceGetSingle p1) in
  sequenceBuildSingle (_double(log i1))

(* Galax function glx:log10 *)

let _glx_log10 nse schema n =
  let p1 = Args.get_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else *)
  let i1 = get_double (sequenceGetSingle p1) in
  sequenceBuildSingle (_double(log10 i1))

(* Galax function glx:cos *)

let _glx_cos nse schema n =
  let p1 = Args.get_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else *)
  let i1 = get_double (sequenceGetSingle p1) in
  sequenceBuildSingle (_double(cos i1))

(* Galax function glx:sin *)

let _glx_sin nse schema n =
  let p1 = Args.get_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else *)
  let i1 = get_double (sequenceGetSingle p1) in
  sequenceBuildSingle (_double(sin i1))

(* Galax function glx:tan *)

let _glx_tan nse schema n =
  let p1 = Args.get_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else *)
  let i1 = get_double (sequenceGetSingle p1) in
  sequenceBuildSingle (_double(tan i1))

(* Galax function glx:acos *)

let _glx_acos nse schema n =
  let p1 = Args.get_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else *)
  let i1 = get_double (sequenceGetSingle p1) in
  sequenceBuildSingle (_double(acos i1))

(* Galax function glx:asin *)

let _glx_asin nse schema n =
  let p1 = Args.get_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else *)
  let i1 = get_double (sequenceGetSingle p1) in
  sequenceBuildSingle (_double(asin i1))

(* Galax function glx:atan *)

let _glx_atan nse schema n =
  let p1 = Args.get_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else *)
  let i1 = get_double (sequenceGetSingle p1) in
  sequenceBuildSingle (_double(atan i1))

(* Galax function glx:atan2 *)

let _glx_atan2 nse schema n =
  let (p1,p2) = Args.get_param2 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else *)
  let (i1,i2) = (get_double (sequenceGetSingle p1), get_double (sequenceGetSingle p2)) in
  sequenceBuildSingle (_double(atan2 i1 i2))

(* Galax function glx:cosh *)

let _glx_cosh nse schema n =
  let p1 = Args.get_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else *)
  let i1 = get_double (sequenceGetSingle p1) in
  sequenceBuildSingle (_double(cosh i1))

(* Galax function glx:sinh *)

let _glx_sinh nse schema n =
  let p1 = Args.get_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else *)
  let i1 = get_double (sequenceGetSingle p1) in
  sequenceBuildSingle (_double(sinh i1))

(* Galax function glx:tanh *)

let _glx_tanh nse schema n =
  let p1 = Args.get_param1 n in
(* Not clear whether it should work on double? instead of double - Jerome
  if (sequenceIsEmpty p1) || (sequenceIsEmpty p2) then sequenceEmpty
  else *)
  let i1 = get_double (sequenceGetSingle p1) in
  sequenceBuildSingle (_double(tanh i1))

let _glx_soap_call nse schema n =
  let (p1,p2) = Args.get_param2 n in
    (* server URI *)
  let uri_server = get_anyURI (sequenceGetSingle p1) in
  let string_server = AnyURI._string_of_uri uri_server in
    (* soap message *)
  let request = Serialization.bserialize_datamodel p2 in
  let response = Glx_http.http_soap request string_server in
  let xml_stream = Sax_parser.open_xml_stream_from_string response in
    Load.load_xml_document_from_sax uri_server xml_stream    
      (* 
       * For the time being, it is a simple http POST message
       * 
       * Nicola
       *)

(*******************************)
(* Table of built-in functions *)
(*******************************)

let fun_table = [
  (* F&O Section 2. Accessors *)
  ((fn_uri,"node-kind"), 1),  _fn_node_kind;
  ((fn_uri,"node-name"), 1),  _fn_node_name;
  ((fn_uri,"string"), 1),     _fn_string;
  ((fn_uri,"base-uri"), 0),   _fn_base_uri_zero;
  ((fn_uri,"base-uri"), 1),   _fn_base_uri_one;
  ((fn_uri,"data"), 1),       _fn_data;
  ((fn_uri,"boolean"), 1),    _fn_boolean;
  ((fn_uri,"unique-ID"), 1),  _DUMMY "unique-ID";   (* Dummy *)

  (* F&O Section 5. Functions and Operators on Numbers *)
  (* Note: arithmetic functions are monomorphic in Galax for now... *)
  ((op_uri,"int-add"), 2), _fn_int_add;
  ((op_uri,"int-subtract"), 2), _fn_int_sub;
  ((op_uri,"int-multiply"), 2), _fn_int_mult;
  ((op_uri,"int-divide"), 2), _fn_int_div;
  ((op_uri,"int-mod"), 2), _fn_int_mod;

  ((op_uri,"int-unary-plus"), 1), _fn_int_unary_plus;
  ((op_uri,"int-unary-minus"), 1), _fn_int_unary_minus;

  ((op_uri,"int-equal"), 2), _fn_int_equal;
  ((op_uri,"int-nequal"), 2), _fn_int_nequal;
  ((op_uri,"int-lt"), 2), _fn_int_lt;
  ((op_uri,"int-gt"), 2), _fn_int_gt;
  ((op_uri,"int-le"), 2), _fn_int_le;
  ((op_uri,"int-ge"), 2), _fn_int_ge;

  ((op_uri,"integer-add"), 2), _fn_integer_add;
  ((op_uri,"integer-subtract"), 2), _fn_integer_sub;
  ((op_uri,"integer-multiply"), 2), _fn_integer_mult;
  ((op_uri,"integer-divide"), 2), _fn_integer_div;
  ((op_uri,"integer-mod"), 2), _fn_integer_mod;

  ((op_uri,"integer-unary-plus"), 1), _fn_integer_unary_plus;
  ((op_uri,"integer-unary-minus"), 1), _fn_integer_unary_minus;

  ((op_uri,"integer-equal"), 2), _fn_integer_equal;
  ((op_uri,"integer-nequal"), 2), _fn_integer_nequal;
  ((op_uri,"integer-lt"), 2), _fn_integer_lt;
  ((op_uri,"integer-gt"), 2), _fn_integer_gt;
  ((op_uri,"integer-le"), 2), _fn_integer_le;
  ((op_uri,"integer-ge"), 2), _fn_integer_ge;

  ((op_uri,"decimal-add"), 2), _fn_decimal_add;
  ((op_uri,"decimal-subtract"), 2), _fn_decimal_sub;
  ((op_uri,"decimal-multiply"), 2), _fn_decimal_mult;
  ((op_uri,"decimal-divide"), 2), _fn_decimal_div;
  ((op_uri,"decimal-mod"), 2), _fn_decimal_mod;

  ((op_uri,"decimal-unary-plus"), 1), _fn_decimal_unary_plus;
  ((op_uri,"decimal-unary-minus"), 1), _fn_decimal_unary_minus;

  ((op_uri,"decimal-equal"), 2), _fn_decimal_equal;
  ((op_uri,"decimal-nequal"), 2), _fn_decimal_nequal;
  ((op_uri,"decimal-lt"), 2), _fn_decimal_lt;
  ((op_uri,"decimal-gt"), 2), _fn_decimal_gt;
  ((op_uri,"decimal-le"), 2), _fn_decimal_le;
  ((op_uri,"decimal-ge"), 2), _fn_decimal_ge;

  ((op_uri,"float-add"), 2), _fn_float_add;
  ((op_uri,"float-subtract"), 2), _fn_float_sub;
  ((op_uri,"float-multiply"), 2), _fn_float_mult;
  ((op_uri,"float-divide"), 2), _fn_float_div;
  ((op_uri,"float-mod"), 2), _fn_float_mod;

  ((op_uri,"float-unary-plus"), 1), _fn_float_unary_plus;
  ((op_uri,"float-unary-minus"), 1), _fn_float_unary_minus;

  ((op_uri,"float-equal"), 2), _fn_float_equal;
  ((op_uri,"float-nequal"), 2), _fn_float_nequal;
  ((op_uri,"float-lt"), 2), _fn_float_lt;
  ((op_uri,"float-gt"), 2), _fn_float_gt;
  ((op_uri,"float-le"), 2), _fn_float_le;
  ((op_uri,"float-ge"), 2), _fn_float_ge;

  ((op_uri,"double-add"), 2), _fn_double_add;
  ((op_uri,"double-subtract"), 2), _fn_double_sub;
  ((op_uri,"double-multiply"), 2), _fn_double_mult;
  ((op_uri,"double-divide"), 2), _fn_double_div;
  ((op_uri,"double-mod"), 2), _fn_double_mod;

  ((op_uri,"double-unary-plus"), 1), _fn_double_unary_plus;
  ((op_uri,"double-unary-minus"), 1), _fn_double_unary_minus;

  ((op_uri,"double-equal"), 2), _fn_double_equal;
  ((op_uri,"double-nequal"), 2), _fn_double_nequal;
  ((op_uri,"double-lt"), 2), _fn_double_lt;
  ((op_uri,"double-gt"), 2), _fn_double_gt;
  ((op_uri,"double-le"), 2), _fn_double_le;
  ((op_uri,"double-ge"), 2), _fn_double_ge;

  ((fn_uri,"floor"), 1), _fn_floor;
  ((fn_uri,"ceiling"), 1), _fn_ceiling;
  ((fn_uri,"round"), 1), _fn_round;

  (* F&O Section 4. Trace *)

  ((fn_uri,"trace"), 2), _fn_trace;

  (* F&O Section 6. Functions on Strings *)

  (* Note:
       (**) indicates the functions which have been partially
       implemented.
     BChoi *)
  ((fn_uri,"concat"), 2),             _fn_concat;
  ((fn_uri,"string-join"), 2),        _fn_string_join;
  ((fn_uri,"starts-with"), 2),        _fn_starts_with;                         (**)
  ((fn_uri,"ends-with"), 2),          _fn_ends_with;                           (**)
  ((fn_uri,"contains"), 2),           _fn_contains;
  ((fn_uri,"substring"), 3),          _fn_substring;                           (**)
  ((fn_uri,"string-length"), 1),      _fn_string_length;
  ((fn_uri,"substring-before"), 2),   _fn_substring_before;
  ((fn_uri,"substring-after"), 2),    _fn_substring_after;
  ((fn_uri,"normalize-space"), 1),    _fn_normalize_space;
  ((fn_uri,"upper-case"), 1),         _fn_upper_case;
  ((fn_uri,"lower-case"), 1),         _fn_lower_case;
  ((fn_uri,"translate"), 3),          _fn_translate;
  ((fn_uri,"string-pad"), 3),         _fn_string_pad;

  ((op_uri,"string-equal"), 2),       _op_string_equal;
  ((op_uri,"string-nequal"), 2),      _op_string_nequal;
  ((op_uri,"string-lt"), 2),          _op_string_lt;
  ((op_uri,"string-gt"), 2),          _op_string_gt;
  ((op_uri,"string-le"), 2),          _op_string_le;
  ((op_uri,"string-ge"), 2),          _op_string_ge;


  (* F&O Section 7. Functions and Operators on Booleans *)
  ((fn_uri,"true"), 0),  _fn_true;
  ((fn_uri,"false"), 0), _fn_false;
  ((fn_uri,"not"), 1),   _fn_not;
  ((op_uri,"or"), 2),    _op_or;
  ((op_uri,"and"), 2),   _op_and;
  ((op_uri,"boolean-equal"), 2), _op_boolean_equal;
  ((op_uri,"boolean-nequal"), 2), _op_boolean_nequal;
  ((op_uri,"boolean-lt"), 2), _op_boolean_lt;
  ((op_uri,"boolean-gt"), 2), _op_boolean_gt;
  ((op_uri,"boolean-le"), 2), _op_boolean_le;
  ((op_uri,"boolean-ge"), 2), _op_boolean_ge;

  (* F&O Section 9. Functions on QNames *)
  ((fn_uri,"expanded-QName"), 2), _fn_expanded_qname;
  ((fn_uri,"get-local-name-from-QName"), 1), _fn_get_local_name_from_qname;
  ((fn_uri,"get-namespace-from-QName"), 1), _fn_get_namespace_from_qname;
  ((op_uri, "QName-equal"), 2), _op_QName_equal;
  ((op_uri, "QName-nequal"), 2), _op_QName_nequal;

  (* F&O Section 10. Functions and Operators for anyURI *)
  ((fn_uri,"resolve-uri"), 2), _fn_resolve_uri;
  ((op_uri, "anyURI-equal"), 2), _op_anyURI_equal;
  ((op_uri, "anyURI-nequal"), 2), _op_anyURI_nequal;

  (* F&O Section 13. Functions and Operators on Nodes *)
  ((fn_uri,"name"), 1), _fn_name;
  ((fn_uri,"local-name"), 1), _fn_local_name;
  ((fn_uri,"namespace-uri"), 1), _fn_namespace_uri;
  ((fn_uri,"lang"), 2), _fn_lang;
  ((op_uri,"node-equal"), 2), _op_node_equal;
  ((op_uri,"node-before"), 2), _fn_node_before;
  ((op_uri,"node-after"), 2), _fn_node_after;
  ((fn_uri,"root"), 1), _fn_root;

  (* F&O Section 15. Functions and Operators on Sequences *)
  ((fn_uri,"exactly-one"), 1),             _glx_unique;
  ((fn_uri,"zero-or-one"), 1),           _glx_optional;
  ((fn_uri,"one-or-more"), 1),           _glx_one_plus;

  ((op_uri,"to"), 2), _op_to;

  ((fn_uri,"item-at"), 2), _fn_item_at;
  ((fn_uri,"index-of"), 2), _fn_index_of;
  ((fn_uri,"empty"), 1), _fn_empty;
  ((fn_uri,"exists"), 1), _fn_exists;
  ((fn_uri,"distinct-nodes"), 1), _fn_distinct_nodes;
  ((fn_uri,"distinct-values"), 1), _fn_distinct_value;
  ((fn_uri,"insert-before"), 3), _fn_insert;
  ((fn_uri,"remove"), 2), _fn_remove;
  ((fn_uri,"subsequence"), 2), _fn_subsequence_2;
  ((fn_uri,"subsequence"), 3), _fn_subsequence_3;

  ((fn_uri,"deep-equal"), 2), _fn_sequence_deep_equal;
  ((fn_uri,"sequence-node-identical"), 2), _fn_sequence_node_equal;
  ((op_uri,"union"), 2), _op_union;
  ((op_uri,"intersect"), 2), _op_intersect;
  ((op_uri,"except"), 2), _op_except;

  ((fn_uri,"count"), 1), _fn_count;

  ((fn_uri,"avg-int"), 1), _fn_avg_int;
  ((fn_uri,"max-int"), 1), _fn_max_int;
  ((fn_uri,"min-int"), 1), _fn_min_int;
  ((fn_uri,"sum-int"), 1), _fn_sum_int;

  ((fn_uri,"avg-integer"), 1), _fn_avg_integer;
  ((fn_uri,"max-integer"), 1), _fn_max_integer;
  ((fn_uri,"min-integer"), 1), _fn_min_integer;
  ((fn_uri,"sum-integer"), 1), _fn_sum_integer;

  ((fn_uri,"avg-decimal"), 1), _fn_avg_decimal;
  ((fn_uri,"max-decimal"), 1), _fn_max_decimal;
  ((fn_uri,"min-decimal"), 1), _fn_min_decimal;
  ((fn_uri,"sum-decimal"), 1), _fn_sum_decimal;

  ((fn_uri,"avg-float"), 1), _fn_avg_float;
  ((fn_uri,"max-float"), 1), _fn_max_float;
  ((fn_uri,"min-float"), 1), _fn_min_float;
  ((fn_uri,"sum-float"), 1), _fn_sum_float;

  ((fn_uri,"avg-double"), 1), _fn_avg_double;
  ((fn_uri,"max-double"), 1), _fn_max_double;
  ((fn_uri,"min-double"), 1), _fn_min_double;
  ((fn_uri,"sum-double"), 1), _fn_sum_double;

  ((fn_uri,"document"), 1), _fn_document;

  ((glx_uri,"docorder"), 1), _glx_docorder;
  ((glx_uri,"distinct-docorder"), 1), _glx_distinct_docorder;
  ((glx_uri,"document-validate"), 2), _glx_document_validate;
  ((glx_uri,"node-sequence"), 1), _glx_node_sequence;

  (* Comparison functions *)

  (* Note: The following functions are effectively polymorphic. *)

  ((op_uri,"equal"), 2), _op_equal;
  ((op_uri,"nequal"), 2), _op_notequal;
  ((op_uri,"le"), 2), _op_lteq;
  ((op_uri,"ge"), 2), _op_gteq;
  ((op_uri,"gt"), 2), _op_gt;
  ((op_uri,"lt"), 2), _op_lt;

  (* Galax functions *)

  (* Note:
       Those functions are not part of the F&O spec. They are added in
       Galax for user's convenience.
     - Jerome *)
  
  ((glx_uri,"union-values"), 2), _glx_union_values;
  ((glx_uri,"intersect-values"), 2), _glx_intersect_values;
  ((glx_uri,"except-values"), 2), _glx_except_values;

  ((glx_uri,"print-string"), 1), _glx_print_string;
  ((glx_uri,"print-item"), 1), _glx_print_item;

  ((glx_uri,"print-string-err"), 1), _glx_print_string_err;
  ((glx_uri,"print-item-err"), 1), _glx_print_item_err;

  ((glx_uri,"save-document"), 2), _glx_save_document;

  ((glx_uri,"get-lang"), 1), _glx_get_lang;

  ((glx_uri,"exponent"), 2), _glx_exponent;
  ((glx_uri,"sqrt"), 1), _glx_sqrt;
  ((glx_uri,"exp"), 1), _glx_exp;
  ((glx_uri,"log"), 1), _glx_log;
  ((glx_uri,"log10"), 1), _glx_log10;
  ((glx_uri,"cos"), 1), _glx_cos;
  ((glx_uri,"sin"), 1), _glx_sin;
  ((glx_uri,"tan"), 1), _glx_tan;
  ((glx_uri,"acos"), 1), _glx_acos;
  ((glx_uri,"asin"), 1), _glx_asin;
  ((glx_uri,"atan"), 1), _glx_atan;
  ((glx_uri,"atan2"), 2), _glx_atan2;
  ((glx_uri,"cosh"), 1), _glx_cosh;
  ((glx_uri,"sinh"), 1), _glx_sinh;
  ((glx_uri,"tanh"), 1), _glx_tanh;

  ((glx_uri,"soap-call"), 2), _glx_soap_call;
]

(* creation of the table for built-in functions *)

let bltin_fctns : ((rfname * int), bltin) Hashtbl.t = Hashtbl.create 137

let add_bltin_fctn ((rfname, i), f) = Hashtbl.add bltin_fctns (rfname,i) f

let lookup_bltin_fctn (rfname,i) =
  try
    Hashtbl.find bltin_fctns (rfname,i)
  with _ ->
    raise (Query (Unknown ("Builtin function "
			   ^ (Namespace.string_of_rqname rfname)
			   ^ " with arity "
			   ^ (string_of_int i)
			   ^ " not found")))

let _ = 
  try
    List.iter add_bltin_fctn fun_table
  with exn -> printf_error " " exn

