(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: fn_static.ml,v 1.3 2003/06/04 20:58:17 simeon Exp $ *)

(* Module Fn_static
   Description:
     This module implements static typing for F&O functions.
*)

open Error

open Sym
open Namespace

open Occurrence
open Datatypes

open Schema
open Schemautil

open Typeutil

open Xquery_ast
open Xquery_astutil

(* Typing for built-in functions types *)

type bltin_type_rule = schema -> Finfo.finfo -> model list -> model -> model

let _data_type schema fi arguments_types output_type =
  let argument_type = Args.get_param1 arguments_types in
  let input_type = Simplification.simplify_ty schema argument_type in

  (* See: wrap_factor_type for the rationale for this logic *)
  if (is_empty_sequence schema input_type) then fmkmodel TEmpty fi
  else if (is_empty_choice schema input_type) then fmkmodel TNone fi
  else
    let (pl,m,n) = Typeutil.factor_with_units schema input_type in
    let data_types_fun t =
      begin
	let t'' = (dereference_elem_attr schema (dereference_model schema t)) in
	match (t''.pmodel_desc) with 
	| TElemLocal(l,_,t') -> 
	    begin
	      match t'.ptyped_model_desc with 
	      | TAnonymousType m -> 
		  let content_m = extract_content_model schema m in
		  if (Subsumption.isSubtype schema content_m (fmkmodel (TGroupRef (RGroupSym Sym.anysimpletype)) fi))
		  then content_m 
		    (* MFF: Note that this code should change once named typing is implemented *)
    	            (* MFF: data() on arbitrary node always returns Untyped *)
		  else fmkmodel (TBase ATUntyped) fi 
	    end
	| TAttrLocal(l,t') -> 
	    begin
	      match t'.ptyped_model_desc with 
	      | TAnonymousType m ->
		  m
	    end
	| TText -> 
	    begin
	      fmkmodel (TBase ATUntyped) fi
	    end
	| TDocument m ->
	    fmkmodel TEmpty fi
	| TBase b -> t'' 
	| _ ->
	    raise(Query (Static_Type_Error (fi, ("Factored type of 'fn:data' argument:\n"
						 ^ (Print.bprintf_model "" t)
						 ^ " not a prime type"
						     ))))
      end
    in
    let p' = Schemautil.weakunion_types (List.map data_types_fun pl)
    in fmkmodel (TBound(p', m, n)) fi
(*  apply_to_unit_types schema data_types_fun input_type *)

let _copy_type schema fi arguments_types output_type =
  Args.get_param1 arguments_types

let _copy_first_type schema fi arguments_types output_type =
  let (argument_type1, argument_type2) = Args.get_param2 arguments_types in
  argument_type1

let _distinct_type schema fi arguments_types  output_type =
  let argument_type = Args.get_param1 arguments_types in
  let input_type = Simplification.simplify_ty schema argument_type in

  (* See: wrap_factor_type for the rationale for this logic *)
  if (Typeutil.is_empty_sequence schema input_type) then fmkmodel TEmpty fi
  else if (Typeutil.is_empty_choice schema input_type) then fmkmodel TNone fi
  else
    let (p,m,n) = Typeutil.factor schema input_type in
    fmkmodel (TBound(p,ub_min m (occurs 1),n)) fi

(*
  _node_seq_type implements a more precise type rule for glx:node-sequence:

          Expr : Type
         Type <: Node*
  ------------------------------
  glx:node-sequence(Expr) : Type

*)
let  _node_seq_type schema fi arguments_types  output_type =
  let argument_type = Args.get_param1 arguments_types in
  let input_type = Simplification.simplify_ty schema argument_type in

  (* See: wrap_factor_type for the rationale for this logic *)
  if (Typeutil.is_empty_sequence schema input_type) then fmkmodel TEmpty fi
  else if (Typeutil.is_empty_choice schema input_type) then fmkmodel TNone fi
  else if (Typeutil.is_anynode_sequence schema input_type) then input_type 
  else raise(Query (Static_Type_Error (fi, ("Type of argument to 'glx:node-sequence' contains atomic types.\n"))))


(*
  _arithmetic_type implements more precise type rules for arithmetic operators:

        Expr_i : empty
        1 <= i <= 2
  ---------------------------
  Expr1 ArithOp Expr2 : empty

          Expr1 <: numeric
          Expr2 <: numeric
  ----------------------------------
  Expr1 ArithOp Expr2 : type(ArithOp)

*)
let _arithmetic_type schema fi arguments_types  output_type =
  let (argument_type1, argument_type2) = Args.get_param2 arguments_types in
  let input_type1 = Simplification.simplify_ty schema argument_type1 in
  let input_type2 = Simplification.simplify_ty schema argument_type2 in
  let (output_type',_,_) = Typeutil.factor schema output_type in

  if (is_empty_sequence schema input_type1) || (is_empty_sequence schema input_type2) then fmkmodel TEmpty fi
  else if (is_anynumeric schema input_type1) && (is_anynumeric schema input_type2) then output_type'
  else output_type

(*
  _unary_type implements more precise type rules for unary arithmetic operators:

        Expr : empty
   --------------------
   UnaryOp Expr : empty

       Expr <: numeric
  ----------------------------
  UnaryOp Expr : type(UnaryOp)

*)
let _unary_arithmetic_type schema fi arguments_types  output_type =
  let argument_type = Args.get_param1 arguments_types in
  let input_type = Simplification.simplify_ty schema argument_type in

  if (is_empty_sequence schema input_type) then fmkmodel TEmpty fi
  else if (is_anynumeric schema input_type) then input_type
  else output_type

(* creation of the table for built-in functions *)

let bltin_fctns_type_rules : ((rfname * int), bltin_type_rule) Hashtbl.t = Hashtbl.create 137

let add_bltin_fctn_type_rule ((rfname,i), type_rule) = Hashtbl.add bltin_fctns_type_rules (rfname,i) type_rule

let lookup_bltin_fctn_type_rule (rfname,i) =
  try
    Hashtbl.find bltin_fctns_type_rules (rfname,i)
  with _ ->
    raise (Query (Internal_Error ("Looking up type rule for a built-in function. This error should be trapped and never be visible")))

(* Type relief functions *)

let _zero_or_one schema fi arguments_types  output_type =
  let argument_type = Args.get_param1 arguments_types in
  let input_type = Simplification.simplify_ty schema argument_type in

  (* See: wrap_factor_type for the rationale for this logic *)
  if (Typeutil.is_empty_sequence schema input_type) then fmkmodel TEmpty fi
  else if (Typeutil.is_empty_choice schema input_type) then fmkmodel TNone fi
  else
    let (p,m,n) = Typeutil.factor schema input_type in
    if (Occurrence.equal m (occurs 1)) then p 
    else fmkmodel (TBound(p, (occurs 0), (occurs 1))) fi

(* MF: 
   We use approximate occurrences in the type-refief functions, so
   even we compute a type T(m,n) where m and n are known integers
   s.t. m > 1, we approximate as (1, unbounded).
*)
let _exactly_one schema fi arguments_types  output_type =
  let argument_type = Args.get_param1 arguments_types in
  let input_type = Simplification.simplify_ty schema argument_type in

  (* See: wrap_factor_type for the rationale for this logic *)
  if (Typeutil.is_empty_sequence schema input_type) then 
    raise(Query (Static_Type_Error (fi, ("Type of argument to 'fn:exactly-one' is empty\n"))))
  else if (Typeutil.is_empty_choice schema input_type) then 
    fmkmodel TNone fi
  else
    let (p,m,n) = Typeutil.factor schema input_type in 
    p

let _one_or_more schema fi arguments_types  output_type =
  let argument_type = Args.get_param1 arguments_types in
  let input_type = Simplification.simplify_ty schema argument_type in

  (* See: wrap_factor_type for the rationale for this logic *)
  if (Typeutil.is_empty_sequence schema input_type) then 
    raise(Query (Static_Type_Error (fi, ("Type of 'fn:one-or-more' argument is empty\n"))))
  else if (Typeutil.is_empty_choice schema input_type) then 
    fmkmodel TNone fi
  else
    let (p,m,n) = Typeutil.factor schema input_type in
    fmkmodel (TBound(p, (occurs 1), n)) fi

let type_rule_table = [
  (* data model operations *)
  ((fn_uri,"data"), 1), _data_type;

  (* node operations *)
  ((fn_uri,"copy"), 1), _copy_type;

  (* arithmetic operators *)
  ((op_uri,"numeric-add"), 2), _arithmetic_type;
  ((op_uri, "double-add"), 2), _arithmetic_type; 
  ((op_uri, "float-add"), 2), _arithmetic_type; 
  ((op_uri, "decimal-add"), 2), _arithmetic_type; 
  ((op_uri, "integer-add"), 2), _arithmetic_type;  
  ((op_uri, "int-add"), 2), _arithmetic_type; 

  ((op_uri,"numeric-subtract"), 2), _arithmetic_type;
  ((op_uri, "double-subtract"), 2), _arithmetic_type; 
  ((op_uri, "float-subtract"), 2), _arithmetic_type; 
  ((op_uri, "decimal-subtract"), 2), _arithmetic_type; 
  ((op_uri, "integer-subtract"), 2), _arithmetic_type;  
  ((op_uri, "int-subtract"), 2), _arithmetic_type; 

  ((op_uri,"numeric-multiply"), 2), _arithmetic_type;
  ((op_uri, "double-multiply"), 2), _arithmetic_type; 
  ((op_uri, "float-multiply"), 2), _arithmetic_type; 
  ((op_uri, "decimal-multiply"), 2), _arithmetic_type; 
  ((op_uri, "integer-multiply"), 2), _arithmetic_type; 
  ((op_uri, "int-multiply"), 2), _arithmetic_type; 

  ((op_uri,"numeric-divide"), 2), _arithmetic_type;
  ((op_uri, "double-divide"), 2), _arithmetic_type; 
  ((op_uri, "float-divide"), 2), _arithmetic_type;  
  ((op_uri, "decimal-divide"), 2), _arithmetic_type; 
  ((op_uri, "integer-divide"), 2), _arithmetic_type;  
  ((op_uri, "int-divide"), 2), _arithmetic_type; 

  ((op_uri,"numeric-mod"), 2), _arithmetic_type; 
  ((op_uri, "double-mod"), 2), _arithmetic_type; 
  ((op_uri, "float-mod"), 2), _arithmetic_type; 
  ((op_uri, "decimal-mod"), 2), _arithmetic_type; 
  ((op_uri, "integer-mod"), 2), _arithmetic_type; 
  ((op_uri, "int-mod"), 2), _arithmetic_type; 

  ((op_uri,"numeric-idivide"), 2), _arithmetic_type; 
  ((op_uri, "integer-divide"), 2), _arithmetic_type; 
  ((op_uri, "int-divide"), 2), _arithmetic_type; 

  ((op_uri,"numeric-unary-plus"), 1), _unary_arithmetic_type; 
  ((op_uri,"double-unary-plus"), 1), _unary_arithmetic_type;
  ((op_uri,"float-unary-plus"), 1), _unary_arithmetic_type;
  ((op_uri,"decimal-unary-plus"), 1), _unary_arithmetic_type;
  ((op_uri,"integer-unary-plus"), 1), _unary_arithmetic_type; 
  ((op_uri,"int-unary-plus"), 1), _unary_arithmetic_type;

  ((op_uri,"numeric-unary-minus"), 1), _unary_arithmetic_type;
  ((op_uri,"double-unary-minus"), 1), _unary_arithmetic_type;
  ((op_uri,"float-unary-minus"), 1), _unary_arithmetic_type;
  ((op_uri,"decimal-unary-minus"), 1), _unary_arithmetic_type;
  ((op_uri,"integer-unary-minus"), 1), _unary_arithmetic_type; 
  ((op_uri,"int-unary-minus"), 1), _unary_arithmetic_type; 

  (* type relief functions *)
  ((fn_uri,"exactly-one"),1), _exactly_one;
  ((fn_uri,"zero-or-one"),1), _zero_or_one;
  ((fn_uri,"one-or-more"),1), _one_or_more;
  ((fn_uri,"trace"),2), _copy_first_type;

  (* collection operations *)
  ((fn_uri,"distinct-values"),1), _distinct_type;
  ((fn_uri,"distinct-nodes"),1), _distinct_type;

  (* document order functions *)
  ((glx_uri,"docorder"),1), _distinct_type;
  ((glx_uri,"distinct-docorder"),1), _distinct_type;
  ((glx_uri,"node-sequence"),1), _node_seq_type;
]

let _ = List.map add_bltin_fctn_type_rule type_rule_table

