(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: sax_parser.ml,v 1.7.2.1 2003/12/02 04:08:48 simeon Exp $ *)

(* Module: Sax_parser
   Description:
     This module implements a SAX parser for Galax.
*)

open Glx_pxp
  (* Note: PXP is used to build a well-behaved lexing buffer. - Jerome *)

open Error

open Xquery_parser
  (* Note: This contains the definition of lexing tokens. - Jerome *)
open Parseutil

open Sax_types

(* Note:
     The SAX parser is using two levels of streams, one consuming the
     other.

       * The lexer stream is a stream of tokens taken out of the
         lexers. It is low-level and internal to the SAX
         implementation.
       * The XML event stream is a stream of XML SAX events. It is
         high-level and accessible to the user of the SAX module.

     Most of the code deals with consumption of the low-level stream
     to recognized it is a well-formed sequence of tokens, and if so,
     to generate the corresponding high-level events.

   - Jerome
 *)

(*********************)
(* Entity resolution *)
(*********************)

(* Requests a token from the lexer and builds an optional result, as
   required for the stream interface *)

let xml_lexer_stream_fun_build lexbuf n =
  let lexer_token = Parseutil.xml_lexfun lexbuf in
  match lexer_token with
  | EOF -> None
  | _ -> Some lexer_token

let open_local_stream_from_string push_lexstate_fun s =
  push_lexstate_fun ();
  let resolver = Glx_pxp.resolve_as_string s in
  Glx_pxp.init_resolver resolver (Encoding.get_internal_encoding ());
  let lexbuf = Glx_pxp.open_resolver resolver s in
  let xml_lexer_stream_fun =
    xml_lexer_stream_fun_build lexbuf
  in
  (Stream.from xml_lexer_stream_fun, resolver)

let close_local_stream resolver =
  Glx_pxp.close_resolver resolver;
  Lexutil.pop_state_keep_buffer()

let close_local_stream_keep_string_buffer resolver =
  Glx_pxp.close_resolver resolver;
  Lexutil.pop_state_keep_buffer()


let compute_replacement_text parse_context fi x =
  let ended_text = (x ^ "\"") in (* HACK -- to make the lexer terminate *)
  let (stream,resolver) = open_local_stream_from_string Lexutil.push_entity_value ended_text in
  let buffer = ref "" in
  begin
    let rec compute_replacement_text_aux stream =
      begin
	match Stream.next stream with
	| TEXTENTITYREF (text,entityref) ->
	    buffer := !buffer ^ text ^ entityref;
	    compute_replacement_text_aux stream
	| TEXTCHARREF (text,char_ref) ->
	    let char_ref_text = Glx_pxp.character (Encoding.get_internal_encoding ()) char_ref in
	    buffer := !buffer ^ text ^ char_ref_text;
	    compute_replacement_text_aux stream
	| ATTRIBUTETEXT text ->
	    buffer := !buffer ^ text
	| _ ->
	    raise (Query (Parsing (fi,"Cannot parse general entity: \"" ^ x ^ "\"")))
      end
    in
    try
      compute_replacement_text_aux stream
    with
    | Stream.Failure ->
	()
  end;
  begin
    close_local_stream resolver;
    !buffer
  end


(*********************************)
(* Error generation              *)
(*********************************)

let string_of_token token =
  match token with
  | XQNAME uqname -> "the QName: " ^ (Namespace.string_of_uqname uqname)
  | XNCNAME ncname -> "the QName: " ^ ncname
  | XSTRING s -> "the string: \"" ^ s ^ "\""
  | XMLPI -> "'<?xml'"
  | XVERSION -> "an XML version"
  | XSDDECL -> "standalone"
  | XSINGLEQUOTE -> "'''"
  | XDOUBLEQUOTE -> "'\"'"
  | XDECLNAME s -> "'" ^ s ^ "'"
  | XENCODING s -> "an XML encoding ('" ^ s ^ "')"
  | DOCTYPE -> "'<!DOCTYPE'"
  | DOCTYPECLOSE -> "'>'"
  | DOCTYPE_SYSTEM -> "'SYSTEM'"
  | DOCTYPE_PUBLIC -> "'PUBLIC'"
  | DTDDECLOPEN -> "'['"
  | DTDDECLCLOSE -> "']'"
  | S -> "some whitespace"
  | LOPENINGTAG -> "'<'"
  | LCLOSINGTAG -> "'</'"
  | ROPENINGTAG -> "'>'"
  | RCLOSINGTAG -> "'>'"
  | REMPTYELEMENT -> "'/>'"
  | TEXTLCLOSINGTAG u -> "the text '" ^ u ^ "'" 
  | TEXTLOPENINGTAG u -> "the text '" ^ u ^ "'" 
  | TEXTLCURLY u -> "the text '" ^ u ^ "'" 
  | OPENINGPI _ -> "'<?'"
  | CLOSINGPI -> "'?>'"
  | LOPENINGCOMMENT -> "'<!--'"
  | TEXTCLOSINGPI u -> "the text '" ^ u ^ "'"
  | TEXTRCLOSINGCOMMENT u -> "the text '" ^ u ^ "'"
  | TEXTOPENINGPI (u,_) -> "the text '" ^ u ^ "'"
  | TEXTLOPENINGCOMMENT u -> "the text '" ^ u ^ "'"
  | ATTRIBUTETEXT u -> "the text '" ^ u ^ "'"
  | ATTRIBUTETEXTLCURLY u -> "the text '" ^ u ^ "'"
  | TEXTCHARREF (u,_) -> "the text '" ^ u ^ "'"
  | TEXTENTITYREF (u,_) -> "the text '" ^ u ^ "'"
  | NMTOKEN _ -> "an NMTOKEN"
  | PEREF _ -> "an entity reference"
  | BODYDECL -> "'['"
  | ENDDECL -> "']]>'"
  | BEGINDECL -> "'<!['"
  | NOTATIONDECL -> "'<!NOTATION'"
  | ENTITYDECL -> "'<!ENTITY'"
  | ATTLISTDECL -> "'<!ATTLIST'"
  | ELEMENTDECL -> "'<!ELEMENTDECL'"
  | PCDATA -> "'#PCDATA'"
  | FIXED -> "'#FIXED'"
  | IMPLIED -> "'#IMPLIED'"
  | REQUIRED -> "'#REQUIRED'"
  | PERCENT -> "'%'"
  | NAMESPACE -> "'namespace'"
  | DECLARENAMESPACE -> "'declare namespace'"
  | DEFAULTELEMENT -> "'default element'"
  | DEFAULTFUNCTION -> "'default function'"
  | DEFINEKEY -> "'default key'"
  | DEFINEFUNCTION -> "'default function'"
  | DEFINESIGNATURE -> "'default signature'"
  | DEFINEGLOBAL -> "'default global'"
  | IMPORTSCHEMA -> "'import schema'"
  | DEFINETYPE _ -> "'define type'"
  | DEFINEELEMENT _ -> "'define element'"
  | DEFINEATTRIBUTE _ -> "'define attribute'"
  | DOCUMENTOF -> "'document of'"
  | ELEMENT -> "'element'"
  | ELEMENTOF -> "'element of'"
  | ATTRIBUTE -> "'attribute'"
  | DOCUMENT -> "'document'"
  | EMPTY -> "'empty'"
  | TYPE -> "'type'"
  | NONE -> "'none'"
  | QUESTION -> "'?'"
  | NODE -> "'node'"
  | ITEM -> "'item'"
  | TEXTNODE -> "'text'"
  | ATOMIC -> "'atomic'"
  | NUMERIC -> "'numeric'"
  | STABLE -> "'stable'"
  | ASCENDING -> "'ascending'"
  | DESCENDING -> "'descending'"
  | EMPTYGREATEST -> "'empty greatest'"
  | EMPTYLEAST -> "'empty least'"
  | OR -> "'or'"
  | AND -> "'and'"
  | STAR -> "'*'"
  | MULT -> "'*'"
  | DIV -> "'/'"
  | IDIV -> "'idiv'"
  | MOD -> "'mod'"
  | MINUS -> "'-'"
  | PLUS -> "'+'"
  | IN -> "'in'"
  | SATISFIES -> "'satisfies'"
  | RETURN -> "'return'"
  | THEN -> "'then'"
  | ELSE -> "'else'"
  | TO -> "'to'"
  | WHERE -> "'where'"
  | END -> "'end'"
  | INTERSECT -> "'intersect'"
  | UNION -> "'union'"
  | EXCEPT -> "'except'"
  | PRECEDES -> "'<<'"
  | FOLLOWS -> "'>>'"
  | CASE -> "'case'"
  | INSTANCEOF -> "'instance of'"
  | DEFAULT -> "'default'"
  | IF -> "'if'"
  | TYPESWITCH -> "'typeswitch'"
  | STABLESORTBY -> "'stable'"
  | SORTBY -> "'sort by'"
  | ORDERBY -> "'order by'"
  | ASSERTAS -> "'assert as'"
  | CASTAS -> "'cast as'"
  | CASTABLEAS -> "'castable as'"
  | TREATAS -> "'treat as'"
  | VALIDATEAS -> "'validate as'"
  | SOME -> "'some'"
  | EVERY -> "'every'"
  | FOR -> "'for'"
  | LET -> "'let'"
  | AS -> "'as'"
  | AT -> "'at'"
  | INSERT -> "'insert'"
  | REPLACE -> "'replace'"
  | DELETE -> "'delete'"
  | ASFIRST -> "'as first'"
  | ASLAST -> "'as last'"
  | INTO -> "'into'"
  | WITH -> "'with'"
  | VALUEOF -> "'value of'"
  | BEFORE -> "'before'"
  | AFTER -> "'after'"
  | UPDATE -> "'update'"
  | NCNAME ncname ->  "the QName: " ^ ncname
  | QNAME uqname -> "the QName: " ^ (Namespace.string_of_uqname uqname)
  | AXIS _ -> "an Axis"
  | PROCESSINGINSTRUCTIONLPAR -> "processing-instruction("
  | COMMENTLPAR -> "comment("
  | TEXTLPAR -> "text("
  | NODELPAR -> "node("
  | COLON -> ":"
  | LTCOLON -> "<:"
  | BACKSLASHSLASH -> "\\"
  | SLASH -> "/"
  | SLASHSLASH -> "//"
  | DOT -> "."
  | DOTDOT -> ".."
  | EQ -> "eq"
  | EQUALS -> "="
  | NE -> "ne"
  | NOTEQUALS -> "!="
  | IS -> "is"
  | ISNOT -> "isnot"
  | LTE -> "le"
  | GTE -> "ge"
  | LTEQUALS -> "<="
  | GTEQUALS -> ">="
  | COLONEQUALS -> ":="
  | LTOP -> "lt"
  | GTOP -> "gt"
  | LT -> "<"
  | GT -> ">"
  | BAR -> "|"
  | ATSIGN -> "@"
  | COMMA -> ","
  | AMPERSAND -> "&"
  | SEMICOLON -> ";"
  | LPAR -> "("
  | RPAR -> ")"
  | RPARLPAR -> ")("
  | LBRACK -> "["
  | RBRACK -> "]"
  | LCURLY -> "{"
  | RCURLY -> "}"
  | DOLLAR -> "$"
  | FUNCTIONNAMELPAR uqname -> "the function call: " ^ (Namespace.string_of_uqname uqname) ^ "("
  | DECIMAL d -> "the decimal value: " ^ (Decimal._string_of_decimal d)
  | INT i -> "the integer value: " ^ (Decimal._string_of_integer i)
  | DOUBLE f -> "the double value: " ^ (string_of_float f)
  | STRING s -> "the string \"" ^ s ^ "\""
  | EOF -> "EOF"


(******************************)
(* Internal types and streams *)
(******************************)

(* Stream of tokens from the lexer *)

type token_stream =
    Xquery_parser.token Stream.t

(* Type used internally to build the final SAX even stream *)

type sax_struct =
    { mutable sax_local_buffer : sax_event option list;    (* Small buffer used to store events locally *)
      sax_stream       : Xquery_parser.token Stream.t;     (* Actual stream of tokens from the lexer *)
      sax_stack        : Namespace.uqname Stack.t;         (* Stack to check opening-closing tags *)
      sax_lexbuf       : Lexing.lexbuf;                    (* Keep a handle on the lexing buffer *)
      sax_resolver     : resolver option;                  (* Keep a handle to the PXP resolver *)
      mutable sax_local_stream : (Xquery_parser.token Stream.t) option;
      mutable sax_local_resolver : resolver option }

(* Basic operations on SAX structures *)

let sax_close_local_stream ss =
  match ss.sax_local_resolver with
  | Some r ->
      close_local_stream r;
      ss.sax_local_stream <- None;
      ss.sax_local_resolver <- None
  | None ->
      raise (Query (Internal_Error "Where is the local resolver??"))

let sax_close_local_stream_keep_string_buffer ss =
  match ss.sax_local_resolver with
  | Some r ->
      close_local_stream_keep_string_buffer r;
      ss.sax_local_stream <- None;
      ss.sax_local_resolver <- None
  | None ->
      raise (Query (Internal_Error "Where is the local resolver??"))

let sax_close_opt_resolver ro =
  match ro with
  | Some r ->
      Glx_pxp.close_resolver r
  | None ->
      ()
  

let sax_close_sax_struct ss =
  begin
    Stack.clear ss.sax_stack;
    sax_close_opt_resolver ss.sax_resolver
  end

let sax_raise_parsing_error ss msg =
  let lexbuf = ss.sax_lexbuf in
  let locinfo = Finfo.lexing_locinfo lexbuf in
  begin
    sax_close_sax_struct ss;
    raise (Query (Parsing (locinfo, msg)))
  end

let sax_raise_transition_error ss expected_token_message actual_token =
  let msg =
    "Was expecting a "
    ^ expected_token_message
    ^ " but got a "
    ^ (string_of_token actual_token)
    ^ " instead"
  in
  sax_raise_parsing_error ss msg 

let sax_next ss =
  try
    match ss.sax_local_stream with
    | None ->
	Stream.next ss.sax_stream
    | Some local_stream ->
	begin
	  try
	    Stream.next local_stream
	  with
	  | Stream.Failure ->
	      begin
		sax_close_local_stream ss;
		Stream.next ss.sax_stream
	      end
	end
  with
  | Stream.Failure ->
      sax_raise_parsing_error ss "Unexpected end of file"

let sax_next_keep_buffer ss =
  try
    match ss.sax_local_stream with
    | None ->
	Stream.next ss.sax_stream
    | Some local_stream ->
	begin
	  try
	    Stream.next local_stream
	  with
	  | Stream.Failure ->
	      begin
		sax_close_local_stream_keep_string_buffer ss;
		Stream.next ss.sax_stream
	      end
	end
  with
  | Stream.Failure ->
      sax_raise_parsing_error ss "Unexpected end of file"

let sax_skip ss =
  ignore(sax_next ss)

let rec sax_next_discard_whitespace ss =
  match sax_next ss with
  | S -> sax_next_discard_whitespace ss
  | x -> x

let sax_peek ss =
  match ss.sax_local_stream with
  | None ->
      Stream.peek ss.sax_stream
  | Some local_stream ->
      begin
	match Stream.peek local_stream with
	| None ->
	    begin
	      sax_close_local_stream ss;
	      Stream.peek ss.sax_stream
	    end
	| event ->
	    event
      end

let rec sax_peek_discard_whitespace ss =
  match sax_peek ss with
  | Some S ->
      begin
	sax_skip ss;
	sax_peek_discard_whitespace ss
      end
  | x -> x

let sax_pop ss =
  Stack.pop ss.sax_stack

let sax_push qname ss =
  Stack.push qname ss.sax_stack

let sax_encoding ss enc_string =
  begin
    match ss.sax_resolver with
    | Some resolver ->
	change_encoding resolver enc_string
    | None ->
	()
  end;
  encoding_of_string enc_string
  

(************************)
(* Stream of SAX events *)
(************************)

type doc_stream =
    (sax_event Stream.t * sax_struct)

let next_xml_event (sax_stream,ss) =
  try
    Stream.next sax_stream
  with
  | Stream.Failure ->
      sax_close_sax_struct ss;
      raise Stream.Failure


(*******************************)
(* XML lexer stream operations *)
(*******************************)

let xml_lexer_stream_fun_build lexbuf n =
  let lexer_token = xml_lexfun lexbuf in
  match lexer_token with
  | EOF -> None
  | _ -> Some lexer_token

(* Opens a new document from a file *)

let open_doc_stream_from_file fname =
  (* Initialize the lexing environment for XML parsing *)
  Lexutil.init_xml_lexing ();

  let parse_context = Parse_context.build_parse_context () in

  (* Make sure the finfo locator know which file is processed *)
  Finfo.current_file := fname;

  (* This is where the lexing buffer is created using the nice PXP reader *)
  let resolver = resolve_as_file () in
  init_resolver resolver (Encoding.get_internal_encoding ());
  let lexbuf = open_resolver resolver fname in
  let xml_lexer_stream_fun =
    xml_lexer_stream_fun_build lexbuf
  in
  (Stream.from xml_lexer_stream_fun, resolver, lexbuf, parse_context)

(* Open a new document from a string *)

let open_doc_stream_from_string s =
  Lexutil.init_xml_lexing ();

  let parse_context = Parse_context.build_parse_context () in
  (* This is where the lexing buffer is created using the nice PXP reader *)
  let resolver = resolve_as_string s in
  init_resolver resolver (Encoding.get_internal_encoding ());
  let lexbuf = open_resolver resolver s in
(*  let lexbuf = Lexing.from_string s in *)
  let xml_lexer_stream_fun =
    xml_lexer_stream_fun_build lexbuf
  in
  (Stream.from xml_lexer_stream_fun,lexbuf,parse_context)

(**************************************)
(* XML patterns recognition functions *)
(**************************************)

(* Recognizes a list of attributes *)

let rec recognize_attribute_content parse_context ss previous_content =
  let next_lexer_token = sax_next_keep_buffer ss in
  match next_lexer_token with
  | ATTRIBUTETEXT text ->
      previous_content ^ text
  | TEXTENTITYREF (text, entityref) ->
      (* Note:
	   In that case, the we need to do the parsing with the
	   special case that single/double quote are not treated as
	   markup.

	   Here is the corresponding section of the XML 1.0 spec.

         --------------------------------------------------------------
	 4.4.5 Included in Literal

	 When an entity reference appears in an attribute value, or a
	 parameter entity reference appears in a literal entity value,
	 its replacement text is processed in place of the reference
	 itself as though it were part of the document at the location
	 the reference was recognized, except that a single or double
	 quote character in the replacement text is always treated as
	 a normal data character and will not terminate the
	 literal. For example, this is well-formed:

	 <!--  -->
	 <!ENTITY % YN '"Yes"' >
	 <!ENTITY WhatHeSaid "He said %YN;" >

	 while this is not:

	 <!ENTITY EndAttr "27'" >
	 <element attribute='a-&EndAttr;>
         --------------------------------------------------------------

	 - Jerome
	 *) 
      let lexbuf = ss.sax_lexbuf in
      let locinfo = Finfo.lexing_locinfo lexbuf in
      let entity_ref_text = Parse_context.get_general_entity parse_context locinfo entityref in
      let real_entity_ref_text = compute_replacement_text parse_context locinfo entity_ref_text in
      let push_fun = Lexutil.push_entity_included_in_literal in
      let (local_stream, local_resolver) = open_local_stream_from_string push_fun real_entity_ref_text in
      (* Switch the parsing to apply on the entity directly *)
      begin
	ss.sax_local_stream <- Some local_stream;
	ss.sax_local_resolver <- Some local_resolver;
	recognize_attribute_content parse_context ss (previous_content ^ text)
      end
  | TEXTCHARREF (text, charref) ->
      let char_ref_text = Glx_pxp.character (Encoding.get_internal_encoding ()) charref in
      let new_attribute_content = previous_content ^ text ^ char_ref_text in
      recognize_attribute_content parse_context ss new_attribute_content
  | wrong_token ->
      sax_raise_transition_error ss "string" wrong_token

let rec recognize_attributes parse_context ss =
  let next_lexer_token = sax_next_discard_whitespace ss in
  match next_lexer_token with
  | ROPENINGTAG ->
      ([],[])
  | REMPTYELEMENT ->
      begin
	(* In the case of elements with empty content: pop to make the stack happy! *)
	ignore(sax_pop ss);
	([], (Some SAX_endElement) :: [])
      end
  | QNAME qname ->
      let equal_token  = sax_next_discard_whitespace ss in
      let string_token = sax_peek_discard_whitespace ss in
      begin
	match (equal_token, string_token) with 
	| (EQUALS, Some _) ->
	    let attribute_content =
	      recognize_attribute_content parse_context ss ""
	    in
	    let (more_attributes,end_tokens) = recognize_attributes parse_context ss in
	    ((qname,attribute_content) :: more_attributes,end_tokens)
  (* The following are transition errors - Jerome *)
	| (EQUALS, None) ->
	    sax_raise_transition_error ss "string" EOF
	| (wrong_token, _) ->
	    sax_raise_transition_error ss "'='" wrong_token
      end
  | wrong_token ->
      sax_raise_transition_error ss "'>'" wrong_token

(* Recognizes an opening tag *)

let recognize_start_element parse_context ss =
  let opening_tag = sax_next ss in
  let qname       = sax_next ss in
  match (opening_tag, qname) with
  | (LOPENINGTAG, QNAME qname) ->
      begin
	sax_push qname ss;  (* adds one element tag in the stack *)
	let (attributes,other_tokens) = recognize_attributes parse_context ss in
	(Some (SAX_startElement (qname, attributes, false))) :: other_tokens
      end
  | (TEXTLOPENINGTAG text, QNAME qname) ->
      begin
	sax_push qname ss;  (* adds one element tag in the stack *)
	let (attributes,other_tokens) = recognize_attributes parse_context ss in
	(* Remove empty text nodes *)
	(Some (SAX_characters text)) :: (Some (SAX_startElement (qname, attributes, false))) :: other_tokens
      end
  (* The following are transition errors - Jerome *)
  | (LOPENINGTAG, wrong_token)
  | (TEXTLOPENINGTAG _, wrong_token) ->
      sax_raise_transition_error ss "QName" wrong_token
  | (wrong_token,_) ->
      sax_raise_transition_error ss "'<'" wrong_token

(* Recognizes a closing tag *)

let recognize_end_element parse_context ss =
  let lclosing_tag = sax_next ss in
  let qname = sax_next ss in
  let rclosing_tag = sax_next_discard_whitespace ss in
  match (lclosing_tag, qname, rclosing_tag) with
  | (LCLOSINGTAG, QNAME qname, RCLOSINGTAG) ->
      begin
	let corresponding_opening_tag = sax_pop ss in
	if not(qname = corresponding_opening_tag)
	then
	  sax_raise_parsing_error ss ("opening tag '"
				      ^ (Namespace.string_of_uqname corresponding_opening_tag)
				      ^ "' and closing tag '"
				      ^ (Namespace.string_of_uqname qname)
				      ^ "' mismatched")
	else
	  (Some SAX_endElement) :: []
      end
  | (TEXTLCLOSINGTAG text, QNAME qname, RCLOSINGTAG) ->
      begin
	let corresponding_opening_tag = sax_pop ss in
	if not(qname = corresponding_opening_tag)
	then
	  sax_raise_parsing_error ss ("opening tag '"
				      ^ (Namespace.string_of_uqname corresponding_opening_tag)
				      ^ "' and closing tag '"
				      ^ (Namespace.string_of_uqname qname)
				      ^ "' mismatched")
	else
	  (Some (SAX_characters text)) :: (Some SAX_endElement) :: []
      end
  (* The following are transition errors - Jerome *)
  | (LCLOSINGTAG, QNAME _, wrong_token)
  | (TEXTLCLOSINGTAG _, QNAME _, wrong_token) ->
      sax_raise_transition_error ss "'>'" wrong_token
  | (LCLOSINGTAG, wrong_token, _)
  | (TEXTLCLOSINGTAG _, wrong_token, _) ->
      sax_raise_transition_error ss "QName" wrong_token
  | (wrong_token,_,_) ->
      sax_raise_transition_error ss "'</'" wrong_token

(* Recognizes an character reference *)

let recognize_char_ref parse_context ss =
  let char_ref_token = sax_next ss in
  match char_ref_token with
  | TEXTCHARREF (text, charref) ->
      let charref_text = Glx_pxp.character (Encoding.get_internal_encoding ()) charref in
      (Some (SAX_characters (text ^ charref_text))) :: []
  | wrong_token ->
      sax_raise_transition_error ss "'a character reference'" wrong_token

(* Recognizes an entity reference *)

let recognize_entity_ref parse_context ss =
  let char_ref_token = sax_next ss in
  match char_ref_token with
  | TEXTENTITYREF (text, entityref) ->
      let lexbuf = ss.sax_lexbuf in
      let locinfo = Finfo.lexing_locinfo lexbuf in
      let entity_ref_text = Parse_context.get_general_entity parse_context locinfo entityref in
      let real_entity_ref_text = compute_replacement_text parse_context locinfo entity_ref_text in
      let push_fun = Lexutil.push_entity_included in
      let (local_stream, local_resolver) = open_local_stream_from_string push_fun real_entity_ref_text in
      (* Switch the parsing to apply on the entity directly *)
      begin
	ss.sax_local_stream <- Some local_stream;
	ss.sax_local_resolver <- Some local_resolver;
	(Some (SAX_characters text)) :: []
      end
  | wrong_token ->
      sax_raise_transition_error ss "'a character reference'" wrong_token

(* Recognizes a comment tag *)

let recognize_comment parse_context ss =
  let opening_comment = sax_next ss in
  let closing_comment = sax_next ss in
  match (opening_comment, closing_comment) with
  | (LOPENINGCOMMENT, TEXTRCLOSINGCOMMENT text) ->
      if !Conf.xml_pis_and_comments then
	(Some (SAX_comment text)) :: []
      else
	[]
  | (TEXTLOPENINGCOMMENT text1, TEXTRCLOSINGCOMMENT text2) ->
      let comment_tokens =
	if !Conf.xml_pis_and_comments then
	  (Some (SAX_comment text2)) :: []
	else
	  []
      in
      (Some (SAX_characters text1)) :: comment_tokens
  (* The following are transition errors - Jerome *)
  | (LOPENINGCOMMENT, wrong_token)
  | (TEXTLOPENINGCOMMENT _, wrong_token) ->
      sax_raise_transition_error ss "'-->'" wrong_token
  | (wrong_token, _) ->
      sax_raise_transition_error ss "'<!--'" wrong_token

(* Recognizes a pi tag *)

let recognize_pi parse_context ss =
  let openingpi = sax_next ss in
  let closingpi = sax_next ss in
  match (openingpi, closingpi) with
  | (OPENINGPI pi_ncname, TEXTCLOSINGPI text) ->
      if !Conf.xml_pis_and_comments then
	(Some (SAX_processingInstruction (pi_ncname, text))) :: []
      else
	[]
  | (TEXTOPENINGPI (text1,pi_ncname), TEXTCLOSINGPI text2) ->
      let pi_token =
        if !Conf.xml_pis_and_comments then
	  (Some (SAX_processingInstruction (pi_ncname, text2))) :: []
	else
	  []
      in
      (Some (SAX_characters text1)) :: pi_token
  (* The following are transition errors - Jerome *)
  | (OPENINGPI _, wrong_token)
  | (TEXTOPENINGPI _, wrong_token) ->
      sax_raise_transition_error ss "'?>'" wrong_token
  | (wrong_token, _) ->
      sax_raise_transition_error ss "'<?'" wrong_token

(* General function which recognizes the next event tag *)

let rec recognize_next parse_context ss =
  let next_lexer_token = sax_peek_discard_whitespace ss in
  match next_lexer_token with
  | None ->
      (* Cool, this is normal termination here - Jerome *)
      begin
	sax_close_sax_struct ss;
	Some (SAX_endDocument) :: None :: []
      end
  | Some LOPENINGTAG
  | Some (TEXTLOPENINGTAG _) ->
      (recognize_start_element parse_context ss)
  | Some (TEXTCHARREF _) ->
      (recognize_char_ref parse_context ss)
  | Some (TEXTENTITYREF _) ->
      (recognize_entity_ref parse_context ss)
  | Some LCLOSINGTAG
  | Some (TEXTLCLOSINGTAG _) ->
      (recognize_end_element parse_context ss)
  | Some LOPENINGCOMMENT
  | Some (TEXTLOPENINGCOMMENT _) ->
      (recognize_comment parse_context ss)
  | Some (OPENINGPI _)
  | Some (TEXTOPENINGPI _) ->
      (recognize_pi parse_context ss)
  | Some wrong_token ->
      sax_raise_transition_error ss "'<'" wrong_token


(*******************)
(* DTD Declaration *)
(*******************)

type external_id =
  | System of string
  | Public of string * string

let analyze_system_id parse_context ss =
  let sys_token = sax_next ss in
  let system_literal = sax_next_discard_whitespace ss in
  match (sys_token,system_literal) with
  | (DOCTYPE_SYSTEM, STRING system) ->
      System system
      (* The following are transition errors - Jerome *)
  | (DOCTYPE_SYSTEM, wrong_token) ->
      sax_raise_transition_error ss "a system literal" wrong_token
  | (wrong_token, _) ->
      sax_raise_transition_error ss "'SYSTEM'" wrong_token

let analyze_public_id parse_context ss =
  let sys_token = sax_next ss in
  let pubid_literal = sax_next_discard_whitespace ss in
  let system_literal = sax_next_discard_whitespace ss in
  match (sys_token,pubid_literal,system_literal) with
  | (DOCTYPE_PUBLIC, STRING pubid, STRING system) ->
      Public (pubid,system)
      (* The following are transition errors - Jerome *)
  | (DOCTYPE_PUBLIC, STRING _, wrong_token) ->
      sax_raise_transition_error ss "a system literal" wrong_token
  | (DOCTYPE_PUBLIC, wrong_token, _) ->
      sax_raise_transition_error ss "a pubid" wrong_token
  | (wrong_token, _, _) ->
      sax_raise_transition_error ss "'PUBLIC'" wrong_token

let analyze_doctypeclose parse_context ss =
  let doctypeclose_token = sax_next ss in
  match doctypeclose_token with
  | DOCTYPECLOSE ->
      ()
  | wrong_token ->
      sax_raise_transition_error ss "'>'" wrong_token

let recognize_entity_value parse_context ss =
  let attribute_token = sax_next ss in
  let end_token = sax_next_discard_whitespace ss in
  match (attribute_token, end_token) with
  | (ATTRIBUTETEXT v, ROPENINGTAG) ->
      v
  | (ATTRIBUTETEXT v, wrong_token) ->
      sax_raise_transition_error ss "'>'" wrong_token
  | (wrong_token,_) ->
      sax_raise_transition_error ss "'An entity value'" wrong_token

let analyze_pe_def parse_context ss name =
  let next_token = sax_peek ss in
  match next_token with
  | Some ATTRIBUTETEXT _ ->
      begin
	let entity_value = recognize_entity_value parse_context ss in
	Parse_context.add_parameter_entity_to_parse_context parse_context name entity_value
      end
  | Some (NCNAME "SYSTEM") ->
      raise (Query (Prototype "SYSTEM not supported in parameter entities"))
  | Some wrong_token ->
      sax_raise_transition_error ss "'\"'" wrong_token
  | None ->
      sax_raise_transition_error ss "'\"'" EOF

let analyze_pe_decl parse_context ss =
  let percent_token = sax_next ss in
  let whitespace_token1 = sax_next ss in
  let name_token = sax_next ss in
  let whitespace_token2 = sax_next ss in
  match (percent_token,whitespace_token1,name_token,whitespace_token2) with
  | (PERCENT,S,NCNAME name,S) ->
      analyze_pe_def parse_context ss name
  | (PERCENT,S,NCNAME _,wrong_token) ->
      sax_raise_transition_error ss "' '" wrong_token
  | (PERCENT,S,wrong_token,_) ->
      sax_raise_transition_error ss "'An NCNAME'" wrong_token
  | (PERCENT,wrong_token,_,_) ->
      sax_raise_transition_error ss "' '" wrong_token
  | (wrong_token,_,_,_) ->
      sax_raise_transition_error ss "'%'" wrong_token

let analyze_entity_def parse_context ss name =
  let next_token = sax_peek ss in
  match next_token with
  | Some (ATTRIBUTETEXT _) ->
      begin
	let entity_value = recognize_entity_value parse_context ss in
	Parse_context.add_general_entity_to_parse_context parse_context name entity_value
      end
  | Some (NCNAME "SYSTEM") ->
      raise (Query (Prototype "SYSTEM not supported in general entities"))
  | Some wrong_token ->
      sax_raise_transition_error ss "'\"'" wrong_token
  | None ->
      sax_raise_transition_error ss "'\"'" EOF

let analyze_ge_decl parse_context ss =
  let name_token = sax_next ss in
  let whitespace_token = sax_next ss in
  match (name_token, whitespace_token) with
  | (NCNAME name,S) ->
      analyze_entity_def parse_context ss name
  | (NCNAME _, wrong_token) ->
      sax_raise_transition_error ss "' '" wrong_token
  | (wrong_token,_) ->
      sax_raise_transition_error ss "'An NCNAME'" wrong_token

let analyze_entity_decl parse_context ss =
  begin
    ignore(sax_next ss);
    match sax_next ss with
    | S -> ()
    | wrong_token ->
	sax_raise_transition_error ss "' '" wrong_token
  end;
  let next_token = sax_peek ss in
  match next_token with
  | Some PERCENT ->
      analyze_pe_decl parse_context ss
  | Some (NCNAME _) ->
      analyze_ge_decl parse_context ss
  | Some wrong_token ->
      sax_raise_transition_error ss "'%'" wrong_token
  | None ->
      sax_raise_transition_error ss "'%'" EOF

let rec analyze_declarations parse_context ss =
  let next_lexer_token = sax_peek_discard_whitespace ss in
  match next_lexer_token with
  | Some DTDDECLCLOSE ->
      ignore(sax_next_discard_whitespace ss)
  | _ ->
      begin
	match next_lexer_token with
	| Some ENTITYDECL ->
	    analyze_entity_decl parse_context ss;
	| Some LOPENINGCOMMENT ->
	    ignore(recognize_comment parse_context ss)
	| Some PEREF _
	| Some PERCENT
	| Some NCNAME _
	| Some NMTOKEN _
	| Some PLUS
	| Some STAR
	| Some BAR
	| Some COMMA
	| Some QUESTION
	| Some LPAR
	| Some RPAR
	| Some REQUIRED
	| Some IMPLIED
	| Some FIXED
	| Some PCDATA
	| Some ELEMENTDECL
	| Some ATTLISTDECL
	| Some NOTATIONDECL
	| Some BEGINDECL
	| Some ENDDECL
	| Some BODYDECL
	| Some OPENINGPI _
	| Some ATTRIBUTETEXT _
	| Some ROPENINGTAG ->
	    ignore(sax_next_discard_whitespace ss)
	| Some wrong_token ->
	    sax_raise_transition_error ss "']'" wrong_token
	| None ->
	    sax_raise_transition_error ss "']'" EOF
      end;
      analyze_declarations parse_context ss


let analyze_doctypeclose parse_context ss =
  let doctype_token =  sax_next_discard_whitespace ss in
  match doctype_token with
  | DOCTYPECLOSE ->
      ()
  | wrong_token ->
      sax_raise_transition_error ss "'>'" wrong_token

let recognize_declarations parse_context ss =
  let next_lexer_token = sax_peek_discard_whitespace ss in
  match next_lexer_token with
  | Some DTDDECLOPEN ->
      sax_skip ss;
      let _ = analyze_declarations parse_context ss in
      analyze_doctypeclose parse_context ss
  | Some DOCTYPECLOSE ->
      analyze_doctypeclose parse_context ss
  | Some wrong_token ->
      sax_raise_transition_error ss "'['" wrong_token
  | None ->
      sax_raise_transition_error ss "'>'" EOF

let recognize_external_id parse_context ss =
  let next_lexer_token = sax_peek_discard_whitespace ss in
  match next_lexer_token with
  | Some DOCTYPE_SYSTEM ->
      let extid = analyze_system_id parse_context ss in
      (Some extid, recognize_declarations parse_context ss)
  | Some DOCTYPE_PUBLIC ->
      let extid = analyze_public_id parse_context ss in
      (Some extid, recognize_declarations parse_context ss)
  | _ ->
      (None, recognize_declarations parse_context ss)

(* Recognizes a DTD *)

let analyze_doctype parse_context ss =
  let doctype      = sax_next ss in
  let doctype_name = sax_next_discard_whitespace ss in
  match (doctype,doctype_name) with
  | (DOCTYPE, QNAME dtd_name) ->
      (dtd_name, recognize_external_id parse_context ss)
      (* The following are transition errors - Jerome *)
  | (DOCTYPE, wrong_token) ->
      sax_raise_transition_error ss "a QName" wrong_token
  | (wrong_token, _) ->
      sax_raise_transition_error ss "'<!DOCTYPE'" wrong_token

let rec recognize_dtddecl parse_context xmldecl buffer ss =
  let next_lexer_token = sax_peek_discard_whitespace ss in
  match next_lexer_token with
  | Some DOCTYPE ->
      begin
	let doctype = analyze_doctype parse_context ss in
	(Some (SAX_startDocument (xmldecl, Some ()))) :: buffer
      end
  | Some LOPENINGCOMMENT
  | Some (TEXTLOPENINGCOMMENT _) ->
      let comment_buffer = (recognize_comment parse_context ss) in
      recognize_dtddecl parse_context xmldecl (buffer @ comment_buffer) ss
  | Some (OPENINGPI _)
  | Some (TEXTOPENINGPI _) ->
      let pi_buffer = (recognize_pi parse_context ss) in
      recognize_dtddecl parse_context xmldecl (buffer @ pi_buffer) ss
  | _ ->
      (Some (SAX_startDocument (xmldecl, None))) :: buffer

(*******************)
(* XML Declaration *)
(*******************)

let analyze_end_of_xmldecl ss =
  let next_lexer_token = sax_next_discard_whitespace ss in
  match next_lexer_token with
  | CLOSINGPI ->
      ()
  | wrong_token ->
      sax_raise_transition_error ss "'?>'" wrong_token

let analyze_sddecl ss =
  let next_lexer_token = sax_next_discard_whitespace ss in
  let quote1_token = sax_next_discard_whitespace ss in
  let declname_token = sax_next ss in
  let quote2_token = sax_next ss in
  match (next_lexer_token,quote1_token,declname_token,quote2_token) with
  | (XSDDECL, XSINGLEQUOTE, XDECLNAME sddecl, XSINGLEQUOTE) ->
      begin
        if (sddecl = "yes") || (sddecl = "no")
	then
	  sddecl
	else
	  sax_raise_parsing_error ss ("Wrong declaration: " ^ sddecl)
      end
        (* The following are transition errors - Jerome *)
  | (XSDDECL, XDOUBLEQUOTE, XDECLNAME sddecl, XDOUBLEQUOTE) ->
      begin
        if (sddecl = "yes") || (sddecl = "no")
	then
	  sddecl
	else
	  sax_raise_parsing_error ss ("Wrong declaration: " ^ sddecl)
      end
        (* The following are transition errors - Jerome *)
  | (XSDDECL,XSINGLEQUOTE,XDECLNAME _,wrong_token) ->
      sax_raise_transition_error ss "'" wrong_token
  | (XSDDECL,XDOUBLEQUOTE,XDECLNAME _,wrong_token) ->
      sax_raise_transition_error ss "\"" wrong_token
  | (XSDDECL,XSINGLEQUOTE, wrong_token, _) ->
      sax_raise_transition_error ss "an XML declaration" wrong_token
  | (XSDDECL,XDOUBLEQUOTE, wrong_token, _) ->
      sax_raise_transition_error ss "an XML declaration" wrong_token
  | (XSDDECL,wrong_token, _, _) ->
      sax_raise_transition_error ss "' or \"" wrong_token
  | (wrong_token, _, _, _) ->
      sax_raise_transition_error ss "'standalone'" wrong_token

let recognize_sddecl ss =
  let next_lexer_token = sax_peek_discard_whitespace ss in
  match next_lexer_token with
  | Some XSDDECL ->
      let sddecl = analyze_sddecl ss in
      begin
	analyze_end_of_xmldecl ss;
	Some sddecl
      end
  | Some CLOSINGPI ->
      begin
	analyze_end_of_xmldecl ss;
	None
      end
  | Some wrong_token ->
      sax_raise_transition_error ss "'?>'" wrong_token
  | None ->
      sax_raise_transition_error ss "'?>'" EOF

let analyze_encoding ss =
  let next_lexer_token = sax_next_discard_whitespace ss in
  match next_lexer_token with
  | XENCODING enc_string ->
      sax_encoding ss enc_string
  | wrong_token ->
      sax_raise_transition_error ss "'encoding'" wrong_token

let recognize_encoding ss =
  let next_lexer_token = sax_peek_discard_whitespace ss in
  match next_lexer_token with
  | Some XENCODING _ ->
      let enc = analyze_encoding ss in
      let sddecl = recognize_sddecl ss in
      (Some enc, sddecl)
  | Some CLOSINGPI ->
      begin
	analyze_end_of_xmldecl ss;
	(None,None)
      end
  | Some _ ->
      (None,recognize_sddecl ss)
  | None ->
      sax_raise_transition_error ss "'?>'" EOF

let recognize_version ss =
  let quote1_token = sax_next ss in
  let declname_token = sax_next ss in
  let quote2_token = sax_next ss in
  match (quote1_token, declname_token, quote2_token) with
  | (XSINGLEQUOTE, XDECLNAME decl, XSINGLEQUOTE) ->
      if decl = "1.0"
      then
	decl
      else
	sax_raise_parsing_error ss ("XML version: " ^ decl ^ " not supported")
  | (XDOUBLEQUOTE, XDECLNAME decl, XDOUBLEQUOTE) ->
      if decl = "1.0"
      then
	decl
      else
	sax_raise_parsing_error ss ("XML version: " ^ decl ^ " not supported")
  (* The following are transition errors - Jerome *)
  | (XSINGLEQUOTE, XDECLNAME _, wrong_token) ->
      sax_raise_transition_error ss "'" wrong_token
  | (XDOUBLEQUOTE, XDECLNAME _, wrong_token) ->
      sax_raise_transition_error ss "\"" wrong_token
  | (XSINGLEQUOTE, wrong_token, _) ->
      sax_raise_transition_error ss "an XML declaration" wrong_token
  | (XDOUBLEQUOTE, wrong_token, _) ->
      sax_raise_transition_error ss "an XML declaration" wrong_token
  | (wrong_token, _, _) ->
      sax_raise_transition_error ss "' or \"" wrong_token

let recognize_xmldecl parse_context ss =
  let xmlpi   = sax_next ss in
  let version = sax_next_discard_whitespace ss in
  match (xmlpi,version) with
  | (XMLPI,XVERSION) ->
      let version = recognize_version ss in
      let (enc,decl) = recognize_encoding ss in
      (version, enc, decl)
	(* The following are transition errors - Jerome *)
  | (XMLPI, wrong_token) ->
      sax_raise_transition_error ss "'a version number'" wrong_token
  | (wrong_token, _) ->
      sax_raise_transition_error ss "'<?xml'" wrong_token

(******************)
(* The XML prolog *)
(******************)

let recognize_prolog parse_context ss =
  let next_lexer_token = sax_peek ss in
  match next_lexer_token with
  | Some XMLPI ->
      let xmldecl = recognize_xmldecl parse_context ss in
      recognize_dtddecl parse_context (Some xmldecl) [] ss
  | _ ->
      recognize_dtddecl parse_context None [] ss


(***************************)
(* Top level SAX functions *)
(***************************)

let rec next_event_internal parse_context t previous_event =
  match t.sax_local_buffer with
  | [] ->
      let new_local_buffer = recognize_next parse_context t in
      begin
	t.sax_local_buffer <- new_local_buffer;
	next_event_internal parse_context t previous_event
      end
  | (Some (SAX_characters characters)) :: rest ->
      begin
	t.sax_local_buffer <- rest;
	if characters = ""
	then
	  next_event_internal parse_context t previous_event
	else
	  match previous_event with
	  | None ->
	      next_event_internal parse_context t (Some characters)
	  | Some text ->
	      next_event_internal parse_context t (Some (text ^ characters))
      end
  | other_event :: rest ->
      begin
	match previous_event with
	| None ->
	    t.sax_local_buffer <- rest;
	    other_event
	| Some text ->
	    Some (SAX_characters text)
      end

let next_event parse_context t n =
  next_event_internal parse_context t None

let open_event_stream fname =
  let (actual_token_stream,resolver,lexbuf,parse_context) = open_doc_stream_from_file fname in
  let stack = Stack.create () in
  let stream_structure =
    { sax_local_buffer = [];
      sax_stream = actual_token_stream;
      sax_stack = stack;
      sax_lexbuf = lexbuf;
      sax_resolver = Some resolver;
      sax_local_stream = None;
      sax_local_resolver = None }
  in
  let start_buffer = recognize_prolog parse_context stream_structure in
  stream_structure.sax_local_buffer <- start_buffer;
  (Stream.from (next_event parse_context stream_structure), stream_structure)

let open_internal_xml_stream_from_file fname =
  let (actual_token_stream,resolver,lexbuf,parse_context) = open_doc_stream_from_file fname in
  let stack = Stack.create () in
  let stream_structure=
    { sax_local_buffer = [];
      sax_stream = actual_token_stream;
      sax_stack = stack;
      sax_lexbuf = lexbuf;
      sax_resolver = Some resolver;
      sax_local_stream = None;
      sax_local_resolver = None }
  in
  let start_buffer = recognize_prolog parse_context stream_structure in
  stream_structure.sax_local_buffer <- start_buffer;
  (Stream.from (next_event parse_context stream_structure),stream_structure)

let open_internal_xml_stream_from_string string =
  let (actual_token_stream,lexbuf,parse_context) = open_doc_stream_from_string string in
  let stack = Stack.create () in
  let stream_structure=
    { sax_local_buffer = [];
      sax_stream = actual_token_stream;
      sax_stack = stack;
      sax_lexbuf = lexbuf;
      sax_resolver = None;
      sax_local_stream = None;
      sax_local_resolver = None }
  in
  let start_buffer = recognize_prolog parse_context stream_structure in
  stream_structure.sax_local_buffer <- start_buffer;
  (Stream.from (next_event parse_context stream_structure),stream_structure)


(************************************)
(* The top level stream constructor *)
(************************************)

let open_xml_stream_from_file file =
  let internal_doc_stream = open_internal_xml_stream_from_file file in
  let internal_next x =
    try
      Some (next_xml_event internal_doc_stream)
    with
    | Stream.Failure ->
	None
  in
  Stream.from internal_next

let open_xml_stream_from_string string =
  let internal_doc_stream = open_internal_xml_stream_from_string string in
  let internal_next x =
    try
      Some (next_xml_event internal_doc_stream)
    with
    | Stream.Failure ->
	None
  in
  Stream.from internal_next

