(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: sax_typesutil.ml,v 1.4 2003/06/16 15:34:34 simeon Exp $ *)

(* Module: Sax_typesutil
   Description:
     This module implements a few operations on SAX events.
*)

open Error

open Sax_types

open Namespace
open Datatypes


(************************)
(* SAX events accessors *)
(************************)

let get_start_document xml_stream =
  let event = Stream.next xml_stream in
  match event with
  | SAX_startDocument (sax_xml_decl,_) -> sax_xml_decl
  | _ ->
      raise (Query (Internal_Error ("Was expecting a start document event")))

let get_end_document xml_stream =
  let event = Stream.next xml_stream in
  match event with
  | SAX_endDocument -> ()      
  | _ ->
      raise (Query (Internal_Error ("Was expecting an end document event")))


(*******************************)
(* Extracts special attributes *)
(*******************************)

(* Process an xmlns attribute *)

let build_prefix_uri_pair prefix_thing content_thing =
  let ns_prefix =
    match prefix_thing with
    | Some ncame ->
	NSPrefix ncame
    | None ->
	NSDefaultElementPrefix
  in
  let uri =
    try
      (* Hack for now! - Jerome *)
      ignore(anyURI_of_untyped content_thing);
      NSUri content_thing
    with
    | _ ->
	let ns_attribute =
	  match prefix_thing with
	  | None ->
	      "xmlns"
	  | Some ncname ->
	      "xmlns:" ^ ncname
	in
	raise (Query (Mapping_Failure ("Namespace attribute " ^ ns_attribute ^ " does not contain a URI")))
  in
  (ns_prefix, uri)

(* This function extracts special attributes after parsing *)

let rec extract_special_attributes attributes =
  match attributes with
  | [] -> (Whitespace.Default, [], [])
  | att1 :: attributes' ->
      let (whitespace_mode, namespace_decls, other_attributes') =
	extract_special_attributes attributes'
      in
      match att1 with
	(* Note: in the case of xmlns attributes, those do not get
	   preserved in the final Infoset.
	   - Jerome *)
      | ((NSPrefix "xmlns", ncname), att_content) ->
	  let new_namespace_decl =
	    build_prefix_uri_pair (Some ncname) att_content
	  in
	  (whitespace_mode, new_namespace_decl :: namespace_decls, other_attributes')
      | ((NSDefaultElementPrefix, "xmlns"), att_content) ->
	  let new_namespace_decl =
	    build_prefix_uri_pair None att_content
	  in
	  (whitespace_mode, new_namespace_decl :: namespace_decls, other_attributes')
	    (* Note:
	         Other special attributes are preserved in the XML
	         Infoset.
               - Jerome *)
      | ((NSPrefix "xml", "space"), "preserve") ->
	  (Whitespace.Preserve, namespace_decls, att1 :: other_attributes')
      | ((NSPrefix "xml", "space"), "default") ->
	  (Whitespace.Default, namespace_decls, att1 :: other_attributes')
	    (* For normal attributes, just add them to the final list of attributes *)
      | _ ->
	  (whitespace_mode, namespace_decls, att1 :: other_attributes')


(***********************)
(* Generic SAX handler *)
(***********************)

type 'a sax_handler =
    { handle_start_document : 'a -> sax_xml_decl option * sax_dtd_decl -> 'a;
      handle_end_document : 'a -> unit -> 'a;
      handle_start_element : 'a -> element_desc -> 'a;
      handle_end_element : 'a -> uqname -> 'a;
      handle_processing_instruction : 'a -> pi_desc -> 'a;
      handle_comment : 'a -> comment_desc -> 'a;
      handle_characters : 'a -> text_desc -> 'a;
      handle_attribute : 'a -> sax_xml_attribute -> 'a;
      handle_atomic_value : 'a -> Datatypes._AtomicValue -> 'a }

let xml_generic_parse sax_handler xml_stream init_answer =
  let elem_name_stack =
    Stack.create ()
  in
  let rec parse old_answer =
    (* Process the next event *)
    try
      let xml_event = Stream.next xml_stream in
      match xml_event with
      | SAX_startDocument (xml_dec, dtd_dec) ->
	  let new_answer =
	    sax_handler.handle_start_document old_answer (xml_dec, dtd_dec)
	  in
	  parse new_answer
	    
      | SAX_endDocument ->
	  let new_answer =
	    sax_handler.handle_end_document old_answer ()
	  in
	  parse new_answer
	    
      | SAX_startElement (elem_uqname, att_el, has_elem_content) ->
	  let new_answer =
	    sax_handler.handle_start_element old_answer (elem_uqname, att_el, has_elem_content)
	  in
   	  begin
	    Stack.push elem_uqname elem_name_stack;
	    parse new_answer
	  end

      | SAX_endElement ->
	  let elem_uqname =
	    try
	      Stack.pop elem_name_stack
	    with
	    | _ ->
		raise (Query (Internal_Error "Empty element stack in SAX handler"))
	  in
	  let new_answer =
	    sax_handler.handle_end_element old_answer elem_uqname
	  in
	  parse old_answer

      | SAX_processingInstruction (target,content) ->
	  let new_answer =
	    sax_handler.handle_processing_instruction old_answer (target,content)
	  in
	  parse new_answer
	    
      | SAX_comment c ->
	  let new_answer =
	    sax_handler.handle_comment old_answer c
	  in
	  parse new_answer
	    
      | SAX_characters s ->
	  let new_answer =
	    sax_handler.handle_characters old_answer s
	  in
	  parse new_answer

      | SAX_attribute (attr_uqname,content) ->
	  let new_answer =
	    sax_handler.handle_attribute old_answer (attr_uqname,content)
	  in
	  parse new_answer

      | SAX_atomicValue av ->
	  let new_answer =
	    sax_handler.handle_atomic_value old_answer av
	  in
	  parse new_answer

    with
    | Stream.Failure ->
	old_answer
  in
  parse init_answer

