(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: stream_ops.ml,v 1.3 2003/06/09 15:28:27 simeon Exp $ *)

(* Module: Stream_ops
   Description:
     Some basic operations on XML streams.
*)

open Sax_types


(************************************)
(* Discards the current XML subtree *)
(************************************)

(* Note:
     The first operation populates namespace context even though
     parsing events are discarded.
   - Jerome *)

let rec discard_xml_stream_remember_namespaces_aux nsenv xml_stream =
  let xml_event = Stream.next xml_stream in
  match xml_event with
  | SAX_startDocument _ ->
      discard_xml_stream_remember_namespaces_aux nsenv xml_stream
  | SAX_endDocument ->
      ()    (* Terminates recursion for the children of an element.
	       This gives the hand back to the calling element node creation. *)
  | SAX_startElement (_, att_el, _) ->
      (* Then, extract the namespace attributes,
	 update the namespace environment accordingly,
	 and create local namespace nodes ?? *)
      let (_, new_nss, _) = Sax_typesutil.extract_special_attributes att_el in
      let nsenv' = Namespace.add_all_ns nsenv new_nss in
      begin
	discard_xml_stream_remember_namespaces_aux nsenv' xml_stream;
	discard_xml_stream_remember_namespaces_aux nsenv  xml_stream
      end
  | SAX_endElement ->
      ()    (* Terminates recursion for the children of an element.
	       This gives the hand back to the calling element node creation. *)
  | SAX_processingInstruction _ ->
      discard_xml_stream_remember_namespaces_aux nsenv xml_stream
  | SAX_comment _ ->
      discard_xml_stream_remember_namespaces_aux nsenv xml_stream
  | SAX_characters s ->
      discard_xml_stream_remember_namespaces_aux nsenv xml_stream
  | SAX_attribute a ->
      discard_xml_stream_remember_namespaces_aux nsenv xml_stream
  | SAX_atomicValue a ->
      discard_xml_stream_remember_namespaces_aux nsenv xml_stream

let discard_xml_stream_remember_namespaces xml_stream =
  let nsenv = Namespace.empty_nsenv () in
  discard_xml_stream_remember_namespaces_aux nsenv xml_stream

let rec discard_xml_stream xml_stream =
  let xml_event = Stream.next xml_stream in
  match xml_event with
  | SAX_startDocument _ ->
      discard_xml_stream xml_stream
  | SAX_endDocument ->
      ()    (* Terminates recursion for the children of an element.
	       This gives the hand back to the calling element node creation. *)
  | SAX_startElement _ ->
      begin
	discard_xml_stream xml_stream;
	discard_xml_stream xml_stream
      end
  | SAX_endElement ->
      ()    (* Terminates recursion for the children of an element.
	       This gives the hand back to the calling element node creation. *)
  | SAX_processingInstruction _ ->
      discard_xml_stream xml_stream
  | SAX_comment _ ->
      discard_xml_stream xml_stream
  | SAX_characters s ->
      discard_xml_stream xml_stream
  | SAX_attribute a ->
      discard_xml_stream xml_stream
  | SAX_atomicValue a ->
      discard_xml_stream xml_stream


(************************************************)
(* Turns a typed XML stream into an untyped one *)
(************************************************)

(* Note:
     Below, some namespace magic will have to occur eventually.
   - Jerome *)

let untype_attribute (rsym,content,_) =
  let (uri,ncname) = Sym.rattr_name rsym in
  let prefix = Sym.rattr_prefix rsym in
  let uqname = (prefix,ncname) in
  (uqname,content)

let untype_attributes typed_attributes =
  List.map untype_attribute typed_attributes

let untype_event typed_event =
  match typed_event with
  | TSAX_startDocument (xmldecl,dtddecl) ->
      Some (SAX_startDocument (xmldecl,dtddecl))
  | TSAX_endDocument ->
      Some (SAX_endDocument)
  | TSAX_startElement (rsym,typed_sax_xml_attributes,_,has_element_content) ->
      let (uri,ncname) = Sym.relem_name rsym in
      let prefix = Sym.relem_prefix rsym in
      let uqname = (prefix,ncname) in
      let sax_xml_attributes =
	untype_attributes typed_sax_xml_attributes
      in
      Some (SAX_startElement (uqname,sax_xml_attributes,has_element_content))
  | TSAX_endElement ->
      Some (SAX_endElement)
  | TSAX_processingInstruction pi_desc ->
      Some (SAX_processingInstruction pi_desc)
  | TSAX_comment comment_desc ->
      Some (SAX_comment comment_desc)
  | TSAX_characters text_desc ->
      Some (SAX_characters text_desc)
  | TSAX_namespace ->
      None
  | TSAX_attribute typed_sax_xml_attribute ->
      let sax_xml_attribute =
	untype_attribute typed_sax_xml_attribute
      in
      Some (SAX_attribute sax_xml_attribute)
  | TSAX_atomicValue av ->
      Some (SAX_atomicValue av)

   let rec untype_xml_stream_next_event typed_xml_stream =
  try
    let next_event = Stream.next typed_xml_stream in
    match untype_event next_event with
    | None ->
	untype_xml_stream_next_event typed_xml_stream
    | Some event ->
	Some event
  with
  | Stream.Failure ->
      None

let next_event typed_xml_stream n =
  untype_xml_stream_next_event typed_xml_stream
  
let rec untype_xml_stream typed_xml_stream =
  Stream.from (next_event typed_xml_stream)


