(***********************************************************************)
(*                                                                     *)
(*                               Galax                                 *)
(*                            XQuery Engine                            *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: validate.ml,v 1.27 2003/06/21 15:30:55 simeon Exp $ *)

open Format
open Print
open Error
open Xquery_ast


(***************************)
(* Configuration variables *)
(***************************)

let wf            = ref false     (* True if just checking well-formedness *)

let schema_file   = ref None      (* Schema against which to validate - required except with -wf option *)
let schema_root   = ref None      (* Schema root type from which to start validation *)
let schema_prefix = ref (Namespace.NSDefaultElementPrefix)  (* Schema namespace prefix *)

let typesyntax    = ref false     (* True if using XQuery type syntax, false is using XML Schema *)

let print      	  = ref false     (* True if pretty printing the the result back *)
let dm         	  = ref false     (* True if building the data model instance *) 

let serialization_kind_of_param ser =
  match String.lowercase ser with
  | "wf" -> Conf.Serialize_As_Well_Formed
  | "xquery" -> Conf.Serialize_As_XQuery
  | "canonical" -> Conf.Serialize_As_Canonical
  | _ -> raise (Query(Unknown("Expected: 'wf', 'xquery', or 'canonical'.  Found: "^ser)))


(****************)
(* Command-line *)
(****************)

let usage_msg = 
  sprintf "Usage: %s [-schema file -root qname [-prefix prefix]] [-typesyntax] [-serialize [wf/xquery]] [-xml-whitespace] [-xml-pic] [-dm] [-wf] file(s)" Sys.argv.(0)

let process_args () =
  let args = ref [] in
  Arg.parse
    [ "-serialize",  Arg.String   (fun ser -> print := true; Conf.serialize := serialization_kind_of_param ser), " Set serialization kind [wf or xquery]\n\n // Processing phases configuration\n";
      "-print-error-code",   Arg.Clear Conf.verbose_error, " Print only the error code instead of the full error message";
      "-output-encoding",   Arg.String (fun s -> Encoding.set_output_encoding (Glx_pxp.encoding_of_string s)), " Set the outputx encoding representation\n\n // Input context configuration\n";
      "-xml-whitespace",    Arg.Set Conf.xml_whitespace, " Preserves whitespace in XML documents";
      "-xml-pic",           Arg.Set Conf.xml_pis_and_comments, " Preserves PI's and comments in XML documents";
      "-dm", Arg.Set dm, "Also builds the data model instance";
      "-wf", Arg.Set wf, "Checks only for well-formedness";
      "-prefix", Arg.String (fun f  -> schema_prefix := (Namespace.NSPrefix f)), "Schema namespace prefix";
      "-typesyntax", Arg.Set typesyntax, "Uses XQuery type system syntax instead of XML Schema syntax";
      "-schema", Arg.String (fun x -> schema_file := Some x), "XML Schema file";
      "-root", Arg.String (fun x -> schema_root := Some x), "Schema root type" ]
    (fun arg -> args := arg :: !args) usage_msg;
  match !args with
  | [] -> failwith "No input XML file(s) specified"
  | fnames ->
      fnames

let initialization () =
  (* Then initialize the namespace environment and the schema *)
  if not(!wf)
  then
    raise (Query (Toplevel_Error "Sorry: XML Schema validation is disabled from this version of 'validate'"))

(*
    let (nse,local_schema) =
      match !schema_file with
      | None ->
	  raise (Query (Toplevel_Error "Cannot perform validation without a schema"))
      | Some s ->
	  let (nse,sdl) =
	    if !typesyntax
	    then
	      Toputils.convert_xquery_context s
	    else
	      Toputils.convert_schema s !schema_prefix
	  in
	  (nse, Toputils.load_schema (nse,sdl))
    in
    let local_root =
      match !schema_root with
      | None ->
	  raise (Query (Toplevel_Error "Cannot perform validation without a schema root type"))
      | Some r ->
	  try
	    let uqname = (Namespace.uqname_element_of_string r) in
	    let (_,rqname) = Resolve.resolve_element_qname nse uqname in
	    Sym.relem_symbol (fst uqname, rqname)
	  with
	  | Failure _ ->
	      raise (Query (Toplevel_Error "Root type parameter is not a QName"))
    in
    Schema.create_root local_schema (Some local_root);
    (nse,local_schema)
*)
  else
    (Namespace.empty_nsenv (), Schema.empty_schema ())

let process_file uri_string =
  begin
    (* Initialization *)
    let (nse,schema) = initialization () in
    
    (* Parsing, with optionally data model loading *)
    let uri = AnyURI._uri_of_string uri_string in
    
    if (!dm) then
      begin
	let apply_load_document uri_string =
	  let xml_stream = Sax_parser.open_xml_stream_from_file uri_string in
	  Load.load_xml_document_from_sax uri xml_stream
	in
	begin
	  let res =
	    Monitor.wrap_monitor (Monitor.Document_ParsingLoading_Phase uri) apply_load_document uri_string
	  in
	  if (!print)
	  then
	    (* OLD: - Jerome
	       Datamodel.fserialize_result Format.std_formatter res
	     *)
	    begin
              Serialization.serialize_datamodel res;
	      Format.print_newline ()
	    end
	  else
	    ()
	end
      end
    else
      begin
	let apply_load_document uri_string =
	  let xml_stream = Sax_parser.open_xml_stream_from_file uri_string in
	  if (!print)
	  then
	    begin
	      Serialization.serialize_xml_stream xml_stream;
	      Format.print_newline ()
	    end
	  else
	    Stream_ops.discard_xml_stream xml_stream
	in
	Monitor.wrap_monitor (Monitor.Document_ParsingLoading_Phase uri) apply_load_document uri_string
      end

    (* That's all! *)

  end


(********)
(* Main *)
(********)
    
let main () =
  let input_files = process_args() in
  List.iter process_file input_files

(* A catch-all exception handler *)

let exec func arg =
  try
    try
      func arg;
    with
    | Fn_error.Xquery_error msg ->
	Fn_error.downgrade_error msg
  with
  | e ->
      begin
	eprintf_error "  " e;
	fprintf err_formatter "@."
      end

(* Let's roll! *)

let _ =
  exec main()

