(***********************************************************************)
(*                                                                     *)
(*                               Galax                                 *)
(*                            XQuery Engine                            *)
(*                                                                     *)
(*  Copyright 2001-2003, Lucent Technologies.                          *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: xmlquery.ml,v 1.55 2003/06/27 20:34:24 simeon Exp $ *)

open Conf
open Format
open Print
open Error

(* Context input from command-line *)

let context_file = ref (None)
let context_item = ref (None)
let global_vars = ref []

let output_all = ref false

let separator =
  ref "\n\n------------------------------------------------------------------------------\n\n"

let usage_msg = 
  sprintf "Usage: %s [options] xquery-filename(s)

//
// Here is the list of available options:
//

// Misc options
" Sys.argv.(0)

let print_version () =
  begin
    printf "\n";
    printf "  This is the %s XQuery 1.0 and XPath 2.0 interpretor, version %s\n" system version;
    printf "  %s\n\n" copyright;
    printf "  Status of this release is: %s\n" status;
    printf "  Galax library directory: %s\n\n" galax_library
  end

let bool_of_onoff onoff = 
  match String.lowercase onoff with
  | "on" -> true
  | "off" -> false
  | _ -> raise (Query(Unknown("Expected: 'on' or 'off'.  Found: "^onoff)))

let serialization_kind_of_param ser =
  match String.lowercase ser with
  | "wf" -> Serialize_As_Well_Formed
  | "xquery" -> Serialize_As_XQuery
  | "canonical" -> Serialize_As_Canonical
  | _ -> raise (Query(Unknown("Expected: 'wf', 'xquery', or 'canonical'.  Found: "^ser)))

let init_output_refs f f_out f_formatter = 
  f_out := open_out(f); 
  f_formatter := Format.formatter_of_out_channel (!f_out)

(* Check that configuration flags are compatible *)
let check_config () = 
  begin
    if !typing && not(!mapping) then
      raise (Query(Wrong_Args("Static typing requires normalization"))); 
    if !optim && not(!mapping) then 
      raise (Query(Wrong_Args("Optimization requires normalization"))); 
    if !projection && !optimized_projection then 
      raise (Query(Wrong_Args("Cannot apply projection and optimized projection both on at the same time"))); 
    if !eval && not(!mapping) then 
      raise (Query(Wrong_Args("Dynamic evaluation requires normalization"))); 
    if !print_type && not(!typing) then
      raise (Query(Wrong_Args("Cannot print type without static typing on"))); 
    if !print_core_expr && not(!mapping) then
      raise (Query(Wrong_Args("Cannot print normalized expression without normalization on"))); 
    if !print_optimized_expr && not(!optim) then
      raise (Query(Wrong_Args("Cannot pring optimized expresstion without optimization on"))); 
    if !print_typed_expr && not(!typing) then
      raise (Query(Wrong_Args("Cannot print typed expression without static typing on"))); 
    if !print_xml && not(!eval) then
      raise (Query(Wrong_Args("Cannot print xml result without dynamic evaluation on")))
  end
 
let process_args () =
  let args = ref [] in
  Arg.parse [

   (* Misc options *)
  "-version",        Arg.Unit (fun () -> print_version (); exit 0), " Prints the Galax version";
  "-debug",          Arg.Set debugging, " Emit debugging" ;
  "-sbdo",           Arg.String (fun onoff -> do_sbdo := bool_of_onoff(onoff)), " Turns off duplicate elimination and sorting (distinct-doc-order)";
  "-monitor", 	     Arg.String (fun onoff -> monitor := bool_of_onoff(onoff)), " Monitors memory and CPU consumption";
  "-output-monitor", Arg.String (fun f -> init_output_refs f monitor_output monitor_formatter), " Output monitor actibity to file" ;
  "-pretty-print-module", Arg.Set print_module, " Pretty printing the whole module";
  "-pretty-print",   Arg.String (fun f -> print_expr := true; pretty_print := true; pretty_print_output := (open_out (f ^ "_pretty.xq")); pretty_print_formatter := formatter_of_out_channel !pretty_print_output) , " Pretty printing the query to file 'filename_pretty.xq'\n\n // General configuration\n";

  (* General configuration *)
  "-internal-encoding", Arg.String (fun s -> Encoding.set_internal_encoding (Glx_pxp.rep_encoding_of_encoding(Glx_pxp.encoding_of_string s))), " Set the internal encoding representation";
  "-output-encoding",   Arg.String (fun s -> Encoding.set_output_encoding (Glx_pxp.encoding_of_string s)), " Set the outputx encoding representation\n\n // Input context configuration\n";

  (* Input context configuration *)
  "-base-uri", Arg.String (fun s -> AnyURI.set_base_uri s), " Sets the default base URI in the context";
  "-var",      Arg.String (fun vf -> let (v, f) = Gmisc.split_left_on_char vf '=' in global_vars := (v, f)::(!global_vars)), " var=filename, Binds the global variable var to document in filename"; 
  "-context-item",  Arg.String (fun f -> context_item := (Some f)), "Binds the context item to the document in filename"; 
  "-context",  Arg.String (fun f  -> context_file := Some f), " Load the context from an XQuery file\n\n // Data model configuration\n";

  (* Data model configuration *)
  "-xml-whitespace",    Arg.Set xml_whitespace, " Preserves whitespace in XML documents";
  "-xml-pic",           Arg.Set xml_pis_and_comments, " Preserves PI's and comments in XML documents";
  "-xquery-whitespace", Arg.Set xquery_whitespace, " Preserves whitespace in XQuery expressions\n\n // Serialization configuration\n";

  (* Serialization configuration *)
  "-serialize",         Arg.String   (fun ser -> serialize := serialization_kind_of_param ser), " Set serialization kind [wf, canonical, or xquery]\n\n // Processing phases configuration\n";

  "-serialize-id",      Arg.Set serialize_id, " Set serialization kind [wf or xquery]\n\n // Processing phases configuration\n";

  (* Processing phases configuration *)
  "-normalize", 	Arg.String   (fun onoff -> mapping := bool_of_onoff(onoff)), " Normalization phase";
  "-static",    	Arg.String   (fun onoff -> typing := bool_of_onoff(onoff)), " Static analysis phase";
  "-optim",     	Arg.String   (fun onoff -> optim := bool_of_onoff(onoff)), " Optimization phase (use with -static on)";
  "-dynamic",   	Arg.String   (fun onoff -> eval := bool_of_onoff(onoff); let b = bool_of_onoff(onoff) in if not(b) then print_xml := b), " Evaluation phase\n\n // Printing configuration\n";

  (* Printing configuration *)
  "-print-error-code",   Arg.Clear verbose_error, " Print only the error code instead of the full error message";
  "-print-xml",   Arg.String (fun onoff -> print_xml := bool_of_onoff(onoff)), " Print XML result";
  "-output-xml",  Arg.String (fun f -> init_output_refs f xml_output xml_formatter), " Output XML to result file" ;
  
  "-print-type",  Arg.String (fun onoff -> print_type := bool_of_onoff(onoff)), " Print type of expression";
  "-output-type", Arg.String (fun f -> init_output_refs f type_output type_formatter), " Output type to file.\n" ;
  
  "-print-expr",  Arg.String (fun onoff -> print_expr := bool_of_onoff(onoff)), " Print input expression";
  "-output-expr", Arg.String (fun f -> init_output_refs f expr_output expr_formatter), " Output expr to file.\n" ;
  
  "-print-normalized-expr",  Arg.String   (fun onoff -> print_core_expr := bool_of_onoff(onoff)), " Print normalized expression";
  "-output-normalized-expr", Arg.String (fun f -> init_output_refs f core_expr_output core_expr_formatter), " Output normalized expression to file.\n" ;
  
  "-print-typed-expr",       Arg.String   (fun onoff -> print_typed_expr := bool_of_onoff(onoff)), " Print typed expression";
  "-output-typed-expr",      Arg.String (fun f -> init_output_refs f typed_expr_output typed_expr_formatter), " Output typed expression to file.\n" ;
  
  "-print-optimized-expr",   Arg.String   (fun onoff -> print_optimized_expr := bool_of_onoff(onoff)), " Print optimized expression";
  "-output-optimized-expr",  Arg.String (fun f -> init_output_refs f optimized_expr_output optimized_expr_formatter), " Output optimized expression to file.\n" ;
  
  "-print-type-annotations", Arg.String  (fun onoff -> print_type_annotations := bool_of_onoff(onoff)), " Print type annotations of expressions";
  
  "-output-all", Arg.String (fun f -> output_all := true;
    init_output_refs f xml_output xml_formatter; 
    type_formatter := !xml_formatter;
    expr_formatter := !xml_formatter;
    core_expr_formatter := !xml_formatter;
    typed_expr_formatter := !xml_formatter;
    optimized_expr_formatter := !xml_formatter;), " Output everything to file." ;

  "-verbose",    Arg.String (fun onoff -> verbose := bool_of_onoff(onoff)), " Emit descriptive headers in output" ;
  
  "-projection", Arg.String (fun onoff -> projection := bool_of_onoff(onoff)), " Projection of documents";
  "-optimized-projection", Arg.String (fun onoff -> optimized_projection := bool_of_onoff(onoff)), " Optimized projection of documents";
  
  "-print-projection", Arg.String (fun onoff -> print_projection := bool_of_onoff(onoff)), " Prints the projection paths";
  "-output-projection", Arg.String (fun f -> init_output_refs f projection_output projection_formatter), " Output the projections paths to a file.\n" ;
  
  "-print-projected-file", Arg.String (fun onoff -> print_projected_file := bool_of_onoff(onoff)), " Prints back the input document after projection";
  "-output-projected-file", Arg.String (fun f -> init_output_refs f projected_file_output projected_file_formatter), " Output the input document after projection into file.\n" ;

]
    
    (fun arg -> args := arg :: !args) usage_msg;
  match !args with
  | [] -> failwith "Input file not specified"
  | fnames ->
      List.rev fnames

let init_all () =
  begin
    check_config();
    let pc = Galapi.default_processing_context() in
    let mod_ctxt = Galapi.load_standard_library pc in
    mod_ctxt
  end
	
let process_main_module mod_file =
  let (use_case, truncated_name) =
    try
      let mid = String.index mod_file '_' in
      let prefix = String.sub mod_file 0 mid in
      let suffix = String.sub mod_file (mid+1) ((String.length mod_file) - (mid+1)) in
      let mid2 = String.index suffix '.' in
      let prefix2 = String.sub suffix 0 mid2 in
      match prefix2 with
      | "usecase" -> (true, prefix)
      | "usecase_pretty" -> (true, prefix)
      | _ -> (false, mod_file)
    with
    | _ ->
	(false, mod_file)
  in
  if (!verbose) then 
    if (use_case)
    then
      begin
	printf "---------------------------------------\n";
	printf "Processing use case: %s\n" truncated_name;
	printf "---------------------------------------\n";
	printf "%s" !separator;
	flush stdout;
      end
    else
      begin
	printf "---------------------------------------\n";
	printf "Processing file: %s\n" truncated_name;
	printf "---------------------------------------\n";
	printf "%s" !separator;
	flush stdout;
      end; 

  try
    let mod_ctxt = init_all () in
    let mod_ctxt' = 
      match !context_file with
      | None ->
	  mod_ctxt 
      | Some f ->
	  Galapi.import_library_module mod_ctxt f 
    in 
    let ctxt_item = 
      match !context_item with 
      |	None -> None
      |	Some f -> Some (List.hd (Galapi.load_document f))
    in

    (* Load main module *)
    let (mod_ctxt'', stmts) = Galapi.import_main_module mod_ctxt' mod_file in

    (* Load documents bound to global variables *)
    let var_vals = List.map (fun (v, f) -> (v, Galapi.load_document f)) !global_vars in

    (* Build external context of global variables and values *)
    let ext_ctxt = Galapi.build_external_context ctxt_item var_vals in
    let mod_ctxt''' = Galapi.add_external_context mod_ctxt'' ext_ctxt in

    (* Evaluate all global variables *)
    let mod_ctxt'''' = Galapi.eval_global_variables mod_ctxt''' in 

    (* Evaluate all statements in module *)
    List.iter 
      (fun stmt -> 
	let opt_res = Galapi.eval_statement mod_ctxt'''' stmt;
	in
	match opt_res with
	| None -> ()
	| Some res -> Galapi.serialize_to_formatter (!xml_formatter) res
      ) stmts 
  with
  | e ->
      begin
	eprintf_error "  " e;
	fprintf err_formatter "@."
      end

let close_channel_ref chan = 
  if (!chan != stdout) then
    begin
      close_out (!chan);
      chan := stdout;
    end

let main () =
  Conf.serialize := Serialize_As_XQuery;
  let module_files = process_args() in
  Monitor.init_monitor ();
  List.iter process_main_module module_files;
  close_channel_ref(glx_stderr); 
  close_channel_ref(xml_output); 
  close_channel_ref(projection_output); 
  close_channel_ref(projected_file_output); 
  if not(!output_all) then
    begin
      close_channel_ref(expr_output); 
      close_channel_ref(type_output); 
      close_channel_ref(core_expr_output); 
      close_channel_ref(typed_expr_output); 
      close_channel_ref(optimized_expr_output); 
    end;
  if (!pretty_print) then 
    begin
      fprintf (!pretty_print_formatter) "@.";
      close_out !pretty_print_output;
      pretty_print_output := stdout;
    end

let exec func arg =
  try
    try
      func arg;
      Monitor.close_monitor ();
      if !monitor then
	close_channel_ref(monitor_output)
    with
    | Fn_error.Xquery_error msg ->
	Fn_error.downgrade_error msg
  with
  | e ->
      begin
	Monitor.close_monitor ();
	if !monitor then
	  close_channel_ref(monitor_output);
	eprintf_error "  " e;
	fprintf err_formatter "@."
      end

let _ =
  exec main()

