(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: commontype.ml,v 1.4 2003/04/16 19:52:47 mff Exp $ *)

open Error
open Occurrence

open Sym
open Namespace

open Xquery_ast
open Xquery_astutil

open Schema
open Schemautil

open Static_context

(*********************)
(* Typeswitch typing *)
(*********************)

(* Note:
     Local exception that is raised for empty small types.
   - J&M *)

exception Empty_Model 

(* Compute small type *)
(* small_type_unit model datatype: 

   model is type of expression
   datatype occurs in typeswitch case

   (called common-primes in FS Doc)   
   restricted form of intersection between model and datatype
*)
let small_type_unit stat_context m1 dtk2 =
  let schema = Norm_context.schema_from_norm_context(norm_context_from_stat_context stat_context) in
  match (m1.pmodel_desc, dtk2) with

  | (TEmpty, DTEmpty) -> 
      m1

  | (_, DTEmpty) -> 
      raise Empty_Model

    (* item wildcard *)
  | (_, DTItem) ->
      m1
  
    (* text node *)
  | (TText, DTText) ->
    m1 

  | (_, DTText) ->
      raise Empty_Model

    (* node wildcard *)
  | (TBase _, DTNode) ->
      raise Empty_Model

  | (_, DTNode) ->
      m1

    (* attribute wildcard *)
  | (TAttrRef _, DTAttrRef None) ->
      m1
  | (TAttrLocal _, DTAttrRef None) ->
      m1
  | (_, DTAttrRef None) ->
      raise Empty_Model

    (* element wildcard *)
  | (TElemRef _, DTElemRef None) ->
      m1
  | (TElemLocal _, DTElemRef None) ->
      m1
  | (_, DTElemRef None) ->
      raise Empty_Model

    (* numeric types *)
  | (TBase st1, DTNumeric) ->
      if (Typeutil.is_anynumeric schema m1)
      then
	m1
      else
	raise Empty_Model
  | (_, DTNumeric) ->
      raise Empty_Model


    (* atomic types *)
  | (TBase _, DTAtomic) ->
      m1
  | (_, DTAtomic) ->
      raise Empty_Model

    (* simple types *)
  | (TBase st1, DTBase st2) ->
      if (Datatypes.atomic_type_subsumes st1 st2)
      then
	m1
      else
	raise Empty_Model
  | (_, DTBase _) ->
      raise Empty_Model

    (* document of element *)
  | (TDocument m1', DTDocument None) ->
      m1

  | (TDocument m1', DTDocument (Some elem_sym2)) ->
      let m2' = fmkmodel (TElemRef elem_sym2) Finfo.bogus in
      if (Typeutil.is_subtype_of schema m1' m2')
      then
	m1
      else
	fmkmodel (TDocument m2') Finfo.bogus

  | (_, DTDocument _) ->
      raise Empty_Model

    (* attribute refs *)
  | (TAttrRef (RAttrSym attr_sym1), DTAttrRef (Some (RAttrSym attr_sym2))) ->
      if (rattr_subtag attr_sym1 attr_sym2)
      then
	let m2 = fmkmodel (TAttrRef (RAttrSym attr_sym2)) Finfo.bogus in
	if (Typeutil.is_subtype_of schema m1 m2)
	then
	  m1
	else
	  m2
      else
	raise Empty_Model
  (* Note:
       It's not clear what the following case would be under by-name
       subtyping.
     - M&J *)
  | (TAttrLocal (RAttrSym attr_sym1,_), DTAttrRef (Some (RAttrSym attr_sym2))) ->
      if (rattr_subtag attr_sym1 attr_sym2)
      then
	let m2 = fmkmodel (TAttrRef (RAttrSym attr_sym2)) Finfo.bogus in
	if (Typeutil.is_subtype_of schema m1 m2)
	then
	  m1
	else
	  m2
      else
	raise Empty_Model
  | (_, DTAttrRef _) ->
      raise Empty_Model

    (* element refs *)
  | (TElemRef (RElemSym elem_sym1), DTElemRef (Some (RElemSym elem_sym2))) ->
      if (relem_subtag elem_sym1 elem_sym2)
      then
	let m2 = fmkmodel (TElemRef (RElemSym elem_sym2)) Finfo.bogus in
	if (Typeutil.is_subtype_of schema m1 m2)
	then
	  m1
	else
	  m2
      else
	raise Empty_Model
  (* Note:
       It's not clear what the following case would be under by-name
       subtyping.
     - M&J *)
  | (TElemLocal (RElemSym elem_sym1,_,_), DTElemRef (Some (RElemSym elem_sym2))) ->
      if (relem_subtag elem_sym1 elem_sym2)
      then
	let m2 = fmkmodel (TElemRef (RElemSym elem_sym2)) Finfo.bogus in
	if (Typeutil.is_subtype_of schema m1 m2)
	then
	  m1
	else
	  m2
      else
	raise Empty_Model
  | (_, DTElemRef _) ->
      raise Empty_Model

    (* type ref *)
  | (_,DTTypeRef _) ->

      raise (Query (Prototype "typing: Type references not supported in SequenceType"))

    (* document of type *)
  | (TDocument m1', DTDocumentOfType group_sym2) ->
    raise (Query (Internal_Error "document of type not supported in SequenceType"))

  (* See remark labeled **J&M** below *)
  (* HACK! we do not care about the symbol here, but we will in the return type three lines down... *)
  (*
      let elem_sym = 
	begin
	  match m1'.pmodel_desc with
	  | TElemRef elem_sym -> elem_sym 
	  | TElemLocal(elem_sym, _, _) -> elem_sym
	  | _ -> raise (Query (Internal_Error "document type must refer to element"))
	end in
      let gm2 = fmkmodel (TGroupRef group_sym2) Finfo.bogus in
      let m2' =
	fmkmodel (TElemLocal (elem_sym,
			      None,
			      fmktyped_model (TAnonymousType gm2) Finfo.bogus)) Finfo.bogus in
      if (Typeutil.is_subtype_of tr m1' m2')
      then
	m1
      else
	fmkmodel (TDocument m2') Finfo.bogus
  *)

  | (_, DTDocumentOfType _) ->
      raise Empty_Model

    (* element of type *)
  | (TElemRef (RElemSym elem_sym1), DTElemOfType group_sym2) ->
    raise (Query (Prototype "typing: element of type not supported in SequenceType"))

 (* **J&M**: the following typing is correct, however typeswitch evaluation does not yet support 
    element of type.  It trivially compares element tag names and does not recursively 
    check sub-structure.  Thus, we raise a type error here because evaluation will fail. 
    If we support full structural subtyping, then the typing below should be uncommented. 
  *) 

 (* HACK! we do not care about the symbol here, but we will in the return type three lines down... *)
 (*   let gm2 = fmkmodel (TGroupRef group_sym2) Finfo.bogus in
      let m2 =
	fmkmodel (TElemLocal (RElemSym elem_sym1,
			      None,
			      fmktyped_model (TAnonymousType gm2) Finfo.bogus)) Finfo.bogus in
      if (Typeutil.is_subtype_of tr m1 m2)
      then
	m1
      else
	m2
 *)

  (* Note:
       It's not clear what the following case would be under by-name
       subtyping.
     - M&J *)
  | (TElemLocal ((RElemSym elem_sym1),_,_), DTElemOfType group_sym2) ->
      raise (Query (Prototype "typing: element of type not supported in SequenceType"))

 (* See comment labeled **J&M** above *)
 (* HACK! we do not care about the symbol here, but we will in the return type three lines down... *)
 (*   let gm2 = fmkmodel (TGroupRef group_sym2) Finfo.bogus in
      let m2 =
	fmkmodel (TElemLocal (RElemSym elem_sym1,
			      None,
			      fmktyped_model (TAnonymousType gm2) Finfo.bogus)) Finfo.bogus in
      if (Typeutil.is_subtype_of tr m1 m2)
      then
	m1
      else
	m2
 *)

  | (_, DTElemOfType _) ->
      raise (Query (Prototype "typing: element of type not supported in SequenceType")) 
 (* See remark **J&M** above *)
 (*   raise Empty_Model *)


    (* element of type *)
  | (TElemRef (RElemSym elem_sym1), DTElemOfBase st2) ->
      raise (Query (Prototype "typing: element of base type not supported in SequenceType")) 
 
 (* See remark **J&M** above *)
 (* HACK! we do not care about the symbol here, but we will in the return type three lines down... *)
(*
      let gm2 = fmkmodel (TBase st2) Finfo.bogus in
      let m2 =
	fmkmodel (TElemLocal (RElemSym elem_sym1,
			      None,
			      fmktyped_model (TAnonymousType gm2) Finfo.bogus)) Finfo.bogus in
      if (Typeutil.is_subtype_of tr m1 m2)
      then
	m1
      else
	m2
*)


  | (TElemLocal ((RElemSym elem_sym1),_,_), DTElemOfBase st2) ->
      raise (Query (Prototype "typing: element of base type not supported in SequenceType")) 

 (* See remark **J&M** above *)
 (* HACK! we do not care about the symbol here, but we will in the return type three lines down... *)
(*    let gm2 = fmkmodel (TBase st2) Finfo.bogus in
      let m2 =
	fmkmodel (TElemLocal (RElemSym elem_sym1,
			      None,
			      fmktyped_model (TAnonymousType gm2) Finfo.bogus)) Finfo.bogus in
      if (Typeutil.is_subtype_of schema m1 m2)
      then
	m1
      else
	m2
*)

  | (_, DTElemOfBase _) ->
      raise (Query (Prototype "typing: element of base type not supported in SequenceType")) 
 (* See remark **J&M** above *)
 (*    raise Empty_Model  *)

let rec process_small_type stat_context ml1 dtk2 =
  match ml1 with
  | [] -> []
  | m1 :: ml1' ->
      let more = process_small_type stat_context ml1' dtk2 in
      try
	(small_type_unit stat_context m1 dtk2) :: more
      with
      | Empty_Model ->
	  more

(* Call to the intersect type function *)

let small_type stat_context (p1,m1,n1) (dtk2,m2,n2) =
  (* J&M: This is a mess... The next two cases are symmetic tests for
     emptiness on the input model (p1, m1, n1) and datatype DT, which
     is not handled by small_type.  
     if () <: P and m2 = 0 then () 
     else if () <: DT and m1 = 0 then ()
     else ... as before ...
  *)
  if (List.hd p1).pmodel_desc = TEmpty then
    if (equal m2 (occurs 0)) then
      fmkmodel (TEmpty) Finfo.bogus
    else fmkmodel (TNone) Finfo.bogus 
  else if (dtk2 = DTEmpty) then 
    if (equal m1 (occurs 0)) then
      fmkmodel (TEmpty) Finfo.bogus
    else fmkmodel (TNone) Finfo.bogus 
  else
    let small_unit_type_result_list = process_small_type stat_context p1 dtk2 in
  (* J&M: If there is no non-empty type in small_unit_type_result_list, then 
     return empty choice type *)
    if (small_unit_type_result_list = []) then 
      fmkmodel (TNone) Finfo.bogus 

    else 
      let small_unit_type_result = union_types small_unit_type_result_list in
      let m_result = ub_max m1 m2
      and n_result = ub_min n1 n2 in
      if (equal n_result (occurs 0)) then 
	fmkmodel (TEmpty) Finfo.bogus
      else
	fmkmodel (TBound (small_unit_type_result, m_result, n_result)) Finfo.bogus

