(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: tpath.ml,v 1.6 2003/04/16 19:52:47 mff Exp $ *)

(* This module implements static typing for XPath steps, axis and node
   tests *)

open Error
open Occurrence

open Sym
open Namespace

open Datatypes
open Xquery_ast
open Xquery_astutil

open Schema
open Schemautil
open Simplification

open Format
open Print


(*****************************************)
(* Auxiliary functions on content models *)
(*****************************************)

(* Content of a type *)

let content_type schema m =
  match m.pmodel_desc with 
  | TElemLocal (l,_,t'') ->
      begin
	match t''.ptyped_model_desc with
	| TAnonymousType m -> m
      end
  | TAttrLocal(l,t'') ->
      begin
	match t''.ptyped_model_desc with
	| TAnonymousType m -> m
      end
  | TBase _ ->
      fmkmodel TEmpty Finfo.bogus
  | TText ->
      fmkmodel TEmpty Finfo.bogus
  | TDocument m ->
      m
  | _ ->
      raise (Query (Internal_Error ("Content type accessor not supported on type:\n"
				    ^ (Print.bprintf_model "" m))))

(* Auxiliary operations on content models *)

let only_non_atomic_model m =
  match m.pmodel_desc with
  | TBase _ -> fmkmodel TNone Finfo.bogus
  | _ -> m

let extract_node_model s m = 
  try
    apply_to_unit_types s only_non_atomic_model m
  with
  | _ ->
      raise (Query (Internal_Error ("Error processing content model: " ^
				    (Print.bprintf_model "" m))))

let only_atomic_model m =
  match m.pmodel_desc with
  | TBase _ -> m
  | _ -> fmkmodel TNone Finfo.bogus

let extract_atomic_model s m = 
  try
    apply_to_unit_types s only_atomic_model m
  with
  | _ ->
      raise (Query (Internal_Error ("Error processing content model: " ^
				    (Print.bprintf_model "" m))))

(* Content model without attributes *)

let remove_attributes schema m =
  simplify_ty schema (extract_content_model schema m)

let remove_atomic_types schema m =
  simplify_ty schema (extract_node_model schema m)

let remove_non_atomic_types schema m =
  simplify_ty schema (extract_atomic_model schema m)

(*******************)
(* Typing for axis *)
(*******************)

(* Child axis *)

(* TO BE FIXED - Jerome
let child_axis_type schema m =
  let content_model = content_type schema m in
  
*)

let clean_atomic_types schema m =
  let turn_atomic_to_text t =
    match t.pmodel_desc with
    | TBase _ ->
	fmkmodel TText Finfo.bogus
    | _ ->
	t
  in
  apply_to_unit_types schema turn_atomic_to_text m

let child_unit_type schema ut =
  match ut.pmodel_desc with 
  | TElemLocal (l,_,t'') ->
      begin
	match t''.ptyped_model_desc with
	| TAnonymousType m ->
	    clean_atomic_types schema m
      end
  | TAttrLocal(l,t'') ->
      fmkmodel TEmpty Finfo.bogus
  | TBase _ ->
      fmkmodel TEmpty Finfo.bogus
  | TText ->
      fmkmodel TEmpty Finfo.bogus
  | TDocument m ->
      m
  | _ ->
      raise (Query (Static_Type_Error (ut.pmodel_loc, ("Accessor children not supported on type:\n"
						^ (Print.bprintf_model "" ut)))))

(* Content of a prime type *)

let child_prime_type schema pt =
  apply_to_unit_types schema (child_unit_type schema) pt


(* J&M: Not clear whether we should return TNone or TEmpty here. *)

let select_text_or_element t =
  match t.pmodel_desc with
  | TText ->
      t
  | TElemLocal _ ->
      t
  | TBase _ ->
      raise (Query (Internal_Error "Should not contain an atomic type when typing an XPath axis"))
  | _ ->
      fmkmodel TEmpty Finfo.bogus

(* Main typing function for axis *)

let eval_axis_type stat_ctxt a node_type =
  let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
  match a with
  | Ancestor ->
      Typeutil.anyancestor_model
  | Ancestor_or_self ->
      Typeutil.anyancestor_model
  | Attribute ->
      let content = child_prime_type schema node_type in
      let select_attribute t =
	match t.pmodel_desc with
	| TAttrLocal _ -> t
	| _ -> fmkmodel TEmpty Finfo.bogus
      in
      apply_to_unit_types schema select_attribute content
  | Child ->
      if (!Conf.debugging) then Print.printf_model "child " node_type;
      let content = child_prime_type schema node_type in
      let c =
	apply_to_unit_types schema select_text_or_element content
      in
      if (!Conf.debugging) then Print.printf_model " =\n" c;
      c
  | Descendant ->
      let child_type =
	if (!Conf.debugging) then Print.printf_model "child in descendant " node_type;
	let content = child_prime_type schema node_type in
	let c = apply_to_unit_types schema select_text_or_element content in
	if (!Conf.debugging) then Print.printf_model " =\n" c;
	c
      in
      if (!Conf.debugging) then Print.printf_model "recfactor " child_type;
      let f = Typeutil.recfactor schema child_type in 
      if (!Conf.debugging) then Print.printf_model " =\n  " f;
      f
  | Descendant_or_self ->
      if (!Conf.debugging) then Print.printf_model "recfactor " node_type;
      let f = Typeutil.recfactor schema node_type in 
      if (!Conf.debugging) then Print.printf_model " =\n  " f;
      f
  | Namespace ->
      raise (Query (Prototype "Typing for namespace axis not implemented"))
  | Parent ->
      Typeutil.anyparent_model
  | Self ->
      node_type


(*************************)
(* Typing for node tests *)
(*************************)

let eval_node_test_type stat_ctxt a nt forest_type =
  let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
  match nt with
  | CPNameTest l' ->
      let (qn, rsym) = 
	match a with 
	| Attribute -> (l', RUnitAttrSymbol (rattr_symbol (NSDefaultElementPrefix, l')))
	| Namespace -> raise (Query (Prototype "In typing: name tests on namespaces not yet supported"))
	| _ -> (l', RUnitElemSymbol (relem_symbol (NSDefaultElementPrefix, l')))
      in
      begin
	match qn with
	| (NSWildcardUri,"*") ->
	    let select_principal t =
	      match Datamodel.principal_node_kind a with
	      |	Datamodel.ElementNodeKind ->
		  begin
		    match t.pmodel_desc with
		    | TElemLocal _ -> t
		    | _ -> fmkmodel TEmpty Finfo.bogus
		  end
	      |	Datamodel.AttributeNodeKind ->
		  begin
		    match t.pmodel_desc with
		    | TAttrLocal _ -> t
		    | _ -> fmkmodel TEmpty Finfo.bogus
		  end
	      |	Datamodel.NamespaceNodeKind ->
		  raise (Query (Prototype "namespace prefixes not yet supported"))
	      |	_ ->
		  raise (Query (Internal_Error "Should not project along this axis"))
	    in
	    apply_to_unit_types schema select_principal forest_type
	| _ ->
	    let select_principal t =
	      let fi = t.pmodel_loc in
	      match Datamodel.principal_node_kind a with
	      |	Datamodel.ElementNodeKind ->
		  begin
		    match t.pmodel_desc with
		    | TElemLocal (RElemSym l,d,tm) ->
			if subtag rsym (RUnitElemSymbol l)
			then 
			  begin
			    let rsym' = RElemSym (relem_symbol_of_unit_symbol rsym) in
			    fmkmodel (TElemLocal (rsym',d,tm)) fi
			  end
			else fmkmodel TEmpty Finfo.bogus
		    | _ -> fmkmodel TEmpty Finfo.bogus
		  end
	      |	Datamodel.AttributeNodeKind ->
		  begin
		    match t.pmodel_desc with
		    | TAttrLocal (RAttrSym l ,tm) ->
			if subtag rsym (RUnitAttrSymbol l)
			then 
			  begin
			    let rsym' = RAttrSym (rattr_symbol_of_unit_symbol rsym) in
			    fmkmodel (TAttrLocal (rsym',tm)) fi
			  end
			else fmkmodel TEmpty Finfo.bogus
		    | _ -> fmkmodel TEmpty Finfo.bogus
		  end
	      |	Datamodel.NamespaceNodeKind ->
		  raise (Query (Prototype "namespace prefixes not yet supported"))
	      |	_ ->
		  raise (Query (Internal_Error "Should not project along this axis"))
	    in
	    apply_to_unit_types schema select_principal forest_type
(*	| (Some prefix,_) ->
	    raise (Query (Prototype "namespace prefixes not yet supported"))*)
      end
  | CPNodeKindTest node_kind ->
      begin
	match node_kind with
	| PIKind pik ->
	    fmkmodel TEmpty Finfo.bogus
	| CommentKind ->
	    fmkmodel TEmpty Finfo.bogus
	| TextKind ->
	    begin
	    try
	      let select_base t =
		match t.pmodel_desc with
		| TBase _ -> raise (Query(Internal_Error ("Content type accessor not supported on type:\n"
				    ^ (Print.bprintf_model "" t))))
		| TText -> fmkmodel (TText) Finfo.bogus
		| _ -> fmkmodel TEmpty Finfo.bogus
	      in
	      apply_to_unit_types schema select_base forest_type
	    with
	    | Query(Internal_Error "") -> fmkmodel (TText) Finfo.bogus
	    end
	| AnyKind ->
	    forest_type
      end

