(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: typing.ml,v 1.42 2003/05/02 02:09:41 mff Exp $ *)

open Error
open Occurrence

open Sym
open Namespace

open Datatypes
open Xquery_ast
open Xquery_astutil

open Schema
open Schemautil
open Simplification

open Typeutil

open Format
open Print

(***************)
(* Type errors *)
(***************)

let type_error fi ty e xqe = 
  match xqe with
  | None ->
      raise(Query (Type_Error (fi, ("Core expression "
				    ^ (bprintf_cexpr "" e)
				    ^ " has type "
				    ^ (bprintf_model "" ty)))))
  | (Some xq) ->
      raise(Query (Type_Error (fi, ("Expression  "
				    ^ (bprintf_expr "" xq)
				    ^ " has type "
				    ^ (bprintf_model "" ty)))))

  (* Type errors: 
     e : () and e is not trivially the () expression
     e : 0  and e is not trivially the error expression
   *)

(* Check for type errors after typing function declaration or top-level expression.  *)

let check_type e ty = 
  let fi = e.pcexpr_loc in
  let xqe = e.pcexpr_origin in
  if (is_empty_seq_type ty)
  then
    match e.pcexpr_desc with
    | CEEmpty -> ty
    | _ -> type_error fi ty e xqe 
  else
    if (is_empty_choice_type ty) then
      match e.pcexpr_desc with
      | CEError _ -> ty
      | _ -> type_error fi ty e xqe 
  else
    ty

let is_error_expr e =
  match e.pcexpr_desc with
  | CEError _ -> true
  | _ -> false

(* Check type declaration *)
let check_type_declaration schema fi odt t =
  match odt with
  | None -> t
  | Some dt -> 
      let m = Mapping.map_resolved_datatype_to_model dt in 
      if (Typeutil.is_subtype_of schema t m)
      then m  (* The result type should be declared type *)
      else
	raise (Query (Static_Type_Error (fi, ("Type declaration failed.  Type:\n\n"
				       ^ (Print.bprintf_model "    " t)
				       ^ "\n is not subtype of:\n\n"
				       ^ (Print.bprintf_datatype "    " dt)
				       ^ ".\n\n"))))

(* promotes_to checks that a type can be promoted to another type.

2.  statEnv |- Type1 can be promoted to Type2      
    statEnv |- Type2 can be promoted to Type3
    -----------------------------------------
    statEnv |- Type1 can be promoted to Type3
    
  (We don't implement the transitive rule, because can_be_promoted_to
   returns the entire set of types that an atomic type can 
   be promoted to)

*)

let promotes_to stat_ctxt (t1, (dt, t2)) =
  let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
  (* 
     1.        statEnv |- Type1 <: Type2
     -----------------------------------------
     statEnv |- Type1 can be promoted to Type2
  *)
  if (Typeutil.is_subtype_of schema t1 t2) then
    (Some t2)
  else
  (*
     3.
     statEnv |- prime(Type1) can be promoted to prime(Type2)
     statEnv |- quantifier(Type1) < quantifier(Type2)
     -------------------------------------------------------
     statEnv |- Type1 can be promoted to Type2

     4.
     statEnv |- Type1 can be promoted to Type'
     ...
     statEnv |- TypeN can be promoted to Type'
     ---------------------------------------------------
     statEnv |- (Type1 | ... | TypeN) can be promoted to Type'

     5.
     statEnv |- Type1 can be promoted to (Type1' | ... | TypeN')
               statEnv |- Type1 can be promoted to Typei'
                        1 <= i' <= N'
     ---------------------------------------------------
     statEnv |- Type1 can be promoted to (Type1' | ... | TypeN')
   *)
    let (pl1,m1,n1) = Typeutil.factor_with_units schema t1 in
    let (pl2,m2,n2) = Typeutil.factor_with_units schema t2 in
    let promote_itemtype p1 = 
      match (p1.pmodel_desc) with 
      | (TBase a1) -> 
	  let atl = Datatypes.can_be_promoted_to a1 in
	  let pl2' = List.map (fun p -> p.pmodel_desc) pl2 in
	  (List.exists (fun a -> List.mem (TBase a) pl2') atl)
      | _ -> false
    in
(*
(print_string("\n t1 "^(bprintf_model "" t1));
 print_string (" m1 "^(string_of_occurs m1)^" n1 "^(string_of_occurs n1));
 print_string("\n t2 "^(bprintf_model "" t2));
 print_string (" m2 "^(string_of_occurs m2)^" n2 "^(string_of_occurs n2));
 print_string("\n m2 <= m1 "^(string_of_bool (Occurrence.le m2 m1)));
 print_string("\n n1 <= n2 "^(string_of_bool (Occurrence.le n1 n2)));
);
*)
    if (List.for_all promote_itemtype pl1) &&
       (Occurrence.le m2 m1) &&
       (Occurrence.le n1 n2) then 
      (Some t2)
    else
      None

(* match_function_signature finds the first function signature such
   that the types of the functions arguments match the signature, after
   promotion.  

  statEnv |- QName => qname
  statEnv.funcType(qname) = { FuncDecl1, ..., FuncDeclm }
  FuncDecli = define function qname'(Type1, ..., Typen) as Type
  statEnv |- Expr1 : Type1'     Type1' <: Type1''     Type1'' can be promoted to Type1
  ...
  statEnv |- Exprn : Typen'     Typen' <: Typen''     Typen'' can be promoted to Typen
  ------------------------------------------------------------------------------------
  statEnv |- QName (Expr1, ..., Exprn) : Type

*)
let match_function_signature stat_ctxt fi fn targs signatures =
  let rec args_match_signature signatures = 
    match signatures with 
    | [] -> raise Not_found
    | (fn, (input_types, output_type))::sigs ->
	begin
	  let vts = List.combine targs input_types in
	  let targs' = Gmisc.unwrap_option_list(List.map (promotes_to stat_ctxt) vts) in 
	  if (List.length targs = List.length targs') then
	    (fn, (input_types, output_type))
	  else
	    args_match_signature sigs
	end
  in 
  try
    args_match_signature signatures 
  with
    Not_found -> 
      let input_types = String.concat "," (List.map (bprintf_model "") targs) in
      raise (Query (Static_Type_Error (fi, ("Arguments to function '"
						^ (Namespace.string_of_rqname fn) 
						^ "("^ input_types ^")' do not match function's signature.\n"))))

(* Top-level call for function type checking *)

let eval_type_fctn_appl stat_ctxt fi fn arguments_types =
  let norm_ctxt = (Static_context.norm_context_from_stat_context stat_ctxt) in
  let schema = Norm_context.schema_from_norm_context norm_ctxt in
  let arity = List.length arguments_types in
  let signatures = Overloaded.resolve_overloaded_function norm_ctxt (fn, arity) fi in
  (* Pick the first function whose signature matches the types of the function arguments, 
     after type promotion *)
  let (fn', (input_types, (output_datatype, output_type))) = match_function_signature stat_ctxt fi fn arguments_types signatures in
  let type_rule =
    try
      Fn_static.lookup_bltin_fctn_type_rule (fn',arity)
    with
    | _ ->
	(fun s fi x o -> output_type)
  in
  type_rule schema fi arguments_types output_type

(* match_function_signature_name looks up most specific function that
   implements argument function applied to arguments with given types.
*)
let match_function_signature_name stat_ctxt fi fn arguments_types =
  let norm_ctxt = (Static_context.norm_context_from_stat_context stat_ctxt) in
  let schema = Norm_context.schema_from_norm_context norm_ctxt in
  let arity = List.length arguments_types in
  let signatures = Overloaded.resolve_overloaded_function norm_ctxt (fn, arity) fi in
  (* Pick the first function whose signature matches the types of the function arguments, 
     after type promotion *)
  let (fn', (input_types, (output_datatype, output_type))) = match_function_signature stat_ctxt fi fn arguments_types signatures in
  fn'

let atomic_to_text m = 
  match m.pmodel_desc with
  | TBase _ -> fmkmodel TText m.pmodel_loc
  | _ -> m

(********************************)
(* main type-checking functions *)
(********************************)

(*********************************************)
(* Static  semantics of element construction *)
(*   See Xaeval.element_constructor for dynamic semantics  *)
(*********************************************)
let rec element_constructor_type stat_ctxt a e' fi = 
  let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
  let content_type = type_cexpr stat_ctxt e' in
  let element_type =
    fmkmodel (TElemLocal (a, None, mkanonymous_type content_type)) fi
  in
  if not(Typeutil.is_element schema element_type)
  then
    (* We conservatively raise a static type error when the content mixes values and nodes, 
       even though dynamic evaluation is more permissive. 
    *)
    raise (Query (Static_Type_Error (fi, ("Content of element "
					  ^ (Sym.string_of_elem_symbol a)
					  ^ " contains both atomic types and nodes."))))

(*  let (p, m, n) = Typeutil.factor_with_units schema element_type in 
    let p' = Schemautil.weakunion_types (List.map atomic_to_text p) 
    in fmkmodel(TBound(p', m, n)) fi *)
  else
    element_type


(*********************************************)
(* Static  semantics of element construction *)
(*   See Xaeval.element_constructor for dynamic semantics  *)
(*********************************************)
and attribute_constructor_type stat_ctxt a e' fi = 
  let m' = (type_cexpr stat_ctxt e') in
  let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
  if (Typeutil.is_anysimpletype schema m') then  
    fmkmodel (TAttrLocal (a, mkanonymous_type m')) fi
  else if (Typeutil.is_anynode_sequence schema m') then 
    fmkmodel (TAttrLocal (a, mkanonymous_type (fmkmodel (TBase ATUntyped) fi))) fi
  else
    (* We conservatively raise a static type error when attribute content mixes values and nodes, 
       even though dynamic evaluation is more permissive.
    *)
    raise (Query (Static_Type_Error (fi, ("Content of attribute "
					  ^ (Sym.string_of_attr_symbol a)
					  ^ " contains both atomic types and nodes"))))

(* general note: If we want to know if a predicate has type bool it's
   not possible to check this with "match predicate with
   TBase(...)..."  because predicate may contain the type variable for
   the base type...  That's the reason why we have to use the
   subtyping. This also holds for TLab. *)

and eval_type_aux stat_ctxt (e : cexpr) : model =
  let fi = e.pcexpr_loc in
  let xqe = e.pcexpr_origin in
  let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
    begin
      match e.pcexpr_desc with
      |	CEUnordered e' -> 
	  let input_type =  type_cexpr stat_ctxt e' in
	  
          (* See: wrap_factor_type for the rationale for this logic *)
	  if (empty_type schema input_type) then fmkmodel TEmpty fi
	  else if (none_type schema input_type) then fmkmodel TNone fi
	  else
	    let (p,m,n) = Typeutil.factor schema input_type in
	    fmkmodel (TBound(p, m, n)) fi

      | CESort (v1, e1, _, e2, _, _)  ->
	  begin
	    let m1 = type_cexpr stat_ctxt e1 in

	    let factor_fn stat_ctxt m1 v e2 = 
	      let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
	      let (p, m, n) = Typeutil.factor schema m1 in
	      let p' = simplify_ty schema p in
	      
                 (* should return p' min m max n *)
                 (* check e2's type; see whether the sort make sense *)
	      
                 (* bind v to $dot, return a new environment *)
	      let stat_ctxt2 = Static_context.add_var_to_static_context stat_ctxt (v, p') in
	      let t2 = type_cexpr stat_ctxt2 e2 in
	      if (Typeutil.is_sorttype schema t2)
	      then
		fmkmodel (TBound(p',m,n)) fi
	      else
                (* Mary -- shouldn't key type be comparable scalar type? *)
		raise (Query (Static_Type_Error(fi, "'sortby' ordering expression should be an optional item")))
	    in
	    wrap_factor_type stat_ctxt factor_fn m1 v1 e2
	  end
      | CEScalar bv ->
	  fmkmodel (TBase (atomic_type_of_atomic_value bv)) fi
      | CEDocument e ->
	  let m = type_cexpr stat_ctxt e in
	  fmkmodel (TDocument m) fi
      | CEText t ->
          (* This is a horrible hack: empty text nodes are never constructed, therefore they never contribute to a type.  They exist solely to delimit enclosed expressions. *) 
	  if (t = "") then fmkmodel TEmpty fi
	  else fmkmodel TText fi
      | CETextComputed e' ->
	  (* The type of text{ Expr } is always text, because the
	  dynamic semantics converts the value of Expr to a string. *)
	  let m = type_cexpr stat_ctxt e' in
	  fmkmodel TText fi
      | CEPI _ ->
	  fmkmodel TEmpty fi    (* Note: type inference for PI and comments yields empty for now *)
      | CEComment _ ->
	  fmkmodel TEmpty fi
      | CEVar (v) ->
	  Static_context.var_from_static_context fi stat_ctxt v
      | CEElem (a,e') ->
	  element_constructor_type stat_ctxt (RElemSym a) e' fi
      | CEAttr (a,e') ->
	  attribute_constructor_type stat_ctxt (RAttrSym a) e' fi
      | CEAnyElem (e1,e2) ->
	  let m1 = type_cexpr stat_ctxt e1 in
	  if (is_string_type m1)
	  then
	    element_constructor_type stat_ctxt (RElemSym Sym.anyrelem) e2 fi
	  else
	    raise (Query (Static_Type_Error(fi, ("Computed element constructor name of type "
					  ^ (Print.bprintf_model "" m1)
					  ^ " instead of type xsd:string")))) 
      | CEAnyAttr (e1,e2) ->
	  let m1 = type_cexpr stat_ctxt e1 in
	  if (is_string_type m1)
	  then
	    attribute_constructor_type stat_ctxt (RAttrSym Sym.anyrattr) e2 fi
	  else
	    raise (Query (Static_Type_Error(fi, ("Computed attribute constructor name of type "
					  ^ (Print.bprintf_model "" m1)
					  ^ " instead of type xsd:string")))) 
      | CESeq (e1,e2) ->
	  fmkmodel (TComma (type_cexpr stat_ctxt e1, type_cexpr stat_ctxt e2)) fi
      | CEEmpty ->
	  fmkmodel TEmpty fi
      | CEIf (e1,e2,e3) ->
	  begin
            let predicate_type = type_cexpr stat_ctxt e1 in
            if (is_bool_type predicate_type)
            then
	      fmkmodel (TBar(type_cexpr stat_ctxt e2, type_cexpr stat_ctxt e3)) fi
            else
	      raise (Query (Static_Type_Error(fi,("Predicate in 'if' expression is not a boolean"))))
	  end
      | CELet (odt,v,e1,e2) ->
	  let t1 = check_type_declaration schema fi odt (type_cexpr stat_ctxt e1) in
	  let stat_ctxt' = Static_context.add_var_to_static_context stat_ctxt (v,t1) in
	  type_cexpr stat_ctxt' e2
      | CEFor (odt, v, None, e1, e2) ->
	  let m1 = (type_cexpr stat_ctxt e1) in
	  let factor_fn stat_ctxt m1 v e2  =             
	    let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
	    let (p,m,n) = Typeutil.factor schema m1 in
	    let p' = check_type_declaration schema fi odt p in
	    let stat_ctxt' = Static_context.add_var_to_static_context stat_ctxt (v,p') in
	    let m2 = type_cexpr stat_ctxt' e2 in
	    fmkmodel (TBound(m2,m,n)) fi
	  in
	  wrap_factor_type stat_ctxt factor_fn m1 v e2
      | CEFor (odt, v, Some v', e1, e2) ->
	  let m1 = (type_cexpr stat_ctxt e1) in
	  let factor_fn stat_ctxt m1 v e2  =             
	    let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
	    let (p,m,n) = Typeutil.factor schema m1 in
	    let p' = check_type_declaration schema fi odt p in
	    let stat_ctxt' = Static_context.add_var_to_static_context stat_ctxt (v,p') in
	    let stat_ctxt'' = Static_context.add_var_to_static_context stat_ctxt' (v',Typeutil.integer_model) in
	    let m2 = type_cexpr stat_ctxt'' e2 in
	    fmkmodel (TBound(m2,m,n)) fi
	  in
	  wrap_factor_type stat_ctxt factor_fn m1 v e2

      | CEForTuple (v, e1, e2) ->
	  let m1 = (type_cexpr stat_ctxt e1) in
	  let factor_fn stat_ctxt m1 v e2  =             
	    let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
	    let (p,m,n) = Typeutil.factor schema m1 in
	    let stat_ctxt' = Static_context.add_var_to_static_context stat_ctxt (v,p) in
	    let m2 = type_cexpr stat_ctxt' e2 in
	    fmkmodel (TBound(m2,m,n)) fi
	  in
	  wrap_factor_type stat_ctxt factor_fn m1 v e2

      | CETypeswitch (e', pats) ->
	  let t0 = type_cexpr stat_ctxt e' in
	  (* Note: (Typeswitch Fix1)
	       Fixing the input type to be factored.
	     - Jerome
	   *)
	  if (none_type schema t0)
	  then
	    fmkmodel TNone fi
	  else
	    let (p0,m0,n0) = Typeutil.factor_with_units schema t0 in
	    let t1 = eval_type_match stat_ctxt fi (p0,m0,n0) [] pats in
	    let t2 = union_types t1 in
	   (* Note:
	        Why do we need simplification there?
	      - J&M *)
	    let t3 = simplify_ty schema t2 in
	    t3
      | CEApp (fname, arguments) ->
	  let arguments_types = List.map (type_cexpr stat_ctxt) arguments in
	  eval_type_fctn_appl stat_ctxt fi fname arguments_types
      | CEError e ->
	  let m = type_cexpr stat_ctxt e in 
	  if (Typeutil.is_sorttype schema m) then 
	    fmkmodel TNone fi
	  else raise (Query (Static_Type_Error (fi, ("fn:error function expects item?"))))

      | CETreat (e',m) ->
	  let t2 = type_cexpr stat_ctxt e' in
	  m
      | CEValidate (e',m) ->
	  let t2 = type_cexpr stat_ctxt e' in
	  m
      | CECast (e',m) ->
	  let m' = (type_cexpr stat_ctxt e') in
	  if (Typeutil.is_anyatomic_optional schema m')
	  then
	    m
	  else
	    raise (Query (Static_Type_Error (fi, ("Content of cast expression not an atomic type or empty sequence"))))
      	  (* Cast is unsafe !!! *)
      	  (* A Warning should be raised - JS *)
      | CECastable (e',m) ->
	  Typeutil.boolean_model
      | CESubtype (t1,t2) ->
	  fmkmodel (TBase ATBoolean) fi
      | CEIntersects (t1,t2) ->
	  fmkmodel (TBase ATBoolean) fi
      | CEForwardAxis (a,nt) ->
          let dot_type = Static_context.var_from_static_context fi stat_ctxt Xquery_astutil.dot in
	  let forest_type = Tpath.eval_axis_type stat_ctxt a dot_type in
	  Tpath.eval_node_test_type stat_ctxt a nt forest_type
      | CEReverseAxis (a,nt) ->
          let dot_type = Static_context.var_from_static_context fi stat_ctxt Xquery_astutil.dot in
	  let forest_type = Tpath.eval_axis_type stat_ctxt a dot_type in
	  Tpath.eval_node_test_type stat_ctxt a nt forest_type
      | CESome (odt,v, e1, e2)
      | CEEvery (odt, v, e1, e2) ->
	  let m1 = (type_cexpr stat_ctxt e1) in
	  let factor_fn stat_ctxt m1 v e2  =             
	    let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
	    let (p,m,n) = Typeutil.factor schema m1 in
	    let p' = check_type_declaration schema fi odt p in
	    let stat_ctxt' = Static_context.add_var_to_static_context stat_ctxt (v,p') in
	    let m2 = type_cexpr stat_ctxt' e2 in
	    if (is_bool_type m2) then
	      m2
	    else
	      raise (Query (Static_Type_Error(fi,("Predicate in quantified expression is not a boolean"))))
	  in
	  wrap_factor_type stat_ctxt factor_fn m1 v e2
      | CEKeyRef (kn, kv) ->
	  Typeutil.anyparent_model
      | CECreateTuple rne_list ->
	  let rnm_list =
	    List.map (fun (rname,e1) -> (rname,type_cexpr stat_ctxt e1)) rne_list
	  in
	  fmkmodel (TTuple rnm_list) fi
      | CEAccessTuple (e1,rname) ->
	  let m1 = (type_cexpr stat_ctxt e1) in
	  get_tuple_component rname m1 fi
    end
and get_tuple_component rname m fi = 
  match m.pmodel_desc with
  | TTuple [] -> fmkmodel TEmpty fi
  | TTuple rnm_list ->
      begin
	try
	  List.assoc rname rnm_list
	with
	| _ ->
	    raise (Query (Static_Type_Error(fi,("Tuple component "^(Namespace.string_of_uqname rname)^" not found in tuple"))))		    
      end
  | TBar (m1, m2) ->
      let m1' = get_tuple_component rname m1 fi in
      let m2' = get_tuple_component rname m2 fi in
      fmkmodel (TBar(m1', m2')) fi
  | _ ->	
      raise (Query (Static_Type_Error(fi,("Accessing tuple component over a type which is not a tuple"))))


and type_cexpr stat_ctxt (e : cexpr) : model =
  try
    let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
    let fi = e.pcexpr_loc in
    let t = eval_type_aux stat_ctxt e
    in
    if (!Conf.debugging) then begin (printf_cexpr "\n" e); (printf_model "before simplification: " t); printf "\n" end; 
    let restype = (simplify_ty schema t)
    in
    if (!Conf.debugging) then begin (printf_model "after simplification: " restype); printf "\n" end; 
    begin
      (*
      if (empty_type schema restype) && not(e.pcexpr_desc = CEEmpty) then
	printf "%s" ("Warning: Expression has empty type:\n"
		     ^ (bprintf_cexpr "" e));
      *)
      if (none_type schema restype) && not(is_error_expr e) then
	raise (Query (Static_Type_Error (fi, "Expression has none type: "
					 ^ (bprintf_cexpr "" e))));
      e.pcexpr_type <- (Some restype);
      restype
    end
  with
  | (Query (Static_Type_Error (fi, err))) ->
      begin
	let restype = fmkmodel TNone Finfo.bogus in
	e.pcexpr_type <- (Some restype);
	raise (Query (Static_Type_Error (fi, err)))
      end

(* if (is_empty_type stat_ctxt restype) && not(e.pcexpr_desc = CEEmpty) then
    printf "%s" ("Warning: Expression has empty type:\n"
		 ^ (bprintf_cexpr "" e))
   else (); 
*)

(* 

  wrap_factor_type - J&M

  There are several typing rules that factor their input type:
    for, sort, unordered, fn:distinct-values(), fn:data(), and 
    function calls when attempting promotion. 

  If the input type is empty or none, the bodies of these expressions
  need not be type checked.  We require them to be type checked,
  however, so that every core expression has a type annotation.

  wrap_factor_type type checks the bodies of these expressions when
  the input type is empty or none.  If the input type is not empty or
  none, it applies an expression-specific function for type checking
  the expression's body.

 *)
and wrap_factor_type stat_ctxt factor_fn m1 v e2 =
  let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
  if (empty_type schema m1)
  then
    let t2 = 
      let stat_ctxt' = Static_context.add_var_to_static_context stat_ctxt (v,fmkmodel TNone Finfo.bogus) in
      try 
	type_cexpr stat_ctxt' e2 
      with
	Query(Static_Type_Error _) -> fmkmodel TNone Finfo.bogus
    in fmkmodel TEmpty Finfo.bogus
  else if (none_type schema m1)
  then
    let t2 = 
      let stat_ctxt' = Static_context.add_var_to_static_context stat_ctxt (v,fmkmodel TNone Finfo.bogus) in
      try 
	type_cexpr stat_ctxt' e2 
      with
	Query(Static_Type_Error _) -> fmkmodel TNone Finfo.bogus
    in fmkmodel TNone Finfo.bogus
  else
    factor_fn stat_ctxt m1 v e2

and eval_type_match stat_ctxt fi (p0,m0,n0) tprev pats =
  let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
  match pats with
  | (p,ovn,e) :: pats' ->
      begin
	match p.ppattern_desc with
	| Default ->
	    let t0' = Typeutil.defactor schema (weakunion_types p0,m0,n0) in
	    (* Note:
	         Now, we need to make sure we type check the case
	         expression in all cases! I.e., call eval_type, which
	         also updates the AST with types.

                 This must be done even if we know statically
	         the case clause will not be evaluated.
	        - BUG FOUND BY GARGI!
	       *)
	    let stat_ctxt' =
	      match ovn with
	      |	None -> stat_ctxt
	      |	Some vname -> Static_context.add_var_to_static_context stat_ctxt (vname,t0') 
	    in

	    (* Mary & Jerome: type checking may fail on some clause.
               If this occurs, then the type of the case is None and
               it does not contribute to the result type. *)

	    let case_type = 
	      try 
		type_cexpr stat_ctxt' e 
	      with
		Query(Static_Type_Error _) -> fmkmodel TNone Finfo.bogus
	    in

	    (* Note: (Typeswitch Fix2)
	       Making sure to simplify the union of previous
	       types. This is enforcing determinism on the resulting
	       type, as required by the subtyping algorithm.
	       - Jerome
	     *)

		let t1' = simplify_ty schema (union_types (List.map Mapping.map_resolved_datatype_to_model tprev)) in

		let deftype = (Typeutil.is_subtype_of schema t0' t1') in
		if (deftype)
		then []
		else [case_type]
	      
	      
	| Case dt ->
	    let (dtk1, m1, n1) = Typeutil.factor_datatype dt in
            (* *)
	    let t' = Commontype.small_type stat_ctxt (p0,m0,n0) (dtk1,m1,n1) in
            (* Mary & Gargi -- 

               In preparation for optimization, we need to type the
               body of the case clause, even if its type will not
               contribute to the type of the typeswitch expression,
               but we cannot type the case-clause body when the "small
               type" is None as this may raise a type error.
               Therefore, we type the case-clause body
               "conservatively" by using the data type of the case
               clause.  We are not happy about this.
            *)

	    let stat_ctxt' = 
	      match ovn with 
	      |	None -> 
(*		  if (is_empty_choice_type t') 
		  then 	raise (Query (Type_Error (fi, "Cannot statically type check case clause with no variable, because 
                               this clause will never be evaluated.")))
		  else 
*)
		    stat_ctxt
	      |	Some vname -> 
		  if (is_empty_choice_type t') 
		  then Static_context.add_var_to_static_context stat_ctxt (vname, Mapping.map_resolved_datatype_to_model dt)
		  else 
		    Static_context.add_var_to_static_context stat_ctxt (vname,t') 
	    in
	    (* Mary & Jerome: type checking may fail on some clause.  If this occurs, then the
               type of the case is None and it does not contribute to the result type. *)
	    let case_type = 
	      try 
		type_cexpr stat_ctxt' e 
	      with
		Query(Static_Type_Error _) -> fmkmodel TNone Finfo.bogus
	    in
	    let othercases = eval_type_match stat_ctxt fi (p0,m0,n0) (dt :: tprev) pats' in
	    if (is_empty_choice_type t')
	    then othercases
	    else case_type :: othercases
      end
  | _ ->
      raise (Query (Internal_Error "No default in typeswitch"))

(*****************************************)
(* typing for the top-level declarations *)
(*****************************************)

(* adding global variables into the environment *)

let add_var_type_in_env stat_ctxt v =
  match v.pcvar_decl_desc with
  | (vn,ce) ->
      let m = type_cexpr stat_ctxt ce in 
      Static_context.add_var_to_static_context stat_ctxt (vn, m) 

(* typing for function declarations *)

let fctn_type_cexpr stat_ctxt f =
  let schema = Norm_context.schema_from_norm_context (Static_context.norm_context_from_stat_context stat_ctxt) in
  match f.pcfunction_def_desc with
  | (_, _, CEFunctionBltIn _) ->
      ()
  | (fname,(input_types,(output_datatype, output_model)), CEFunctionUser (vars,e)) -> 
      if (List.length vars) != (List.length input_types)
      then
	raise (Query (Static_Type_Error(f.pcfunction_loc,("number of argument variables doesn't correspond to number of types"))))
      else
	(* adds asumptions about variables for functions in the environment *)
	let vars_with_types = (List.combine vars input_types) in
        let stat_ctxt' = List.fold_right (fun (var,(dt,ty)) tenv -> Static_context.add_var_to_static_context tenv (var,ty)) vars_with_types stat_ctxt in
	(* then checks the type of the function's expression with the above asumptions *)
        let t' = type_cexpr stat_ctxt' e in
	(* finaly, checks that the resulting type is a subtype of the expected output type *)
        if (Typeutil.is_subtype_of schema t' output_model)
        then ()
        else
	  let t1' = simplify_ty schema t'
	  and t2' = simplify_ty schema output_model in
	  raise (Query (Static_Type_Error(f.pcfunction_loc,("checking function "
						     ^ (Namespace.string_of_rqname fname)
						     ^ ": body type:\n"
						     ^ (Print.bprintf_model "" t1')
						     ^ "not subtype of function's codomain:\n"
						     ^ (Print.bprintf_model "" t2')))))

let type_cupdate stat_ctxt cu =
  raise (Query (Prototype "Typing: Updates not supported yet"))

let type_cstatement stat_ctxt cs = 
  match cs.pcstatement_desc with
  | CSExpression ce ->
      type_cexpr stat_ctxt ce
  | CSUpdate cu ->
      type_cupdate stat_ctxt cu

(* typing for the whole toplevel declarations *)

let type_module stat_ctxt qm =
  (* adds asumptions about type of global variables in the environment *)
  let stat_ctxt' = List.fold_left add_var_type_in_env stat_ctxt qm.pcmodule_vars in
  (* checks types for functions *)
  List.iter (fctn_type_cexpr stat_ctxt') qm.pcmodule_functions;
  (* returns new type environment (with global variables) if everything works fine *)
  stat_ctxt'

