(***********************************************************************)
(*                                                                     *)
(*                                 GALAX                               *)
(*                            XML Query Engine                         *)
(*                                                                     *)
(*  Copyright 2000, Lucent Technologies.                               *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: querycgi.ml,v 1.20 2003/05/29 21:40:18 mff Exp $ *)

(* Galax modules *)

open Error
open Conf
open Print
open Xquery_ast
open Toputils         (* Note: contains most of what we need here - Jerome *)
open Galapi

(* Caml modules *)

open Format

(* demo modules *)

open Demo_conf

(* Additional global parameters *)

let showresult = ref true
let showtype = ref true

let html_setup () =
  setup 120 100;
  separator := "<HR>"

(* Check that configuration flags are compatible *)
let check_config () = 
  begin
    if !typing && not(!mapping) then
      raise (Query(Wrong_Args("Static typing requires normalization"))); 
    if !optim && not(!mapping) then 
      raise (Query(Wrong_Args("Optimization requires normalization"))); 
    if !projection && !optimized_projection then 
      raise (Query(Wrong_Args("Cannot apply projection and optimized projection both on at the same time"))); 
    if !eval && not(!mapping) then 
      raise (Query(Wrong_Args("Dynamic evaluation requires normalization"))); 
    if !print_type && not(!typing) then
      raise (Query(Wrong_Args("Cannot print type without static typing on"))); 
    if !print_core_expr && not(!mapping) then
      raise (Query(Wrong_Args("Cannot print normalized expression without normalization on"))); 
    if !print_optimized_expr && not(!optim) then
      raise (Query(Wrong_Args("Cannot pring optimized expresstion without optimization on"))); 
    if !print_typed_expr && not(!typing) then
      raise (Query(Wrong_Args("Cannot print typed expression without static typing on"))); 
    if !print_xml && not(!eval) then
      raise (Query(Wrong_Args("Cannot print xml result without dynamic evaluation on")))
  end
 
(* Extract CGI arguments *)

let process_args () =
  Cgi.parse_arguments Cgi.default_config;
  begin
    try
      Conf.debugging := bool_of_string (Cgi.argument_value "debug")
    with
    | _ -> ()
  end;
  begin
    try
      Conf.print_core_expr := bool_of_string (Cgi.argument_value "core-print")
    with
    | _ -> ()
  end;
  begin
    try
      Conf.typing := bool_of_string (Cgi.argument_value "typing")
    with
    | _ -> ()
  end;
  begin
    try
      Conf.print_type := bool_of_string (Cgi.argument_value "typing-print")
    with
    | _ -> ()
  end;
  begin
    try
      Conf.optim := bool_of_string (Cgi.argument_value "optim")
    with
    | _ -> ()
  end;
  begin
    try
      Conf.print_optimized_expr := bool_of_string (Cgi.argument_value "optim-print")
    with
    | _ -> ()
  end;
  begin
    try
      Conf.eval := bool_of_string (Cgi.argument_value "eval")
    with
    | _ -> ()
  end;
  begin
    try
      Conf.print_xml := bool_of_string (Cgi.argument_value "eval-print")
    with
    | _ -> ()
  end;
  check_config ();
  let query_string = (Cgi.argument_value "query")
  (* Load local context file *)
  and query_context = Gmisc.load_file (Cgi.argument_value "context") in
  (query_context,query_string)

(* Special XQuery type/eval/print function for HTML *)

let html_xquery_loop mod_ctxt t cs =
  let optv = eval_statement mod_ctxt cs in
  match optv with
  | None -> ()
  | Some v -> Galapi.serialize_to_formatter !Conf.xml_formatter v

let main() =
  html_setup ();
  demo_init "galax_demo_config";

  Conf.verbose := true; 
 
  (* Set up the output headers and footers *)
  Conf.expr_header := "\n<HR>\n<H2>XQuery Expression:</H2>\n<pre>\n";
  Conf.expr_footer := "</pre><br>\n"^(!separator);

  Conf.core_expr_header := "\n<H2>Normalized Expression (XQuery Core):</H2>\n<pre>\n";
  Conf.core_expr_footer := "</pre><br>\n"^(!separator);

  Conf.type_header := "<H2>Static Type Analysis:</H2>\n<pre>\n";
  Conf.type_footer := "</pre><br>\n"^(!separator);

  Conf.optimized_expr_header := "\n<H2>Optimized Normalized Expression (XQuery Core):</H2>\n<pre>\n";
  Conf.optimized_expr_footer := "</pre><br>\n"^(!separator);

  Conf.xml_header := "<H2>Dynamic Evaluation:</H2>\n<pre>\n";
  Conf.xml_footer := "</pre><br>\n"^(!separator);

  (* Sys.chdir !document_dir; *)

  let (context,queries) = process_args () in

  let full_text = (context ^ queries) in

  if (!Conf.debugging) then
    begin
      printf "Query text is:\n\n";
      printf "%s" context;
      flush stdout;
      begin
	let fi = open_out "/tmp/content.txt" in
	output_string fi "Query text is:\n\n";
	output_string fi context;
	close_out fi
      end
    end;

  begin
    let pc = Galapi.default_processing_context() in
    (* This function should be part of the processing context *)
    Conf.xml_charescape_fn := Netencoding.Html.encode_from_latin1;  
    let mod_ctxt = Galapi.load_standard_library pc in

    (* Load main module *)
    let (mod_ctxt', stmts) = Galapi.import_main_module_from_string mod_ctxt full_text in
    (* Evaluate global variables *)
    let mod_ctxt'' = Galapi.eval_global_variables mod_ctxt' in 

    List.iter (html_xquery_loop mod_ctxt'' full_text) stmts
  end;

  flush stdout

let beg_html =
  "<html><head><title>Galax</title><link href=\"../galax.css\" type=\"text/css\" rel=\"stylesheet\"></head>\n<body>\n<multicol cols=1>\n"

let end_html =
  "</multicol>\n\
</body></html>\n"

let exec func arg  =
  printf ("Content-type: text/html\n\n");
  printf "%s" beg_html;
  flush stdout;
  begin
   try
      func arg
    with
      e ->
        begin
          printf "<H2>An Error Was Raised:</H2>\n<pre>\n";
          printf "%s" (bprintf_error_safe "  " e);
          printf "\n</pre>\n%s" !separator
        end
  end;
  printf "%s" end_html;
  flush stdout

let _ =
  exec main ()

